﻿// Copyright Rayzzor Studios 2011 
// 
// This file is part of 42S.
// 
// 42S is free software: you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
// 
// 42S is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
// GNU General Public License for more details. 
// 
// You should have received a copy of the GNU General Public License 
// along with 42S. If not, see <http://www.gnu.org/licenses/>. 

using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Navigation;
using System.Windows.Shapes;
using Microsoft.Phone.Controls;
using Microsoft.Phone.Shell;

using System.IO;
using System.IO.IsolatedStorage;
using System.Runtime.Serialization;

using System.Text;
using System.Windows.Media.Imaging;
using Microsoft.Xna.Framework.GamerServices;


namespace WP7_42s
{
    public partial class App : Application
    {

        // Easy access to the root frame
        public PhoneApplicationFrame RootFrame { get; private set; }

        // Constructor
        public App()
        {
            // Global handler for uncaught exceptions. 
            // Note that exceptions thrown by ApplicationBarItem.Click will not get caught here.
            UnhandledException += Application_UnhandledException;

            // Standard Silverlight initialization
            InitializeComponent();

            // Phone-specific initialization
            InitializePhoneApplication();

#if DEBUG
//            Guide.SimulateTrialMode = true;
#endif
        }

        // Code to execute when the application is launching (eg, from Start)
        // This code will not execute when the application is reactivated
        private void Application_Launching(object sender, LaunchingEventArgs e)
        {
            trialMode = Guide.IsTrialMode;
            showSplash = true;
        }

        // Code to execute when the application is activated (brought to foreground)
        // This code will not execute when the application is first launched
        private void Application_Activated(object sender, ActivatedEventArgs e)
        {
            trialMode = Guide.IsTrialMode;
            showSplash = false;
        }

        // Code to execute when the application is deactivated (sent to background)
        // This code will not execute when the application is closing
        private void Application_Deactivated(object sender, DeactivatedEventArgs e)
        {
            core_quit();
        }

        // Code to execute when the application is closing (eg, user hit Back)
        // This code will not execute when the application is deactivated
        private void Application_Closing(object sender, ClosingEventArgs e)
        {
            core_quit();
        }


        public static void app_init()
        {
            rand = new Random(Environment.TickCount);
            read_key_map();

            for (int i = 0; i < progmenu_arg.Length; i++)
                progmenu_arg[i] = new App.arg_struct();

            for (int i = 0; i < progmenu_label.Length; i++)
                progmenu_label[i] = new string("".ToCharArray());

            for (int r = 0; r < 3; r++)
                for (int i = 0; i < 6; i++)
                    custommenu_label[r, i] = new string("".ToCharArray());
        }


        public static bool load_state(int ver)
        {
            bool err = false;

            IsolatedStorageFile isf = IsolatedStorageFile.GetUserStoreForApplication();
            IsolatedStorageFileStream isfs = null;

            try
            {
                isfs = new IsolatedStorageFileStream(App.STATE_FILE, FileMode.Open, isf);

                int magic = (int)SilverlightSerializer.Deserialize(isfs);
                if (magic != App.FREE42_MAGIC)
                    throw new Exception("Invalid state magic number");

                int version = (int)SilverlightSerializer.Deserialize(isfs);
                if (version != ver)
                    throw new Exception("Invalid state version number");

                core_settings.matrix_singularmatrix = (bool)SilverlightSerializer.Deserialize(isfs);
                core_settings.matrix_outofrange = (bool)SilverlightSerializer.Deserialize(isfs);
                core_settings.raw_text = (bool)SilverlightSerializer.Deserialize(isfs);
                core_settings.auto_repeat = (bool)SilverlightSerializer.Deserialize(isfs);
                core_settings.enable_ext_copan = (bool)SilverlightSerializer.Deserialize(isfs);
                core_settings.enable_ext_bigstack = (bool)SilverlightSerializer.Deserialize(isfs);
                core_settings.enable_ext_accel = (bool)SilverlightSerializer.Deserialize(isfs);
                core_settings.enable_ext_locat = (bool)SilverlightSerializer.Deserialize(isfs);
                core_settings.enable_ext_heading = (bool)SilverlightSerializer.Deserialize(isfs);
                core_settings.enable_ext_time = (bool)SilverlightSerializer.Deserialize(isfs);

                mode_clall = (bool)SilverlightSerializer.Deserialize(isfs);
                mode_command_entry = (bool)SilverlightSerializer.Deserialize(isfs);
                mode_number_entry = (bool)SilverlightSerializer.Deserialize(isfs);
                mode_alpha_entry = (bool)SilverlightSerializer.Deserialize(isfs);
                mode_shift = (bool)SilverlightSerializer.Deserialize(isfs);

                mode_appmenu = (int)SilverlightSerializer.Deserialize(isfs);
                mode_plainmenu = (int)SilverlightSerializer.Deserialize(isfs);
                mode_plainmenu_sticky = (bool)SilverlightSerializer.Deserialize(isfs);
                mode_transientmenu = (int)SilverlightSerializer.Deserialize(isfs);
                mode_alphamenu = (int)SilverlightSerializer.Deserialize(isfs);
                mode_commandmenu = (int)SilverlightSerializer.Deserialize(isfs);
                mode_running = (bool)SilverlightSerializer.Deserialize(isfs);
                mode_varmenu = (bool)SilverlightSerializer.Deserialize(isfs);
                mode_updown = (bool)SilverlightSerializer.Deserialize(isfs);
                mode_getkey = (bool)SilverlightSerializer.Deserialize(isfs);

                entered_number = (Phloat)SilverlightSerializer.Deserialize(isfs);
                entered_string_length = (int)SilverlightSerializer.Deserialize(isfs);
                entered_string = (char[])SilverlightSerializer.Deserialize(isfs);

                pending_command = (int)SilverlightSerializer.Deserialize(isfs);
                pending_command_arg = (arg_struct)SilverlightSerializer.Deserialize(isfs);
                xeq_invisible = (int)SilverlightSerializer.Deserialize(isfs);

                incomplete_command = (int)SilverlightSerializer.Deserialize(isfs);
                incomplete_ind = (int)SilverlightSerializer.Deserialize(isfs);
                incomplete_alpha = (int)SilverlightSerializer.Deserialize(isfs);
                incomplete_length = (int)SilverlightSerializer.Deserialize(isfs);
                incomplete_maxdigits = (int)SilverlightSerializer.Deserialize(isfs);
                incomplete_argtype = (int)SilverlightSerializer.Deserialize(isfs);
                incomplete_num = (int)SilverlightSerializer.Deserialize(isfs);
                incomplete_str = (char[])SilverlightSerializer.Deserialize(isfs);
                incomplete_saved_pc = (int)SilverlightSerializer.Deserialize(isfs);
                incomplete_saved_highlight_row = (int)SilverlightSerializer.Deserialize(isfs);

                cmdline = (char[])SilverlightSerializer.Deserialize(isfs);
                cmdline_length = (int)SilverlightSerializer.Deserialize(isfs);
                cmdline_row = (int)SilverlightSerializer.Deserialize(isfs);

                matedit_mode = (int)SilverlightSerializer.Deserialize(isfs);
                matedit_name = (char[])SilverlightSerializer.Deserialize(isfs);
                matedit_length = (int)SilverlightSerializer.Deserialize(isfs);

                matedit_x = (vartype)SilverlightSerializer.Deserialize(isfs);
                matedit_i = (int)SilverlightSerializer.Deserialize(isfs);
                matedit_j = (int)SilverlightSerializer.Deserialize(isfs);
                matedit_prev_appmenu = (int)SilverlightSerializer.Deserialize(isfs);

                input_name = (char[])SilverlightSerializer.Deserialize(isfs);
                input_length = (int)SilverlightSerializer.Deserialize(isfs);
                input_arg = (arg_struct)SilverlightSerializer.Deserialize(isfs);

                baseapp = (int)SilverlightSerializer.Deserialize(isfs);
                random_number = (Phloat)SilverlightSerializer.Deserialize(isfs);

                deferred_print = (int)SilverlightSerializer.Deserialize(isfs);

                keybuf_head = (int)SilverlightSerializer.Deserialize(isfs);
                keybuf_tail = (int)SilverlightSerializer.Deserialize(isfs);
                keybuf = (int[])SilverlightSerializer.Deserialize(isfs);



                //
                // Unpersist Display
                //

                is_dirty = false;

                catalogmenu_section = (int[])SilverlightSerializer.Deserialize(isfs);
                catalogmenu_rows = (int[])SilverlightSerializer.Deserialize(isfs);
                catalogmenu_row = (int[])SilverlightSerializer.Deserialize(isfs);

                catalogmenu_item = (int[,])SilverlightSerializer.Deserialize(isfs);
                custommenu_length = (int[,])SilverlightSerializer.Deserialize(isfs);

                custommenu_label = (string[,])SilverlightSerializer.Deserialize(isfs);

                progmenu_arg = (arg_struct[])SilverlightSerializer.Deserialize(isfs);

                progmenu_is_gto = (int[])SilverlightSerializer.Deserialize(isfs);
                progmenu_length = (int[])SilverlightSerializer.Deserialize(isfs);
                progmenu_label = (string[])SilverlightSerializer.Deserialize(isfs);

                display = (byte[])SilverlightSerializer.Deserialize(isfs);

                appmenu_exitcallback = (int)SilverlightSerializer.Deserialize(isfs);



                //
                // Unpersist Globals
                //

                array_count = 0;
                array_list_capacity = 0;
                array_list = null;

                varray_count = 0;
                varray_list_capacity = 0;
                varray_list = null;

                reg_x = (vartype)SilverlightSerializer.Deserialize(isfs);
                reg_y = (vartype)SilverlightSerializer.Deserialize(isfs);
                reg_z = (vartype)SilverlightSerializer.Deserialize(isfs);
                reg_t = (vartype)SilverlightSerializer.Deserialize(isfs);
                reg_lastx = (vartype)SilverlightSerializer.Deserialize(isfs);

                reg_alpha_length = (int)SilverlightSerializer.Deserialize(isfs);
                reg_alpha = (char[])SilverlightSerializer.Deserialize(isfs);

                mode_sigma_reg = (int)SilverlightSerializer.Deserialize(isfs);
                mode_goose = (int)SilverlightSerializer.Deserialize(isfs);
                mode_time_clktd = (bool)SilverlightSerializer.Deserialize(isfs);
                mode_time_clk24 = (bool)SilverlightSerializer.Deserialize(isfs);
                mode_time_dmy = (bool)SilverlightSerializer.Deserialize(isfs);

                flags = (flags_struct)SilverlightSerializer.Deserialize(isfs);

                vars_count = (int)SilverlightSerializer.Deserialize(isfs);
                vars = (var_struct[])SilverlightSerializer.Deserialize(isfs);

                prgms_count = (int)SilverlightSerializer.Deserialize(isfs);
                prgms = (prgm_struct[])SilverlightSerializer.Deserialize(isfs);

                current_prgm = (int)SilverlightSerializer.Deserialize(isfs);
                pc = (int)SilverlightSerializer.Deserialize(isfs);
                prgm_highlight_row = (int)SilverlightSerializer.Deserialize(isfs);

                varmenu_length = (int)SilverlightSerializer.Deserialize(isfs);
                varmenu = (char[])SilverlightSerializer.Deserialize(isfs);
                varmenu_rows = (int)SilverlightSerializer.Deserialize(isfs);
                varmenu_row = (int)SilverlightSerializer.Deserialize(isfs);
                varmenu_labellength = (int[])SilverlightSerializer.Deserialize(isfs);
                varmenu_labeltext = (char[][])SilverlightSerializer.Deserialize(isfs);
                varmenu_role = (int)SilverlightSerializer.Deserialize(isfs);

                rtn_sp = (int)SilverlightSerializer.Deserialize(isfs);
                rtn_prgm = (int[])SilverlightSerializer.Deserialize(isfs);
                rtn_pc = (int[])SilverlightSerializer.Deserialize(isfs);



                //
                // Unpersist Math
                //

                solve = (solve_state)SilverlightSerializer.Deserialize(isfs);
                integ = (integ_state)SilverlightSerializer.Deserialize(isfs);


                // Prefs

                prefBeepVol = (double)SilverlightSerializer.Deserialize(isfs);
                prefKeyClickVol = (double)SilverlightSerializer.Deserialize(isfs);
                prefKeyHighlightEn = (bool)SilverlightSerializer.Deserialize(isfs);


                // Print

                imgNum = (int)SilverlightSerializer.Deserialize(isfs);

                Main.printStack.Children.Clear();
                int count = (int)SilverlightSerializer.Deserialize(isfs);
                for (int x = 0; x < count; x++)
                {
                    byte[] imgBytes = (byte[])SilverlightSerializer.Deserialize(isfs);
                    MemoryStream ms = new MemoryStream(imgBytes);

                    BitmapImage bmi = new BitmapImage();
                    bmi.SetSource(ms);
                    WriteableBitmap wbm = new WriteableBitmap(bmi);
                    
                    Image i = new Image();
                    i.Source = wbm;

                    Main.printStack.Children.Add(i);
                }

                scrollPrint = true;
            }

            catch { err = true; }

            finally
            {
                if (isfs != null)
                    isfs.Close();
            }

            rebuild_label_table();

            return !err;
        }

        public static void hard_reset()
        {
            init_state();

            clear_display();
            clear_custom_menu();
            clear_prgm_menu();

            draw_string(0, 0, "Memory Clear".ToCharArray(), 12);

            display_x(1);
            flush_display();
        }

        public static void init_state()
        {
            prefBeepVol = 1.0;
            prefKeyClickVol = 0.5;
            prefKeyHighlightEn = true;

            // Clear stack
            free_vartype(ref reg_x);
            free_vartype(ref reg_y);
            free_vartype(ref reg_z);
            free_vartype(ref reg_t);
            free_vartype(ref reg_lastx);
            reg_x = App.new_real(0);
            reg_y = App.new_real(0);
            reg_z = App.new_real(0);
            reg_t = App.new_real(0);
            reg_lastx = App.new_real(0);

            // Clear alpha
            reg_alpha_length = 0;
            reg_alpha = new char[44];
            reg_alpha_len_max = 44;

            // Clear variables
            vars = null;
            vars_count = 0;
            vars_capacity = 0;

            vartype regs = App.new_realmatrix(25, 1);
            App.store_var("REGS".ToCharArray(), 4, regs);

            // Clear programs
            prgms = null;
            prgms_capacity = 0;
            prgms_count = 0;

            labels = null;
            labels_capacity = 0;
            labels_count = 0;

            goto_dot_dot();

            pending_command = App.CMD_NONE;

            matedit_mode = 0;
            matedit_name = new char[7];

            input_length = 0;
            input_name = new char[11];

            baseapp = 0;

            random_number = shell_random_seed();

            flags.f00 = flags.f01 = flags.f02 = flags.f03 = flags.f04 = 0;
            flags.f05 = flags.f06 = flags.f07 = flags.f08 = flags.f09 = 0;
            flags.f10 = 0;
            flags.auto_exec = 0;
            flags.double_wide_print = 0;
            flags.lowercase_print = 0;
            flags.f14 = 0;
            flags.trace_print = 0;
            flags.normal_print = 0;
            flags.f17 = flags.f18 = flags.f19 = flags.f20 = 0;
            flags.printer_enable = 1; // HP-42S sets this to 0 on hard reset
            flags.numeric_data_input = 0;
            flags.alpha_data_input = 0;
            flags.range_error_ignore = 0;
            flags.error_ignore = 0;
            flags.audio_enable = 1;
            // flags.VIRTUAL_custom_menu = 0;
            flags.decimal_point = 1;
            flags.thousands_separators = 1;
            flags.stack_lift_disable = 0;
            flags.f31 = flags.f32 = flags.f33 = 0;
            flags.agraph_control1 = 0;
            flags.agraph_control0 = 0;
            flags.digits_bit3 = 0;
            flags.digits_bit2 = 1;
            flags.digits_bit1 = 0;
            flags.digits_bit0 = 0;
            flags.fix_or_all = 1;
            flags.eng_or_all = 0;
            flags.grad = 0;
            flags.rad = 0;
            flags.continuous_on = 0;
            // flags.VIRTUAL_solving = 0;
            // flags.VIRTUAL_integrating = 0;
            // flags.VIRTUAL_variable_menu = 0;
            flags.alpha_mode = 0;
            // flags.VIRTUAL_low_battery = 0;
            flags.message = 1;
            flags.two_line_message = 0;
            flags.prgm_mode = 0;
            // flags.VIRTUAL_input = 0;
            flags.f54 = 0;
            flags.printer_exists = 1; // HP-42S sets this to 0 on hard reset
            flags.lin_fit = 1;
            flags.log_fit = 0;
            flags.exp_fit = 0;
            flags.pwr_fit = 0;
            flags.all_sigma = 1;
            flags.log_fit_invalid = 0;
            flags.exp_fit_invalid = 0;
            flags.pwr_fit_invalid = 0;
            flags.f64 = 0;
            // flags.VIRTUAL_matrix_editor = 0;
            flags.grow = 0;
            flags.f67 = 0;
            flags.base_bit0 = 0;
            flags.base_bit1 = 0;
            flags.base_bit2 = 0;
            flags.base_bit3 = 0;
            flags.local_label = 0;
            flags.polar = 0;
            flags.real_result_only = 0;
            // flags.VIRTUAL_programmable_menu = 0;
            flags.matrix_edge_wrap = 0;
            flags.matrix_end_wrap = 0;
            flags.f78 = flags.f79 = flags.f80 = flags.f81 = flags.f82 = 0;
            flags.f83 = flags.f84 = flags.f85 = flags.f86 = flags.f87 = 0;
            flags.f88 = flags.f89 = flags.f90 = flags.f91 = flags.f92 = 0;
            flags.f93 = flags.f94 = flags.f95 = flags.f96 = flags.f97 = 0;
            flags.f98 = flags.f99 = 0;

            mode_clall = false;
            mode_command_entry = false;
            mode_number_entry = false;
            mode_alpha_entry = false;
            mode_shift = false;

            mode_commandmenu = (int)MENU.NONE;
            mode_alphamenu = (int)MENU.NONE;
            mode_transientmenu = (int)MENU.NONE;
            mode_plainmenu = (int)MENU.NONE;
            mode_appmenu = (int)MENU.NONE;

            mode_running = false;
            mode_getkey = false;
            mode_pause = false;
            mode_varmenu = false;
            prgm_highlight_row = 0;

            varmenu_length = 0;
            varmenu = new char[7];
            varmenu_labellength = new int[6];
            varmenu_labeltext = new char[6][];
            for (int i = 0; i < 6; i++)
                varmenu_labeltext[i] = new char[7];

            mode_updown = false;
            mode_sigma_reg = 11;
            mode_goose = -1;
            mode_time_clktd = false;
            mode_time_clk24 = false;
            mode_time_dmy = false;


            core_settings = new core_settings_struct();
            core_settings.auto_repeat = true;

            // SLR - These flags must also be manually set in core_tables.cs: extension_struct[] extensions
            core_settings.enable_ext_copan = false;
            core_settings.enable_ext_bigstack = false;
            core_settings.enable_ext_accel = false;             // TODO: True in iPhone - Implement?
            core_settings.enable_ext_locat = false;             // TODO: True in iPhone - Implement?
            core_settings.enable_ext_heading = false;           // TODO: True in iPhone - Implement?
            core_settings.enable_ext_time = true;

            reset_math();

            imgNum = 0;
            Main.printStack.Children.Clear();
        }



        public static void core_quit()
        {
            if (mode_interruptible != null)
                stop_interruptible();

            save_state();

            App.free_vartype(ref reg_x);
            App.free_vartype(ref reg_y);
            App.free_vartype(ref reg_z);
            App.free_vartype(ref reg_t);
            App.free_vartype(ref reg_lastx);

            App.purge_all_vars();
            
            clear_all_prgms();

            vars = null;
        }

        private static void save_state()
        {
            if (MainPage.load_complete)
            {
                IsolatedStorageFile isf = IsolatedStorageFile.GetUserStoreForApplication();
                IsolatedStorageFileStream isfs = null;

                try
                {
                    isfs = new IsolatedStorageFileStream(App.STATE_FILE, FileMode.Create, isf);

                    SilverlightSerializer.Serialize(FREE42_MAGIC, isfs);
                    SilverlightSerializer.Serialize(FREE42_VERSION, isfs);

                    SilverlightSerializer.Serialize(core_settings.matrix_singularmatrix, isfs);
                    SilverlightSerializer.Serialize(core_settings.matrix_outofrange, isfs);
                    SilverlightSerializer.Serialize(core_settings.raw_text, isfs);
                    SilverlightSerializer.Serialize(core_settings.auto_repeat, isfs);
                    SilverlightSerializer.Serialize(core_settings.enable_ext_copan, isfs);
                    SilverlightSerializer.Serialize(core_settings.enable_ext_bigstack, isfs);
                    SilverlightSerializer.Serialize(core_settings.enable_ext_accel, isfs);
                    SilverlightSerializer.Serialize(core_settings.enable_ext_locat, isfs);
                    SilverlightSerializer.Serialize(core_settings.enable_ext_heading, isfs);
                    SilverlightSerializer.Serialize(core_settings.enable_ext_time, isfs);

                    SilverlightSerializer.Serialize(mode_clall, isfs);
                    SilverlightSerializer.Serialize(mode_command_entry, isfs);
                    SilverlightSerializer.Serialize(mode_number_entry, isfs);
                    SilverlightSerializer.Serialize(mode_alpha_entry, isfs);
                    SilverlightSerializer.Serialize(mode_shift, isfs);

                    SilverlightSerializer.Serialize(mode_appmenu, isfs);
                    SilverlightSerializer.Serialize(mode_plainmenu, isfs);
                    SilverlightSerializer.Serialize(mode_plainmenu_sticky, isfs);
                    SilverlightSerializer.Serialize(mode_transientmenu, isfs);
                    SilverlightSerializer.Serialize(mode_alphamenu, isfs);
                    SilverlightSerializer.Serialize(mode_commandmenu, isfs);
                    SilverlightSerializer.Serialize(mode_running, isfs);
                    SilverlightSerializer.Serialize(mode_varmenu, isfs);
                    SilverlightSerializer.Serialize(mode_updown, isfs);
                    SilverlightSerializer.Serialize(mode_getkey, isfs);

                    SilverlightSerializer.Serialize(entered_number, isfs);
                    SilverlightSerializer.Serialize(entered_string_length, isfs);
                    SilverlightSerializer.Serialize(entered_string, isfs);

                    SilverlightSerializer.Serialize(pending_command, isfs);
                    SilverlightSerializer.Serialize(pending_command_arg, isfs);
                    SilverlightSerializer.Serialize(xeq_invisible, isfs);

                    SilverlightSerializer.Serialize(incomplete_command, isfs);
                    SilverlightSerializer.Serialize(incomplete_ind, isfs);
                    SilverlightSerializer.Serialize(incomplete_alpha, isfs);
                    SilverlightSerializer.Serialize(incomplete_length, isfs);
                    SilverlightSerializer.Serialize(incomplete_maxdigits, isfs);
                    SilverlightSerializer.Serialize(incomplete_argtype, isfs);
                    SilverlightSerializer.Serialize(incomplete_num, isfs);
                    SilverlightSerializer.Serialize(incomplete_str, isfs);
                    SilverlightSerializer.Serialize(incomplete_saved_pc, isfs);
                    SilverlightSerializer.Serialize(incomplete_saved_highlight_row, isfs);

                    SilverlightSerializer.Serialize(cmdline, isfs);
                    SilverlightSerializer.Serialize(cmdline_length, isfs);
                    SilverlightSerializer.Serialize(cmdline_row, isfs);

                    SilverlightSerializer.Serialize(matedit_mode, isfs);
                    SilverlightSerializer.Serialize(matedit_name, isfs);
                    SilverlightSerializer.Serialize(matedit_length, isfs);

                    SilverlightSerializer.Serialize(matedit_x, isfs);
                    SilverlightSerializer.Serialize(matedit_i, isfs);
                    SilverlightSerializer.Serialize(matedit_j, isfs);
                    SilverlightSerializer.Serialize(matedit_prev_appmenu, isfs);

                    SilverlightSerializer.Serialize(input_name, isfs);
                    SilverlightSerializer.Serialize(input_length, isfs);
                    SilverlightSerializer.Serialize(input_arg, isfs);

                    SilverlightSerializer.Serialize(baseapp, isfs);
                    SilverlightSerializer.Serialize(random_number, isfs);

                    SilverlightSerializer.Serialize(deferred_print, isfs);

                    SilverlightSerializer.Serialize(keybuf_head, isfs);
                    SilverlightSerializer.Serialize(keybuf_tail, isfs);
                    SilverlightSerializer.Serialize(keybuf, isfs);



                    //
                    // Persist Display
                    //

                    SilverlightSerializer.Serialize(catalogmenu_section, isfs);
                    SilverlightSerializer.Serialize(catalogmenu_rows, isfs);
                    SilverlightSerializer.Serialize(catalogmenu_row, isfs);

                    SilverlightSerializer.Serialize(catalogmenu_item, isfs);
                    SilverlightSerializer.Serialize(custommenu_length, isfs);

                    SilverlightSerializer.Serialize(custommenu_label, isfs);

                    SilverlightSerializer.Serialize(progmenu_arg, isfs);

                    SilverlightSerializer.Serialize(progmenu_is_gto, isfs);
                    SilverlightSerializer.Serialize(progmenu_length, isfs);
                    SilverlightSerializer.Serialize(progmenu_label, isfs);

                    SilverlightSerializer.Serialize(display, isfs);
                    SilverlightSerializer.Serialize(appmenu_exitcallback, isfs);



                    //
                    // Persist Globals
                    //

                    array_count = 0;
                    array_list_capacity = 0;
                    array_list = null;

                    varray_count = 0;
                    varray_list_capacity = 0;
                    varray_list = null;

                    SilverlightSerializer.Serialize(reg_x, isfs);
                    SilverlightSerializer.Serialize(reg_y, isfs);
                    SilverlightSerializer.Serialize(reg_z, isfs);
                    SilverlightSerializer.Serialize(reg_t, isfs);
                    SilverlightSerializer.Serialize(reg_lastx, isfs);

                    SilverlightSerializer.Serialize(reg_alpha_length, isfs);
                    SilverlightSerializer.Serialize(reg_alpha, isfs);

                    SilverlightSerializer.Serialize(mode_sigma_reg, isfs);
                    SilverlightSerializer.Serialize(mode_goose, isfs);
                    SilverlightSerializer.Serialize(mode_time_clktd, isfs);
                    SilverlightSerializer.Serialize(mode_time_clk24, isfs);
                    SilverlightSerializer.Serialize(mode_time_dmy, isfs);

                    SilverlightSerializer.Serialize(flags, isfs);

                    SilverlightSerializer.Serialize(vars_count, isfs);
                    SilverlightSerializer.Serialize(vars, isfs);

                    SilverlightSerializer.Serialize(prgms_count, isfs);
                    SilverlightSerializer.Serialize(prgms, isfs);
                    SilverlightSerializer.Serialize(current_prgm, isfs);
                    SilverlightSerializer.Serialize(pc, isfs);
                    SilverlightSerializer.Serialize(prgm_highlight_row, isfs);

                    SilverlightSerializer.Serialize(varmenu_length, isfs);
                    SilverlightSerializer.Serialize(varmenu, isfs);
                    SilverlightSerializer.Serialize(varmenu_rows, isfs);
                    SilverlightSerializer.Serialize(varmenu_row, isfs);
                    SilverlightSerializer.Serialize(varmenu_labellength, isfs);
                    SilverlightSerializer.Serialize(varmenu_labeltext, isfs);
                    SilverlightSerializer.Serialize(varmenu_role, isfs);

                    SilverlightSerializer.Serialize(rtn_sp, isfs);
                    SilverlightSerializer.Serialize(rtn_prgm, isfs);
                    SilverlightSerializer.Serialize(rtn_pc, isfs);



                    //
                    // Persist Math
                    //

                    SilverlightSerializer.Serialize(solve, isfs);
                    SilverlightSerializer.Serialize(integ, isfs);


                    // Prefs

                    SilverlightSerializer.Serialize(prefBeepVol, isfs);
                    SilverlightSerializer.Serialize(prefKeyClickVol, isfs);
                    SilverlightSerializer.Serialize(prefKeyHighlightEn, isfs);


                    // Print

                    SilverlightSerializer.Serialize(imgNum, isfs);

                    int count = Main.printStack.Children.Count;
                    SilverlightSerializer.Serialize(count, isfs);
                    for (int x = 0; x < count; x++)
                    {
                        MemoryStream ms = new MemoryStream();
                        WriteableBitmap wb = (WriteableBitmap)((Image)Main.printStack.Children[x]).Source;
                        wb.SaveJpeg(ms, wb.PixelWidth, wb.PixelHeight, 0, imgQuality);
                        byte[] imgBytes = ms.ToArray();

                        SilverlightSerializer.Serialize(imgBytes, isfs);
                    }
                }
                catch (Exception e)
                {
                    MessageBox.Show("Write error: " + e.ToString());
                }

                if (isfs != null)
                {
                    isfs.Flush();
                    isfs.Close();
                }
            }
        }


        
        // Code to execute if a navigation fails
        void RootFrame_NavigationFailed(object sender, NavigationFailedEventArgs e)
        {
            if (System.Diagnostics.Debugger.IsAttached)
            {
                // A navigation has failed; break into the debugger
                System.Diagnostics.Debugger.Break();
            }
        }


        private class QuitException : Exception { }

        // Code to execute on Unhandled Exceptions
        private void Application_UnhandledException(object sender, ApplicationUnhandledExceptionEventArgs e)
        {
            if (e.ExceptionObject is QuitException)
            {
                core_quit();

                return;
            }

            if (System.Diagnostics.Debugger.IsAttached)
            {
                // An unhandled exception has occurred; break into the debugger
                System.Diagnostics.Debugger.Break();
            }
        }

        #region Phone application initialization

        // Avoid double-initialization
        private bool phoneApplicationInitialized = false;

        // Do not add any additional code to this method
        private void InitializePhoneApplication()
        {
            if (phoneApplicationInitialized)
                return;

            // Create the frame but don't set it as RootVisual yet; this allows the splash
            // screen to remain active until the application is ready to render.
            RootFrame = new PhoneApplicationFrame();
            RootFrame.Navigated += CompleteInitializePhoneApplication;

            // Handle navigation failures
            RootFrame.NavigationFailed += RootFrame_NavigationFailed;

            // Ensure we don't initialize again
            phoneApplicationInitialized = true;

            app_init();
        }

        // Do not add any additional code to this method
        private void CompleteInitializePhoneApplication(object sender, NavigationEventArgs e)
        {
            // Set the root visual to allow the application to render
            if (RootVisual != RootFrame)
                RootVisual = RootFrame;

            // Remove this handler since it is no longer needed
            RootFrame.Navigated -= CompleteInitializePhoneApplication;
        }

        #endregion
    }
}
