//
// Porting and modifications made by Rayzzor Studios, 2011
//

/**
 * Copyright (c) 2005-2009 voidware ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS," WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

using System;


namespace WP7_42s
{
    public class BCD
    {
        public BCDFloat _v;
        public static string _buf;

        #region Constructors

        public BCD()
        {
            _v = new BCDFloat();        // SLR: Added instantiation of data fields for default constructor to avoid null references
        }

        public BCD(string s)
        {
            _v = new BCDFloat(s);
        }

        public BCD(int v)
        {
            _v = new BCDFloat(v);
        }

        public BCD(uint v)
        {
            _v = new BCDFloat(v);
        }
        
        public BCD(BCDFloatData bf)
        {
            _v = new BCDFloat(bf);
        }

        #endregion

        #region Operators

        public static BCD operator +(BCD a, BCD b)
        {
            BCD c = new BCD();
            BCDFloat.add(a._v, b._v, c._v);
            return c;
        }

        public static BCD operator -(BCD a, BCD b)
        {
            BCD c = new BCD();
            BCDFloat.sub(a._v, b._v, c._v);
            return c;
        }

        public static BCD operator *(BCD a, BCD b)
        {
            BCD c = new BCD();
            BCDFloat.mul(a._v, b._v, c._v);
            return c;
        }

        public static BCD operator /(BCD a, BCD b)
        {
            BCD c = new BCD();
            BCDFloat.div(a._v, b._v, c._v);
            return c;
        }

        public static BCD operator ++(BCD a)
        {
            BCD c = new BCD(a._v);
            c += new BCD(1);
            return c;
        }

        public static BCD operator --(BCD a)
        {
            BCD c = new BCD(a._v);
            c -= new BCD(1);
            return c;
        }

        public static bool operator ==(BCD a, BCD b)
        {
            if ((object)a == null && (object)b == null)
                return true;
            else if ((object)a == null && (object)b != null)
                return false;
            else if ((object)a != null && (object)b == null)
                return false;
            else
                return BCDFloat.equal(a._v, b._v);
        }

        public static bool operator !=(BCD a, BCD b)
        {
            if ((object)a == null && (object)b == null)
                return false;
            else if ((object)a == null && (object)b != null)
                return true;
            else if ((object)a != null && (object)b == null)
                return true;
            else
                return !BCDFloat.equal(a._v, b._v);
        }

        public static bool operator <(BCD a, BCD b)
        {
            return BCDFloat.lt(a._v, b._v);
        }

        public static bool operator <=(BCD a, BCD b)
        {
            return BCDFloat.le(a._v, b._v);
        }

        public static bool operator >(BCD a, BCD b)
        {
            return BCDFloat.gt(a._v, b._v);
        }

        public static bool operator >=(BCD a, BCD b)
        {
            return BCDFloat.ge(a._v, b._v);
        }

        #endregion

        #region Members

        public int exponent()
        {
            return _v.exp();
        }

        public void setExponent(int v)
        {
            _v.exp(v);
        }

        public int digit(int n)
        {
            return _v.d_[n];
        }

        public void digit(int n, int v)
        {
            _v.d_[n] = (ushort) v;
        }

        public bool isZero()
        {
            return _v.isZero();
        }

        public bool isNeg()
        {
            return _v.neg();
        }

        public bool isSpecial()
        {
            return _v.isSpecial();
        }

        public bool isInf()
        {
            return  _v.isInf();
        }

        public bool isNan()
        {
            return _v.isNan();
        }

        public bool isInteger()
        {
            return _v.isInteger();
        }

        public string asString()
        {
            _v.asString(ref _buf);
            return _buf;
        }

	    public string asStringFmt(BCDFloat.Format fmt, int precision)
	    {
		    BCD bp = this;
		    BCD rv = new BCD();

		    if (fmt != BCDFloat.Format.format_truncate)
		    {
			    rv = round(precision);
			    bp = rv;
		    }

            bp._v.asStringFmt(ref _buf, fmt, precision);
		    return _buf;
	    }

        public BCD round(int digits)
        {
            BCD rv = new BCD();
            _v._roundDigits((uint) digits, ref rv._v);
            return rv;
        }

        public void negate()
        {
            _v.negate();
        }

        #endregion

        #region Static

        public static BCD inf()
        {
            return new BCD(BCDFloat.posInf());
        }

        public static BCD epsilon(int n)
        {
            BCD v = new BCD();
            BCDFloat.epsilon(n, v._v);
            return v;
        }

        public static int ifloor(BCD a)
        {
            return BCDFloat.ifloor(a._v);
        }

        public static int itrunc(BCD a)
        {
            return BCDFloat.itrunc(a._v);
        }
        
        public static BCD floor(BCD a)
        {
            /* floor, largest integer <= a.
             * eg floor(2.1) = 2.
             *    floor(-2.1) = -3.
             */
            BCD t = new BCD();

            if (a.isSpecial())
                return a;

            BCDFloat.floor(a._v, ref t._v);
            return t;
        }

        public static BCD trunc(BCD a)
        {
            /* truncate towards zero.
             * trunc(2.1) = 2.
             * trunc(-2.1) = -2
             */
            BCD t = new BCD();

            if (a.isSpecial())
                return a;

            BCDFloat.trunc(a._v, ref t._v);
            return t;
        }

        public static BCD fabs(BCD a)
        {
            if (a.isNeg())
            {
                BCD b = new BCD(0);
                b -= a;
                return b;
            }
            else
                return a;
        }

        public static BCD frac(BCD a)
        {
            if (a.isSpecial())
                return a;

            return a - trunc(a);
        }

        public static BCD sqrt(BCD a)
        {
            BCD c = new BCD();

            if (!BCDFloat.sqrt(a._v, ref c._v))
                c._v = BCDFloat.nan();

            return c;
        }

        public static BCD pow(BCD a, int n)
        {
            int m;

            if (n == 0)
                return new BCD(1);

            m = (n < 0) ? -n : n;

            BCD2 s = new BCD2();

            if (m > 1)
            {
                BCD2 r = new BCD2(a._v);
                s = new BCD2(1);

                // Use binary exponentiation 
                for (; ; )
                {
                    if ((m & 1) != 0)
                        s *= r;

                    m >>= 1;

                    if (m == 0)
                        break;
                    r *= r;
                }
            }
            else
            {
                s = new BCD2(a._v);
            }

            // Compute the reciprocal if n is negative. 
            BCD s1 = s.asBCD();
            return n < 0 ? new BCD(1) / s1 : s1;
        }

        #endregion
    }
}
