﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/**
 * Copyright (c) 2005-2009 voidware ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS," WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

using System;


namespace WP7_42s
{
    public class BCD2
    {
        public BCDFloat2 _v;

        #region Constructors

        public BCD2()
        {
            _v = new BCDFloat2();       // SLR: Added instantiation of data fields for default constructor to avoid null references
        }

        public BCD2(int v)
        {
            _v = new BCDFloat2(v);
        }

        public BCD2(uint v)
        {
            _v = new BCDFloat2(v);
        }

        public BCD2(BCDFloatData2 bf)
        {
            _v = new BCDFloat2(bf);
        }

        public BCD2(BCDFloatData bf)
        {
            _v = new BCDFloat2(bf);
        }

        public BCD2(BCD b)
        {
            _v = new BCDFloat2(b._v);
        }

        #endregion

        #region Operators

        public static BCD2 operator +(BCD2 a, BCD2 b)
        {
            BCD2 c = new BCD2();
            BCDFloat2.add(a._v, b._v, c._v);
            return c;
        }

        public static BCD2 operator -(BCD2 a, BCD2 b)
        {
            BCD2 c = new BCD2();
            BCDFloat2.sub(a._v, b._v, c._v);
            return c;
        }

        public static BCD2 operator *(BCD2 a, BCD2 b)
        {
            BCD2 c = new BCD2();
            BCDFloat2.mul(a._v, b._v, c._v);
            return c;
        }

        public static BCD2 operator /(BCD2 a, BCD2 b)
        {
            BCD2 c = new BCD2();
            BCDFloat2.div(a._v, b._v, c._v);
            return c;
        }

        public static BCD2 operator ++(BCD2 a)
        {
            BCD2 c = new BCD2(a._v);
            c += new BCD2(1);
            return c;
        }

        public static BCD2 operator --(BCD2 a)
        {
            BCD2 c = new BCD2(a._v);
            c -= new BCD2(1);
            return c;
        }

        public static bool operator ==(BCD2 a, BCD2 b)
        {
            if ((object)a == null && (object)b == null)
                return true;
            else if ((object)a == null && (object)b != null)
                return false;
            else if ((object)a != null && (object)b == null)
                return false;
            else
                return BCDFloat2.equal(a._v, b._v);
        }

        public static bool operator !=(BCD2 a, BCD2 b)
        {
            if ((object)a == null && (object)b == null)
                return false;
            else if ((object)a == null && (object)b != null)
                return true;
            else if ((object)a != null && (object)b == null)
                return true;
            else
                return !BCDFloat2.equal(a._v, b._v);
        }

        public static bool operator <(BCD2 a, BCD2 b)
        {
            return BCDFloat2.lt(a._v, b._v);
        }

        public static bool operator <=(BCD2 a, BCD2 b)
        {
            return BCDFloat2.le(a._v, b._v);
        }

        public static bool operator >(BCD2 a, BCD2 b)
        {
            return BCDFloat2.gt(a._v, b._v);
        }

        public static bool operator >=(BCD2 a, BCD2 b)
        {
            return BCDFloat2.ge(a._v, b._v);
        }
/*
        public static bool operator ==(BCDFloatData2 a, BCDFloatData2 b)
        {
            return new BCD2(a) == new BCD2(b);
        }

        public static bool operator !=(BCDFloatData2 a, BCDFloatData2 b)
        {
            return new BCD2(a) != new BCD2(b);
        }

        public static bool operator <(BCDFloatData2 a, BCDFloatData2 b)
        {
            return new BCD2(a) < new BCD2(b);
        }

        public static bool operator <=(BCDFloatData2 a, BCDFloatData2 b)
        {
            return new BCD2(a) <= new BCD2(b);
        }

        public static bool operator >(BCDFloatData2 a, BCDFloatData2 b)
        {
            return new BCD2(a) > new BCD2(b);
        }

        public static bool operator >=(BCDFloatData2 a, BCDFloatData2 b)
        {
            return new BCD2(a) >= new BCD2(b);
        }

        public static BCD2 operator -(BCDFloatData2 a, BCDFloatData2 b)
        {
            return new BCD2(a) - new BCD2(b);
        }
*/
        #endregion

        #region Members

        public int exponent()
        {
            return _v.exp();
        }

        public void setExponent(int v)
        {
            _v.exp(v);
        }

        public int digit(int n)
        {
            return _v.d_[n];
        }

        public void digit(int n, int v)
        {
            _v.d_[n] = (ushort) v;
        }

        public bool isZero()
        {
            return _v.isZero();
        }

        public bool isNeg()
        {
            return _v.neg();
        }

        public bool isSpecial()
        {
            return _v.isSpecial();
        }

        public bool isInf()
        {
            return  _v.isInf();
        }

        public bool isNan()
        {
            return _v.isNan();
        }

        public bool isInteger()
        {
            return _v.isInteger();
        }

        public void negate()
        {
            _v.negate();
        }

        public BCD asBCD()
        {
            BCD v = new BCD();
            _v.asBCD(v._v);
            return v;
        }

        public void makeInf()
        {
            _v.makeInf();
        }

        #endregion

        #region Static

        public static BCD2 epsilon(int n)
        {
            BCD2 v = new BCD2();
            BCDFloat2.epsilon(n, v._v);
            return v;
        }

        public static int ifloor(BCD2 a)
        {
            return BCDFloat2.ifloor(a._v);
        }

        public static int itrunc(BCD2 a)
        {
            return BCDFloat2.itrunc(a._v);
        }

        public static BCD2 floor(BCD2 a)
        {
            // floor, largest integer <= a.
            // eg floor(2.1) = 2.
            //    floor(-2.1) = -3.
            //
            BCD2 t = new BCD2();

            if (a.isSpecial())
                return a;

            BCDFloat2.floor(a._v, ref t._v);
            return t;
        }

        public static BCD2 trunc(BCD2 a)
        {
            // truncate towards zero.
            // trunc(2.1) = 2.
            // trunc(-2.1) = -2
            //
            BCD2 t = new BCD2();

            if (a.isSpecial())
                return a;

            BCDFloat2.trunc(a._v, ref t._v);
            return t;
        }

        public static BCD2 fabs(BCD2 a)
        {
            if (a.isNeg())
            {
                BCD2 b = new BCD2(0);
                b -= a;
                return b;
            }
            else
                return a;
        }

        public static BCD2 frac(BCD2 a)
        {
            if (a.isSpecial())
                return a;

            return a - trunc(a);
        }

        public static BCD2 sqrt(BCD2 a)
        {
            BCD2 c = new BCD2();

            if (!BCDFloat2.sqrt(a._v, ref c._v))
                c._v.makeNAN();

            return c;
        }

        public static BCD2 pow(BCD2 a, int n)
        {
            int m;

            if (n == 0)
                return new BCD2(1);

            m = (n < 0) ? -n : n;

            BCD2 s = new BCD2();

            if (m > 1) 
            {
                BCD2 r = new BCD2(a._v);
                s = new BCD2(1);

                /* Use binary exponentiation */
                for (;;) 
                {
                    if ((m & 1) != 0)
                        s *= r;

                    m >>= 1;

                    if (m == 0)
                        break;
                    r *= r;
                }
            }
            else
            {
                s = new BCD2(a._v);
            }

            // Compute the reciprocal if n is negative.
            if (n < 0) 
                return new BCD2(1) / s;

            return s;
        }

        #endregion
    }
}
