﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/**
 * Copyright (c) 2005-2009 voidware ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS," WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */


using System;


namespace WP7_42s
{
    public class BCDFloatData2
    {
        //     store P 4dec `digits', equivalent to P*4 decimal digits.
        //     the last place is the exponent.
        //     
	    public ushort[] d_ = new ushort[2 * App.P + 1 + 1];
    }


    public class BCDFloat2 : BCDFloatData2
    {
        //
        // Constructors
        //
        #region Constructors

	    public BCDFloat2() { }          // Warning: not initialised.

	    public BCDFloat2(int v)
	    {
		    _init();
		    if (v != 0)
		    {
			    bool neg = v < 0;
			    if (neg)
				    v = -v;
			    _fromUInt((uint)v);
			    if (neg)
				    negate();
		    }
	    }

        public BCDFloat2(uint v)
	    {
		    _init();
		    if (v != 0)
			    _fromUInt(v);
	    }

	    public BCDFloat2(BCDFloatData2 d)
	    {
		    _bcdUpdate2(d);
	    }

	    public BCDFloat2(BCDFloatData d)
	    {
		    int i;
		    for (i = 0; i < App.P; ++i)
			    d_[i] = d.d_[i];

		    while (i < 2 * App.P + 1)
			    d_[i++] = 0;

		    d_[2 * App.P + 1] = d.d_[App.P];
	    }

        #endregion


        #region Members

        public void _bcdUpdate2(BCDFloatData2 d)
        {
            d_[0] = d.d_[0];
            d_[1] = d.d_[1];
            d_[2] = d.d_[2];
            d_[3] = d.d_[3];
            d_[4] = d.d_[4];
            d_[5] = d.d_[5];
            d_[6] = d.d_[6];
            d_[7] = d.d_[7];

            d_[8] = d.d_[8];
            d_[9] = d.d_[9];
            d_[10] = d.d_[10];
            d_[11] = d.d_[11];
            d_[12] = d.d_[12];
            d_[13] = d.d_[13];
            d_[14] = d.d_[14];
            d_[15] = d.d_[15];
        }

	    public int exp()
	    {
		    return ((short)(d_[2 * App.P + 1] << 3)) >> 3;
	    }
	    
        public void exp(int v)
	    {
		    d_[2 * App.P + 1] = (ushort) (v & App.EXPMASK);
	    }

	    public bool neg()
	    {
		    return ((d_[2 * App.P + 1] & App.NEG) != 0) && (d_[0] != 0 || isInf());
	    }

	    public void setSign()
	    {
		    d_[2 * App.P + 1] |= App.NEG;
	    }

        public void clearSign()
	    {
            ushort n = App.NEG;
		    d_[2 * App.P + 1] &= (ushort)~n;
	    }

	    public void negate()
	    {
		    d_[2 * App.P + 1] ^= App.NEG;
	    }

	    public bool isSpecial()
	    {
		    return (d_[2 * App.P + 1] & 0x6000) != 0;
	    }

	    public bool isZero()
	    {
		    return (d_[0] == 0) && !isSpecial();
	    }

	    public bool isNan()
	    {
		    return (d_[2 * App.P + 1] & 0x4000) != 0;
	    }

        public bool isInf()
	    {
		    return (d_[2 * App.P + 1] & 0x2000) != 0;
	    }

        public bool isInteger()
	    {
		    if (isZero())
			    return true;
	
		    int e = exp();
		    int i;
		    for (i = 2 * App.P + 1 - 1; i >= 0; --i)
			    if (d_[i] != 0)
				    return e > i;
		    return false;
	    }
        
        public void ldexp(uint mant, int e)
	    {
		    // load the exp as a 4-block
		    _init();
		    _fromUInt(mant);
		    exp(e);
	    }



       	public static void add(BCDFloat2 a, BCDFloat2 b, BCDFloat2 c)
	    {
		    GlobalMembersBcdfloat.bcd_add(a.d_, b.d_, c.d_, 2 * App.P + 1);
	    }

	    public static void sub(BCDFloat2 a, BCDFloat2 b, BCDFloat2 c)
	    {
		    GlobalMembersBcdfloat.bcd_sub(a.d_, b.d_, c.d_, 2 * App.P + 1);
	    }

	    public static void mul(BCDFloat2 a, BCDFloat2 b, BCDFloat2 c)
	    {
		    GlobalMembersBcdfloat.bcd_mul(a.d_, b.d_, c.d_, 2 * App.P + 1);
	    }

	    public static void div(BCDFloat2 a, BCDFloat2 b, BCDFloat2 c)
	    {
		    GlobalMembersBcdfloat.bcd_div(a.d_, b.d_, c.d_, 2 * App.P + 1);
	    }



	    public void asBCD(BCDFloat v)
	    {
		    for (int i = 0; i <= App.P; ++i)
			    v.d_[i] = d_[i];

		    int e = d_[2 * App.P + 1];

		    if (v._round25() != 0) // XX overflow?
			    ++e;

		    v.d_[App.P] = (ushort) e;
	    }

        public static bool sqrt(BCDFloat2 a, ref BCDFloat2 r)
	    {
		    // use single precision version as approx.
		    BCDFloat a1 = new BCDFloat();
		    BCDFloat r1 = new BCDFloat();

            a.asBCD(a1);

		    bool res = BCDFloat.sqrt(a1, ref r1);
	
		    BCDFloat2 rr = new BCDFloat2(r1);
		    BCDFloat2 t = new BCDFloat2();
		    BCDFloat2 t2 = new BCDFloat2();
		    BCDFloat2 two = new BCDFloat2(2);

		    if (res)
		    {
			    if (!rr.isSpecial())
			    {
				    for (int i = 0; i < 3; ++i)
				    {
					    div(a, rr, t); // t = a/x
					    add(t, rr, t2); // t2 = (a/x) + x
					    div(t2, two, rr); // x = (t2)/2
				    }
			    }
			    r = new BCDFloat2(rr);
		    }
		    return res;
	    }



        public static bool lt(BCDFloat2 a, BCDFloat2 b)
	    {
		    // true iff a < b 
		    return GlobalMembersBcdfloat.bcd_cmp(a.d_, b.d_, 2 * App.P + 1) < 0;
	    }

	    public static bool le(BCDFloat2 a, BCDFloat2 b)
	    {
		    // true iff a <= b 
		    return GlobalMembersBcdfloat.bcd_cmp(a.d_, b.d_, 2 * App.P + 1) <= 0;
	    }

	    public static bool gt(BCDFloat2 a, BCDFloat2 b)
	    {
		    // true iff a > b 
		    return GlobalMembersBcdfloat.bcd_cmp(a.d_, b.d_, 2 * App.P + 1) == 1;
	    }

	    public static bool ge(BCDFloat2 a, BCDFloat2 b)
	    {
		    // true iff a >= b 
		    int res = GlobalMembersBcdfloat.bcd_cmp(a.d_, b.d_, 2 * App.P + 1);
	        return res == 0 || res == 1;
	    }

	    public static bool equal(BCDFloat2 a, BCDFloat2 b)
	    {
		    return GlobalMembersBcdfloat.bcd_cmp(a.d_, b.d_, 2 * App.P + 1) == 0;
	    }

	    public static int ifloor(BCDFloat2 x)
	    {
		    BCDFloat2 a = new BCDFloat2();
		    floor(x, ref a);
		    return a.asInt();
	    }

	    public static int itrunc(BCDFloat2 x)
	    {
		    BCDFloat2 a = new BCDFloat2();
		    trunc(x, ref a);
		    return a.asInt();
	    }

	    public static bool floor(BCDFloat2 a, ref BCDFloat2 c)
	    {
	    //     floor, largest integer <= a.
	    //     eg floor(2.1) = 2.
	    //        floor(-2.1) = -3.
	    //
	
		    c = new BCDFloat2(a);
		    int e = c.exp();
		    int i;
		    bool changed = false;
		    for (i = 2 * App.P + 1 - 1; i >= 0; --i)
		    {
			    if (i >= e)
			    {
				    if (c.d_[i] != 0)
					    changed = true;

				    c.d_[i] = 0;
			    }
		    }
	
		    if (c.neg() && changed)
		    {
			    // need to subtract 1
			    for (i = 2 * App.P + 1 - 1; i >= 0; --i)
			    {
				    if (c.d_[i] != 0)
				    {
					    ++c.d_[i];
					    break;
				    }
			    }
		    }
		    return true;
	    }

	    public static bool trunc(BCDFloat2 a, ref BCDFloat2 c)
	    {
	    //     truncate towards zero.
	    //     * trunc(2.1) = 2.
	    //     * trunc(-2.1) = -2
	    //
		    c = new BCDFloat2(a);
		    int e = c.exp();
		    int i;
		    for (i = 2 * App.P + 1 - 1; i >= 0; --i)
			    if (i >= e)
				    c.d_[i] = 0;
	
		    return true;
	    }



        public void _init()
	    {
		    for (int i = 0; i <= 2 * App.P + 1; ++i)
			    d_[i] = 0;
	    }



        public void makeInf()
	    {
		    _init();
		    d_[2 * App.P + 1] = App.POS_INF_EXP;
	    }

	    public void makeNAN()
	    {
		    _init();
		    d_[2 * App.P + 1] = App.NAN_EXP;
	    }

	    public int _round25()
	    {
		    return GlobalMembersBcdfloat.bcd_round25(d_, 2 * App.P + 1);
	    }

	    public void _fromUInt(uint v)
	    {
		    GlobalMembersBcdfloat.bcd_fromUInt(d_, 2 * App.P + 1, v);
	    }



        public static void epsilon(int n, BCDFloat2 v)
	    {
		    // generate 10^-n, 
		    int m = GlobalMembersBcdfloat.BCDDecade[(n - 1) & 3];
		    v.ldexp((uint)m, -(n >> 2));
	    }

        public int asInt()
	    {
		    // if we fit in an int, return it otherwise 0
		    int ea = exp();
		    int v = 0;
		    int i = 0;
		    while (i < ea && i < 2 * App.P + 1)
		    {
			    if (v > 214748L) // too large, bail out.
				    return 0;
			    v *= App.BASE;
			    v += d_[i];
			    ++i;
		    }
		    if (neg()) v = -v;
		    return v;
	    }

        #endregion
    }
}
