﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        public static int docmd_clx(arg_struct arg)
        {
            free_vartype(ref reg_x);
            reg_x = new_real(0);
            mode_disable_stack_lift = true;
            return ERR_NONE;
        }

        public static int docmd_enter(arg_struct arg)
        {
            vartype v = dup_vartype(reg_x);

            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;

            free_vartype(ref reg_t);
            reg_t = reg_z;
            reg_z = reg_y;
            reg_y = v;
            
            mode_disable_stack_lift = true;
            return ERR_NONE;
        }

        public static int docmd_swap(arg_struct arg)
        {
            vartype temp = reg_x;
            reg_x = reg_y;
            reg_y = temp;

            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);

            return ERR_NONE;
        }

        public static int docmd_rdn(arg_struct arg)
        {
            vartype temp = reg_x;
            reg_x = reg_y;
            reg_y = reg_z;
            reg_z = reg_t;
            reg_t = temp;
            
            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_chs(arg_struct arg)
        {
            switch (reg_x.type)
            {
                case TYPE.REAL:
                    {
                        vartype_real r = (vartype_real)reg_x;
                        r.x = -(r.x);
                        break;
                    }
                case TYPE.COMPLEX:
                    {
                        vartype_complex c = (vartype_complex)reg_x;
                        c.re = -(c.re);
                        c.im = -(c.im);
                        break;
                    }
                case TYPE.REALMATRIX:
                    {
                        vartype_realmatrix rm = (vartype_realmatrix)reg_x;
                        int sz = rm.rows * rm.columns;

                        for (int i = 0; i < sz; i++)
                            if (rm.array.is_string[i] != 0)
                                return ERR_ALPHA_DATA_IS_INVALID;
                        
                        if (disentangle((vartype)rm) == 0)
                            return ERR_INSUFFICIENT_MEMORY;
                        
                        for (int i = 0; i < sz; i++)
                            rm.array.data[i] = -(rm.array.data[i]);
                        break;
                    }
                case TYPE.COMPLEXMATRIX:
                    {
                        vartype_complexmatrix cm = (vartype_complexmatrix)reg_x;
                        int sz = cm.rows * cm.columns * 2;

                        if (disentangle((vartype)cm) == 0)
                            return ERR_INSUFFICIENT_MEMORY;

                        for (int i = 0; i < sz; i++)
                            cm.array.data[i] = -(cm.array.data[i]);
                        break;
                    }
                case TYPE.STRING:
                    return ERR_ALPHA_DATA_IS_INVALID;
            }

            if (flags.trace_print != 0 && flags.printer_exists!= 0)
                docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_div(arg_struct arg)
        {
            return generic_div(reg_x, reg_y, docmd_div_completion);
        }

        public static void docmd_div_completion(int error, vartype res)
        {
            if (error == ERR_NONE)
                binary_result(res);
        }

        public static int docmd_add(arg_struct arg)
        {
            vartype res = new vartype();

            int error = generic_add(reg_x, reg_y, ref res);
            
            if (error == ERR_NONE)
                binary_result(res);
            
            return error;
        }

        public static int docmd_sub(arg_struct arg)
        {
            vartype res = new vartype();

            int error = generic_sub(reg_x, reg_y, ref res);
            
            if (error == ERR_NONE)
                binary_result(res);
            
            return error;
        }

        public static int docmd_mul(arg_struct arg)
        {
            return generic_mul(reg_x, reg_y, docmd_mul_completion);
        }

        public static void docmd_mul_completion(int error, vartype res)
        {
            if (error == ERR_NONE)
                binary_result(res);
        }

        public static int docmd_lastx(arg_struct arg)
        {
            vartype v = dup_vartype(reg_lastx);

            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            recall_result(v);
            return ERR_NONE;
        }

        public static int docmd_percent(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING || reg_y.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            if (reg_x.type != TYPE.REAL || reg_y.type != TYPE.REAL)
                return ERR_INVALID_TYPE;
            else
            {
                vartype_real x = (vartype_real)reg_x;
                vartype_real y = (vartype_real)reg_y;

                Phloat res = x.x * y.x;
                
                if (Phloat.p_isinf(res) != 0)
                {
                    // Try different evaluation order
                    res = (x.x / 100.0) * y.x;

                    if (Phloat.p_isinf(res) != 0)
                        return ERR_OUT_OF_RANGE;
                }
                else
                    res /= 100.0;

                vartype new_x = new_real(res);
                
                if (new_x == null)
                    return ERR_INSUFFICIENT_MEMORY;
                
                free_vartype(ref reg_lastx);
                reg_lastx = reg_x;
                reg_x = new_x;

                if (flags.trace_print != 0 && flags.printer_exists != 0)
                    docmd_prx(null);
                
                return ERR_NONE;
            }
        }

        public static int docmd_pi(arg_struct arg)
        {
            vartype v = new_real(PI);
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            recall_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_complex(arg_struct arg)
        {
            switch (reg_x.type)
            {
                case TYPE.REAL:
                    {
                        if (reg_y.type == TYPE.STRING)
                            return ERR_ALPHA_DATA_IS_INVALID;
                        else if (reg_y.type != TYPE.REAL)
                            return ERR_INVALID_TYPE;

                        vartype v;
                        
                        if (flags.polar != 0)
                        {
                            Phloat re = new Phloat();
                            Phloat im = new Phloat();

                            generic_p2r(((vartype_real)reg_y).x, ((vartype_real)reg_x).x, ref re, ref im);
                            
                            v = new_complex(re, im);
                        }
                        else
                        {
                            v = new_complex(((vartype_real)reg_y).x, ((vartype_real)reg_x).x);
                        }

                        if (v == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        
                        free_vartype(ref reg_lastx);
                        reg_lastx = reg_x;
                        reg_x = v;
                        
                        free_vartype(ref reg_y);
                        reg_y = reg_z;
                        reg_z = dup_vartype(reg_t);
                        
                        break;
                    }

                case TYPE.COMPLEX:
                    {
                        vartype new_x = new_real(0);
                        vartype new_y = new_real(0);
                        
                        if (new_x == null || new_y == null)
                        {
                            free_vartype(ref new_x);
                            free_vartype(ref new_y);
                            
                            return ERR_INSUFFICIENT_MEMORY;
                        }

                        if (flags.polar != 0)
                        {
                            Phloat r = new Phloat();
                            Phloat phi = new Phloat();

                            generic_r2p(((vartype_complex)reg_x).re, ((vartype_complex)reg_x).im, ref r, ref phi);

                            if (Phloat.p_isinf(r) != 0)
                            {
                                if (flags.range_error_ignore != 0)
                                    r = Phloat.POS_HUGE_PHLOAT;
                                else
                                {
                                    free_vartype(ref new_x);
                                    free_vartype(ref new_y);

                                    return ERR_OUT_OF_RANGE;
                                }
                            }
                            
                            ((vartype_real)new_y).x = r;
                            ((vartype_real)new_x).x = phi;
                        }
                        else
                        {
                            ((vartype_real)new_y).x = ((vartype_complex)reg_x).re;
                            ((vartype_real)new_x).x = ((vartype_complex)reg_x).im;
                        }

                        free_vartype(ref reg_lastx);
                        reg_lastx = reg_x;
                        free_vartype(ref reg_t);
                        reg_t = reg_z;
                        reg_z = reg_y;

                        reg_y = new_y;
                        reg_x = new_x;
                        
                        break;
                    }

                case TYPE.REALMATRIX:
                    {
                        if (reg_y.type == TYPE.STRING)
                            return ERR_ALPHA_DATA_IS_INVALID;
                        else if (reg_y.type != TYPE.REALMATRIX)
                            return ERR_INVALID_TYPE;
                        else
                        {
                            vartype_realmatrix re_m = (vartype_realmatrix)reg_y;
                            vartype_realmatrix im_m = (vartype_realmatrix)reg_x;
                            vartype_complexmatrix cm;
                            int sz, i;
                            
                            if (re_m.rows != im_m.rows || re_m.columns != im_m.columns)
                                return ERR_DIMENSION_ERROR;

                            sz = re_m.rows * re_m.columns;

                            for (i = 0; i < sz; i++)
                                if (re_m.array.is_string[i] != 0)
                                    return ERR_ALPHA_DATA_IS_INVALID;

                            for (i = 0; i < sz; i++)
                                if (im_m.array.is_string[i] != 0)
                                    return ERR_ALPHA_DATA_IS_INVALID;

                            cm = (vartype_complexmatrix) new_complexmatrix(re_m.rows, re_m.columns);
                            
                            if (cm == null)
                                return ERR_INSUFFICIENT_MEMORY;
                            
                            if (flags.polar != 0)
                            {
                                for (i = 0; i < sz; i++)
                                {
                                    generic_p2r(re_m.array.data[i], im_m.array.data[i], ref cm.array.data[2 * i], ref cm.array.data[2 * i + 1]);
                                }
                            }
                            else
                            {
                                for (i = 0; i < sz; i++)
                                {
                                    cm.array.data[2 * i] = re_m.array.data[i];
                                    cm.array.data[2 * i + 1] = im_m.array.data[i];
                                }
                            }

                            free_vartype(ref reg_lastx);
                            reg_lastx = reg_x;
                            free_vartype(ref reg_y);
                            reg_y = reg_z;
                            reg_z = dup_vartype(reg_t);
                            reg_x = (vartype)cm;
                            
                            break;
                        }
                    }

                case TYPE.COMPLEXMATRIX:
                    {
                        vartype_complexmatrix cm = (vartype_complexmatrix)reg_x;
                        
                        int rows = cm.rows;
                        int columns = cm.columns;
                        int sz = rows * columns;
                        int i;
                        
                        vartype_realmatrix re_m = (vartype_realmatrix) new_realmatrix(rows, columns);
                        vartype_realmatrix im_m;
                        
                        if (re_m == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        
                        im_m = (vartype_realmatrix) new_realmatrix(rows, columns);
                        
                        if (im_m == null)
                        {
                            vartype _vt = (vartype)re_m;
                            re_m = null;
                            free_vartype(ref _vt);
                            return ERR_INSUFFICIENT_MEMORY;
                        }

                        if (flags.polar != 0)
                        {
                            for (i = 0; i < sz; i++)
                            {
                                Phloat r = new Phloat();
                                Phloat phi = new Phloat();

                                generic_r2p(cm.array.data[2 * i], cm.array.data[2 * i + 1], ref r, ref phi);

                                if (Phloat.p_isinf(r) != 0)
                                {
                                    if (flags.range_error_ignore != 0)
                                        r = Phloat.POS_HUGE_PHLOAT;
                                    else
                                    {
                                        vartype _vt = (vartype)re_m;
                                        re_m = null;
                                        free_vartype(ref _vt);

                                        _vt = (vartype)im_m;
                                        im_m = null;
                                        free_vartype(ref _vt);

                                        return ERR_OUT_OF_RANGE;
                                    }
                                }
                                
                                re_m.array.data[i] = r;
                                im_m.array.data[i] = phi;
                            }
                        }
                        else
                        {
                            for (i = 0; i < sz; i++)
                            {
                                re_m.array.data[i] = cm.array.data[2 * i];
                                im_m.array.data[i] = cm.array.data[2 * i + 1];
                            }
                        }

                        free_vartype(ref reg_lastx);
                        reg_lastx = reg_x;
                        free_vartype(ref reg_t);
                        reg_t = reg_z;
                        reg_z = reg_y;
                        reg_y = (vartype)re_m;
                        reg_x = (vartype)im_m;
                        
                        break;
                    }

                case TYPE.STRING:
                    return ERR_ALPHA_DATA_IS_INVALID;
            }

            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);

            return ERR_NONE;
        }

        public static int docmd_sto(arg_struct arg)
        {
            return generic_sto(arg, (char)0);
        }

        public static int docmd_sto_div(arg_struct arg)
        {
            return generic_sto(arg, '/');
        }

        public static int docmd_sto_mul(arg_struct arg)
        {
            return generic_sto(arg, '*');
        }

        public static int docmd_sto_sub(arg_struct arg)
        {
            return generic_sto(arg, '-');
        }

        public static int docmd_sto_add(arg_struct arg)
        {
            return generic_sto(arg, '+');
        }

        public static int docmd_rcl(arg_struct arg)
        {
            vartype v = new vartype();

            int err = generic_rcl(arg, ref v);
            
            if (err == ERR_NONE)
                recall_result(v);
            
            return err;
        }

        public static int docmd_rcl_div(arg_struct arg)
        {
            int err = generic_rcl(arg, ref temp_v);

            if (err != ERR_NONE)
                return err;
            
            return generic_div(temp_v, reg_x, docmd_rcl_div_completion);
        }

        public static int docmd_rcl_mul(arg_struct arg)
        {
            int err = generic_rcl(arg, ref temp_v);

            if (err != ERR_NONE)
                return err;
            
            return generic_mul(temp_v, reg_x, docmd_rcl_mul_completion);
        }

        public static int docmd_rcl_sub(arg_struct arg)
        {
            vartype v = new vartype();
            vartype w = new vartype();
        
            int err = generic_rcl(arg, ref v);
            
            if (err != ERR_NONE)
	            return err;
            
            err = generic_sub(v, reg_x, ref w);
            free_vartype(ref v);
            
            if (err == ERR_NONE)
	            unary_result(w);
            
            return err;
        }

        public static int docmd_rcl_add(arg_struct arg)
        {
            vartype v = new vartype();
            vartype w = new vartype();

            int err = generic_rcl(arg, ref v);
            
            if (err != ERR_NONE)
	            return err;
            
            err = generic_add(v, reg_x, ref w);
            free_vartype(ref v);
            
            if (err == ERR_NONE)
	            unary_result(w);
            
            return err;
        }

        public static void docmd_rcl_div_completion(int error, vartype res)
        {
            free_vartype(ref temp_v);

            if (error == ERR_NONE)
                unary_result(res);
        }

        public static void docmd_rcl_mul_completion(int error, vartype res)
        {
            free_vartype(ref temp_v);

            if (error == ERR_NONE)
                unary_result(res);
        }

        public static int docmd_fix(arg_struct arg)
        {
            int err;
            int num = 0;
            
            err = arg_to_num(arg, ref num);
            
            if (err != ERR_NONE)
                return err;
            
            if (num > 11)
                return ERR_INVALID_DATA;
            
            flags.digits_bit3 = (byte)((num & 8) != 0 ? 1 : 0);
            flags.digits_bit2 = (byte)((num & 4) != 0 ? 1 : 0);
            flags.digits_bit1 = (byte)((num & 2) != 0 ? 1 : 0);
            flags.digits_bit0 = (byte)((num & 1) != 0 ? 1 : 0);
            flags.fix_or_all = 1;
            flags.eng_or_all = 0;
            
            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_sci(arg_struct arg)
        {
            int err;
            int num = 0;
            
            err = arg_to_num(arg, ref num);
            
            if (err != ERR_NONE)
                return err;
            
            if (num > 11)
                return ERR_INVALID_DATA;

            flags.digits_bit3 = (byte)((num & 8) != 0 ? 1 : 0);
            flags.digits_bit2 = (byte)((num & 4) != 0 ? 1 : 0);
            flags.digits_bit1 = (byte)((num & 2) != 0 ? 1 : 0);
            flags.digits_bit0 = (byte)((num & 1) != 0 ? 1 : 0);
            flags.fix_or_all = 0;
            flags.eng_or_all = 0;
            
            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_eng(arg_struct arg)
        {
            int err;
            int num = 0;

            err = arg_to_num(arg, ref num);
            
            if (err != ERR_NONE)
                return err;
            
            if (num > 11)
                return ERR_INVALID_DATA;

            flags.digits_bit3 = (byte)((num & 8) != 0 ? 1 : 0);
            flags.digits_bit2 = (byte)((num & 4) != 0 ? 1 : 0);
            flags.digits_bit1 = (byte)((num & 2) != 0 ? 1 : 0);
            flags.digits_bit0 = (byte)((num & 1) != 0 ? 1 : 0);
            flags.fix_or_all = 0;
            flags.eng_or_all = 1;
            
            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_all(arg_struct arg)
        {
            flags.digits_bit3 = 0;
            flags.digits_bit2 = 0;
            flags.digits_bit1 = 0;
            flags.digits_bit0 = 0;
            flags.fix_or_all = 1;
            flags.eng_or_all = 1;
            
            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_null(arg_struct arg)
        {
            return ERR_NONE;
        }

        public static int docmd_asto(arg_struct arg)
        {
            // I'm lazy enough to spot that ASTO has exactly the same side effects as
            // STO with the first 6 characters of ALPHA in ST X (as long as the
            // destination is not ST X or IND ST X). If you're also aware that
            // new_string() constructs a string of no more than 6 characters,
            // you'll see that this code does the job quite nicely.
            //
            vartype s = new_string(reg_alpha, reg_alpha_length);

            if (s == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            if (arg.type == ARGTYPE_STK && arg.val.stk == 'X')
            {
                // Special case for ASTO ST X
                free_vartype(ref reg_x);
                reg_x = s;

                return ERR_NONE;
            }
            else
            {
                int err;
                
                if (arg.type == ARGTYPE_IND_STK && arg.val.stk == 'X')
                {
                    // Special case for ASTO IND ST X
                    err = resolve_ind_arg(arg);
                    
                    if (err != ERR_NONE)
                    {
                        free_vartype(ref s);
                        return err;
                    }
                }

                vartype saved_x = reg_x;
                reg_x = s;
                err = docmd_sto(arg);
                free_vartype(ref s);
                reg_x = saved_x;

                return err;
            }
        }

        public static int docmd_arcl(arg_struct arg)
        {
            // Do some contortions to use docmd_rcl() to get the variable,
            // and do it without affecting the stack.
            //
            vartype saved_x = dup_vartype(reg_x);
            
            if (saved_x == null)
	            return ERR_INSUFFICIENT_MEMORY;

            int saved_nostacklift = flags.stack_lift_disable;
            int saved_trace = flags.trace_print;
            
            flags.stack_lift_disable = 1;
            flags.trace_print = 0;
            
            int err = docmd_rcl(arg);
            
            flags.stack_lift_disable = (byte)saved_nostacklift;
            flags.trace_print = (byte)saved_trace;
            
            vartype v;

            if (err != ERR_NONE)
            {
	            free_vartype(ref saved_x);
	            return err;
            }
            else
            {
	            v = reg_x;
	            reg_x = saved_x;
            }

            // Convert the variable to a string, using the same conversion
            // used when displaying the variable in 'normal' mode -- except
            // for strings, which we append with no quotes.
            //
            if (v.type == TYPE.STRING)
            {
	            vartype_string s = (vartype_string) v;
	            append_alpha_string(s.text, s.length, 0);
            }
            else
            {
	            char[] buf = new char[100];
	            int bufptr = vartype2string(v, buf, 100);
	            append_alpha_string(buf, bufptr, 0);
            }

            free_vartype(ref v);

            if (flags.alpha_mode != 0 && !program_running())
	            set_alpha_entry(true);
            
            if (flags.trace_print != 0 && flags.printer_exists != 0)
	            docmd_pra(null);
            
            return ERR_NONE;
        }

        public static int docmd_cla(arg_struct arg)
        {
            reg_alpha_length = 0;
            set_alpha_entry(false);

            return ERR_NONE;
        }

        public static int docmd_deg(arg_struct arg)
        {
            flags.grad = 0;
            flags.rad = 0;
            shell_annunciators(-1, -1, -1, -1, 0, 0);

            return ERR_NONE;
        }

        public static int docmd_rad(arg_struct arg)
        {
            flags.grad = 0;
            flags.rad = 1;
            shell_annunciators(-1, -1, -1, -1, 0, 1);
            
            return ERR_NONE;
        }

        public static int docmd_grad(arg_struct arg)
        {
            flags.grad = 1;
            flags.rad = 0;
            shell_annunciators(-1, -1, -1, -1, 1, 1);
            
            return ERR_NONE;
        }

        public static int docmd_rect(arg_struct arg)
        {
            flags.polar = 0;

            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_polar(arg_struct arg)
        {
            flags.polar = 1;

            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_size(arg_struct arg)
        {
            if (arg.type != ARGTYPE_NUM)
                return ERR_INVALID_TYPE;
            
            return dimension_array("REGS".ToCharArray(), 4, arg.val.num, 1);
        }

        public static int docmd_quiet(arg_struct arg)
        {
            flags.audio_enable = (byte)(flags.audio_enable == 0 ? 1 : 0);
            return ERR_NONE;
        }

        public static int docmd_cpxres(arg_struct arg)
        {
            flags.real_result_only = 0;
            return ERR_NONE;
        }

        public static int docmd_realres(arg_struct arg)
        {
            flags.real_result_only = 1;
            return ERR_NONE;
        }

        public static int docmd_keyasn(arg_struct arg)
        {
            flags.local_label = 0;
            return ERR_NONE;
        }

        public static int docmd_lclbl(arg_struct arg)
        {
            flags.local_label = 1;
            return ERR_NONE;
        }

        public static int docmd_rdxdot(arg_struct arg)
        {
            flags.decimal_point = 1;
            
            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);

            return ERR_NONE;
        }

        public static int docmd_rdxcomma(arg_struct arg)
        {
            flags.decimal_point = 0;

            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_clsigma(arg_struct arg)
        {
            vartype regs = recall_var("REGS".ToCharArray(), 4);
            vartype_realmatrix r;

            int first = mode_sigma_reg;
            int last = first + (flags.all_sigma != 0 ? 13 : 6);
            int size, i;
            
            if (regs == null)
                return ERR_SIZE_ERROR;
            
            if (regs.type != TYPE.REALMATRIX)
                return ERR_INVALID_TYPE;
            
            r = (vartype_realmatrix)regs;
            size = r.rows * r.columns;
            
            if (last > size)
                return ERR_SIZE_ERROR;
            
            for (i = first; i < last; i++)
            {
                r.array.is_string[i] = (char)0;
                r.array.data[i] = 0;
            }
            
            flags.log_fit_invalid = 0;
            flags.exp_fit_invalid = 0;
            flags.pwr_fit_invalid = 0;
            
            return ERR_NONE;
        }

        public static int docmd_clp(arg_struct arg)
        {
            return clear_prgm(arg);
        }

        public static int docmd_clv(arg_struct arg)
        {
            int err;
            
            if (arg.type == ARGTYPE_IND_NUM
                || arg.type == ARGTYPE_IND_STK
                || arg.type == ARGTYPE_IND_STR)
            {
                err = resolve_ind_arg(arg);
            
                if (err != ERR_NONE)
                    return err;
            }
            
            if (arg.type == ARGTYPE_STR)
            {
                // When EDITN is active, don't allow the matrix being
                // edited to be deleted.
                if (matedit_mode == 3 && arg.length == matedit_length)
                {
                    bool equal = true;
                    
                    for (int i = 0; i < arg.length; i++)
                        if (arg.val.text[i] != matedit_name[i])
                        {
                            equal = false;
                            break;
                        }
                    
                    if (equal)
                        return ERR_RESTRICTED_OPERATION;
                }
                
                purge_var(arg.val.text, arg.length);
                remove_shadow(arg.val.text, arg.length);
                
                return ERR_NONE;
            }
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_clst(arg_struct arg)
        {
            free_vartype(ref reg_x);
            free_vartype(ref reg_y);
            free_vartype(ref reg_z);
            free_vartype(ref reg_t);

            reg_x = new_real(0);
            reg_y = new_real(0);
            reg_z = new_real(0);
            reg_t = new_real(0);
            
            return ERR_NONE;
        }

        public static int docmd_clrg(arg_struct arg)
        {
            vartype regs = recall_var("REGS".ToCharArray(), 4);

            if (regs == null)
                return ERR_NONEXISTENT;

            if (regs.type == TYPE.REALMATRIX)
            {
                vartype_realmatrix rm;
                int sz, i;
                
                if (disentangle(regs) == 0)
                    return ERR_INSUFFICIENT_MEMORY;
                
                rm = (vartype_realmatrix)regs;
                sz = rm.rows * rm.columns;
                
                for (i = 0; i < sz; i++)
                    rm.array.data[i] = 0;
                
                for (i = 0; i < sz; i++)
                    rm.array.is_string[i] = (char)0;
                
                return ERR_NONE;
            }
            else if (regs.type == TYPE.COMPLEXMATRIX)
            {
                vartype_complexmatrix cm;
                int sz, i;
                
                if (disentangle(regs) == 0)
                    return ERR_INSUFFICIENT_MEMORY;
                
                cm = (vartype_complexmatrix)regs;
                sz = 2 * cm.rows * cm.columns;
                
                for (i = 0; i < sz; i++)
                    cm.array.data[i] = 0;
                
                return ERR_NONE;
            }
            else
            {
                // Should not happen; STO does not allow anything other
                // than a matrix to be stored in 'REGS'.
                //
                return ERR_INTERNAL_ERROR;
            }
        }

        public static int docmd_del(arg_struct arg)
        {
            if (arg.type != ARGTYPE_NUM)
                return ERR_INVALID_TYPE;

            clear_prgm_lines(arg.val.num);
            
            return ERR_NONE;
        }

        public static int docmd_clkeys(arg_struct arg)
        {
            clear_custom_menu();
            return ERR_NONE;
        }

        public static int docmd_cllcd(arg_struct arg)
        {
            clear_display();
            flush_display();
            flags.message = 1;
            flags.two_line_message = 1;
            
            return ERR_NONE;
        }

        public static int docmd_clmenu(arg_struct arg)
        {
            clear_prgm_menu();
            return ERR_NONE;
        }

        public static int docmd_to_deg(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.REALMATRIX)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_to_deg, null);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int mappable_to_deg(Phloat x, ref Phloat y)
        {
            Phloat r;
            int inf;
            
            r = rad_to_deg(x);
            
            if ((inf = Phloat.p_isinf(r)) != 0)
            {
                if (flags.range_error_ignore != 0)
                    r = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            y = r;
            
            return ERR_NONE;
        }

        public static int docmd_to_rad(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.REALMATRIX)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_to_rad, null);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int mappable_to_rad(Phloat x, ref Phloat y)
        {
            y = deg_to_rad(x);
            
            return ERR_NONE;
        }

        public static int docmd_to_hr(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.REALMATRIX)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_to_hr, null);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int mappable_to_hr(Phloat x, ref Phloat y)
        {
            bool neg = x < 0;

            Phloat res;
            Phloat point01 = new Phloat(1, 100);
            Phloat point36 = new Phloat(36, 100);

            if (neg)
	            x = -x;
    
            if (x == x + 1)
	            res = x;
            else
            {
	            if (x < point01)
		            res = x / point36;
	            else
                {
		            Phloat xh = Phloat.floor(x);
		            Phloat t = (x - xh) * 100;
                    Phloat xm = Phloat.floor(t);
		            Phloat xs = (t - xm) * 100;
		            
                    res = xh + xm / 60 + xs / 3600;
	            }
            }
            
            y = neg ? -res : res;
            
            return ERR_NONE;
        }

        public static int docmd_to_hms(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.REALMATRIX)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_to_hms, null);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int mappable_to_hms(Phloat x, ref Phloat y)
        {
            bool neg = x < 0;
            
            Phloat r, t;
            
            if (neg)
                x = -x;
            
            r = Phloat.floor(x);
            x = (x - r) * 60;
            t = Phloat.floor(x);
            r += t / 100 + (x - t) * 3 / 500;

            // Round-off may have caused the minutes or seconds to reach 60;
            // detect this and fix...
            //
            r = fix_hms(r);

            y = neg ? -r : r;
            
            return ERR_NONE;
        }

        public static int docmd_to_rec(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL)
            {
	            if (reg_y.type == TYPE.REAL || reg_y.type == TYPE.COMPLEX)
                {
	                // Note: the strange behavior re: real number in X, and
	                // complex number in Y, is for bug-compatibility with the
	                // real HP-42S. It's not very useful, but it doesn't really
	                // hurt either, I suppose. If I ever implement an "enhanced"
	                // Free42 mode, I'll probably make that combination of
	                // arguments return ERR_INVALID_TYPE, or at least offer the
	                // option of selecting that behavior.
	                //
	                Phloat r = ((vartype_real) reg_x).x;
	                Phloat phi = reg_y.type == TYPE.REAL ? ((vartype_real) reg_y).x : ((vartype_complex) reg_y).re;
	                
                    Phloat x = new Phloat();
                    Phloat y = new Phloat();
	                vartype vx, vy;
	                
                    generic_p2r(r, phi, ref x, ref y);

	                vx = new_real(x);

	                if (vx == null)
		                return ERR_INSUFFICIENT_MEMORY;
	                
                    vy = new_real(y);
	                
                    if (vy == null)
                    {
                        vx = null;
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                free_vartype(ref reg_y);
	                reg_y = vy;
	                free_vartype(ref reg_lastx);
	                reg_lastx = reg_x;
	                reg_x = vx;

	                if (flags.trace_print != 0 && flags.printer_exists != 0)
		                docmd_prx(null);

	                return ERR_NONE;
	            }
                else if (reg_y.type == TYPE.STRING)
	                return ERR_ALPHA_DATA_IS_INVALID;
                else
	                // The original HP-42S has a bug here: it accepts real and complex
	                // matrices, and also complex numbers, in Y, when X is real.
	                // I allow the Y-is-complex while X-is-real behavior (see above),
	                // but the Y-is-matrix case does not yield anything useful or even
	                // recognizable on the real HP-42S, so I feel I'm not going to
	                // break anything (that wasn't broken to begin with) by returning
	                // an error message here.
	                //
	                return ERR_INVALID_TYPE;
            }
            else if (reg_x.type == TYPE.COMPLEX)
            {
	            vartype_complex c = (vartype_complex) reg_x;
	            
                Phloat x = new Phloat();
                Phloat y = new Phloat();
	            vartype v;

	            generic_p2r(c.re, c.im, ref x, ref y);
	            v = new_complex(x, y);

	            if (v == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            
                unary_result(v);
	            
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
	            return ERR_ALPHA_DATA_IS_INVALID;
            else
	            return ERR_INVALID_TYPE;
        }

        public static int docmd_to_pol(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL)
            {
	            if (reg_y.type == TYPE.REAL || reg_y.type == TYPE.COMPLEX)
                {
	                // Note: the strange behavior re: real number in X, and
	                // complex number in Y, is for bug-compatibility with the
	                // real HP-42S. It's not very useful, but it doesn't really
	                // hurt either, I suppose. If I ever implement an "enhanced"
	                // Free42 mode, I'll probably make that combination of
	                // arguments return ERR_INVALID_TYPE, or at least offer the
	                // option of selecting that behavior.
	                ///
	                Phloat x = ((vartype_real) reg_x).x;
	                Phloat y = reg_y.type == TYPE.REAL ? ((vartype_real) reg_y).x : ((vartype_complex) reg_y).re;
	                
                    Phloat r = new Phloat();
                    Phloat phi = new Phloat();
                    vartype vx, vy;

	                generic_r2p(x, y, ref r, ref phi);
	                
                    if (Phloat.p_isinf(r) != 0)
                    {
		                if (flags.range_error_ignore != 0)
                            r = Phloat.POS_HUGE_PHLOAT;
		                else
		                    return ERR_OUT_OF_RANGE;
	                }

                    vx = new_real(r);
	                
                    if (vx == null)
		                return ERR_INSUFFICIENT_MEMORY;
	                
                    vy = new_real(phi);
	                
                    if (vy == null)
                    {
		                free_vartype(ref vx);
                        return ERR_INSUFFICIENT_MEMORY;
    	            }
	
                    free_vartype(ref reg_y);
	                reg_y = vy;
	                free_vartype(ref reg_lastx);
	                reg_lastx = reg_x;
	                reg_x = vx;

	                if (flags.trace_print != 0 && flags.printer_exists != 0)
		                docmd_prx(null);
	                
                    return ERR_NONE;
	            }
                else if (reg_y.type == TYPE.STRING)
	                return ERR_ALPHA_DATA_IS_INVALID;
                else
	                // The original HP-42S has a bug here: it accepts real and complex
	                // matrices, and also complex numbers, in Y, when X is real.
	                // I allow the Y-is-complex while X-is-real behavior (see above),
	                // but the Y-is-matrix case does not yield anything useful or even
	                // recognizable on the real HP-42S, so I feel I'm not going to
	                // break anything (that wasn't broken to begin with) by returning
	                // an error message here.
	                //
	                return ERR_INVALID_TYPE;
            }
            else if (reg_x.type == TYPE.COMPLEX)
            {
	            vartype_complex c = (vartype_complex) reg_x;
	            
                Phloat r = new Phloat();
                Phloat phi = new Phloat();
	            vartype v;
	            
                generic_r2p(c.re, c.im, ref r, ref phi);
	            
                if (Phloat.p_isinf(r) != 0)
                {
	                if (flags.range_error_ignore != 0)
                        r = Phloat.POS_HUGE_PHLOAT;
	                else
                        return ERR_OUT_OF_RANGE;
	            }

                v = new_complex(r, phi);
	        
                if (v == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            
                unary_result(v);
	            return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
	            return ERR_ALPHA_DATA_IS_INVALID;
            else
	            return ERR_INVALID_TYPE;
        }

        public static int docmd_ip(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.REALMATRIX)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_ip, null);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int mappable_ip(Phloat x, ref Phloat y)
        {
            if (x < 0)
                y = -Phloat.floor(-x);
            else
                y = Phloat.floor(x);
            
            return ERR_NONE;
        }

        public static int docmd_fp(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.REALMATRIX)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_fp, null);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int mappable_fp(Phloat x, ref Phloat y)
        {
            if (x < 0)
                y = x + Phloat.floor(-x);
            else
                y = x - Phloat.floor(x);
            
            return ERR_NONE;
        }

        public static int docmd_rnd(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
            {
                vartype v = new vartype();

                int err;
                int digits = 0;
                
                if (flags.digits_bit3 != 0) digits += 8;
                if (flags.digits_bit2 != 0) digits += 4;
                if (flags.digits_bit1 != 0) digits += 2;
                if (flags.digits_bit0 != 0) digits += 1;
                
                rnd_multiplier = Phloat.pow(10, digits);

                err = map_unary(reg_x, ref v, mappable_rnd_r, mappable_rnd_c);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
        }

        public static int mappable_rnd_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            int err = mappable_rnd_r(xre, ref yre);

            if (err != ERR_NONE)
                return err;
            
            return mappable_rnd_r(xim, ref yim);
        }

        public static int mappable_rnd_r(Phloat x, ref Phloat y)
        {
            if (flags.fix_or_all != 0)
            {
                if (flags.eng_or_all != 0)
                    y = x;
                else
                {
                    Phloat t = new Phloat(x);

                    bool neg = t < 0;
                    
                    if (neg)
                        t = -t;
                    
                    if (t > 1e20)
                        y = x;
                    else
                    {
                        t = Phloat.floor(t * rnd_multiplier + 0.5) / rnd_multiplier;
                        y = neg ? -t : t;
                    }
                }

                return ERR_NONE;
            }
            else
            {
                Phloat t = new Phloat(x);

                int neg;
                Phloat scale = new Phloat();
                
                if (t == 0)
                    y = 0;
                else
                {
                    if (t < 0)
                    {
                        t = -t;
                        neg = 1;
                    }
                    else
                        neg = 0;
                    
                    scale = Phloat.pow(10, Phloat.floor(Phloat.log10(t)));
                    
                    if (scale > t)
                    {
                        // Theoretically, this can't happen, but due to limited
                        // precision, the log of something like 9.999999999+
                        // might be computed as 1 instead of 0.99999999+,
                        // hence the extra check.
                        //
                        scale /= 10;
                    }

                    t = Phloat.floor(t / scale * rnd_multiplier + 0.5) / rnd_multiplier * scale;

                    if (Phloat.p_isinf(t) != 0)
                    {
                        if (flags.range_error_ignore != 0)
                            y = neg != 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                        else
                            return ERR_OUT_OF_RANGE;
                    }
                    else
                        y = neg != 0 ? -t : t;
                }

                return ERR_NONE;
            }
        }

        public static int docmd_abs(arg_struct arg)
        {
            switch (reg_x.type)
            {
                case TYPE.REAL:
                    {
                        vartype r = new vartype();

                        Phloat x = ((vartype_real)reg_x).x;
                        
                        if (x < 0)
                            x = -x;

                        r = new_real(x);
                        
                        if (r == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        
                        unary_result(r);
                        
                        return ERR_NONE;
                    }

                case TYPE.COMPLEX:
                    {
                        vartype r = new vartype();

                        Phloat a = Phloat.hypot(((vartype_complex)reg_x).re, ((vartype_complex)reg_x).im);

                        if (Phloat.p_isinf(a) == 0)
                            r = new_real(a);
                        else if (flags.range_error_ignore != 0)
                            r = new_real(Phloat.POS_HUGE_PHLOAT);
                        else
                            return ERR_OUT_OF_RANGE;

                        if (r == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        
                        unary_result(r);
                        
                        return ERR_NONE;
                    }

                case TYPE.STRING:
                    return ERR_ALPHA_DATA_IS_INVALID;

                case TYPE.REALMATRIX:
                    {
                        if (is_pure_real(reg_x) == 0)
                            return ERR_ALPHA_DATA_IS_INVALID;

                        vartype_realmatrix src = new vartype_realmatrix();
                        vartype_realmatrix dst = new vartype_realmatrix();
                        
                        int size;
                        int i;

                        src = (vartype_realmatrix) reg_x;
                        dst = (vartype_realmatrix) new_realmatrix(src.rows, src.columns);

                        if (dst == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        
                        size = src.rows * src.columns;
                        
                        for (i = 0; i < size; i++)
                        {
                            Phloat x = src.array.data[i];

                            if (x < 0)
                                x = -x;
                            
                            dst.array.data[i] = x;
                        }

                        unary_result((vartype)dst);
                        
                        return ERR_NONE;
                    }

                case TYPE.COMPLEXMATRIX:
                    return ERR_INVALID_TYPE;
                
                default:
                    return ERR_INTERNAL_ERROR;
            }
        }

        public static int docmd_sign(arg_struct arg)
        {
            switch (reg_x.type)
            {
                case TYPE.REAL:
                    {
                        // Note that this implementation has sign(0) = 1, which is not
                        // how most programming languages handle zero. This is a holdover
                        // from the HP-41C, where SIGN did double duty for recognizing
                        // strings: they would return zero and could be recognized in this
                        // way.
                        // The HP-42S has the STR? function for recognizing strings, but
                        // it still supports the HP-41C's SIGN behavior for compatibility.
                        // And so does Free42, of course.
                        //
                        vartype r = new_real(((vartype_real)reg_x).x < 0 ? -1 : 1);

                        if (r == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        
                        unary_result(r);
                        
                        return ERR_NONE;
                    }

                case TYPE.STRING:
                    {
                        vartype r = new_real(0);

                        if (r == null)
                            return ERR_INSUFFICIENT_MEMORY;

                        unary_result(r);
                        
                        return ERR_NONE;
                    }

                case TYPE.REALMATRIX:
                    {
                        // Can't use the mapper here because it won't handle strings.
                        vartype_realmatrix src = new vartype_realmatrix();
                        vartype_realmatrix dst = new vartype_realmatrix();

                        int size;
                        int i;
                        
                        src = (vartype_realmatrix) reg_x;
                        dst = (vartype_realmatrix) new_realmatrix(src.rows, src.columns);

                        if (dst == null)
                            return ERR_INSUFFICIENT_MEMORY;

                        size = src.rows * src.columns;
                        
                        for (i = 0; i < size; i++)
                        {
                            if (src.array.is_string[i] != 0)
                                dst.array.data[i] = 0;
                            else
                                dst.array.data[i] = src.array.data[i] < 0 ? -1 : 1;
                        }

                        unary_result((vartype)dst);
                        
                        return ERR_NONE;
                    }

                case TYPE.COMPLEX:
                case TYPE.COMPLEXMATRIX:
                    {
                        vartype v = new vartype();

                        map_unary(reg_x, ref v, null, mappable_sign);
                        unary_result((vartype)v);
                        
                        return ERR_NONE;
                    }

                default:
                    return ERR_INTERNAL_ERROR;
            }
        }

        public static int mappable_sign(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            Phloat h = Phloat.hypot(xre, xim);

            if (h == 0)
            {
                yre = 0;
                yim = 0;
            }
            else if (Phloat.p_isinf(h) == 0)
            {
                yre = xre / h;
                yim = xim / h;
            }
            else
            {
                xre /= 10000;
                xim /= 10000;
                h = Phloat.hypot(xre, xim);
                yre = xre / h;
                yim = xim / h;
            }

            return ERR_NONE;
        }

        public static int docmd_mod(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL && reg_y.type == TYPE.REAL)
            {
                Phloat x = ((vartype_real)reg_x).x;
                Phloat y = ((vartype_real)reg_y).x;

                Phloat res = new Phloat();
                vartype v = new vartype();

                if (x == 0)
                    res = y;
                else if (y == 0)
                    res = 0;
                else
                {
                    res = Phloat.fmod(y, x);

                    if (res != 0 && ((x > 0 && y < 0) || (x < 0 && y > 0)))
                        res += x;
                }

                v = new_real(res);
                
                if (v == null)
                    return ERR_INSUFFICIENT_MEMORY;

                binary_result(v);
                
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;
            else if (reg_y.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_clall(arg_struct arg)
        {
            vartype regs;

            // Clear all registers
            free_vartype(ref reg_x);
            free_vartype(ref reg_y);
            free_vartype(ref reg_z);
            free_vartype(ref reg_t);
            free_vartype(ref reg_lastx);

            reg_x = new_real(0);
            reg_y = new_real(0);
            reg_z = new_real(0);
            reg_t = new_real(0);
            reg_lastx = new_real(0);
            reg_alpha_length = 0;

            // Exit all menus (even leaving the matrix editor
            // is guaranteed not to fail because there's 0 in X,
            // and that is always valid).
            //
            set_menu(MENULEVEL_APP, (int)MENU.NONE);
            flags.prgm_mode = 0;

            // Clear all programs and variables
            clear_all_prgms();
            goto_dot_dot();
            purge_all_vars();

            regs = new_realmatrix(25, 1);
            store_var("REGS".ToCharArray(), 4, regs);

            // Clear the CUSTOM and programmable menus
            clear_custom_menu();
            clear_prgm_menu();

            return ERR_NONE;
        }
    }
}
