﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        //*******************************************************
        // Implementations of HP-42S built-in functions, part 2
        //*******************************************************

        public static char[] virtual_flags = new char[100]
        {
            '0', '0', '0', '0', '0', '0', '0', '0', '0', '0',
            '0', '0', '0', '0', '0', '0', '0', '0', '0', '0',
            '0', '0', '0', '0', '0', '0', '0', '1', '0', '0',
            '0', '0', '0', '0', '0', '0', '0', '0', '0', '0',
            '0', '0', '0', '0', '0', '1', '1', '1', '0', '1',

            '0', '0', '0', '1', '0', '0', '0', '0', '0', '0',
            '0', '0', '0', '0', '0', '1', '0', '0', '0', '0',
            '0', '0', '0', '0', '0', '1', '0', '0', '0', '0',
            '0', '0', '0', '0', '0', '0', '0', '0', '0', '0',
            '0', '0', '0', '0', '0', '0', '0', '0', '0', '0'
        };

        public static int docmd_sf(arg_struct arg)
        {
            int err;
            int num = 0;

            err = arg_to_num(arg, ref num);
            
            if (err != ERR_NONE)
	            return err;
            
            if (num >= 100)
	            return ERR_NONEXISTENT;
            
            if (num >= 36 && num <= 80)
	            return ERR_RESTRICTED_OPERATION;
            
            if (virtual_flags[num] == '1')
	            return virtual_flag_handler(FLAGOP_SF, num);
            else
            {
	            //flags.farray[num] = 1;
                set_flag(num, 1);

                // This is the stack_lift_disable flag.
                // Since we automatically enable stack lift after every command,
                // unless mode_disable_stack_lift is set, we must intervene for
                // this to actually have the intended effect.
                //
                if (num == 30)
	                mode_disable_stack_lift = true;
	            
                return ERR_NONE;
            }
        }

        public static int docmd_cf(arg_struct arg)
        {
            int err;
            int num = 0;

            err = arg_to_num(arg, ref num);
            
            if (err != ERR_NONE)
                return err;
            
            if (num >= 100)
                return ERR_NONEXISTENT;
            
            if (num >= 36 && num <= 80)
                return ERR_RESTRICTED_OPERATION;
            
            if (virtual_flags[num] == '1')
                return virtual_flag_handler(FLAGOP_CF, num);
            else
            {
                //flags.farray[num] = 0;
                set_flag(num, 0);

                return ERR_NONE;
            }
        }

        public static int docmd_prx(arg_struct arg)
        {
            // arg == null if we're called to do TRACE mode auto-print
            if (arg != null && (flags.printer_enable == 0) && program_running())
                return ERR_NONE;

            if (flags.printer_exists == 0)
                return ERR_PRINTING_IS_DISABLED;
            else
            {
                char[] buf = new char[100];
                int len;

                shell_annunciators(-1, -1, 0, -1, -1, -1);

                len = vartype2string(reg_x, buf, 100);

                if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.STRING)
                    print_right(buf, len, "***".ToCharArray(), 3);
                else
                {
                    // Normally we print X right-justified, but if it doesn't fit on
                    // one line, we print it left-justified, because having the excess
                    // go near the right margin looks weird and confusing.
                    //
                    int left = (len > ((flags.double_wide_print != 0) ? 12 : 24)) ? 1 : 0;

                    print_lines(buf, len, left);
                }

                if (arg != null && (reg_x.type == TYPE.REALMATRIX || reg_x.type == TYPE.COMPLEXMATRIX))
                {
                    prv_var = reg_x;
                    prv_index = 0;

                    mode_interruptible = prv_worker;
                    mode_stoppable = true;

                    return ERR_INTERRUPTIBLE;
                }
                else
                {
                    shell_annunciators(-1, -1, 0, -1, -1, -1);

                    return ERR_NONE;
                }
            }
        }

        public static int docmd_pra(arg_struct arg)
        {
            char[] tmp;

            // arg == NULL if we're called to do TRACE mode auto-print
            if (arg != null && flags.printer_enable == 0 && program_running())
                return ERR_NONE;

            if (flags.printer_exists == 0)
                return ERR_PRINTING_IS_DISABLED;

            shell_annunciators(-1, -1, 1, -1, -1, -1);
            
            if (reg_alpha_length == 0)
                print_text(null, 0, 1);
            else
            {
                int line_start = 0;
                int width = flags.double_wide_print != 0 ? 12 : 24;

                for (int i = 0; i < reg_alpha_length; i++)
                {
                    if (reg_alpha[i] == 10)
                    {
                        tmp = new char[i - line_start];
                        Array.Copy(reg_alpha, line_start, tmp, 0, i - line_start);
                        print_text(tmp, i - line_start, 1);

                        line_start = i + 1;
                    }
                    else if (i == line_start + width)
                    {
                        tmp = new char[i - line_start];
                        Array.Copy(reg_alpha, line_start, tmp, 0, i - line_start);
                        print_text(tmp, i - line_start, 1);
                        
                        line_start = i;
                    }
                }

                if (line_start < reg_alpha_length || (line_start > 0 && reg_alpha[line_start - 1] == 10))
                {
                    tmp = new char[reg_alpha_length - line_start];
                    Array.Copy(reg_alpha, line_start, tmp, 0, reg_alpha_length - line_start);
                    print_text(tmp, reg_alpha_length - line_start, 1);
                }
            }

            shell_annunciators(-1, -1, 0, -1, -1, -1);
            return ERR_NONE;
        }

        public static int docmd_gto(arg_struct arg)
        {
            int running = program_running() ? 1 : 0;

            if (running == 0)
                clear_all_rtns();

            if (arg.type == ARGTYPE_NUM || arg.type == ARGTYPE_LCLBL)
            {
                if (running == 0 || arg.target == -1)
                    arg.target = find_local_label(arg);
                if (arg.target == -2)
                    return ERR_LABEL_NOT_FOUND;
                else
                {
                    pc = arg.target;
                    prgm_highlight_row = 1;
                    return ERR_NONE;
                }
            }

            if (arg.type == ARGTYPE_STR)
            {
                int new_prgm = 0;
                int new_pc = 0;

                if (find_global_label(arg, ref new_prgm, ref new_pc) != 0)
                {
                    current_prgm = new_prgm;
                    pc = new_pc;
                    prgm_highlight_row = 1;
                    return ERR_NONE;
                }
                else
                    return ERR_LABEL_NOT_FOUND;
            }

            if (arg.type == ARGTYPE_IND_NUM || arg.type == ARGTYPE_IND_STK || arg.type == ARGTYPE_IND_STR)
            {
                int err = resolve_ind_arg(arg);
                
                if (err != ERR_NONE)
                    return err;
                
                if (arg.type == ARGTYPE_NUM)
                {
                    int target_pc = find_local_label(arg);
                    
                    if (target_pc == -2)
                        return ERR_LABEL_NOT_FOUND;
                    else
                    {
                        pc = target_pc;
                        prgm_highlight_row = 1;
                        return ERR_NONE;
                    }
                }
                else
                {
                    int newprgm = 0;
                    int newpc = 0;
                    
                    if (find_global_label(arg, ref newprgm, ref newpc) != 0)
                    {
                        current_prgm = newprgm;
                        pc = newpc;
                        prgm_highlight_row = 1;
                        return ERR_NONE;
                    }
                    else
                        return ERR_LABEL_NOT_FOUND;
                }
            }

            if (arg.type == ARGTYPE_LBLINDEX)
            {
                int labelindex = arg.val.num;
                current_prgm = labels[labelindex].prgm;
                pc = labels[labelindex].pc;
                prgm_highlight_row = 1;
                return ERR_NONE;
            }

            return ERR_INTERNAL_ERROR;
        }

        public static int view_helper(arg_struct arg, int print)
        {
            int err;
            char[] buf = new char[22];
            int bufptr = 0;
            int part2;

            vartype v = new vartype();
            
            if (arg.type == ARGTYPE_IND_NUM || arg.type == ARGTYPE_IND_STK || arg.type == ARGTYPE_IND_STR)
            {
    	        err = resolve_ind_arg(arg);
	
                if (err != ERR_NONE)
	                return err;
            }

            switch (arg.type)
            {
    	        case ARGTYPE_NUM:
	                int num = arg.val.num;
	                char2buf(buf, 22, ref bufptr, 'R');
	                
                    if (num < 10)
		                char2buf(buf, 22, ref bufptr, '0');
	                
                    bufptr += int2string(num, buf, bufptr, 22);
	                break;
	        
	            case ARGTYPE_STK:
	                string2buf(buf, 22, ref bufptr, "ST ".ToCharArray(), 3);
	                char2buf(buf, 22, ref bufptr, arg.val.stk);
	                break;

	            case ARGTYPE_STR:
	                string2buf(buf, 22, ref bufptr, arg.val.text, arg.length);
	                break;
            }
            
            char2buf(buf, 22, ref bufptr, '=');
            
            part2 = bufptr;
            err = generic_rcl(arg, ref v);
            
            if (err != ERR_NONE)
	            return err;
            
            bufptr += vartype2string(v, buf, bufptr, 22);
            
            free_vartype(ref v);
            
            clear_row(0);
            draw_string(0, 0, buf, bufptr);
            flush_display();
            
            flags.message = 1;
            flags.two_line_message = 0;

            if (print != 0 && (flags.printer_enable != 0 || !program_running()))
            {
                if (flags.printer_exists != 0)
                {
                    char[] tmp = new char[bufptr - part2];
                    Array.Copy(buf, part2, tmp, 0, bufptr - part2);
                    print_wide(buf, part2, tmp, bufptr - part2);
                }
                else
                    return ERR_STOP;
            }

            return ERR_NONE;
        }

        public static int docmd_fs_t(arg_struct arg)
        {
            int err;
            int num = 0;

            err = arg_to_num(arg, ref num);
            
            if (err != ERR_NONE)
                return err;
            
            if (num >= 100)
                return ERR_NONEXISTENT;
            
            if (virtual_flags[num] == '1')
                return virtual_flag_handler(FLAGOP_FS_T, num);
            else
                return get_flag(num) != 0 ? ERR_YES : ERR_NO;
        }

        public static int docmd_fc_t(arg_struct arg)
        {
            int err;
            int num = 0;

            err = arg_to_num(arg, ref num);

            if (err != ERR_NONE)
                return err;
            
            if (num >= 100)
                return ERR_NONEXISTENT;
            
            if (virtual_flags[num] == '1')
                return virtual_flag_handler(FLAGOP_FC_T, num);
            else
                return get_flag(num) != 0 ? ERR_NO : ERR_YES;
        }

        public static int docmd_fsc_t(arg_struct arg)
        {
            int err;
            int num = 0;

            err = arg_to_num(arg, ref num);

            if (err != ERR_NONE)
                return err;
            
            if (num >= 100)
                return ERR_NONEXISTENT;
            
            if (num >= 36 && num <= 80)
                return ERR_RESTRICTED_OPERATION;
            
            if (virtual_flags[num] == '1')
                return virtual_flag_handler(FLAGOP_FSC_T, num);
            else
            {
                err = get_flag(num) != 0 ? ERR_YES : ERR_NO;
                set_flag(num, 0);

                return err;
            }
        }

        public static int docmd_fcc_t(arg_struct arg)
        {
            int err;
            int num = 0;

            err = arg_to_num(arg, ref num);

            if (err != ERR_NONE)
                return err;
            
            if (num >= 100)
                return ERR_NONEXISTENT;
            
            if (num >= 36 && num <= 80)
                return ERR_RESTRICTED_OPERATION;
            
            if (virtual_flags[num] == '1')
                return virtual_flag_handler(FLAGOP_FCC_T, num);
            else
            {
                err = get_flag(num) != 0 ? ERR_NO : ERR_YES;
                set_flag(num, 0);

                return err;
            }
        }

        public static int docmd_comb(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL && reg_y.type == TYPE.REAL)
            {
                Phloat y = ((vartype_real)reg_y).x;
                Phloat x = ((vartype_real)reg_x).x;

                Phloat r = new Phloat(1);
                Phloat q = new Phloat(1);
                vartype v = new vartype();

                if (x < 0 || x != Phloat.floor(x) || x == x - 1 || y < 0 || y != Phloat.floor(y))
                    return ERR_INVALID_DATA;
                
                if (y < x)
                    return ERR_INVALID_DATA;
                
                if (x > y / 2)
                    x = y - x;
                
                while (q <= x)
                {
                    r *= y--;
                    if (Phloat.p_isinf(r) != 0)
                    {
                        if (flags.range_error_ignore != 0)
                        {
                            r = Phloat.POS_HUGE_PHLOAT;
                            break;
                        }
                        else
                            return ERR_OUT_OF_RANGE;
                    }

                    r /= q++;
                }

                v = new_real(r);
                
                if (v == null)
                    return ERR_INSUFFICIENT_MEMORY;
                
                binary_result(v);
                
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;
            else if (reg_y.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_perm(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL && reg_y.type == TYPE.REAL)
            {
                Phloat y = ((vartype_real)reg_y).x;
                Phloat x = ((vartype_real)reg_x).x;

                Phloat r = new Phloat(1);
                vartype v = new vartype();

                if (x < 0 || x != Phloat.floor(x) || x == x - 1 || y < 0 || y != Phloat.floor(y))
                    return ERR_INVALID_DATA;
                
                if (y < x)
                    return ERR_INVALID_DATA;
                
                while (x > 0)
                {
                    r *= y--;
                    if (Phloat.p_isinf(r) != 0)
                    {
                        if (flags.range_error_ignore != 0)
                        {
                            r = Phloat.POS_HUGE_PHLOAT;
                            break;
                        }
                        else
                            return ERR_OUT_OF_RANGE;
                    }

                    x--;
                }

                v = new_real(r);
                
                if (v == null)
                    return ERR_INSUFFICIENT_MEMORY;
                
                binary_result(v);
                
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;
            else if (reg_y.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_fact(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.REALMATRIX)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_fact, null);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int mappable_fact(Phloat x, ref Phloat y)
        {
            Phloat f = new Phloat(1);

            if (x < 0 || x != Phloat.floor(x))
                return ERR_INVALID_DATA;
            
            while (x > 1)
            {
                f *= x--;
                if (Phloat.p_isinf(f) != 0)
                {
                    if (flags.range_error_ignore != 0)
                    {
                        y = Phloat.POS_HUGE_PHLOAT;
                        return ERR_NONE;
                    }
                    else
                        return ERR_OUT_OF_RANGE;
                }
            }

            y = f;
            
            return ERR_NONE;
        }

        public static int docmd_gamma(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type == TYPE.COMPLEX || reg_x.type == TYPE.COMPLEXMATRIX)
                return ERR_INVALID_TYPE;
            else
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, math_gamma, null);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
        }

        public static int docmd_ran(arg_struct arg)
        {
            vartype v = new_real(math_random());

            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            recall_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_seed(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL)
            {
                Phloat x = ((vartype_real)reg_x).x;

                int i;
                
                if (x == 0)
                    random_number = shell_random_seed();
                else
                    random_number = x - Phloat.floor(x);

                for (i = 0; i < 10; i++)
                    math_random();

                return ERR_NONE;
            }
            else if (arg.type == TYPE_STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_lbl(arg_struct arg)
        {
            if (flags.message == 0)
                fly_goose();
            
            return ERR_NONE;
        }

        public static int docmd_rtn(arg_struct arg)
        {
            if (program_running())
            {
                int newprgm = 0;
                int newpc = 0;

                pop_rtn_addr(ref newprgm, ref newpc);
                
                if (newprgm == -3)
                    return return_to_integ(0);
                else if (newprgm == -2)
                    return return_to_solve(0);
                else if (newprgm == -1)
                {
                    if (pc >= prgms[current_prgm].size)
                        pc = -1;            // It's an END; go to line 0

                    return ERR_STOP;
                }
                else
                {
                    current_prgm = newprgm;
                    pc = newpc;
                    
                    return ERR_NONE;
                }
            }
            else
            {
                clear_all_rtns();
                pc = -1;
                
                return ERR_NONE;
            }
        }

        public static int docmd_input(arg_struct arg)
        {
            vartype v = new vartype();
            int err;

            if (arg.type == ARGTYPE_IND_NUM || arg.type == ARGTYPE_IND_STK || arg.type == ARGTYPE_IND_STR)
            {
                err = resolve_ind_arg(arg);

                if (err != ERR_NONE)
                    return err;
            }

            err = generic_rcl(arg, ref v);

            if (err == ERR_NONEXISTENT)
            {
                v = new_real(0);
            
                if (v == null)
                    return ERR_INSUFFICIENT_MEMORY;
            }
            else if (err != ERR_NONE)
                return err;

            input_length = 0;
            input_arg = arg;

            if (arg.type == ARGTYPE_NUM)
            {
                char2buf(input_name, 11, ref input_length, 'R');
                
                if (arg.val.num < 10)
                    char2buf(input_name, 11, ref input_length, '0');
                
                input_length += int2string(arg.val.num, input_name, input_length, 11);
            }
            else if (arg.type == ARGTYPE_STK)
            {
                string2buf(input_name, 11, ref input_length, "ST ".ToCharArray(), 3);
                char2buf(input_name, 11, ref input_length, arg.val.stk);
            }
            else if (arg.type == ARGTYPE_STR)
            {
                string2buf(input_name, 11, ref input_length, arg.val.text, arg.length);
            }
            else
            {
                v = null;
                return ERR_INVALID_TYPE;
            }

            recall_result(v);
            
            return ERR_STOP;
        }

        public static int docmd_view(arg_struct arg)
        {
            return view_helper(arg, 1);
        }

        public static int docmd_aview(arg_struct arg)
        {
            aview_helper();

            if (flags.printer_enable != 0 || !program_running())
            {
                if (flags.printer_exists != 0)
                    docmd_pra(arg);
                else
                    return ERR_STOP;
            }

            return ERR_NONE;
        }

        public static void aview_helper()
        {
            const int DISP_ROWS = 2;
            const int DISP_COLUMNS = 22;

            int[] line_start = new int[DISP_ROWS];
            int[] line_length = new int[DISP_ROWS];

            int line = 0;
            int i;
            
            line_start[0] = 0;
            
            for (i = 0; i < reg_alpha_length; i++)
            {
	            if (reg_alpha[i] == 10)
                {
	                if (line == DISP_ROWS - 1)
		                break;
	            
                    line_length[line] = i - line_start[line];
                    line_start[++line] = i + 1;
    	        }
                else if (i == line_start[line] + DISP_COLUMNS)
                {
	                if (line == DISP_ROWS - 1)
		                break;
	            
                    line_length[line] = i - line_start[line];
	                line_start[++line] = i;
	            }
            }

            line_length[line] = i - line_start[line];
            
            flags.message = 1;
            flags.two_line_message = (byte)(line == 1 ? 1 : 0);

            clear_row(0);
            
            if (flags.two_line_message != 0 || program_running())
	            clear_row(1);

            for (i = 0; i <= line; i++)
            {
                string tmp = new string(reg_alpha);
                draw_string(0, i, tmp.Substring(line_start[i], line_length[i]));
            }
            
            flush_display();
        }

        public static int docmd_xeq(arg_struct arg)
        {
            if (program_running())
            {
                int oldprgm = current_prgm;
                int oldpc = pc;
                int error = docmd_gto(arg);

                if (error != ERR_NONE)
                    return error;
                
                return push_rtn_addr(oldprgm, oldpc);
            }
            else
            {
                int err = docmd_gto(arg);
                
                if (err != ERR_NONE)
                    return err;
                
                clear_all_rtns();
                return ERR_RUN;
            }
        }

        public static int docmd_prompt(arg_struct arg)
        {
            aview_helper();

            if (flags.printer_enable != 0 && flags.printer_exists != 0 && (flags.trace_print != 0 || flags.normal_print != 0))
                docmd_pra(arg);

            return ERR_STOP;
        }

        public static int docmd_pse(arg_struct arg)
        {
            if (program_running())
            {
	            int saved_command = pending_command;
	            pending_command = CMD_NONE;

	            redisplay();
	            
                pending_command = saved_command;
	            mode_pause = true;
            }

            return ERR_NONE;
        }

        public static int docmd_isg(arg_struct arg)
        {
            return generic_loop(arg, true);
        }

        public static int generic_loop(arg_struct arg, bool isg)
        {
            int err;

            if (arg.type == ARGTYPE_IND_NUM || arg.type == ARGTYPE_IND_STK || arg.type == ARGTYPE_IND_STR)
            {
                err = resolve_ind_arg(arg);

                if (err != ERR_NONE)
                    return err;
            }

            switch (arg.type)
            {
                case ARGTYPE_NUM:
                    {
                        vartype regs = recall_var("REGS".ToCharArray(), 4);

                        if (regs == null)
                            return ERR_SIZE_ERROR;
                        else if (regs.type == TYPE.REALMATRIX)
                        {
                            vartype_realmatrix rm = (vartype_realmatrix)regs;

                            int size = rm.rows * rm.columns;
                            int index = arg.val.num;
                            
                            if (index >= size)
                                return ERR_SIZE_ERROR;
                            
                            if (rm.array.is_string[index] != 0)
                                return ERR_ALPHA_DATA_IS_INVALID;
                            else
                            {
                                if (disentangle(regs) == 0)
                                    return ERR_INSUFFICIENT_MEMORY;
                                
                                return generic_loop_helper(ref rm.array.data[index], isg);
                            }
                        }
                        else if (regs.type == TYPE.COMPLEXMATRIX)
                        {
                            return ERR_INVALID_TYPE;
                        }
                        else
                        {
                            // This should never happen; STO should prevent
                            // "REGS" from being any other type than a real or
                            // complex matrix.
                            //
                            return ERR_INTERNAL_ERROR;
                        }
                    }

                case ARGTYPE_STK:
                    {
                        vartype v = new vartype();

                        switch (arg.val.stk)
                        {
                            case 'X': v = reg_x; break;
                            case 'Y': v = reg_y; break;
                            case 'Z': v = reg_z; break;
                            case 'T': v = reg_t; break;
                            case 'L': v = reg_lastx; break;
                        }

                        if (v.type == TYPE.REAL)
                            return generic_loop_helper(ref ((vartype_real)v).x, isg);
                        else if (v.type == TYPE.STRING)
                            return ERR_ALPHA_DATA_IS_INVALID;
                        else
                            return ERR_INVALID_TYPE;
                    }

                case ARGTYPE_STR:
                    {
                        vartype v = recall_var(arg.val.text, arg.length);

                        if (v.type == TYPE.REAL)
                            return generic_loop_helper(ref ((vartype_real)v).x, isg);
                        else if (v.type == TYPE.STRING)
                            return ERR_ALPHA_DATA_IS_INVALID;
                        else
                            return ERR_INVALID_TYPE;
                    }

                default:
                    return ERR_INTERNAL_ERROR;
            }
        }

        public static int generic_loop_helper(ref Phloat x, bool isg)
        {
            Phloat t;
            long i, j, k;
            int s;
            
            if (x == x + 1)
            {
	            // Too big to do anything useful with; this is what the real
	            // HP-42S does in this case:
	            //
	            return isg == (x < 0) ? ERR_YES : ERR_NO;
            }

            // Break number up as follows: II.JJJKKRRRRR
            // The sign goes with I; everything else is considered positive.
            //
            t = x;

            if (t < 0)
            {
	            t = -t;
	            s = -1;
            }
            else
	            s = 1;

            i = Phloat.to_int8(t);
            t = (t - i) * 100000;

            // The 0.0000005 is a precaution to prevent the loop increment
            // value from being taken to be 1 lower than what the user intended;
            // this can happen because the decimal fractions used here cannot,
            // in general, be represented exactly in binary, so that what should
            // be 10.00902 may actually end up being approximated as something
            // fractionally lower -- and 10.0090199999999+ would be interpreted
            // as having a loop increment of 1, not the 2 that was intended.
            // By adding 0.0000005 before truncating, we effectively round to
            // 7 decimals, which is all that a real HP-42S would have left after
            // the multiplication by 100000. So, we sacrifice some of the range
            // of an IEEE-754 double, but maintain HP-42S compatibility.
            //
            k = Phloat.to_int8(t);
            j = k / 100;
            k -= j * 100;

            if (k == 0)
	            k = 1;

            // Update the 'real' loop control value separately from the components
            // we have just separated out. I'm very paranoid about cumulative errors,
            // so I don't rebuild the loop control value from i, j, k, etc.
            // This way is computationally cheaper, anyway.
            //
            if (isg)
            {
	            if (x < 0 && x > -k)
	                x = -(x) + k - 2 * i;
	            else
	                x += k;
            }
            else
            {
	            if (x > 0 && x < k)
	                x = -(x) - k + 2 * i;
	            else
	                x -= k;
            }

            // Now we do what you would expect ISG/DSE to do...
            if (isg)
            {
	            if (s == -1)
	                i = k - i;
	            else
	                i = k + i;
	            
                return i > j ? ERR_NO : ERR_YES;
            }
            else
            {
	            if (s == -1)
	                i = -i - k;
	            else
	                i = i - k;
	            
                return i <= j ? ERR_NO : ERR_YES;
            }
        }

        public static int docmd_dse(arg_struct arg)
        {
            return generic_loop(arg, false);
        }

        public static int docmd_aip(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL)
            {
	            char[] buf = new char[44];

	            int size = ip2revstring(((vartype_real) reg_x).x, buf, 44);
	            append_alpha_string(buf, size, 1);
	        
                if (flags.trace_print != 0 && flags.printer_exists != 0)
	                docmd_pra(null);
	        
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
	            return ERR_ALPHA_DATA_IS_INVALID;
            else
	            return ERR_INVALID_TYPE;
        }

        public static int docmd_xtoa(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL)
            {
	            Phloat x = ((vartype_real) reg_x).x;

	            if (x < 0)
	                x = -x;
	        
                if (x >= 256)
	                return ERR_INVALID_DATA;
	        
                append_alpha_char(Phloat.to_char(x));
            }
            else if (reg_x.type == TYPE.STRING)
            {
	            vartype_string s = (vartype_string) reg_x;
	            append_alpha_string(s.text, s.length, 0);
            }
            else if (reg_x.type == TYPE.REALMATRIX)
            {
	            vartype_realmatrix m = (vartype_realmatrix) reg_x;

	            int size = m.rows * m.columns;
	            int i;
	            char[] buf = new char[44];
	            int buflen = 0;

	            for (i = size - 1; i >= 0; i--)
                {
	                if (m.array.is_string[i] != 0)
                    {
		                int j;
		                
                        for (j = Phloat.phloat_length_get(m.array.data[i]) - 1; j >= 0; j--)
                        {
        		            buf[buflen++] = Phloat.phloat_text_get(m.array.data[i])[j];
		        
                            if (buflen == 44)
			                    goto done;
		                }
	                }
                    else
                    {
        		        Phloat d = m.array.data[i];
		        
                        if (d < 0)
		                    d = -d;
		        
                        if (d >= 256)
		                    buf[buflen++] = (char) 255;
		                else
		                    buf[buflen++] = Phloat.to_char(d);
		        
                        if (buflen == 44)
		                    goto done;
	                }
	            }

	        done:
    
                append_alpha_string(buf, buflen, 1);
            }
            else
	            return ERR_INVALID_TYPE;
            
            if (flags.trace_print != 0 && flags.printer_exists != 0)
	            docmd_pra(null);
            
            return ERR_NONE;
        }

        public static int docmd_agraph(arg_struct arg)
        {
            switch (reg_x.type)
            {
                case TYPE.REAL:
                    {
                        if (reg_y.type == TYPE.REAL)
                        {
                            Phloat x = ((vartype_real)reg_x).x;
                            Phloat y = ((vartype_real)reg_y).x;

                            draw_pattern(x, y, reg_alpha, reg_alpha_length);
                            flush_display();
                            
                            flags.message = flags.two_line_message = 1;
                            return ERR_NONE;
                        }
                        else if (reg_y.type == TYPE.STRING)
                            return ERR_ALPHA_DATA_IS_INVALID;
                        else
                            return ERR_INVALID_TYPE;
                    }

                case TYPE.COMPLEX:
                    {
                        Phloat x = ((vartype_complex)reg_x).re;
                        Phloat y = ((vartype_complex)reg_y).im;

                        draw_pattern(x, y, reg_alpha, reg_alpha_length);
                        flush_display();
                        
                        flags.message = flags.two_line_message = 1;
                        return ERR_NONE;
                    }

                case TYPE.COMPLEXMATRIX:
                    {
                        vartype_complexmatrix cm = (vartype_complexmatrix)reg_x;

                        int size = 2 * cm.rows * cm.columns;
                        int i;
                        
                        for (i = 0; i < size; i += 2)
                            draw_pattern(cm.array.data[i], cm.array.data[i + 1], reg_alpha, reg_alpha_length);
                        
                        flush_display();
                        
                        flags.message = flags.two_line_message = 1;
                        return ERR_NONE;
                    }

                case TYPE.STRING:
                    return ERR_ALPHA_DATA_IS_INVALID;
                
                default:
                    return ERR_INVALID_TYPE;
            }
        }

        public static int docmd_pixel(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL)
            {
                if (reg_y.type == TYPE.REAL)
                {
                    pixel_helper(((vartype_real)reg_x).x, ((vartype_real)reg_y).x);
                    flush_display();

                    flags.message = flags.two_line_message = 1;
                    return ERR_NONE;
                }
                else if (reg_y.type == TYPE.STRING)
                    return ERR_ALPHA_DATA_IS_INVALID;
                else
                    return ERR_INVALID_TYPE;
            }
            else if (reg_x.type == TYPE.COMPLEX)
            {
                pixel_helper(((vartype_complex)reg_x).re, ((vartype_complex)reg_x).im);
                flush_display();
                
                flags.message = flags.two_line_message = 1;
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.COMPLEXMATRIX)
            {
                vartype_complexmatrix m = (vartype_complexmatrix)reg_x;

                int size = 2 * m.rows * m.columns;
                int i;
                
                for (i = 0; i < size; i += 2)
                    pixel_helper(m.array.data[i], m.array.data[i + 1]);

                flush_display();

                flags.message = flags.two_line_message = 1;
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static void pixel_helper(Phloat dx, Phloat dy)
        {
            dx = dx < 0 ? -Phloat.floor(-dx + 0.5) : Phloat.floor(dx + 0.5);
            dy = dy < 0 ? -Phloat.floor(-dy + 0.5) : Phloat.floor(dy + 0.5);
            int x = dx < -132 ? -132 : dx > 132 ? 132 : Phloat.to_int(dx);
            int y = dy < -132 ? -132 : dy > 132 ? 132 : Phloat.to_int(dy);

            int i;
            int dot = 1;
            
            if (x < 0)
            {
                x = -x;
                
                if (x >= 1 && x <= 131)
                    for (i = 0; i < 16; i++)
                        draw_pixel(x - 1, i);
                
                dot = 0;
            }

            if (y < 0)
            {
                y = -y;
            
                if (y >= 1 && y <= 16)
                    for (i = 0; i < 131; i++)
                        draw_pixel(i, y - 1);
                
                dot = 0;
            }

            if (dot != 0 && x >= 1 && x <= 131 && y >= 1 && y <= 16)
                draw_pixel(x - 1, y - 1);
        }

        public static int docmd_beep(arg_struct arg)
        {
            tone(8);
            tone(5);
            tone(9);
            tone(8);

            return ERR_NONE;
        }

        public static int docmd_tone(arg_struct arg)
        {
            int err;
            int num = 0;

            err = arg_to_num(arg, ref num);
            
            if (err != ERR_NONE)
                return err;
            
            if (num >= 10)
                return ERR_INVALID_DATA;
            
            tone(num);
            
            return ERR_NONE;
        }

        public static int docmd_mvar(arg_struct arg)
        {
            return ERR_NONE;
        }

        public static int docmd_varmenu(arg_struct arg)
        {
            int saved_prgm = current_prgm;
            int prgm = 0;
            int pc = 0;
            int command = 0;

            arg_struct arg2 = new arg_struct();
            int i;

            if (arg.type == ARGTYPE_IND_NUM || arg.type == ARGTYPE_IND_STK || arg.type == ARGTYPE_IND_STR)
            {
                int err = resolve_ind_arg(arg);
                if (err != ERR_NONE)
                    return err;
            }

            if (find_global_label(arg, ref prgm, ref pc) == 0)
                return ERR_LABEL_NOT_FOUND;

            pc += get_command_length(prgm, pc);
            current_prgm = prgm;
            
            get_next_command(ref pc, ref command, arg2, 0);
            current_prgm = saved_prgm;

            if (command != CMD_MVAR)
                return ERR_NO_MENU_VARIABLES;
            
            varmenu_length = arg.length;
            
            for (i = 0; i < varmenu_length; i++)
                varmenu[i] = arg.val.text[i];

            varmenu_row = 0;
            varmenu_role = 0;
            
            return set_menu_return_err(MENULEVEL_APP, MENU_VARMENU);
        }

        public static int docmd_getkey(arg_struct arg)
        {
            mode_getkey = true;
            return ERR_NONE;
        }

        public static int docmd_menu(arg_struct arg)
        {
            set_menu(MENULEVEL_PLAIN, MENU_PROGRAMMABLE);
            mode_plainmenu_sticky = true;
            
            return ERR_NONE;
        }

        public static int docmd_x_eq_0(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;

            return ((vartype_real)reg_x).x == 0 ? ERR_YES : ERR_NO;
        }

        public static int docmd_x_ne_0(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;

            return ((vartype_real)reg_x).x != 0 ? ERR_YES : ERR_NO;
        }

        public static int docmd_x_lt_0(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;

            return ((vartype_real)reg_x).x < 0 ? ERR_YES : ERR_NO;
        }

        public static int docmd_x_gt_0(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;

            return ((vartype_real)reg_x).x > 0 ? ERR_YES : ERR_NO;
        }

        public static int docmd_x_le_0(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;

            return ((vartype_real)reg_x).x <= 0 ? ERR_YES : ERR_NO;
        }

        public static int docmd_x_ge_0(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;

            return ((vartype_real)reg_x).x >= 0 ? ERR_YES : ERR_NO;
        }

        public static int docmd_x_eq_y(arg_struct arg)
        {
            if (reg_x.type != reg_y.type)
                return ERR_NO;

            switch (reg_x.type)
            {
                case TYPE.REAL:
                    {
                        vartype_real x = (vartype_real)reg_x;
                        vartype_real y = (vartype_real)reg_y;

                        return x.x == y.x ? ERR_YES : ERR_NO;
                    }

                case TYPE.COMPLEX:
                    {
                        vartype_complex x = (vartype_complex)reg_x;
                        vartype_complex y = (vartype_complex)reg_y;

                        return x.re == y.re && x.im == y.im ? ERR_YES : ERR_NO;
                    }

                case TYPE.REALMATRIX:
                    {
                        vartype_realmatrix x = (vartype_realmatrix)reg_x;
                        vartype_realmatrix y = (vartype_realmatrix)reg_y;

                        int sz;
                        int i;

                        if (x.rows != y.rows || x.columns != y.columns)
                            return ERR_NO;

                        sz = x.rows * x.columns;
                        
                        for (i = 0; i < sz; i++)
                        {
                            int xstr = x.array.is_string[i];
                            int ystr = y.array.is_string[i];

                            if (xstr != ystr)
                                return ERR_NO;
                            
                            if (xstr != 0)
                            {
                                if (string_equals(Phloat.phloat_text_get(x.array.data[i]), Phloat.phloat_length_get(x.array.data[i]),
                                           Phloat.phloat_text_get(y.array.data[i]), Phloat.phloat_length_get(y.array.data[i])) == 0)
                                    return ERR_NO;
                            }
                            else
                            {
                                if (x.array.data[i] != y.array.data[i])
                                    return ERR_NO;
                            }
                        }
                        return ERR_YES;
                    }

                case TYPE.COMPLEXMATRIX:
                    {
                        vartype_complexmatrix x = (vartype_complexmatrix)reg_x;
                        vartype_complexmatrix y = (vartype_complexmatrix)reg_y;

                        int sz;
                        int i;

                        if (x.rows != y.rows || x.columns != y.columns)
                            return ERR_NO;
                        
                        sz = 2 * x.rows * x.columns;
                        
                        for (i = 0; i < sz; i++)
                            if (x.array.data[i] != y.array.data[i])
                                return ERR_NO;
                        
                        return ERR_YES;
                    }

                case TYPE.STRING:
                    {
                        vartype_string x = (vartype_string)reg_x;
                        vartype_string y = (vartype_string)reg_y;

                        if (string_equals(x.text, x.length, y.text, y.length) != 0)
                            return ERR_YES;
                        else
                            return ERR_NO;
                    }
                default:
                    // Can't happen
                    return ERR_INTERNAL_ERROR;
            }
        }

        public static int docmd_x_ne_y(arg_struct arg)
        {
            int err = docmd_x_eq_y(arg);

            switch (err)
            {
                case ERR_YES:
                    return ERR_NO;
            
                case ERR_NO:
                    return ERR_YES;
                
                default:
                    return err;
            }
        }

        public static int docmd_x_lt_y(arg_struct arg)
        {
            switch (reg_x.type)
            {
                case TYPE.REAL:
                    {
                        if (reg_y.type == TYPE.STRING)
                            return ERR_ALPHA_DATA_IS_INVALID;
                        else if (reg_y.type != TYPE.REAL)
                            return ERR_INVALID_TYPE;

                        if (((vartype_real)reg_x).x < ((vartype_real)reg_y).x)
                            return ERR_YES;
                        else
                            return ERR_NO;
                    }

                case TYPE.STRING:
                    return ERR_ALPHA_DATA_IS_INVALID;
                
                default:
                    return ERR_INVALID_TYPE;
            }
        }

        public static int docmd_x_gt_y(arg_struct arg)
        {
            switch (reg_x.type)
            {
                case TYPE.REAL:
                    {
                        if (reg_y.type == TYPE.STRING)
                            return ERR_ALPHA_DATA_IS_INVALID;
                        else if (reg_y.type != TYPE.REAL)
                            return ERR_INVALID_TYPE;

                        if (((vartype_real)reg_x).x > ((vartype_real)reg_y).x)
                            return ERR_YES;
                        else
                            return ERR_NO;
                    }

                case TYPE.STRING:
                    return ERR_ALPHA_DATA_IS_INVALID;
                
                default:
                    return ERR_INVALID_TYPE;
            }
        }

        public static int docmd_x_le_y(arg_struct arg)
        {
            switch (reg_x.type)
            {
                case TYPE.REAL:
                    {
                        if (reg_y.type == TYPE.STRING)
                            return ERR_ALPHA_DATA_IS_INVALID;
                        else if (reg_y.type != TYPE.REAL)
                            return ERR_INVALID_TYPE;

                        if (((vartype_real)reg_x).x <= ((vartype_real)reg_y).x)
                            return ERR_YES;
                        else
                            return ERR_NO;
                    }

                case TYPE.STRING:
                    return ERR_ALPHA_DATA_IS_INVALID;

                default:
                    return ERR_INVALID_TYPE;
            }
        }

        public static int docmd_x_ge_y(arg_struct arg)
        {
            switch (reg_x.type)
            {
                case TYPE.REAL:
                    {
                        if (reg_y.type == TYPE.STRING)
                            return ERR_ALPHA_DATA_IS_INVALID;
                        else if (reg_y.type != TYPE.REAL)
                            return ERR_INVALID_TYPE;

                        if (((vartype_real)reg_x).x >= ((vartype_real)reg_y).x)
                            return ERR_YES;
                        else
                            return ERR_NO;
                    }

                case TYPE.STRING:
                    return ERR_ALPHA_DATA_IS_INVALID;

                default:
                    return ERR_INVALID_TYPE;
            }
        }

        public static int docmd_prsigma(arg_struct arg)
        {
            vartype regs = recall_var("REGS".ToCharArray(), 4);
            vartype_realmatrix rm;
            int nr;
            int size;
            int max;
            int i;
            char[] buf = new char[100];
            int bufptr;
            int dispmode;
            int digits = 0;

            if (regs == null)
	            return ERR_NONEXISTENT;

            if (regs.type != TYPE.REALMATRIX)
	            return ERR_INVALID_TYPE;
            
            rm = (vartype_realmatrix) regs;
            nr = flags.all_sigma != 0 ? 13 : 6;

            size = rm.rows * rm.columns;
            max = mode_sigma_reg + nr;
            
            if (max > size)
	            return ERR_SIZE_ERROR;

            if (flags.printer_enable == 0 && program_running())
	            return ERR_NONE;

            if (flags.printer_exists == 0)
	            return ERR_PRINTING_IS_DISABLED;

            if (flags.fix_or_all != 0)
	            dispmode = flags.eng_or_all != 0 ? 3 : 0;
            else
	            dispmode = flags.eng_or_all != 0 ? 2 : 1;

            if (flags.digits_bit3 != 0)
	            digits += 8;
            
            if (flags.digits_bit2 != 0)
	            digits += 4;
            
            if (flags.digits_bit1 != 0)
	            digits += 2;
            
            if (flags.digits_bit0 != 0)
	            digits += 1;
    
            shell_annunciators(-1, -1, 1, -1, -1, -1);
            print_text(null, 0, 1);

            for (i = 0; i < nr; i++)
            {
	            int j = i + mode_sigma_reg;

	            if (rm.array.is_string[j] != 0)
                {
	                bufptr = 0;
	                char2buf(buf, 100, ref bufptr, '"');
	                string2buf(buf, 100, ref bufptr, Phloat.phloat_text_get(rm.array.data[j]), Phloat.phloat_length_get(rm.array.data[j]));
	                char2buf(buf, 100, ref bufptr, '"');
	            }
                else
	                bufptr = easy_phloat2string(rm.array.data[j], buf, 100, 0);
	                print_wide(sigma_labels[i].text, sigma_labels[i].length, buf, bufptr);
            }
            
            shell_annunciators(-1, -1, 0, -1, -1, -1);
            
            return ERR_NONE;
        }

        public static int docmd_prp(arg_struct arg)
        {
            int prgm_index = 0;
            
            if (arg.type == ARGTYPE_LBLINDEX)
                prgm_index = labels[arg.val.num].prgm;
            else if (arg.type == ARGTYPE_STR)
            {
                if (arg.length == 0)
                    prgm_index = current_prgm;
                else
                {
                    int pc = 0;

                    if (find_global_label(arg, ref prgm_index, ref pc) == 0)
                        return ERR_LABEL_NOT_FOUND;
                }
            }
            else
                return ERR_INVALID_TYPE;

            if (flags.printer_exists == 0)
                return ERR_PRINTING_IS_DISABLED;
            
            return print_program(prgm_index, -1, -1, 0);
        }

        public static int docmd_prv(arg_struct arg)
        {
            if (arg.type == ARGTYPE_IND_NUM || arg.type == ARGTYPE_IND_STK || arg.type == ARGTYPE_IND_STR)
            {
	            int err = resolve_ind_arg(arg);
	            
                if (err != ERR_NONE)
	                return err;
            }

            if (arg.type != ARGTYPE_STR)
	            return ERR_INVALID_TYPE;
            else
            {
	            vartype v = recall_var(arg.val.text, arg.length);

	            char[] lbuf = new char[32];
                char[] rbuf = new char[100];
	            int llen = 0;
                int rlen = 0;
	        
                if (v == null)
	                return ERR_NONEXISTENT;

	            if (flags.printer_enable == 0 && program_running())
	                return ERR_NONE;
	        
                if (flags.printer_exists == 0)
	                return ERR_PRINTING_IS_DISABLED;

	            shell_annunciators(-1, -1, 1, -1, -1, -1);
	        
                string2buf(lbuf, 8, ref llen, arg.val.text, arg.length);
	            char2buf(lbuf, 8, ref llen, '=');
	        
                rlen = vartype2string(v, rbuf, 100);
	            print_wide(lbuf, llen, rbuf, rlen);

	            if (v.type == TYPE.REALMATRIX || v.type == TYPE.COMPLEXMATRIX)
                {
	                prv_var = v;
	                prv_index = 0;
	                mode_interruptible = prv_worker;
	                mode_stoppable = true;
	                
                    return ERR_INTERRUPTIBLE;
	            }
                else
                {
	                shell_annunciators(-1, -1, 0, -1, -1, -1);
	                return ERR_NONE;
	            }
            }
        }

        public static int prv_worker(int interrupted)
        {
            char[] lbuf = new char[32];
            char[] rbuf = new char[100];
            
            int llen = 0;
            int rlen = 0;
            int i;
            int j;
            int sz;

            if (interrupted != 0)
            {
	            shell_annunciators(-1, -1, 0, -1, -1, -1);
	            return ERR_STOP;
            }

            if (prv_var.type == TYPE.REALMATRIX)
            {
	            vartype_realmatrix rm = (vartype_realmatrix) prv_var;

	            i = prv_index / rm.columns;
	            j = prv_index % rm.columns;
	            sz = rm.rows * rm.columns;

	            llen = int2string(i + 1, lbuf, 0, 32);
	            char2buf(lbuf, 32, ref llen, ':');
	            llen += int2string(j + 1, lbuf, llen, 32);
	            char2buf(lbuf, 32, ref llen, '=');
	        
                if (rm.array.is_string[prv_index] != 0)
                {
	                rlen = 0;
	                char2buf(rbuf, 100, ref rlen, '"');
	                string2buf(rbuf, 100, ref rlen, Phloat.phloat_text_get(rm.array.data[prv_index]), Phloat.phloat_length_get(rm.array.data[prv_index]));
	                char2buf(rbuf, 100, ref rlen, '"');
	            }
                else
	                rlen = easy_phloat2string(rm.array.data[prv_index], rbuf, 100, 0);

	            print_wide(lbuf, llen, rbuf, rlen);
            }
            else // prv_var->type == TYPE_COMPLEXMATRIX)
            {
	            vartype_complexmatrix cm = (vartype_complexmatrix) prv_var;
                vartype_complex cpx = new vartype_complex();

	            cpx.type = TYPE.COMPLEX;
	            i = prv_index / cm.columns;
	            j = prv_index % cm.columns;
	            sz = cm.rows * cm.columns;

	            llen = int2string(i + 1, lbuf, 0, 32);
	            char2buf(lbuf, 32, ref llen, ':');
	            llen += int2string(j + 1, lbuf, llen, 32);
	            char2buf(lbuf, 32, ref llen, '=');

	            cpx.re = cm.array.data[2 * prv_index];
	            cpx.im = cm.array.data[2 * prv_index + 1];
	            
                rlen = vartype2string((vartype) cpx, rbuf, 100);
	            
                print_wide(lbuf, llen, rbuf, rlen);
            }

            if (++prv_index < sz)
    	        return ERR_INTERRUPTIBLE;
            else
            {
	            shell_annunciators(-1, -1, 0, -1, -1, -1);
	            return ERR_NONE;
            }
        }

        public static int docmd_prstk(arg_struct arg)
        {
            char[] buf = new char[100];
            int len;

            if (flags.printer_enable == 0 && program_running())
	            return ERR_NONE;
            
            if (flags.printer_exists == 0)
	            return ERR_PRINTING_IS_DISABLED;
            
            shell_annunciators(-1, -1, 1, -1, -1, -1);
            print_text(null, 0, 1);
            len = vartype2string(reg_t, buf, 100);
            print_wide("T=".ToCharArray(), 2, buf, len);
            len = vartype2string(reg_z, buf, 100);
            print_wide("Z=".ToCharArray(), 2, buf, len);
            len = vartype2string(reg_y, buf, 100);
            print_wide("Y=".ToCharArray(), 2, buf, len);
            len = vartype2string(reg_x, buf, 100);
            print_wide("X=".ToCharArray(), 2, buf, len);
            shell_annunciators(-1, -1, 0, -1, -1, -1);
            
            return ERR_NONE;
        }

        public static int docmd_prusr(arg_struct arg)
        {
            if (flags.printer_enable == 0 && program_running())
                return ERR_NONE;
            
            if (flags.printer_exists == 0)
                return ERR_PRINTING_IS_DISABLED;
            else
            {
                shell_annunciators(-1, -1, 1, -1, -1, -1);
                print_text(null, 0, 1);
                prusr_state = 0;
                prusr_index = vars_count - 1;
                mode_interruptible = prusr_worker;
                mode_stoppable = true;
                return ERR_INTERRUPTIBLE;
            }
        }

        public static int prusr_worker(int interrupted)
        {
            if (interrupted != 0)
            {
	            shell_annunciators(-1, -1, 0, -1, -1, -1);
	            return ERR_STOP;
            }

            if (prusr_state == 0)
            {
	            char[] lbuf = new char[8];
	            char[] rbuf = new char[100];
	            int llen = 0;
                int rlen = 0;

	            if (prusr_index < 0)
                {
	                if (vars_count > 0)
		            print_text(null, 0, 1);
	                prusr_state = 1;
	                prusr_index = 0;
	            
                    // goto state1;                 State1 replicated here to avoid goto
                    char[] buf = new char[13];
                    int len;

                    len = 0;

                    if (prusr_index >= labels_count)
                    {
                        shell_annunciators(-1, -1, 0, -1, -1, -1);
                        return ERR_NONE;
                    }

                    if (labels[prusr_index].length == 0)
                    {
                        if (prusr_index == labels_count - 1)
                            string2buf(buf, 13, ref len, ".END.".ToCharArray(), 5);
                        else
                            string2buf(buf, 13, ref len, "END".ToCharArray(), 3);
                    }
                    else
                    {
                        string2buf(buf, 13, ref len, "LBL \"".ToCharArray(), 5);
                        string2buf(buf, 13, ref len, labels[prusr_index].name, labels[prusr_index].length);
                        char2buf(buf, 13, ref len, '"');
                    }

                    print_text(buf, len, 1);
                    prusr_index++;


                    return ERR_INTERRUPTIBLE;       // Exit as in state1 rather than falling through
                }

	            llen = 0;
	            string2buf(lbuf, 8, ref llen, vars[prusr_index].name, vars[prusr_index].length);
	            char2buf(lbuf, 8, ref llen, '=');
	            rlen = vartype2string(vars[prusr_index].value, rbuf, 100);
	        
                print_wide(lbuf, llen, rbuf, rlen);
	            prusr_index--;
            }
            else
            {
	            char[] buf = new char[13];
	            int len;

	        // state1:          State1 replicated above
	            len = 0;
	        
                if (prusr_index >= labels_count)
                {
	                shell_annunciators(-1, -1, 0, -1, -1, -1);
	                return ERR_NONE;
	            }

	            if (labels[prusr_index].length == 0)
                {
	                if (prusr_index == labels_count - 1)
		                string2buf(buf, 13, ref len, ".END.".ToCharArray(), 5);
	                else
		                string2buf(buf, 13, ref len, "END".ToCharArray(), 3);
	            }
                else
                {
	                string2buf(buf, 13, ref len, "LBL \"".ToCharArray(), 5);
	                string2buf(buf, 13, ref len, labels[prusr_index].name, labels[prusr_index].length);
	                char2buf(buf, 13, ref len, '"');
	            }

	            print_text(buf, len, 1);
	            prusr_index++;
            }

            return ERR_INTERRUPTIBLE;
        }

        public static int docmd_list(arg_struct arg)
        {
            if (arg.type != ARGTYPE_NUM)
                return ERR_INVALID_TYPE;
            
            if (arg.val.num == 0)
                return ERR_NONE;
            
            if (flags.printer_exists == 0)
                return ERR_PRINTING_IS_DISABLED;
            
            return print_program(current_prgm, pc, arg.val.num, 0);
        }

        public static int docmd_adv(arg_struct arg)
        {
            if (flags.printer_exists != 0 && (flags.printer_enable != 0 || !program_running()))
            {
                shell_annunciators(-1, -1, 1, -1, -1, -1);
                print_text(null, 0, 1);
                shell_annunciators(-1, -1, 0, -1, -1, -1);
            }

            return ERR_NONE;
        }

        public static int docmd_prlcd(arg_struct arg)
        {
            if (flags.printer_enable == 0 && program_running())
                return ERR_NONE;
            
            if (flags.printer_exists == 0)
                return ERR_PRINTING_IS_DISABLED;
            else
            {
                shell_annunciators(-1, -1, 1, -1, -1, -1);
                print_display();
                shell_annunciators(-1, -1, 0, -1, -1, -1);
                
                return ERR_NONE;
            }
        }

        public static int docmd_delay(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL)
            {
                Phloat x = ((vartype_real)reg_x).x;

                if (x < 0)
                    x = -x;
                
                if (x >= 1.95)
                    return ERR_INVALID_DATA;
                else
                    // We don't actually use the delay value...
                    return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_pon(arg_struct arg)
        {
            flags.printer_exists = 1;
            flags.printer_enable = 1;

            return ERR_NONE;
        }

        public static int docmd_poff(arg_struct arg)
        {
            flags.printer_exists = 0;
            flags.printer_enable = 0;

            return ERR_NONE;
        }

        public static int docmd_man(arg_struct arg)
        {
            flags.trace_print = 0;
            flags.normal_print = 0;

            return ERR_NONE;
        }

        public static int docmd_norm(arg_struct arg)
        {
            flags.trace_print = 0;
            flags.normal_print = 1;

            return ERR_NONE;
        }

        public static int docmd_trace(arg_struct arg)
        {
            flags.trace_print = 1;
            flags.normal_print = 0;

            return ERR_NONE;
        }

        public static int docmd_end(arg_struct arg)
        {
            return docmd_rtn(arg);
        }

        public static int docmd_number(arg_struct arg)
        {
            vartype new_x = new_real(arg.val_d);

            if (new_x == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            if (flags.stack_lift_disable != 0)
                free_vartype(ref reg_x);
            else
            {
                free_vartype(ref reg_t);
                reg_t = reg_z;
                reg_z = reg_y;
                reg_y = reg_x;
            }

            reg_x = new_x;
            
            return ERR_NONE;
        }

        public static int docmd_string(arg_struct arg)
        {
            bool append = arg.length > 0 && arg.val.text[0] == 127;
            
            if (append)
            {
                char[] tmp = new char[arg.length - 1];
                Array.Copy(arg.val.text, 1, tmp, 0, arg.length - 1);
                append_alpha_string(tmp, arg.length - 1, 0);
            
                if (flags.trace_print != 0 && flags.printer_exists != 0)
                    docmd_pra(null);
            }
            else
            {
                reg_alpha_length = 0;
                append_alpha_string(arg.val.text, arg.length, 0);
            }
            
            return ERR_NONE;
        }

        public static int docmd_gtodot(arg_struct arg)
        {
            if (arg.type == ARGTYPE_NUM)
            {
                pc = line2pc(arg.val.num);
                clear_all_rtns();
                prgm_highlight_row = 1;
                
                return ERR_NONE;
            }
            else if (arg.type == ARGTYPE_STR)
            {
                int new_prgm = 0;
                int new_pc = 0;
                
                if (find_global_label(arg, ref new_prgm, ref new_pc) != 0)
                {
                    current_prgm = new_prgm;
                    pc = new_pc;
                    clear_all_rtns();
                    prgm_highlight_row = 1;
                
                    return ERR_NONE;
                }
                else
                    return ERR_LABEL_NOT_FOUND;
            }
            else if (arg.type == ARGTYPE_LBLINDEX)
            {
                int labelindex = arg.val.num;
                current_prgm = labels[labelindex].prgm;
                pc = labels[labelindex].pc;
                clear_all_rtns();
                prgm_highlight_row = 1;
                
                return ERR_NONE;
            }
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_gtodotdot(arg_struct arg)
        {
            goto_dot_dot();

            return ERR_NONE;
        }

        public static int docmd_stop(arg_struct arg)
        {
            return ERR_STOP;
        }

        public static int docmd_newmat(arg_struct arg)
        {
            vartype m;

            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;
            if (reg_y.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_y.type != TYPE.REAL)
                return ERR_INVALID_TYPE;

            Phloat x = ((vartype_real)reg_x).x;

            if (x <= -2147483648.0 || x >= 2147483648.0)
                return ERR_DIMENSION_ERROR;
            
            int xx = Phloat.to_int(x);

            if (xx == 0)
                return ERR_DIMENSION_ERROR;
            
            if (xx < 0)
                xx = -xx;

            Phloat y = ((vartype_real)reg_y).x;

            if (y <= -2147483648.0 || y >= 2147483648.0)
                return ERR_DIMENSION_ERROR;
            
            int yy = Phloat.to_int(y);

            if (yy == 0)
                return ERR_DIMENSION_ERROR;
            
            if (yy < 0)
                yy = -yy;

            m = new_realmatrix(yy, xx);
            
            if (m == null)
                return ERR_INSUFFICIENT_MEMORY;
            else
            {
                binary_result(m);
                return ERR_NONE;
            }
        }

        public static int docmd_rup(arg_struct arg)
        {
            vartype temp = reg_x;
            
            reg_x = reg_t;
            reg_t = reg_z;
            reg_z = reg_y;
            reg_y = temp;

            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_real_t(arg_struct arg)
        {
            return reg_x.type == TYPE.REAL ? ERR_YES : ERR_NO;
        }

        public static int docmd_cpx_t(arg_struct arg)
        {
            return reg_x.type == TYPE.COMPLEX ? ERR_YES : ERR_NO;
        }

        public static int docmd_str_t(arg_struct arg)
        {
            return reg_x.type == TYPE.STRING ? ERR_YES : ERR_NO;
        }

        public static int docmd_mat_t(arg_struct arg)
        {
            return reg_x.type == TYPE.REALMATRIX || reg_x.type == TYPE.COMPLEXMATRIX ? ERR_YES : ERR_NO;
        }

        public static int docmd_dim_t(arg_struct arg)
        {
            int rows;
            int columns;

            if (reg_x.type == TYPE.REALMATRIX)
            {
                rows = ((vartype_realmatrix)reg_x).rows;
                columns = ((vartype_realmatrix)reg_x).columns;
            }
            else if (reg_x.type == TYPE.COMPLEXMATRIX)
            {
                rows = ((vartype_complexmatrix)reg_x).rows;
                columns = ((vartype_complexmatrix)reg_x).columns;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;

            vartype new_y = new_real(rows);
            vartype new_x = new_real(columns);

            if (new_x == null || new_y == null)
            {
                free_vartype(ref new_x);
                free_vartype(ref new_y);

                return ERR_INSUFFICIENT_MEMORY;
            }

            free_vartype(ref reg_lastx);
            reg_lastx = reg_x;
            free_vartype(ref reg_t);
            reg_t = reg_z;
            reg_z = reg_y;
            reg_y = new_y;
            reg_x = new_x;

            return ERR_NONE;
        }

        public static int assign_helper(int num, arg_struct arg)
        {
            if (arg.type == ARGTYPE_COMMAND)
            {
                // For backward compatibility only; we don't allow this type
                // of assignment command to be created anymore, but we do want
                // programs that already contain such commands to continue
                // working.
                //
                command_spec cs = cmd_array[arg.val.cmd];
                assign_custom_key(num, cs.name.ToCharArray(), cs.name_length);
            }
            else
                assign_custom_key(num, arg.val.text, arg.length);

            flags.local_label = 0;
            
            return ERR_NONE;
        }

        public static int docmd_asgn01(arg_struct arg)
        {
            return assign_helper(1, arg);
        }

        public static int docmd_asgn02(arg_struct arg)
        {
            return assign_helper(2, arg);
        }

        public static int docmd_asgn03(arg_struct arg)
        {
            return assign_helper(3, arg);
        }

        public static int docmd_asgn04(arg_struct arg)
        {
            return assign_helper(4, arg);
        }

        public static int docmd_asgn05(arg_struct arg)
        {
            return assign_helper(5, arg);
        }

        public static int docmd_asgn06(arg_struct arg)
        {
            return assign_helper(6, arg);
        }

        public static int docmd_asgn07(arg_struct arg)
        {
            return assign_helper(7, arg);
        }

        public static int docmd_asgn08(arg_struct arg)
        {
            return assign_helper(8, arg);
        }

        public static int docmd_asgn09(arg_struct arg)
        {
            return assign_helper(9, arg);
        }

        public static int docmd_asgn10(arg_struct arg)
        {
            return assign_helper(10, arg);
        }

        public static int docmd_asgn11(arg_struct arg)
        {
            return assign_helper(11, arg);
        }

        public static int docmd_asgn12(arg_struct arg)
        {
            return assign_helper(12, arg);
        }

        public static int docmd_asgn13(arg_struct arg)
        {
            return assign_helper(13, arg);
        }

        public static int docmd_asgn14(arg_struct arg)
        {
            return assign_helper(14, arg);
        }

        public static int docmd_asgn15(arg_struct arg)
        {
            return assign_helper(15, arg);
        }

        public static int docmd_asgn16(arg_struct arg)
        {
            return assign_helper(16, arg);
        }

        public static int docmd_asgn17(arg_struct arg)
        {
            return assign_helper(17, arg);
        }

        public static int docmd_asgn18(arg_struct arg)
        {
            return assign_helper(18, arg);
        }

        public static int docmd_on(arg_struct arg)
        {
            flags.continuous_on = 1;

            return ERR_NONE;
        }

        public static int docmd_off(arg_struct arg)
        {
            shell_powerdown();
            
            return ERR_NONE;
        }

        public static int docmd_key1g(arg_struct arg)
        {
            assign_prgm_key(1, 1, arg);
            return ERR_NONE;
        }

        public static int docmd_key2g(arg_struct arg)
        {
            assign_prgm_key(2, 1, arg);
            return ERR_NONE;
        }

        public static int docmd_key3g(arg_struct arg)
        {
            assign_prgm_key(3, 1, arg);
            return ERR_NONE;
        }

        public static int docmd_key4g(arg_struct arg)
        {
            assign_prgm_key(4, 1, arg);
            return ERR_NONE;
        }

        public static int docmd_key5g(arg_struct arg)
        {
            assign_prgm_key(5, 1, arg);
            return ERR_NONE;
        }

        public static int docmd_key6g(arg_struct arg)
        {
            assign_prgm_key(6, 1, arg);
            return ERR_NONE;
        }

        public static int docmd_key7g(arg_struct arg)
        {
            assign_prgm_key(7, 1, arg);
            return ERR_NONE;
        }

        public static int docmd_key8g(arg_struct arg)
        {
            assign_prgm_key(8, 1, arg);
            return ERR_NONE;
        }

        public static int docmd_key9g(arg_struct arg)
        {
            assign_prgm_key(9, 1, arg);
            return ERR_NONE;
        }

        public static int docmd_key1x(arg_struct arg)
        {
            assign_prgm_key(1, 0, arg);
            return ERR_NONE;
        }

        public static int docmd_key2x(arg_struct arg)
        {
            assign_prgm_key(2, 0, arg);
            return ERR_NONE;
        }

        public static int docmd_key3x(arg_struct arg)
        {
            assign_prgm_key(3, 0, arg);
            return ERR_NONE;
        }

        public static int docmd_key4x(arg_struct arg)
        {
            assign_prgm_key(4, 0, arg);
            return ERR_NONE;
        }

        public static int docmd_key5x(arg_struct arg)
        {
            assign_prgm_key(5, 0, arg);
            return ERR_NONE;
        }

        public static int docmd_key6x(arg_struct arg)
        {
            assign_prgm_key(6, 0, arg);
            return ERR_NONE;
        }

        public static int docmd_key7x(arg_struct arg)
        {
            assign_prgm_key(7, 0, arg);
            return ERR_NONE;
        }

        public static int docmd_key8x(arg_struct arg)
        {
            assign_prgm_key(8, 0, arg);
            return ERR_NONE;
        }

        public static int docmd_key9x(arg_struct arg)
        {
            assign_prgm_key(9, 0, arg);
            return ERR_NONE;
        }

        public static int docmd_vmsto(arg_struct arg)
        {
            // STO variant that is invoked from a VARMENU
            int err = docmd_sto(arg);

            if (err == ERR_NONE)
                err = view_helper(arg, 0);
            
            if (err == ERR_NONE)
                mode_varmenu = true;
            
            return err;
        }

        public static int docmd_sigma_reg(arg_struct arg)
        {
            if (arg.type == ARGTYPE_IND_NUM || arg.type == ARGTYPE_IND_STK || arg.type == ARGTYPE_IND_STR)
            {
                int err = resolve_ind_arg(arg);

                if (err != ERR_NONE)
                    return err;
            }

            if (arg.type != ARGTYPE_NUM)
                return ERR_INTERNAL_ERROR;
            
            mode_sigma_reg = (int)arg.val.num;
            
            if (mode_sigma_reg < 0)
                mode_sigma_reg = -mode_sigma_reg;
            
            return ERR_NONE;
        }

        public static int docmd_sigma_reg_t(arg_struct arg)
        {
            vartype v = new_real(mode_sigma_reg);

            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            recall_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_cld(arg_struct arg)
        {
            if (flags.message != 0 && mode_goose >= 0)
                mode_goose = -1 - mode_goose;

            flags.message = 0;
            flags.two_line_message = 0;
            
            return ERR_NONE;
        }

        public static int docmd_vmsto2(arg_struct arg)
        {
            // Special-purpose STO variant that is invoked from the Solver's VARMENU.
            // It saves the previous value of the target variable; this feature is used
            // by docmd_vmsolve() to provide the second initial guess (the first is
            // taken from the target variable, that is, the variable named as SOLVE's
            // parameter).
            //
            vartype v;

            if (arg.type != ARGTYPE_STR)
                return ERR_INVALID_TYPE;
            
            v = recall_var(arg.val.text, arg.length);
            
            if (v == null || v.type != TYPE.REAL)
                remove_shadow(arg.val.text, arg.length);
            else
                put_shadow(arg.val.text, arg.length, ((vartype_real)v).x);

            
            return docmd_vmsto(arg);
        }
    }
}