﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        public static int docmd_rclel(arg_struct arg)
        {
            vartype m;
            vartype v;

            switch (matedit_mode)
            {
	            case 0:
	                return ERR_NONEXISTENT;
	            case 1:
	            case 3:
	                m = recall_var(matedit_name, matedit_length);
	                break;
	            case 2:
	                m = matedit_x;
	                break;
	            default:
	                return ERR_INTERNAL_ERROR;
            }

            if (m == null)
	            return ERR_NONEXISTENT;

            if (m.type == TYPE.REALMATRIX)
            {
    	        vartype_realmatrix rm = (vartype_realmatrix) m;
	            int n = matedit_i * rm.columns + matedit_j;
	        
                if (rm.array.is_string[n] != 0)
	                v = new_string(Phloat.phloat_text_get(rm.array.data[n]), Phloat.phloat_length_get(rm.array.data[n]));
	            else
	                v = new_real(rm.array.data[n]);
            }
            else if (m.type == TYPE.COMPLEXMATRIX)
            {
	            vartype_complexmatrix cm = (vartype_complexmatrix) m;
	            int n = matedit_i * cm.columns + matedit_j;
	            
                v = new_complex(cm.array.data[2 * n], cm.array.data[2 * n + 1]);
            }
            else
    	        return ERR_INVALID_TYPE;
    
            if (v == null)
	            return ERR_INSUFFICIENT_MEMORY;
            
            recall_result(v);
            return ERR_NONE;
        }

        public static int docmd_stoel(arg_struct arg)
        {
            vartype m;

            switch (matedit_mode)
            {
                case 0:
                    return ERR_NONEXISTENT;
                case 1:
                case 3:
                    m = recall_var(matedit_name, matedit_length);
                    break;
                case 2:
                    m = matedit_x;
                    break;
                default:
                    return ERR_INTERNAL_ERROR;
            }

            if (m == null)
                return ERR_NONEXISTENT;

            if (m.type != TYPE.REALMATRIX && m.type != TYPE.COMPLEXMATRIX)
                // Should not happen, but could, as long as I don't implement
                // matrix locking.
                //
                return ERR_INVALID_TYPE;

            if (disentangle(m) == 0)
                return ERR_INSUFFICIENT_MEMORY;

            if (m.type == TYPE.REALMATRIX)
            {
                vartype_realmatrix rm = (vartype_realmatrix)m;
                int n = matedit_i * rm.columns + matedit_j;

                if (reg_x.type == TYPE.REAL)
                {
                    rm.array.is_string[n] = (char)0;
                    rm.array.data[n] = ((vartype_real)reg_x).x;
                    return ERR_NONE;
                }
                else if (reg_x.type == TYPE.STRING)
                {
                    vartype_string s = (vartype_string)reg_x;
                    int i;

                    rm.array.is_string[n] = (char)1;
                    Phloat.phloat_length_set(rm.array.data[n], s.length);
                    
                    for (i = 0; i < s.length; i++)
                        Phloat.phloat_text_set(rm.array.data[n], i, s.text[i]);

                    return ERR_NONE;
                }
                else
                    return ERR_INVALID_TYPE;
            }
            else // m.type == TYPE_COMPLEXMATRIX
            {
                vartype_complexmatrix cm = (vartype_complexmatrix)m;

                int n = matedit_i * cm.columns + matedit_j;

                if (reg_x.type == TYPE.REAL)
                {
                    cm.array.data[2 * n] = ((vartype_real)reg_x).x;
                    cm.array.data[2 * n + 1] = 0;
                    return ERR_NONE;
                }
                else if (reg_x.type == TYPE.COMPLEX)
                {
                    vartype_complex c = (vartype_complex)reg_x;
                    cm.array.data[2 * n] = c.re;
                    cm.array.data[2 * n + 1] = c.im;
                    return ERR_NONE;
                }
                else
                    return ERR_INVALID_TYPE;
            }
        }

        public static int docmd_insr(arg_struct arg)
        {
            vartype m;
            vartype newx = new vartype();
            vartype_realmatrix rm = new vartype_realmatrix();
            vartype_complexmatrix cm = new vartype_complexmatrix();

            int rows;
            int columns;
            int i;
            int err;
            int refcount;
            int interactive;

            switch (matedit_mode)
            {
	            case 0:
	                return ERR_NONEXISTENT;
	        
                case 1:
	            case 3:
	                m = recall_var(matedit_name, matedit_length);
	                break;
	            
                case 2:
	                m = matedit_x;
	                break;
	            
                default:
	                return ERR_INTERNAL_ERROR;
            }

            if (m == null)
	            return ERR_NONEXISTENT;
            
            if (m.type != TYPE.REALMATRIX && m.type != TYPE.COMPLEXMATRIX)
	            return ERR_INVALID_TYPE;

            interactive = matedit_mode == 2 || matedit_mode == 3 ? 1 : 0;
            
            if (interactive != 0)
            {
	            err = docmd_stoel(null);
	        
                if (err != ERR_NONE)
	                return err;
            }

            if (m.type == TYPE.REALMATRIX)
            {
	            rm = (vartype_realmatrix) m;
	            rows = rm.rows;
	            columns = rm.columns;
	            refcount = rm.array.refcount;

	            if (interactive != 0)
                {
	                newx = new_real(0);
	            
                    if (newx == null)
		                return ERR_INSUFFICIENT_MEMORY;
	            }
            }
            else
            {
	            cm = (vartype_complexmatrix) m;
	            rows = cm.rows;
	            columns = cm.columns;
	            refcount = cm.array.refcount;
	            
                if (interactive != 0)
                {
	                newx = new_complex(0, 0);
	            
                    if (newx == null)
		                return ERR_INSUFFICIENT_MEMORY;
	            }
            }

            if (matedit_i >= rows)
	            matedit_i = rows - 1;

            if (matedit_j >= columns)
	            matedit_j = columns - 1;

            if (refcount == 1)
            {
	            // We have this array to ourselves so we can modify it in place
	            err = dimension_array_ref(m, rows + 1, columns);

	            if (err != ERR_NONE)
                {
	                if (interactive != 0)
		                free_vartype(ref newx);

	                return err;
	            }
	        
                rows++;
	        
                if (m.type == TYPE.REALMATRIX)
                {
	                for (i = rows * columns - 1; i >= (matedit_i + 1) * columns; i--)
                    {
		                rm.array.is_string[i] = rm.array.is_string[i - columns];
		                rm.array.data[i] = rm.array.data[i - columns];
	                }
	            
                    for (i = matedit_i * columns; i < (matedit_i + 1) * columns; i++)
                    {
		                rm.array.is_string[i] = (char)0;
		                rm.array.data[i] = 0;
	                }
	            }
                else
                {
	                for (i = 2 * rows * columns - 1; i >= 2 * (matedit_i + 1) * columns; i--)
		                cm.array.data[i] = cm.array.data[i - 2 * columns];
	            
                    for (i = 2 * matedit_i * columns; i < 2 * (matedit_i + 1) * columns; i++)
                        cm.array.data[i] = 0;
    	        }
            }
            else
            {
	            // We're sharing this array. I don't use disentangle() because it
	            // does not deal with resizing.
	            int newsize = (rows + 1) * columns;
	        
                if (m.type == TYPE.REALMATRIX)
                {
	                realmatrix_data array = new realmatrix_data();

                    if (array == null)
                    {
		                if (interactive != 0)
		                    free_vartype(ref newx);
		        
                        return ERR_INSUFFICIENT_MEMORY;
	                }
	            
                    array.data = new Phloat[newsize];
	            
                    if (array.data == null)
                    {
		                if (interactive != 0)
		                    free_vartype(ref newx);
		                
                        array = null;
		                
                        return ERR_INSUFFICIENT_MEMORY;
	                }
	            
                    array.is_string = new char[newsize];
	            
                    if (array.is_string == null)
                    {
		                if (interactive != 0)
		                    free_vartype(ref newx);
		        
                        array.data = null;
		                array = null;
		        
                        return ERR_INSUFFICIENT_MEMORY;
	                }
	            
                    for (i = 0; i < matedit_i * columns; i++)
                    {
		                array.is_string[i] = rm.array.is_string[i];
		                array.data[i] = rm.array.data[i];
	                }
	            
                    for (i = matedit_i * columns; i < (matedit_i + 1) * columns; i++)
                    {
		                array.is_string[i] = (char)0;
		                array.data[i] = 0;
	                }
	            
                    for (i = (matedit_i + 1) * columns; i < newsize; i++)
                    {
		                array.is_string[i] = rm.array.is_string[i - columns];
		                array.data[i] = rm.array.data[i - columns];
	                }

	                array.refcount = 1;
	                rm.array.refcount--;
	                rm.array = array;
	                rm.rows++;
	            }
                else
                {
                    complexmatrix_data array = new complexmatrix_data();
	                
                    if (array == null)
                    {
		                if (interactive != 0)
		                    free_vartype(ref newx);

		                return ERR_INSUFFICIENT_MEMORY;
	                }
	                
                    array.data = new Phloat[2 * newsize];
	                
                    if (array.data == null)
                    {
		                if (interactive != 0)
		                    free_vartype(ref newx);
		                
                        array = null;
		                
                        return ERR_INSUFFICIENT_MEMORY;
	                }
	                
                    for (i = 0; i < 2 * matedit_i * columns; i++)
		                array.data[i] = cm.array.data[i];
	                
                    for (i = 2 * matedit_i * columns; i < 2 * (matedit_i + 1) * columns; i++)
                            array.data[i] = 0;

                    for (i = 2 * (matedit_i + 1) * columns; i < 2 * newsize; i++)
		                array.data[i] = cm.array.data[i - 2 * columns];
	                
                    array.refcount = 1;
	                cm.array.refcount--;
	                cm.array = array;
	                cm.rows++;
	            }
            }

            if (interactive != 0)
            {
                free_vartype(ref reg_x);
	            reg_x = newx;
            }

            return ERR_NONE;
        }

        public static int docmd_invrt(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.COMPLEX)
                return ERR_INVALID_TYPE;
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return linalg_inv(reg_x, invrt_completion);
        }

        public static void invrt_completion(int error, vartype res)
        {
            if (error == ERR_NONE)
                unary_result(res);
        }

        public static int docmd_j_add(arg_struct arg)
        {
            int rows = 0;
            int columns = 0;
            int oldi = matedit_i;
            int oldj = matedit_j;
            
            int err = matedit_get_dim(ref rows, ref columns);
            
            if (err != ERR_NONE)
                return err;
            
            if (++matedit_j >= columns)
            {
                flags.matrix_edge_wrap = 1;
                matedit_j = 0;
                
                if (++matedit_i >= rows)
                {
                    flags.matrix_end_wrap = 1;
                    if (flags.grow != 0)
                    {
                        if (matedit_mode == 2)
                            err = dimension_array_ref(matedit_x, rows + 1, columns);
                        else
                            err = dimension_array(matedit_name, matedit_length, rows + 1, columns);
                        
                        if (err != ERR_NONE)
                        {
                            matedit_i = oldi;
                            matedit_j = oldj;
                            return err;
                        }
                        
                        matedit_i = rows;
                    }
                    else
                        matedit_i = 0;
                }
                else
                    flags.matrix_end_wrap = 0;
            }
            else
            {
                flags.matrix_edge_wrap = 0;
                flags.matrix_end_wrap = 0;
            }

            return ERR_NONE;
        }

        public static int docmd_j_sub(arg_struct arg)
        {
            int rows = 0;
            int columns = 0;
            
            int err = matedit_get_dim(ref rows, ref columns);
            
            if (err != ERR_NONE)
                return err;
            
            if (--matedit_j < 0)
            {
                flags.matrix_edge_wrap = 1;
                matedit_j = columns - 1;
                
                if (--matedit_i < 0)
                {
                    flags.matrix_end_wrap = 1;
                    matedit_i = rows - 1;
                }
                else
                    flags.matrix_end_wrap = 0;
            }
            else
            {
                flags.matrix_edge_wrap = 0;
                flags.matrix_end_wrap = 0;
            }

            return ERR_NONE;
        }

        public static int docmd_ln_1_x(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.REALMATRIX)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_ln_1_x, null);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int mappable_ln_1_x(Phloat x, ref Phloat y)
        {
            if (x <= -1)
                return ERR_INVALID_DATA;
            
            y = Phloat.log1p(x);
            
            return ERR_NONE;
        }

        public static int docmd_old(arg_struct arg)
        {
            return docmd_rclel(null);
        }

        public static int docmd_posa(arg_struct arg)
        {
            int pos = -1;
            vartype v;

            if (reg_x.type == TYPE.REAL)
            {
                Phloat x = ((vartype_real)reg_x).x;
                char c;
                int i;

                if (x < 0)
                    x = -x;
                
                if (x >= 256)
                    return ERR_INVALID_DATA;
                
                c = Phloat.to_char(x);
                
                for (i = 0; i < reg_alpha_length; i++)
                    if (reg_alpha[i] == c)
                    {
                        pos = i;
                        break;
                    }
            }
            else if (reg_x.type == TYPE.STRING)
            {
                vartype_string s = (vartype_string)reg_x;

                if (s.length != 0)
                {
                    int i, j;
                    
                    for (i = 0; i < reg_alpha_length - s.length; i++)
                    {
                        for (j = 0; j < s.length; j++)
                            if (reg_alpha[i + j] != s.text[j])
                                goto notfound;

                        pos = i;
                        break;

                    notfound: ;
                    }
                }
            }
            else
                return ERR_INVALID_TYPE;

            v = new_real(pos);
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            unary_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_putm(arg_struct arg)
        {
            vartype m;
            int i;
            int j;

            switch (matedit_mode)
            {
	            case 0:
	                return ERR_NONEXISTENT;

    	        case 1:
	            case 3:
	                m = recall_var(matedit_name, matedit_length);
	                break;
	        
                case 2:
	                m = matedit_x;
	                break;
	        
                default:
	                return ERR_INTERNAL_ERROR;
            }

            if (m == null)
	            return ERR_NONEXISTENT;

            if (m.type != TYPE.REALMATRIX && m.type != TYPE.COMPLEXMATRIX)
	            // Shouldn't happen, but could, as long as I don't
	            // implement matrix locking
	            //
	            return ERR_INVALID_TYPE;

            if (reg_x.type == TYPE.STRING)
	            return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.COMPLEX)
	            return ERR_INVALID_TYPE;

            if (m.type == TYPE.REALMATRIX)
            {
	            vartype_realmatrix src;
                vartype_realmatrix dst;
	        
                if (reg_x.type == TYPE.COMPLEXMATRIX)
	                return ERR_INVALID_TYPE;
	        
                src = (vartype_realmatrix) reg_x;
	            dst = (vartype_realmatrix) m;
	        
                if (src.rows + matedit_i > dst.rows || src.columns + matedit_j > dst.columns)
	                return ERR_DIMENSION_ERROR;

	            if (disentangle(m) == 0)
	                return ERR_INSUFFICIENT_MEMORY;

	            for (i = 0; i < src.rows; i++)
	                for (j = 0; j < src.columns; j++)
                    {
		                int n1 = i * src.columns + j;
		                int n2 = (i + matedit_i) * dst.columns + j + matedit_j;
		                dst.array.is_string[n2] = src.array.is_string[n1];
		                dst.array.data[n2] = src.array.data[n1];
	                }
	        
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.REALMATRIX)
            {
	            vartype_realmatrix src = (vartype_realmatrix) reg_x;
	            vartype_complexmatrix dst = (vartype_complexmatrix) m;

	            if (src.rows + matedit_i > dst.rows || src.columns + matedit_j > dst.columns)
	                return ERR_DIMENSION_ERROR;

	            for (i = 0; i < src.rows * src.columns; i++)
	                if (src.array.is_string[i] != 0)
		                return ERR_ALPHA_DATA_IS_INVALID;
	        
                if (disentangle(m) == 0)
	                return ERR_INSUFFICIENT_MEMORY;

                for (i = 0; i < src.rows; i++)
	                for (j = 0; j < src.columns; j++)
                    {
        		        int n1 = i * src.columns + j;
		                int n2 = (i + matedit_i) * dst.columns + j + matedit_j;
		                dst.array.data[n2 * 2] = src.array.data[n1];
		                dst.array.data[n2 * 2 + 1] = 0;
	                }
	        
                return ERR_NONE;
            }
            else
            {
	            vartype_complexmatrix src = (vartype_complexmatrix) reg_x;
	            vartype_complexmatrix dst = (vartype_complexmatrix) m;

	            if (src.rows + matedit_i > dst.rows || src.columns + matedit_j > dst.columns)
	                return ERR_DIMENSION_ERROR;

	            if (disentangle(m) == 0)
	                return ERR_INSUFFICIENT_MEMORY;

	            for (i = 0; i < src.rows; i++)
	                for (j = 0; j < src.columns; j++)
                    {
		                int n1 = i * src.columns + j;
		                int n2 = (i + matedit_i) * dst.columns + j + matedit_j;
		                dst.array.data[n2 * 2] = src.array.data[n1 * 2];
		                dst.array.data[n2 * 2 + 1] = src.array.data[n1 * 2 + 1];
	                }

	            return ERR_NONE;
            }
        }

        public static int docmd_rclij(arg_struct arg)
        {
            vartype i;
            vartype j;

            if (matedit_mode == 0)
	            return ERR_NONEXISTENT;

            i = new_real(matedit_i + 1);
            j = new_real(matedit_j + 1);
            
            if (i == null || j == null)
            {
	            free_vartype(ref i);
	            free_vartype(ref j);

	            return ERR_INSUFFICIENT_MEMORY;
            }

            recall_two_results(ref j, ref i);
            
            return ERR_NONE;
        }

        public static int docmd_rnrm(arg_struct arg)
        {
            if (reg_x.type == TYPE.REALMATRIX)
            {
                vartype v;
                vartype_realmatrix rm = (vartype_realmatrix)reg_x;

                int size = rm.rows * rm.columns;
                int i;
                int j;

                Phloat max = new Phloat(0);

                for (i = 0; i < size; i++)
                    if (rm.array.is_string[i] != 0)
                        return ERR_ALPHA_DATA_IS_INVALID;
                
                for (i = 0; i < rm.rows; i++)
                {
                    Phloat nrm = new Phloat(0);

                    for (j = 0; j < rm.columns; j++)
                    {
                        Phloat x = rm.array.data[i * rm.columns + j];

                        if (x >= 0)
                            nrm += x;
                        else
                            nrm -= x;
                    }

                    if (Phloat.p_isinf(nrm) != 0)
                    {
                        if (flags.range_error_ignore != 0)
                            max = Phloat.POS_HUGE_PHLOAT;
                        else
                            return ERR_OUT_OF_RANGE;
                        break;
                    }

                    if (nrm > max)
                        max = nrm;
                }

                v = new_real(max);
                
                if (v == null)
                    return ERR_INSUFFICIENT_MEMORY;

                unary_result(v);
                
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.COMPLEXMATRIX)
            {
                vartype v;
                vartype_complexmatrix cm = (vartype_complexmatrix)reg_x;
                int i;
                int j;
                Phloat max = new Phloat(0);

                for (i = 0; i < cm.rows; i++)
                {
                    Phloat nrm = new Phloat(0);

                    for (j = 0; j < cm.columns; j++)
                    {
                        Phloat re = cm.array.data[2 * (i * cm.columns + j)];
                        Phloat im = cm.array.data[2 * (i * cm.columns + j) + 1];
                        
                        nrm += Phloat.hypot(re, im);
                    }

                    if (Phloat.p_isinf(nrm) != 0)
                    {
                        if (flags.range_error_ignore != 0)
                            max = Phloat.POS_HUGE_PHLOAT;
                        else
                            return ERR_OUT_OF_RANGE;
                        break;
                    }

                    if (nrm > max)
                        max = nrm;
                }

                v = new_real(max);
                
                if (v == null)
                    return ERR_INSUFFICIENT_MEMORY;
                
                unary_result(v);
                
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_rsum(arg_struct arg)
        {
            if (reg_x.type == TYPE.REALMATRIX)
            {
                vartype_realmatrix rm = (vartype_realmatrix)reg_x;
                vartype_realmatrix res;
                int size = rm.rows * rm.columns;
                int i;
                int j;

                for (i = 0; i < size; i++)
                    if (rm.array.is_string[i] != 0)
                        return ERR_ALPHA_DATA_IS_INVALID;

                res = (vartype_realmatrix)new_realmatrix(rm.rows, 1);
                
                if (res == null)
                    return ERR_INSUFFICIENT_MEMORY;
                
                for (i = 0; i < rm.rows; i++)
                {
                    Phloat sum = new Phloat(0);
                    int inf;

                    for (j = 0; j < rm.columns; j++)
                        sum += rm.array.data[i * rm.columns + j];
                    
                    if ((inf = Phloat.p_isinf(sum)) != 0)
                    {
                        if (flags.range_error_ignore != 0)
                            sum = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                        else
                        {
                            vartype tmp = (vartype)res;
                            free_vartype(ref tmp);

                            return ERR_OUT_OF_RANGE;
                        }
                    }

                    res.array.data[i] = sum;
                }

                unary_result((vartype)res);

                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.COMPLEXMATRIX)
            {
                vartype_complexmatrix cm = (vartype_complexmatrix)reg_x;
                vartype_complexmatrix res;
                int i;
                int j;
                
                res = (vartype_complexmatrix)new_complexmatrix(cm.rows, 1);
                
                if (res == null)
                    return ERR_INSUFFICIENT_MEMORY;

                for (i = 0; i < cm.rows; i++)
                {
                    Phloat sum_re = 0, sum_im = 0;
                    int inf;

                    for (j = 0; j < cm.columns; j++)
                    {
                        sum_re += cm.array.data[2 * (i * cm.columns + j)];
                        sum_im += cm.array.data[2 * (i * cm.columns + j) + 1];
                    }

                    if ((inf = Phloat.p_isinf(sum_re)) != 0)
                    {
                        if (flags.range_error_ignore != 0)
                            sum_re = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                        else
                        {
                            vartype tmp = (vartype)res;
                            free_vartype(ref tmp);

                            return ERR_OUT_OF_RANGE;
                        }
                    }

                    if ((inf = Phloat.p_isinf(sum_im)) != 0)
                    {
                        if (flags.range_error_ignore != 0)
                            sum_im = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                        else
                        {
                            vartype tmp = (vartype)res;
                            free_vartype(ref tmp);

                            return ERR_OUT_OF_RANGE;
                        }
                    }

                    res.array.data[2 * i] = sum_re;
                    res.array.data[2 * i + 1] = sum_im;
                }

                unary_result((vartype)res);

                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_swap_r(arg_struct arg)
        {
            vartype m;
            Phloat xx;
            Phloat yy;
            int x;
            int y;
            int i;

            switch (matedit_mode)
            {
                case 0:
                    return ERR_NONEXISTENT;
                
                case 1:
                case 3:
                    m = recall_var(matedit_name, matedit_length);
                    break;
                
                case 2:
                    m = matedit_x;
                    break;
                
                default:
                    return ERR_INTERNAL_ERROR;
            }

            if (m == null)
                return ERR_NONEXISTENT;

            if (m.type != TYPE.REALMATRIX && m.type != TYPE.COMPLEXMATRIX)
                // Should not happen, but could, as long as I don't implement
                // matrix locking.
                return ERR_INVALID_TYPE;

            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;
            
            if (reg_y.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_y.type != TYPE.REAL)
                return ERR_INVALID_TYPE;

            xx = ((vartype_real)reg_x).x;

            if (xx <= -2147483648.0 || xx >= 2147483648.0)
                return ERR_DIMENSION_ERROR;
            
            x = Phloat.to_int(xx);

            if (x == 0)
                return ERR_DIMENSION_ERROR;
            
            if (x < 0)
                x = -x;
            
            x--;

            yy = ((vartype_real)reg_y).x;

            if (yy <= -2147483648.0 || yy >= 2147483648.0)
                return ERR_DIMENSION_ERROR;
            
            y = Phloat.to_int(yy);
            
            if (y == 0)
                return ERR_DIMENSION_ERROR;
            
            if (y < 0)
                y = -y;
            
            y--;

            if (m.type == TYPE.REALMATRIX)
            {
                vartype_realmatrix rm = (vartype_realmatrix)m;

                if (x > rm.rows || y > rm.rows)
                    return ERR_DIMENSION_ERROR;
                else if (x == y)
                    return ERR_NONE;

                if (disentangle(m) == 0)
                    return ERR_INSUFFICIENT_MEMORY;

                for (i = 0; i < rm.columns; i++)
                {
                    int n1 = x * rm.columns + i;
                    int n2 = y * rm.columns + i;
                    char tempc = rm.array.is_string[n1];
                    Phloat tempds = rm.array.data[n1];

                    rm.array.is_string[n1] = rm.array.is_string[n2];
                    rm.array.data[n1] = rm.array.data[n2];
                    rm.array.is_string[n2] = tempc;
                    rm.array.data[n2] = tempds;
                }

                return ERR_NONE;
            }
            else // m.type == TYPE_COMPLEXMATRIX
            {
                vartype_complexmatrix cm = (vartype_complexmatrix)m;

                if (x > cm.rows || y > cm.rows)
                    return ERR_DIMENSION_ERROR;
                else if (x == y)
                    return ERR_NONE;

                if (disentangle(m) == 0)
                    return ERR_INSUFFICIENT_MEMORY;

                for (i = 0; i < 2 * cm.columns; i++)
                {
                    int n1 = x * 2 * cm.columns + i;
                    int n2 = y * 2 * cm.columns + i;
                    Phloat tempd = cm.array.data[n1];
                    
                    cm.array.data[n1] = cm.array.data[n2];
                    cm.array.data[n2] = tempd;
                }

                return ERR_NONE;
            }
        }

        public static int docmd_sinh(arg_struct arg)
        {
            if (reg_x.type != TYPE.STRING)
            {
                vartype v = new vartype();
                
                int err = map_unary(reg_x, ref v, mappable_sinh_r, mappable_sinh_c);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else
                return ERR_ALPHA_DATA_IS_INVALID;
        }

        public static int mappable_sinh_r(Phloat x, ref Phloat y)
        {
            int inf;
            
            y = Phloat.sinh(x);
            
            if ((inf = Phloat.p_isinf(y)) != 0)
            {
                if (flags.range_error_ignore != 0)
                    y = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            return ERR_NONE;
        }

        public static int mappable_sinh_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            Phloat sinhxre;
            Phloat coshxre;
            Phloat sinxim = new Phloat();
            Phloat cosxim = new Phloat();
            int inf;

            sinhxre = Phloat.sinh(xre);
            coshxre = Phloat.cosh(xre);
            
            Phloat.sincos(xim, ref sinxim, ref cosxim);
            yre = sinhxre * cosxim;
            
            if ((inf = Phloat.p_isinf(yre)) != 0)
            {
                if (flags.range_error_ignore != 0)
                    yre = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }
            
            yim = coshxre * sinxim;
            
            if ((inf = Phloat.p_isinf(yim)) != 0)
            {
                if (flags.range_error_ignore != 0)
                    yim = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            return ERR_NONE;
        }

        public static int docmd_stoij(arg_struct arg)
        {
            vartype m;
            Phloat x;
            Phloat y;
            int i;
            int j;

            switch (matedit_mode)
            {
                case 0:
                    return ERR_NONEXISTENT;
                
                case 1:
                case 3:
                    m = recall_var(matedit_name, matedit_length);
                    break;
                
                case 2:
                    m = matedit_x;
                    break;
                
                default:
                    return ERR_INTERNAL_ERROR;
            }

            if (m == null)
                return ERR_NONEXISTENT;

            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;
            
            if (reg_y.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_y.type != TYPE.REAL)
                return ERR_INVALID_TYPE;

            x = ((vartype_real)reg_x).x;

            if (x <= -2147483648.0 || x >= 2147483648.0)
                return ERR_DIMENSION_ERROR;
            
            j = Phloat.to_int(x);
            
            if (j < 0)
                j = -j;

            y = ((vartype_real)reg_y).x;

            if (y <= -2147483648.0 || y >= 2147483648.0)
                return ERR_DIMENSION_ERROR;
            
            i = Phloat.to_int(y);
            
            if (i < 0)
                i = -i;

            if (m.type == TYPE.REALMATRIX)
            {
                vartype_realmatrix rm = (vartype_realmatrix)m;

                if (i == 0 || i > rm.rows || j == 0 || j > rm.columns)
                    return ERR_DIMENSION_ERROR;
            }
            else if (m.type == TYPE.COMPLEXMATRIX)
            {
                vartype_complexmatrix cm = (vartype_complexmatrix)m;

                if (i == 0 || i > cm.rows || j == 0 || j > cm.columns)
                    return ERR_DIMENSION_ERROR;
            }
            else
                // Should not happen, but could, as long as I don't implement
                // matrix locking.
                return ERR_INVALID_TYPE;

            matedit_i = i - 1;
            matedit_j = j - 1;

            return ERR_NONE;
        }

        public static int docmd_tanh(arg_struct arg)
        {
            if (reg_x.type != TYPE.STRING)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_tanh_r, mappable_tanh_c);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else
                return ERR_ALPHA_DATA_IS_INVALID;
        }

        public static int mappable_tanh_r(Phloat x, ref Phloat y)
        {
            y = Phloat.tanh(x);
            
            return ERR_NONE;
        }

        public static int mappable_tanh_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            Phloat sinhxre;
            Phloat coshxre;
            Phloat sinxim = new Phloat();
            Phloat cosxim = new Phloat();
            Phloat re_sinh;
            Phloat re_cosh;
            Phloat im_sinh;
            Phloat im_cosh;
            Phloat abs_cosh;
            int inf;

            sinhxre = Phloat.sinh(xre);
            coshxre = Phloat.cosh(xre);

            Phloat.sincos(xim, ref sinxim, ref cosxim);

            re_sinh = sinhxre * cosxim;
            im_sinh = coshxre * sinxim;
            re_cosh = coshxre * cosxim;
            im_cosh = sinhxre * sinxim;
            abs_cosh = Phloat.hypot(re_cosh, im_cosh);

            if (abs_cosh == 0)
            {
                if (flags.range_error_ignore != 0)
                {
                    yre = re_sinh * im_sinh + re_cosh * im_cosh > 0 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                    yim = im_sinh * re_cosh - re_sinh * im_cosh > 0 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                }
                else
                    return ERR_OUT_OF_RANGE;
            }

            yre = (re_sinh * re_cosh + im_sinh * im_cosh) / abs_cosh / abs_cosh;

            if ((inf = Phloat.p_isinf(yre)) != 0)
            {
                if (flags.range_error_ignore != 0)
                    yre = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            yim = (im_sinh * re_cosh - re_sinh * im_cosh) / abs_cosh / abs_cosh;

            if ((inf = Phloat.p_isinf(yim)) != 0)
            {
                if (flags.range_error_ignore != 0)
                    yim = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            return ERR_NONE;
        }

        public static int docmd_trans(arg_struct arg)
        {
            if (reg_x.type == TYPE.REALMATRIX)
            {
                vartype_realmatrix src = (vartype_realmatrix)reg_x;
                vartype_realmatrix dst;

                int rows = src.rows;
                int columns = src.columns;
                int i;
                int j;

                dst = (vartype_realmatrix)new_realmatrix(columns, rows);
                
                if (dst == null)
                    return ERR_INSUFFICIENT_MEMORY;
                
                for (i = 0; i < rows; i++)
                    for (j = 0; j < columns; j++)
                    {
                        int n1 = i * columns + j;
                        int n2 = j * rows + i;
                        dst.array.is_string[n2] = src.array.is_string[n1];
                        dst.array.data[n2] = src.array.data[n1];
                    }

                unary_result((vartype)dst);

                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.COMPLEXMATRIX)
            {
                vartype_complexmatrix src = (vartype_complexmatrix)reg_x;
                vartype_complexmatrix dst;

                int rows = src.rows;
                int columns = src.columns;
                int i;
                int j;

                dst = (vartype_complexmatrix)new_complexmatrix(columns, rows);
                
                if (dst == null)
                    return ERR_INSUFFICIENT_MEMORY;

                for (i = 0; i < rows; i++)
                    for (j = 0; j < columns; j++)
                    {
                        int n1 = 2 * (i * columns + j);
                        int n2 = 2 * (j * rows + i);
                        dst.array.data[n2] = src.array.data[n1];
                        dst.array.data[n2 + 1] = src.array.data[n1 + 1];
                    }

                unary_result((vartype)dst);

                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_wrap(arg_struct arg)
        {
            flags.grow = 0;
            return ERR_NONE;
        }

        public static int docmd_x_swap(arg_struct arg)
        {
            vartype v = new vartype();
            int err = generic_rcl(arg, ref v);

            if (err != ERR_NONE)
                return err;
            
            err = generic_sto(arg, (char)0);
            
            if (err != ERR_NONE)
                free_vartype(ref v);
            else
            {
                free_vartype(ref reg_x);
                reg_x = v;
            
                if (flags.trace_print != 0 && flags.printer_exists != 0)
                    docmd_prx(null);
            }

            return err;
        }

        public static int docmd_left(arg_struct arg)
        {
            return matedit_move(DIR_LEFT);
        }

        public static int docmd_up(arg_struct arg)
        {
            return matedit_move(DIR_UP);
        }

        public static int docmd_down(arg_struct arg)
        {
            return matedit_move(DIR_DOWN);
        }

        public static int docmd_right(arg_struct arg)
        {
            return matedit_move(DIR_RIGHT);
        }

        public static int matedit_move(int direction)
        {
            vartype m;
            vartype v = new vartype();
            vartype_realmatrix rm = new vartype_realmatrix();
            vartype_complexmatrix cm = new vartype_complexmatrix();

            int rows;
            int columns;
            int new_i;
            int new_j;
            int old_n;
            int new_n;
            int edge_flag = 0;
            int end_flag = 0;

            switch (matedit_mode)
            {
	            case 0:
	                return ERR_NONEXISTENT;
	        
                case 1:
	            case 3:
	                m = recall_var(matedit_name, matedit_length);
	                break;
	            
                case 2:
	                m = matedit_x;
	                break;
	            
                default:
	                return ERR_INTERNAL_ERROR;
            }

            if (m == null)
	            return ERR_NONEXISTENT;

            if (m.type == TYPE.REALMATRIX)
            {
	            rm = (vartype_realmatrix) m;
	            rows = rm.rows;
	            columns = rm.columns;
            }
            else if (m.type == TYPE.COMPLEXMATRIX)
            {
	            cm = (vartype_complexmatrix) m;
	            rows = cm.rows;
	            columns = cm.columns;
            }
            else
	            return ERR_INVALID_TYPE;

            if (disentangle(m) == 0)
	            return ERR_INSUFFICIENT_MEMORY;

            new_i = matedit_i;
            new_j = matedit_j;
            
            switch (direction)
            {
	            case DIR_LEFT:
	                if (--new_j < 0)
                    {
		                edge_flag = 1;
		                new_j = columns - 1;
		                
                        if (--new_i < 0)
                        {
		                    end_flag = 1;
		                    new_i = rows - 1;
		                }
	                }
	                break;

	            case DIR_RIGHT:
	                if (++new_j >= columns)
                    {
		                edge_flag = 1;
		                new_j = 0;
		                
                        if (++new_i >= rows)
                        {
		                    end_flag = 1;
		                
                            if (flags.grow != 0)
                            {
			                    int err;
			                    if (matedit_mode == 2)
			                        err = dimension_array_ref(matedit_x, rows + 1, columns);
			                    else
			                        err = dimension_array(matedit_name, matedit_length, rows + 1, columns);
			                
                                if (err != ERR_NONE)
			                        return err;
			                
                                new_i = rows++;
		                    }
                            else
			                    new_i = 0;
		                }
	                }
	                break;

	            case DIR_UP:
	                if (--new_i < 0)
                    {
		                edge_flag = 1;
		                new_i = rows - 1;
		                
                        if (--new_j < 0)
                        {
		                    end_flag = 1;
		                    new_j = columns - 1;
		                }
	                }
	                break;

	            case DIR_DOWN:
	                if (++new_i >= rows)
                    {
		                edge_flag = 1;
		                new_i = 0;
		                
                        if (++new_j >= columns)
                        {
		                    end_flag = 1;
		                    new_j = 0;
		                }
	                }
	                break;
            }

            old_n = matedit_i * columns + matedit_j;
            new_n = new_i * columns + new_j;

            if (m.type == TYPE.REALMATRIX)
            {
	            if (old_n != new_n)
                {
	                if (rm.array.is_string[new_n] != 0)
		                v = new_string(Phloat.phloat_text_get(rm.array.data[new_n]), Phloat.phloat_length_get(rm.array.data[new_n]));
	                else
		                v = new_real(rm.array.data[new_n]);
	                    
                    if (v == null)
		                return ERR_INSUFFICIENT_MEMORY;
	            }
	        
                if (reg_x.type == TYPE.REAL)
                {
	                rm.array.is_string[old_n] = (char)0;
	                rm.array.data[old_n] = ((vartype_real) reg_x).x;
	            }
                else if (reg_x.type == TYPE.STRING)
                {
	                vartype_string s = (vartype_string) reg_x;
	                int i;

	                rm.array.is_string[old_n] = (char)1;
	                Phloat.phloat_length_set(rm.array.data[old_n], s.length);

	                for (i = 0; i < s.length; i++)
		                Phloat.phloat_text_set(rm.array.data[old_n], i, s.text[i]);
	            }
                else
                {
	                free_vartype(ref v);
	                return ERR_INVALID_TYPE;
	            }
            }
            else // m.type == TYPE_COMPLEXMATRIX
            {
        	    if (old_n != new_n)
                {
        	        v = new_complex(cm.array.data[2 * new_n], cm.array.data[2 * new_n + 1]);
	            
                    if (v == null)
		                return ERR_INSUFFICIENT_MEMORY;
	            }
	        
                if (reg_x.type == TYPE.REAL)
                {
	                cm.array.data[2 * old_n] = ((vartype_real) reg_x).x;
	                cm.array.data[2 * old_n + 1] = 0;
	            }
                else if (reg_x.type == TYPE.COMPLEX)
                {
	                vartype_complex c = (vartype_complex) reg_x;

	                cm.array.data[2 * old_n] = c.re;
	                cm.array.data[2 * old_n + 1] = c.im;
	            }
                else
                {
	                free_vartype(ref v);
	                return ERR_INVALID_TYPE;
	            }
            }

            matedit_i = new_i;
            matedit_j = new_j;

            flags.matrix_edge_wrap = (byte)edge_flag;
            flags.matrix_end_wrap = (byte)end_flag;
            
            if (old_n != new_n)
            {
	            free_vartype(ref reg_x);
	            reg_x = v;
            }

            mode_disable_stack_lift = true;
            
            if (flags.trace_print != 0 && flags.printer_enable != 0)
	            docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_percent_ch(arg_struct arg)
        {
            Phloat x;
            Phloat y;
            Phloat r;
            int inf;
            vartype v;

            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;
            
            if (reg_y.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_y.type != TYPE.REAL)
                return ERR_INVALID_TYPE;
            
            x = ((vartype_real)reg_x).x;
            y = ((vartype_real)reg_y).x;
            
            if (y == 0)
                return ERR_DIVIDE_BY_0;
            
            r = (x - y) / y * 100;
            
            if ((inf = Phloat.p_isinf(r)) != 0)
            {
                if (flags.range_error_ignore != 0)
                    r = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }
            
            v = new_real(r);
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            // Binary function, but unary result, like %
            unary_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_simq(arg_struct arg)
        {
            vartype m;
            vartype mata;
            vartype matb;
            vartype matx;
            
            int dim;
            int err;

            if (arg.type != ARGTYPE_NUM)
	            return ERR_INVALID_TYPE;

            dim = arg.val.num;
            
            if (dim <= 0)
	            return ERR_DIMENSION_ERROR;

            m = recall_var("MATA".ToCharArray(), 4);
            
            if (m == null)
            {
	            mata = new_realmatrix(dim, dim);
	            
                if (mata == null)
	                return ERR_INSUFFICIENT_MEMORY;
            }
            else
            {
	            mata = dup_vartype(m);
	        
                if (mata == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            
                err = dimension_array_ref(mata, dim, dim);
	            
                if (err != ERR_NONE)
	                return err;
            }

            m = recall_var("MATB".ToCharArray(), 4);

            if (m == null)
            {
	            matb = new_realmatrix(dim, 1);
	        
                if (matb == null)
	                return ERR_INSUFFICIENT_MEMORY;
            }
            else
            {
	            matb = dup_vartype(m);
	            
                if (matb == null)
	                return ERR_INSUFFICIENT_MEMORY;

	            err = dimension_array_ref(matb, dim, 1);

                if (err != ERR_NONE)
	                return err;
            }

            m = recall_var("MATX".ToCharArray(), 4);
            
            if (m == null)
            {
	            matx = new_realmatrix(dim, 1);

	            if (matx == null)
	                return ERR_INSUFFICIENT_MEMORY;
            }
            else
            {
	            matx = dup_vartype(m);

	            if (matx == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            
                err = dimension_array_ref(matx, dim, 1);
	            
                if (err != ERR_NONE)
	                return err;
            }

            err = set_menu_return_err(MENULEVEL_APP, MENU_MATRIX_SIMQ);
            
            if (err != ERR_NONE)
            {
	            // Didn't work; we're stuck in the matrix editor
	            // waiting for the user to put something valid into X.
	            // (Then again, how can anyone issue the SIMQ command if
	            // they're in the matrix editor? SIMQ has the 'hidden'
	            // command property. Oh, well, better safe than sorry...)
	            //
	            free_vartype(ref mata);
	            free_vartype(ref matb);
	            free_vartype(ref matx);
	            
                return err;
            }

            store_var("MATX".ToCharArray(), 4, matx);
            store_var("MATB".ToCharArray(), 4, matb);
            store_var("MATA".ToCharArray(), 4, mata);

            return ERR_NONE;
        }

        public static int docmd_mata(arg_struct arg)
        {
            return matabx(0);
        }

        public static int docmd_matb(arg_struct arg)
        {
            return matabx(1);
        }

        public static int docmd_matx(arg_struct arg)
        {
            return matabx(2);
        }

        public static int matabx(int which)
        {
            vartype mat = new vartype();
            vartype v;

            switch (which)
            {
	            case 0:
	                mat = recall_var("MATA".ToCharArray(), 4);
	                break;

	            case 1:
	                mat = recall_var("MATB".ToCharArray(), 4);
	                break;

	            case 2:
                    {
	                    vartype mata;
                        vartype matb;

	                    mata = recall_var("MATA".ToCharArray(), 4);

	                    if (mata == null)
		                    return ERR_NONEXISTENT;
	                    
                        if (mata.type != TYPE.REALMATRIX && mata.type != TYPE.COMPLEXMATRIX)
		                    return ERR_INVALID_TYPE;

	                    matb = recall_var("MATB".ToCharArray(), 4);

	                    if (matb == null)
                            return ERR_NONEXISTENT;

                        if (matb.type != TYPE.REALMATRIX && matb.type != TYPE.COMPLEXMATRIX)
		                    return ERR_INVALID_TYPE;

	                    if (mata.type == TYPE.REALMATRIX && matb.type == TYPE.REALMATRIX)
		                    matx_v = new_real(0);
	                    else
		                    matx_v = new_complex(0, 0);

	                    if (matx_v == null)
		                    return ERR_INSUFFICIENT_MEMORY;
	                    
                        return linalg_div(matb, mata, matx_completion);
	                }
            }

            if (mat.type == TYPE.REALMATRIX)
            {
	            vartype_realmatrix rm = (vartype_realmatrix) mat;

	            if (rm.array.is_string[0] != 0)
	                v = new_string(Phloat.phloat_text_get(rm.array.data[0]), Phloat.phloat_length_get(rm.array.data[0]));
	            else
	                v = new_real(rm.array.data[0]);
            }
            else
            {
	            vartype_complexmatrix cm = (vartype_complexmatrix) mat;
	            v = new_complex(cm.array.data[0], cm.array.data[1]);
            }
            
            if (v == null)
	            return ERR_INSUFFICIENT_MEMORY;

            matedit_prev_appmenu = MENU_MATRIX_SIMQ;
            set_menu(MENULEVEL_APP, MENU_MATRIX_EDIT1);
            // NOTE: no need to use set_menu_return_err() here, since the MAT[ABX]
            // commands can only be invoked from the SIMQ menu; the SIMQ menu
            // has no exit callback, so leaving it never fails.
            //
            set_appmenu_exitcallback(1);

            free_vartype(ref reg_x);
            reg_x = v;
            
            matedit_mode = 3;
            matedit_length = 4;
            matedit_name[0] = 'M';
            matedit_name[1] = 'A';
            matedit_name[2] = 'T';
            matedit_name[3] = which == 0 ? 'A' : 'B';
            matedit_i = 0;
            matedit_j = 0;
            
            return ERR_NONE;
        }

        public static void matx_completion(int error, vartype res)
        {
            if (error != ERR_NONE)
            {
                free_vartype(ref matx_v);
                return;
            }

            store_var("MATX".ToCharArray(), 4, res);

            matedit_prev_appmenu = MENU_MATRIX_SIMQ;
            set_menu(MENULEVEL_APP, MENU_MATRIX_EDIT1);
            // NOTE: no need to use set_menu_return_err() here, since the MAT[ABX]
            // commands can only be invoked from the SIMQ menu; the SIMQ menu
            // has no exit callback, so leaving it never fails.
            //
            set_appmenu_exitcallback(1);

            if (res.type == TYPE.REALMATRIX)
            {
                vartype_realmatrix m = (vartype_realmatrix)res;
                vartype_real v = (vartype_real)matx_v;
                v.x = m.array.data[0];
            }
            else
            {
                vartype_complexmatrix m = (vartype_complexmatrix)res;
                vartype_complex v = (vartype_complex)matx_v;
                v.re = m.array.data[0];
                v.im = m.array.data[1];
            }

            free_vartype(ref reg_x);
            reg_x = matx_v;

            matedit_mode = 3;
            matedit_length = 4;
            matedit_name[0] = 'M';
            matedit_name[1] = 'A';
            matedit_name[2] = 'T';
            matedit_name[3] = 'X';
            matedit_i = 0;
            matedit_j = 0;
        }

        public static int docmd_max(arg_struct arg)
        {
            return max_min_helper(1);
        }

        public static int max_min_helper(int do_max)
        {
            vartype m;
            vartype_realmatrix rm;

            Phloat max_or_min_value = do_max != 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;

            int i;
            int max_or_min_index = 0;
            
            vartype new_x;
            vartype new_y;

            switch (matedit_mode)
            {
	            case 0:
	                return ERR_NONEXISTENT;
	        
                case 1:
	            case 3:
	                m = recall_var(matedit_name, matedit_length);
	                break;
	            
                case 2:
	                m = matedit_x;
	                break;
	            
                default:
	                return ERR_INTERNAL_ERROR;
            }

            if (m == null)
	            return ERR_NONEXISTENT;

            if (m.type != TYPE.REALMATRIX)
	            return ERR_INVALID_TYPE;
            
            rm = (vartype_realmatrix) m;

            for (i = matedit_i; i < rm.rows; i++)
            {
	            int index = i * rm.columns + matedit_j;
	            Phloat e;
	            
                if (rm.array.is_string[index] != 0)
	                return ERR_ALPHA_DATA_IS_INVALID;

                e = rm.array.data[index];
	            
                if (do_max != 0 ? e >= max_or_min_value : e <= max_or_min_value)
                {
	                max_or_min_value = e;
	                max_or_min_index = i;
	            }
            }

            new_x = new_real(max_or_min_value);
            
            if (new_x == null)
	            return ERR_INSUFFICIENT_MEMORY;
            
            new_y = new_real(max_or_min_index + 1);
            
            if (new_y == null)
            {
	            free_vartype(ref new_x);
	            return ERR_INSUFFICIENT_MEMORY;
            }

            recall_two_results(ref new_x, ref new_y);
            
            return ERR_NONE;
        }

        public static int docmd_min(arg_struct arg)
        {
            return max_min_helper(0);
        }

        public static int docmd_find(arg_struct arg)
        {
            vartype m;

            if (reg_x.type == TYPE.REALMATRIX || reg_x.type == TYPE.COMPLEXMATRIX)
                return ERR_INVALID_TYPE;

            switch (matedit_mode)
            {
                case 0:
                    return ERR_NONEXISTENT;
            
                case 1:
                case 3:
                    m = recall_var(matedit_name, matedit_length);
                    break;
                
                case 2:
                    m = matedit_x;
                    break;
                
                default:
                    return ERR_INTERNAL_ERROR;
            }

            if (m == null)
                return ERR_NONEXISTENT;

            if (m.type == TYPE.REALMATRIX)
            {
                vartype_realmatrix rm;
                int i;
                int j;
                int p = 0;

                if (reg_x.type == TYPE.COMPLEX)
                    return ERR_NO;

                rm = (vartype_realmatrix)m;

                if (reg_x.type == TYPE.REAL)
                {
                    Phloat d = ((vartype_real)reg_x).x;

                    for (i = 0; i < rm.rows; i++)
                        for (j = 0; j < rm.columns; j++)
                            if (rm.array.is_string[p] == 0 && rm.array.data[p] == d)
                            {
                                matedit_i = i;
                                matedit_j = j;

                                return ERR_YES;
                            }
                            else
                                p++;
                }
                else // reg_x->type == TYPE_STRING
                {
                    vartype_string s = (vartype_string)reg_x;

                    for (i = 0; i < rm.rows; i++)
                        for (j = 0; j < rm.columns; j++)
                            if (rm.array.is_string[p] != 0 && string_equals(s.text, s.length,
                                Phloat.phloat_text_get(rm.array.data[p]), Phloat.phloat_length_get(rm.array.data[p])) != 0)
                            {
                                matedit_i = i;
                                matedit_j = j;
                                return ERR_YES;
                            }
                            else
                                p++;
                }
            }
            else // m->type == TYPE_COMPLEXMATRIX
            {
                vartype_complexmatrix cm;
                int i;
                int j;
                int p = 0;
                Phloat re;
                Phloat im;

                if (reg_x.type != TYPE.COMPLEX)
                    return ERR_NO;

                cm = (vartype_complexmatrix)m;
                re = ((vartype_complex)reg_x).re;
                im = ((vartype_complex)reg_x).im;
                
                for (i = 0; i < cm.rows; i++)
                    for (j = 0; j < cm.columns; j++)
                        if (cm.array.data[p] == re && cm.array.data[p + 1] == im)
                        {
                            matedit_i = i;
                            matedit_j = j;
                            return ERR_YES;
                        }
                        else
                            p += 2;
            }

            return ERR_NO;
        }

        public static int docmd_xrom(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
    }
}
