﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        public static int appmenu_exitcallback_2(int menuid)
        {
            if (menuid == MENU_BASE || menuid == MENU_BASE_A_THRU_F || menuid == MENU_BASE_LOGIC)
            {
                mode_appmenu = menuid;
                set_appmenu_exitcallback(2);
            }
            else
            {
                set_base(10);
                mode_appmenu = menuid;
                baseapp = 0;
            }

            return ERR_NONE;
        }

        public static int appmenu_exitcallback_3(int menuid)
        {
            if (menuid == (int)MENU.NONE)
            {
                set_menu(MENULEVEL_APP, MENU_CATALOG);
                set_cat_section(CATSECT_PGM_SOLVE);
            }
            else
                mode_appmenu = menuid;
            
            return ERR_NONE;
        }

        public static int appmenu_exitcallback_4(int menuid)
        {
            if (menuid == (int)MENU.NONE)
            {
                set_menu(MENULEVEL_APP, MENU_CATALOG);
                set_cat_section(CATSECT_PGM_INTEG);
            }
            else
                mode_appmenu = menuid;
            
            return ERR_NONE;
        }

        public static int appmenu_exitcallback_5(int menuid)
        {
            if (menuid == (int)MENU.NONE)
            {
                get_integ_prgm(varmenu, ref varmenu_length);

                varmenu_row = 0;
                varmenu_role = 2;
                
                set_menu(MENULEVEL_APP, MENU_VARMENU);
                set_appmenu_exitcallback(4);
            }
            else
                mode_appmenu = menuid;
            
            return ERR_NONE;
        }

        public static int sigma_helper_1(int weight)
        {
            // Check if summation registers are OK
            int first = mode_sigma_reg;
            int last = first + (flags.all_sigma != 0 ? 13 : 6);
            int size;
            int i;

            vartype regs = recall_var("REGS".ToCharArray(), 4);
            vartype_realmatrix r;
            // Phloat[] sigmaregs;

            if (regs == null)
                return ERR_SIZE_ERROR;
            
            if (regs.type != TYPE.REALMATRIX)
                return ERR_INVALID_TYPE;
            
            r = (vartype_realmatrix)regs;
            size = r.rows * r.columns;

            if (last > size)
                return ERR_SIZE_ERROR;
            
            for (i = first; i < last; i++)
                if (r.array.is_string[i] != 0)
                    return ERR_ALPHA_DATA_IS_INVALID;
            
            // sigmaregs = r.array.data + first;            Use first as offset parameter in sigma_helper_2 calls

            // All summation registers present, real-valued, non-string.
            switch (reg_x.type)
            {
                case TYPE.REAL:
                    {
                        if (reg_y.type == TYPE.REAL)
                        {
                            vartype_real x = (vartype_real)new_real(0);

                            if (x == null)
                                return ERR_INSUFFICIENT_MEMORY;

                            x.x = sigma_helper_2(r.array.data, first, ((vartype_real)reg_x).x, ((vartype_real)reg_y).x, weight);

                            free_vartype(ref reg_lastx);
                            reg_lastx = reg_x;
                            reg_x = (vartype)x;

                            mode_disable_stack_lift = true;
                            
                            return ERR_NONE;
                        }
                        else if (reg_y.type == TYPE.STRING)
                            return ERR_ALPHA_DATA_IS_INVALID;
                        else
                            return ERR_INVALID_TYPE;
                    }

                case TYPE.REALMATRIX:
                    {
                        vartype_realmatrix rm = (vartype_realmatrix)reg_x;
                        vartype_real x;
                        
                        if (rm.columns != 2)
                            return ERR_DIMENSION_ERROR;
                        
                        for (i = 0; i < rm.rows * 2; i++)
                            if (rm.array.is_string[i] != 0)
                                return ERR_ALPHA_DATA_IS_INVALID;

                        x = (vartype_real)new_real(0);
                        
                        if (x == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        
                        for (i = 0; i < rm.rows; i++)
                            x.x = sigma_helper_2(r.array.data, first, rm.array.data[i * 2], rm.array.data[i * 2 + 1], weight);

                        free_vartype(ref reg_lastx);
                        reg_lastx = reg_x;
                        reg_x = (vartype)x;

                        mode_disable_stack_lift = true;
                        
                        return ERR_NONE;
                    }

                case TYPE.STRING:
                    return ERR_ALPHA_DATA_IS_INVALID;

                default:
                    return ERR_INVALID_TYPE;
            }
        }

        public static void accum(ref Phloat sum, Phloat term, int weight)
        {
            int inf;
            Phloat s;

            if (weight == 1)
                s = sum + term;
            else
                s = sum - term;

            if ((inf = Phloat.p_isinf(s)) != 0)
                s = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
            
            sum = s;
        }

        public static Phloat sigma_helper_2(Phloat[] sigmaregs, int first, Phloat x, Phloat y, int weight)
        {

            accum(ref sigmaregs[0 + first], x, weight);
            accum(ref sigmaregs[1 + first], x * x, weight);
            accum(ref sigmaregs[2 + first], y, weight);
            accum(ref sigmaregs[3 + first], y * y, weight);
            accum(ref sigmaregs[4 + first], x * y, weight);
            accum(ref sigmaregs[5 + first], 1, weight);

            if (flags.all_sigma != 0)
            {
                if (x > 0)
                {
                    Phloat lnx = Phloat.log(x);

                    if (y > 0)
                    {
                        Phloat lny = Phloat.log(y);

                        accum(ref sigmaregs[8 + first], lny, weight);
                        accum(ref sigmaregs[9 + first], lny * lny, weight);
                        accum(ref sigmaregs[10 + first], lnx * lny, weight);
                        accum(ref sigmaregs[11 + first], x * lny, weight);
                    }
                    else
                    {
                        flags.exp_fit_invalid = 1;
                        flags.pwr_fit_invalid = 1;
                    }

                    accum(ref sigmaregs[6 + first], lnx, weight);
                    accum(ref sigmaregs[7 + first], lnx * lnx, weight);
                    accum(ref sigmaregs[12 + first], lnx * y, weight);
                }
                else
                {
                    if (y > 0)
                    {
                        Phloat lny = Phloat.log(y);

                        accum(ref sigmaregs[8 + first], lny, weight);
                        accum(ref sigmaregs[9 + first], lny * lny, weight);
                        accum(ref sigmaregs[11 + first], x * lny, weight);
                    }
                    else
                        flags.exp_fit_invalid = 1;

                    flags.log_fit_invalid = 1;
                    flags.pwr_fit_invalid = 1;
                }
            }
            else
            {
                flags.log_fit_invalid = 1;
                flags.exp_fit_invalid = 1;
                flags.pwr_fit_invalid = 1;
            }

            return sigmaregs[5 + first];
        }

        public static int docmd_sigmaadd(arg_struct arg)
        {
            int err = sigma_helper_1(1);

            if (err == ERR_NONE && flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);

            return err;
        }

        public static int docmd_sigmasub(arg_struct arg)
        {
            int err = sigma_helper_1(-1);

            if (err == ERR_NONE && flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
            
            return err;
        }

        public static int docmd_allsigma(arg_struct arg)
        {
            flags.all_sigma = 1;

            return ERR_NONE;
        }

        public static int docmd_and(arg_struct arg)
        {
            long x = 0;
            long y = 0;
            int err;
            vartype v;

            if ((err = get_base_param(reg_x, ref x)) != ERR_NONE)
                return err;

            if ((err = get_base_param(reg_y, ref y)) != ERR_NONE)
                return err;
            
            v = new_real((Phloat)(x & y));
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            binary_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_baseadd(arg_struct arg)
        {
            long x = 0;
            long y = 0;
            long res = 0;
            int err;
            vartype v;

            if ((err = get_base_param(reg_x, ref x)) != ERR_NONE)
                return err;
            
            if ((err = get_base_param(reg_y, ref y)) != ERR_NONE)
                return err;
            
            res = x + y;
            
            if ((err = base_range_check(ref res)) != ERR_NONE)
                return err;
            
            v = new_real(new Phloat(res));
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            binary_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_basesub(arg_struct arg)
        {
            long x = 0;
            long y = 0;
            long res = 0;
            int err;
            vartype v;

            if ((err = get_base_param(reg_x, ref x)) != ERR_NONE)
                return err;
            
            if ((err = get_base_param(reg_y, ref y)) != ERR_NONE)
                return err;
            
            res = y - x;
            
            if ((err = base_range_check(ref res)) != ERR_NONE)
                return err;
            
            v = new_real(new Phloat(res));
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            binary_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_basemul(arg_struct arg)
        {
            long x = 0;
            long y = 0;
            double res = 0;
            int err;
            vartype v;

            if ((err = get_base_param(reg_x, ref x)) != ERR_NONE)
                return err;
            
            if ((err = get_base_param(reg_y, ref y)) != ERR_NONE)
                return err;
            
            // I compute the result in 'double' arithmetic, because doing it
            // in int8 arithmetic could cause me to overlook an out-of-range
            // condition (e.g. 2^32 * 2^32).
            //
            res = ((double)x) * ((double)y);
            
            if (res < -34359738368.0)
            {
                if (flags.range_error_ignore != 0)
                    res = -34359738368.0;
                else
                    return ERR_OUT_OF_RANGE;
            }
            else if (res > 34359738367.0)
            {
                if (flags.range_error_ignore != 0)
                    res = 34359738367.0;
                else
                    return ERR_OUT_OF_RANGE;
            }
            
            v = new_real(res);
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            binary_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_basediv(arg_struct arg)
        {
            long x = 0;
            long y = 0;
            long res = 0;
            int err;
            vartype v;

            if ((err = get_base_param(reg_x, ref x)) != ERR_NONE)
                return err;
            
            if ((err = get_base_param(reg_y, ref y)) != ERR_NONE)
                return err;
            
            if (x == 0)
                return ERR_DIVIDE_BY_0;
            
            res = y / x;
            
            if ((err = base_range_check(ref res)) != ERR_NONE)
                return err;
            
            v = new_real(new Phloat(res));
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            binary_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_basechs(arg_struct arg)
        {
            long x = 0;
            int err;

            if ((err = get_base_param(reg_x, ref x)) != ERR_NONE)
                return err;
            
            if (x == -34359738368)
            {
                if (flags.range_error_ignore != 0)
                    x = 34359738367;
                else
                    return ERR_OUT_OF_RANGE;
            }
            else
                x = -x;
            
            ((vartype_real)reg_x).x = new Phloat(x);
            
            return ERR_NONE;
        }

        public static int docmd_best(arg_struct arg)
        {
            int best = MODEL_NONE;
            Phloat bestr = new Phloat(0);
            int firsterr = ERR_NONE;
            int i;

            for (i = MODEL_LIN; i <= MODEL_PWR; i++)
            {
                Phloat r = new Phloat();
                int err = corr_helper(i, ref r);

                if (err == ERR_NONE)
                {
                    if (r < 0)
                        r = -r;
                
                    if (r > bestr)
                    {
                        best = i;
                        bestr = r;
                    }
                }
                else
                {
                    if (firsterr == ERR_NONE)
                        firsterr = err;
                }
            }
            
            if (best == MODEL_NONE)
                best = MODEL_LIN;
            else
                firsterr = ERR_NONE;
            
            flags.lin_fit = (byte)(best == MODEL_LIN ? 1 : 0);
            flags.log_fit = (byte)(best == MODEL_LOG ? 1 : 0);
            flags.exp_fit = (byte)(best == MODEL_EXP ? 1 : 0);
            flags.pwr_fit = (byte)(best == MODEL_PWR ? 1 : 0);
            
            return firsterr;
        }

        public static int corr_helper(int modl, ref Phloat r)
        {
            Phloat cov;
            Phloat varx;
            Phloat vary;
            Phloat v;
            Phloat tr;

            int err = get_model_summation(modl);
            
            if (err != ERR_NONE)
                return err;
            
            if (model.n == 0 || model.n == 1)
                return ERR_STAT_MATH_ERROR;
            
            cov = model.xy - model.x * model.y / model.n;
            varx = model.x2 - model.x * model.x / model.n;
            vary = model.y2 - model.y * model.y / model.n;
            
            if (varx <= 0 || vary <= 0)
                return ERR_STAT_MATH_ERROR;
            
            v = varx * vary;
            
            if (v == 0)
                return ERR_STAT_MATH_ERROR;
            
            tr = cov / Phloat.sqrt(v);
            
            if (tr < -1)
                tr = -1;
            else if (tr > 1)
                tr = 1;
            
            r = tr;
            
            return ERR_NONE;
        }

        public static int get_model_summation(int modl)
        {
            int err = get_summation();
         
            if (err != ERR_NONE)
                return err;
            
            switch (modl)
            {
                case MODEL_LIN:
                    model.xy = sum.xy;
                    model.ln_before = 0;
                    model.exp_after = 0;
                    break;
            
                case MODEL_LOG:
                    if (flags.log_fit_invalid != 0)
                        return ERR_INVALID_FORECAST_MODEL;
                
                    model.xy = sum.ylnx;
                    model.ln_before = 1;
                    model.exp_after = 0;
                    break;
                
                case MODEL_EXP:
                    if (flags.exp_fit_invalid != 0)
                        return ERR_INVALID_FORECAST_MODEL;
                
                    model.xy = sum.xlny;
                    model.ln_before = 0;
                    model.exp_after = 1;
                    break;
                
                case MODEL_PWR:
                    if (flags.pwr_fit_invalid != 0)
                        return ERR_INVALID_FORECAST_MODEL;
                
                    model.xy = sum.lnxlny;
                    model.ln_before = 1;
                    model.exp_after = 1;
                    break;
                
                default:
                    return ERR_INVALID_FORECAST_MODEL;
            }

            if (model.ln_before != 0)
            {
                model.x = sum.lnx;
                model.x2 = sum.lnx2;
            }
            else
            {
                model.x = sum.x;
                model.x2 = sum.x2;
            }

            if (model.exp_after != 0)
            {
                model.y = sum.lny;
                model.y2 = sum.lny2;
            }
            else
            {
                model.y = sum.y;
                model.y2 = sum.y2;
            }

            model.n = sum.n;
            
            return ERR_NONE;
        }

        public static int get_summation()
        {
            // Check if summation registers are OK
            int first = mode_sigma_reg;
            int last = first + (flags.all_sigma != 0 ? 13 : 6);
            int size;
            int i;

            vartype regs = recall_var("REGS".ToCharArray(), 4);
            vartype_realmatrix r;
            
            Phloat[] sigmaregs;
            
            if (regs == null)
                return ERR_SIZE_ERROR;

            if (regs.type != TYPE.REALMATRIX)
                return ERR_INVALID_TYPE;
            
            r = (vartype_realmatrix)regs;
            size = r.rows * r.columns;
            
            if (last > size)
                return ERR_SIZE_ERROR;
            
            for (i = first; i < last; i++)
                if (r.array.is_string[i] != 0)
                    return ERR_ALPHA_DATA_IS_INVALID;

            sigmaregs = r.array.data;       // use "+first" as offset
            
            sum.x = sigmaregs[0 + first];
            sum.x2 = sigmaregs[1 + first];
            sum.y = sigmaregs[2 + first];
            sum.y2 = sigmaregs[3 + first];
            sum.xy = sigmaregs[4 + first];
            sum.n = sigmaregs[5 + first];

            if (flags.all_sigma != 0)
            {
                sum.lnx = sigmaregs[6 + first];
                sum.lnx2 = sigmaregs[7 + first];
                sum.lny = sigmaregs[8 + first];
                sum.lny2 = sigmaregs[9 + first];
                sum.lnxlny = sigmaregs[10 + first];
                sum.xlny = sigmaregs[11 + first];
                sum.ylnx = sigmaregs[12 + first];
            }

            return ERR_NONE;
        }

        public static int docmd_binm(arg_struct arg)
        {
            return base_helper(2);
        }

        public static int base_helper(int _base)
        {
            if (program_running())
            {
	            int err = set_menu_return_err(MENULEVEL_APP, MENU_BASE);
	        
                if (err != ERR_NONE)
	                return err;
	        
                set_appmenu_exitcallback(2);
	        
                baseapp = 1;
            }
            
            set_base(_base);
            
            return ERR_NONE;
        }

        public static int docmd_bit_t(arg_struct arg)
        {
            long x = 0;
            long y = 0;
            int err;
            
            if ((err = get_base_param(reg_x, ref x)) != ERR_NONE)
                return err;
            
            if ((err = get_base_param(reg_y, ref y)) != ERR_NONE)
                return err;
            
            if (x < 0 || x > 35)
                return ERR_INVALID_DATA;
            
            return (y & ((long)1 << (int)x)) != 0 ? ERR_YES : ERR_NO;
        }

        public static int docmd_corr(arg_struct arg)
        {
            Phloat r = new Phloat();
            int err;
            vartype rv;

            err = corr_helper(get_model(), ref r);
            
            if (err != ERR_NONE)
                return err;
            
            rv = new_real(r);
            
            if (rv == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            recall_result(rv);
            
            return ERR_NONE;
        }

        public static int get_model()
        {
            if (flags.lin_fit != 0)
                return MODEL_LIN;
            else if (flags.log_fit != 0)
                return MODEL_LOG;
            else if (flags.exp_fit != 0)
                return MODEL_EXP;
            else if (flags.pwr_fit != 0)
                return MODEL_PWR;
            else
                return MODEL_NONE;
        }

        public static int docmd_decm(arg_struct arg)
        {
            return base_helper(10);
        }

        public static int docmd_expf(arg_struct arg)
        {
            flags.lin_fit = 0;
            flags.log_fit = 0;
            flags.exp_fit = 1;
            flags.pwr_fit = 0;
            
            return ERR_NONE;
        }

        public static int docmd_fcstx(arg_struct arg)
        {
            int err = get_model_summation(get_model());
            
            vartype v = new vartype();
            
            if (err != ERR_NONE)
                return err;
            
            err = slope_yint_helper();
            
            if (err != ERR_NONE)
                return err;
            
            if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.REALMATRIX)
            {
                err = map_unary(reg_x, ref v, mappable_fcstx, null);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int slope_yint_helper()
        {
            // The caller should have made sure that 'model' is up to date
            // by calling get_model_summation() first.
            //
            Phloat cov;
            Phloat varx;
            Phloat meanx;
            Phloat meany;
            int inf;

            if (model.n == 0 || model.n == 1)
                return ERR_STAT_MATH_ERROR;
            
            cov = model.xy - model.x * model.y / model.n;
            varx = model.x2 - model.x * model.x / model.n;
            
            if (varx == 0)
                return ERR_STAT_MATH_ERROR;
            
            model.slope = cov / varx;
            
            if ((inf = Phloat.p_isinf(model.slope)) != 0)
                model.slope = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
            
            meanx = model.x / model.n;
            meany = model.y / model.n;
            model.yint = meany - model.slope * meanx;
            
            if ((inf = Phloat.p_isinf(model.yint)) != 0)
                model.yint = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
            
            return ERR_NONE;
        }

        public static int mappable_fcstx(Phloat x, ref Phloat y)
        {
            int inf;
            
            if (model.exp_after != 0)
            {
                if (x <= 0)
                    return ERR_INVALID_FORECAST_MODEL;
                
                x = Phloat.log(x);
            }

            if (model.slope == 0)
                return ERR_STAT_MATH_ERROR;
            
            x = (x - model.yint) / model.slope;
            
            if (model.ln_before != 0)
                x = Phloat.exp(x);
            
            if ((inf = Phloat.p_isinf(x)) != 0)
                x = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
            
            y = x;
            
            return ERR_NONE;
        }

        public static int docmd_fcsty(arg_struct arg)
        {
            int err = get_model_summation(get_model());
            vartype v = new vartype();

            if (err != ERR_NONE)
                return err;
            
            err = slope_yint_helper();
            
            if (err != ERR_NONE)
                return err;
            
            if (reg_x.type == TYPE.REAL || reg_x.type == TYPE.REALMATRIX)
            {
                err = map_unary(reg_x, ref v, mappable_fcsty, null);
            
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;
        }

        public static int mappable_fcsty(Phloat x, ref Phloat y)
        {
            int inf;

            if (model.ln_before != 0)
            {
                if (x <= 0)
                    return ERR_INVALID_FORECAST_MODEL;

                x = Phloat.log(x);
            }

            x = x * model.slope + model.yint;
            
            if (model.exp_after != 0)
                x = Phloat.exp(x);
            
            if ((inf = Phloat.p_isinf(x)) != 0)
                x = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
            
            y = x;
            
            return ERR_NONE;
        }

        public static int docmd_hexm(arg_struct arg)
        {
            return base_helper(16);
        }

        public static int docmd_integ(arg_struct arg)
        {
            int err;
            if (arg.type == ARGTYPE_IND_NUM || arg.type == ARGTYPE_IND_STK || arg.type == ARGTYPE_IND_STR)
            {
                err = resolve_ind_arg(arg);
                
                if (err != ERR_NONE)
                    return err;
            }

            if (arg.type != ARGTYPE_STR)
                return ERR_INVALID_TYPE;
            
            if (!program_running())
                clear_all_rtns();
            
            string_copy(reg_alpha, ref reg_alpha_length, arg.val.text, arg.length);

            return start_integ(arg.val.text, arg.length);
        }

        public static int docmd_linf(arg_struct arg)
        {
            flags.lin_fit = 1;
            flags.log_fit = 0;
            flags.exp_fit = 0;
            flags.pwr_fit = 0;
            
            return ERR_NONE;
        }

        public static int docmd_linsigma(arg_struct arg)
        {
            flags.all_sigma = 0;
            return ERR_NONE;
        }

        public static int docmd_logf(arg_struct arg)
        {
            flags.lin_fit = 0;
            flags.log_fit = 1;
            flags.exp_fit = 0;
            flags.pwr_fit = 0;
            
            return ERR_NONE;
        }

        public static int docmd_mean(arg_struct arg)
        {
            Phloat m;
            int inf;
            vartype mx;
            vartype my;

            int err = get_summation();
            
            if (err != ERR_NONE)
	            return err;
            
            if (sum.n == 0)
	            return ERR_STAT_MATH_ERROR;

            m = sum.x / sum.n;
            
            if ((inf = Phloat.p_isinf(m)) != 0)
	            m = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
            
            mx = new_real(m);
            
            if (mx == null)
	            return ERR_INSUFFICIENT_MEMORY;
            
            m = sum.y / sum.n;
            
            if ((inf = Phloat.p_isinf(m)) != 0)
                m = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
            
            my = new_real(m);
            
            if (my == null)
            {
	            free_vartype(ref mx);
	            return ERR_INSUFFICIENT_MEMORY;
            }

            free_vartype(ref reg_y);
            reg_y = my;
            free_vartype(ref reg_lastx);
            reg_lastx = reg_x;
            reg_x = mx;
            
            if (flags.trace_print != 0 && flags.printer_enable != 0)
	            docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_not(arg_struct arg)
        {
            long x = 0;
            int err;
            vartype v;

            if ((err = get_base_param(reg_x, ref x)) != ERR_NONE)
                return err;
            
            v = new_real(new Phloat(~x));
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            unary_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_octm(arg_struct arg)
        {
            return base_helper(8);
        }

        public static int docmd_or(arg_struct arg)
        {
            long x = 0;
            long y = 0;
            int err;
            vartype v;

            if ((err = get_base_param(reg_x, ref x)) != ERR_NONE)
                return err;
            
            if ((err = get_base_param(reg_y, ref y)) != ERR_NONE)
                return err;
            
            v = new_real(new Phloat(x | y));
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            binary_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_pgmslv(arg_struct arg)
        {
            int err;

            if (arg.type == ARGTYPE_IND_NUM || arg.type == ARGTYPE_IND_STK || arg.type == ARGTYPE_IND_STR)
            {
                err = resolve_ind_arg(arg);

                if (err != ERR_NONE)
                    return err;
            }
            if (arg.type == ARGTYPE_STR)
            {
                int prgm = 0;
                int pc = 0;

                if (find_global_label(arg, ref prgm, ref pc) == 0)
                    return ERR_LABEL_NOT_FOUND;
                
                set_solve_prgm(arg.val.text, arg.length);
                
                return ERR_NONE;
            }
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_pgmint(arg_struct arg)
        {
            int err;

            if (arg.type == ARGTYPE_IND_NUM || arg.type == ARGTYPE_IND_STK || arg.type == ARGTYPE_IND_STR)
            {
                err = resolve_ind_arg(arg);

                if (err != ERR_NONE)
                    return err;
            }

            if (arg.type == ARGTYPE_STR)
            {
                int prgm = 0;
                int pc = 0;
            
                if (find_global_label(arg, ref prgm, ref pc) == 0)
                    return ERR_LABEL_NOT_FOUND;
                
                set_integ_prgm(arg.val.text, arg.length);
                
                return ERR_NONE;
            }
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_pwrf(arg_struct arg)
        {
            flags.lin_fit = 0;
            flags.log_fit = 0;
            flags.exp_fit = 0;
            flags.pwr_fit = 1;
            
            return ERR_NONE;
        }

        public static int docmd_sdev(arg_struct arg)
        {
            int err = get_summation();
        
            Phloat _var;
            vartype sx;
            vartype sy;

            if (err != ERR_NONE)
	            return err;

            if (sum.n == 0 || sum.n == 1)
	            return ERR_STAT_MATH_ERROR;

            _var = (sum.x2 - (sum.x * sum.x / sum.n)) / (sum.n - 1);
            
            if (_var < 0)
	            return ERR_STAT_MATH_ERROR;

            if (Phloat.p_isinf(_var) != 0)
	            sx = new_real(Phloat.POS_HUGE_PHLOAT);
            else
	            sx = new_real(Phloat.sqrt(_var));

            if (sx == null)
	            return ERR_INSUFFICIENT_MEMORY;
            
            _var = (sum.y2 - (sum.y * sum.y / sum.n)) / (sum.n - 1);
            
            if (_var < 0)
	            return ERR_STAT_MATH_ERROR;

            if (Phloat.p_isinf(_var) != 0)
	            sy = new_real(Phloat.POS_HUGE_PHLOAT);
            else
	            sy = new_real(Phloat.sqrt(_var));
            
            if (sy == null)
            {
	            free_vartype(ref sx);
	            return ERR_INSUFFICIENT_MEMORY;
            }

            free_vartype(ref reg_y);
            reg_y = sy;
            free_vartype(ref reg_lastx);
            reg_lastx = reg_x;
            reg_x = sx;
            
            if (flags.trace_print != 0 && flags.printer_enable != 0)
	            docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_slope(arg_struct arg)
        {
            int err = get_model_summation(get_model());
            vartype v;

            if (err != ERR_NONE)
                return err;

            err = slope_yint_helper();
            
            if (err != ERR_NONE)
                return err;
            
            v = new_real(model.slope);
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            recall_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_solve(arg_struct arg)
        {
            int err;
            vartype v;
            Phloat x1;
            Phloat x2;

            if (arg.type == ARGTYPE_IND_NUM || arg.type == ARGTYPE_IND_STK || arg.type == ARGTYPE_IND_STR)
            {
                err = resolve_ind_arg(arg);

                if (err != ERR_NONE)
                    return err;
            }

            if (arg.type != ARGTYPE_STR)
                return ERR_INVALID_TYPE;

            v = recall_var(arg.val.text, arg.length);
            
            if (v == null)
                x1 = 0;
            else if (v.type == TYPE.REAL)
                x1 = ((vartype_real)v).x;
            else if (v.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;

            if (reg_x.type == TYPE.REAL)
                x2 = ((vartype_real)reg_x).x;
            else if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;

            if (!program_running())
                clear_all_rtns();
            
            string_copy(reg_alpha, ref reg_alpha_length, arg.val.text, arg.length);
            
            return start_solve(arg.val.text, arg.length, x1, x2);
        }

        public static int docmd_sum(arg_struct arg)
        {
            int err = get_summation();
        
            vartype sx;
            vartype sy;

            if (err != ERR_NONE)
	            return err;
            
            sx = new_real(sum.x);
            
            if (sx == null)
	            return ERR_INSUFFICIENT_MEMORY;
            
            sy = new_real(sum.y);
            
            if (sy == null)
            {
	            free_vartype(ref sx);
	            return ERR_INSUFFICIENT_MEMORY;
            }

            free_vartype(ref reg_lastx);
            free_vartype(ref reg_y);
            reg_y = sy;
            reg_lastx = reg_x;
            reg_x = sx;
            
            if (flags.trace_print != 0 && flags.printer_enable != 0)
	            docmd_prx(null);
            
            return ERR_NONE;
        }

        public static int docmd_wmean(arg_struct arg)
        {
            Phloat wm;
            int inf;
            vartype v;
            
            int err = get_summation();
            
            if (err != ERR_NONE)
                return err;
            
            if (sum.y == 0)
                return ERR_STAT_MATH_ERROR;
            
            wm = sum.xy / sum.y;
            
            if ((inf = Phloat.p_isinf(wm)) != 0)
                wm = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
            
            v = new_real(wm);
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            recall_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_xor(arg_struct arg)
        {
            long x = 0;
            long y = 0;
            int err;
            vartype v;

            if ((err = get_base_param(reg_x, ref x)) != ERR_NONE)
                return err;
            
            if ((err = get_base_param(reg_y, ref y)) != ERR_NONE)
                return err;
            
            v = new_real(new Phloat(x ^ y));
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            binary_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_yint(arg_struct arg)
        {
            int err = get_model_summation(get_model());
            Phloat yint;
            vartype v;

            if (err != ERR_NONE)
                return err;
            
            err = slope_yint_helper();
            
            if (err != ERR_NONE)
                return err;
            
            if (model.exp_after != 0)
            {
                yint = Phloat.exp(model.yint);

                if (Phloat.p_isinf(yint) != 0)
                    yint = Phloat.POS_HUGE_PHLOAT;
            }
            else
                yint = model.yint;
            
            v = new_real(yint);
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            recall_result(v);
            
            return ERR_NONE;
        }

        public static int docmd_to_dec(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL)
            {
	            Phloat oct = ((vartype_real) reg_x).x;
	            Phloat res;
	        
                bool neg = oct < 0;
	        
                if (neg)
	                oct = -oct;
	        
                if (oct > 777777777777.0 || oct != Phloat.floor(oct))
	                return ERR_INVALID_DATA;
	        
                vartype v;

                Phloat dec = new Phloat(0);
                Phloat mul = new Phloat(1);

	            while (oct != 0)
                {
		            int digit = Phloat.to_digit(oct);

		            if (digit > 7)
		                return ERR_INVALID_DATA;
		            
                    oct = Phloat.floor(oct / 10);
		            dec += digit * mul;
		            mul *= 8;
	            }

	            res = neg ? -dec : dec;
                v = new_real(res);

                if (v == null)
	                return ERR_INSUFFICIENT_MEMORY;
	        
                unary_result(v);
	        
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
	            return ERR_ALPHA_DATA_IS_INVALID;
            else
	            return ERR_INVALID_TYPE;
        }

        public static int docmd_to_oct(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL)
            {
	            Phloat dec = ((vartype_real) reg_x).x;
	            Phloat res;
	            
                bool neg = dec < 0;
	            
                if (neg)
	                dec = -dec;
	            
                if (dec > 68719476735.0 || dec != Phloat.floor(dec))
	                return ERR_INVALID_DATA;
	            
                vartype v;

	            Phloat oct = new Phloat(0);
                Phloat mul = new Phloat(1);

	            while (dec != 0)
                {
		            int digit = Phloat.to_int(Phloat.fmod(dec, 8));
		            dec = Phloat.floor(dec / 8);
		            oct += digit * mul;
		            mul *= 10;
	            }

	            res = neg ? -oct : oct;
	            v = new_real(res);
	            
                if (v == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            
                unary_result(v);
	            
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
	            return ERR_ALPHA_DATA_IS_INVALID;
            else
	            return ERR_INVALID_TYPE;
        }

        public static int docmd_pgmslvi(arg_struct arg)
        {
            // This command can only be invoked from a menu; we assume that
            // the menu handler only gives us valid arguments. We do check
            // the argument type, but the existence of the named label, and
            // whether it actually has MVAR instructions, we just assume.
            //
            if (arg.type == ARGTYPE_STR)
            {
                set_solve_prgm(arg.val.text, arg.length);

                string_copy(varmenu, ref varmenu_length, arg.val.text, arg.length);
                
                varmenu_row = 0;
                varmenu_role = 1;
                
                set_menu(MENULEVEL_APP, MENU_VARMENU);
                set_appmenu_exitcallback(3);
                
                return ERR_NONE;
            }
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_pgminti(arg_struct arg)
        {
            // This command can only be invoked from a menu; we assume that
            // the menu handler only gives us valid arguments. We do check
            // the argument type, but the existence of the named label, and
            // whether it actually has MVAR instructions, we just assume.
            //
            if (arg.type == ARGTYPE_STR)
            {
                set_integ_prgm(arg.val.text, arg.length);

                string_copy(varmenu, ref varmenu_length, arg.val.text, arg.length);

                varmenu_row = 0;
                varmenu_role = 2;
                
                set_menu(MENULEVEL_APP, MENU_VARMENU);
                set_appmenu_exitcallback(4);
                clear_row(0);
                draw_string(0, 0, "Set Vars; Select \x0003var".ToCharArray(), 21);
                
                flags.message = 1;
                flags.two_line_message = 0;
                mode_varmenu = true;
                
                return ERR_NONE;
            }
            else
                return ERR_INVALID_TYPE;
        }

        public static int docmd_vmsolve(arg_struct arg)
        {
            vartype v;
            Phloat x1 = new Phloat();
            Phloat x2 = new Phloat();

            if (arg.type != ARGTYPE_STR)
                return ERR_INVALID_TYPE;

            v = recall_var(arg.val.text, arg.length);
            
            if (v == null)
            {
                x1 = 0;
                x2 = 1;
            }
            else if (v.type == TYPE.REAL)
            {
                x1 = ((vartype_real)v).x;

                if (get_shadow(arg.val.text, arg.length, ref x2) == 0)
                    x2 = x1;
            }
            else if (v.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
                return ERR_INVALID_TYPE;

            clear_all_rtns();
            string_copy(reg_alpha, ref reg_alpha_length, arg.val.text, arg.length);
            
            return start_solve(arg.val.text, arg.length, x1, x2);
        }
    }
}
