﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        public static int docmd_sin(arg_struct arg)
        {
            if (reg_x.type != TYPE.STRING)
            {
                vartype v = new vartype();
                int err = map_unary(reg_x, ref v, mappable_sin_r, mappable_sin_c);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else
                return ERR_ALPHA_DATA_IS_INVALID;
        }

        public static int mappable_sin_r(Phloat x, ref Phloat y)
        {
            if (flags.rad != 0)
            {
	            y = Phloat.sin(x);
            }
            else if (flags.grad != 0)
            {
                x = Phloat.fmod(x, 400);

                if (x < 0)
	                x += 400;
	            
                if (x == 0 || x == 200)
	                y = 0;
	            else if (x == 100)
	                y = 1;
	            else if (x == 300)
	                y = -1;
	            else
                    y = Phloat.sin(x / (200 / PI));
            }
            else
            {
                x = Phloat.fmod(x, 360);
	        
                if (x < 0)
	                x += 360;
	            
                if (x == 0 || x == 180)
	                y = 0;
	            else if (x == 90)
	                y = 1;
	            else if (x == 270)
	                y = -1;
	            else
                    y = Phloat.sin(x / (180 / PI));
            }

            return ERR_NONE;
        }

        public static int mappable_sin_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            // NOTE: DEG/RAD/GRAD mode does not apply here.
            Phloat sinxre = new Phloat();
            Phloat cosxre = new Phloat();
            Phloat sinhxim;
            Phloat coshxim;
            int inf;
            
            Phloat.sincos(xre, ref sinxre, ref cosxre);
            sinhxim = Phloat.sinh(xim);
            coshxim = Phloat.cosh(xim);
            
            yre = sinxre * coshxim;

            if ((inf = Phloat.p_isinf(yre)) != 0)
            {
	            if (flags.range_error_ignore != 0)
                    yre = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            else
	                return ERR_OUT_OF_RANGE;
            }
            
            yim = cosxre * sinhxim;

            if ((inf = Phloat.p_isinf(yim)) != 0)
            {
	            if (flags.range_error_ignore != 0)
                    yim = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            else
	                return ERR_OUT_OF_RANGE;
            }
            
            return ERR_NONE;
        }

        public static int docmd_cos(arg_struct arg)
        {
            if (reg_x.type != TYPE.STRING)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_cos_r, mappable_cos_c);
                
                if (err == ERR_NONE)
                    unary_result(v);

                return err;
            }
            else
                return ERR_ALPHA_DATA_IS_INVALID;
        }

        public static int mappable_cos_r(Phloat x, ref Phloat y)
        {
            if (flags.rad != 0)
            {
    	        y = Phloat.cos(x);
            }
            else if (flags.grad != 0)
            {
                x = Phloat.fmod(x, 400);

                if (x < 0)
	                x += 400;
	            
                if (x == 0)
	                y = 1;
	            else if (x == 100 || x == 300)
	                y = 0;
	            else if (x == 200)
	                y = -1;
	            else
                    y = Phloat.cos(x / (200 / PI));
            }
            else
            {
                x = Phloat.fmod(x, 360);
	        
                if (x < 0)
	                x += 360;
	            
                if (x == 0)
	                y = 1;
	            else if (x == 90 || x == 270)
	                y = 0;
	            else if (x == 180)
	                y = -1;
	            else
                    y = Phloat.cos(x / (180 / PI));
            }

            return ERR_NONE;
        }

        public static int mappable_cos_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            // NOTE: DEG/RAD/GRAD mode does not apply here.
            Phloat sinxre = new Phloat();
            Phloat cosxre = new Phloat();
            Phloat sinhxim;
            Phloat coshxim;
            int inf;

            Phloat.sincos(xre, ref sinxre, ref cosxre);
            sinhxim = Phloat.sinh(xim);
            coshxim = Phloat.cosh(xim);
            
            yre = cosxre * coshxim;

            if ((inf = Phloat.p_isinf(yre)) != 0)
            {
	            if (flags.range_error_ignore != 0)
                    yre = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            else
	                return ERR_OUT_OF_RANGE;
            }

            yim = -sinxre * sinhxim;

            if ((inf = Phloat.p_isinf(yim)) != 0)
            {
	            if (flags.range_error_ignore != 0)
                    yim = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            else
	                return ERR_OUT_OF_RANGE;
            }

            return ERR_NONE;
        }

        public static int docmd_tan(arg_struct arg)
        {
            if (reg_x.type != TYPE.STRING)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_tan_r, mappable_tan_c);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else
                return ERR_ALPHA_DATA_IS_INVALID;
        }

        public static int mappable_tan_r(Phloat x, ref Phloat y)
        {
            int inf = 1;

            if (flags.rad != 0)
            {
	            y = Phloat.tan(x);
            }
            else if (flags.grad != 0)
            {
	            x = Phloat.fmod(x, 200);
	        
                if (x < 0)
	                x += 200;
	            
                if (x == 0)
	                y = 0;
	            else if (x == 50)
	                y = 1;
	            else if (x == 100)
                {
	                //goto infinite;
	                if (flags.range_error_ignore != 0)
	                    y = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	                else
	                    return ERR_OUT_OF_RANGE;
            
                    return ERR_NONE;
                }
	            else if (x == 150)
	                y = -1;
	            else
	                y = Phloat.tan(x / (200 / PI));
            }
            else
            {
	            x = Phloat.fmod(x, 180);
	        
                if (x < 0)
	                x += 180;
	            
                if (x == 0)
	                y = 0;
	            else if (x == 45)
	                y = 1;
	            else if (x == 90)
                {
	                //goto infinite;
	                if (flags.range_error_ignore != 0)
	                    y = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	                else
	                    return ERR_OUT_OF_RANGE;
            
                    return ERR_NONE;
                }
	            else if (x == 135)
	                y = -1;
	            else
	                y = Phloat.tan(x / (180 / PI));
            }

            if (Phloat.p_isnan(y) != 0)
            {
	            //goto infinite;
	            if (flags.range_error_ignore != 0)
	                y = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            else
	                return ERR_OUT_OF_RANGE;
            
                return ERR_NONE;
            }

            if ((inf = Phloat.p_isinf(y)) != 0)
            {
	        infinite:
	            if (flags.range_error_ignore != 0)
	                y = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            else
	                return ERR_OUT_OF_RANGE;
            }
            
            return ERR_NONE;
        }

        public static int mappable_tan_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            // NOTE: DEG/RAD/GRAD mode does not apply here.
            Phloat sinxre = new Phloat();
            Phloat cosxre = new Phloat();
            Phloat sinhxim;
            Phloat coshxim;
            Phloat re_sin;
            Phloat im_sin;
            Phloat re_cos;
            Phloat im_cos;
            Phloat abs_cos;
            int inf;

            Phloat.sincos(xre, ref sinxre, ref cosxre);
            sinhxim = Phloat.sinh(xim);
            coshxim = Phloat.cosh(xim);

            re_sin = sinxre * coshxim;
            im_sin = cosxre * sinhxim;
            re_cos = cosxre * coshxim;
            im_cos = -sinxre * sinhxim;
            abs_cos = Phloat.hypot(re_cos, im_cos);

            if (abs_cos == 0)
            {
	            if (flags.range_error_ignore != 0)
                {
                    yre = re_sin * re_cos + im_sin * im_cos > 0 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                    yim = im_sin * re_cos - re_sin * im_cos > 0 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	                return ERR_NONE;
	            }
                else
	                return ERR_OUT_OF_RANGE;
            }

            yre = (re_sin * re_cos + im_sin * im_cos) / abs_cos / abs_cos;

            if ((inf = Phloat.p_isinf(yre)) != 0)
            {
	            if (flags.range_error_ignore != 0)
                    yre = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            else
	                return ERR_OUT_OF_RANGE;
            }

            yim = (im_sin * re_cos - re_sin * im_cos) / abs_cos / abs_cos;

            if ((inf = Phloat.p_isinf(yim)) != 0)
            {
	            if (flags.range_error_ignore != 0)
                    yim = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            else
	                return ERR_OUT_OF_RANGE;
            }

            return ERR_NONE;
        }

        public static int docmd_asin(arg_struct arg)
        {
            vartype v = new vartype();

            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_x.type == TYPE.REAL)
            {
                Phloat x = ((vartype_real)reg_x).x;

                if (x < -1 || x > 1)
                {
                    if (flags.real_result_only != 0)
                        return ERR_INVALID_DATA;
                    else
                    {
                        // (orig author note) TO DO: review
                        if (x > 0)
                            v = new_complex(PI / 2, -Phloat.acosh(x));
                        else
                            v = new_complex(-PI / 2, Phloat.acosh(-x));
                    }
                }
                else
                    v = new_real(rad_to_angle(Phloat.asin(x)));
                if (v == null)
                    return ERR_INSUFFICIENT_MEMORY;
            }
            else
            {
                int err = map_unary(reg_x, ref v, mappable_asin_r, mappable_asin_c);

                if (err != ERR_NONE)
                    return err;
            }

            unary_result(v);
            return ERR_NONE;
        }

        public static int mappable_asin_r(Phloat x, ref Phloat y)
        {
            if (x < -1 || x > 1)
	            return ERR_INVALID_DATA;
            
            y = rad_to_angle(Phloat.asin(x));
            
            return ERR_NONE;
        }

        public static int mappable_asin_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            return math_asinh_acosh(xim, xre, ref yim, ref yre, 1);
        }

        public static int docmd_acos(arg_struct arg)
        {
            vartype v = new vartype();

            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;

            if (reg_x.type == TYPE.REAL)
            {
                Phloat x = ((vartype_real)reg_x).x;

                if (x < -1 || x > 1)
                {
                    if (flags.real_result_only != 0)
                        return ERR_INVALID_DATA;
                    else
                    {
                        // (orig authors note) TO DO: review
                        if (x > 0)
                            v = new_complex(0, Phloat.acosh(x));
                        else
                            v = new_complex(PI, -Phloat.acosh(-x));
                    }
                }
                else
                    v = new_real(rad_to_angle(Phloat.acos(x)));
                
                if (v == null)
                    return ERR_INSUFFICIENT_MEMORY;
            }
            else
            {
                int err = map_unary(reg_x, ref v, mappable_acos_r, mappable_acos_c);
                
                if (err != ERR_NONE)
                    return err;
            }

            unary_result(v);
            return ERR_NONE;
        }

        public static int mappable_acos_r(Phloat x, ref Phloat y)
        {
            if (x < -1 || x > 1)
	            return ERR_INVALID_DATA;
            
            y = rad_to_angle(Phloat.acos(x));
            
            return ERR_NONE;
        }

        public static int mappable_acos_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            Phloat tre = new Phloat();
            Phloat tim = new Phloat();

            int err = math_asinh_acosh(xre, xim, ref tre, ref tim, 0);
            
            yre = tim;
            yim = -tre;
            
            return err;
        }

        public static int docmd_atan(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_atan_r, mappable_atan_c);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
        }

        public static int mappable_atan_r(Phloat x, ref Phloat y)
        {
            y = rad_to_angle(Phloat.atan(x));
            
            return ERR_NONE;
        }

        public static int mappable_atan_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            Phloat tre = new Phloat();
            Phloat tim = new Phloat();

            int err = math_atanh(xim, -xre, ref tre, ref tim);
            
            yre = -tim;
            yim = tre;
            
            return err;
        }

        public static int docmd_log(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type == TYPE.REAL)
            {
                vartype_real x = (vartype_real)reg_x;

                if (x.x == 0)
                    return ERR_INVALID_DATA;
                else if (x.x < 0)
                {
                    if (flags.real_result_only != 0)
                        return ERR_INVALID_DATA;
                    else
                    {
                        vartype r = new_complex(Phloat.log10(-x.x), PI / Phloat.log(10));
                        
                        if (r == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        else
                        {
                            unary_result(r);
                            return ERR_NONE;
                        }
                    }
                }
                else
                {
                    vartype r = new_real(Phloat.log10(x.x));

                    if (r == null)
                        return ERR_INSUFFICIENT_MEMORY;
                    else
                    {
                        unary_result(r);
                        return ERR_NONE;
                    }
                }
            }
            else
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_log_r, mappable_log_c);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
        }

        public static int mappable_log_r(Phloat x, ref Phloat y)
        {
            if (x <= 0)
	            return ERR_INVALID_DATA;
            else
            {
	            y = Phloat.log10(x);
	            
                return ERR_NONE;
            }
        }

        public static int mappable_log_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            Phloat h = Phloat.hypot(xre, xim);
            
            if (h == 0)
	            return ERR_INVALID_DATA;
            else
            {
                if (Phloat.p_isinf(h) != 0)
                {
                    Phloat s = new Phloat(10000);

                    h = Phloat.hypot(xre / s, xim / s);
                    yre = Phloat.log10(h) + 4;
	            }
                else
                    yre = Phloat.log10(h);

                yim = Phloat.atan2(xim, xre) / Phloat.log(10);
	            
                return ERR_NONE;
            }
        }

        public static int docmd_10_pow_x(arg_struct arg)
        {
            if (reg_x.type != TYPE.STRING)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_10_pow_x_r, mappable_10_pow_x_c);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else
                return ERR_ALPHA_DATA_IS_INVALID;
        }

        public static int mappable_10_pow_x_r(Phloat x, ref Phloat y)
        {
            y = Phloat.pow(10, x);

            if (Phloat.p_isinf(y) != 0)
            {
	            if (flags.range_error_ignore == 0)
	                return ERR_OUT_OF_RANGE;

                y = Phloat.POS_HUGE_PHLOAT;
            }

            return ERR_NONE;
        }

        public static int mappable_10_pow_x_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            int inf;
            Phloat h;

            xim *= Phloat.log(10);

            if ((inf = Phloat.p_isinf(xim)) != 0)
                xim = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;

            h = Phloat.pow(10, xre);

            if ((inf = Phloat.p_isinf(h)) == 0)
            {
                yre = Phloat.cos(xim) * h;
                yim = Phloat.sin(xim) * h;
	            
                return ERR_NONE;
            }
            else if (flags.range_error_ignore != 0)
            {
                Phloat t = Phloat.cos(xim);
	            
                if (t == 0)
	                yre = 0;
	            else if (t < 0)
                    yre = inf < 0 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	            else
                    yre = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;

                t = Phloat.sin(xim);
	            
                if (t == 0)
	                yim = 0;
	            else if (t < 0)
                    yim = inf < 0 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	            else
                    yim = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            
                return ERR_NONE;
            }
            else
	            return ERR_OUT_OF_RANGE;
        }

        public static int docmd_ln(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type == TYPE.REAL)
            {
                vartype_real x = (vartype_real)reg_x;
                
                if (x.x == 0)
                    return ERR_INVALID_DATA;
                else if (x.x < 0)
                {
                    if (flags.real_result_only != 0)
                        return ERR_INVALID_DATA;
                    else
                    {
                        vartype r = new_complex(Phloat.log(-x.x), PI);
                        
                        if (r == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        else
                        {
                            unary_result(r);
                            return ERR_NONE;
                        }
                    }
                }
                else
                {
                    vartype r = new_real(Phloat.log(x.x));

                    if (r == null)
                        return ERR_INSUFFICIENT_MEMORY;
                    else
                    {
                        unary_result(r);
                        return ERR_NONE;
                    }
                }
            }
            else
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_ln_r, mappable_ln_c);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
        }

        public static int mappable_ln_r(Phloat x, ref Phloat y)
        {
            if (x <= 0)
	            return ERR_INVALID_DATA;
            else
            {
	            y = Phloat.log(x);
	            
                return ERR_NONE;
            }
        }

        public static int mappable_ln_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            Phloat h = Phloat.hypot(xre, xim);
            
            if (h == 0)
	            return ERR_INVALID_DATA;
            else
            {
	            if (Phloat.p_isinf(h) != 0)
                {
	                Phloat s = new Phloat(10000);
                    h = Phloat.hypot(xre / s, xim / s);
                    yre = Phloat.log(h) + Phloat.log(s);
	            }
                else
                    yre = Phloat.log(h);

                yim = Phloat.atan2(xim, xre);
	            
                return ERR_NONE;
            }
        }

        public static int docmd_e_pow_x(arg_struct arg)
        {
            if (reg_x.type != TYPE.STRING)
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_e_pow_x_r, mappable_e_pow_x_c);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
            else
                return ERR_ALPHA_DATA_IS_INVALID;
        }

        public static int mappable_e_pow_x_r(Phloat x, ref Phloat y)
        {
            y = Phloat.exp(x);

            if (Phloat.p_isinf(y) != 0)
            {
	            if (flags.range_error_ignore == 0)
	                return ERR_OUT_OF_RANGE;

                y = Phloat.POS_HUGE_PHLOAT;
            }

            return ERR_NONE;
        }

        public static int mappable_e_pow_x_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            Phloat h = Phloat.exp(xre);
            int inf = Phloat.p_isinf(h);
            
            if (inf == 0)
            {
                yre = Phloat.cos(xim) * h;
                yim = Phloat.sin(xim) * h;
	            
                return ERR_NONE;
            }
            else if (flags.range_error_ignore != 0)
            {
                Phloat t = Phloat.cos(xim);

	            if (t == 0)
	                yre = 0;
	            else if (t < 0)
                    yre = inf < 0 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	            else
                    yre = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;

                t = Phloat.sin(xim);
	            
                if (t == 0)
	                yim = 0;
	            else if (t < 0)
                    yim = inf < 0 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	            else
                    yim = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            
                return ERR_NONE;
            }
            else
	            return ERR_OUT_OF_RANGE;
        }

        public static int docmd_sqrt(arg_struct arg)
        {
            if (reg_x.type == TYPE.REAL)
            {
                Phloat x = ((vartype_real)reg_x).x;
                
                vartype v;
                
                if (x < 0)
                {
                    if (flags.real_result_only != 0)
                        return ERR_INVALID_DATA;

                    v = new_complex(0, Phloat.sqrt(-x));
                }
                else
                    v = new_real(Phloat.sqrt(x));
                
                if (v == null)
                    return ERR_INSUFFICIENT_MEMORY;
                
                unary_result(v);
                return ERR_NONE;
            }
            else if (reg_x.type == TYPE.STRING)
            {
                return ERR_ALPHA_DATA_IS_INVALID;
            }
            else
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_sqrt_r, mappable_sqrt_c);
                
                if (err != ERR_NONE)
                    return err;
                
                unary_result(v);
                return ERR_NONE;
            }
        }

        public static int mappable_sqrt_r(Phloat x, ref Phloat y)
        {
            if (x < 0)
	            return ERR_INVALID_DATA;
            else
            {
	            y = Phloat.sqrt(x);
	            
                return ERR_NONE;
            }
        }

        public static int mappable_sqrt_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            // (orig authors note) TO DO: review -- is there a better way, without all the trig?
            Phloat r = Phloat.sqrt(Phloat.hypot(xre, xim));
            Phloat phi = Phloat.atan2(xim, xre) / 2;
            
            Phloat s = new Phloat();
            Phloat c = new Phloat();

            Phloat.sincos(phi, ref s, ref c);
            
            yre = r * c;
            yim = r * s;
            
            return ERR_NONE;
        }

        public static int docmd_square(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_square_r, mappable_square_c);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
        }

        public static int mappable_square_r(Phloat x, ref Phloat y)
        {
            Phloat r = x * x;
            int inf;
        
            if ((inf = Phloat.p_isinf(r)) != 0)
            {
	            if (flags.range_error_ignore != 0)
                    r = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	            else
	                return ERR_OUT_OF_RANGE;
            }

            y = r;
            
            return ERR_NONE;
        }

        public static int mappable_square_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            Phloat rre = xre * xre - xim * xim;
            Phloat rim = 2 * xre * xim;
            int inf;

            if ((inf = Phloat.p_isinf(rre)) != 0)
            {
	            if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	            else
	                return ERR_OUT_OF_RANGE;
            }

            if ((inf = Phloat.p_isinf(rim)) != 0)
            {
	            if (flags.range_error_ignore != 0)
                    rim = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	            else
	                return ERR_OUT_OF_RANGE;
            }

            yre = rre;
            yim = rim;
            
            return ERR_NONE;
        }

        public static int docmd_inv(arg_struct arg)
        {
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else
            {
                vartype v = new vartype();

                int err = map_unary(reg_x, ref v, mappable_inv_r, mappable_inv_c);
                
                if (err == ERR_NONE)
                    unary_result(v);
                
                return err;
            }
        }

        public static int mappable_inv_r(Phloat x, ref Phloat y)
        {
            int inf;
        
            if (x == 0)
	            return ERR_DIVIDE_BY_0;
            
            y = 1 / x;
            
            if ((inf = Phloat.p_isinf(y)) != 0)
            {
	            if (flags.range_error_ignore == 0)
	                return ERR_OUT_OF_RANGE;

                y = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
            }

            return ERR_NONE;
        }

        public static int mappable_inv_c(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            int inf;

            Phloat h = Phloat.hypot(xre, xim);
            
            if (h == 0)
	            return ERR_DIVIDE_BY_0;
            
            yre = xre / h / h;

            if ((inf = Phloat.p_isinf(yre)) != 0)
            {
	            if (flags.range_error_ignore == 0)
	                return ERR_OUT_OF_RANGE;

                yre = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
            }

            yim = (-xim) / h / h;

            if ((inf = Phloat.p_isinf(yim)) != 0)
            {
	            if (flags.range_error_ignore == 0)
	                return ERR_OUT_OF_RANGE;

                yim = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
            }

            return ERR_NONE;
        }

        public static int docmd_y_pow_x(arg_struct arg)
        {
            Phloat yr;
            Phloat yphi;
            int inf;
            vartype res;
            Phloat x;

            if (reg_x.type == TYPE.STRING || reg_y.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            else if (reg_x.type == TYPE.REALMATRIX
                || reg_x.type == TYPE.COMPLEXMATRIX
                || reg_y.type == TYPE.REALMATRIX
                || reg_y.type == TYPE.COMPLEXMATRIX)
                return ERR_INVALID_TYPE;
            else if (reg_x.type == TYPE.REAL)
            {
                x = ((vartype_real)reg_x).x;
                if (x == Phloat.floor(x))
                {
                    // Integer exponent
                    if (reg_y.type == TYPE.REAL)
                    {
                        // Real number to integer power
                        Phloat y = ((vartype_real)reg_y).x;
                        Phloat r = Phloat.pow(y, x);
                        
                        if (Phloat.p_isnan(r) != 0)
                            // Should not happen; pow() is supposed to be able
                            // to raise negative numbers to integer exponents
                            //
                            return ERR_INVALID_DATA;

                        if ((inf = Phloat.p_isinf(r)) != 0)
                        {
                            if (flags.range_error_ignore == 0)
                                return ERR_OUT_OF_RANGE;

                            r = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                        }

                        res = new_real(r);
                        
                        if (res == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        else
                        {
                            binary_result(res);
                            return ERR_NONE;
                        }
                    }
                    else
                    {
                        // Complex number to integer power
                        Phloat rre;
                        Phloat rim;
                        Phloat yre;
                        Phloat yim;
                        int ex;
                        
                        if (x < -2147483647.0 || x > 2147483647.0)
                            // For really huge exponents, the repeated-squaring
                            // algorithm for integer exponents loses its accuracy
                            // and speed advantage, and we switch to the general
                            // complex-to-real-power code instead.
                            //
                            goto complex_pow_real_1;

                        rre = 1;
                        rim = 0;
                        yre = ((vartype_complex)reg_y).re;
                        yim = ((vartype_complex)reg_y).im;
                        ex = Phloat.to_int(x);
                        
                        if (yre == 0 && yim == 0)
                        {
                            if (ex < 0)
                                return ERR_INVALID_DATA;
                            else if (ex == 0)
                            {
                                res = new_complex(1, 0);
                        
                                if (res == null)
                                    return ERR_INSUFFICIENT_MEMORY;
                                else
                                {
                                    binary_result(res);
                                    return ERR_NONE;
                                }
                            }
                        }

                        if (ex < 0)
                        {
                            Phloat h = Phloat.hypot(yre, yim);
                            yre = yre / h / h;
                            yim = (-yim) / h / h;
                            ex = -ex;
                        }

                        while (true)
                        {
                            Phloat tmp;
                            
                            if ((ex & 1) != 0)
                            {
                                tmp = rre * yre - rim * yim;
                                rim = rre * yim + rim * yre;
                                rre = tmp;
                                // TO DO: can one component be infinite while
                                // the other is zero? If yes, how do we handle
                                // that?
                                //
                                if (Phloat.p_isinf(rre) != 0 && Phloat.p_isinf(rim) != 0)
                                    break;

                                if (rre == 0 && rim == 0)
                                    break;
                            }

                            ex >>= 1;
                            
                            if (ex == 0)
                                break;
                            
                            tmp = yre * yre - yim * yim;
                            yim = 2 * yre * yim;
                            yre = tmp;
                        }

                        if ((inf = Phloat.p_isinf(rre)) != 0)
                        {
                            if (flags.range_error_ignore == 0)
                                return ERR_OUT_OF_RANGE;

                            rre = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                        }

                        if ((inf = Phloat.p_isinf(rim)) != 0)
                        {
                            if (flags.range_error_ignore == 0)
                                return ERR_OUT_OF_RANGE;

                            rim = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                        }

                        res = new_complex(rre, rim);
                        
                        if (res == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        else
                        {
                            binary_result(res);
                            return ERR_NONE;
                        }
                    }
                }
                else if (reg_y.type == TYPE.REAL)
                {
                    // Real number to noninteger real power
                    Phloat y = ((vartype_real)reg_y).x;
                    Phloat r;

                    if (y < 0)
                    {
                        if (flags.real_result_only != 0)
                            return ERR_INVALID_DATA;

                        yr = -y;
                        yphi = PI;
                        
                        goto complex_pow_real_2;
                    }

                    r = Phloat.pow(y, x);
                    inf = Phloat.p_isinf(r);
                    
                    if (inf != 0)
                    {
                        if (flags.range_error_ignore == 0)
                            return ERR_OUT_OF_RANGE;

                        r = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                    }

                    res = new_real(r);
                    
                    if (res == null)
                        return ERR_INSUFFICIENT_MEMORY;
                    else
                    {
                        binary_result(res);
                        return ERR_NONE;
                    }
                }
                else
                {
                    goto complex_pow_real_1;
                }
            }
            else
            {
                // Real or complex number to complex power
                Phloat xre = ((vartype_complex)reg_x).re;
                Phloat xim = ((vartype_complex)reg_x).im;
                Phloat yre;
                Phloat yim;
                Phloat lre = new Phloat();
                Phloat lim = new Phloat();
                Phloat tmp;
                int err;

                if (reg_y.type == TYPE.REAL)
                {
                    yre = ((vartype_real)reg_y).x;
                    yim = 0;
                }
                else
                {
                    yre = ((vartype_complex)reg_y).re;
                    yim = ((vartype_complex)reg_y).im;
                }

                if (yre == 0 && yim == 0)
                {
                    if (xre < 0 || (xre == 0 && xim != 0))
                        return ERR_INVALID_DATA;
                    else if (xre == 0)
                        res = new_complex(1, 0);
                    else
                        res = new_complex(0, 0);
                
                    if (res == null)
                        return ERR_INSUFFICIENT_MEMORY;
                    else
                    {
                        binary_result(res);
                        return ERR_NONE;
                    }
                }

                err = mappable_ln_c(yre, yim, ref lre, ref lim);

                if (err != ERR_NONE)
                    return err;
                
                tmp = lre * xre - lim * xim;
                lim = lre * xim + lim * xre;
                lre = tmp;
                err = mappable_e_pow_x_c(lre, lim, ref xre, ref xim);

                if (err != ERR_NONE)
                    return err;
                
                res = new_complex(xre, xim);
                
                if (res == null)
                    return ERR_INSUFFICIENT_MEMORY;
                else
                {
                    binary_result(res);
                    return ERR_NONE;
                }
            }


            // SLR: Moved these to end to get goto statements in scope
            // No fall through from above

            // Complex (or negative real) number to noninteger real power
        complex_pow_real_1:
            {
                Phloat yre = ((vartype_complex)reg_y).re;
                Phloat yim = ((vartype_complex)reg_y).im;

                yr = Phloat.hypot(yre, yim);
                yphi = Phloat.atan2(yim, yre);
            }

        complex_pow_real_2:
            {
                yr = Phloat.pow(yr, x);
                yphi *= x;

                Phloat rre;
                Phloat rim;

                if ((inf = Phloat.p_isinf(yr)) != 0)
                {
                    if (flags.range_error_ignore == 0)
                        return ERR_OUT_OF_RANGE;
                    else
                    {
                        Phloat re = new Phloat();
                        Phloat im = new Phloat();

                        Phloat.sincos(yphi, ref im, ref re);

                        rre = re == 0 ? 0 : re < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                        rim = im == 0 ? 0 : im < 0 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                    }
                }
                else
                {
                    Phloat re = new Phloat();
                    Phloat im = new Phloat();

                    Phloat.sincos(yphi, ref im, ref re);

                    rre = yr * re;
                    rim = yr * im;
                }

                res = new_complex(rre, rim);

                if (res == null)
                    return ERR_INSUFFICIENT_MEMORY;
                else
                {
                    binary_result(res);
                    return ERR_NONE;
                }
            }
        }

        public static int docmd_rotxy(arg_struct arg)
        {
            long x = 0;
            long y = 0;
            long res;
            int err;
            vartype v;

            if ((err = get_base_param(reg_x, ref x)) != ERR_NONE)
                return err;
            
            if ((err = get_base_param(reg_y, ref y)) != ERR_NONE)
                return err;
            
            if (x < -35 || x > 35)
                return ERR_INVALID_DATA;
            
            if (x == 0)
                res = y;
            else
            {
                y &= 0xfffffffff;

                if (x > 0)
                    res = (y >> (int)x) | (y << (int)(36 - x));
                else
                {
                    x = -x;
                    res = (y << (int)x) | (y >> (int)(36 - x));
                }

                if ((res & 0x800000000) == 0)
                    res &= 0x7ffffffff;
                else
                    res = (long)((ulong)res | (ulong)0xfffffff000000000);
            }

            v = new_real(new Phloat(res));
            
            if (v == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            binary_result(v);
            
            return ERR_NONE;
        }
    }
}
