﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        // Underhill's COPAN
        public static int docmd_openf(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_closef(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_readp(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_writp(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_getxy(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_putxy(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_clrp(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_clrd(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_appd(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_getn(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_putn(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_getz(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_putz(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_delp(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }


        // Byron Foster's DROP for Bigstack
        public static int docmd_drop(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }


        // Accelerometer, GPS, and compass support
        public static int docmd_accel(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_locat(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_heading(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }


        // Time Module & CX Time support
        public static int docmd_adate(arg_struct arg)
        {
            if (!core_settings.enable_ext_time)
    	        return ERR_NONEXISTENT;
    
            if (reg_x.type == TYPE.STRING)
	            return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_x.type != TYPE.REAL)
	            return ERR_INVALID_TYPE;

            Phloat x = ((vartype_real) reg_x).x;
            
            if (x < 0)
	            x = -x;
            
            if (x >= 100)
	            return ERR_INVALID_DATA;
            
            int m = Phloat.to_int(Phloat.floor(x));
            int dy = Phloat.to_int(Phloat.floor((x - Phloat.floor(x)) * 1000000));
            int d = (int) (dy / 10000);
            int c = (int) (dy / 100 % 100);
            int y = (int) (dy % 100);

            int digits;
            if (flags.fix_or_all != 0 && flags.eng_or_all != 0)
	            digits = 11;
            else
            {
	            digits = 0;
	         
                if (flags.digits_bit3 != 0)
	                digits += 8;
	            
                if (flags.digits_bit2 != 0)
	                digits += 4;
	            
                if (flags.digits_bit1 != 0)
	                digits += 2;
	            
                if (flags.digits_bit0 != 0)
	                digits += 1;
            }

            char[] buf = new char[10];
            int bufptr = 0;

            if (m < 10)
	            char2buf(buf, 10, ref bufptr, '0');
            
            bufptr += int2string(m, buf, bufptr, 10);
            
            if (digits > 0)
            {
	            char2buf(buf, 10, ref bufptr, mode_time_dmy ? '.' : '/');
	            
                if (d < 10)
	                char2buf(buf, 10, ref bufptr, '0');
	            
                bufptr += int2string(d, buf, bufptr, 10);
	            
                if (digits > 2)
                {
	                char2buf(buf, 10, ref bufptr, mode_time_dmy ? '.' : '/');
	            
                    if (digits > 4)
                    {
		                if (c < 10)
		                    char2buf(buf, 10, ref bufptr, '0');
		                
                        bufptr += int2string(c, buf, bufptr, 10);
	                }
	                
                    if (y < 10)
		                char2buf(buf, 10, ref bufptr, '0');

	                bufptr += int2string(y, buf, bufptr, 10);
	            }
            }

            append_alpha_string(buf, bufptr, 0);
            
            return ERR_NONE;
        }

        public static int docmd_almcat(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_almnow(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }

        public static int docmd_atime(arg_struct arg)
        {
            if (!core_settings.enable_ext_time)
    	        return ERR_NONEXISTENT;
    
            if (reg_x.type == TYPE.STRING)
	            return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_x.type != TYPE.REAL)
	            return ERR_INVALID_TYPE;

            Phloat x = ((vartype_real) reg_x).x;
            
            bool neg = x < 0;
            
            if (neg)
	            x = -x;
            
            if (x >= 100)
	            return ERR_INVALID_DATA;

            int h = Phloat.to_int(Phloat.floor(x));

            if (h == 0)
	            neg = false;

            int ms = Phloat.to_int(Phloat.floor((x - Phloat.floor(x)) * 1000000));

            int m = (int) (ms / 10000);
            int s = (int) (ms / 100 % 100);
            int cs = (int) (ms % 100);

            bool am = false;
            bool pm = false;

            if (mode_time_clk24)
            {
	            if (neg && h >= 1 && h <= 11)
	                h += 12;
            }
            else if (h < 24)
            {
	            if (!neg && h < 12)
	                am = true;
	            else
	                pm = true;

                if (h == 0)
	                h = 12;
	            else if (h > 12)
	                h -= 12;
            }

            int digits;

            if (flags.fix_or_all != 0 && flags.eng_or_all != 0)
	            digits = 11;
            else
            {
	            digits = 0;

                if (flags.digits_bit3 != 0)
	                digits += 8;
	            
                if (flags.digits_bit2 != 0)
	                digits += 4;
	            
                if (flags.digits_bit1 != 0)
	                digits += 2;
	            
                if (flags.digits_bit0 != 0)
	                digits += 1;
            }

            char[] buf = new char[14];
            int bufptr = 0;

            if (h < 10)
	            char2buf(buf, 14, ref bufptr, mode_time_clk24 ? '0' : ' ');

            bufptr += int2string(h, buf, bufptr, 14);
            
            if (digits > 0)
            {
	            char2buf(buf, 14, ref bufptr, ':');

                if (m < 10)
	                char2buf(buf, 14, ref bufptr, '0');
	            
                bufptr += int2string(m, buf, bufptr, 14);
	            
                if (digits > 2)
                {
	                char2buf(buf, 14, ref bufptr, ':');

                    if (s < 10)
		                char2buf(buf, 14, ref bufptr, '0');

	                bufptr += int2string(s, buf, bufptr, 14);
	                
                    if (digits > 4)
                    {
    		            char2buf(buf, 14, ref bufptr, '.');
	
                        if (cs < 10)
		                    char2buf(buf, 14, ref bufptr, '0');
		            
                        bufptr += int2string(cs, buf, bufptr, 14);
	                }
	            }
            }

            if (am)
	            string2buf(buf, 14, ref bufptr, " AM".ToCharArray(), 3);
            else if (pm)
	            string2buf(buf, 14, ref bufptr, " PM".ToCharArray(), 3);

            append_alpha_string(buf, bufptr, 0);

            return ERR_NONE;
        }

        public static int docmd_atime24(arg_struct arg)
        {
            if (!core_settings.enable_ext_time)
                return ERR_NONEXISTENT;
        
            bool saved_clk24 = mode_time_clk24;

            mode_time_clk24 = true;
            int res = docmd_atime(arg);
            mode_time_clk24 = saved_clk24;
            
            return res;
        }
        
        public static int docmd_clk12(arg_struct arg)
        {
            if (!core_settings.enable_ext_time)
                return ERR_NONEXISTENT;
        
            mode_time_clk24 = false;
            
            return ERR_NONE;
        }
        
        public static int docmd_clk24(arg_struct arg)
        {
            if (!core_settings.enable_ext_time)
                return ERR_NONEXISTENT;
            
            mode_time_clk24 = true;
            
            return ERR_NONE;
        }

        public static int docmd_clkt(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_clktd(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_clock(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_correct(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }

        public static int docmd_date(arg_struct arg)
        {
            if (!core_settings.enable_ext_time)
    	        return ERR_NONEXISTENT;

            uint time = 0;
            uint date = 0;
            int weekday = 0;
            
            shell_get_time_date(ref time, ref date, ref weekday);
            
            int y = (int)(date / 10000);
            int m = (int)(date / 100 % 100);
            int d = (int)(date % 100);
            
            if (mode_time_dmy)
	            date = (uint)(y + m * 10000L + d * 1000000);
            else
	            date = (uint)(y + m * 1000000 + d * 10000L);
            
            vartype new_x = new_real((int) date);
            
            if (new_x == null)
	            return ERR_INSUFFICIENT_MEMORY;
            
            ((vartype_real) new_x).x /= 1000000;
            recall_result(new_x);
            
            if (!program_running())
            {
	            // Note: I'm not completely faithful to the HP-41 here. It formats the
	            // date as "14.03.2010 SUN" in DMY mode, and as "03/14/2010:SU" in MDY
	            // mode. I mimic the former, but the latter I changed to
	            // "03/14/2010 SUN"; the MDY display format used on the HP-41 is the
	            // way it is because that was all they could fit in its 12-character
	            // display. (Note that the periods in the DMY format and the colon in
	            // the MDY format don't take up a character position on the HP-41.)
	            //
	            char[] buf = new char[22];
	            int bufptr = 0;

                int n = mode_time_dmy ? d : m;
	            
                if (n < 10)
	                char2buf(buf, 22, ref bufptr, '0');
	            
                bufptr += int2string(n, buf, bufptr, 22);
	            char2buf(buf, 22, ref bufptr, mode_time_dmy ? '.' : '/');
	            
                n = mode_time_dmy ? m : d;
	            
                if (n < 10)
	                char2buf(buf, 22, ref bufptr, '0');
	            
                bufptr += int2string(n, buf, bufptr, 22);
	            char2buf(buf, 22, ref bufptr, mode_time_dmy ? '.' : '/');
	            bufptr += int2string(y, buf, bufptr, 22);
	            char2buf(buf, 22, ref bufptr, ' ');
	            string2buf(buf, 22, ref bufptr, weekdaynames[weekday].ToCharArray(), 3);
	            
                clear_row(0);
	            draw_string(0, 0, buf, bufptr);
	            flush_display();
	            
                flags.message = 1;
	            flags.two_line_message = 0;
            }
            // TODO: Trace-mode printing. What should I print, the contents of X,
            // or, when not in a running program, the nicely formatted date?
            return ERR_NONE;
        }

        public static int docmd_date_plus(arg_struct arg)
        {
            if (!core_settings.enable_ext_time)
                return ERR_NONEXISTENT;
        
            // TODO: Accept real matrices as well?
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;
            
            if (reg_y.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_y.type != TYPE.REAL)
                return ERR_INVALID_TYPE;

            Phloat date = ((vartype_real)reg_y).x;
            
            if (date < 0 || date > 100)
                return ERR_INVALID_DATA;
            
            Phloat days = ((vartype_real)reg_x).x;
            
            if (days < -1000000 || days > 1000000)
                return ERR_OUT_OF_RANGE;

            int y = 0;
            int m = 0;
            int d = 0;
            int jd = 0;
            
            int err = date2comps(date, ref y, ref m, ref d);
            
            if (err != ERR_NONE)
                return err;
            
            err = greg2jd(y, m, d, ref jd);
            
            if (err != ERR_NONE)
                return err;
            
            jd += Phloat.to_int(Phloat.floor(days));
            err = jd2greg(jd, ref y, ref m, ref d);
            
            if (err != ERR_NONE)
                return err;
            
            date = comps2date(y, m, d);

            vartype new_x = new_real(date);
            
            if (new_x == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            binary_result(new_x);

            return ERR_NONE;
        }

        public static int docmd_ddays(arg_struct arg)
        {
            if (!core_settings.enable_ext_time)
                return ERR_NONEXISTENT;
        
            // TODO: Accept real matrices as well?
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;
            
            if (reg_y.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;
            
            if (reg_y.type != TYPE.REAL)
                return ERR_INVALID_TYPE;

            Phloat date1 = ((vartype_real)reg_y).x;
            
            if (date1 < 0 || date1 > 100)
                return ERR_INVALID_DATA;
            
            Phloat date2 = ((vartype_real)reg_x).x;
            
            if (date2 < 0 || date2 > 100)
                return ERR_INVALID_DATA;
            
            int y = 0;
            int m = 0;
            int d = 0;
            int jd1 = 0;
            int jd2 = 0;
            
            int err = date2comps(date1, ref y, ref m, ref d);
            
            if (err != ERR_NONE)
                return err;
            
            err = greg2jd(y, m, d, ref jd1);
            
            if (err != ERR_NONE)
                return err;
            
            err = date2comps(date2, ref y, ref m, ref d);
            
            if (err != ERR_NONE)
                return err;
            
            err = greg2jd(y, m, d, ref jd2);
            
            if (err != ERR_NONE)
                return err;

            vartype new_x = new_real(jd2 - jd1);
            
            if (new_x == null)
                return ERR_INSUFFICIENT_MEMORY;
            
            binary_result(new_x);
            
            return ERR_NONE;
        }

        public static int docmd_dmy(arg_struct arg)
        {
            if (!core_settings.enable_ext_time)
                return ERR_NONEXISTENT;
        
            mode_time_dmy = true;
            
            return ERR_NONE;
        }

        public static int docmd_dow(arg_struct arg)
        {
            if (!core_settings.enable_ext_time)
                return ERR_NONEXISTENT;
        
            // TODO: Accept real matrices as well?
            if (reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;

            if (reg_x.type != TYPE.REAL)
                return ERR_INVALID_TYPE;

            Phloat x = ((vartype_real)reg_x).x;

            if (x < 0 || x > 100)
                return ERR_INVALID_DATA;

            int y = 0;
            int m = 0;
            int d = 0;
            int jd = 0;

            int err = date2comps(x, ref y, ref m, ref d);
            
            if (err != ERR_NONE)
                return err;
            
            err = greg2jd(y, m, d, ref jd);
            
            if (err != ERR_NONE)
                return err;
            
            jd = (jd + 1) % 7;

            vartype new_x = new_real(jd);

            if (new_x == null)
                return ERR_INSUFFICIENT_MEMORY;

            unary_result(new_x);

            if (!program_running())
            {
                clear_row(0);
                draw_string(0, 0, weekdaynames[jd].ToCharArray(), 3);
                flush_display();
            
                flags.message = 1;
                flags.two_line_message = 0;
            }

            return ERR_NONE;
        }

        public static int docmd_mdy(arg_struct arg)
        {
            if (!core_settings.enable_ext_time)
                return ERR_NONEXISTENT;
        
            mode_time_dmy = false;
            
            return ERR_NONE;
        }

        public static int docmd_rclaf(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_rclsw(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_runsw(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_setaf(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_setdate(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_setime(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_setsw(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_stopsw(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_sw(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_t_plus_x(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }

        public static int docmd_time(arg_struct arg)
        {
            if (!core_settings.enable_ext_time)
	            return ERR_NONEXISTENT;

            uint time = 0;
            uint date = 0;
            int weekday = 0;

            shell_get_time_date(ref time, ref date, ref weekday);
            
            vartype new_x = new_real((int) time);
            
            if (new_x == null)
	            return ERR_INSUFFICIENT_MEMORY;

            ((vartype_real) new_x).x /= 1000000;
            recall_result(new_x);
            
            if (!program_running())
            {
	            int h = (int)(time / 1000000);
	            bool am = false;

                if (!mode_time_clk24)
                {
	                am = h < 12;
	                h = h % 12;

	                if (h == 0)
		                h = 12;
	            }

                int m = (int)(time / 10000 % 100);
	            int s = (int)(time / 100 % 100);

                char[] buf = new char[22];
	            int bufptr = 0;
	            
                if (h < 10)
	                char2buf(buf, 22, ref bufptr, ' ');

	            bufptr += int2string(h, buf, bufptr, 22);
	            char2buf(buf, 22, ref bufptr, ':');
	            
                if (m < 10)
	                char2buf(buf, 22, ref bufptr, '0');
	            
                bufptr += int2string(m, buf, bufptr, 22);
	            char2buf(buf, 22, ref bufptr, ':');
	            
                if (s < 10)
	                char2buf(buf, 22, ref bufptr, '0');
	            
                bufptr += int2string(s, buf, bufptr, 22);
	            
                if (!mode_time_clk24)
                {
	                char2buf(buf, 22, ref bufptr, ' ');
	                char2buf(buf, 22, ref bufptr, am ? 'A' : 'P');
	                char2buf(buf, 22, ref bufptr, 'M');
	            }
	            
                clear_row(0);
	            draw_string(0, 0, buf, bufptr);
	            flush_display();
	            
                flags.message = 1;
	            flags.two_line_message = 0;
            }
            // TODO: Trace-mode printing. What should I print, the contents of X,
            // or, when not in a running program, the nicely formatted time?
            return ERR_NONE;
        }

        public static int docmd_xyzalm(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_clalma(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_clalmx(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_clralms(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_rclalm(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }
        public static int docmd_swpt(arg_struct arg)
        {
            return ERR_NONEXISTENT;
        }

        // Gregorian Date <-> Day Number conversion functions
        // Algorithm due to Henry F. Fliegel and Thomas C. Van Flandern,
        // Communications of the ACM, Vol. 11, No. 10 (October, 1968).
        public static int greg2jd(int y, int m, int d, ref int jd)
        {
            jd = (1461 * (y + 4800 + (m - 14) / 12)) / 4 +
              (367 * (m - 2 - 12 * ((m - 14) / 12))) / 12 -
              (3 * ((y + 4900 + (m - 14) / 12) / 100)) / 4 +
              d - 32075;

            return ERR_NONE;
        }

        public static int jd2greg(int jd, ref int y, ref int m, ref int d)
        {
            if (jd < 2299161 || jd > 3299160)
                return ERR_OUT_OF_RANGE;

            int l = jd + 68569;
            int n = (4 * l) / 146097;
            l = l - (146097 * n + 3) / 4;
            int i = (4000 * (l + 1)) / 1461001;
            l = l - (1461 * i) / 4 + 31;
            int j = (80 * l) / 2447;
            d = l - (2447 * j) / 80;
            l = j / 11;
            m = j + 2 - (12 * l);
            y = 100 * (n - 49) + i + l;
            return ERR_NONE;
        }

        public static int date2comps(Phloat x, ref int yy, ref int mm, ref int dd)
        {
            int m = Phloat.to_int(Phloat.floor(x));

            int d = Phloat.to_int(Phloat.floor((x - m) * 100));
            int y = Phloat.to_int(x * 1000000) % 10000;

            if (mode_time_dmy)
            {
                int t = m;
                m = d;
                d = t;
            }

            if (y < 1582 || y > 4320 || m < 1 || m > 12 || d < 1 || d > 31)
                return ERR_INVALID_DATA;

            if ((m == 4 || m == 6 || m == 9 || m == 11) && d == 31)
                return ERR_INVALID_DATA;

            if (m == 2 && d > ((y % 4 == 0 && (y % 100 != 0 || y % 400 == 0)) ? 29 : 28))
                return ERR_INVALID_DATA;

            if (y == 1582 && (m < 10 || m == 10 && d < 15) || y == 4320 && (m > 9 || m == 9 && d > 10))
                return ERR_INVALID_DATA;

            yy = y;
            mm = m;
            dd = d;

            return ERR_NONE;
        }

        public static Phloat comps2date(int y, int m, int d)
        {
            if (mode_time_dmy)
            {
                int t = m;
                m = d;
                d = t;
            }

            return new Phloat(m * 1000000 + d * 10000 + y) / 1000000;
        }
    }
}
