﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Text.RegularExpressions;
using Microsoft.Xna.Framework.GamerServices;

namespace WP7_42s
{
    public partial class App
    {
        private static int MUNGE_LEN = 1;

        public static string munge(string s)
        {
            if (trialMode)
            {
                Regex re = new Regex(@"(?<ok>\.\d{" + MUNGE_LEN.ToString() + @"})(?<del>\d+)");
                return re.Replace(s, new MatchEvaluator(reMatch));
            }
            else
                return s;
        }

        public static string reMatch(Match m)
        {
            GroupCollection gc = m.Groups;
            return gc["ok"].Value + new string('-', gc["del"].Value.Length);
        }


        
        public static char[] munge(char[] s)
        {
            return munge(new string(s)).ToCharArray();
        }




        public static void clear_display()
        {
            for (int i = 0; i < 272; i++)
                display[i] = 0;

            mark_dirty(0, 0, 16, 131);
        }

        public static void flush_display()
        {
            if (!is_dirty)
                return;

            display_blitter(display, 17, dirty_left, dirty_top, dirty_right - dirty_left, dirty_bottom - dirty_top);
            is_dirty = false;
        }

        public static void repaint_display()
        {
            display_blitter(display, 17, 0, 0, 131, 16);
        }

        public static void draw_pixel(int x, int y)
        {
            display[y * 17 + (x >> 3)] |= (byte)(1 << (x & 7));
            mark_dirty(y, x, y + 1, x + 1);
        }

        public static void draw_char(int x, int y, char c)
        {
            if (x < 0 || x >= 22 || y < 0 || y >= 2)
                return;

            byte uc = (byte)c;
            if (uc >= 130)
                uc -= 128;

            int X = x * 6;
            int Y = y * 8;

            for (int v = 0; v < 8; v++)
            {
                int YY = Y + v;

                for (int h = 0; h < 5; h++)
                {
                    int XX = X + h;

                    byte mask = (byte)(1 << (XX & 7));

                    if ((bigchars[uc, h] & (1 << v)) != 0)
                        display[YY * 17 + (XX >> 3)] |= mask;
                    else
                        display[YY * 17 + (XX >> 3)] &= (byte)~mask;
                }
            }

            mark_dirty(Y, X, Y + 8, X + 5);
        }

        public static void draw_string(int x, int y, string s)
        {
            // Zero fractional amounts in trial mode
            s = munge(s);

            foreach (char c in s)
                draw_char(x++, y, c);
        }

        public static void draw_string(int x, int y, char[] s, int length)
        {
            int i = 0;

            // Zero fractional amounts in trial mode
            s = munge(s);

            while (length != 0 && x < 22)
            {
                draw_char(x++, y, s[i++]);
                length--;
            }
        }

        public static void fill_rect(int x, int y, int width, int height, double color)
        {
            for (int v = y; v < y + height; v++)
                for (int h = x; h < x + width; h++)
                    if (color != 0)
                        display[v * 17 + (h >> 3)] |= (byte)(1 << (h & 7));
                    else
                        display[v * 17 + (h >> 3)] &= (byte)~(1 << (h & 7));

            mark_dirty(y, x, y + height, x + width);
        }

        public static void clear_row(int row)
        {
            fill_rect(0, row * 8, 131, 8, 0);
        }

        public static void mark_dirty(int top, int left, int bottom, int right)
        {
            if (is_dirty)
            {
                if (top < dirty_top) dirty_top = top;
                if (left < dirty_left) dirty_left = left;
                if (bottom > dirty_bottom) dirty_bottom = bottom;
                if (right > dirty_right) dirty_right = right;
            }
            else
            {
                dirty_top = top;
                dirty_left = left;
                dirty_bottom = bottom;
                dirty_right = right;
                is_dirty = true;
            }
        }

        public static void draw_key(int n, bool highlight, bool hide_meta, char[] s, int length)
        {
            int swidth = 0;
            int len = 0;
            byte fatdot = 31;


            /* Note: the SST handling code uses a magic value of 2 in prgm_mode
             * so that we know *not* to suppress menu highlights while stepping. */

            if (flags.prgm_mode == 1)
                highlight = false;

            if (highlight)
            {
                int f = smallchars_map[fatdot];
                swidth = smallchars_offset[f + 1] - smallchars_offset[f];
            }

            while (len < length)
            {
                byte c = (byte)s[len++];

                if (hide_meta && c >= 128)
                    continue;

                c &= 127;

                int m = smallchars_map[c];
                int cw = smallchars_offset[m + 1] - smallchars_offset[m];

                if (swidth != 0)
                    cw++;

                if (swidth + cw > 19)
                {
                    len--;
                    break;
                }

                swidth += cw;
            }

            fill_rect(n * 22, 9, 21, 7, 1);

            int x = n * 22 + 10 - swidth / 2;
            int len2 = highlight ? len + 1 : len;


            for (int i = 0; i < len2; i++)
            {
                byte c;

                if (i == len)
                    c = fatdot;
                else
                    c = (byte)s[i];

                if (hide_meta && c >= 128)
                    continue;

                c &= 127;

                int m = smallchars_map[c];
                int o = smallchars_offset[m];
                int cw = smallchars_offset[m + 1] - o;

                for (int j = 0; j < cw; j++)
                {
                    int b = smallchars[o + j];

                    for (int k = 0; k < 8; k++)
                        if (((b >> k) & 1) != 0)
                            display[(k + 8) * 17 + (x >> 3)] &= (byte)~(1 << (x & 7));

                    x++;
                }
                x++;
            }
            // No need for mark_dirty(); fill_rect() took care of that already.
        }

        public static void display_x(int row)
        {
            char[] buf = new char[22];
            int bufptr = 0;

            clear_row(row);

            if (matedit_mode == 2 || matedit_mode == 3)
            {
                bufptr += int2string(matedit_i + 1, buf, bufptr, 22);
                char2buf(buf, 22, ref bufptr, ':');
                bufptr += int2string(matedit_j + 1, buf, bufptr, 22);
                char2buf(buf, 22, ref bufptr, '=');
            }
            else if (input_length > 0)
            {
                string2buf(buf, 22, ref bufptr, input_name, input_length);
                char2buf(buf, 22, ref bufptr, '?');
            }
            else
                string2buf(buf, 22, ref bufptr, "x\x0080".ToCharArray(), 2);

            bufptr += vartype2string(reg_x, buf, bufptr, 22);

            draw_string(0, row, buf, bufptr);
        }

        public static void display_y(int row)
        {
            char[] buf = new char[20];
            int len;

            clear_row(row);

            len = vartype2string(reg_y, buf, 20);

            draw_string(0, row, "\x0081\x0080".ToCharArray(), 2);

            if (len > 20)
            {
                draw_string(2, row, buf, 19);
                draw_char(21, row, (char)26);
            }
            else
                draw_string(2, row, buf, len);
        }

        public static void update_catalog()
        {
            int the_menu;

            if (mode_commandmenu != (int)MENU.NONE)
                the_menu = mode_commandmenu;
            else if (mode_alphamenu != (int)MENU.NONE)
                the_menu = mode_alphamenu;
            else if (mode_transientmenu != (int)MENU.NONE)
                the_menu = mode_transientmenu;
            else if (mode_plainmenu != (int)MENU.NONE)
                the_menu = mode_plainmenu;
            else if (mode_appmenu != (int)MENU.NONE)
                the_menu = mode_appmenu;
            else
                return;

            if (the_menu != (int)MENU.CATALOG)
                return;

            switch (get_cat_section())
            {
                case CATSECT_TOP:
                case CATSECT_FCN:
                    return;

                case CATSECT_PGM:
                case CATSECT_PGM_ONLY:
                    break;

                case CATSECT_REAL:
                    if (vars_exist(1, 0, 0) == 0)
                        set_cat_section(CATSECT_TOP);
                    break;

                case CATSECT_CPX:
                    if (vars_exist(0, 1, 0) == 0)
                        set_cat_section(CATSECT_TOP);
                    break;

                case CATSECT_MAT:
                    if (vars_exist(0, 0, 1) == 0)
                        set_cat_section(CATSECT_TOP);
                    break;

                case CATSECT_REAL_ONLY:
                    if (vars_exist(1, 0, 0) == 0)
                    {
                        the_menu = (int)MENU.NONE;
                        redisplay();
                        return;
                    }
                    break;

                case CATSECT_MAT_ONLY:
                    if (vars_exist(0, 0, 1) == 0)
                    {
                        the_menu = (int)MENU.NONE;
                        redisplay();
                        return;
                    }
                    break;

                case CATSECT_VARS_ONLY:
                    if (vars_exist(1, 1, 1) == 0)
                    {
                        the_menu = (int)MENU.NONE;
                        redisplay();
                        return;
                    }
                    break;

                case CATSECT_PGM_SOLVE:
                case CATSECT_PGM_INTEG:
                    if (mvar_prgms_exist() == 0)
                    {
                        the_menu = (int)MENU.NONE;
                        display_error((int)ERR.NO_MENU_VARIABLES, 0);
                        redisplay();
                        return;
                    }
                    break;
            }

            draw_catalog();
        }

        public static int get_cat_section()
        {
            int index = get_cat_index();

            if (index != -1)
                return catalogmenu_section[index];
            else
                return CATSECT_TOP;
        }

        public static int get_cat_item(int menukey)
        {
            int index = get_cat_index();

            if (index == -1)
                return -1;
            else
                return catalogmenu_item[index, menukey];
        }

        public static int get_cat_index()
        {
            if (mode_commandmenu != (int)MENU.NONE)
                return 0;
            else if (mode_alphamenu != (int)MENU.NONE)
                return 1;
            else if (mode_transientmenu != (int)MENU.NONE)
                return 2;
            else if (mode_plainmenu != (int)MENU.NONE)
                return 3;
            else if (mode_appmenu != (int)MENU.NONE)
                return 4;
            else
                return -1;
        }

        public static void set_cat_section(int section)
        {
            int index = get_cat_index();

            if (index != -1)
                catalogmenu_section[index] = section;
        }

        public static void display_error(int error, int print)
        {
            clear_row(0);

            draw_string(0, 0, errors[error]);

            flags.message = 1;
            flags.two_line_message = 0;

            if ((print != 0) && ((flags.trace_print != 0) || (flags.normal_print != 0)) && (flags.printer_exists != 0))
                print_text(errors[error], 1);
        }

        public static void redisplay()
        {
            int menu_id;
            int avail_rows = 2;
            int i;

            if (mode_clall)
            {
                clear_display();

                draw_string(0, 0, "Clear All Memory?");

                draw_key(0, false, false, "YES".ToCharArray(), 3);
                draw_key(1, false, false, "".ToCharArray(), 0);
                draw_key(2, false, false, "".ToCharArray(), 0);
                draw_key(3, false, false, "".ToCharArray(), 0);
                draw_key(4, false, false, "".ToCharArray(), 0);
                draw_key(5, false, false, "NO".ToCharArray(), 2);

                flush_display();
                return;
            }

            if (flags.two_line_message != 0)
                return;

            if (flags.message != 0)
                clear_row(1);
            else
                clear_display();

            if (mode_commandmenu != (int)MENU.NONE)
                menu_id = mode_commandmenu;
            else if (mode_alphamenu != (int)MENU.NONE)
                menu_id = mode_alphamenu;
            else if (mode_transientmenu != (int)MENU.NONE)
                menu_id = mode_transientmenu;
            else if (mode_plainmenu != (int)MENU.NONE)
                menu_id = mode_plainmenu;
            else if (mode_appmenu != (int)MENU.NONE)
                menu_id = mode_appmenu;
            else
                menu_id = (int)MENU.NONE;

            if (menu_id == (int)MENU.CATALOG)
            {
                draw_catalog();
                avail_rows = 1;
            }
            else if (menu_id == (int)MENU.VARMENU)
            {
                draw_varmenu();

                if (varmenu_length == 0)
                {
                    redisplay();
                    return;
                }

                avail_rows = 1;
            }
            else if (menu_id == (int)MENU.CUSTOM1 || menu_id == (int)MENU.CUSTOM2 || menu_id == (int)MENU.CUSTOM3)
            {
                int r = menu_id - (int)MENU.CUSTOM1;

                if ((flags.local_label != 0) && !(mode_command_entry && incomplete_argtype == ARG_CKEY))
                {
                    for (i = 0; i < 5; i++)
                    {
                        char c = (char)((r == 0 ? 'A' : 'F') + i);
                        draw_key(i, false, false, c.ToString().ToCharArray(), 1);
                    }

                    draw_key(5, false, false, "XEQ".ToCharArray(), 3);
                }
                else
                {
                    for (i = 0; i < 6; i++)
                    {
                        draw_key(i, false, true, custommenu_label[r, i].ToCharArray(), custommenu_length[r, i]);
                    }
                }
                avail_rows = 1;
            }
            else if (menu_id == (int)MENU.PROGRAMMABLE)
            {
                for (i = 0; i < 6; i++)
                    draw_key(i, false, false, progmenu_label[i].ToCharArray(), progmenu_length[i]);

                avail_rows = 1;
            }
            else if (menu_id != (int)MENU.NONE)
            {
                menu_spec m = menus[menu_id];

                for (i = 0; i < 6; i++)
                {
                    menu_item_spec mi = m.child[i];

                    if (mi.menuid == MENU.NONE || (((int)(mi.menuid)) & 0x3000) == 0)
                        draw_key(i, false, false, mi.title.ToCharArray(), mi.title.Length);
                    else
                    {
                        int cmd_id = ((int)(mi.menuid)) & 0xfff;

                        command_spec cmd = cmd_array[cmd_id];

                        bool is_flag = ((int)mi.menuid & 0x2000) != 0;
                        if (is_flag)
                        {
                            // Take a closer look at the command ID and highlight
                            // the menu item if appropriate -- that is, clear 'is_flag'
                            // if highlighting is *not* appropriate
                            //
                            switch (cmd_id)
                            {
                                case CMD_FIX:
                                    is_flag = (flags.fix_or_all != 0) && (flags.eng_or_all == 0);
                                    break;

                                case CMD_SCI:
                                    is_flag = (flags.fix_or_all == 0) && (flags.eng_or_all == 0);
                                    break;

                                case CMD_ENG:
                                    is_flag = (flags.fix_or_all == 0) && (flags.eng_or_all != 0);
                                    break;

                                case CMD_ALL:
                                    is_flag = (flags.fix_or_all != 0) && (flags.eng_or_all != 0);
                                    break;

                                case CMD_RDXDOT:
                                    is_flag = (flags.decimal_point != 0);
                                    break;

                                case CMD_RDXCOMMA:
                                    is_flag = (flags.decimal_point == 0);
                                    break;

                                case CMD_DEG:
                                    is_flag = (flags.rad == 0) && (flags.grad == 0);
                                    break;

                                case CMD_RAD:
                                    is_flag = (flags.rad != 0) && (flags.grad == 0);
                                    break;

                                case CMD_GRAD:
                                    is_flag = (flags.grad != 0);
                                    break;

                                case CMD_POLAR:
                                    is_flag = (flags.polar != 0);
                                    break;

                                case CMD_RECT:
                                    is_flag = (flags.polar == 0);
                                    break;

                                case CMD_QUIET:
                                    is_flag = (flags.audio_enable == 0);
                                    break;

                                case CMD_CPXRES:
                                    is_flag = (flags.real_result_only == 0);
                                    break;

                                case CMD_REALRES:
                                    is_flag = (flags.real_result_only != 0);
                                    break;

                                case CMD_KEYASN:
                                    is_flag = (flags.local_label == 0);
                                    break;

                                case CMD_LCLBL:
                                    is_flag = (flags.local_label != 0);
                                    break;

                                case CMD_PON:
                                    is_flag = (flags.printer_exists != 0);
                                    break;

                                case CMD_POFF:
                                    is_flag = (flags.printer_exists == 0);
                                    break;

                                case CMD_MAN:
                                    is_flag = (flags.trace_print == 0) && (flags.normal_print == 0);
                                    break;

                                case CMD_NORM:
                                    is_flag = (flags.trace_print == 0) && (flags.normal_print != 0);
                                    break;

                                case CMD_TRACE:
                                    is_flag = (flags.trace_print != 0);
                                    break;

                                case CMD_ALLSIGMA:
                                    is_flag = (flags.all_sigma != 0);
                                    break;

                                case CMD_LINSIGMA:
                                    is_flag = (flags.all_sigma == 0);
                                    break;

                                case CMD_LINF:
                                    is_flag = (flags.lin_fit != 0);
                                    break;

                                case CMD_LOGF:
                                    is_flag = (flags.log_fit != 0);
                                    break;

                                case CMD_EXPF:
                                    is_flag = (flags.exp_fit != 0);
                                    break;

                                case CMD_PWRF:
                                    is_flag = (flags.pwr_fit != 0);
                                    break;

                                case CMD_WRAP:
                                    is_flag = (flags.grow == 0);
                                    break;

                                case CMD_GROW:
                                    is_flag = (flags.grow != 0);
                                    break;

                                case CMD_BINM:
                                    is_flag = (get_base() == 2);
                                    break;

                                case CMD_OCTM:
                                    is_flag = (get_base() == 8);
                                    break;

                                case CMD_DECM:
                                    is_flag = (get_base() == 10);
                                    break;

                                case CMD_HEXM:
                                    is_flag = (get_base() == 16);
                                    break;
                            }
                        }

                        draw_key(i, is_flag, true, cmd.name.ToCharArray(), cmd.name_length);
                    }
                }

                avail_rows = 1;
            }

            if ((flags.prgm_mode == 0) && (mode_command_entry || (pending_command != CMD_NONE)))
            {
                int cmd_row;

                if (menu_id == (int)MENU.NONE)
                {
                    cmd_row = 1;
                    avail_rows = 1;
                }
                else
                {
                    cmd_row = 0;
                    avail_rows = 0;
                }

                if (mode_command_entry)
                    display_incomplete_command(cmd_row);
                else
                    display_command(cmd_row);
            }

            if ((flags.alpha_mode == 0) && (flags.prgm_mode == 0))
            {
                if (avail_rows == 1)
                {
                    if (flags.message == 0)
                        display_x(0);
                }
                else if (avail_rows == 2)
                {
                    if (flags.message == 0)
                        display_y(0);

                    display_x(1);
                }
            }
            else if ((flags.prgm_mode != 0) && (avail_rows != 0))
            {
                if (mode_command_entry)
                {
                    if (avail_rows == 1)
                        display_incomplete_command(0);
                    else
                    {
                        display_prgm_line(0, -1);
                        display_incomplete_command(1);
                    }
                }
                else
                {
                    if (avail_rows == 1)
                    {
                        if (flags.message == 0)
                            display_prgm_line(0, 0);
                    }
                    else if (avail_rows == 2)
                    {
                        if (flags.message == 0)
                        {
                            if (pc == -1)
                                prgm_highlight_row = 0;
                            else
                            {
                                if (prgms[current_prgm].text[pc] == CMD_END)
                                    prgm_highlight_row = 1;
                            }
                            if (prgm_highlight_row == 0)
                            {
                                display_prgm_line(0, 0);
                                display_prgm_line(1, 1);
                            }
                            else
                            {
                                display_prgm_line(0, -1);
                                display_prgm_line(1, 0);
                            }
                        }
                        else
                            display_prgm_line(1, 0);
                    }
                }
            }
            else if ((flags.alpha_mode != 0) && (avail_rows != 0) && (flags.message == 0))
            {
                int avail = mode_alpha_entry ? 21 : 22;

                if (reg_alpha_length <= avail)
                {
                    draw_string(0, 0, reg_alpha, reg_alpha_length);
                    if (mode_alpha_entry)
                        draw_char(reg_alpha_length, 0, '_');
                }
                else
                {
                    avail--;

                    draw_char(0, 0, (char)26);

                    char[] tmp = new char[avail];
                    Array.Copy(reg_alpha, reg_alpha_length - avail, tmp, 0, avail);
                    draw_string(1, 0, tmp, avail);

                    if (mode_alpha_entry)
                        draw_char(21, 0, '_');
                }
            }

            flush_display();
        }

        public static void display_incomplete_command(int row)
        {
            char[] buf = new char[40];
            int bufptr = 0;

            command_spec cmd = cmd_array[incomplete_command];

            if ((flags.prgm_mode != 0) && (cmd.flags & FLAG_IMMED) == 0)
            {
                int line = pc2line(pc);

                if (line < 10)
                    char2buf(buf, 40, ref bufptr, '0');

                bufptr += int2string(line, buf, bufptr, 40);
                char2buf(buf, 40, ref bufptr, (char)6);
            }

            if (incomplete_command == CMD_ASSIGNb)
            {
                string2buf(buf, 40, ref bufptr, "ASSIGN \"".ToCharArray(), 8);
                string2buf(buf, 40, ref bufptr, pending_command_arg.val.text, pending_command_arg.length);
                string2buf(buf, 40, ref bufptr, "\" TO _".ToCharArray(), 6);
                goto done;
            }

            if (incomplete_argtype == ARG_MKEY)
            {
                // KEYG/KEYX
                string2buf(buf, 40, ref bufptr, "KEY _".ToCharArray(), 5);
                goto done;
            }

            if (incomplete_command == CMD_SIMQ)
                string2buf(buf, 40, ref bufptr, "Number of Unknowns ".ToCharArray(), 19);
            else
            {
                string2buf(buf, 40, ref bufptr, cmd.name.ToCharArray(), cmd.name.Length);
                char2buf(buf, 40, ref bufptr, ' ');
            }

            if (incomplete_ind != 0)
                string2buf(buf, 40, ref bufptr, "IND ".ToCharArray(), 4);

            if (incomplete_alpha != 0)
            {
                char2buf(buf, 40, ref bufptr, '"');
                string2buf(buf, 40, ref bufptr, incomplete_str, incomplete_length);
                char2buf(buf, 40, ref bufptr, '_');
            }
            else
            {
                int d = 1;
                int i;

                for (i = 0; i < incomplete_length - 1; i++)
                    d *= 10;

                for (i = 0; i < incomplete_maxdigits; i++)
                {
                    if (i < incomplete_length)
                    {
                        char2buf(buf, 40, ref bufptr, (char)('0' + (incomplete_num / d) % 10));
                        d /= 10;
                    }
                    else
                        char2buf(buf, 40, ref bufptr, '_');
                }
            }

        done:

            clear_row(row);

            if (bufptr <= 22)
                draw_string(0, row, buf, bufptr);
            else
            {
                draw_char(0, row, (char)26);

                char[] tmp = new char[21];
                Array.Copy(buf, bufptr - 21, tmp, 0, 21);
                draw_string(1, row, tmp, 21);
            }
        }

        public static int get_front_menu()
        {                                                   // Returning value rather than reference
            if (mode_commandmenu != (int)MENU.NONE)         // Orig code returned pointers to mode_*
                return mode_commandmenu;                    // Should be ok if caller does not modify or pass on

            if (mode_alphamenu != (int)MENU.NONE)
                return mode_alphamenu;

            if (mode_transientmenu != (int)MENU.NONE)
                return mode_transientmenu;

            if (mode_plainmenu != (int)MENU.NONE)
                return mode_plainmenu;

            if (mode_appmenu != (int)MENU.NONE)
                return mode_appmenu;

            return (int)MENU.NONE;
        }

        public static void display_command(int row)
        {
            char[] buf = new char[22];
            int bufptr = 0;

            command_spec cmd = cmd_array[pending_command];

            int the_menu;
            int catsect;

            bool hide = pending_command == CMD_XEQ;
            hide &= xeq_invisible != 0;

            the_menu = get_front_menu();
            hide &= the_menu == MENU_CATALOG;

            catsect = get_cat_section();
            hide &= catsect == CATSECT_PGM || catsect == CATSECT_PGM_ONLY;

            hide |= pending_command == CMD_VMEXEC;

            if (pending_command >= CMD_ASGN01 && pending_command <= CMD_ASGN18)
                string2buf(buf, 22, ref bufptr, "ASSIGN ".ToCharArray(), 7);
            else if (!hide)
            {
                if (pending_command == CMD_SIMQ)
                    string2buf(buf, 22, ref bufptr, "Number of Unknowns ".ToCharArray(), 19);
                else
                {
                    string2buf(buf, 22, ref bufptr, cmd.name.ToCharArray(), cmd.name.Length);
                    char2buf(buf, 22, ref bufptr, ' ');
                }
            }

            if (cmd.argtype == ARG_NONE)
                goto done;

            if (pending_command_arg.type == ARGTYPE_IND_NUM
                || pending_command_arg.type == ARGTYPE_IND_STK
                || pending_command_arg.type == ARGTYPE_IND_STR)
                string2buf(buf, 22, ref bufptr, "IND ".ToCharArray(), 4);

            if (pending_command_arg.type == ARGTYPE_NUM
                || pending_command_arg.type == ARGTYPE_IND_NUM)
            {
                int d = 1;
                int i;
                int leadingzero = 1;

                for (i = 0; i < pending_command_arg.length - 1; i++)
                    d *= 10;

                for (i = 0; i < pending_command_arg.length; i++)
                {
                    int digit = (pending_command_arg.val.num / d) % 10;

                    if (digit != 0 || i >= pending_command_arg.length - 2)
                        leadingzero = 0;

                    if (leadingzero == 0)
                        char2buf(buf, 22, ref bufptr, (char)('0' + digit));

                    d /= 10;
                }
            }
            else if (pending_command_arg.type == ARGTYPE_STK
              || pending_command_arg.type == ARGTYPE_IND_STK)
            {
                string2buf(buf, 22, ref bufptr, "ST ".ToCharArray(), 3);
                char2buf(buf, 22, ref bufptr, pending_command_arg.val.stk);
            }
            else if (pending_command_arg.type == ARGTYPE_STR
                || pending_command_arg.type == ARGTYPE_IND_STR)
            {
                char2buf(buf, 22, ref bufptr, '"');
                string2buf(buf, 22, ref bufptr, pending_command_arg.val.text, pending_command_arg.length);
                char2buf(buf, 22, ref bufptr, '"');
            }
            else if (pending_command_arg.type == ARGTYPE_LBLINDEX)
            {
                int labelindex = pending_command_arg.val.num;

                if (labels[labelindex].length == 0)
                    if (labelindex == labels_count - 1)
                        string2buf(buf, 22, ref bufptr, ".END.".ToCharArray(), 5);
                    else
                        string2buf(buf, 22, ref bufptr, "END".ToCharArray(), 3);
                else
                {
                    char2buf(buf, 22, ref bufptr, '"');
                    string2buf(buf, 22, ref bufptr, labels[labelindex].name, labels[labelindex].length);
                    char2buf(buf, 22, ref bufptr, '"');
                }
            }
            else // ARGTYPE_LCLBL
            {
                char2buf(buf, 22, ref bufptr, pending_command_arg.val.lclbl);
            }

            if (pending_command >= CMD_ASGN01 && pending_command <= CMD_ASGN18)
            {
                int keynum = pending_command - CMD_ASGN01 + 1;

                string2buf(buf, 22, ref bufptr, " TO ".ToCharArray(), 4);
                char2buf(buf, 22, ref bufptr, (char)('0' + keynum / 10));
                char2buf(buf, 22, ref bufptr, (char)('0' + keynum % 10));
            }

        done:
            clear_row(row);
            draw_string(0, row, buf, bufptr);
        }

        public static void display_prgm_line(int row, int line_offset)
        {
            int tmppc = pc;
            int tmpline = pc2line(pc);
            int cmd = 0;
            arg_struct arg = new arg_struct();
            char[] buf = new char[44];
            int bufptr;
            int len = 22;

            if (row == -1)          // This means use both lines; used by SHOW
                len = 44;

            if (tmpline != 0)
            {
                if (line_offset == 0)
                    get_next_command(ref tmppc, ref cmd, arg, 0);
                else if (line_offset == 1)
                {
                    tmppc += get_command_length(current_prgm, tmppc);
                    tmpline++;
                    get_next_command(ref tmppc, ref cmd, arg, 0);
                }
                else                // line_offset == -1
                {
                    tmpline--;

                    if (tmpline != 0)
                    {
                        tmppc = line2pc(tmpline);
                        get_next_command(ref tmppc, ref cmd, arg, 0);
                    }
                }
            }
            else
            {
                if (line_offset == 0)
                {
                    // Nothing to do
                }
                else if (line_offset == 1)
                {
                    tmppc = 0;
                    get_next_command(ref tmppc, ref cmd, arg, 0);
                    tmpline++;
                }

                // Should not get offset == -1 when at line 0!
            }

            bufptr = prgmline2buf(buf, len, tmpline, line_offset == 0, cmd, arg);

            if (row == -1)
            {
                clear_display();

                if (bufptr <= 22)
                    draw_string(0, 0, buf, bufptr);
                else
                {
                    draw_string(0, 0, buf, 22);

                    char[] tmp = new char[44];
                    Array.Copy(buf, 22, tmp, 0, bufptr - 22);
                    draw_string(0, 1, tmp, bufptr - 22);
                }
            }
            else
            {
                clear_row(row);
                draw_string(0, row, buf, bufptr);
            }
        }

        public static int prgmline2buf(char[] buf, int len, int line, bool highlight, int cmd, arg_struct arg)
        {
            int bufptr = 0;

            if (line != -1)
            {
                if (line < 10)
                    char2buf(buf, len, ref bufptr, '0');

                bufptr += int2string(line, buf, bufptr, len);

                if (highlight)
                    char2buf(buf, len, ref bufptr, (char)6);
                else
                    char2buf(buf, len, ref bufptr, ' ');
            }

            if (line == 0)
            {
                int size = core_program_size(current_prgm);

                string2buf(buf, len, ref bufptr, "{ ".ToCharArray(), 2);

                bufptr += int2string(size, buf, bufptr, len);
                string2buf(buf, len, ref bufptr, "-Byte Prgm }".ToCharArray(), 12);
            }
            else if ((flags.alpha_mode != 0) && mode_alpha_entry && highlight)
            {
                bool append = entered_string_length > 0 && entered_string[0] == 127;

                if (append)
                {
                    string2buf(buf, len, ref bufptr, "\x007f\"".ToCharArray(), 2);

                    char[] tmp = new char[buf.Length];
                    Array.Copy(entered_string, 1, tmp, 0, entered_string_length - 1);
                    string2buf(buf, len, ref bufptr, tmp, entered_string_length - 1);
                }
                else
                {
                    char2buf(buf, len, ref bufptr, '"');
                    string2buf(buf, len, ref bufptr, entered_string, entered_string_length);
                }

                char2buf(buf, len, ref bufptr, '_');
            }
            else if (cmd == CMD_END && current_prgm == prgms_count - 1)
            {
                string2buf(buf, len, ref bufptr, ".END.".ToCharArray(), 5);
            }
            else if (cmd == CMD_NUMBER)
            {
                char[] num = phloat2program(arg.val_d);

                char c;
                int i = 0;
                while ((c = num[i++]) != 0 && bufptr < len)
                    buf[bufptr++] = c;

                if (c != 0)
                    buf[bufptr - 1] = (char)26;
            }
            else if (cmd == CMD_STRING)
            {
                int append = (arg.length > 0 && arg.val.text[0] == 127) ? 1 : 0;

                if (append != 0)
                    char2buf(buf, len, ref bufptr, (char)127);

                char2buf(buf, len, ref bufptr, '"');

                char[] tmp = new char[arg.length - append];
                Array.Copy(arg.val.text, append, tmp, 0, arg.length - append);
                string2buf(buf, len, ref bufptr, tmp, arg.length - append);

                char2buf(buf, len, ref bufptr, '"');
            }
            else
            {
                char[] tmp = new char[len - bufptr];
                int chars = command2buf(tmp, len - bufptr, cmd, arg);
                Array.Copy(tmp, 0, buf, bufptr, len - bufptr);
                bufptr += chars;
            }

            return bufptr;
        }

        public static int command2buf(char[] buf, int len, int cmd, arg_struct arg)
        {
            int bufptr = 0;
            int xrom_arg = 0;

            if (!core_settings.enable_ext_copan && cmd >= CMD_OPENF && cmd <= CMD_DELP
                || !core_settings.enable_ext_bigstack && cmd == CMD_DROP
                || !core_settings.enable_ext_accel && cmd == CMD_ACCEL
                || !core_settings.enable_ext_locat && cmd == CMD_LOCAT
                || !core_settings.enable_ext_heading && cmd == CMD_HEADING
                || !core_settings.enable_ext_time && cmd >= CMD_ADATE && cmd <= CMD_SWPT
                || ((cmd_array[cmd].hp42s_code & 0xfffff800) == 0x0000a000) && ((cmd_array[cmd].flags & FLAG_HIDDEN) != 0))
            {
                xrom_arg = (int)cmd_array[cmd].hp42s_code;
                cmd = CMD_XROM;
            }
            else if (cmd == CMD_XROM)
                xrom_arg = arg.val.num;

            command_spec cmdspec = cmd_array[cmd];

            if (cmd >= CMD_ASGN01 && cmd <= CMD_ASGN18)
                string2buf(buf, len, ref bufptr, "ASSIGN ".ToCharArray(), 7);
            else
                string2buf(buf, len, ref bufptr, cmdspec.name.ToCharArray(), cmdspec.name.Length);

            if (cmd == CMD_XROM)
            {
                int n = xrom_arg & 0x7FF;
                int rom = n >> 6;
                int instr = n & 63;

                char2buf(buf, len, ref bufptr, ' ');
                char2buf(buf, len, ref bufptr, (char)('0' + rom / 10));
                char2buf(buf, len, ref bufptr, (char)('0' + rom % 10));
                char2buf(buf, len, ref bufptr, ',');
                char2buf(buf, len, ref bufptr, (char)('0' + instr / 10));
                char2buf(buf, len, ref bufptr, (char)('0' + instr % 10));
            }
            else if (cmdspec.argtype != ARG_NONE)
            {
                if (cmdspec.name_length > 0)
                    char2buf(buf, len, ref bufptr, ' ');

                if (arg.type == ARGTYPE_IND_NUM || arg.type == ARGTYPE_IND_STK || arg.type == ARGTYPE_IND_STR)
                    string2buf(buf, len, ref bufptr, "IND ".ToCharArray(), 4);

                if (arg.type == ARGTYPE_NUM || arg.type == ARGTYPE_IND_NUM)
                {
                    int digits = arg.type == ARGTYPE_IND_NUM ? 2 : cmdspec.argtype == ARG_NUM9 ? 1 : 2;

                    int d = 1;
                    int i;

                    for (i = 0; i < digits - 1; i++)
                        d *= 10;

                    while (arg.val.num >= d * 10)
                    {
                        d *= 10;
                        digits++;
                    }

                    for (i = 0; i < digits; i++)
                    {
                        char2buf(buf, len, ref bufptr, (char)('0' + (arg.val.num / d) % 10));
                        d /= 10;
                    }
                }
                else if (arg.type == ARGTYPE_STK || arg.type == ARGTYPE_IND_STK)
                {
                    string2buf(buf, len, ref bufptr, "ST ".ToCharArray(), 3);
                    char2buf(buf, len, ref bufptr, arg.val.stk);
                }
                else if (arg.type == ARGTYPE_STR || arg.type == ARGTYPE_IND_STR)
                {
                    char2buf(buf, len, ref bufptr, '"');
                    string2buf(buf, len, ref bufptr, arg.val.text, arg.length);
                    char2buf(buf, len, ref bufptr, '"');
                }
                else if (arg.type == ARGTYPE_LCLBL)
                {
                    char2buf(buf, len, ref bufptr, arg.val.lclbl);
                }
                else if (arg.type == ARGTYPE_LBLINDEX)
                {
                    label_struct lbl = labels[arg.val.num];

                    if (lbl.length == 0)
                    {
                        if (arg.val.num == labels_count - 1)
                            string2buf(buf, len, ref bufptr, ".END.".ToCharArray(), 5);
                        else
                            string2buf(buf, len, ref bufptr, "END".ToCharArray(), 3);
                    }
                    else
                    {
                        char2buf(buf, len, ref bufptr, '"');
                        string2buf(buf, len, ref bufptr, lbl.name, lbl.length);
                        char2buf(buf, len, ref bufptr, '"');
                    }
                }
                else // ARGTYPE_COMMAND; for backward compatibility only
                {
                    command_spec cs = cmd_array[arg.val.cmd];

                    char2buf(buf, len, ref bufptr, '"');
                    string2buf(buf, len, ref bufptr, cs.name.ToCharArray(), cs.name.Length);
                    char2buf(buf, len, ref bufptr, '"');
                }
            }

            if (cmd >= CMD_ASGN01 && cmd <= CMD_ASGN18)
            {
                int keynum = cmd - CMD_ASGN01 + 1;

                string2buf(buf, len, ref bufptr, " TO ".ToCharArray(), 4);
                char2buf(buf, len, ref bufptr, (char)('0' + keynum / 10));
                char2buf(buf, len, ref bufptr, (char)('0' + keynum % 10));
            }

            return bufptr;
        }

        public static void draw_varmenu()
        {
            arg_struct arg = new arg_struct();
            int saved_prgm;
            int prgm = 0;
            int pc = 0;
            int pc2;
            int command = 0;
            int i;
            int row;
            int key;
            int num_mvars = 0;

            if (mode_appmenu != (int)MENU.VARMENU)
                return;

            arg.type = ARGTYPE_STR;
            arg.length = (byte)varmenu_length;

            for (i = 0; i < arg.length; i++)
                arg.val.text[i] = varmenu[i];

            if (find_global_label(arg, ref prgm, ref pc) == 0)
            {
                set_appmenu((int)MENU.NONE);
                varmenu_length = 0;

                return;
            }

            saved_prgm = current_prgm;
            current_prgm = prgm;

            pc += get_command_length(prgm, pc);
            pc2 = pc;

            get_next_command(ref pc, ref command, arg, 0);

            while (command == CMD_MVAR)
            {
                num_mvars++;
                get_next_command(ref pc, ref command, arg, 0);
            }

            if (num_mvars == 0)
            {
                current_prgm = saved_prgm;
                set_appmenu((int)MENU.NONE);
                varmenu_length = 0;
                return;
            }

            varmenu_rows = (num_mvars + 5) / 6;

            if (varmenu_row >= varmenu_rows)
                varmenu_row = varmenu_rows - 1;

            row = 0;
            key = 0;

            get_next_command(ref pc2, ref command, arg, 0);

            while (command == CMD_MVAR)
            {
                if (row == varmenu_row)
                {
                    varmenu_labellength[key] = arg.length;

                    for (i = 0; i < arg.length; i++)
                        varmenu_labeltext[key][i] = arg.val.text[i];

                    draw_key(key, false, false, arg.val.text, arg.length);
                }

                if (key++ == 5)
                {
                    if (row++ == varmenu_row)
                        break;
                    else
                        key = 0;
                }

                get_next_command(ref pc2, ref command, arg, 0);
            }

            current_prgm = saved_prgm;

            while (key < 6)
            {
                varmenu_labellength[key] = 0;
                draw_key(key, false, false, "".ToCharArray(), 0);
                key++;
            }
        }

        public static void draw_catalog()
        {
            int catsect = get_cat_section();
            int catindex = get_cat_index();

            if (catsect == CATSECT_TOP)
            {
                // draw_top:                        // Eliminated use of goto
                draw_key(0, false, false, "FCN".ToCharArray(), 3);
                draw_key(1, false, false, "PGM".ToCharArray(), 3);
                draw_key(2, false, false, "REAL".ToCharArray(), 4);
                draw_key(3, false, false, "CPX".ToCharArray(), 3);
                draw_key(4, false, false, "MAT".ToCharArray(), 3);
                draw_key(5, false, false, "MEM".ToCharArray(), 3);

                mode_updown = false;
                shell_annunciators(0, -1, -1, -1, -1, -1);
            }
            else if (catsect == CATSECT_PGM || catsect == CATSECT_PGM_ONLY || catsect == CATSECT_PGM_SOLVE || catsect == CATSECT_PGM_INTEG)
            {
                // Show menu of alpha labels
                int lcount = 0;
                int i, j, k = -1;

                if (catsect == CATSECT_PGM || catsect == CATSECT_PGM_ONLY)
                {
                    int lastprgm = -1;

                    for (i = 0; i < labels_count; i++)
                    {
                        if (labels[i].length > 0 || labels[i].prgm != lastprgm)
                            lcount++;

                        lastprgm = labels[i].prgm;
                    }
                }
                else // CATSECT_PGM_SOLVE, CATSECT_PGM_INTEG
                {
                    for (i = 0; i < labels_count; i++)
                        if (label_has_mvar(i) != 0)
                            lcount++;
                }

                catalogmenu_rows[catindex] = (lcount + 5) / 6;

                if (catalogmenu_row[catindex] >= catalogmenu_rows[catindex])
                    catalogmenu_row[catindex] = catalogmenu_rows[catindex] - 1;

                j = -1;

                for (i = labels_count - 1; i >= 0; i--)
                {
                    int show_this_label = 0;

                    if (catsect == CATSECT_PGM || catsect == CATSECT_PGM_ONLY)
                        show_this_label = (labels[i].length > 0 || i == 0 || labels[i - 1].prgm != labels[i].prgm) ? 1 : 0;
                    else
                        show_this_label = label_has_mvar(i);

                    if (show_this_label != 0)
                    {
                        j++;

                        if (j / 6 == catalogmenu_row[catindex])
                        {
                            int len = labels[i].length;
                            k = j % 6;

                            if (len == 0)
                            {
                                if (i == labels_count - 1)
                                    draw_key(k, false, false, ".END.".ToCharArray(), 5);
                                else
                                    draw_key(k, false, false, "END".ToCharArray(), 3);
                            }
                            else
                                draw_key(k, false, false, labels[i].name, labels[i].length);

                            catalogmenu_item[catindex, k] = i;

                            if (k == 5)
                                break;
                        }
                    }
                }

                while (k < 5)
                {
                    draw_key(++k, false, false, "".ToCharArray(), 0);
                    catalogmenu_item[catindex, k] = -1;
                }

                mode_updown = catalogmenu_rows[catindex] > 1;

                shell_annunciators(mode_updown ? 1 : 0, -1, -1, -1, -1, -1);
            }
            else if (catsect == CATSECT_FCN)
            {
                int desired_row = catalogmenu_row[catindex];

                if (desired_row < 42)
                {
	                for (int i = 0; i < 6; i++)
                    {
		                int cmd = fcn_cat[desired_row * 6 + i];
		                catalogmenu_item[catindex, i] = cmd;

		                draw_key(i, false, true, cmd_array[cmd].name.ToCharArray(), cmd_array[cmd].name_length);
	                }

	                // Setting number of rows to a ridiculously large value;
	                // this value only comes into play when the user presses "up"
	                // while on the first row of the FCN catalog, and the extension-handling
	                // code in the desired_row >= 42 case takes care of handling the
	                // ever-changing number of rows in this menu, and will clip the row
	                // number to the actual number of rows.

                    // SLR - Boundary fix for no extensions
                    if (core_settings.enable_ext_copan ||
                        core_settings.enable_ext_bigstack ||
                        core_settings.enable_ext_accel ||
                        core_settings.enable_ext_locat ||
                        core_settings.enable_ext_heading ||
                        core_settings.enable_ext_time)
                        catalogmenu_rows[catindex] = 1000;
                    else
                        catalogmenu_rows[catindex] = 42;
            	}
                else
                {
	                int curr_row = 41;
	                int curr_pos = 5;
	                bool menu_full = false;
	                int menu_length = 0;
	            
                    for (int extno = 0; extensions[extno].first_cmd != CMD_NULL; extno++)
                    {
		                if (extensions[extno].enable_flag)
                        {
		                    for (int cmd = extensions[extno].first_cmd; cmd <= extensions[extno].last_cmd; cmd++)
                            {
			                    if ((cmd_array[cmd].flags & FLAG_HIDDEN) == 0)
                                {
			                        if (curr_pos == 5)
                                    {
				                        curr_pos = 0;
				                        curr_row++;
			                        }
                                    else
            				            curr_pos++;

        			                if (menu_full)
		            		            goto done;

			                        catalogmenu_item[catindex, curr_pos] = cmd;
			                        catalogmenu_row[catindex] = curr_row;

			                        menu_length = curr_pos + 1;

                                    if (curr_pos == 5 && curr_row == desired_row)
				                        menu_full = true;
			                    }
                            }
                        }
                    }

                done:
                    catalogmenu_rows[catindex] = curr_row + 1;
                    
                    for (int i = 0; i < 6; i++)
                    {
                        if (i >= menu_length)
                            catalogmenu_item[catindex, i] = -1;
                    
                        int cmd = catalogmenu_item[catindex, i];
                        
                        if (cmd == -1)
                            draw_key(i, false, false, "".ToCharArray(), 0);
                        else
                            draw_key(i, false, true, cmd_array[cmd].name.ToCharArray(), cmd_array[cmd].name_length);
                    }
                }

                mode_updown = true;

                shell_annunciators(1, -1, -1, -1, -1, -1);
            }
            else
            {
                int vcount = 0;
                int i, j, k = -1;
                int show_real = 1;
                int show_cpx = 1;
                int show_mat = 1;

                switch (catsect)
                {
                    case CATSECT_REAL:
                    case CATSECT_REAL_ONLY:
                        show_cpx = show_mat = 0;
                        break;

                    case CATSECT_CPX:
                        show_real = show_mat = 0;
                        break;

                    case CATSECT_MAT:
                    case CATSECT_MAT_ONLY:
                        show_real = show_cpx = 0;
                        break;
                }

                for (i = 0; i < vars_count; i++)
                {
                    int type = (int)vars[i].value.type;

                    switch (type)
                    {
                        case TYPE_REAL:
                        case TYPE_STRING:
                            if (show_real != 0)
                                vcount++;
                            break;

                        case TYPE_COMPLEX:
                            if (show_cpx != 0)
                                vcount++;
                            break;

                        case TYPE_REALMATRIX:
                        case TYPE_COMPLEXMATRIX:
                            if (show_mat != 0)
                                vcount++;
                            break;
                    }
                }
                if (vcount == 0)
                {
                    // We should only get here if the 'plainmenu' catalog is
                    // in operation; the other catalogs only operate during
                    // command entry mode, or are label catalogs -- so in those
                    // cases, it is possible to prevent empty catalogs from
                    // being displayed *in advance* (i.e., check if any real
                    // variables exist before enabling MENU_CATALOG with
                    // catalogmenu_section = CATSECT_REAL, etc.).
                    // When a catalog becomes empty while displayed, we move
                    // to the top level silently. The "No XXX Variables" message
                    // is only displayed if the user actively tries to enter
                    // an empty catalog section.
                    //
                    set_cat_section(CATSECT_TOP);



                    // draw_top code replicated here and then exit

                    // goto draw_top;

                    draw_key(0, false, false, "FCN".ToCharArray(), 3);
                    draw_key(1, false, false, "PGM".ToCharArray(), 3);
                    draw_key(2, false, false, "REAL".ToCharArray(), 4);
                    draw_key(3, false, false, "CPX".ToCharArray(), 3);
                    draw_key(4, false, false, "MAT".ToCharArray(), 3);
                    draw_key(5, false, false, "MEM".ToCharArray(), 3);

                    mode_updown = false;
                    shell_annunciators(0, -1, -1, -1, -1, -1);
                    return;
                }

                catalogmenu_rows[catindex] = (vcount + 5) / 6;

                if (catalogmenu_row[catindex] >= catalogmenu_rows[catindex])
                    catalogmenu_row[catindex] = catalogmenu_rows[catindex] - 1;

                j = -1;

                for (i = vars_count - 1; i >= 0; i--)
                {
                    int type = (int)vars[i].value.type;

                    switch (type)
                    {
                        case TYPE_REAL:
                        case TYPE_STRING:
                            if (show_real != 0)
                                break;
                            else
                                continue;

                        case TYPE_COMPLEX:
                            if (show_cpx != 0)
                                break;
                            else
                                continue;

                        case TYPE_REALMATRIX:
                        case TYPE_COMPLEXMATRIX:
                            if (show_mat != 0)
                                break;
                            else
                                continue;
                    }

                    j++;

                    if (j / 6 == catalogmenu_row[catindex])
                    {
                        k = j % 6;

                        draw_key(k, false, false, vars[i].name, vars[i].length);

                        catalogmenu_item[catindex, k] = i;

                        if (k == 5)
                            break;
                    }
                }

                while (k < 5)
                {
                    draw_key(++k, false, false, "".ToCharArray(), 0);

                    catalogmenu_item[catindex, k] = -1;
                }

                mode_updown = catalogmenu_rows[catindex] > 1;

                shell_annunciators(mode_updown ? 1 : 0, -1, -1, -1, -1, -1);
            }
        }

        public static int set_menu_return_err(int level, int menuid)
        {
            int newmenu;
            int err;

            switch (level)
            {
                case MENULEVEL_COMMAND:
                    mode_commandmenu = menuid;
                    goto lbl_03;

                case MENULEVEL_ALPHA:
                    mode_alphamenu = menuid;
                    goto lbl_02;

                case MENULEVEL_TRANSIENT:
                    mode_transientmenu = menuid;
                    goto lbl_01;

                case MENULEVEL_PLAIN:
                    mode_plainmenu = menuid;
                    goto lbl_00;

                case MENULEVEL_APP:
                    err = set_appmenu(menuid);

                    if (err != (int)ERR.NONE)
                        return err;
                    break;
            }

            mode_plainmenu = (int)MENU.NONE;

        lbl_00: mode_transientmenu = (int)MENU.NONE;
        lbl_01: mode_alphamenu = (int)MENU.NONE;
        lbl_02: mode_commandmenu = (int)MENU.NONE;
        lbl_03:

            newmenu = get_front_menu();

            if (newmenu != (int)MENU.NONE)
            {
                if (newmenu == (int)MENU.CATALOG)
                {
                    int index = get_cat_index();
                    mode_updown = index != -1 && catalogmenu_rows[index] > 1;
                }
                else if (newmenu == (int)MENU.PROGRAMMABLE)
                {
                    // The programmable menu's up/down annunciator is on if the UP
                    // and/or DOWN keys have been assigned to.
                    // This is something the original HP-42S doesn't do, but I couldn't
                    // resist this little improvement, perfect compatibility be damned.
                    // In my defense, the Programming Examples and Techniques book,
                    // bottom of page 34, does state that this should work.
                    // Can't say whether the fact that it doesn't work on the real
                    // HP-42S is a bug, or whether the coders and the documentation
                    // writers just had a misunderstanding.
                    //
                    mode_updown = progmenu_arg[6].type != ARGTYPE_NONE || progmenu_arg[7].type != ARGTYPE_NONE;
                }
                else
                {
                    // The up/down annunciator for catalogs depends on how many
                    // items they contain; this is handled in draw_catalog().
                    //
                    mode_updown = newmenu == (int)MENU.VARMENU ? varmenu_rows > 1 : menus[newmenu].next != MENU.NONE;
                }
            }
            else
                mode_updown = false;

            shell_annunciators(mode_updown ? 1 : 0, -1, -1, -1, -1, -1);

            return (int)ERR.NONE;
        }

        public static void set_menu(int level, int menuid)
        {
            int err = set_menu_return_err(level, menuid);

            if (err != ERR_NONE)
            {
                display_error(err, 1);
                flush_display();
            }
        }

        public static void move_cat_row(int direction)
        {
            int index = get_cat_index();

            if (index == -1)
                return;

            if (direction == 0)
                catalogmenu_row[index] = 0;
            else if (direction == -1)
            {
                catalogmenu_row[index]--;

                if (catalogmenu_row[index] < 0)
                    catalogmenu_row[index] = catalogmenu_rows[index] - 1;
            }
            else
            {
                catalogmenu_row[index]++;

                if (catalogmenu_row[index] >= catalogmenu_rows[index])
                    catalogmenu_row[index] = 0;
            }
        }

        public static void set_catalog_menu(int section)
        {
            mode_commandmenu = (int)MENU.CATALOG;

            move_cat_row(0);
            set_cat_section(section);

            switch (section)
            {
                case CATSECT_TOP:
                case CATSECT_FCN:
                case CATSECT_PGM:
                case CATSECT_PGM_ONLY:
                    return;
                case CATSECT_REAL:
                case CATSECT_REAL_ONLY:
                    if (vars_exist(1, 0, 0) == 0)
                        mode_commandmenu = (int)MENU.NONE;
                    return;

                case CATSECT_CPX:
                    if (vars_exist(0, 1, 0) == 0)
                        mode_commandmenu = (int)MENU.NONE;
                    return;

                case CATSECT_MAT:
                case CATSECT_MAT_ONLY:
                    if (vars_exist(0, 0, 1) == 0)
                        mode_commandmenu = (int)MENU.NONE;
                    return;

                case CATSECT_VARS_ONLY:
                    if (vars_exist(1, 1, 1) == 0)
                        mode_commandmenu = (int)MENU.NONE;
                    return;

                case CATSECT_PGM_SOLVE:
                case CATSECT_PGM_INTEG:
                default:
                    mode_commandmenu = (int)MENU.NONE;
                    return;
            }
        }

        public static int get_cat_row()
        {
            int index = get_cat_index();

            if (index == -1)
                return 0;
            else
                return catalogmenu_row[index];
        }

        public static void set_cat_row(int row)
        {
            int index = get_cat_index();

            if (index == -1)
                return;
            catalogmenu_row[index] = row;
        }

        public static void get_char(char[] bits, char c)
        {
            // TO DO - it should be possible to simply return bigchars[c],
            // but that crashes on the Palm. Looks like a compiler bug.
            //
            byte uc = (byte)c;

            if (uc >= 130)
                uc -= 128;

            for (int i = 0; i < 5; i++)
                bits[i] = (char)bigchars[uc, i];
        }

        public static int set_appmenu(int menuid)
        {
            if (mode_appmenu != (int)MENU.NONE && appmenu_exitcallback != 0)
            {
                // We delegate the set_menu() call to the callback,
                // but only once. If the callback wants to stay active,
                // it will have to call set_appmenu_callback() itself
                // to reinstate itself.
                //

                int cb = appmenu_exitcallback;
                appmenu_exitcallback = 0;

                // NOTE: I don't use a 'traditional' callback pointer
                // because appmenu_exitcallback has to be persistable,
                // and pointers to code do not have that property.
                //

                switch (cb)
                {
                    case 1: return appmenu_exitcallback_1(menuid);
                    case 2: return appmenu_exitcallback_2(menuid);
                    case 3: return appmenu_exitcallback_3(menuid);
                    case 4: return appmenu_exitcallback_4(menuid);
                    case 5: return appmenu_exitcallback_5(menuid);
                    default: return (int)ERR.INTERNAL_ERROR;
                }
            }
            else
            {
                mode_appmenu = menuid;
                appmenu_exitcallback = 0;

                return (int)ERR.NONE;
            }
        }

        public static void squeak()
        {
            if (flags.audio_enable != 0)
                shell_beeper(10);
        }

        public static void tone(int n)
        {
            if (flags.audio_enable != 0)
                shell_beeper(n);
        }

        public static void display_mem()
        {
            uint bytes = shell_get_mem();

            char[] buf = new char[16];
            int buflen;

            clear_display();
            //            draw_string(0, 0, "Available Memory:");
            draw_string(0, 0, "Used Memory:");

            buflen = uint2string(bytes, buf, 0, 16);

            draw_string(0, 1, buf, buflen);
            draw_string(buflen + 1, 1, "Bytes");
            flush_display();
        }

        public static void set_plainmenu(int menuid)
        {
            mode_commandmenu = (int)MENU.NONE;
            mode_alphamenu = (int)MENU.NONE;
            mode_transientmenu = (int)MENU.NONE;

            if (menuid == mode_plainmenu)
                mode_plainmenu_sticky = true;
            else if (menuid == MENU_CUSTOM1 || menuid == MENU_CUSTOM2 || menuid == MENU_CUSTOM3)
            {
                mode_plainmenu = menuid;
                mode_plainmenu_sticky = true;
                redisplay();
                mode_updown = true;
                shell_annunciators(1, -1, -1, -1, -1, -1);
            }
            else
            {
                // Even if it's a different menu than the current one, it should
                // still stick if it belongs to the same group.
                //
                if (mode_plainmenu != (int)MENU.NONE)
                {
                    int menu1 = mode_plainmenu;
                    int menu2 = menuid;
                    int parent;

                    while ((parent = (int)menus[menu1].parent) != (int)MENU.NONE)
                        menu1 = parent;

                    while ((parent = (int)menus[menu2].parent) != (int)MENU.NONE)
                        menu2 = parent;

                    if (menu1 == menu2)
                        mode_plainmenu_sticky = true;
                    else if (menus[menu1].next == MENU.NONE)
                        mode_plainmenu_sticky = false;
                    else
                    {
                        int m = menu1;

                        mode_plainmenu_sticky = false;

                        do
                        {
                            m = (int)menus[m].next;

                            if (m == menu2)
                            {
                                mode_plainmenu_sticky = true;
                                break;
                            }
                        } while (m != menu1);
                    }
                }
                else
                    mode_plainmenu_sticky = false;

                if (!mode_plainmenu_sticky)
                {
                    mode_plainmenu = menuid;

                    if (mode_plainmenu == MENU_CATALOG)
                        set_cat_section(CATSECT_TOP);

                    redisplay();
                }

                mode_updown = mode_plainmenu != (int)MENU.NONE && menus[mode_plainmenu].next != MENU.NONE;

                shell_annunciators(mode_updown ? 1 : 0, -1, -1, -1, -1, -1);
            }
        }

        public static void set_appmenu_exitcallback(int callback_id)
        {
            appmenu_exitcallback = callback_id;
        }

        public static void get_custom_key(int keynum, char[] name, ref int length)
        {
            int row = (keynum - 1) / 6;
            int key = (keynum - 1) % 6;

            string_copy(name, ref length, custommenu_label[row, key].ToCharArray(), custommenu_length[row, key]);
        }

        public static void do_prgm_menu_key(int keynum)
        {
            int err;
            int oldprgm;
            int oldpc;

            keynum--;

            if (keynum == 8)
                set_menu(MENULEVEL_PLAIN, (int)MENU.NONE);

            if (progmenu_arg[keynum].type == ARGTYPE_NONE)
            {
                if (keynum < 6)
                    pending_command = CMD_NULL;
                else if (keynum == 8)
                    pending_command = CMD_CANCELLED;
                return;
            }

            if ((flags.trace_print != 0 || flags.normal_print != 0) && flags.printer_exists != 0)
                print_command(progmenu_is_gto[keynum] != 0 ? CMD_GTO : CMD_XEQ, progmenu_arg[keynum]);

            oldprgm = current_prgm;
            oldpc = pc;

            set_running(true);

            progmenu_arg[keynum].target = -1; // Force docmd_gto() to search
            err = docmd_gto(progmenu_arg[keynum]);

            if (err != ERR_NONE)
            {
                set_running(false);
                display_error(err, 1);
                flush_display();
                return;
            }

            if (progmenu_is_gto[keynum] == 0)
            {
                if (oldpc == -1)
                    oldpc = 0;

                err = push_rtn_addr(oldprgm, oldpc);

                if (err != ERR_NONE)
                {
                    // Solve/Integ RTN Lost. Someone is writing weird programs
                    // if they're using a programmable menu in the middle of
                    // a solver invocation...
                    //
                    set_running(false);
                    display_error(err, 1);
                    flush_display();
                    return;
                }
            }
        }

        public static void show()
        {
            char[] tmp;

            if (flags.prgm_mode != 0)
                display_prgm_line(-1, 0);
            else if (flags.alpha_mode != 0)
            {
                clear_display();

                if (reg_alpha_length <= 22)
                    draw_string(0, 0, reg_alpha, reg_alpha_length);
                else
                {
                    draw_string(0, 0, reg_alpha, 22);

                    tmp = new char[reg_alpha_length - 22];
                    Array.Copy(reg_alpha, 22, tmp, 0, reg_alpha_length - 22);
                    draw_string(0, 1, tmp, reg_alpha_length - 22);
                }
            }
            else
            {
                char[] buf = new char[44];
                int bufptr;

                clear_display();

                switch (reg_x.type)
                {
                    case TYPE.REAL:
                        bufptr = Phloat.phloat2string(((vartype_real)reg_x).x, buf, 44,
                                       2, 0, 3,
                                       flags.thousands_separators);

                        if (bufptr <= 22)
                            draw_string(0, 0, buf, bufptr);
                        else
                        {
                            draw_string(0, 0, buf, 22);

                            tmp = new char[bufptr - 22];
                            Array.Copy(buf, 22, tmp, 0, bufptr - 22);
                            draw_string(0, 1, tmp, bufptr - 22);
                        }
                        break;

                    case TYPE.STRING:
                        vartype_string s = (vartype_string)reg_x;

                        draw_char(0, 0, '"');
                        draw_string(1, 0, s.text, s.length);
                        draw_char(s.length + 1, 0, '"');
                        break;

                    case TYPE.COMPLEX:
                        vartype_complex c = (vartype_complex)reg_x;

                        Phloat x = new Phloat();
                        Phloat y = new Phloat();

                        if (flags.polar != 0)
                        {
                            generic_r2p(c.re, c.im, ref x, ref y);

                            if (Phloat.p_isinf(x) != 0)
                                x = Phloat.POS_HUGE_PHLOAT;
                        }
                        else
                        {
                            x = c.re;
                            y = c.im;
                        }

                        bufptr = Phloat.phloat2string(x, buf, 22,
                                       0, 0, 3,
                                       flags.thousands_separators);
                        draw_string(0, 0, buf, bufptr);

                        bufptr = Phloat.phloat2string(y, buf, 22,
                                       0, 0, 3,
                                       flags.thousands_separators);

                        if (flags.polar != 0)
                        {
                            draw_char(0, 1, (char)23);
                            draw_string(1, 1, buf, bufptr);
                        }
                        else
                        {
                            if (buf[0] == '-')
                            {
                                draw_string(0, 1, "-i".ToCharArray(), 2);

                                tmp = new char[bufptr - 1];
                                Array.Copy(buf, 1, tmp, 0, bufptr - 1);
                                draw_string(2, 1, tmp, bufptr - 1);
                            }
                            else
                            {
                                draw_char(0, 1, 'i');
                                draw_string(1, 1, buf, bufptr);
                            }
                        }
                        break;

                    case TYPE.REALMATRIX:
                        vartype_realmatrix rm = (vartype_realmatrix)reg_x;

                        Phloat[] d = rm.array.data;

                        bufptr = vartype2string(reg_x, buf, 22);
                        draw_string(0, 0, buf, bufptr);
                        draw_string(0, 1, "1:1=".ToCharArray(), 4);

                        if (rm.array.is_string[0] != 0)
                        {
                            draw_char(4, 1, '"');
                            draw_string(5, 1, Phloat.phloat_text_get(d[0]), Phloat.phloat_length_get(d[0]));
                            draw_char(5 + Phloat.phloat_length_get(d[0]), 1, '"');
                        }
                        else
                        {
                            bufptr = Phloat.phloat2string(d[0], buf, 18,
                                       0, 0, 3,
                                       flags.thousands_separators);

                            draw_string(4, 1, buf, bufptr);
                        }
                        break;

                    case TYPE.COMPLEXMATRIX:
                        vartype_complexmatrix cm = (vartype_complexmatrix)reg_x;
                        vartype_complex _c = new vartype_complex();

                        bufptr = vartype2string(reg_x, buf, 22);
                        draw_string(0, 0, buf, bufptr);
                        draw_string(0, 1, "1:1=".ToCharArray(), 4);

                        _c.type = TYPE.COMPLEX;
                        _c.re = cm.array.data[0];
                        _c.im = cm.array.data[1];

                        bufptr = vartype2string((vartype)_c, buf, 18);
                        draw_string(4, 1, buf, bufptr);
                        break;
                }
            }

            flush_display();
        }

        public static void clear_custom_menu()
        {
            int row, key;

            for (row = 0; row < 3; row++)
                for (key = 0; key < 6; key++)
                    custommenu_length[row, key] = 0;
        }

        public static void clear_prgm_menu()
        {
            int i;

            for (i = 0; i < 9; i++)
                progmenu_arg[i].type = ARGTYPE_NONE;

            for (i = 0; i < 6; i++)
                progmenu_length[i] = 0;
        }

        public static void fly_goose()
        {
            uint goosetime = shell_milliseconds();

            // No goose movements if the most recent one was less than 100 ms
            // ago; in other words, maximum goose speed is 10 positions/second
            //
            if (goosetime - 100 < lastgoosetime)
                return;

            lastgoosetime = goosetime;

            if (mode_goose < 0)
            {
                clear_row(0);
                mode_goose = (-mode_goose) % 22;
                draw_char(mode_goose, 0, (char)6);
            }
            else
            {
                draw_char(mode_goose, 0, ' ');
                mode_goose = (mode_goose + 1) % 22;
                draw_char(mode_goose, 0, (char)6);
            }

            flush_display();
        }

        public static void draw_pattern(Phloat dx, Phloat dy, char[] pattern, int pattern_width)
        {
            int x, y, h, v, hmin, hmax, vmin, vmax;

            if (dx + pattern_width < 1 || dx > 131 || dy + 8 < 1 || dy > 16)
                return;

            x = dx < 0 ? Phloat.to_int(-Phloat.floor(-dx + 0.5)) : Phloat.to_int(Phloat.floor(dx + 0.5));
            y = dy < 0 ? Phloat.to_int(-Phloat.floor(-dy + 0.5)) : Phloat.to_int(Phloat.floor(dy + 0.5));

            hmin = x < 1 ? 1 - x : 0;
            hmax = x + pattern_width > 132 ? 132 - x : pattern_width;
            vmin = y < 1 ? 1 - y : 0;
            vmax = y + 8 > 17 ? 17 - y : 8;

            x--;
            y--;
            if (flags.agraph_control1 != 0)
            {
                if (flags.agraph_control0 != 0)
                    // dst = dst ^ src
                    for (h = hmin; h < hmax; h++)
                    {
                        char c = (char)(pattern[h] >> vmin);

                        for (v = vmin; v < vmax; v++)
                        {
                            if ((c & 1) != 0)
                            {
                                int X = h + x;
                                int Y = v + y;
                                display[Y * 17 + (X >> 3)] ^= (byte)(1 << (X & 7));
                            }

                            c >>= 1;
                        }
                    }
                else
                    // dst = dst & ~src
                    for (h = hmin; h < hmax; h++)
                    {
                        char c = (char)(pattern[h] >> vmin);

                        for (v = vmin; v < vmax; v++)
                        {
                            if ((c & 1) != 0)
                            {
                                int X = h + x;
                                int Y = v + y;
                                display[Y * 17 + (X >> 3)] &= (byte)(~(1 << (X & 7)));
                            }

                            c >>= 1;
                        }
                    }
            }
            else
            {
                if (flags.agraph_control0 != 0)
                    // dst = src
                    for (h = hmin; h < hmax; h++)
                    {
                        char c = (char)(pattern[h] >> vmin);

                        for (v = vmin; v < vmax; v++)
                        {
                            int X = h + x;
                            int Y = v + y;

                            if ((c & 1) != 0)
                                display[Y * 17 + (X >> 3)] |= (byte)(1 << (X & 7));
                            else
                                display[Y * 17 + (X >> 3)] &= (byte)(~(1 << (X & 7)));

                            c >>= 1;
                        }
                    }
                else
                    // dst = dst | src
                    for (h = hmin; h < hmax; h++)
                    {
                        char c = (char)(pattern[h] >> vmin);

                        for (v = vmin; v < vmax; v++)
                        {
                            if ((c & 1) != 0)
                            {
                                int X = h + x;
                                int Y = v + y;
                                display[Y * 17 + (X >> 3)] |= (byte)(1 << (X & 7));
                            }

                            c >>= 1;
                        }
                    }
            }

            mark_dirty(y + vmin, x + hmin, y + vmax, x + hmax);
        }

        public static void print_display()
        {
            shell_print("<lcd>".ToCharArray(), 5, display, 17, 0, 0, 131, 16);
        }

        public static void assign_custom_key(int keynum, char[] name, int length)
        {
            int row = (keynum - 1) / 6;
            int key = (keynum - 1) % 6;
            int i;

            custommenu_length[row, key] = length;
            custommenu_label[row, key] = new string(name, 0, length);
        }

        public static void assign_prgm_key(int keynum, int is_gto, arg_struct arg)
        {
            int length;
            int i;

            keynum--;
            progmenu_arg[keynum] = (arg_struct)arg;
            progmenu_is_gto[keynum] = is_gto;
            length = reg_alpha_length;

            if (keynum < 6)
            {
                if (length > 7)
                    length = 7;

                progmenu_label[keynum] = new string(reg_alpha);
                progmenu_length[keynum] = length;
            }
        }

        public static void print_program_line(int prgm_index, int pc)
        {
            print_program(prgm_index, pc, 1, 1);
        }

        public static int print_program(int prgm_index, int pc, int lines, int normal)
        {
            prp_data_struct dat = new prp_data_struct();

            if (dat == null)
                return ERR_INSUFFICIENT_MEMORY;

            shell_annunciators(-1, -1, 1, -1, -1, -1);
            
            dat.len = 0;
            dat.saved_prgm = current_prgm;
            dat.cmd = CMD_NONE;
            dat.line = pc2line(pc);
            dat.pc = pc;
            dat.lines = lines;
            dat.width = flags.double_wide_print != 0 ? 12 : 24;
            dat.first = 1;
            
            if (normal != 0)
            {
                dat.trace = 0;
                dat.normal = 1;
            }
            else
            {
                dat.trace = flags.trace_print;
                dat.normal = flags.normal_print;
            }

            current_prgm = prgm_index;
            prp_data = dat;

            if (normal != 0)
            {
                // Printing just one line for NORM and TRACE mode;
                // we don't do the 'interruptible' thing in this case.
                //
                int err;
                while ((err = print_program_worker(0)) == ERR_INTERRUPTIBLE) ;
                return err;
            }
            else
            {
                print_text(null, 0, 1);
                mode_interruptible = print_program_worker;
                mode_stoppable = true;
                return ERR_INTERRUPTIBLE;
            }
        }

        static int print_program_worker(int interrupted)
        {
            prp_data_struct dat = prp_data;
            int printed = 0;

            if (interrupted != 0)
                goto done;

            do
            {
                if (dat.line == 0)
                    dat.pc = 0;
                else
                    get_next_command(ref dat.pc, ref dat.cmd, dat.arg, 0);

                if (dat.trace != 0)
                {
                    if (dat.cmd == CMD_LBL || dat.first != 0)
                    {
                        if (dat.len > 0)
                        {
                            print_lines(dat.buf, dat.len, 1);
                            printed = 1;
                        }

                        if (dat.first == 0)
                            print_text(null, 0, 1);

                        dat.first = 0;
                        dat.buf[0] = ' ';

                        char[] tmp = new char[99];
                        dat.len = 1 + prgmline2buf(tmp, 100 - 1, dat.line, dat.cmd == CMD_LBL, dat.cmd, dat.arg);
                        Array.Copy(tmp, 0, dat.buf, 1, tmp.Length);
                        
                        if (dat.cmd == CMD_LBL || dat.cmd == CMD_END || dat.lines == 1)
                        {
                            print_lines(dat.buf, dat.len, 1);
                            printed = 1;
                            dat.len = 0;
                        }
                    }
                    else
                    {
                        int i, len2;

                        if (dat.len > 0)
                        {
                            dat.buf[dat.len++] = ' ';
                            dat.buf[dat.len++] = ' ';
                        }

                        char[] tmp = new char[100 - dat.len];
                        len2 = prgmline2buf(tmp, 100 - dat.len, -1, false, dat.cmd, dat.arg);
                        Array.Copy(tmp, 0, dat.buf, dat.len, tmp.Length);

                        if (dat.len > 0 && dat.len + len2 > dat.width)
                        {
                            // Break line before current instruction
                            print_lines(dat.buf, dat.len - 2, 1);
                            printed = 1;

                            for (i = 0; i < len2; i++)
                                dat.buf[i] = dat.buf[dat.len + i];
                            
                            dat.len = len2;
                        }
                        else
                            dat.len += len2;

                        if (dat.lines == 1 || dat.cmd == CMD_END)
                        {
                            print_lines(dat.buf, dat.len, 1);
                            printed = 1;
                        }
                        else if (dat.len >= dat.width)
                        {
                            len2 = (dat.len / dat.width) * dat.width;
                            print_lines(dat.buf, len2, 1);
                            printed = 1;
                        
                            for (i = len2; i < dat.len; i++)
                                dat.buf[i - len2] = dat.buf[i];
                            
                            dat.len -= len2;
                        }
                    }
                }
                else
                {
                    dat.len = prgmline2buf(dat.buf, 100, dat.line, dat.cmd == CMD_LBL, dat.cmd, dat.arg);

                    if (dat.normal != 0)
                    {
                        // In normal mode, programs are printed right-justified;
                        // we pad the instuctions to a minimum of 8 characters so
                        // the listing won't look too ragged.
                        // First, find the beginning of the instruction -- it starts
                        // right after the first space or 'goose' (6) character.
                        //
                        int p = 0;
                        
                        while (dat.buf[p] != ' ' && dat.buf[p] != 6)
                            p++;
                        
                        while (dat.len < p + 9)
                            dat.buf[dat.len++] = ' ';
                        
                        // Insert blank line above LBLs
                        if (dat.cmd == CMD_LBL && dat.first == 0)
                            print_text(null, 0, 1);

                        dat.first = 0;
                    }

                    print_lines(dat.buf, dat.len, dat.normal == 0 ? 1 : 0);
                    printed = 1;
                }

                dat.line++;
                dat.lines--;

            } while (printed == 0);

            if (dat.lines != 0 && dat.cmd != CMD_END)
                return ERR_INTERRUPTIBLE;

        done:
            current_prgm = dat.saved_prgm;
            dat = null;
            
            shell_annunciators(-1, -1, 0, -1, -1, -1);
            
            return ERR_STOP;
        }
        
        public static void display_blitter(byte[] bits, int bytesperline, int x, int y, int width, int height)
        {
            for (int v = y; v < y + height; v++)
                for (int h = x; h < x + width; h++)
                {
                    bool pixel = (bits[v * bytesperline + (h >> 3)] & (1 << (h & 7))) != 0;

                    if (pixel)
                        Disp.SetPixel(h, v, 1);
                    else
                        Disp.SetPixel(h, v, 0);
                }
        }
    }
}
