﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;


namespace WP7_42s
{
    public partial class App
    {
        public static void recall_result(vartype v)
        {
            if (flags.stack_lift_disable != 0)
                App.free_vartype(ref reg_x);
            else
            {
                App.free_vartype(ref reg_t);
                reg_t = reg_z;
                reg_z = reg_y;
                reg_y = reg_x;
            }

            reg_x = v;

            if ((flags.trace_print != 0) && (flags.printer_exists != 0))
                docmd_prx(null);
        }

        public static void char2buf(char[] buf, int buflen, ref int bufptr, char c)
        {
            if (bufptr < buflen)
	            buf[bufptr++] = c;
            else
	            buf[buflen - 1] = (char) 26;
        }

        public static void string2buf(char[] buf, int buflen, ref int bufptr, char[] s, int slen)
        {
            for (int i = 0; i < slen; i++)
	            char2buf(buf, buflen, ref bufptr, s[i]);
        }

        public static int get_base()
        {
            int _base = 0;

            if (App.flags.base_bit0 != 0) _base += 1;
            if (App.flags.base_bit1 != 0) _base += 2;
            if (App.flags.base_bit2 != 0) _base += 4;
            if (App.flags.base_bit3 != 0) _base += 8;

            if (_base == 1 || _base == 7 || _base == 15)
	            return _base + 1;
            else
	            return 10;
        }

        public static int uint2string(uint n, char[] buf, int offset, int buflen)
        {
            uint pt = 1;
            int count = offset;

            while (n / pt >= 10)
                pt *= 10;

            while (pt != 0)
            {
                char2buf(buf, buflen, ref count, (char)('0' + (n / pt) % 10));
                pt /= 10;
            }
            return count - offset;
        }

        public static int int2string(int n, char[] buf, int offset, int buflen)
        {
            uint u;
            int count = offset;

            if (n < 0)
            {
                char2buf(buf, buflen, ref count, '-');
                u = (uint) -n;
            }
            else
                u = (uint) n;

            return count - offset + uint2string(u, buf, count, buflen);
        }

        public static Phloat rad_to_angle(Phloat x)
        {
            if (App.flags.rad != 0)
                return x;
            else if (App.flags.grad != 0)
                return x * (new Phloat(200) / Phloat.PI);
            else
                return x * (new Phloat(180) / Phloat.PI);
        }

        public static void generic_r2p(Phloat re, Phloat im, ref Phloat r, ref Phloat phi)
        {
            if (im == new Phloat(0))
            {
                if (re >= new Phloat(0))
                {
                    r = re;
                    phi = new Phloat(0);
                }
                else
                {
                    r = -re;
                    phi = App.flags.grad != 0 ? new Phloat(200) : App.flags.rad != 0 ? Phloat.PI : new Phloat(180);
                }
            }
            else if (re == new Phloat(0))
            {
                if (im > new Phloat(0))
                {
                    r = im;
                    phi = App.flags.grad != 0 ? new Phloat(100) : App.flags.rad != 0 ? Phloat.PI / new Phloat(2) : new Phloat(90);
                }
                else
                {
                    r = -im;
                    phi = App.flags.grad != 0 ? new Phloat(-100) : App.flags.rad != 0 ? -Phloat.PI / new Phloat(2) : new Phloat(-90);
                }
            }
            else
            {
                r = Phloat.hypot(re, im);
                phi = rad_to_angle(Phloat.atan2(im, re));
            }
        }

        public static void generic_p2r(Phloat r, Phloat phi, ref Phloat re, ref Phloat im)
        {
            Phloat tre = new Phloat();
            Phloat tim = new Phloat();

            if (flags.rad != 0)
            {
                Phloat.sincos(phi, ref tim, ref tre);
            }
            else if (flags.grad != 0)
            {
                phi = Phloat.fmod(phi, 400);
                
                if (phi < 0)
                    phi += 400;
                
                if (phi == 0)
                {
                    tre = 1;
                    tim = 0;
                }
                else if (phi == 100)
                {
                    tre = 0;
                    tim = 1;
                }
                else if (phi == 200)
                {
                    tre = -1;
                    tim = 0;
                }
                else if (phi == 300)
                {
                    tre = 0;
                    tim = -1;
                }
                else
                    Phloat.sincos(phi / (200 / PI), ref tim, ref tre);
            }
            else
            {
                phi = Phloat.fmod(phi, 360);
                
                if (phi < 0)
                    phi += 360;
                
                if (phi == 0)
                {
                    tre = 1;
                    tim = 0;
                }
                else if (phi == 90)
                {
                    tre = 0;
                    tim = 1;
                }
                else if (phi == 180)
                {
                    tre = -1;
                    tim = 0;
                }
                else if (phi == 270)
                {
                    tre = 0;
                    tim = -1;
                }
                else
                    Phloat.sincos(phi / (180 / PI), ref tim, ref tre);
            }

            re = r * tre;
            im = r * tim;
        }

        public static int vartype2string(App.vartype v, char[] buf, int offset, int buflen)
        {
            char[] _buf = new char[buflen - offset];
            int added = vartype2string(v, _buf, buflen - offset);
            _buf.CopyTo(buf, offset);

            return added;
        }

        public static int vartype2string(App.vartype v, char[] buf, int buflen)
        {
            int dispmode;
            int digits = 0;

            if (App.flags.fix_or_all != 0)
	            dispmode = App.flags.eng_or_all != 0 ? 3 : 0;
            else
	            dispmode = App.flags.eng_or_all != 0 ? 2 : 1;

            if (App.flags.digits_bit3 != 0)
	            digits += 8;

            if (App.flags.digits_bit2 != 0)
	            digits += 4;

            if (App.flags.digits_bit1 != 0)
	            digits += 2;

            if (App.flags.digits_bit0 != 0)
	            digits += 1;

            switch (v.type)
            {
	            case App.TYPE.REAL:
                    return Phloat.phloat2string(((App.vartype_real) v).x, buf, buflen, 1, digits, dispmode, App.flags.thousands_separators);

	            case App.TYPE.COMPLEX:
                {
	                Phloat x = new Phloat();
                    Phloat y = new Phloat();

	                char[] x_buf = new char[22];
	                int x_len;

	                char[] y_buf = new char[22];
	                int y_len;

	                int i;
                    int ret_len;

	                if (App.flags.polar != 0)
                    {
		                generic_r2p(((App.vartype_complex) v).re, ((App.vartype_complex) v).im, ref x, ref y);

		                if (Phloat.p_isinf(x) != 0)
		                    x = Phloat.POS_HUGE_PHLOAT;
	                }
                    else
                    {
		                x = ((App.vartype_complex) v).re;
		                y = ((App.vartype_complex) v).im;
	                }

	                x_len = Phloat.phloat2string(x, x_buf, 22, 0, digits, dispmode, App.flags.thousands_separators);
	                y_len = Phloat.phloat2string(y, y_buf, 22, 0, digits, dispmode, App.flags.thousands_separators);

	                if (x_len + y_len + 2 > buflen)
                    {
		                /* Too long? Fall back on ENG 2 */
		                x_len = Phloat.phloat2string(x, x_buf, 22, 0, 2, 2, App.flags.thousands_separators);
		                y_len = Phloat.phloat2string(y, y_buf, 22, 0, 2, 2, App.flags.thousands_separators);
    	            }

	                for (i = 0; i < buflen; i++)
                    {
                        if (i < x_len)
                            buf[i] = x_buf[i];
                        else if (i < x_len + 1)
                            buf[i] = ' ';
                        else if (i < x_len + 2)
                        {
                            if (y_buf[0] == '-' && App.flags.polar == 0)
                                buf[i] = '-';
                            else
                                buf[i] = App.flags.polar != 0 ? (char)23 : 'i';
                        }
                        else if (i < x_len + 3)
                        {
                            if (y_buf[0] == '-' && App.flags.polar == 0)
                                buf[i] = App.flags.polar != 0 ? (char)23 : 'i';
                            else
                                buf[i] = y_buf[0];
                        }
                        else
                            if (i - x_len - 2 < 22)                 // SLR - Added bounds check on y_buf
                                buf[i] = y_buf[i - x_len - 2];
	                }

	                ret_len = x_len + y_len + 2;

	                if (ret_len > buflen)
                    {
		                buf[buflen - 1] = (char) 26;
		                ret_len = buflen;
	                }

	                return ret_len;
	            }

	        case App.TYPE.REALMATRIX:
                {
                    App.vartype_realmatrix m = (App.vartype_realmatrix) v;

	                int i;
	                int chars_so_far = 0;

	                string2buf(buf, buflen, ref chars_so_far, "[ ".ToCharArray(), 2);

	                i = int2string(m.rows, buf, chars_so_far, buflen);
	                chars_so_far += i;

	                char2buf(buf, buflen, ref chars_so_far, 'x');
	                
                    i = int2string(m.columns, buf, chars_so_far, buflen);
	                chars_so_far += i;

	                string2buf(buf, buflen, ref chars_so_far, " Matrix ]".ToCharArray(), 9);

	                return chars_so_far;
	            }

	        case App.TYPE.COMPLEXMATRIX:
                {
	                App.vartype_complexmatrix m = (App.vartype_complexmatrix) v;

	                int i;
	                int chars_so_far = 0;

	                string2buf(buf, buflen, ref chars_so_far, "[ ".ToCharArray(), 2);

	                i = int2string(m.rows, buf, chars_so_far, buflen);
	                chars_so_far += i;

	                char2buf(buf, buflen, ref chars_so_far, 'x');
	                
                    i = int2string(m.columns, buf, chars_so_far, buflen);
	                chars_so_far += i;
	                
                    string2buf(buf, buflen, ref chars_so_far, " Cpx Matrix ]".ToCharArray(), 13);
	                
                    return chars_so_far;
	            }

	        case App.TYPE.STRING:
                {
	                App.vartype_string s = (App.vartype_string) v;

	                int i;
	                int chars_so_far = 0;

	                char2buf(buf, buflen, ref chars_so_far, '"');
	                
                    for (i = 0; i < s.length; i++)
		                char2buf(buf, buflen, ref chars_so_far, s.text[i]);

	                char2buf(buf, buflen, ref chars_so_far, '"');
	                
                    return chars_so_far;
	            }

	        default:
                {
	                char[] msg = "UnsuppVarType".ToCharArray();
	                int msglen = 13;
	                
	                for (int i = 0; i < msglen; i++)
    		            buf[i] = msg[i];

	                return msglen;
	            }
            }
        }

        public static int string_equals(char[] s1, int s1len, char[] s2, int s2len)
        {
            if (s1len != s2len)
    	        return 0;

            for (int i = 0; i < s1len; i++)
	            if (s1[i] != s2[i])
	                return 0;

            return 1;
        }

        public static char[] phloat2program(Phloat d)
        {
            // Converts a phloat to its most compact representation;
            // used for generating HP-42S style number literals in programs.
            //
            char[] allbuf = new char[25];
            char[] scibuf = new char[25];
            int alllen;
            int scilen;

            char dot = (flags.decimal_point != 0) ? '.' : ',';

            int _decimal;
            int zeroes = 0;
            int last_nonzero = 0;
            int exponent;
            int i;

            alllen = Phloat.phloat2string(d, allbuf, 24, 0, 0, 3, 0);
            scilen = Phloat.phloat2string(d, scibuf, 24, 0, 11, 1, 0);
            
            // Shorten SCI representation by removing trailing zeroes,
            // and decreasing the exponent until the decimal point
            // shifts out of the mantissa.
            //
            
            _decimal = -1;
            exponent = -1;
            
            for (i = 0; i < scilen; i++)
            {
	            char c = scibuf[i];

    	        if (c == dot)
                {
	                _decimal = i;
	                last_nonzero = i;
	                zeroes = 0;
	            }
                else if (c == 24)
                {
	                exponent = i;
	                break;
	            }
                else if (c != '0')
                {
	                last_nonzero = i;
	                zeroes = 0;
	            }
                else
	                zeroes++;
            }

            if (_decimal != -1)
            {
	            if (zeroes > 0)
                {
	                for (i = last_nonzero + 1; i < scilen - zeroes; i++)
		                scibuf[i] = scibuf[i + zeroes];

	                scilen -= zeroes;
	            }

	            if ((exponent == -1 && _decimal == scilen - 1) || (exponent != -1 && exponent - _decimal == 1))
                {
	                for (i = _decimal; i < scilen - 1; i++)
		                scibuf[i] = scibuf[i + 1];

    	            scilen--;
	            }
                else if (exponent != -1)
                {
	                int offset, ex, neg, newexplen, t;

	                exponent -= zeroes;
	                offset = exponent - _decimal - 1;
	                ex = 0;
	                neg = 0;

	                for (i = exponent + 1; i < scilen; i++)
                    {
		                char c = scibuf[i];

		                if (c == '-')
		                    neg = 1;
		                else
		                    ex = ex * 10 + c - '0';
	                }

    	            if (neg != 0)
        		        ex = -ex;

    	            ex -= offset;

    	            if (ex < 0)
                    {
		                ex = -ex;
		                neg = 1;
	                }
                    else
		                neg = 0;

	                newexplen = (neg != 0) ? 2 : 1;
	                t = 10;

	                while (ex >= t)
                    {
		                newexplen++;
		                t *= 10;
	                }

	                if (newexplen <= scilen - exponent - 1)
                    {
		                for (i = _decimal; i < exponent; i++)
		                    scibuf[i] = scibuf[i + 1];

		                scilen = exponent;
		            
                        if (neg != 0)
		                    ex = -ex;

		                scilen += int2string(ex, scibuf, exponent, 50);     // TODO: is 50 really correct?
	                }
	            }
            }

            if (scilen < alllen)
            {
	            scibuf[scilen] = (char)0;
	            return scibuf;
            }
            else
            {
	            allbuf[alllen] = (char)0;
	            return allbuf;
            }
        }

        public static void string_copy(char[] dst, ref int dstlen, char[] src, int srclen)
        {
            dstlen = srclen;

            for (int i = 0; i < srclen; i++)
    	        dst[i] = src[i];
        }

        public static void print_right(char[] left, int leftlen, char[] right, int rightlen)
        {
            char[] buf = new char[100];
            int len = 0;
            int width = (flags.double_wide_print != 0) ? 12 : 24;
            int i, pad;

            string_copy(buf, ref len, left, leftlen);
            
            if (len + rightlen + 1 <= width)
            {
	            buf[len++] = ' ';
	            pad = width - len - rightlen;

	            if (pad > 6 - rightlen)
	                pad = 6 - rightlen;
	        
                if (pad < 0)
	                pad = 0;

	            for (i = 0; i < pad; i++)
	                buf[len++] = ' ';
	        
                for (i = 0; i < rightlen; i++)
	                buf[len++] = right[i];

	            print_text(buf, len, 0);
            }
            else
            {
	            int line_start = 0;
                char[] tmp;

	            while (line_start + width < len)
                {
                    tmp = new char[width];
                    Array.Copy(buf, line_start, tmp, 0, width);
                    print_text(tmp, width, 1);

	                line_start += width;
	            }

	            pad = width - (len - line_start) - rightlen;

	            if (pad < 1)
	                pad = 1;

	            for (i = 0; i < pad; i++)
	                buf[len++] = ' ';

	            for (i = 0; i < rightlen; i++)
	                buf[len++] = right[i];

                tmp = new char[len - line_start];
                Array.Copy(buf, line_start, tmp, 0, len - line_start);
	            print_lines(tmp, len - line_start, 1);
            }
        }

        public static void print_text(char[] text, int length, int left_justified)
        {
            // Zero fractional amounts in trial mode
            text = munge(text);

            // TO DO: check preferences so we don't waste any time generating
            // print-outs that aren't going to be accepted by the shell anyway
            //
            char[] buf = new char[24];
            int width = (flags.double_wide_print != 0) ? 12 : 24;
            int bufptr = 0, i;
            char[] bitmap = new char[162];

            // Handle left/right justification and lowercase printing
            if (length > width)
	            length = width;

            if (left_justified == 0)
	            for (i = length; i < width; i++)
	                buf[bufptr++] = ' ';

            for (i = 0; i < length; i++)
            {
	            char c = text[i];

	            if ((flags.lowercase_print != 0) && (c & 127) >= 'A' && (c & 127) <= 'Z')
	                c += (char)32;
	            else if (c == 10)
	                c = (char) 138;

    	        buf[bufptr++] = c;
            }

            // Do bit-mapped printing
            for (i = 0; i < 162; i++)
	            bitmap[i] = (char)0;

            for (i = 0; i < bufptr; i++)
            {
	            char[] charbits = new char[5];

                get_char(charbits, buf[i]);
	            
                for (int j = 0; j < 5; j++)
                {
	                int x1 = i * 6 + j;
	                int x2 = x1 + 1;
	                int x, y;
	                if (flags.double_wide_print != 0)
                    {
		                x1 <<= 1;
		                x2 <<= 1;
	                }

	                for (x = x1; x < x2; x++)
		                for (y = 0; y < 8; y++)
		                    if ((charbits[j] & (1 << y)) != 0)
			                    bitmap[y * 18 + (x >> 3)] |= (char)(1 << (x & 7));
	            }
            }

            // Handle text-mode double-width printing by inserting spaces and
            // underscores; do text-mode printing
            if (flags.double_wide_print != 0)
            {
	            for (i = bufptr - 1; i >= 0; i--)
                {
	                char c = buf[i];
	                buf[2 * i] = c;
	                buf[2 * i + 1] = c == ' ' ? ' ' : '_';
	            }

    	        bufptr *= 2;
            }

            shell_print(buf, bufptr, bitmap, 18, 0, 0, 143, 9);
        }

        public static void print_text(string text, int left_justified)
        {
            print_text(text.ToCharArray(), text.Length, left_justified);
        }

        public static void print_lines(char[] text, int length, int left_justified)
        {
            int line_start = 0;
            int width = (flags.double_wide_print != 0) ? 12 : 24;
            char[] tmp;

            while (line_start + width < length)
            {
                tmp = new char[width];
                Array.Copy(text, line_start, tmp, 0, width);
                print_text(tmp, width, left_justified);

	            line_start += width;
            }

            tmp = new char[length - line_start];
            Array.Copy(text, line_start, tmp, 0, length - line_start);
            print_text(tmp, length - line_start, left_justified);
        }

        public static void print_wide(char[] left, int leftlen, char[] right, int rightlen)
        {
            int width = (flags.double_wide_print != 0) ? 12 : 24;

            char[] buf = new char[24];
            int bufptr = 0;
            int i;

            if (leftlen + rightlen <= width)
            {
	            for (i = 0; i < leftlen; i++)
	                buf[bufptr++] = left[i];

	            for (i = width - leftlen - rightlen; i > 0; i--)
	                buf[bufptr++] = ' ';

	            for (i = 0; i < rightlen; i++)
	                buf[bufptr++] = right[i];

	            print_text(buf, bufptr, 1);
            }
            else
            {
	            for (i = 0; i < leftlen; i++)
                {
	                buf[bufptr++] = left[i];

                    if (bufptr == width)
                    {
		                print_text(buf, width, 1);
		                bufptr = 0;
	                }
	            }
	        
                for (i = 0; i < rightlen; i++)
                {
	                buf[bufptr++] = right[i];
	                
                    if (bufptr == width)
                    {
		                print_text(buf, width, 1);
		                bufptr = 0;
	                }
	            }
	        
                if (bufptr > 0)
	                print_text(buf, bufptr, 1);
            }
        }

        public static void print_command(int cmd, arg_struct arg)
        {
            char[] buf = new char[100];
            int bufptr = 0;
    
            if (cmd == CMD_NULL && deferred_print == 0)
    	        return;

            shell_annunciators(-1, -1, 1, -1, -1, -1);

            if (cmd != CMD_NULL)
	            bufptr += command2buf(buf, 100, cmd, arg);

            if (deferred_print != 0)
            {
	            // If the display mode is FIX n, and the user has not entered
	            // an exponent, pad the fractional part to n, in order to get
	            // decimal points to line up.
	            //
	            if ((flags.fix_or_all != 0) && (flags.eng_or_all == 0))
                {
	                int i;
                    int n;
                    int dot_found = 0;
                    int ip_digits = 0;
                    int fp_digits = 0;

	                char dot = (flags.decimal_point != 0) ? '.' : ',';

	                for (i = 0; i < cmdline_length; i++)
                    {
		                char c = cmdline[i];

		                if (c == 24)
		                    goto never_mind;

		                if (c == dot)
		                    dot_found = 1;
		                
                        if (c >= '0' && c <= '9')
                        {
		                    if (dot_found != 0)
			                    fp_digits++;
		                    else
			                    ip_digits++;
		                }
	                }

	                n = 0;

	                if (flags.digits_bit3 != 0) n += 8;
	                if (flags.digits_bit2 != 0) n += 4;
	                if (flags.digits_bit1 != 0) n += 2;
	                if (flags.digits_bit0 != 0) n += 1;

	                if (dot_found == 0)
		                cmdline[cmdline_length++] = dot;

	                while (fp_digits < n && ip_digits + fp_digits < 12)
                    {
		                cmdline[cmdline_length++] = '0';
		                fp_digits++;
	                }

	            never_mind:;
	           }
	    
                print_right(cmdline, cmdline_length, buf, bufptr);
            }
            else
            {
	            // Normally we print commands right-justified, but if they don't fit
	            // on one line, we print them left-justified, because having the excess
	            // go near the right margin looks weird and confusing.
	            //
	            int left = (bufptr > (flags.double_wide_print != 0 ? 12 : 24)) ? 1 : 0;
	            print_lines(buf, bufptr, left);
            }

            deferred_print = 0;
            shell_annunciators(-1, -1, 0, -1, -1, -1);
        }

        public static void set_base(int _base)
        {
            int oldbase = 0;

            if (flags.base_bit0 != 0) oldbase += 1;
            if (flags.base_bit1 != 0) oldbase += 2;
            if (flags.base_bit2 != 0) oldbase += 4;
            if (flags.base_bit3 != 0) oldbase += 8;

            if (_base == 2 || _base == 8 || _base == 16)
	            _base--;
            else
	            _base = 0;

            flags.base_bit0 = (byte)((_base & 1) != 0 ? 1 : 0);
            flags.base_bit1 = (byte)((_base & 2) != 0 ? 1 : 0);
            flags.base_bit2 = (byte)((_base & 4) != 0 ? 1 : 0);
            flags.base_bit3 = (byte)((_base & 8) != 0 ? 1 : 0);

            if (mode_appmenu == MENU_BASE_A_THRU_F)
	            set_menu(MENULEVEL_APP, MENU_BASE);

            if (_base != oldbase && flags.trace_print != 0 && flags.printer_exists != 0)
	            docmd_prx(null);
        }

        public static void append_alpha_char(char c)
        {
            if (reg_alpha_length == 44)
            {
                for (int i = 0; i < 43; i++)
                    reg_alpha[i] = reg_alpha[i + 1];

                reg_alpha[43] = c;
            }
            else
                reg_alpha[reg_alpha_length++] = c;
        }

        public static int resolve_ind_arg(arg_struct arg)
        {
            vartype v = new vartype();

            switch (arg.type)
            {
                case ARGTYPE_IND_NUM:
                    {
                        vartype regs = recall_var("REGS".ToCharArray(), 4);

                        if (regs == null)
                            return ERR_SIZE_ERROR;
                        if (regs.type != TYPE.REALMATRIX)
                            return ERR_INVALID_TYPE;
                        else
                        {
                            vartype_realmatrix rm = (vartype_realmatrix)regs;
                            int size = rm.rows * rm.columns;
                            int num = arg.val.num;
                            
                            if (num >= size)
                                return ERR_SIZE_ERROR;
                            
                            if (rm.array.is_string[num] != 0)
                            {
                                Phloat d = rm.array.data[num];
                                
                                arg.type = ARGTYPE_STR;
                                arg.length = (byte)Phloat.phloat_length_get(d);

                                for (int i = 0; i < Phloat.phloat_length_get(d); i++)
                                    arg.val.text[i] = Phloat.phloat_text_get(d)[i];
                            }
                            else
                            {
                                Phloat x = rm.array.data[num];

                                if (x < 0)
                                    x = -x;
                                
                                if (x >= 2147483648.0)
                                    arg.val.num = 2147483647;
                                else
                                    arg.val.num = Phloat.to_int(x);
                                
                                arg.type = ARGTYPE_NUM;
                            }

                            return ERR_NONE;
                        }
                    }

                case ARGTYPE_IND_STK:
                    {
                        switch (arg.val.stk)
                        {
                            case 'X': v = reg_x; break;
                            case 'Y': v = reg_y; break;
                            case 'Z': v = reg_z; break;
                            case 'T': v = reg_t; break;
                            case 'L': v = reg_lastx; break;
                        }

//                        goto finish_resolve;      // Replicate finish_resolve code below to avoid goto
                        if (v.type == TYPE.REAL)
                        {
                            Phloat x = ((vartype_real)v).x;

                            if (x < 0)
                                x = -x;

                            if (x >= 2147483648.0)
                                arg.val.num = 2147483647;
                            else
                                arg.val.num = Phloat.to_int(x);

                            arg.type = ARGTYPE_NUM;
                            return ERR_NONE;
                        }
                        else if (v.type == TYPE.STRING)
                        {
                            vartype_string s = (vartype_string)v;

                            arg.type = ARGTYPE_STR;
                            arg.length = (byte)s.length;

                            for (int i = 0; i < s.length; i++)
                                arg.val.text[i] = s.text[i];
                            return ERR_NONE;
                        }
                        else
                            return ERR_INVALID_TYPE;
                    }

                case ARGTYPE_IND_STR:
                    {
                        v = recall_var(arg.val.text, arg.length);
                        
                        if (v == null)
                            return ERR_NONEXISTENT;

//                    finish_resolve:
                        if (v.type == TYPE.REAL)
                        {
                            Phloat x = ((vartype_real)v).x;
                            
                            if (x < 0)
                                x = -x;
                            
                            if (x >= 2147483648.0)
                                arg.val.num = 2147483647;
                            else
                                arg.val.num = Phloat.to_int(x);
                            
                            arg.type = ARGTYPE_NUM;
                            return ERR_NONE;
                        }
                        else if (v.type == TYPE.STRING)
                        {
                            vartype_string s = (vartype_string)v;

                            arg.type = ARGTYPE_STR;
                            arg.length = (byte)s.length;
                            
                            for (int i = 0; i < s.length; i++)
                                arg.val.text[i] = s.text[i];
                            return ERR_NONE;
                        }
                        else
                            return ERR_INVALID_TYPE;
                    }

                default:
                    // Should not happen; this function should only
                    // be called to resolve indirect arguments.
                    //
                    return ERR_INTERNAL_ERROR;
            }
        }

        public static void binary_result(vartype x)
        {
            free_vartype(ref reg_lastx);
            
            reg_lastx = reg_x;
            reg_x = x;
            
            free_vartype(ref reg_y);
            
            reg_y = reg_z;
            reg_z = dup_vartype(reg_t);
            
            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
        }

        public static void unary_result(vartype x)
        {
            free_vartype(ref reg_lastx);

            reg_lastx = reg_x;
            reg_x = x;
            
            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
        }

        public static int arg_to_num(arg_struct arg, ref int num)
        {
            if (arg.type == ARGTYPE_IND_NUM
                || arg.type == ARGTYPE_IND_STK
                || arg.type == ARGTYPE_IND_STR)
            {
                int err = resolve_ind_arg(arg);
                
                if (err != ERR_NONE)
                    return err;
            }

            if (arg.type == ARGTYPE_NUM)
            {
                num = arg.val.num;
                
                return ERR_NONE;
            }
            else
                return ERR_INVALID_TYPE;
        }

        public static void append_alpha_string(char[] buf, int buflen, int reverse)
        {
            int needed, i;
            int offset = 0;

            if (buflen > 44)
            {
                if (reverse == 0)
                    offset = buflen - 44;       // TODO: used offset to emulate pointer increment - verify
	                //buf += buflen - 44;
	            
                buflen = 44;
            }

            needed = reg_alpha_length + buflen - 44;
            
            if (needed > 0)
            {
	            for (i = 0; i < reg_alpha_length - needed; i++)
	                reg_alpha[i] = reg_alpha[i + needed];
	            
                reg_alpha_length -= needed;
            }

            if (reverse != 0)
	            for (i = buflen - 1; i >= 0; i--)
	                reg_alpha[reg_alpha_length++] = buf[i + offset];
            else
	            for (i = 0; i < buflen; i++)
	                reg_alpha[reg_alpha_length++] = buf[i + offset];
        }

        public static int dimension_array(char[] name, int namelen, int rows, int columns)
        {
            vartype matrix = recall_var(name, namelen);
            
            // NOTE: 'size' will only ever be 0 when we're called from
            // docmd_size(); docmd_dim() does not allow 0-size matrices.
            //
            int size = rows * columns;
            
            if (matrix == null || (matrix.type != TYPE.REALMATRIX && matrix.type != TYPE.COMPLEXMATRIX))
            {
	            vartype newmatrix;
	            
                if (size == 0)
	                return ERR_NONE;
	            
                newmatrix = new_realmatrix(rows, columns);
	            
                if (newmatrix == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            
                store_var(name, namelen, newmatrix);
	            
                return ERR_NONE;
            }
            else if (size == 0)
            {
	            purge_var(name, namelen);
	            
                return ERR_NONE;
            }
            else
	            return dimension_array_ref(matrix, rows, columns);
        }

        public static int dimension_array_ref(vartype matrix, int rows, int columns)
        {
            int size = rows * columns;
            
            if (matrix.type == TYPE.REALMATRIX)
            {
                vartype_realmatrix oldmatrix = (vartype_realmatrix)matrix;
                
                if (oldmatrix.rows == rows && oldmatrix.columns == columns)
                    return ERR_NONE;
                
                if (oldmatrix.array.refcount == 1)
                {
                    // Since there are no shared references to this array,
                    // I can modify it in place using a realloc(). However, I
                    // only use realloc() on the 'data' array, not on the
                    // 'is_string' array -- if I used it on both, and the second
                    // call fails, I might be unable to roll back the first.
                    // So, playing safe -- shouldn't be too big a handicap since
                    // 'is_string' is a lot smaller than 'data', so the transient
                    // memory overhead is only about 12.5%.
                    //
                    char[] new_is_string = new char[size];

                    if (new_is_string == null)
                        return ERR_INSUFFICIENT_MEMORY;
                    
                    int i, s, oldsize;

                    oldsize = oldmatrix.rows * oldmatrix.columns;
                    s = oldsize < size ? oldsize : size;

                    Phloat[] new_data = new Phloat[size];
                    Array.Copy(oldmatrix.array.data, new_data, s);
                    
                    if (new_data == null)
                    {
                        new_is_string = null;
                        return ERR_INSUFFICIENT_MEMORY;
                    }

                    for (i = 0; i < s; i++)
                        new_is_string[i] = oldmatrix.array.is_string[i];
                    
                    for (i = s; i < size; i++)
                    {
                        new_is_string[i] = (char)0;
                        new_data[i] = 0;
                    }
                    
                    oldmatrix.array.is_string = new_is_string;
                    oldmatrix.array.data = new_data;
                    oldmatrix.rows = rows;
                    oldmatrix.columns = columns;

                    return ERR_NONE;
                }
                else
                {
                    // There are shared references to the matrix. This means I
                    // can't realloc() it; I'm going to allocate a brand-new instance,
                    // and copy the contents from the old instance. I don't use
                    // disentangle(); that's only useful if you want to eliminate
                    // shared references without resizing.
                    //
                    int i, s, oldsize;

                    realmatrix_data new_array = new realmatrix_data();
                    
                    if (new_array == null)
                        return ERR_INSUFFICIENT_MEMORY;
                    
                    new_array.data = new Phloat[size];
                    
                    if (new_array.data == null)
                    {
                        new_array = null;
                        return ERR_INSUFFICIENT_MEMORY;
                    }
                    
                    new_array.is_string = new char[size];
                    
                    if (new_array.is_string == null)
                    {
                        new_array.data = null;
                        new_array = null;
                        return ERR_INSUFFICIENT_MEMORY;
                    }
                    
                    oldsize = oldmatrix.rows * oldmatrix.columns;
                    s = oldsize < size ? oldsize : size;
                    
                    for (i = 0; i < s; i++)
                    {
                        new_array.is_string[i] = oldmatrix.array.is_string[i];
                        new_array.data[i] = oldmatrix.array.data[i];
                    }
                    
                    for (i = s; i < size; i++)
                    {
                        new_array.is_string[i] = (char)0;
                        new_array.data[i] = 0;
                    }
                    
                    new_array.refcount = 1;
                    oldmatrix.array.refcount--;
                    oldmatrix.array = new_array;
                    oldmatrix.rows = rows;
                    oldmatrix.columns = columns;
                    
                    return ERR_NONE;
                }
            }
            else // matrix.type == TYPE_COMPLEXMATRIX
            {
                vartype_complexmatrix oldmatrix = (vartype_complexmatrix)matrix;
                
                if (oldmatrix.rows == rows && oldmatrix.columns == columns)
                    return ERR_NONE;
                
                if (oldmatrix.array.refcount == 1)
                {
                    // Since there are no shared references to this array,
                    // I can modify it in place using a realloc().
                    //
                    int i, oldsize;

                    Phloat[] new_data = new Phloat[2 * size];
                    Array.Copy(oldmatrix.array.data, new_data, oldmatrix.array.data.Length);

                    if (new_data == null)
                        return ERR_INSUFFICIENT_MEMORY;
                    
                    oldsize = oldmatrix.rows * oldmatrix.columns;
                    
                    for (i = 2 * oldsize; i < 2 * size; i++)
                        new_data[i] = 0;
                    
                    oldmatrix.array.data = new_data;
                    oldmatrix.rows = rows;
                    oldmatrix.columns = columns;
                    
                    return ERR_NONE;
                }
                else
                {
                    // There are shared references to the matrix. This means I
                    // can't realloc() it; I'm going to allocate a brand-new instance,
                    // and copy the contents from the old instance. I don't use
                    // disentangle(); that's only useful if you want to eliminate
                    // shared references without resizing.
                    //
                    int i, s, oldsize;

                    complexmatrix_data new_array = new complexmatrix_data();
                    
                    if (new_array == null)
                        return ERR_INSUFFICIENT_MEMORY;

                    new_array.data = new Phloat[2 * size];
                    
                    if (new_array.data == null)
                    {
                        new_array = null;
                        return ERR_INSUFFICIENT_MEMORY;
                    }

                    oldsize = oldmatrix.rows * oldmatrix.columns;
                    s = oldsize < size ? oldsize : size;
                    
                    for (i = 0; i < 2 * s; i++)
                        new_array.data[i] = oldmatrix.array.data[i];
                    
                    for (i = 2 * s; i < 2 * size; i++)
                        new_array.data[i] = 0;
                    
                    new_array.refcount = 1;
                    oldmatrix.array.refcount--;
                    oldmatrix.array = new_array;
                    oldmatrix.rows = rows;
                    oldmatrix.columns = columns;
                    
                    return ERR_NONE;
                }
            }
        }

        public static Phloat rad_to_deg(Phloat x)
        {
            return x * (180 / PI);
        }

        public static Phloat deg_to_rad(Phloat x)
        {
            return x / (180 / PI);
        }

        public static Phloat fix_hms(Phloat x)
        {
            Phloat sec_corr = new Phloat(4, 1000);
            Phloat min_corr = new Phloat(4, 10);

            bool neg = x < 0;
            
            if (neg)
	            x = -x;
            
            if (x == x + 1)
	            return x;
            
            if (x < 0.0059)
	            return x;

            if (Phloat.floor(Phloat.fmod(x * 10000, 100)) == 60)
	            x += sec_corr;

            if (Phloat.floor(Phloat.fmod(x * 100, 100)) == 60)
            {
	            x += min_corr;

                if (Phloat.floor(Phloat.fmod(x * 10000, 100)) == 60)
		            x += sec_corr;
	        }

            return neg ? -x : x;
        }

        public static int is_pure_real(vartype matrix)
        {
            vartype_realmatrix rm = new vartype_realmatrix();

            int size;
            int i;

            if (matrix.type != TYPE.REALMATRIX)
	            return 0;

            rm = (vartype_realmatrix) matrix;
            size = rm.rows * rm.columns;

            for (i = 0; i < size; i++)
	            if (rm.array.is_string[i] != 0)
	                return 0;

            return 1;
        }

        public static void set_flag(int num, int val)
        {
            switch (num)
            {
                case 0: flags.f00 = (byte)val; return;
                case 1: flags.f01 = (byte)val; return;
                case 2: flags.f02 = (byte)val; return;
                case 3: flags.f03 = (byte)val; return;
                case 4: flags.f04 = (byte)val; return;
                case 5: flags.f05 = (byte)val; return;
                case 6: flags.f06 = (byte)val; return;
                case 7: flags.f07 = (byte)val; return;
                case 8: flags.f08 = (byte)val; return;
                case 9: flags.f09 = (byte)val; return;

                case 10: flags.f10 = (byte)val; return;
                case 11: flags.auto_exec = (byte)val; return;
                case 12: flags.double_wide_print = (byte)val; return;
                case 13: flags.lowercase_print = (byte)val; return;
                case 14: flags.f14 = (byte)val; return;
                case 15: flags.trace_print = (byte)val; return;
                case 16: flags.normal_print = (byte)val; return;
                case 17: flags.f17 = (byte)val; return;
                case 18: flags.f18 = (byte)val; return;
                case 19: flags.f19 = (byte)val; return;

                case 20: flags.f20 = (byte)val; return;
                case 21: flags.printer_enable = (byte)val; return;
                case 22: flags.numeric_data_input = (byte)val; return;
                case 23: flags.alpha_data_input = (byte)val; return;
                case 24: flags.range_error_ignore = (byte)val; return;
                case 25: flags.error_ignore = (byte)val; return;
                case 26: flags.audio_enable = (byte)val; return;
                case 27: flags.VIRTUAL_custom_menu = (byte)val; return;
                case 28: flags.decimal_point = (byte)val; return;
                case 29: flags.thousands_separators = (byte)val; return;

                case 30: flags.stack_lift_disable = (byte)val; return;
                case 31: flags.f31 = (byte)val; return;
                case 32: flags.f32 = (byte)val; return;
                case 33: flags.f33 = (byte)val; return;
                case 34: flags.agraph_control1 = (byte)val; return;
                case 35: flags.agraph_control0 = (byte)val; return;
                case 36: flags.digits_bit3 = (byte)val; return;
                case 37: flags.digits_bit2 = (byte)val; return;
                case 38: flags.digits_bit1 = (byte)val; return;
                case 39: flags.digits_bit0 = (byte)val; return;

                case 40: flags.fix_or_all = (byte)val; return;
                case 41: flags.eng_or_all = (byte)val; return;
                case 42: flags.grad = (byte)val; return;
                case 43: flags.rad = (byte)val; return;
                case 44: flags.continuous_on = (byte)val; return;
                case 45: flags.VIRTUAL_solving = (byte)val; return;
                case 46: flags.VIRTUAL_integrating = (byte)val; return;
                case 47: flags.VIRTUAL_variable_menu = (byte)val; return;
                case 48: flags.alpha_mode = (byte)val; return;
                case 49: flags.VIRTUAL_low_battery = (byte)val; return;

                case 50: flags.message = (byte)val; return;
                case 51: flags.two_line_message = (byte)val; return;
                case 52: flags.prgm_mode = (byte)val; return;
                case 53: flags.VIRTUAL_input = (byte)val; return;
                case 54: flags.f54 = (byte)val; return;
                case 55: flags.printer_exists = (byte)val; return;
                case 56: flags.lin_fit = (byte)val; return;
                case 57: flags.log_fit = (byte)val; return;
                case 58: flags.exp_fit = (byte)val; return;
                case 59: flags.pwr_fit = (byte)val; return;

                case 60: flags.all_sigma = (byte)val; return;
                case 61: flags.log_fit_invalid = (byte)val; return;
                case 62: flags.exp_fit_invalid = (byte)val; return;
                case 63: flags.pwr_fit_invalid = (byte)val; return;
                case 64: flags.f64 = (byte)val; return;
                case 65: flags.VIRTUAL_matrix_editor = (byte)val; return;
                case 66: flags.grow = (byte)val; return;
                case 67: flags.f67 = (byte)val; return;
                case 68: flags.base_bit0 = (byte)val; return;
                case 69: flags.base_bit1 = (byte)val; return;

                case 70: flags.base_bit2 = (byte)val; return;
                case 71: flags.base_bit3 = (byte)val; return;
                case 72: flags.local_label = (byte)val; return;
                case 73: flags.polar = (byte)val; return;
                case 74: flags.real_result_only = (byte)val; return;
                case 75: flags.VIRTUAL_programmable_menu = (byte)val; return;
                case 76: flags.matrix_edge_wrap = (byte)val; return;
                case 77: flags.matrix_end_wrap = (byte)val; return;
                case 78: flags.f78 = (byte)val; return;
                case 79: flags.f79 = (byte)val; return;

                case 80: flags.f80 = (byte)val; return;
                case 81: flags.f81 = (byte)val; return;
                case 82: flags.f82 = (byte)val; return;
                case 83: flags.f83 = (byte)val; return;
                case 84: flags.f84 = (byte)val; return;
                case 85: flags.f85 = (byte)val; return;
                case 86: flags.f86 = (byte)val; return;
                case 87: flags.f87 = (byte)val; return;
                case 88: flags.f88 = (byte)val; return;
                case 89: flags.f89 = (byte)val; return;

                case 90: flags.f90 = (byte)val; return;
                case 91: flags.f91 = (byte)val; return;
                case 92: flags.f92 = (byte)val; return;
                case 93: flags.f93 = (byte)val; return;
                case 94: flags.f94 = (byte)val; return;
                case 95: flags.f95 = (byte)val; return;
                case 96: flags.f96 = (byte)val; return;
                case 97: flags.f97 = (byte)val; return;
                case 98: flags.f98 = (byte)val; return;
                case 99: flags.f99 = (byte)val; return;
            }
        }

        public static int get_flag(int num)
        {
            switch (num)
            {
                case 0: return flags.f00;
                case 1: return flags.f01;
                case 2: return flags.f02;
                case 3: return flags.f03;
                case 4: return flags.f04;
                case 5: return flags.f05;
                case 6: return flags.f06;
                case 7: return flags.f07;
                case 8: return flags.f08;
                case 9: return flags.f09;

                case 10: return flags.f10;
                case 11: return flags.auto_exec;
                case 12: return flags.double_wide_print;
                case 13: return flags.lowercase_print;
                case 14: return flags.f14;
                case 15: return flags.trace_print;
                case 16: return flags.normal_print;
                case 17: return flags.f17;
                case 18: return flags.f18;
                case 19: return flags.f19;

                case 20: return flags.f20;
                case 21: return flags.printer_enable;
                case 22: return flags.numeric_data_input;
                case 23: return flags.alpha_data_input;
                case 24: return flags.range_error_ignore;
                case 25: return flags.error_ignore;
                case 26: return flags.audio_enable;
                case 27: return flags.VIRTUAL_custom_menu;
                case 28: return flags.decimal_point;
                case 29: return flags.thousands_separators;

                case 30: return flags.stack_lift_disable;
                case 31: return flags.f31;
                case 32: return flags.f32;
                case 33: return flags.f33;
                case 34: return flags.agraph_control1;
                case 35: return flags.agraph_control0;
                case 36: return flags.digits_bit3;
                case 37: return flags.digits_bit2;
                case 38: return flags.digits_bit1;
                case 39: return flags.digits_bit0;

                case 40: return flags.fix_or_all;
                case 41: return flags.eng_or_all;
                case 42: return flags.grad;
                case 43: return flags.rad;
                case 44: return flags.continuous_on;
                case 45: return flags.VIRTUAL_solving;
                case 46: return flags.VIRTUAL_integrating;
                case 47: return flags.VIRTUAL_variable_menu;
                case 48: return flags.alpha_mode;
                case 49: return flags.VIRTUAL_low_battery;

                case 50: return flags.message;
                case 51: return flags.two_line_message;
                case 52: return flags.prgm_mode;
                case 53: return flags.VIRTUAL_input;
                case 54: return flags.f54;
                case 55: return flags.printer_exists;
                case 56: return flags.lin_fit;
                case 57: return flags.log_fit;
                case 58: return flags.exp_fit;
                case 59: return flags.pwr_fit;

                case 60: return flags.all_sigma;
                case 61: return flags.log_fit_invalid;
                case 62: return flags.exp_fit_invalid;
                case 63: return flags.pwr_fit_invalid;
                case 64: return flags.f64;
                case 65: return flags.VIRTUAL_matrix_editor;
                case 66: return flags.grow;
                case 67: return flags.f67;
                case 68: return flags.base_bit0;
                case 69: return flags.base_bit1;

                case 70: return flags.base_bit2;
                case 71: return flags.base_bit3;
                case 72: return flags.local_label;
                case 73: return flags.polar;
                case 74: return flags.real_result_only;
                case 75: return flags.VIRTUAL_programmable_menu;
                case 76: return flags.matrix_edge_wrap;
                case 77: return flags.matrix_end_wrap;
                case 78: return flags.f78;
                case 79: return flags.f79;

                case 80: return flags.f80;
                case 81: return flags.f81;
                case 82: return flags.f82;
                case 83: return flags.f83;
                case 84: return flags.f84;
                case 85: return flags.f85;
                case 86: return flags.f86;
                case 87: return flags.f87;
                case 88: return flags.f88;
                case 89: return flags.f89;

                case 90: return flags.f90;
                case 91: return flags.f91;
                case 92: return flags.f92;
                case 93: return flags.f93;
                case 94: return flags.f94;
                case 95: return flags.f95;
                case 96: return flags.f96;
                case 97: return flags.f97;
                case 98: return flags.f98;
                case 99: return flags.f99;
                default: return 0;
            }
        }

        public static int virtual_flag_handler(int flagop, int flagnum)
        {
            // NOTE: the determination which flag numbers are handled by this
            // function is made by docmd_sf() etc.; they do this based on a constant
            // string 'virtual_flags' defined locally in core_commands.c.
            //
            switch (flagnum)
            {
                case 27: // custom_menu
                    {
                        bool its_on = mode_plainmenu == MENU_CUSTOM1 || mode_plainmenu == MENU_CUSTOM2 || mode_plainmenu == MENU_CUSTOM3;

                        switch (flagop)
                        {
                            case FLAGOP_SF:
                                if (!its_on)
                                    set_menu(MENULEVEL_PLAIN, MENU_CUSTOM1);
                                return ERR_NONE;
                            
                            case FLAGOP_CF:
                                if (its_on)
                                    set_menu(MENULEVEL_PLAIN, (int)MENU.NONE);
                                return ERR_NONE;
                            
                            case FLAGOP_FS_T:
                                return its_on ? ERR_YES : ERR_NO;
                            
                            case FLAGOP_FC_T:
                                return its_on ? ERR_NO : ERR_YES;
                            
                            case FLAGOP_FSC_T:
                                if (its_on)
                                    set_menu(MENULEVEL_PLAIN, (int)MENU.NONE);
                                return its_on ? ERR_YES : ERR_NO;
                            
                            case FLAGOP_FCC_T:
                                if (its_on)
                                    set_menu(MENULEVEL_PLAIN, (int)MENU.NONE);
                                return its_on ? ERR_NO : ERR_YES;
                            
                            default:
                                return ERR_INTERNAL_ERROR;
                        }
                    }

                case 45: // solving
                    {
                        switch (flagop)
                        {
                            case FLAGOP_FS_T:
                                return solve_active() ? ERR_YES : ERR_NO;
                            case FLAGOP_FC_T:
                                return solve_active() ? ERR_NO : ERR_YES;
                            default:
                                return ERR_INTERNAL_ERROR;
                        }
                    }

                case 46: // integrating
                    {
                        switch (flagop)
                        {
                            case FLAGOP_FS_T:
                                return integ_active() ? ERR_YES : ERR_NO;
                            case FLAGOP_FC_T:
                                return integ_active() ? ERR_NO : ERR_YES;
                            default:
                                return ERR_INTERNAL_ERROR;
                        }
                    }

                case 47: // variable_menu
                    {
                        bool its_on = mode_appmenu == MENU_VARMENU && varmenu_role == 0;

                        switch (flagop)
                        {
                            case FLAGOP_FS_T:
                                return its_on ? ERR_YES : ERR_NO;
                            case FLAGOP_FC_T:
                                return its_on ? ERR_NO : ERR_YES;
                            default:
                                return ERR_INTERNAL_ERROR;
                        }
                    }

                case 49: // low_battery
                    {
                        bool lowbat = shell_low_battery() != 0;

                        switch (flagop)
                        {
                            case FLAGOP_FS_T:
                                return lowbat ? ERR_YES : ERR_NO;
                            case FLAGOP_FC_T:
                                return lowbat ? ERR_NO : ERR_YES;
                            default:
                                return ERR_INTERNAL_ERROR;
                        }
                    }

                case 53: // input
                    {
                        switch (flagop)
                        {
                            case FLAGOP_FS_T:
                                return input_length > 0 ? ERR_YES : ERR_NO;
                            case FLAGOP_FC_T:
                                return input_length > 0 ? ERR_NO : ERR_YES;
                            default:
                                return ERR_INTERNAL_ERROR;
                        }
                    }

                case 65: // matrix_editor
                    {
                        bool its_on = matedit_mode == 2 || matedit_mode == 3;

                        switch (flagop)
                        {
                            case FLAGOP_FS_T:
                                return its_on ? ERR_YES : ERR_NO;
                            case FLAGOP_FC_T:
                                return its_on ? ERR_NO : ERR_YES;
                            default:
                                return ERR_INTERNAL_ERROR;
                        }
                    }

                case 75: // programmable_menu
                    {
                        bool its_on = (mode_plainmenu = MENU_PROGRAMMABLE) != 0;

                        switch (flagop)
                        {
                            case FLAGOP_FS_T:
                                return its_on ? ERR_YES : ERR_NO;
                            case FLAGOP_FC_T:
                                return its_on ? ERR_NO : ERR_YES;
                            default:
                                return ERR_INTERNAL_ERROR;
                        }
                    }

                default:
                    return ERR_INTERNAL_ERROR;
            }
        }

        public static void recall_two_results(ref vartype x, ref vartype y)
        {
            if (flags.stack_lift_disable != 0)
            {
                free_vartype(ref reg_t);
                free_vartype(ref reg_x);

                reg_t = reg_z;
                reg_z = reg_y;
            }
            else
            {
                free_vartype(ref reg_t);
                free_vartype(ref reg_z);
                
                reg_t = reg_y;
                reg_z = reg_x;
            }

            reg_y = y;
            reg_x = x;
            
            if (flags.trace_print != 0 && flags.printer_exists != 0)
                docmd_prx(null);
        }

        public static int ip2revstring(Phloat d, char[] buf, int buflen)
        {
            int s = 1;
            int bufpos = 0;

            if (d < 0)
            {
                d = -d;
                s = -1;
            }

            d = Phloat.floor(d);
            
            while (d != 0 && bufpos < buflen)
            {
                char c = (char)('0' + Phloat.to_digit(d));
                buf[bufpos++] = c;
                d = Phloat.floor(d / 10);
            }
            
            if (bufpos == 0)
                buf[bufpos++] = '0';
            
            if (s == -1 && bufpos < buflen)
                buf[bufpos++] = '-';
            
            return bufpos;
        }

        public static int easy_phloat2string(Phloat d, char[] buf, int buflen, int base_mode)
        {
            int dispmode;
            int digits = 0;

            if (flags.fix_or_all != 0)
                dispmode = flags.eng_or_all != 0 ? 3 : 0;
            else
                dispmode = flags.eng_or_all != 0 ? 2 : 1;

            if (flags.digits_bit3 != 0)
                digits += 8;
            
            if (flags.digits_bit2 != 0)
                digits += 4;
            
            if (flags.digits_bit1 != 0)
                digits += 2;
            
            if (flags.digits_bit0 != 0)
                digits += 1;

            return Phloat.phloat2string(d, buf, buflen, base_mode, digits, dispmode, flags.thousands_separators);
        }

        public static int get_base_param(vartype v, ref long n)
        {
            if (v.type == TYPE.STRING)
	            return ERR_ALPHA_DATA_IS_INVALID;
            else if (v.type != TYPE.REAL)
	            return ERR_INVALID_TYPE;

            Phloat x = ((vartype_real) v).x;
            
            if (x > 34359738367.0 || x < -34359738368.0)
	            return ERR_INVALID_DATA;

            long t = Phloat.to_int8(x);
            
            if ((t & 0x800000000) != 0)
	            n = (long)((ulong)t | (ulong)0xfffffff000000000);
            else
	            n = t;
            
            return ERR_NONE;
        }

        public static int base_range_check(ref long n)
        {
            if (n < -34359738368)
            {
                if (flags.range_error_ignore != 0)
                    n = -34359738368;
                else
                    return ERR_OUT_OF_RANGE;
            }
            else if (n > 34359738367)
            {
                if (flags.range_error_ignore != 0)
                    n = 34359738367;
                else
                    return ERR_OUT_OF_RANGE;
            }

            return ERR_NONE;
        }
    }
}
