﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        public static void keydown(int shift, int key)
        {
            int menu;

            pending_command = CMD_NONE;

            if (key >= 1024)
            {
                // TO DO: is this actually needed?
                // Filtering out ASCII key events if the alpha menu is not actually
                // active, just in case the subsequent key handling code doesn't
                // handle them properly in all cases. Using a code coverage tool on
                // the keydown handler might be an idea...
                //
                menu = get_front_menu();

                if (menu == (int)MENU.NONE || menu < (int)MENU.ALPHA1 || menu > (int)MENU.ALPHA_MISC2)
                    return;
            }
            else if (key < 1 || key > 37)
            {
                // Bad key code
                squeak();
                return;
            }

            if (mode_clall)
            {
                if (shift == 0 && key == KEY_SIGMA)
                    pending_command = CMD_CLALLb;
                else if (key == KEY_EXIT)
                    pending_command = CMD_CANCELLED;
                else
                    pending_command = CMD_NULL;

                mode_clall = false;
                return;
            }

            if (mode_getkey)
            {
                vartype result = App.new_real(shift != 0 ? key + 37 : key);

                if (result != null)
                {
                    recall_result(result);
                    flags.stack_lift_disable = 0;
                }
                else
                {
                    display_error(ERR_INSUFFICIENT_MEMORY, 1);
                    set_running(false);
                }

                if (key == KEY_EXIT || (shift == 0 && key == KEY_RUN))
                    set_running(false);

                mode_getkey = false;
                
                if (!mode_running)
                    redisplay();
                
                return;
            }

            if (shift != 0 && key == KEY_EXIT)
            {
                pending_command = CMD_SILENT_OFF;
                return;
            }

            if (shift != 0 && key == KEY_RUN)
            {
                if (mode_command_entry)
                {
                    squeak();
                    return;
                }

                if (flags.prgm_mode != 0)
                {
                    if (mode_alpha_entry)
                        finish_alpha_prgm_line();
                    else if (mode_number_entry)
                    {
                        arg_struct arg = new arg_struct();
                        
                        arg.type = ARGTYPE_DOUBLE;
                        arg.val_d = entered_number;
                        
                        store_command(pc, CMD_NUMBER, arg);
                        prgm_highlight_row = 1;
                    }
                }
                else if (mode_alpha_entry)
                {
                    if ((flags.trace_print != 0 || flags.normal_print != 0) && flags.printer_exists != 0)
                        docmd_pra(null);
                }
                else if (mode_number_entry)
                {
                    if ((flags.trace_print != 0 || flags.normal_print != 0) && flags.printer_exists != 0)
                    {
                        deferred_print = 1;
                        print_command(CMD_NULL, null);
                    }
                }

                mode_alpha_entry = false;
                mode_number_entry = false;
                
                flags.prgm_mode = (byte)((flags.prgm_mode == 0) ? 1 : 0);

                mode_varmenu = false;

                if (flags.prgm_mode != 0)
                {
                    if (mode_appmenu == MENU_BASE_A_THRU_F)
                        set_menu(MENULEVEL_APP, (int)MENU.BASE);
                    else if (mode_plainmenu == MENU_PROGRAMMABLE)
                        set_menu(MENULEVEL_PLAIN, (int)MENU.NONE);

                    input_length = 0;
                }

                flags.message = 0;
                flags.two_line_message = 0;

                redisplay();
                return;
            }

            if (flags.message != 0 && shift == 0 && key == KEY_BSP)
            {
                flags.message = 0;
                flags.two_line_message = 0;

                redisplay();
                return;
            }

            flags.message = 0;
            flags.two_line_message = 0;

            menu = get_front_menu();

            if (mode_number_entry && get_base() == 16 && key == KEY_SIGMA
                && menu != (int)MENU.NONE && menu == (int)MENU.BASE)
            {
                // Special case -- entering the A...F menu while in base 16
                // does *not* cancel number entry mode (unlike all other menu
                // keys)... So we intercept and handle it before all the other
                // logic can mess things up.
                //
                keydown_number_entry(-1, -1);
                return;
            }

            if (mode_number_entry
                && !calc_off
                && is_number_key(shift, key) == 0
                && (key != KEY_CHS || shift != 0 || basekeys() != 0 || get_base() != 10)
                && (key != KEY_BSP || shift != 0))
            {
                // Leaving number entry mode
                mode_number_entry = false;

                if (flags.prgm_mode != 0)
                {
                    arg_struct arg = new arg_struct();

                    arg.type = ARGTYPE_DOUBLE;
                    arg.val_d = entered_number;
                    
                    store_command(pc, CMD_NUMBER, arg);
                    prgm_highlight_row = 1;
                }
                else if ((flags.trace_print != 0|| flags.normal_print != 0) && flags.printer_exists != 0)
                    deferred_print = 1;
            }

            if (mode_command_entry && shift != 0 && (key == KEY_UP || key == KEY_DOWN))
            {
                // Trying to do SST or BST while in command entry mode
                squeak();
                return;
            }

            if (key == KEY_UP || (key == KEY_DOWN && (flags.prgm_mode != 0 || (shift == 0 && get_front_menu() != (int)MENU.NONE))))
            {
                // UP, DOWN, BST, or prgm-mode SST
                repeating = 1;
                repeating_shift = shift;
                repeating_key = key;
            }

            if (flags.prgm_mode != 0 && (key == KEY_UP || key == KEY_DOWN) && (shift != 0 || get_front_menu() == (int)MENU.NONE))
            {
                // Stepping through the program in prgm mode
                if (flags.prgm_mode != 0 && mode_alpha_entry)
                    finish_alpha_prgm_line();

                clear_all_rtns();
                
                if (key == KEY_UP)
                    bst();
                else
                    sst();
                
                redisplay();
                return;
            }

            if (flags.prgm_mode == 0 && key == KEY_UP && (shift != 0 || get_front_menu() == (int)MENU.NONE))
            {
                // BST in normal or alpha mode
                if (mode_alpha_entry && (flags.trace_print != 0 || flags.normal_print != 0) && flags.printer_exists != 0)
                    docmd_pra(null);

                mode_alpha_entry = false;
                clear_all_rtns();
                bst();
                flags.prgm_mode = 1;
                
                redisplay();
                
                flags.prgm_mode = 0;
                pending_command = CMD_CANCELLED;
                return;
            }

            if (mode_number_entry)
            {
                if (!calc_off)
                    keydown_number_entry(shift, key);
            }
            else if (mode_command_entry)
                keydown_command_entry(shift, key);
            else if (flags.alpha_mode != 0)
                keydown_alpha_mode(shift, key);
            else
                keydown_normal_mode(shift, key);
        }

        public static void keydown_number_entry(int shift, int key)
        {
            Phloat x = new Phloat();
            char[] buf = new char[100];
            int bufptr;
            int _base = get_base();

            if (key == -1)
            {
	            // Hack... The user is switching to the A...F menu
	            set_menu(MENULEVEL_APP, MENU_BASE_A_THRU_F);
	            redisplay();

	            goto draw_number;
            }

            if (_base != 10 && (key == KEY_E || key == KEY_DOT))
	            return;

            // NOTE: 'key' can only be KEY_CHS at this point in the code
            // if 'baseapp' is false and get_base() returns 10; in all other
            // cases, the +/- key will end number entry mode and we won't
            // get here, and the CMD_CHS or CMD_BASECHS function will be
            // invoked instead.
            //

            if (key == KEY_BSP && cmdline_length == 1)
            {
	            mode_number_entry = false;
	        
                if (flags.prgm_mode != 0)
                {
	                pc = line2pc(pc2line(pc) - 1);
	                prgm_highlight_row = 0;
	             
                    redisplay();
	                
                    return;
	            }
                else
                {
	                pending_command = CMD_CLX;
	                return;
	            }
            }

            if (key == KEY_BSP)
            {
	            cmdline_length--;
	        
                if (flags.prgm_mode == 0 && _base == 10)
	                fix_thousands_separators(cmdline, ref cmdline_length);

	            if (core_settings.auto_repeat)
                {
	                repeating = 2;
	                repeating_key = key;
	                repeating_shift = shift;
	            }
            }
            else if (key == KEY_CHS)
            {
	            // Check if mantissa or exponent gets the sign change
	            int i;
	            int exp_pos = -1;

	            for (i = 0; i < cmdline_length; i++)
	                if (cmdline[i] == 24)
                    {
		                exp_pos = i;
		                break;
	                }
	        
                if (exp_pos != -1)
                {
	                // Change exponent sign
	                int failed = 0;
	                Phloat d = new Phloat();

	            undo_chs_exp:
	                if (cmdline_length > exp_pos + 1 && cmdline[exp_pos + 1] == '-')
                    {
		                for (i = exp_pos + 1; i < cmdline_length - 1; i++)
		                    cmdline[i] = cmdline[i + 1];
		        
                        cmdline_length--;
	                }
                    else
                    {
        		        for (i = cmdline_length; i > exp_pos + 1; i--)
		                    cmdline[i] = cmdline[i - 1];
		        
                        cmdline[exp_pos + 1] = '-';
		                cmdline_length++;
	                }

    	            if (failed == 0)
                    {
		                if (Phloat.string2phloat(cmdline, cmdline_length, ref d) != 0)
                        {
    		                failed = 1;
	        	
                            goto undo_chs_exp;
		                }
	                }
                    else
		                return;
                }
                else
                {
	                // Change mantissa sign
	                if (cmdline[0] == '-')
                    {
		                if (cmdline_length == 1)
                        {
		                    if (flags.prgm_mode != 0)
                            {
			                    mode_number_entry = false;
			                    pc = line2pc(pc2line(pc) - 1);
			                    prgm_highlight_row = 0;
			                    
                                redisplay();
			                    return;
		                    }
                            else
                            {
			                    // This is a bit odd, but it's how the HP-42S
			                    // does it, so there.
			                    //
			                    mode_number_entry = false;
			                    free_vartype(ref reg_x);
			                    reg_x = new_real(0);
			                    pending_command = CMD_CLX;
			                    
                                return;
		                    }
		                }
		        
                        for (i = 0; i < cmdline_length - 1; i++)
		                    cmdline[i] = cmdline[i + 1];

		                cmdline_length--;
                    }
                    else
                    {
        		        for (i = cmdline_length; i > 0; i--)
		                    cmdline[i] = cmdline[i - 1];
		        
                        cmdline[0] = '-';
		                cmdline_length++;
	                }
	            }
            }
            else if (key == KEY_E)
            {
	            int exp_pos = -1;
	            int only_zeroes = 1;
	            int seen_dot = 0;

                char dot = flags.decimal_point != 0 ? '.' : ',';

                int i;
	            for (i = 0; i < cmdline_length; i++)
                {
	                char c = cmdline[i];
	                
                    if (c >= '1' && c <= '9')
		                only_zeroes = 0;
	                else if (c == dot)
		                seen_dot = 1;
	                else if (c == 24)
                    {
		                exp_pos = i;
		                break;
	                }
	            }
	        
                if (exp_pos == -1)
                {
	                if (only_zeroes != 0)
                    {
		                if (cmdline_length > 0 && cmdline[0] == '-')
		                    cmdline_length = 1;
		                else
		                    cmdline_length = 0;

		                cmdline[cmdline_length++] = '1';

		                if (seen_dot != 0)
		                    cmdline[cmdline_length++] = dot;
	                }

	                cmdline[cmdline_length++] = (char)24;
	            }
                    else
	                    return;
            }
            else if (key == KEY_DOT)
            {
	            if (cmdline_length == 0 || (cmdline_length == 1 && cmdline[0] == '-'))
                {
	                cmdline[cmdline_length++] = '0';
	                cmdline[cmdline_length++] = flags.decimal_point != 0 ? '.' : ',';
	            }
                else
                {
	                // Only allow dot if there isn't one already, and
	                // there is no exponent either
	                //
	                int dot_or_exp_pos = -1;
	                int i;
	                char dot = flags.decimal_point != 0 ? '.' : ',';
	            
                    for (i = 0; i < cmdline_length; i++)
		                if (cmdline[i] == dot || cmdline[i] == 24)
                        {
		                    dot_or_exp_pos = i;
		                    break;
		                }
	            
                    if (dot_or_exp_pos == -1)
		                cmdline[cmdline_length++] = dot;
	                else
		                return;
	            }
            }
            else // KEY_0 .. KEY_9 or hex A-F
            {
	            int digit = 0;
	            char c;

                switch (key)
                {
                    case KEY_0: digit = 0; break;
	                case KEY_1: digit = 1; break;
	                case KEY_2: digit = 2; break;
	                case KEY_3: digit = 3; break;
	                case KEY_4: digit = 4; break;
	                case KEY_5: digit = 5; break;
	                case KEY_6: digit = 6; break;
	                case KEY_7: digit = 7; break;
	                case KEY_8: digit = 8; break;
	                case KEY_9: digit = 9; break;
	                case KEY_SIGMA: digit = 10; break;
	                case KEY_INV:   digit = 11; break;
	                case KEY_SQRT:  digit = 12; break;
	                case KEY_LOG:   digit = 13; break;
	                case KEY_LN:    digit = 14; break;
	                case KEY_XEQ:   digit = 15; break;
	            }

                if (digit >= _base)
	                return;
	        
                if (core_settings.auto_repeat)
                {
	                repeating = 2;
	                repeating_key = key;
	                repeating_shift = shift;
	            }

	            c = (char)(digit < 10 ? '0' + digit : 'A' + digit - 10);
	            cmdline[cmdline_length++] = c;

    	        if (_base == 10)
                {
	                if (Phloat.string2phloat(cmdline, cmdline_length, ref x) == 0)
                    {
		                if (flags.prgm_mode == 0)
		                    fix_thousands_separators(cmdline, ref cmdline_length);
	                }
                    else
                    {
		                cmdline_length--;
		                return;
	                }
	            }
                else
                {
	                int bits = _base == 2 ? 1 : _base == 8 ? 3 : 4;
	                bits *= cmdline_length;

                    if (bits > 36)
                    {
		                cmdline_length--;
		                return;
	                }
	            }
            }

            if (_base == 10)
            {
    	        if (Phloat.string2phloat(cmdline, cmdline_length, ref x) != 0)
	                x = 0;          // Should never happen
            }
            else
            {
	            Int64 n = 0;
	            int i;
	        
                for (i = 0; i < cmdline_length; i++)
                {
	                char c = cmdline[i];
	                int digit = c <= '9' ? c - '0' : c - 'A' + 10;
	                n = n * _base + digit;
	            }

                if ((n & 0x800000000) != 0)
	                n = (long)((ulong)n | (ulong)0xfffffff000000000);
	        
                x = new Phloat(n);
            }

            if (flags.prgm_mode != 0)
	            entered_number = x;
            else
            {
	            free_vartype(ref reg_x);
	            reg_x = new_real(x);
            }

        draw_number:

            bufptr = 0;
        
            if (flags.prgm_mode != 0)
            {
	            int line = pc2line(pc);

	            if (line < 10)
	                char2buf(buf, 100, ref bufptr, '0');

	            bufptr += int2string(line, buf, bufptr, 100);
    	        char2buf(buf, 100, ref bufptr, (char)6);
            }
            else if (matedit_mode == 2 || matedit_mode == 3)
            {
	            bufptr += int2string(matedit_i + 1, buf, bufptr, 100);
	            char2buf(buf, 100, ref bufptr, ':');
	            bufptr += int2string(matedit_j + 1, buf, bufptr, 100);
	            char2buf(buf, 100, ref bufptr, '=');
            }
            else if (input_length > 0)
            {
	            string2buf(buf, 100, ref bufptr, input_name, input_length);
	            char2buf(buf, 100, ref bufptr, '?');
            }
            else
	            string2buf(buf, 100, ref bufptr, "x\x0080".ToCharArray(), 2);
            
            string2buf(buf, 100, ref bufptr, cmdline, cmdline_length);
            char2buf(buf, 100, ref bufptr, '_');

            clear_row(cmdline_row);

            if (bufptr <= 22)
	            draw_string(0, cmdline_row, buf, bufptr);
            else
            {
	            draw_char(0, cmdline_row, (char)26);

                char[] tmp = new char[21];
                Array.Copy(buf, bufptr - 21, tmp, 0, 21);
	            draw_string(1, cmdline_row, tmp, 21);
            }

            flush_display();
            return;
        }

        public static int is_number_key(int shift, int key)
        {
            int menu = get_front_menu();

            if (menu != (int)MENU.NONE && menu == (int)MENU.BASE_A_THRU_F
                && (key == KEY_SIGMA || key == KEY_INV || key == KEY_SQRT
                || key == KEY_LOG || key == KEY_LN || key == KEY_XEQ))
                return 1;

            return (shift == 0 && (key == KEY_0 || key == KEY_1 || key == KEY_2
                || key == KEY_3 || key == KEY_4 || key == KEY_5
                || key == KEY_6 || key == KEY_7 || key == KEY_8
                || key == KEY_9 || key == KEY_DOT || key == KEY_E)) ? 1 : 0;
        }

        public static int basekeys()
        {
            if (baseapp == 0)
                return 0;

            int menu = get_front_menu();

            return (menu != (int)MENU.NONE && (menu == (int)MENU.BASE || menu == (int)MENU.BASE_A_THRU_F || menu == (int)MENU.BASE_LOGIC)) ? 1 : 0;
        }

        public static void keydown_command_entry(int shift, int key)
        {
            if (mode_commandmenu == MENU_ST || mode_commandmenu == MENU_IND_ST)
            {
	            int menukey;

                if (shift == 0 && key == KEY_BSP)
                {
	                pending_command = CMD_NULL;
	                finish_command_entry(false);
	                return;
	            }

	            if (key == KEY_EXIT)
                {
	                pending_command = CMD_CANCELLED;
	                finish_command_entry(false);
	                return;
	            }

	            menukey = find_menu_key(key);

	            if (shift == 0 && ((menukey >= 0 && menukey <= 4) || (menukey == 5 && mode_commandmenu == MENU_IND_ST)))
                {
	                if (mode_commandmenu == MENU_IND_ST && menukey == 0)
                    {
		                incomplete_ind = 1;
		                incomplete_maxdigits = 2;
		                
                        set_catalog_menu(CATSECT_REAL_ONLY);
		                redisplay();
		                return;
	                }

                    if (mode_commandmenu == MENU_IND_ST)
		                menukey--;

	                pending_command = incomplete_command;
	                pending_command_arg.type = (byte)(incomplete_ind != 0 ? ARGTYPE_IND_STK : ARGTYPE_STK);
	            
                    pending_command_arg.val.stk = "LXYZT"[menukey];
	            
                    finish_command_entry(true);
	                return;
	            }

	            squeak();
	            return;
            }

            if (incomplete_command == CMD_LBL && incomplete_length == 0 && mode_commandmenu != MENU_CATALOG)
            {
	            // LBL is weird. It's sort of like you have alpha and numeric
	            // at the same time. When we're at length 0, we have to handle
	            // both possibilities and pick the right one.
	            //

	            // On the Palm, we accept 0-9 as alphanumeric in this case;
	            // if the user wants numeric, it's easy enough to just tap
	            // the appropriate keys on the virtual keyboard.
	            // On PCs, on the other hand, we want the number keys to
	            // behave like they always do, so you can enter LBL 00 without
	            // having to use the virtual keyboard. If an alpha LBL with
	            // a name starting with '0'-'9' is desired, the user will have
	            // to use the usual trick: activate a submenu of ALPHA by
	            // pressing any menu key (F1-F6), and then typing the number.
	            //
	            if ((mode_commandmenu == MENU_ALPHA1 || mode_commandmenu == MENU_ALPHA2) && key >= 1024 + '0' && key <= 1024 + '9')
	                switch (key - 1024)
                    {
		                case '0': key = KEY_0; break;
		                case '1': key = KEY_1; break;
		                case '2': key = KEY_2; break;
		                case '3': key = KEY_3; break;
		                case '4': key = KEY_4; break;
		                case '5': key = KEY_5; break;
		                case '6': key = KEY_6; break;
		                case '7': key = KEY_7; break;
		                case '8': key = KEY_8; break;
		                case '9': key = KEY_9; break;
	                }

                if (key >= 1024
                    || (key == KEY_SIGMA || key == KEY_INV || key == KEY_SQRT || key == KEY_LOG || key == KEY_LN || key == KEY_XEQ)
                    || (shift == 0 &&
                        (key == KEY_E || key == KEY_UP || key == KEY_DOWN || key == KEY_DIV || key == KEY_MUL || key == KEY_SUB || key == KEY_ADD || key == KEY_DOT))
                    || (shift != 0 && (key == KEY_RCL || key == KEY_RDN))
                    || (mode_commandmenu >= MENU_ALPHA_ABCDE1 && mode_commandmenu <= MENU_ALPHA_MISC2 &&
                        (key == KEY_0 || key == KEY_1 || key == KEY_2 || key == KEY_3 || key == KEY_4 || key == KEY_5
                        || key == KEY_6 || key == KEY_7 || key == KEY_8 || key == KEY_9)))
                // goto do_incomplete_alpha;
                {
                    do_incomplete_alpha(shift, key);
                    return;
                }
	        
                if (shift != 0 && key == KEY_ADD)
                {
	                if (mode_commandmenu == MENU_CATALOG)
		                squeak();
	                else
                    {
		                incomplete_alpha = 1;
		                set_catalog_menu(CATSECT_TOP);
		                redisplay();
	                }

                    return;
	            }

                if (key == KEY_EXIT && mode_commandmenu >= MENU_ALPHA_ABCDE1 && mode_commandmenu <= MENU_ALPHA_MISC2)
                {
    	            if (mode_commandmenu <= MENU_ALPHA_WXYZ)
	        	        mode_commandmenu = MENU_ALPHA1;
	                else
		                mode_commandmenu = MENU_ALPHA2;

	                redisplay();
	                return;
	            }
            }



            if ((incomplete_command == CMD_ASTO || incomplete_command == CMD_ARCL)
                && incomplete_length == 1 && mode_commandmenu == (int)MENU.NONE
                && mode_alphamenu >= MENU_ALPHA1 && mode_alphamenu <= MENU_ALPHA_MISC2)
            {
                // A similar oddity are ASTO and ARCL, when entered using the
                // STO and RCL keys while ALPHA mode is active. Initially,
                // you'll get a menu of all variables, but when you type a
                // digit, the variables menu disappears, and the ALPHA menu
                // returns... but we still want 0-9 to be treated as numeric.
                
                switch (key - 1024)
                {
                    case '0': key = KEY_0; break;
                    case '1': key = KEY_1; break;
                    case '2': key = KEY_2; break;
                    case '3': key = KEY_3; break;
                    case '4': key = KEY_4; break;
                    case '5': key = KEY_5; break;
                    case '6': key = KEY_6; break;
                    case '7': key = KEY_7; break;
                    case '8': key = KEY_8; break;
                    case '9': key = KEY_9; break;
                }
            }


    
            // Another oddity: ASSIGN...
            if (incomplete_argtype == ARG_CKEY)
            {
	            int menukey = find_menu_key(key);

	            if (menukey != -1)
                {
	                pending_command = CMD_ASGN01 + menukey + 6 * (mode_commandmenu - MENU_CUSTOM1);
	                finish_command_entry(true);
	            }
                else if (shift == 0 && key == KEY_BSP)
                {
	                pending_command = CMD_NULL;
	                finish_command_entry(false);
	            }
                else if (key == KEY_EXIT)
                {
	                pending_command = CMD_CANCELLED;
	                finish_command_entry(false);
	            }
                else if (shift == 0 && (key == KEY_UP || key == KEY_DOWN))
                {
	                mode_commandmenu += key == KEY_UP ? -1 : 1;
	            
                    if (mode_commandmenu < MENU_CUSTOM1)
		                set_menu(MENULEVEL_COMMAND, MENU_CUSTOM3);
	                else if (mode_commandmenu > MENU_CUSTOM3)
		                set_menu(MENULEVEL_COMMAND, MENU_CUSTOM1);
	            
                    redisplay();
	            }
                else
	                squeak();
	            
                return;
            }

            // And yet another oddity: KEYG and KEYX
            if (incomplete_argtype == ARG_MKEY)
            {
	            int cmd = incomplete_command == CMD_KEYG ? CMD_KEY1G : CMD_KEY1X;

	            if (shift != 0)
                {
	                squeak();
	                return;
	            }

	            switch (key)
                {
	                case KEY_SIGMA: case KEY_1: break;
	                case KEY_INV: case KEY_2: cmd++; break;
	                case KEY_SQRT: case KEY_3: cmd += 2; break;
	                case KEY_LOG: case KEY_4: cmd += 3; break;
	                case KEY_LN: case KEY_5: cmd += 4; break;
	                case KEY_XEQ: case KEY_6: cmd += 5; break;
	                case KEY_UP: case KEY_7: cmd += 6; break;
	                case KEY_DOWN: case KEY_8: cmd += 7; break;
	                case KEY_EXIT: case KEY_9: cmd += 8; break;
	                case KEY_BSP:
		                pending_command = CMD_NULL;
		                finish_command_entry(false);
		                return;
	                default:
		                squeak();
		                return;
	            }

	            start_incomplete_command(cmd);
	            return;
            }

            if (mode_commandmenu == MENU_CATALOG)
            {
	            int menukey = find_menu_key(key);
	            int catsect = get_cat_section();

	            if (menukey != -1)
                {
	                if (catsect == CATSECT_TOP)
                    {
        		        switch (menukey)
                        {
		                    case 0:
			                    set_cat_section(CATSECT_FCN);
			                    move_cat_row(0);
			                    break;

                            case 1:
			                    set_cat_section(CATSECT_PGM);
			                    move_cat_row(0);
			                    break;
		            
                            case 2:
			                    if (vars_exist(1, 0, 0) == 0)
                                {
			                        squeak();
			                        return;
			                    }
                                else
                                {
			                        set_cat_section(CATSECT_REAL);
			                        move_cat_row(0);
			                        break;
			                    }
		            
                            case 3:
			                    if (vars_exist(0, 1, 0) == 0)
                                {
			                        squeak();
			                        return;
			                    }
                                else
                                {
			                        set_cat_section(CATSECT_CPX);
			                        move_cat_row(0);
			                        break;
			                    }
		            
                            case 4:
			                    if (vars_exist(0, 0, 1) == 0)
                                {
			                        squeak();
			                        return;
			                    }
                                else
                                {
			                        set_cat_section(CATSECT_MAT);
			                        move_cat_row(0);
			                        break;
			                    }
		            
                            case 5:
                                display_mem();
			                    pending_command = CMD_LINGER1;
			                    shell_request_timeout3(2000);
			                    return;
                        }

                        redisplay();
		                return;
	                }
                    else
                    {
        		        int i;
                        int itemindex = get_cat_item(menukey);

        		        if (itemindex == -1)
                        {
		                    squeak();
		                    return;
		                }

		                if (catsect == CATSECT_PGM || catsect == CATSECT_PGM_ONLY)
                        {
        		            if (labels[itemindex].length == 0)
                            {   // END or .END.
			                    if (incomplete_command != CMD_GTODOT && incomplete_command != CMD_PRP &&
                                    (flags.prgm_mode != 0 || (incomplete_command != CMD_GTO && incomplete_command != CMD_XEQ && incomplete_command != CMD_CLP)))
                                {
			                        squeak();
			                        return;
			                    }
		                    }
    
                            pending_command = incomplete_command;
		            
                            if (incomplete_command == CMD_GTO || incomplete_command == CMD_GTODOT || incomplete_command == CMD_XEQ
			                    || incomplete_command == CMD_CLP || incomplete_command == CMD_PRP)
                            {
			                    pending_command_arg.type = ARGTYPE_LBLINDEX;
			                    pending_command_arg.val.num = itemindex;
			                    xeq_invisible = 0;
		                    }
                            else
                            {
            			        pending_command_arg.type = (byte)(incomplete_ind != 0 ? ARGTYPE_IND_STR : ARGTYPE_STR);
			        
                                pending_command_arg.length = labels[itemindex].length;
			        
                                for (i = 0; i < pending_command_arg.length; i++)
			                        pending_command_arg.val.text[i] = labels[itemindex].name[i];
                            }

                            finish_command_entry(true);
		                    return;
                        }

                        pending_command = incomplete_command;
		                pending_command_arg.type = (byte)(incomplete_ind != 0 ? ARGTYPE_IND_STR : ARGTYPE_STR);
		        
                        if (catsect == CATSECT_FCN)
                        {
        		            command_spec cs = cmd_array[itemindex];

		                    pending_command_arg.length = (byte)(cs.name_length);
		            
                            for (i = 0; i < pending_command_arg.length; i++)
			                    pending_command_arg.val.text[i] = cs.name[i];
                        }
                        else
                        {
        		            pending_command_arg.length = vars[itemindex].length;
		
                            for (i = 0; i < pending_command_arg.length; i++)
			                    pending_command_arg.val.text[i] = vars[itemindex].name[i];
                        }

                        if (incomplete_ind == 0 && incomplete_command == CMD_XEQ)
		                    finish_xeq();
		                else
		                    finish_command_entry(true);
		                
                        return;
                    }
    	        }
	
                if (shift == 0 && (key == KEY_UP || key == KEY_DOWN))
                {
	                move_cat_row(key == KEY_UP ? -1 : 1);
	                redisplay();
	                return;
	            }

	            if (key == KEY_EXIT)
                {
	                if (catsect == CATSECT_FCN
		                || catsect == CATSECT_PGM
		                || catsect == CATSECT_REAL
		                || catsect == CATSECT_CPX
		                || catsect == CATSECT_MAT)
                    {
		                set_cat_section(CATSECT_TOP);
		                redisplay();
	                }
                    else
                    {
		                pending_command = CMD_CANCELLED;
		                finish_command_entry(false);
    	            }
	
                    return;
	            }
            }

            if (mode_commandmenu == MENU_VARMENU)
            {
	            int menukey = find_menu_key(key);
	        
                if (menukey != -1)
                {
	                int i;
	            
                    if (varmenu_labellength[menukey] == 0)
                    {
		                squeak();
		                return;
    	            }
	            
                    pending_command = incomplete_command;
	                pending_command_arg.type = ARGTYPE_STR;
	                pending_command_arg.length = (byte)varmenu_labellength[menukey];
	                
                    for (i = 0; i < pending_command_arg.length; i++)
		                pending_command_arg.val.text[i] = varmenu_labeltext[menukey][i];
	            
                    finish_command_entry(false);
	                return;
	            }
	        
                if (shift == 0 && (key == KEY_UP || key == KEY_DOWN))
                {
	                if (varmenu_rows > 1)
                    {
        		        if (key == KEY_UP)
                        {
        		            varmenu_row--;
		
                            if (varmenu_row < 0)
			                    varmenu_row = varmenu_rows - 1;
		                }
                        else
                        {
        		            varmenu_row++;
		        
                            if (varmenu_row >= varmenu_rows)
			                    varmenu_row = 0;
		                }

                        redisplay();
	                }

                    return;
	            }

                if (key == KEY_EXIT)
                {
	                pending_command = CMD_CANCELLED;
	                finish_command_entry(false);
	                return;
	            }
            }
            else if (mode_commandmenu == MENU_INTEG_PARAMS)
            {
	            int menukey = find_menu_key(key);

	            if (menukey != -1)
                {
	                char[] name;
	                int length;
                    int i;

	                switch (menukey)
                    {
		                case 0: name = "LLIM".ToCharArray(); length = 4; break;
		                case 1: name = "ULIM".ToCharArray(); length = 4; break;
		                case 2: name = "ACC".ToCharArray();  length = 3; break;
		                default: squeak(); return;
	                }

	                pending_command = incomplete_command;
	                pending_command_arg.type = ARGTYPE_STR;
	                pending_command_arg.length = (byte)length;
	                
                    for (i = 0; i < length; i++)
		                pending_command_arg.val.text[i] = name[i];
	                
                    finish_command_entry(false);
	                return;
	            }
	        
                if (key == KEY_EXIT)
                {
	                pending_command = CMD_CANCELLED;
	                finish_command_entry(false);
	                return;
	            }
            }

            if (incomplete_alpha == 0)
            {
                if (key == KEY_EXIT)
                {
                    pending_command = CMD_CANCELLED;
                    finish_command_entry(false);
                    return;
                }

                if (mode_commandmenu == MENU_IND)
                {
                    if (shift == 0 && key == KEY_SIGMA)
                    {
                        incomplete_command = CMD_GTO;
                        incomplete_argtype = ARG_LBL;
                        incomplete_ind = 1;
                        incomplete_maxdigits = 2;

                        set_catalog_menu(CATSECT_REAL_ONLY);
                        redisplay();

                        return;
                    }
                    else if (key == KEY_ENTER)
                    {
                        incomplete_argtype = ARG_LBL;
                        incomplete_alpha = 1;

                        set_menu(MENULEVEL_COMMAND, MENU_ALPHA1);
                        redisplay();

                        return;
                    }
                }

                if (incomplete_command == CMD_STO && shift == 0 && incomplete_length == 0
                    && (key == KEY_DIV || key == KEY_MUL || key == KEY_SUB || key == KEY_ADD))
                {
                    switch (key)
                    {
                        case KEY_DIV: incomplete_command = CMD_STO_DIV; break;
                        case KEY_MUL: incomplete_command = CMD_STO_MUL; break;
                        case KEY_SUB: incomplete_command = CMD_STO_SUB; break;
                        case KEY_ADD: incomplete_command = CMD_STO_ADD; break;
                    }

                    redisplay();
                    return;
                }

                if (incomplete_command == CMD_RCL && shift == 0 && incomplete_length == 0
                    && (key == KEY_DIV || key == KEY_MUL || key == KEY_SUB || key == KEY_ADD))
                {
                    switch (key)
                    {
                        case KEY_DIV: incomplete_command = CMD_RCL_DIV; break;
                        case KEY_MUL: incomplete_command = CMD_RCL_MUL; break;
                        case KEY_SUB: incomplete_command = CMD_RCL_SUB; break;
                        case KEY_ADD: incomplete_command = CMD_RCL_ADD; break;
                    }

                    redisplay();
                    return;
                }

                if (incomplete_command == CMD_GTO && incomplete_ind == 0 && shift == 0 && incomplete_length == 0 && key == KEY_DOT)
                {
                    incomplete_command = CMD_GTODOT;
                    incomplete_argtype = ARG_OTHER;
                    incomplete_maxdigits = 4;

                    set_menu(MENULEVEL_COMMAND, MENU_IND);
                    redisplay();

                    return;
                }

                if (incomplete_command == CMD_GTODOT && shift == 0 && incomplete_length == 0 && key == KEY_DOT)
                {
                    pending_command = CMD_GTODOTDOT;
                    pending_command_arg.type = ARGTYPE_NONE;
                    finish_command_entry(false);
                    return;
                }


                if (key == KEY_ENTER)
                {
                    if (incomplete_length == 0)
                    {
                        if (incomplete_ind != 0 && flags.prgm_mode == 0 && vars_exist(1, 0, 0) == 0)
                            squeak();
                        else if (incomplete_ind != 0
                            || incomplete_argtype == ARG_VAR
                            || incomplete_argtype == ARG_REAL
                            || incomplete_argtype == ARG_NAMED
                            || incomplete_argtype == ARG_LBL
                            || incomplete_argtype == ARG_PRGM)
                        {
                            incomplete_alpha = 1;

                            set_menu(MENULEVEL_COMMAND, MENU_ALPHA1);
                            redisplay();
                        }
                        else
                            squeak();

                        return;
                    }
                    else if (shift == 0)
                    {
                        if (incomplete_command == CMD_GOTOROW)
                        {
                            pending_command_arg.val.num = incomplete_num;
                            start_incomplete_command(CMD_GOTOCOLUMN);
                            return;
                        }
                        else if (incomplete_command == CMD_GOTOCOLUMN)
                        {
                            matedit_goto(pending_command_arg.val.num, incomplete_num);

                            pending_command = CMD_NONE;
                            finish_command_entry(true);

                            return;
                        }

                        pending_command = incomplete_command;
                        pending_command_arg.type = (byte)(incomplete_ind != 0 ? ARGTYPE_IND_NUM : ARGTYPE_NUM);
                        pending_command_arg.length = (byte)(incomplete_maxdigits);
                        pending_command_arg.val.num = incomplete_num;

                        finish_command_entry(false);
                        return;
                    }
                }

                if (incomplete_length == 0 && shift == 0 && key == KEY_DOT)
                {
                    if (incomplete_ind != 0)
                    {
                        if (incomplete_argtype == ARG_VAR
                            || incomplete_argtype == ARG_REAL
                            || incomplete_argtype == ARG_MAT
                            || incomplete_argtype == ARG_RVAR
                            || incomplete_argtype == ARG_NAMED
                            || incomplete_argtype == ARG_LBL
                            || incomplete_argtype == ARG_PRGM
                            || incomplete_argtype == ARG_NUM9
                            || incomplete_argtype == ARG_NUM11
                            || incomplete_argtype == ARG_NUM99)
                        {
                            set_menu(MENULEVEL_COMMAND, MENU_ST);
                            redisplay();
                        }
                    }
                    else
                    {
                        if (incomplete_argtype == ARG_VAR || incomplete_argtype == ARG_REAL)
                        {
                            set_menu(MENULEVEL_COMMAND, MENU_IND_ST);
                            redisplay();
                        }
                        else if (incomplete_argtype == ARG_NUM9
                            || incomplete_argtype == ARG_NUM11
                            || incomplete_argtype == ARG_NUM99
                            || incomplete_argtype == ARG_LBL)
                        {
                            incomplete_ind = 1;
                            incomplete_maxdigits = 2;

                            set_catalog_menu(CATSECT_REAL_ONLY);
                            redisplay();
                        }
                        else
                        {
                            squeak();
                        }
                    }

                    return;
                }

                if (incomplete_length < incomplete_maxdigits && shift == 0 &&
                    (key == KEY_0 || key == KEY_1 || key == KEY_2
                    || key == KEY_3 || key == KEY_4 || key == KEY_5
                    || key == KEY_6 || key == KEY_7 || key == KEY_8
                    || key == KEY_9))
                {
                    int digit = 0;

                    switch (key)
                    {
                        case KEY_0: digit = 0; break;
                        case KEY_1: digit = 1; break;
                        case KEY_2: digit = 2; break;
                        case KEY_3: digit = 3; break;
                        case KEY_4: digit = 4; break;
                        case KEY_5: digit = 5; break;
                        case KEY_6: digit = 6; break;
                        case KEY_7: digit = 7; break;
                        case KEY_8: digit = 8; break;
                        case KEY_9: digit = 9; break;
                    }

                    incomplete_num = incomplete_num * 10 + digit;
                    incomplete_length++;

                    if (incomplete_length == incomplete_maxdigits)
                    {
                        if (incomplete_command == CMD_GOTOROW)
                        {
                            pending_command_arg.val.num = incomplete_num;
                            start_incomplete_command(CMD_GOTOCOLUMN);
                            return;
                        }
                        else if (incomplete_command == CMD_GOTOCOLUMN)
                        {
                            matedit_goto(pending_command_arg.val.num, incomplete_num);

                            pending_command = CMD_NONE;
                            finish_command_entry(true);
                            return;
                        }

                        pending_command = incomplete_command;
                        pending_command_arg.type = (byte)(incomplete_ind != 0 ? ARGTYPE_IND_NUM : ARGTYPE_NUM);
                        pending_command_arg.length = (byte)incomplete_maxdigits;

                        if (incomplete_ind == 0 && incomplete_argtype == ARG_NUM11 && incomplete_num > 11)
                            incomplete_num = 11;

                        pending_command_arg.val.num = incomplete_num;
                        finish_command_entry(true);
                        return;
                    }
                    else
                        set_menu(MENULEVEL_COMMAND, (int)MENU.NONE);

                    redisplay();
                    return;
                }

                if (incomplete_length < incomplete_maxdigits && shift == 0 && key == KEY_BSP)
                {
                    if (incomplete_length == 0)
                    {
                        pending_command = CMD_NULL;
                        finish_command_entry(false);
                        return;
                    }
                    else
                    {
                        incomplete_length--;
                        incomplete_num /= 10;

                        if (incomplete_length == 0)
                        {
                            if (incomplete_command >= CMD_KEY1G && incomplete_command <= CMD_KEY9X)
                                start_incomplete_command(incomplete_command <= CMD_KEY9G ? CMD_KEYG : CMD_KEYX);
                            else if (incomplete_argtype == ARG_VAR)
                                if (mode_appmenu == MENU_VARMENU)
                                    set_menu(MENULEVEL_COMMAND, (int)MENU.VARMENU);
                                else if (mode_appmenu == MENU_INTEG_PARAMS)
                                    set_menu(MENULEVEL_COMMAND, MENU_INTEG_PARAMS);
                                else
                                    set_catalog_menu(CATSECT_VARS_ONLY);
                            else if (incomplete_ind != 0)
                                set_catalog_menu(CATSECT_REAL_ONLY);
                            else if (incomplete_argtype == ARG_REAL)
                                if (mode_appmenu == MENU_VARMENU)
                                    set_menu(MENULEVEL_COMMAND, MENU_VARMENU);
                                else if (mode_appmenu == MENU_INTEG_PARAMS)
                                    set_menu(MENULEVEL_COMMAND, MENU_INTEG_PARAMS);
                                else
                                    set_catalog_menu(CATSECT_REAL_ONLY);
                            else if (incomplete_argtype == ARG_LBL)
                                set_catalog_menu(CATSECT_PGM_ONLY);
                            else if (incomplete_command == CMD_GTODOT)
                                set_menu(MENULEVEL_COMMAND, MENU_IND);
                            else if (incomplete_command == CMD_LBL)
                                set_menu(MENULEVEL_COMMAND, MENU_ALPHA1);
                        }

                        redisplay();
                        return;
                    }
                }

                // Some bad key...
                squeak();
                return;
            }
            else
                do_incomplete_alpha(shift, key);
        }

        public static void out_of_alpha()
        {
            if (incomplete_ind != 0 || incomplete_argtype != ARG_RVAR)
                incomplete_alpha = 0;

            if (incomplete_ind != 0)
                set_catalog_menu(CATSECT_REAL_ONLY);
            else if (incomplete_argtype == ARG_VAR)
                if (mode_appmenu == MENU_VARMENU)
                    set_menu(MENULEVEL_COMMAND, MENU_VARMENU);
                else if (mode_appmenu == MENU_INTEG_PARAMS)
                    set_menu(MENULEVEL_COMMAND, MENU_INTEG_PARAMS);
                else
                    set_catalog_menu(CATSECT_VARS_ONLY);
            else if (incomplete_argtype == ARG_REAL)
                if (mode_appmenu == MENU_VARMENU)
                    set_menu(MENULEVEL_COMMAND, MENU_VARMENU);
                else if (mode_appmenu == MENU_INTEG_PARAMS)
                    set_menu(MENULEVEL_COMMAND, MENU_INTEG_PARAMS);
                else
                    set_catalog_menu(CATSECT_REAL_ONLY);
            else if (incomplete_argtype == ARG_LBL)
                set_catalog_menu(CATSECT_PGM_ONLY);
            else if (incomplete_command == CMD_GTODOT)
                set_menu(MENULEVEL_COMMAND, MENU_IND);
            else if (incomplete_command == CMD_LBL)
                set_menu(MENULEVEL_COMMAND, MENU_ALPHA1);
            else
                set_menu(MENULEVEL_COMMAND, (int)MENU.NONE);

            redisplay();
        }

        public static void do_incomplete_alpha(int shift, int key)
        {
            {
                // incomplete_alpha

                menu_spec m;
                int menukey;
                char c;

//            do_incomplete_alpha:
                m = null;

                if (key >= 1024)
                {
                    c = (char)(key - 1024);

//                    goto handle_char;     // Replicate handle_char code here to avoid goto

                    if (incomplete_length < 7)
                        incomplete_str[incomplete_length++] = c;

                    if (m != null)
                        set_menu(MENULEVEL_COMMAND, (int)m.parent);

                    // incomplete_alpha can be 0 at this point if
                    // the command is CMD_LBL.
                    incomplete_alpha = 1;
                    redisplay();
                    return;
                }
                else if (mode_commandmenu != (int)MENU.NONE && (menukey = find_menu_key(key)) != -1)
                {
                    menu_item_spec mi;

                    m = menus[mode_commandmenu];
                    mi = m.child[menukey];

                    if (mi.menuid != MENU.NONE)
                    {
                        set_menu(MENULEVEL_COMMAND, (int)mi.menuid);
                        redisplay();
                        return;
                    }

                    c = mi.title[0];

                    if (shift != 0 && c >= 'A' && c <= 'Z')
                        c += (char)('a' - 'A');

//                handle_char:
                    if (incomplete_length < 7)
                        incomplete_str[incomplete_length++] = c;

                    if (m != null)
                        set_menu(MENULEVEL_COMMAND, (int)m.parent);

                    // incomplete_alpha can be 0 at this point if
                    // the command is CMD_LBL.
                    incomplete_alpha = 1;
                    redisplay();
                    return;
                }

                if ((incomplete_argtype == ARG_NAMED
                        || incomplete_argtype == ARG_PRGM
                        || incomplete_argtype == ARG_RVAR
                        || incomplete_argtype == ARG_MAT)
                    && incomplete_command != CMD_ASSIGNa
                    && incomplete_command != CMD_CLP
                    && incomplete_command != CMD_PRP
                    && incomplete_command != CMD_MVAR
                    && incomplete_ind == 0
                    && incomplete_length == 0
                    && (mode_commandmenu < MENU_ALPHA1
                        || mode_commandmenu > MENU_ALPHA_MISC2)
                    && shift == 0 && key == KEY_DOT)
                {
                    incomplete_ind = 1;
                    incomplete_alpha = 0;
                    set_catalog_menu(CATSECT_REAL_ONLY);
                    redisplay();
                    return;
                }

                if (incomplete_length == 0 && shift != 0 && key == KEY_ADD)
                {
                    if (mode_commandmenu == MENU_CATALOG)
                        squeak();
                    else
                    {
                        set_catalog_menu(CATSECT_TOP);
                        redisplay();
                    }

                    return;
                }

                // Handle keys that represent characters
                if (mode_commandmenu == MENU_CATALOG)
                    goto nocharkey1;

                if (shift == 0)
                {
                    switch (key)
                    {
                        case KEY_0: c = '0'; break;
                        case KEY_1: c = '1'; break;
                        case KEY_2: c = '2'; break;
                        case KEY_3: c = '3'; break;
                        case KEY_4: c = '4'; break;
                        case KEY_5: c = '5'; break;
                        case KEY_6: c = '6'; break;
                        case KEY_7: c = '7'; break;
                        case KEY_8: c = '8'; break;
                        case KEY_9: c = '9'; break;
                        case KEY_DOT: c = '.'; break;
                        case KEY_E: c = (char)24; break;
                        case KEY_DIV: c = (char)0; break;
                        case KEY_MUL: c = (char)1; break;
                        case KEY_SUB: c = '-'; break;
                        case KEY_ADD: c = '+'; break;
                        default: goto nocharkey1;
                    }
                }
                else
                {
                    switch (key)
                    {
                        case KEY_RCL: c = '%'; break;
                        case KEY_RDN: c = (char)7; break;
                        default: goto nocharkey1;
                    }
                }

                if (incomplete_length < 7)
                    incomplete_str[incomplete_length++] = c;

                // incomplete_alpha can be 0 at this point if
                // the command is CMD_LBL.
                incomplete_alpha = 1;
                redisplay();
                return;

            nocharkey1:
                // End of handling keys that represent characters

                if (shift == 0 && (key == KEY_UP || key == KEY_DOWN))
                {
                    m = menus[mode_commandmenu];
                    int nextmenu = (int)(key == KEY_UP ? m.prev : m.next);

                    if (nextmenu != (int)MENU.NONE)
                    {
                        set_menu(MENULEVEL_COMMAND, nextmenu);
                        redisplay();
                    }

                    return;
                }

                if (key == KEY_EXIT)
                {
                    if (mode_commandmenu == (int)MENU.NONE)
                    {
                        pending_command = CMD_CANCELLED;
                        finish_command_entry(false);
                        return;
                    }

                    if (mode_commandmenu == MENU_CATALOG)
                    {
                        int catsect = get_cat_section();

                        if (catsect == CATSECT_PGM
                            || catsect == CATSECT_REAL
                            || catsect == CATSECT_CPX
                            || catsect == CATSECT_MAT)
                        {
                            set_cat_section(CATSECT_TOP);
                            redisplay();
                            return;
                        }
                    }

                    m = menus[mode_commandmenu];
                    set_menu(MENULEVEL_COMMAND, (int)m.parent);

                    if (mode_commandmenu == (int)MENU.NONE)
                    {
                        pending_command = CMD_CANCELLED;
                        finish_command_entry(false);
                    }
                    else
                        redisplay();

                    return;
                }

                if (shift == 0 && key == KEY_BSP)
                {
                    if (incomplete_command >= CMD_KEY1G && incomplete_command <= CMD_KEY9X)
                    {
                        if (incomplete_length == 0)
                        {
                            if (mode_commandmenu >= MENU_ALPHA1 && mode_commandmenu <= MENU_ALPHA_MISC2)
                                start_incomplete_command(incomplete_command <= CMD_KEY9G ? CMD_KEYG : CMD_KEYX);
                            else
                            {
                                pending_command = CMD_NULL;
                                finish_command_entry(false);
                            }
                        }
                        else
                        {
                            incomplete_length--;
                            redisplay();
                        }

                        return;
                    }

                    if (incomplete_length == 0)
                    {
                        int catsect;

                        if (mode_commandmenu >= MENU_ALPHA1 && mode_commandmenu <= MENU_ALPHA_MISC2)
                        {
                            if (incomplete_command == CMD_GTODOT)
                            {
                                incomplete_argtype = ARG_OTHER;
                                incomplete_maxdigits = 4;
                                incomplete_alpha = 0;

                                set_menu(MENULEVEL_COMMAND, MENU_IND);
                                redisplay();

                                return;
                            }
                            else if (incomplete_argtype == ARG_NAMED)
                            {
                                if (incomplete_command == CMD_ASSIGNa)
                                    set_catalog_menu(CATSECT_TOP);
                                else
                                    set_catalog_menu(CATSECT_VARS_ONLY);

                                if (incomplete_ind != 0)
                                {
//                                    goto out_of_alpha;
                                    out_of_alpha();
                                    return;
                                }
                            }
                            else if (incomplete_argtype == ARG_RVAR)
                            {
                                if (incomplete_command == CMD_MVAR)
                                {
                                    pending_command = CMD_NULL;
                                    finish_command_entry(false);
                                }
                                else
                                {
//                                    goto out_of_alpha;
                                    out_of_alpha();
                                    return;
                                }
                            }
                            else if (incomplete_argtype == ARG_MAT)
                            {
                                if (vars_exist(0, 0, 1) != 0)
                                    set_catalog_menu(CATSECT_MAT_ONLY);
                                else
                                    set_menu(MENULEVEL_COMMAND, (int)MENU.NONE);
                            }
                            else if (incomplete_argtype == ARG_PRGM)
                                set_catalog_menu(CATSECT_PGM_ONLY);
                            else
                            {
//                                goto out_of_alpha;
                                out_of_alpha();
                                return;
                            }

                            redisplay();
                        }
                        else
                        {
                            catsect = get_cat_section();

                            if (mode_commandmenu == MENU_CATALOG
                                && (catsect == CATSECT_FCN
                                || catsect == CATSECT_PGM
                                || catsect == CATSECT_REAL
                                || catsect == CATSECT_CPX
                                || catsect == CATSECT_MAT))
                            {
                                set_catalog_menu(CATSECT_TOP);
                                redisplay();
                            }
                            else
                            {
                                pending_command = CMD_NULL;
                                finish_command_entry(false);
                            }
                        }

                        return;
                    }

                    incomplete_length--;

                    if (incomplete_length == 0)
                    {
                        if (incomplete_command == CMD_GTODOT)
                        {
                            incomplete_argtype = ARG_OTHER;
                            incomplete_maxdigits = 4;
                            incomplete_alpha = 0;

                            set_menu(MENULEVEL_COMMAND, MENU_IND);
                            redisplay();

                            return;
                        }
                        else if (incomplete_argtype == ARG_NAMED)
                        {
                            if (incomplete_command == CMD_ASSIGNa)
                                set_catalog_menu(CATSECT_TOP);
                            else
                                set_catalog_menu(CATSECT_VARS_ONLY);

                            if (incomplete_ind != 0)
                            {
//                                goto out_of_alpha;
                                out_of_alpha();
                                return;
                            }
                        }
                        else if (incomplete_argtype == ARG_RVAR)
                        {
                            if (vars_exist(1, 0, 0) != 0)
                                set_catalog_menu(CATSECT_REAL_ONLY);
                        }
                        else if (incomplete_argtype == ARG_MAT)
                        {
                            if (vars_exist(0, 0, 1) != 0)
                                set_catalog_menu(CATSECT_MAT_ONLY);
                        }
                        else if (incomplete_argtype == ARG_PRGM)
                            set_catalog_menu(CATSECT_PGM_ONLY);
                        else
                        {
//		            out_of_alpha:               // Replace goto out_of_alpha with call to out_of_alpha routine followed by a return (redisplay already done)
                            out_of_alpha();
                            return;
                        }
                    }

                    redisplay();
                    return;
                }

                if (key == KEY_ENTER)
                {
                    int i;

                    if (incomplete_length == 0)
                    {
                        int catsect = get_cat_section();

                        if (mode_commandmenu == (int)MENU.NONE
                            || (mode_commandmenu == MENU_CATALOG
                                && catsect != CATSECT_FCN
                                && catsect != CATSECT_PGM
                                && catsect != CATSECT_REAL
                                && catsect != CATSECT_CPX
                                && catsect != CATSECT_MAT))
                        {
                            set_menu(MENULEVEL_COMMAND, MENU_ALPHA1);
                            redisplay();
                            return;
                        }
                        else if ((incomplete_command != CMD_ASSIGNa && incomplete_command != CMD_CLP && incomplete_command != CMD_PRP)
                            || mode_commandmenu < MENU_ALPHA1 || mode_commandmenu > MENU_ALPHA_MISC2)
                        {
                            squeak();
                            return;
                        }

                        // ASSIGN, CLP, or PRP, alpha menu active, zero-length string:
                        // these are the only cases where an empty string is allowed
                        // as an argument. We fall through to the command completion
                        // code below.
                        //
                    }

                    pending_command = incomplete_command;

                    if (incomplete_ind == 0
                        && (incomplete_command == CMD_GTO
                        || incomplete_command == CMD_XEQ
                        || incomplete_command == CMD_LBL
                        || (incomplete_command >= CMD_KEY1G
                            && incomplete_command <= CMD_KEY9X))
                        && incomplete_length == 1
                        && ((incomplete_str[0] >= 'A'
                            && incomplete_str[0] <= 'J')
                        || (incomplete_str[0] >= 'a'
                            && incomplete_str[0] <= 'e')))
                    {   // Display XEQ "A" briefly before changing to XEQ A
                        mode_command_entry = false;
                        pending_command_arg.type = ARGTYPE_STR;
                        pending_command_arg.length = 1;
                        pending_command_arg.val.text[0] = incomplete_str[0];

                        if (flags.prgm_mode != 0)
                        {
                            flags.prgm_mode = 0;
                            redisplay();
                            flags.prgm_mode = 1;
                            shell_delay(125);
                        }
                        else
                            redisplay();

                        pending_command_arg.type = ARGTYPE_LCLBL;
                        pending_command_arg.val.lclbl = incomplete_str[0];
                        set_menu(MENULEVEL_COMMAND, (int)MENU.NONE);
                        finish_command_entry(false);

                        return;
                    }
                    else
                    {
                        pending_command_arg.type = (byte)(incomplete_ind != 0 ? ARGTYPE_IND_STR : ARGTYPE_STR);
                        pending_command_arg.length = (byte)incomplete_length;

                        for (i = 0; i < incomplete_length; i++)
                            pending_command_arg.val.text[i] = incomplete_str[i];

                        if (incomplete_ind == 0 && incomplete_command == CMD_XEQ)
                            finish_xeq();
                        else
                            finish_command_entry(true);

                        return;
                    }
                }

                squeak();
                return;
            }
        }

        public static void keydown_alpha_mode(int shift, int key)
        {
            int menukey;
            menu_spec m = null;
            char c;
            int command;

            if (key >= 1024)
            {
	            c = (char)(key - 1024);

//	            goto handle_char;       // Replicate handle_char code here to avoid goto
    	        if (flags.prgm_mode != 0)
                {
	                if (!mode_alpha_entry)
		                start_alpha_prgm_line();

	                if (entered_string_length < 15)
		                entered_string[entered_string_length++] = c;
	            }
                else
                {
	                if (!mode_alpha_entry)
                    {
		                reg_alpha_length = 0;
		                flags.alpha_data_input = 1;
		                mode_alpha_entry = true;
	                }
	            
                    append_alpha_char(c);

	                if (reg_alpha_length == 44)
		                squeak();
	            }

                if (core_settings.auto_repeat)
                {
	                repeating = 2;
	                repeating_key = c + 1024;
	                repeating_shift = 0;
	            }

	            if (m != null)
	                set_menu(MENULEVEL_ALPHA, (int)m.parent);

                redisplay();
	            return;
            }

            menukey = find_menu_key(key);

            if (menukey != -1)
            {
	            menu_item_spec mi;

	            m = menus[mode_alphamenu];
	            mi = m.child[menukey];

	            if (mi.menuid != MENU.NONE)
                {
	                set_menu(MENULEVEL_ALPHA, (int)mi.menuid);
	                redisplay();
	                return;
	            }

                c = mi.title[0];

	            if (shift != 0 && c >= 'A' && c <= 'Z')
	                c += (char)('a' - 'A');

//	        handle_char:
    	        if (flags.prgm_mode != 0)
                {
	                if (!mode_alpha_entry)
		                start_alpha_prgm_line();

	                if (entered_string_length < 15)
		                entered_string[entered_string_length++] = c;
	            }
                else
                {
	                if (!mode_alpha_entry)
                    {
		                reg_alpha_length = 0;
		                flags.alpha_data_input = 1;
		                mode_alpha_entry = true;
	                }
	            
                    append_alpha_char(c);

	                if (reg_alpha_length == 44)
		                squeak();
	            }

                if (core_settings.auto_repeat)
                {
	                repeating = 2;
	                repeating_key = c + 1024;
	                repeating_shift = 0;
	            }

	            if (m != null)
	                set_menu(MENULEVEL_ALPHA, (int)m.parent);

                redisplay();
	            return;
            }

            // Handle keys that represent characters
            if (shift == 0)
            {
                switch (key)
                {
	                case KEY_0:   c = '0'; break;
	                case KEY_1:   c = '1'; break;
	                case KEY_2:   c = '2'; break;
	                case KEY_3:   c = '3'; break;
	                case KEY_4:   c = '4'; break;
	                case KEY_5:   c = '5'; break;
	                case KEY_6:   c = '6'; break;
	                case KEY_7:   c = '7'; break;
	                case KEY_8:   c = '8'; break;
	                case KEY_9:   c = '9'; break;
	                case KEY_DOT: c = '.'; break;
	                case KEY_E:   c = (char)24;  break;
                    case KEY_DIV: c = (char)0; break;
                    case KEY_MUL: c = (char)1; break;
	                case KEY_SUB: c = '-'; break;
	                case KEY_ADD: c = '+'; break;
	                default: goto nocharkey2;
	            }
            }
            else
            {
	            switch (key)
                {
	                case KEY_RCL: c = '%'; break;
	                case KEY_RDN: c = (char)7;   break;
	                default: goto nocharkey2;
	            }
            }

            if (flags.prgm_mode != 0)
            {
                if (!mode_alpha_entry)
    	            start_alpha_prgm_line();
	        
                if (entered_string_length < 15)
	                entered_string[entered_string_length++] = c;
            }
            else
            {
    	        if (!mode_alpha_entry)
                {
	                reg_alpha_length = 0;
	                mode_alpha_entry = true;
	                flags.alpha_data_input = 1;
	            }
	        
                append_alpha_char(c);
	        
                if (reg_alpha_length == 44)
	                squeak();
            }

            if (core_settings.auto_repeat)
            {
	            repeating = 2;
	            repeating_key = key;
	            repeating_shift = shift;
            }

            redisplay();
            return;

        nocharkey2:
            // End of handling keys that represent characters

            if (shift == 0 && (key == KEY_UP || key == KEY_DOWN))
            {
	            m = menus[mode_alphamenu];

                int nextmenu = (int)(key == KEY_UP ? m.prev : m.next);
	        
                if (nextmenu != (int)MENU.NONE)
                {
	                set_menu(MENULEVEL_ALPHA, nextmenu);
	                redisplay();
	            }
	        
                return;
            }

            if (key == KEY_EXIT)
            {
	            m = menus[mode_alphamenu];
	        
                set_menu(MENULEVEL_ALPHA, (int)m.parent);
	        
                if (mode_alphamenu == (int)MENU.NONE)
                {
    	            if (mode_alpha_entry)
                    {
		                if (flags.prgm_mode != 0)
		                    finish_alpha_prgm_line();
		                else if ((flags.trace_print != 0 || flags.normal_print != 0) && flags.printer_exists != 0)
		                    docmd_pra(null);
		        
                        mode_alpha_entry = false;
	                }
	            
                    flags.alpha_mode = 0;
	                pending_command = CMD_CANCELLED;
	            }
                else
	                redisplay();
	        
                return;
            }

            if (shift == 0 && key == KEY_BSP)
            {
	            if (flags.prgm_mode != 0)
                {
    	            if (mode_alpha_entry)
                    {
		                if (entered_string_length > 0)
                        {
		                    if (core_settings.auto_repeat)
                            {
			                    repeating = 2;
			                    repeating_key = key;
			                    repeating_shift = shift;
        		            }
		
                            entered_string_length--;
		                }
                        else
		                    finish_alpha_prgm_line();
	                }
                    else
                    {
        		        int line = pc2line(pc);

		                if (line != 0 && (current_prgm != prgms_count - 1 || prgms[current_prgm].text[pc] != CMD_END))
                        {
		                    delete_command(pc);
		                    pc = line2pc(line - 1);
		                }
		            
                        prgm_highlight_row = 0;
		        
                        if (mode_alphamenu != MENU_ALPHA1 && mode_alphamenu != MENU_ALPHA2)
		                    set_menu(MENULEVEL_ALPHA, (int)(menus[mode_alphamenu].parent));
                    }

                    redisplay();
	                return;
                }
                else
                {
    	            if (mode_alpha_entry && reg_alpha_length > 0)
                    {
		                if (core_settings.auto_repeat)
                        {
		                    repeating = 2;
		                    repeating_key = key;
		                    repeating_shift = shift;
		                }
		        
                        reg_alpha_length--;
		                redisplay();
	                }
                    else
		                pending_command = CMD_CLA;
	            
                    return;
	            }
            }

            if (key == KEY_ENTER)
            {
	            if (flags.prgm_mode != 0)
                {
    	            if (mode_alpha_entry)
                    {
		                finish_alpha_prgm_line();
                        flags.alpha_mode = 0;

                        set_menu(MENULEVEL_ALPHA, (int)MENU.NONE);
	                }
                    else if (shift != 0)
                    {
        		        flags.alpha_mode = 0;
		        
                        set_menu(MENULEVEL_ALPHA, (int)MENU.NONE);
	                }
                    else
                    {
		                start_alpha_prgm_line();
		                entered_string[0] = (char)127;
		                entered_string_length = 1;
    	            }
	            }
                else
                {
	                if (shift != 0 || mode_alpha_entry)
                    {
		                if (mode_alpha_entry && (flags.trace_print != 0 || flags.normal_print != 0) && flags.printer_exists != 0)
		                    docmd_pra(null);
		        
                        flags.alpha_mode = 0;
		                mode_alpha_entry = false;
		        
                        set_menu(MENULEVEL_ALPHA, (int)MENU.NONE);
	                }
                    else
		                mode_alpha_entry = true;
	            }
	        
                redisplay();
	            return;
            }

            command = CMD_CANCELLED;

            if (shift == 0)
            {
	            switch (key)
                {
	                case KEY_STO: command = CMD_ASTO; break;
	                case KEY_RCL: command = CMD_ARCL; break;
	                case KEY_RUN: command = CMD_RUN; break;
	                default: command = CMD_NONE; break;
	            }
            }
            else
            {
	            switch (key)
                {
	                case KEY_CHS: set_plainmenu(MENU_MODES1); break;
	                case KEY_E: set_plainmenu(MENU_DISP); break;
	                case KEY_BSP: set_plainmenu(MENU_CLEAR1); break;
	                case KEY_7: set_solve_integ(1); break;
	                case KEY_8: set_solve_integ(0); break;
	                case KEY_9: set_menu(MENULEVEL_APP, MENU_MATRIX1); break;
	                case KEY_DIV: set_menu(MENULEVEL_APP, MENU_STAT1); break;
	                case KEY_DOWN: command = CMD_SST; break;
	                
                    case KEY_4:
                        set_menu(MENULEVEL_APP, MENU_BASE);
			            
                        if (mode_appmenu == MENU_BASE)
                        {
			                set_appmenu_exitcallback(2);
			                baseapp = 1;
			            }
			            break;
	                
                    case KEY_5: set_plainmenu(MENU_CONVERT1); break;
	                case KEY_6: set_plainmenu(MENU_FLAGS); break;
	                case KEY_MUL: set_plainmenu(MENU_PROB); break;
	                case KEY_2: set_plainmenu(MENU_CUSTOM1); break;
	                case KEY_3: set_plainmenu(MENU_PGM_FCN1); break;
	                case KEY_SUB: set_plainmenu(MENU_PRINT1); break;
	                case KEY_0: set_plainmenu(MENU_TOP_FCN); break;
	                
                    case KEY_DOT:
                        show();
			            pending_command = CMD_LINGER1;
			            shell_request_timeout3(2000);
			            return;
	                
                    case KEY_ADD: set_plainmenu(MENU_CATALOG); break;
	                default: command = CMD_NONE; break;
	            }
            }

            if (command == CMD_NONE)
            {
	            squeak();
	            return;
            }
            
            if (mode_alpha_entry)
            {
	            if (flags.prgm_mode != 0)
	                finish_alpha_prgm_line();
	            else if ((flags.trace_print != 0 || flags.normal_print != 0) && flags.printer_exists != 0)
	                docmd_pra(null);
	        
                mode_alpha_entry = false;
            }

            if (command == CMD_CANCELLED)
            {
	            // plainmenu or appmenu switch
	            flags.alpha_mode = 0;
	            set_menu(MENULEVEL_ALPHA, (int)MENU.NONE);
	            redisplay();
	            return;
            }
            else
	            do_interactive(command);
        }

        public static void keydown_normal_mode(int shift, int key)
        {
            int command;

            if (is_number_key(shift, key) != 0)
            {
	            // Entering number entry mode
	            if (deferred_print != 0)
	                print_command(CMD_NULL, null);

	            cmdline_length = 0;
	            
                if (get_front_menu() != (int)MENU.NONE)
	                cmdline_row = 0;
	            else
	                cmdline_row = 1;

	            mode_number_entry = true;
	        
                if (flags.prgm_mode != 0)
                {
    	            if (pc == -1)
	        	        pc = 0;
	                else if (prgms[current_prgm].text[pc] != CMD_END)
		                pc += get_command_length(current_prgm, pc);

                    prgm_highlight_row = 1;
		        
                    if (cmdline_row == 1)
		                display_prgm_line(0, -1);
	            }
                else
                {
    	            if (flags.stack_lift_disable == 0)
                    {
		                free_vartype(ref reg_t);
		                reg_t = reg_z;
		                reg_z = reg_y;
		                reg_y = dup_vartype(reg_x);
	                }
                    else
		                flags.stack_lift_disable = 0;
	            
                    flags.numeric_data_input = 1;
	                mode_varmenu = false;
	            
                    if (cmdline_row == 1)
		                display_y(0);
	                else
		                // Force repaint of menu; it could be hidden due to a recent
		                // two-line AVIEW command
		                redisplay();
	            }

                keydown_number_entry(shift, key);
	            return;
            }

            if (flags.prgm_mode != 0 && shift == 0 && key == KEY_BSP)
            {
	            int line = pc2line(pc);
	        
                if (line == 0)
	                return;
	        
                if (current_prgm != prgms_count - 1 || prgms[current_prgm].text[pc] != CMD_END)
	                delete_command(pc);
	        
                pc = line2pc(line - 1);
	            prgm_highlight_row = 0;
	            redisplay();
	            return;
            }

            if ((mode_appmenu != (int)MENU.NONE
		        || mode_plainmenu != (int)MENU.NONE
		        || mode_transientmenu != (int)MENU.NONE)
	            && mode_alphamenu == (int)MENU.NONE
	            && mode_commandmenu == (int)MENU.NONE)
            {
    	        int menukey = find_menu_key(key);
	            int menu;
                int level;
	        
                if (mode_transientmenu != (int)MENU.NONE)
                {
    	            menu = mode_transientmenu;
	                level = MENULEVEL_TRANSIENT;
	            }
                else if (mode_plainmenu != (int)MENU.NONE)
                {
	                menu = mode_plainmenu;
	                level = MENULEVEL_PLAIN;
	            }
                else
                {
	                menu = mode_appmenu;
	                level = MENULEVEL_APP;
	            }

                if (menu == MENU_PROGRAMMABLE)
                {
    	            int keynum;
	
                    if (menukey != -1)
		                keynum = menukey + 1;
	                else if (shift == 0)
                    {
        		        switch (key)
                        {
		                    case KEY_UP: keynum = 7; break;
		                    case KEY_DOWN: keynum = 8; break;
		                    case KEY_EXIT: keynum = 9; break;
		                    default: goto notprogmenu;
		                }
	                }
                    else
		                goto notprogmenu;
	            
                    do_prgm_menu_key(keynum);
	                return;
	            
                notprogmenu:;
	            }

                if (menu == MENU_VARMENU)
                {
    	            if (menukey != -1)
                    {
        		        if (varmenu_labellength[menukey] == 0)
                        {
        		            pending_command = CMD_NULL;
		                }
                        else if (shift != 0 && flags.prgm_mode == 0)
                        {
        		            view(varmenu_labeltext[menukey], varmenu_labellength[menukey]);
		                }
                        else
                        {
        		            int i;
		                    pending_command_arg.type = ARGTYPE_STR;
		                    pending_command_arg.length = (byte)varmenu_labellength[menukey];
		            
                            for (i = 0; i < pending_command_arg.length; i++)
			                    pending_command_arg.val.text[i] = varmenu_labeltext[menukey][i];
		            
                            if (flags.prgm_mode != 0)
                            {
            			        pending_command = shift != 0 ? CMD_VIEW : CMD_STO;
			                    store_command_after(ref pc, pending_command, pending_command_arg);
			                    prgm_highlight_row = 1;
			                    pending_command = CMD_NONE;
			                    redisplay();
		                    }
                            else
                            {
            			        switch (varmenu_role)
                                {
			                        case 0: // Plain ol' VARMENU
				                        pending_command = mode_varmenu ? CMD_VMEXEC : CMD_VMSTO;
				                        break;

                                    case 1: // Solver
				                        pending_command = mode_varmenu ? CMD_VMSOLVE : CMD_VMSTO2;
				                        break;

                                    case 2: // Integrator
				                        if (mode_varmenu)
                                        {
				                            set_integ_var(varmenu_labeltext[menukey], varmenu_labellength[menukey]);
				                            
                                            set_menu(MENULEVEL_APP, MENU_INTEG_PARAMS);
				                            set_appmenu_exitcallback(5);
				                            redisplay();
				                            
                                            return;
				                        }
                                        else
				                            pending_command = CMD_VMSTO;
                                        break;
			                    }
		                    }
                        }

                        return;
                    }

                    if (shift == 0 && key == KEY_UP)
                    {
        		        if (varmenu_rows > 1)
                        {
        		            if (++varmenu_row >= varmenu_rows)
		            	        varmenu_row = 0;
		            
                            pending_command = CMD_CANCELLED;
		                }
		        
                        return;
	                }

                    if (shift == 0 && key == KEY_DOWN)
                    {
        		        if (varmenu_rows > 1)
                        {
        		            if (--varmenu_row < 0)
		            	        varmenu_row = varmenu_rows - 1;
		            
                            pending_command = CMD_CANCELLED;
		                }
		        
                        return;
	                }

                    if (key == KEY_EXIT)
                    {
		                set_menu(MENULEVEL_APP, (int)MENU.NONE);
		                pending_command = CMD_CANCELLED;
		                return;
    	            }
                }

    	        if (menukey != -1)
                {
	                if (menu == MENU_CUSTOM1 || menu == MENU_CUSTOM2 || menu == MENU_CUSTOM3)
                    {
		                if (flags.local_label != 0)
                        {
        		            if (menukey == 5)
                            {
            			        int cmd = shift != 0 ? CMD_GTO : CMD_XEQ;
			                    do_interactive(cmd);
			                    return;
		                    }
                            else
                            {
            			        pending_command = CMD_XEQ;
			                    pending_command_arg.type = ARGTYPE_LCLBL;
			        
                                if (menu == MENU_CUSTOM1)
			                        pending_command_arg.val.lclbl = (char)(shift != 0 ? 'a' + menukey : 'A' + menukey);
                                else
            			            pending_command_arg.val.lclbl = (char)('F' + menukey);
			
                                if (flags.prgm_mode != 0)
                                {
            			            store_command_after(ref pc, pending_command, pending_command_arg);
			                        prgm_highlight_row = 1;
			                        pending_command = CMD_NONE;
			            
                                    set_menu(MENULEVEL_COMMAND, (int)MENU.NONE);
			                        redisplay();
                                }

                                return;
		                    }
                        }
                        else
                        {
        		            int keynum;
                            int length = 0;
		                    char[] name = new char[7];
		            
                            keynum = menukey + 6 * (menu - MENU_CUSTOM1) + 1;
		                    get_custom_key(keynum, name, ref length);
		            
                            if (length == 0)
			                    squeak();
		                    else
                            {
			                    int dummyprgm = 0;
			                    int dummypc = 0;
			                    int i;
			        
                                pending_command_arg.type = ARGTYPE_STR;
			                    pending_command_arg.length = (byte)length;
			        
                                for (i = 0; i < length; i++)
			                        pending_command_arg.val.text[i] = name[i];
			        
                                if (find_global_label(pending_command_arg, ref dummyprgm, ref dummypc) != 0)
			                        pending_command = CMD_XEQ;
			                    else if (lookup_var(name, length) != -1)
			                        pending_command = CMD_RCL;
                                else
                                {
			                        int cmd = find_builtin(name, length);

                                    if (cmd == -1)
				                        pending_command = CMD_XEQ;
			                        else if (cmd == CMD_CLALLa)
                                    {
                				        mode_clall = true;
				                        pending_command = CMD_NONE;
				                        redisplay();
				                        return;
                                    }
                                    else if (cmd == CMD_CLV || cmd == CMD_PRV)
                                    {
                				        if (flags.prgm_mode == 0 && vars_count == 0)
                                        {
				                            display_error(ERR_NO_VARIABLES, 0);
				                            pending_command = CMD_NONE;
				                            redisplay();
				                            return;
				                        }
				        
                                        pending_command = CMD_NONE;
				                        do_interactive(cmd);
				                        return;
			                        }
                                    else if (cmd == CMD_SST && flags.prgm_mode != 0)
                                    {
				                        sst();

                                        pending_command = CMD_NONE;
				                        redisplay();
				                        
                                        repeating = 1;
				                        repeating_shift = 1;
				                        repeating_key = KEY_DOWN;
				                        
                                        return;
			                        }
                                    else if (cmd == CMD_BST)
                                    {
				                        bst();

                                        if (flags.prgm_mode == 0)
                                        {
				                            flags.prgm_mode = 1;
				                            redisplay();
				                            flags.prgm_mode = 0;
				                            pending_command = CMD_CANCELLED;
				                        }
                                        else
                                        {
				                            redisplay();
				                            pending_command = CMD_NONE;
				                        }

				                        repeating = 1;
				                        repeating_shift = 1;
				                        repeating_key = KEY_UP;
				                        
                                        return;
                                    }
                                    else if (cmd_array[cmd].argtype == ARG_NONE)
                                    {
				                        pending_command = cmd;
				                        pending_command_arg.type = ARGTYPE_NONE;
			                        }
                                    else
                                    {
				                        pending_command = CMD_NONE;
				                        do_interactive(cmd);
				                        return;
			                        }
			                    }

//                                goto send_it_off;     // Replicate send_it_off code below to avoid goto
                                if (flags.prgm_mode != 0 && (cmd_array[pending_command].flags & FLAG_IMMED) == 0)
                                {
                                    store_command_after(ref pc, pending_command, pending_command_arg);

                                    prgm_highlight_row = 1;
                                    pending_command = CMD_NONE;

                                    if (level == MENULEVEL_TRANSIENT || (level == MENULEVEL_PLAIN && !mode_plainmenu_sticky))
                                        set_menu(level, (int)MENU.NONE);

                                    redisplay();
                                }

                                return;
                            }

                            return;
                        }
                    }
                    else if (menu == MENU_CATALOG)
                    {
            	        int catsect = get_cat_section();

		                if (catsect == CATSECT_TOP)
                        {
        		            switch (menukey)
                            {
			                    case 0:
			                        set_cat_section(CATSECT_FCN);
			                        move_cat_row(0);
			                        break;
			                    
                                case 1:
			                        set_cat_section(CATSECT_PGM);
			                        move_cat_row(0);
			                        break;

			                    case 2:
			                        if (vars_exist(1, 0, 0) != 0)
                                    {
				                        set_cat_section(CATSECT_REAL);
				                        move_cat_row(0);
			                        }
                                    else
                                    {
				                        display_error(ERR_NO_REAL_VARIABLES, 0);
				                        flush_display();
				                        return;
			                        }
			                        break;

			                    case 3:
			                        if (vars_exist(0, 1, 0) != 0)
                                    {
				                        set_cat_section(CATSECT_CPX);
				                        move_cat_row(0);
			                        }
                                    else
                                    {
				                        display_error(ERR_NO_COMPLEX_VARIABLES, 0);
				                        flush_display();
				                        return;
			                        }
			                        break;

			                    case 4:
			                        if (vars_exist(0, 0, 1) != 0)
                                    {
				                        set_cat_section(CATSECT_MAT);
				                        move_cat_row(0);
			                        }
                                    else
                                    {
				                        display_error(ERR_NO_MATRIX_VARIABLES, 0);
				                        flush_display();
				                        return;
			                        }
			                        break;

			                    case 5:
			                        display_mem();
			                        pending_command = CMD_LINGER1;
			                        shell_request_timeout3(2000);
			                        return;
                            }
		                    
                            redisplay();
                            return;
                        }
                        else if (catsect == CATSECT_PGM || catsect == CATSECT_PGM_ONLY)
                        {
		                    int labelindex = get_cat_item(menukey);
		                    
                            if (labelindex == -1)
                            {
			                    pending_command = CMD_NULL;
			                    return;
		                    }

                            if (flags.prgm_mode != 0 && labels[labelindex].length == 0)
                            {
			                    display_error(ERR_RESTRICTED_OPERATION, 0);
			                    flush_display();
			                    pending_command = CMD_NONE;
			                    return;
		                    }

		                    pending_command = CMD_XEQ;
		                    pending_command_arg.type = ARGTYPE_LBLINDEX;
		                    pending_command_arg.val.num = labelindex;

                            xeq_invisible = 1;
		                    
                            if (flags.prgm_mode == 0 && (level == MENULEVEL_TRANSIENT || !mode_plainmenu_sticky))
                            {
    			                if (level == MENULEVEL_PLAIN)
                                {
			                        int row = get_cat_row();
			                        set_menu(MENULEVEL_PLAIN, (int)MENU.NONE);
			                        set_menu(MENULEVEL_TRANSIENT, MENU_CATALOG);
			                        set_cat_section(catsect);
			                        set_cat_row(row);
			                    }
			                
                                remove_program_catalog = 1;
		                    }
                        }
                        else if (catsect == CATSECT_PGM_SOLVE || catsect == CATSECT_PGM_INTEG)
                        {
        		            int labelindex = get_cat_item(menukey);
		            
                            if (labelindex == -1)
                            {
			                    pending_command = CMD_NULL;
			                    return;
		                    }
		            
                            if (catsect == CATSECT_PGM_SOLVE)
			                    pending_command = flags.prgm_mode != 0 ? CMD_PGMSLV : CMD_PGMSLVi;
		                    else
			                    pending_command = flags.prgm_mode != 0 ? CMD_PGMINT : CMD_PGMINTi;
		            
                            pending_command_arg.type = ARGTYPE_STR;
		                    pending_command_arg.length = labels[labelindex].length;
		            
                            for (int i = 0; i < pending_command_arg.length; i++)
			                    pending_command_arg.val.text[i] = labels[labelindex].name[i];
                        }
                        else if (catsect == CATSECT_FCN)
                        {
        		            int cmd = get_cat_item(menukey);
		            
                            if (cmd == -1)
			                    cmd = CMD_NULL;
		            
                            if (level == MENULEVEL_TRANSIENT || !mode_plainmenu_sticky)
			                    set_menu(level, (int)MENU.NONE);
		            
                            do_interactive(cmd);
		                    return;
		                }
                        else
                        {
		                    int varindex = get_cat_item(menukey);

		                    if (varindex == -1)
                            {
			                    pending_command = CMD_NULL;
			                    return;
		                    }
		            
                            pending_command = CMD_RCL;
		                    pending_command_arg.type = ARGTYPE_STR;
		                    pending_command_arg.length = vars[varindex].length;
		            
                            for (int i = 0; i < pending_command_arg.length; i++)
			                    pending_command_arg.val.text[i] = vars[varindex].name[i];
		            
                            if (level == MENULEVEL_TRANSIENT || !mode_plainmenu_sticky)
			                    set_menu(level, (int)MENU.NONE);
                        }

//		        send_it_off:
                        if (flags.prgm_mode != 0 && (cmd_array[pending_command].flags & FLAG_IMMED) == 0)
                        {
		                    store_command_after(ref pc, pending_command, pending_command_arg);
		            
                            prgm_highlight_row = 1;
		                    pending_command = CMD_NONE;
		            
                            if (level == MENULEVEL_TRANSIENT || (level == MENULEVEL_PLAIN && !mode_plainmenu_sticky))
			                    set_menu(level, (int)MENU.NONE);
		            
                            redisplay();
		                }

                        return;
                    }
                    else if (menu == MENU_INTEG_PARAMS)
                    {
                        if (menukey <= 2)
                        {
		                    char[] name = new char[10];
		                    int length = 0;

		                    switch (menukey)
                            {
			                    case 0: name = "LLIM".ToCharArray(); length = 4; break;
			                    case 1: name = "ULIM".ToCharArray(); length = 4; break;
			                    case 2: name = "ACC".ToCharArray();  length = 3; break;
    		                }
	
                            
                            if (shift != 0 && flags.prgm_mode == 0)
			                    view(name, length);
		                    else
                            {
			                    pending_command_arg.type = ARGTYPE_STR;
			                    pending_command_arg.length = (byte)length;

                                for (int i = 0; i < length; i++)
			                        pending_command_arg.val.text[i] = name[i];
			        
                                if (flags.prgm_mode != 0)
                                {
            			            pending_command = shift != 0 ? CMD_VIEW : CMD_STO;
                                    store_command_after(ref pc, pending_command, pending_command_arg);

                                    prgm_highlight_row = 1;
			                        pending_command = CMD_NONE;
			            
                                    redisplay();
                                }
                                else
                                {
            			            pending_command = CMD_VMSTO;
			                    }
		                    }
                        }
                        else if (menukey == 5)
                        {
        		            int tmp = 0;
		        
                            pending_command_arg.type = ARGTYPE_STR;
		                    get_integ_var(pending_command_arg.val.text, ref tmp);
		                
                            pending_command_arg.length = (byte)tmp;
		                    pending_command = CMD_INTEG;
                        } else
                            pending_command = CMD_NULL;

                        return;
                    }
                    else
                    {
		                menu_item_spec mi = menus[menu].child[menukey];
		                int cmd_id = (int)mi.menuid;
		                
                        command_spec cmd;

		                if ((cmd_id & 0x3000) == 0)
                        {
		                    set_menu(level, cmd_id);
		                    redisplay();
		                    return;
		                }

		                if (menu == MENU_TOP_FCN && shift != 0)
                        {
		                    switch (menukey)
                            {
			                    case 0: cmd_id = CMD_SIGMASUB; break;
			                    case 1: cmd_id = CMD_Y_POW_X; break;
			                    case 2: cmd_id = CMD_SQUARE; break;
			                    case 3: cmd_id = CMD_10_POW_X; break;
			                    case 4: cmd_id = CMD_E_POW_X; break;
			                    case 5: cmd_id = CMD_GTO; break;
		                    }
		                }
                        else if (menu == MENU_PGM_FCN1 && menukey == 5 && shift != 0)
		                    cmd_id = CMD_GTO;
		                else if (menu == MENU_STAT1 && menukey == 0 && shift != 0)
		                    cmd_id = CMD_SIGMASUB;
		                else
		                    cmd_id &= 0xfff;

                        cmd = cmd_array[cmd_id];

		                if (level == MENULEVEL_TRANSIENT || (level == MENULEVEL_PLAIN && !mode_plainmenu_sticky))
		                    set_menu(level, (int)MENU.NONE);
		                
                        do_interactive(cmd_id);
		                return;
                    }
                }

                if (shift == 0 && (key == KEY_UP || key == KEY_DOWN))
                {
    	            if (menu == MENU_CATALOG)
                    {
		                move_cat_row(key == KEY_UP ? -1 : 1);
		                redisplay();
	                }
                    else if (flags.local_label != 0 && (menu == MENU_CUSTOM1 || menu == MENU_CUSTOM2 || menu == MENU_CUSTOM3))
                    {
		                set_menu(level, menu == MENU_CUSTOM1 ? MENU_CUSTOM2 : MENU_CUSTOM1);
		                redisplay();
	                }
                    else
                    {
        		        menu_spec m = menus[menu];

                        int nextmenu = (int)(key == KEY_UP ? m.prev : m.next);
		        
                        if (nextmenu != (int)MENU.NONE)
                        {
		                    set_menu(level, nextmenu);
		                    redisplay();
		                }
	                }
	            
                    return;
	            }
	
	        
                if (key == KEY_EXIT && !calc_off)
                {
	                if (menu == MENU_CATALOG)
                    {
        		        int catsect = get_cat_section();
		        
                        if (catsect == CATSECT_FCN
			                || catsect == CATSECT_PGM
			                || catsect == CATSECT_REAL
			                || catsect == CATSECT_CPX
			                || catsect == CATSECT_MAT)
		                    set_cat_section(CATSECT_TOP);
		                else
		                    set_menu(level, (int)MENU.NONE);
                    }
                    else
                    {
		                menu_spec m = menus[menu];
		                set_menu(level, (int)m.parent);
    	            }
	    
                    pending_command = CMD_CANCELLED;
	                return;
	            }
            }

            if (shift != 0 && key == KEY_ENTER)
            {
    	        if (deferred_print != 0)
	                print_command(CMD_NULL, null);
	        
                flags.alpha_mode = 1;
	            mode_alpha_entry = false;
	            set_menu(MENULEVEL_ALPHA, MENU_ALPHA1);
	            redisplay();
	            return;
            }

            if (key == KEY_EXIT && flags.prgm_mode != 0)
            {
	            flags.prgm_mode = 0;
	            pending_command = CMD_CANCELLED;
	            return;
            }

            if (key == KEY_UP)
            {
	            // Either shift == 1, or there is no menu; this
	            // means BST. This requires special care because it's
	            // one of the rare cases of auto-repeat.
	            // TO DO.
	            //
	            clear_all_rtns();
	            squeak();
	            return;
            }

            if (shift == 0)
            {
    	        switch (key)
                {
	                case KEY_SIGMA: command = CMD_SIGMAADD; break;
	                case KEY_INV: command = CMD_INV; break;
	                case KEY_SQRT: command = CMD_SQRT; break;
	                case KEY_LOG: command = CMD_LOG; break;
	                case KEY_LN: command = CMD_LN; break;
	                case KEY_XEQ: command = CMD_XEQ; break;
	                case KEY_STO: command = CMD_STO; break;
	                case KEY_RCL: command = CMD_RCL; break;
	                case KEY_RDN: command = CMD_RDN; break;
	                case KEY_SIN: command = CMD_SIN; break;
	                case KEY_COS: command = CMD_COS; break;
	                case KEY_TAN: command = CMD_TAN; break;
	                case KEY_ENTER: command = CMD_ENTER; break;
	                case KEY_SWAP: command = CMD_SWAP; break;
	                case KEY_CHS: command = basekeys() != 0 ? CMD_BASECHS : CMD_CHS; break;
	                case KEY_BSP: command = CMD_CLX; break;
	                case KEY_DIV: command = basekeys() != 0 ? CMD_BASEDIV : CMD_DIV; break;
	                case KEY_DOWN: command = CMD_SST; break;
	                case KEY_MUL: command = basekeys() != 0 ? CMD_BASEMUL : CMD_MUL; break;
	                case KEY_SUB: command = basekeys() != 0 ? CMD_BASESUB : CMD_SUB; break;
	                case KEY_EXIT:
		                input_length = 0;
		                pending_command = CMD_CANCELLED;
		                return;
	                case KEY_RUN: command = CMD_RUN; break;
	                case KEY_ADD: command = basekeys() != 0 ? CMD_BASEADD : CMD_ADD; break;
	                default: command = CMD_NONE; break;
	            }
            }
            else
            {
	            switch (key)
                {
	                case KEY_SIGMA: command = CMD_SIGMASUB; break;
	                case KEY_INV: command = CMD_Y_POW_X; break;
	                case KEY_SQRT: command = CMD_SQUARE; break;
	                case KEY_LOG: command = CMD_10_POW_X; break;
	                case KEY_LN: command = CMD_E_POW_X; break;
	                case KEY_XEQ: command = CMD_GTO; break;
	                case KEY_STO: command = CMD_COMPLEX; break;
	                case KEY_RCL: command = CMD_PERCENT; break;
	                case KEY_RDN: command = CMD_PI; break;
	                case KEY_SIN: command = CMD_ASIN; break;
	                case KEY_COS: command = CMD_ACOS; break;
	                case KEY_TAN: command = CMD_ATAN; break;
	                case KEY_SWAP: command = CMD_LASTX; break;
	                case KEY_CHS: set_plainmenu(MENU_MODES1); return;
	                case KEY_E: set_plainmenu(MENU_DISP); return;
	                case KEY_BSP: set_plainmenu(MENU_CLEAR1); return;
	                case KEY_7: set_solve_integ(1); redisplay(); return;
	                case KEY_8: set_solve_integ(0); redisplay(); return;
	                case KEY_9:
                        set_menu(MENULEVEL_APP, MENU_MATRIX1);
			            redisplay();
			            return;
	                case KEY_DIV:
                        set_menu(MENULEVEL_APP, MENU_STAT1);
			            redisplay();
			            return;
	                case KEY_DOWN: command = CMD_SST; break;
	                case KEY_4:
                        set_menu(MENULEVEL_APP, MENU_BASE);
			            if (mode_appmenu == MENU_BASE)
                        {
			                set_appmenu_exitcallback(2);
			                baseapp = 1;
			                redisplay();
			            }
			            return;
	                case KEY_5: set_plainmenu(MENU_CONVERT1); return;
                    case KEY_6: set_plainmenu(MENU_FLAGS); return;
                    case KEY_MUL: set_plainmenu(MENU_PROB); return;
	                case KEY_1: command = CMD_ASSIGNa; break;
                    case KEY_2: set_plainmenu(MENU_CUSTOM1); return;
                    case KEY_3: set_plainmenu(MENU_PGM_FCN1); return;
                    case KEY_SUB: set_plainmenu(MENU_PRINT1); return;
                    case KEY_DOT:
                        show();
			            pending_command = CMD_LINGER1;
			            shell_request_timeout3(2000);
			            return;
                    case KEY_0: set_plainmenu(MENU_TOP_FCN); return;
                    case KEY_ADD: set_plainmenu(MENU_CATALOG); return;
	                default: command = CMD_NONE; break;
	            }
            }

            if (command == CMD_NONE)
	            return;
            else
	            do_interactive(command);
        }

        public static void view(char[] varname, int varlength)
        {
            arg_struct arg = new arg_struct();
            int i;
            int err;
            
            arg.type = ARGTYPE_STR;
            arg.length = (byte)varlength;

            for (i = 0; i < varlength; i++)
	            arg.val.text[i] = varname[i];
            
            err = view_helper(arg, 0);
            
            if (err != ERR_NONE)
            {
	            display_error(err, 1);
	            flush_display();
	            
                pending_command = CMD_NONE;
            }
            else
            {
	            pending_command = CMD_LINGER1;
	            shell_request_timeout3(2000);
            }
        }

        public static void set_solve_integ(int solve)
        {
            if (flags.prgm_mode != 0 || mvar_prgms_exist() == 0)
            {
                set_menu(MENULEVEL_APP, solve != 0 ? MENU_SOLVE : MENU_INTEG);

                if (flags.prgm_mode == 0)
                    display_error(ERR_NO_MENU_VARIABLES, 0);
            }
            else
            {
                int err = set_menu_return_err(MENULEVEL_APP, MENU_CATALOG);

                if (err == ERR_NONE)
                {
                    set_cat_section(solve != 0 ? CATSECT_PGM_SOLVE : CATSECT_PGM_INTEG);
                    move_cat_row(0);
                    clear_row(0);
                    
                    if (solve != 0)
                        draw_string(0, 0, "Select Solve Program");
                    else
                        draw_string(0, 0, "Select \x0003f(x) Program");
                    
                    flags.message = 1;
                    flags.two_line_message = 0;
                }
                else
                    display_error(err, 1);
            }

            redisplay();
        }
    }
}
