﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        public static int div_rr_completion1(int error, ref vartype_realmatrix a, ref int[] perm, Phloat det)
        {
            if (error != ERR_NONE)
            {
                vartype _vt = (vartype)a;
                a = null;
                free_vartype(ref _vt);
                
                perm = null;
                free_vartype(ref linalg_div_result);
                
                return error;
            }
            else
            {
                matrix_copy(ref linalg_div_result, linalg_div_left);
                
                return lu_backsubst_rr(a, perm, (vartype_realmatrix)linalg_div_result, div_rr_completion2);
            }
        }

        public static void div_rr_completion2(int error, ref vartype_realmatrix a, ref int[] perm, vartype_realmatrix b)
        {
            if (error != ERR_NONE)
                free_vartype(ref linalg_div_result); // Note: linalg_div_result == b

            vartype _vt = (vartype)a;
            a = null;
            free_vartype(ref _vt);

            perm = null;
            
            linalg_div_completion(error, linalg_div_result);
        }

        public static int div_rc_completion1(int error, ref vartype_complexmatrix a, ref int[] perm, Phloat det_re, Phloat det_im)
        {
            if (error != ERR_NONE)
            {
                vartype _vt = (vartype)a;
                a = null;
                free_vartype(ref _vt);
                
                perm = null;
                free_vartype(ref linalg_div_result);
                
                return error;
            }
            else
            {
                matrix_copy(ref linalg_div_result, linalg_div_left);
                
                return lu_backsubst_cc(a, perm, (vartype_complexmatrix)linalg_div_result, div_rc_completion2);
            }
        }

        public static void div_rc_completion2(int error, ref vartype_complexmatrix a, ref int[] perm, vartype_complexmatrix b)
        {
            if (error != ERR_NONE)
                free_vartype(ref linalg_div_result); // Note: linalg_div_result == b

            vartype _vt = (vartype)a;
            a = null;
            free_vartype(ref _vt);

            perm = null;

            linalg_div_completion(error, linalg_div_result);
        }

        public static int div_cr_completion1(int error, ref vartype_realmatrix a, ref int[] perm, Phloat det)
        {
            if (error != ERR_NONE)
            {
                vartype _vt = (vartype)a;
                a = null;
                free_vartype(ref _vt);

                perm = null;
                free_vartype(ref linalg_div_result);
                
                return error;
            }
            else
            {
                matrix_copy(ref linalg_div_result, linalg_div_left);
                
                return lu_backsubst_rc(a, perm, (vartype_complexmatrix)linalg_div_result, div_cr_completion2);
            }
        }

        public static void div_cr_completion2(int error, ref vartype_realmatrix a, ref int[] perm, vartype_complexmatrix b)
        {
            if (error != ERR_NONE)
                free_vartype(ref linalg_div_result); // Note: linalg_div_result == b

            vartype _vt = (vartype)a;
            a = null;
            free_vartype(ref _vt);

            perm = null;

            linalg_div_completion(error, linalg_div_result);
        }

        public static int div_cc_completion1(int error, ref vartype_complexmatrix a, ref int[] perm, Phloat det_re, Phloat det_im)
        {
            if (error != ERR_NONE)
            {
                vartype _vt = (vartype)a;
                a = null;
                free_vartype(ref _vt);

                perm = null;
                free_vartype(ref linalg_div_result);

                return error;
            }
            else
            {
                matrix_copy(ref linalg_div_result, linalg_div_left);

                return lu_backsubst_cc(a, perm, (vartype_complexmatrix)linalg_div_result, div_cc_completion2);
            }
        }

        public static void div_cc_completion2(int error, ref vartype_complexmatrix a, ref int[] perm, vartype_complexmatrix b)
        {
            if (error != ERR_NONE)
                free_vartype(ref linalg_div_result); // Note: linalg_div_result == b

            vartype _vt = (vartype)a;
            a = null;
            free_vartype(ref _vt);

            perm = null;

            linalg_div_completion(error, linalg_div_result);
        }

        public static int linalg_div(vartype left, vartype right, fp_completion completion)
        {
            if (left.type == TYPE.REALMATRIX)
            {
	            if (right.type == TYPE.REALMATRIX)
                {
	                vartype_realmatrix num = (vartype_realmatrix) left;
	                vartype_realmatrix denom = (vartype_realmatrix) right;
	                vartype lu;
                    vartype res;

	                int rows = num.rows;
	                int columns = num.columns;
	                
                    int[] perm;
	                
                    if (denom.rows != rows || denom.columns != rows)
                    {
		                completion(ERR_DIMENSION_ERROR, null);
		                return ERR_DIMENSION_ERROR;
	                }

	                perm = new int[rows];

	                if (perm == null)
                    {
		                completion(ERR_INSUFFICIENT_MEMORY, null);
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                lu = new_realmatrix(rows, rows);
	                
                    if (lu == null)
                    {
		                perm = null;
		                completion(ERR_INSUFFICIENT_MEMORY, null);
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                res = new_realmatrix(rows, columns);
	                
                    if (res == null)
                    {
		                perm = null;
		                free_vartype(ref lu);
		                completion(ERR_INSUFFICIENT_MEMORY, null);
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                matrix_copy(ref lu, right);

	                linalg_div_completion = completion;
	                linalg_div_left = left;
	                linalg_div_result = res;
	                
                    return lu_decomp_r((vartype_realmatrix) lu, perm, div_rr_completion1); 
	            }
                else
                {
	                vartype_realmatrix num = (vartype_realmatrix) left;
	                vartype_complexmatrix denom = (vartype_complexmatrix) right;
	                vartype lu;
                    vartype res;

	                int rows = num.rows;
	                int columns = num.columns;
	                
                    int[] perm;
	                
                    if (denom.rows != rows || denom.columns != rows)
                    {
		                completion(ERR_DIMENSION_ERROR, null);
		                return ERR_DIMENSION_ERROR;
	                }

	                perm = new int[rows];

	                if (perm == null)
                    {
		                completion(ERR_INSUFFICIENT_MEMORY, null);
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                lu = new_complexmatrix(rows, rows);
	                
                    if (lu == null)
                    {
		                perm = null;
		                completion(ERR_INSUFFICIENT_MEMORY, null);
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                res = new_complexmatrix(rows, columns);
	                
                    if (res == null)
                    {
		                perm = null;
		                free_vartype(ref lu);
		                completion(ERR_INSUFFICIENT_MEMORY, null);
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                matrix_copy(ref lu, right);

	                linalg_div_completion = completion;
	                linalg_div_left = left;
	                linalg_div_result = res;
	                
                    return lu_decomp_c((vartype_complexmatrix) lu, perm, div_rc_completion1);
	            }
            }
            else
            {
	            if (right.type == TYPE.REALMATRIX)
                {
	                vartype_complexmatrix num = (vartype_complexmatrix) left;
	                vartype_realmatrix denom = (vartype_realmatrix) right;
	                vartype lu;
                    vartype res;

	                int rows = num.rows;
	                int columns = num.columns;
	                
                    int[] perm;
	                
                    if (denom.rows != rows || denom.columns != rows)
                    {
		                completion(ERR_DIMENSION_ERROR, null);
		                return ERR_DIMENSION_ERROR;
	                }

	                perm = new int[rows];

	                if (perm == null)
                    {
		                completion(ERR_INSUFFICIENT_MEMORY, null);
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                lu = new_realmatrix(rows, rows);
	                
                    if (lu == null)
                    {
		                perm = null;
		                completion(ERR_INSUFFICIENT_MEMORY, null);
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                res = new_complexmatrix(rows, columns);
	                
                    if (res == null)
                    {
		                perm = null;
		                free_vartype(ref lu);
		                completion(ERR_INSUFFICIENT_MEMORY, null);
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                matrix_copy(ref lu, right);
	                
                    linalg_div_completion = completion;
	                linalg_div_left = left;
	                linalg_div_result = res;
	                
                    return lu_decomp_r((vartype_realmatrix) lu, perm, div_cr_completion1);
	            }
                else
                {
	                vartype_complexmatrix num = (vartype_complexmatrix) left;
	                vartype_complexmatrix denom = (vartype_complexmatrix) right;
	                vartype lu;
                    vartype res;

	                int rows = num.rows;
	                int columns = num.columns;
	                
                    int[] perm;
	                
                    if (denom.rows != rows || denom.columns != rows)
                    {
		                completion(ERR_DIMENSION_ERROR, null);
		                return ERR_DIMENSION_ERROR;
	                }

	                perm = new int[rows];
	                
                    if (perm == null)
                    {
		                completion(ERR_INSUFFICIENT_MEMORY, null);
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                lu = new_complexmatrix(rows, rows);
	                
                    if (lu == null)
                    {
		                perm = null;
		                completion(ERR_INSUFFICIENT_MEMORY, null);
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                res = new_complexmatrix(rows, columns);
	                
                    if (res == null)
                    {
		                perm = null;
		                free_vartype(ref lu);
		                completion(ERR_INSUFFICIENT_MEMORY, null);
		                return ERR_INSUFFICIENT_MEMORY;
	                }

	                matrix_copy(ref lu, right);
	                
                    linalg_div_completion = completion;
	                linalg_div_left = left;
	                linalg_div_result = res;
	                
                    return lu_decomp_c((vartype_complexmatrix) lu, perm, div_cc_completion1);
	            }
            }
        }

        public static int linalg_mul(vartype left, vartype right, fp_completion completion)
        {
            if (left.type == TYPE.REALMATRIX)
            {
	            if (right.type == TYPE.REALMATRIX)
	                return matrix_mul_rr((vartype_realmatrix) left, (vartype_realmatrix) right, completion);
	            else
	                return matrix_mul_rc((vartype_realmatrix) left, (vartype_complexmatrix) right, completion);
            }
            else
            {
	            if (right.type == TYPE.REALMATRIX)
	                return matrix_mul_cr((vartype_complexmatrix) left, (vartype_realmatrix) right, completion);
	            else
	                return matrix_mul_cc((vartype_complexmatrix) left, (vartype_complexmatrix) right, completion);
            }
        }

        public static int matrix_mul_rr(vartype_realmatrix left, vartype_realmatrix right, fp_completion completion)
        {
            mul_rr_data_struct dat;
            int error;

            if (left.columns != right.rows)
            {
	            error = ERR_DIMENSION_ERROR;
	            goto finished;
            }

            if (contains_no_strings(left) == 0 || contains_no_strings(right) == 0)
            {
	            error = ERR_ALPHA_DATA_IS_INVALID;
	            goto finished;
            }

            dat = new mul_rr_data_struct();

            if (dat == null)
            {
	            error = ERR_INSUFFICIENT_MEMORY;
	            goto finished;
            }

            dat.result = new_realmatrix(left.rows, right.columns);
            
            if (dat.result == null)
            {
	            dat = null;
	            error = ERR_INSUFFICIENT_MEMORY;
	            goto finished;
            }

            dat.left = left;
            dat.right = right;
            dat.i = 0;
            dat.j = 0;
            dat.k = 0;
            dat.sum = 0;
            dat.completion = completion;

            mul_rr_data = dat;
            mode_interruptible = matrix_mul_rr_worker;
            mode_stoppable = false;
            
            return ERR_INTERRUPTIBLE;

            finished:
            completion(error, null);
            return error;
        }

        public static int matrix_mul_rr_worker(int interrupted)
        {
            mul_rr_data_struct dat = mul_rr_data;
            
            int count = 0;
            int inf;
            
            Phloat[] l = dat.left.array.data;
            Phloat[] r = dat.right.array.data;
            Phloat[] p = ((vartype_realmatrix) dat.result).array.data;
            
            int i = dat.i;
            int j = dat.j;
            int k = dat.k;
            int m = dat.left.rows;
            int n = dat.right.columns;
            int q = dat.left.columns;
            
            Phloat sum = dat.sum;

            if (interrupted != 0)
            {
	            dat.completion(ERR_INTERRUPTED, null);
	            free_vartype(ref dat.result);
	            dat = null;
	            return ERR_INTERRUPTED;
            }

            while (count++ < 1000)
            {
	            sum += l[i * q + k] * r[k * n + j];
	        
                if (++k < q)
	                continue;
	        
                k = 0;
	        
                if ((inf = Phloat.p_isinf(sum)) != 0)
                {
	                if (core_settings.matrix_outofrange && flags.range_error_ignore == 0)
                    {
		                dat.completion(ERR_OUT_OF_RANGE, null);
		                free_vartype(ref dat.result);
		                dat = null;
		                return ERR_OUT_OF_RANGE;
	                }
                    else
                        sum = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            }
	        
                p[i * n + j] = sum;
	            sum = 0;
	        
                if (++j < n)
	                continue;
	        
                j = 0;
	        
                if (++i < m)
	                continue;
	            else
                {
	                dat.completion(ERR_NONE, dat.result);
	                dat = null;
	                return ERR_NONE;
	            }
            }

            dat.i = i;
            dat.j = j;
            dat.k = k;
            dat.sum = sum;

            return ERR_INTERRUPTIBLE;
        }

        public static int matrix_mul_cc(vartype_complexmatrix left, vartype_complexmatrix right, fp_completion completion)
        {
            mul_cc_data_struct dat;
            int error;

            if (left.columns != right.rows)
            {
	            error = ERR_DIMENSION_ERROR;
	            goto finished;
            }

            dat = new mul_cc_data_struct();

            if (dat == null)
            {
	            error = ERR_INSUFFICIENT_MEMORY;
	            goto finished;
            }

            dat.result = new_complexmatrix(left.rows, right.columns);
            
            if (dat.result == null)
            {
	            dat = null;
	            error = ERR_INSUFFICIENT_MEMORY;
	            goto finished;
            }

            dat.left = left;
            dat.right = right;
            dat.i = 0;
            dat.j = 0;
            dat.k = 0;
            dat.sum_re = 0;
            dat.sum_im = 0;
            dat.completion = completion;

            mul_cc_data = dat;
            mode_interruptible = matrix_mul_cc_worker;
            mode_stoppable = false;

            return ERR_INTERRUPTIBLE;

            finished:
            completion(error, null);
            return error;
        }

        public static int matrix_mul_cc_worker(int interrupted)
        {
            mul_cc_data_struct dat = mul_cc_data;

            int count = 0;
            int inf;
            
            Phloat[] l = dat.left.array.data;
            Phloat[] r = dat.right.array.data;
            Phloat[] p = ((vartype_complexmatrix) dat.result).array.data;
            
            int i = dat.i;
            int j = dat.j;
            int k = dat.k;
            int m = dat.left.rows;
            int n = dat.right.columns;
            int q = dat.left.columns;
            
            Phloat sum_re = dat.sum_re;
            Phloat sum_im = dat.sum_im;

            if (interrupted != 0)
            {
	            dat.completion(ERR_INTERRUPTED, null);
	            free_vartype(ref dat.result);
	            dat = null;
	            return ERR_INTERRUPTED;
            }

            while (count++ < 1000)
            {
	            Phloat l_re = l[2 * (i * q + k)];
	            Phloat l_im = l[2 * (i * q + k) + 1];
	            Phloat r_re = r[2 * (k * n + j)];
	            Phloat r_im = r[2 * (k * n + j) + 1];
	            
                sum_re += l_re * r_re - l_im * r_im;
	            sum_im += l_im * r_re + l_re * r_im;
	        
                if (++k < q)
	                continue;
	        
                k = 0;
	        
                if ((inf = Phloat.p_isinf(sum_re)) != 0)
                {
	                if (core_settings.matrix_outofrange && flags.range_error_ignore == 0)
                    {
		                dat.completion(ERR_OUT_OF_RANGE, null);
		                free_vartype(ref dat.result);
		                dat = null;
		                return ERR_OUT_OF_RANGE;
	                }
                    else
		                sum_re = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            }

                if ((inf = Phloat.p_isinf(sum_im)) != 0)
                {
	                if (core_settings.matrix_outofrange && flags.range_error_ignore == 0)
                    {
		                dat.completion(ERR_OUT_OF_RANGE, null);
		                free_vartype(ref dat.result);
		                dat = null;
		                return ERR_OUT_OF_RANGE;
	                }
                    else
		                sum_im = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            }
	            
                p[2 * (i * n + j)] = sum_re;
	            p[2 * (i * n + j) + 1] = sum_im;
	            sum_re = 0;
	            sum_im = 0;
	            
                if (++j < n)
	                continue;
	            
                j = 0;
	            
                if (++i < m)
	                continue;
	            else
                {
	                dat.completion(ERR_NONE, dat.result);
	                dat = null;
	                return ERR_NONE;
	            }
            }

            dat.i = i;
            dat.j = j;
            dat.k = k;
            dat.sum_re = sum_re;
            dat.sum_im = sum_im;
            
            return ERR_INTERRUPTIBLE;
        }

        public static int matrix_mul_cr(vartype_complexmatrix left, vartype_realmatrix right, fp_completion completion)
        {
            mul_cr_data_struct dat;
            int error;

            if (left.columns != right.rows)
            {
	            error = ERR_DIMENSION_ERROR;
	            goto finished;
            }

            if (contains_no_strings(right) == 0)
            {
	            error = ERR_ALPHA_DATA_IS_INVALID;
	            goto finished;
            }

            dat = new mul_cr_data_struct();

            if (dat == null)
            {
	            error = ERR_INSUFFICIENT_MEMORY;
	            goto finished;
            }

            dat.result = new_complexmatrix(left.rows, right.columns);
            
            if (dat.result == null)
            {
	            dat = null;
	            error = ERR_INSUFFICIENT_MEMORY;
	            goto finished;
            }

            dat.left = left;
            dat.right = right;
            dat.i = 0;
            dat.j = 0;
            dat.k = 0;
            dat.sum_re = 0;
            dat.sum_im = 0;
            dat.completion = completion;

            mul_cr_data = dat;
            mode_interruptible = matrix_mul_cr_worker;
            mode_stoppable = false;
            
            return ERR_INTERRUPTIBLE;

            finished:
            completion(error, null);
            return error;
        }

        public static int matrix_mul_cr_worker(int interrupted)
        {
            mul_cr_data_struct dat = mul_cr_data;

            int count = 0;
            int inf;
            
            Phloat[] l = dat.left.array.data;
            Phloat[] r = dat.right.array.data;
            Phloat[] p = ((vartype_complexmatrix) dat.result).array.data;
            
            int i = dat.i;
            int j = dat.j;
            int k = dat.k;
            int m = dat.left.rows;
            int n = dat.right.columns;
            int q = dat.left.columns;
            
            Phloat sum_re = dat.sum_re;
            Phloat sum_im = dat.sum_im;

            if (interrupted != 0)
            {
	            dat.completion(ERR_INTERRUPTED, null);
	            free_vartype(ref dat.result);
	            dat = null;
	            return ERR_INTERRUPTED;
            }

            while (count++ < 1000)
            {
	            Phloat tmp = r[k * n + j];
	            sum_re += tmp * l[2 * (i * q + k)];
	            sum_im += tmp * l[2 * (i * q + k) + 1];
	            
                if (++k < q)
	                continue;
	            
                k = 0;
	            
                if ((inf = Phloat.p_isinf(sum_re)) != 0)
                {
	                if (core_settings.matrix_outofrange && flags.range_error_ignore == 0)
                    {
		                dat.completion(ERR_OUT_OF_RANGE, null);
		                free_vartype(ref dat.result);
		                dat = null;
		                return ERR_OUT_OF_RANGE;
	                }
                    else
		                sum_re = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            }

	            if ((inf = Phloat.p_isinf(sum_im)) != 0)
                {
	                if (core_settings.matrix_outofrange && flags.range_error_ignore == 0)
                    {
		                dat.completion(ERR_OUT_OF_RANGE, null);
		                free_vartype(ref dat.result);
		                dat = null;
		                return ERR_OUT_OF_RANGE;
	                }
                    else
		                sum_im = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            }

                p[2 * (i * n + j)] = sum_re;
	            p[2 * (i * n + j) + 1] = sum_im;
	            sum_re = 0;
	            sum_im = 0;
	            
                if (++j < n)
	                continue;
	            
                j = 0;
	            
                if (++i < m)
	                continue;
	            else
                {
	                dat.completion(ERR_NONE, dat.result);
	                dat = null;
	                return ERR_NONE;
	            }
            }

            dat.i = i;
            dat.j = j;
            dat.k = k;
            dat.sum_re = sum_re;
            dat.sum_im = sum_im;
            return ERR_INTERRUPTIBLE;
        }

        public static int matrix_mul_rc(vartype_realmatrix left, vartype_complexmatrix right, fp_completion completion)
        {
            mul_rc_data_struct dat;
            int error;

            if (left.columns != right.rows)
            {
	            error = ERR_DIMENSION_ERROR;
	            goto finished;
            }

            if (contains_no_strings(left) == 0)
            {
	            error = ERR_ALPHA_DATA_IS_INVALID;
	            goto finished;
            }

            dat = new mul_rc_data_struct();

            if (dat == null)
            {
	            error = ERR_INSUFFICIENT_MEMORY;
	            goto finished;
            }

            dat.result = new_complexmatrix(left.rows, right.columns);
            
            if (dat.result == null)
            {
	            dat = null;
	            error = ERR_INSUFFICIENT_MEMORY;
	            goto finished;
            }

            dat.left = left;
            dat.right = right;
            dat.i = 0;
            dat.j = 0;
            dat.k = 0;
            dat.sum_re = 0;
            dat.sum_im = 0;
            dat.completion = completion;

            mul_rc_data = dat;
            mode_interruptible = matrix_mul_rc_worker;
            mode_stoppable = false;
            
            return ERR_INTERRUPTIBLE;

            finished:
            completion(error, null);
            return error;
        }

        public static int matrix_mul_rc_worker(int interrupted)
        {
            mul_rc_data_struct dat = mul_rc_data;
            
            int count = 0;
            int inf;
            
            Phloat[] l = dat.left.array.data;
            Phloat[] r = dat.right.array.data;
            Phloat[] p = ((vartype_complexmatrix) dat.result).array.data;
            
            int i = dat.i;
            int j = dat.j;
            int k = dat.k;
            int m = dat.left.rows;
            int n = dat.right.columns;
            int q = dat.left.columns;
            
            Phloat sum_re = dat.sum_re;
            Phloat sum_im = dat.sum_im;

            if (interrupted != 0)
            {
	            dat.completion(ERR_INTERRUPTED, null);
	            free_vartype(ref dat.result);
	            dat = null;
	            return ERR_INTERRUPTED;
            }

            while (count++ < 1000)
            {
	            Phloat tmp = l[i * q + k];
	            sum_re += tmp * r[2 * (k * n + j)];
	            sum_im += tmp * r[2 * (k * n + j) + 1];
	            
                if (++k < q)
	                continue;
	            
                k = 0;

                if ((inf = Phloat.p_isinf(sum_re)) != 0)
                {
	                if (core_settings.matrix_outofrange && flags.range_error_ignore == 0)
                    {
		                dat.completion(ERR_OUT_OF_RANGE, null);
		                free_vartype(ref dat.result);
		                dat = null;
		                return ERR_OUT_OF_RANGE;
	                }
                    else
                        sum_re = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            }

                if ((inf = Phloat.p_isinf(sum_im)) != 0)
                {
	                if (core_settings.matrix_outofrange && flags.range_error_ignore == 0)
                    {
		                dat.completion(ERR_OUT_OF_RANGE, null);
		                free_vartype(ref dat.result);
		                dat = null;
		                return ERR_OUT_OF_RANGE;
	                }
                    else
                        sum_im = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	            }
	            
                p[2 * (i * n + j)] = sum_re;
	            p[2 * (i * n + j) + 1] = sum_im;
	            sum_re = 0;
	            sum_im = 0;
	            
                if (++j < n)
	                continue;
	            
                j = 0;
	            
                if (++i < m)
	                continue;
	            else
                {
	                dat.completion(ERR_NONE, dat.result);
	                dat = null;
	                return ERR_NONE;
	            }
            }

            dat.i = i;
            dat.j = j;
            dat.k = k;
            dat.sum_re = sum_re;
            dat.sum_im = sum_im;
            
            return ERR_INTERRUPTIBLE;
        }

        public static int linalg_det(vartype src, fp_completion completion)
        {
            int n;
            int[] perm;
            
            if (src.type == TYPE.REALMATRIX)
            {
	            vartype_realmatrix ma = (vartype_realmatrix) src;
	            n = ma.rows;
	        
                if (n != ma.columns)
                {
	                completion(ERR_DIMENSION_ERROR, null);
	                return ERR_DIMENSION_ERROR;
	            }
	        
                if (contains_no_strings(ma) == 0)
                {
	                completion(ERR_ALPHA_DATA_IS_INVALID, null);
	                return ERR_ALPHA_DATA_IS_INVALID;
	            }
	        
                ma = (vartype_realmatrix) dup_vartype(src);
	        
                if (ma == null)
                {
	                completion(ERR_INSUFFICIENT_MEMORY, null);
	                return ERR_INSUFFICIENT_MEMORY;
	            }
	        
                if (disentangle((vartype) ma) == 0)
                {
	                vartype tmp = (vartype) ma;
                    free_vartype(ref tmp);

	                completion(ERR_INSUFFICIENT_MEMORY, null);
	                return ERR_INSUFFICIENT_MEMORY;
	            }
	        
                perm = new int[n];
	        
                if (perm == null)
                {
                    vartype tmp = (vartype)ma;
	                free_vartype(ref tmp);

	                completion(ERR_INSUFFICIENT_MEMORY, null);
	                return ERR_INSUFFICIENT_MEMORY;
	            }

	            // Before calling lu_decomp_r, make sure the 'singular matrix'
	            // error reporting mode is on; we don't want the HP-42S compatible
	            // zero-pivot-fudging to take place when all we're doing is computing
	            // the determinant.
	            // The completion routine will restore the 'singular matrix' error
	            // mode to its original value.
	            //
	            linalg_det_prev_sm_err = core_settings.matrix_singularmatrix;
	            core_settings.matrix_singularmatrix = true;

	            linalg_det_completion = completion;
	            return lu_decomp_r(ma, perm, det_r_completion); 
            }
            else // src->type == TYPE_COMPLEXMATRIX
            {
	            vartype_complexmatrix ma = (vartype_complexmatrix) src;
	        
                n = ma.rows;
	        
                if (n != ma.columns)
	                return ERR_DIMENSION_ERROR;
	        
                ma = (vartype_complexmatrix) dup_vartype(src);
	        
                if (ma == null)
	                return ERR_INSUFFICIENT_MEMORY;
	        
                if (disentangle((vartype) ma) == 0)
                {
                    vartype tmp = (vartype)ma;
                    free_vartype(ref tmp);

	                return ERR_INSUFFICIENT_MEMORY;
	            }
	        
                n = ma.rows;
                perm = new int[n];

                if (perm == null)
                {
                    vartype tmp = (vartype)ma;
	                free_vartype(ref tmp);

	                return ERR_INSUFFICIENT_MEMORY;
	            }

	            // Before calling lu_decomp_c, make sure the 'singular matrix'
	            // error reporting mode is on; we don't want the HP-42S compatible
	            // zero-pivot-fudging to take place when all we're doing is computing
	            // the determinant.
	            // The completion routine will restore the 'singular matrix' error
	            // mode to its original value.
	            //
	            linalg_det_prev_sm_err = core_settings.matrix_singularmatrix;
	            core_settings.matrix_singularmatrix = true;

	            linalg_det_completion = completion;
	            return lu_decomp_c(ma, perm, det_c_completion); 
            }
        }

        public static int det_r_completion(int error, ref vartype_realmatrix a, ref int[] perm, Phloat det)
        {
            vartype det_v = new vartype();

            core_settings.matrix_singularmatrix = linalg_det_prev_sm_err;

            vartype tmp = (vartype)a;
            free_vartype(ref tmp);
            perm = null;

            if (error == ERR_SINGULAR_MATRIX)
            {
                det = 0;
                error = ERR_NONE;
            }
            
            if (error == ERR_NONE)
            {
                int inf = Phloat.p_isinf(det);
                
                if (inf != 0)
                {
                    if (flags.range_error_ignore != 0)
                        det = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                    else
                        error = ERR_OUT_OF_RANGE;
                }
            }
            
            if (error == ERR_NONE)
            {
                det_v = new_real(det);
            
                if (det_v == null)
                    error = ERR_INSUFFICIENT_MEMORY;
            }

            linalg_det_completion(error, det_v);
            return error;
        }

        public static int det_c_completion(int error, ref vartype_complexmatrix a, ref int[] perm, Phloat det_re, Phloat det_im)
        {
            vartype det_v = new vartype();

            core_settings.matrix_singularmatrix = linalg_det_prev_sm_err;

            vartype tmp = (vartype)a;
            free_vartype(ref tmp);
            perm = null;

            if (error == ERR_SINGULAR_MATRIX)
            {
                det_re = 0;
                det_im = 0;
                error = ERR_NONE;
            }
            
            if (error == ERR_NONE)
            {
                int inf;
                if ((inf = Phloat.p_isinf(det_re)) != 0)
                {
                    if (flags.range_error_ignore != 0)
                        det_re = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                    else
                        error = ERR_OUT_OF_RANGE;
                }

                if ((inf = Phloat.p_isinf(det_im)) != 0)
                {
                    if (flags.range_error_ignore != 0)
                        det_im = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                    else
                        error = ERR_OUT_OF_RANGE;
                }
            }
            
            if (error == ERR_NONE)
            {
                det_v = new_complex(det_re, det_im);
            
                if (det_v == null)
                    error = ERR_INSUFFICIENT_MEMORY;
            }

            linalg_det_completion(error, det_v);
            return error;
        }

        public static int inv_r_completion1(int error, ref vartype_realmatrix a, ref int[] perm, Phloat det)
        {
            if (error != ERR_NONE)
            {
                free_vartype(ref linalg_inv_result);
                
                vartype tmp = (vartype)a;
                free_vartype(ref tmp);
                
                perm = null;

                linalg_inv_completion(error, null);
                return error;
            }
            else
            {
                int i;
                int n = a.rows;

                vartype_realmatrix inv = (vartype_realmatrix)linalg_inv_result;
                
                for (i = 0; i < n; i++)
                    inv.array.data[i * (n + 1)] = 1;

                return lu_backsubst_rr(a, perm, inv, inv_r_completion2);
            }
        }

        public static void inv_r_completion2(int error, ref vartype_realmatrix a, ref int[] perm, vartype_realmatrix b)
        {
            if (error != ERR_NONE)
                free_vartype(ref linalg_inv_result); // Note: linalg_inv_result == b

            vartype tmp = (vartype)a;
            free_vartype(ref tmp);
            
            perm = null;
            
            linalg_inv_completion(error, linalg_inv_result);
        }

        public static int inv_c_completion1(int error, ref vartype_complexmatrix a, ref int[] perm, Phloat det_re, Phloat det_im)
        {
            if (error != ERR_NONE)
            {
                free_vartype(ref linalg_inv_result);

                vartype tmp = (vartype)a;
                free_vartype(ref tmp);
                
                perm = null;
                linalg_inv_completion(error, null);
                
                return error;
            }
            else
            {
                int i;
                int n = a.rows;

                vartype_complexmatrix inv = (vartype_complexmatrix)linalg_inv_result;

                for (i = 0; i < n; i++)
                    inv.array.data[2 * (i * (n + 1))] = 1;
                
                return lu_backsubst_cc(a, perm, inv, inv_c_completion2);
            }
        }

        public static void inv_c_completion2(int error, ref vartype_complexmatrix a, ref int[] perm, vartype_complexmatrix b)
        {
            if (error != ERR_NONE)
                free_vartype(ref linalg_inv_result); // Note: linalg_inv_result == b

            vartype tmp = (vartype)a;
            free_vartype(ref tmp);
            
            perm = null;
            
            linalg_inv_completion(error, linalg_inv_result);
        }

        public static int linalg_inv(vartype src, fp_completion completion)
        {
            int n;
            int[] perm;

            if (src.type == TYPE.REALMATRIX)
            {
	            vartype_realmatrix ma = (vartype_realmatrix) src;
	            vartype lu;
                vartype inv;
	        
                n = ma.rows;
	        
                if (n != ma.columns)
	                return ERR_DIMENSION_ERROR;
	        
                if (contains_no_strings(ma) == 0)
	                return ERR_ALPHA_DATA_IS_INVALID;
	        
                lu = new_realmatrix(n, n);
	        
                if (lu == null)
	                return ERR_INSUFFICIENT_MEMORY;
	        
                inv = new_realmatrix(n, n);
	        
                if (inv == null)
                {
                    free_vartype(ref lu);

	                return ERR_INSUFFICIENT_MEMORY;
	            }
	        
                perm = new int[n];
	        
                if (perm == null)
                {
	                free_vartype(ref lu);
	                free_vartype(ref inv);

	                return ERR_INSUFFICIENT_MEMORY;
	            }

	            matrix_copy(ref lu, src);

	            linalg_inv_completion = completion;
	            linalg_inv_result = inv;
	        
                return lu_decomp_r((vartype_realmatrix) lu, perm, inv_r_completion1);
            }
            else
            {
    	        vartype_complexmatrix ma = (vartype_complexmatrix) src;
	            vartype lu;
                vartype inv;
	        
                n = ma.rows;
	        
                if (n != ma.columns)
	                return ERR_DIMENSION_ERROR;

	            lu = new_complexmatrix(n, n);
	        
                if (lu == null)
	                return ERR_INSUFFICIENT_MEMORY;
	        
                inv = new_complexmatrix(n, n);
	        
                if (inv == null)
                {
	                free_vartype(ref lu);
	                return ERR_INSUFFICIENT_MEMORY;
	            }
	        
                perm = new int[n];
	        
                if (perm == null)
                {
	                free_vartype(ref lu);
	                free_vartype(ref inv);

	                return ERR_INSUFFICIENT_MEMORY;
	            }
	        
                matrix_copy(ref lu, src);
	        
                linalg_inv_completion = completion;
	            linalg_inv_result = inv;
	        
                return lu_decomp_c((vartype_complexmatrix) lu, perm, inv_c_completion1);
            }
        }
    }
}
