﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        public static int lu_backsubst_rr(vartype_realmatrix a, int[] perm, vartype_realmatrix b, fp_completion_rm2 completion)
        {
            backsub_rr_data_struct dat = new backsub_rr_data_struct();

            if (dat == null)
            {
	            completion(ERR_INSUFFICIENT_MEMORY, ref a, ref perm, b);
	            return ERR_INSUFFICIENT_MEMORY;
            }

            dat.a = a;
            dat.perm = perm;
            dat.b = b;
            dat.completion = completion;

            dat.state = 0;

            backsub_rr_data = dat;
            mode_interruptible = lu_backsubst_rr_worker;
            mode_stoppable = false;
            
            return ERR_INTERRUPTIBLE;
        }

        public static int lu_backsubst_rr_worker(int interrupted)
        {
            backsub_rr_data_struct dat = backsub_rr_data;

            Phloat[] a = dat.a.array.data;
            int n = dat.a.rows;
            Phloat[] b = dat.b.array.data;
            int q = dat.b.columns;
            int[] perm = dat.perm;
            int count = 1000;

            int i = dat.i;
            int ii = dat.ii;
            int j = dat.j;
            int ll = dat.ll;
            int k = dat.k;
            
            Phloat sum = dat.sum;

            Phloat t;

            if (interrupted != 0)
            {
                dat.completion(ERR_INTERRUPTED, ref dat.a, ref perm, dat.b);
                dat = null;
                return ERR_INTERRUPTED;
            }

            bool state1 = false;
            bool state2 = false;

            switch (dat.state)
            {
                case 0: break;
                case 1: //goto state1;
                    state1 = true;
                    break;

                case 2: //goto state2;
                    state2 = true;
                    break;
            }

            for (k = state1 || state2 ? k : 0; k < q; k++)
            {
                if (!state2)
                {
                    if (!state1) ii = -1;

                    for (i = state1 ? i : 0; i < n; i++)
                    {
                        if (!state1)
                        {
                            ll = perm[i];
                            sum = b[ll * q + k];
                            b[ll * q + k] = b[i * q + k];
                        }

                        if (ii != -1 || state1)
                        {
                            for (j = state1 ? j : ii; j < i; j++)
                            {
                                if (!state1)
                                {
                                    sum -= a[i * n + j] * b[j * q + k];

                                    // STATE(1);
                                    if (--count <= 0)
                                    {
                                        dat.state = 1;
                                        goto suspend;
                                    }
                                }

                                //state1:
                                state1 = false;
                            }
                        }
                        else if (sum != 0)
                            ii = i;

                        b[i * q + k] = sum;
                    }
                }

                for (i = state2 ? i : n - 1; i >= 0; i--)
                {
                    if (!state2) sum = b[i * q + k];
                
                    for (j = state2 ? j : i + 1; j < n; j++)
                    {
                        if (!state2)
                        {
                            sum -= a[i * n + j] * b[j * q + k];

                            //STATE(2);
                            if (--count <= 0)
                            {
                                dat.state = 2;
                                goto suspend;
                            }
                        }

                        //state2:
                        state2 = false;
                    }
                    
                    t = sum / a[i * n + i];
                    
                    if (Phloat.p_isinf(t) != 0 || Phloat.p_isnan(t) != 0)
                    {
                        if (core_settings.matrix_outofrange && flags.range_error_ignore == 0)
                            return ERR_OUT_OF_RANGE;
                        else
                            t = Phloat.p_isinf(t) < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                    }
                    
                    b[i * q + k] = t;
                }
            }

            dat.completion(ERR_NONE, ref dat.a, ref perm, dat.b);
            dat = null;
            return ERR_NONE;

        suspend:
            dat.i = i;
            dat.ii = ii;
            dat.j = j;
            dat.ll = ll;
            dat.k = k;
            dat.sum = sum;
            return ERR_INTERRUPTIBLE;
        }

        public static int lu_backsubst_cc(vartype_complexmatrix a, int[] perm, vartype_complexmatrix b, fp_completion_cm2 completion)
        {
            backsub_cc_data_struct dat = new backsub_cc_data_struct();

            if (dat == null)
            {
	            completion(ERR_INSUFFICIENT_MEMORY, ref a, ref perm, b);
	            return ERR_INSUFFICIENT_MEMORY;
            }

            dat.a = a;
            dat.perm = perm;
            dat.b = b;
            dat.completion = completion;

            dat.state = 0;

            backsub_cc_data = dat;
            mode_interruptible = lu_backsubst_cc_worker;
            mode_stoppable = false;

            return ERR_INTERRUPTIBLE;
        }

        public static int lu_backsubst_cc_worker(int interrupted)
        {
            backsub_cc_data_struct dat = backsub_cc_data;

            Phloat[] a = dat.a.array.data;
            int n = dat.a.rows;
            Phloat[] b = dat.b.array.data;
            int q = dat.b.columns;
            int[] perm = dat.perm;
            int count = 1000;

            int i = dat.i;
            int ii = dat.ii;
            int j = dat.j;
            int ll = dat.ll;
            int k = dat.k;
            Phloat sum_re = dat.sum_re;
            Phloat sum_im = dat.sum_im;
            Phloat tmp, tmp_re, tmp_im;

            Phloat bre, bim;
            Phloat t_re, t_im;

            if (interrupted != 0)
            {
	            dat.completion(ERR_INTERRUPTED, ref dat.a, ref perm, dat.b);
	            dat = null;
	            return ERR_INTERRUPTED;
            }

            bool state1 = false;
            bool state2 = false;

            switch (dat.state)
            {
	            case 0: break;
	            case 1: //goto state1;
                    state1 = true;
                    break;

	            case 2: //goto state2;
                    state2 = true;
                    break;
            }

            for (k = state1 || state2 ? k : 0; k < q; k++)
            {
                if (!state2)
                {
                    if (!state1) ii = -1;

                    for (i = state1 ? i : 0; i < n; i++)
                    {
                        if (!state1)
                        {
                            ll = perm[i];
                            sum_re = b[2 * (ll * q + k)];
                            sum_im = b[2 * (ll * q + k) + 1];
                            b[2 * (ll * q + k)] = b[2 * (i * q + k)];
                            b[2 * (ll * q + k) + 1] = b[2 * (i * q + k) + 1];
                        }

                        if (ii != -1 || state1)
                        {
                            for (j = state1 ? j : ii; j < i; j++)
                            {
                                if (!state1)
                                {
                                    bre = b[2 * (j * q + k)];
                                    bim = b[2 * (j * q + k) + 1];
                                    tmp_re = a[2 * (i * n + j)];
                                    tmp_im = a[2 * (i * n + j) + 1];
                                    sum_re -= bre * tmp_re - bim * tmp_im;
                                    sum_im -= bim * tmp_re + bre * tmp_im;

                                    //STATE(1);
                                    if (--count <= 0)
                                    {
                                        dat.state = 1;
                                        goto suspend;
                                    }
                                }

                                //state1:
                                state1 = false;

                            }
                        }
                        else if (sum_re != 0 || sum_im != 0)
                            ii = i;

                        b[2 * (i * q + k)] = sum_re;
                        b[2 * (i * q + k) + 1] = sum_im;
                    }
                }

                for (i = state2 ? i : n - 1; i >= 0; i--)
                {
                    if (!state2)
                    {
                        sum_re = b[2 * (i * q + k)];
                        sum_im = b[2 * (i * q + k) + 1];
                    }
	                
                    for (j = state2 ? j : i + 1; j < n; j++)
                    {
                        if (!state2)
                        {
                            bre = b[2 * (j * q + k)];
                            bim = b[2 * (j * q + k) + 1];
                            tmp_re = a[2 * (i * n + j)];
                            tmp_im = a[2 * (i * n + j) + 1];
                            sum_re -= bre * tmp_re - bim * tmp_im;
                            sum_im -= bim * tmp_re + bre * tmp_im;

                            //STATE(2);
                            if (--count <= 0)
                            {
                                dat.state = 2;
                                goto suspend;
                            }
                        }

	                    //state2:
	                    state2 = false;

	                }

                    tmp_re = a[2 * (i * n + i)];
	                tmp_im = a[2 * (i * n + i) + 1];
	                tmp = Phloat.hypot(tmp_re, tmp_im);
	                tmp_re = tmp_re / tmp / tmp;
	                tmp_im = -tmp_im / tmp / tmp;
	                t_re = sum_re * tmp_re - sum_im * tmp_im;
	                t_im = sum_im * tmp_re + sum_re * tmp_im;

                    if (Phloat.p_isinf(t_re) != 0 || Phloat.p_isnan(t_re) != 0)
                    {
        		        if (core_settings.matrix_outofrange && flags.range_error_ignore == 0)
		                    return ERR_OUT_OF_RANGE;
		                else
                            t_re = Phloat.p_isinf(t_re) < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
	                }

                    if (Phloat.p_isinf(t_im) != 0 || Phloat.p_isnan(t_im) != 0)
                    {
		                if (core_settings.matrix_outofrange && flags.range_error_ignore == 0)
		                    return ERR_OUT_OF_RANGE;
		                else
                            t_im = Phloat.p_isinf(t_im) < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                    }

                    b[2 * (i * q + k)] = t_re;
	                b[2 * (i * q + k) + 1] = t_im;
	            }
            }

            dat.completion(ERR_NONE, ref dat.a, ref perm, dat.b);
            dat = null;
            return ERR_NONE;

            suspend:
            dat.i = i;
            dat.ii = ii;
            dat.j = j;
            dat.ll = ll;
            dat.k = k;
            dat.sum_re = sum_re;
            dat.sum_im = sum_im;
            return ERR_INTERRUPTIBLE;
        }

        public static int lu_backsubst_rc(vartype_realmatrix a, int[] perm, vartype_complexmatrix b, fp_completion_rc2 completion)
        {
            backsub_rc_data_struct dat = new backsub_rc_data_struct();

            if (dat == null)
            {
                completion(ERR_INSUFFICIENT_MEMORY, ref a, ref perm, b);
                return ERR_INSUFFICIENT_MEMORY;
            }

            dat.a = a;
            dat.perm = perm;
            dat.b = b;
            dat.completion = completion;

            dat.state = 0;

            backsub_rc_data = dat;
            mode_interruptible = lu_backsubst_rc_worker;
            mode_stoppable = false;

            return ERR_INTERRUPTIBLE;
        }

        public static int lu_backsubst_rc_worker(int interrupted)
        {
            backsub_rc_data_struct dat = backsub_rc_data;

            Phloat[] a = dat.a.array.data;
            int n = dat.a.rows;
            Phloat[] b = dat.b.array.data;
            int q = dat.b.columns;
            int[] perm = dat.perm;
            int count = 1000;

            int i = dat.i;
            int ii = dat.ii;
            int j = dat.j;
            int ll = dat.ll;
            int k = dat.k;
            Phloat sum_re = dat.sum_re;
            Phloat sum_im = dat.sum_im;
            Phloat tmp;

            Phloat t_re, t_im;

            if (interrupted != 0)
            {
                dat.completion(ERR_INTERRUPTED, ref dat.a, ref perm, dat.b);
                dat = null;
                return ERR_INTERRUPTED;
            }

            bool state1 = false;
            bool state2 = false;

            switch (dat.state)
            {
                case 0: break;
                case 1: //goto state1;
                    state1 = true;
                    break;

                case 2: //goto state2;
                    state2 = true;
                    break;
            }

            for (k = state1 || state2 ? k : 0; k < q; k++)
            {
                if (!state1 & !state2) ii = -1;

                if (!state2)
                    for (i = state1 ? i : 0; i < n; i++)
                    {
                        if (!state1)
                        {
                            ll = perm[i];
                            sum_re = b[2 * (ll * q + k)];
                            sum_im = b[2 * (ll * q + k) + 1];
                            b[2 * (ll * q + k)] = b[2 * (i * q + k)];
                            b[2 * (ll * q + k) + 1] = b[2 * (i * q + k) + 1];
                        }

                        if (ii != -1 || state1)
                        {
                            for (j = state1 ? j : ii; j < i; j++)
                            {
                                if (!state1)
                                {
                                    tmp = a[i * n + j];
                                    sum_re -= tmp * b[2 * (j * q + k)];
                                    sum_im -= tmp * b[2 * (j * q + k) + 1];

                                    //STATE(1);
                                    if (--count <= 0)
                                    {
                                        dat.state = 1;
                                        goto suspend;
                                    }
                                }

                                //state1:
                                state1 = false;

                            }
                        }
                        else if (sum_re != 0 || sum_im != 0)
                            ii = i;

                        b[2 * (i * q + k)] = sum_re;
                        b[2 * (i * q + k) + 1] = sum_im;
                    }

                for (i = state2 ? i : n - 1; i >= 0; i--)
                {
                    if (!state2)
                    {
                        sum_re = b[2 * (i * q + k)];
                        sum_im = b[2 * (i * q + k) + 1];
                    }

                    for (j = state2 ? j : i + 1; j < n; j++)
                    {
                        if (!state2)
                        {
                            tmp = a[i * n + j];
                            sum_re -= tmp * b[2 * (j * q + k)];
                            sum_im -= tmp * b[2 * (j * q + k) + 1];

                            //STATE(2);
                            if (--count <= 0)
                            {
                                dat.state = 2;
                                goto suspend;
                            }
                        }

                        //state2:
                        state2 = false; ;

                    }

                    tmp = a[i * n + i];
                    t_re = sum_re / tmp;
                    t_im = sum_im / tmp;

                    if (Phloat.p_isinf(t_re) != 0 || Phloat.p_isnan(t_re) != 0)
                    {
                        if (core_settings.matrix_outofrange && flags.range_error_ignore == 0)
                            return ERR_OUT_OF_RANGE;
                        else
                            t_re = Phloat.p_isinf(t_re) < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                    }

                    if (Phloat.p_isinf(t_im) != 0 || Phloat.p_isnan(t_im) != 0)
                    {
                        if (core_settings.matrix_outofrange && flags.range_error_ignore == 0)
                            return ERR_OUT_OF_RANGE;
                        else
                            t_im = Phloat.p_isinf(t_im) < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                    }

                    b[2 * (i * q + k)] = t_re;
                    b[2 * (i * q + k) + 1] = t_im;
                }
            }

            dat.completion(ERR_NONE, ref dat.a, ref perm, dat.b);
            dat = null;
            return ERR_NONE;

        suspend:
            dat.i = i;
            dat.ii = ii;
            dat.j = j;
            dat.ll = ll;
            dat.k = k;
            dat.sum_re = sum_re;
            dat.sum_im = sum_im;
            return ERR_INTERRUPTIBLE;
        }

        public static int lu_decomp_r(vartype_realmatrix a, int[] perm, fp_completion_rm completion)
        {
            lu_r_data_struct dat = new lu_r_data_struct();

            if (dat == null)
	            return completion(ERR_INSUFFICIENT_MEMORY, ref a, ref perm, 0);     // TODO: should last param be null?

            dat.scale = new Phloat[a.rows];
            
            if (dat.scale == null)
            {
	            dat = null;
	            return completion(ERR_INSUFFICIENT_MEMORY, ref a, ref perm, 0);
            }

            dat.a = a;
            dat.perm = perm;
            dat.completion = completion;

            dat.state = 0;

            lu_r_data = dat;
            mode_interruptible = lu_decomp_r_worker;
            mode_stoppable = false;
            
            return ERR_INTERRUPTIBLE;
        }

        public static int lu_decomp_r_worker(int interrupted)
        {
            lu_r_data_struct dat = lu_r_data;

            Phloat[] a = dat.a.array.data;
            int n = dat.a.rows;
            Phloat[] scale = dat.scale;
            int[] perm = dat.perm;
            int count = 1000;
            int err;

            int i = dat.i;
            int imax = dat.imax;
            int j = dat.j;
            int k = dat.k;
            Phloat max = dat.max;
            Phloat tmp = dat.tmp;
            Phloat sum = dat.sum;

            if (interrupted != 0)
            {
                scale = null;
                err = dat.completion(ERR_INTERRUPTED, ref dat.a, ref perm, 0);
                dat = null;
                return err;
            }

            bool state1 = false;
            bool state2 = false;
            bool state3 = false;
            bool state4 = false;
            bool state5 = false;


            switch (dat.state)
            {
                case 0:
                    break;
                case 1: //goto state1;
                    state1 = true;
                    break;
                case 2: //goto state2;
                    state2 = true;
                    break;
                case 3: //goto state3;
                    state3 = true;
                    break;
                case 4: //goto state4;
                    state4 = true;
                    break;
                case 5: //goto state5;
                    state5 = true;
                    break;
            }

            if (!state1 && !state2 && !state3 && !state4 && !state5) dat.det = 1;

            if (!state2 && !state3 && !state4 && !state5)
                for (i = state1 ? i : 0; i < n; i++)
                {
                    if (!state1) max = 0;

                    for (j = state1 ? j : 0; j < n; j++)
                    {
                        if (!state1)
                        {
                            tmp = a[i * n + j];
                            if (tmp < 0)
                                tmp = -tmp;
                            if (tmp > max)
                                max = tmp;
                     
                            //STATE(1);
                            if (--count <= 0)
                            {
                                dat.state = 1;
                                goto suspend;
                            }
                        }

	                    //state1:
	                    state1 = false;

                    }
                    scale[i] = max;
                }


            for (j = state2 || state3 || state4 || state5 ? j : 0; j < n; j++)
            {
                if (!state3 && !state4 && !state5)
                {
                    for (i = state2 ? i : 0; i < j; i++)
                    {
                        if (!state2) sum = a[i * n + j];

                        for (k = state2 ? k : 0; k < i; k++)
                        {
                            if (!state2)
                            {
                                sum -= a[i * n + k] * a[k * n + j];

                                //STATE(2);
                                if (--count <= 0)
                                {
                                    dat.state = 2;
                                    goto suspend;
                                }
                            }

                            //state2:
                            state2 = false;
                        }
                        a[i * n + j] = sum;
                    }

                    max = 0;
                    imax = j;
                }

                if (!state4 && !state5)
                    for (i = state3 ? i : j; i < n; i++)
                    {
                        if (!state3) sum = a[i * n + j];

                        for (k = state3 ? k : 0; k < j; k++)
                        {
                            if (!state3)
                            {
                                sum -= a[i * n + k] * a[k * n + j];

                                //STATE(3);
                                if (--count <= 0)
                                {
                                    dat.state = 3;
                                    goto suspend;
                                }
                            }
	                    
                            //state3:
	                        state3 = false;

                        }
                        a[i * n + j] = sum;
                        if (scale[i] == 0)
                        {
                            imax = i;
                            break;
                        }
                        tmp = (sum < 0 ? -sum : sum) / scale[i];
                        if (tmp > max)
                        {
                            imax = i;
                            max = tmp;
                        }
                    }

                if (!state5)
                {
                    if (j != imax || state4)
                    {
                        for (k = state4 ? k : 0; k < n; k++)
                        {
                            if (!state4)
                            {
                                tmp = a[imax * n + k];
                                a[imax * n + k] = a[j * n + k];
                                a[j * n + k] = tmp;

                                //STATE(4);
                                if (--count <= 0)
                                {
                                    dat.state = 4;
                                    goto suspend;
                                }
                            }

                            //state4:
                            state4 = false;

                        }
                        dat.det = -dat.det;
                        scale[imax] = scale[j];
                    }

                    perm[j] = imax;
                    if (a[j * n + j] == 0)
                    {
                        if (core_settings.matrix_singularmatrix)
                        {
                            scale = null;
                            err = dat.completion(ERR_SINGULAR_MATRIX, ref dat.a, ref perm, 0);
                            dat = null;
                            return err;
                        }
                        else
                        {
                            // For a zero pivot, substitute a small positive number.
                            // I use a number that's about 10^-20 times the size of
                            // the maximum of the original column, with a minimum of
                            // 10^20 / POS_HUGE_PHLOAT.
                            //
                            Phloat tiniest = 1e20 / Phloat.POS_HUGE_PHLOAT;
                            Phloat tiny;
                            
                            if (scale[j] == 0)
                                tiny = tiniest;
                            else
                            {
                                tiny = Phloat.pow(10, Phloat.floor(Phloat.log10(scale[j])) - 20);
                                
                                if (tiny < tiniest)
                                    tiny = tiniest;
                            }
                            a[j * n + j] = tiny;
                        }
                    }
                    dat.det *= a[j * n + j];
                }

                if (j != n - 1 || state5)
                {
                    if (!state5) tmp = 1 / a[j * n + j];

                    for (i = state5 ? i : j + 1; i < n; i++)
                    {
                        if (!state5)
                        {
                            a[i * n + j] *= tmp;

                            //STATE(5);
                            if (--count <= 0)
                            {
                                dat.state = 5;
                                goto suspend;
                            }
                        }

	                    //state5:
	                    state5 = false;
                    }
                }
            }

            scale = null;
            err = dat.completion(ERR_NONE, ref dat.a, ref perm, dat.det);
            dat = null;
            return err;

        suspend:
            dat.i = i;
            dat.imax = imax;
            dat.j = j;
            dat.k = k;
            dat.max = max;
            dat.tmp = tmp;
            dat.sum = sum;
            return ERR_INTERRUPTIBLE;
        }

        public static int lu_decomp_c(vartype_complexmatrix a, int[] perm, fp_completion_cm completion)
        {
            lu_c_data_struct dat = new lu_c_data_struct();

            if (dat == null)
	            return completion(ERR_INSUFFICIENT_MEMORY, ref a, ref perm, 0, 0);      // TODO: should last part be null?

            dat.scale = new Phloat[a.rows];

            if (dat.scale == null)
            {
	            dat = null;
	            return completion(ERR_INSUFFICIENT_MEMORY, ref a, ref perm, 0, 0);
            }

            dat.a = a;
            dat.perm = perm;
            dat.completion = completion;

            dat.state = 0;

            lu_c_data = dat;
            mode_interruptible = lu_decomp_c_worker;
            mode_stoppable = false;

            return ERR_INTERRUPTIBLE;
        }

        public static int lu_decomp_c_worker(int interrupted)
        {
            lu_c_data_struct dat = lu_c_data;

            Phloat[] a = dat.a.array.data;
            int n = dat.a.rows;
            Phloat[] scale = dat.scale;
            int[] perm = dat.perm;
            int count = 1000;
            int err;

            int i = dat.i;
            int imax = dat.imax;
            int j = dat.j;
            int k = dat.k;
            Phloat max = dat.max;
            Phloat tmp = dat.tmp;
            Phloat tmp_re = dat.tmp_re;
            Phloat tmp_im = dat.tmp_im;
            Phloat sum_re = dat.sum_re;
            Phloat sum_im = dat.sum_im;

            Phloat xre, xim, yre, yim;
            Phloat tiniest = 1e20 / Phloat.POS_HUGE_PHLOAT;
            Phloat tiny;
            Phloat s_re = new Phloat();
            Phloat s_im = new Phloat();

            if (interrupted != 0)
            {
	            scale = null;
	            err = dat.completion(ERR_INTERRUPTED, ref dat.a, ref perm, 0, 0);
	            dat = null;
	            return err;
            }

            bool state1 = false;
            bool state2 = false;
            bool state3 = false;
            bool state4 = false;
            bool state5 = false;

            switch (dat.state)
            {
	            case 0: break;
	            case 1: //goto state1;
                    state1 = true;
                    break;
	            case 2: //goto state2;
                    state2 = true;
                    break;
                case 3: //goto state3;
                    state3 = true;
                    break;
                case 4: //goto state4;
                    state4 = true;
                    break;
                case 5: //goto state5;
                    state5 = true;
                    break;
            }

            if (!state1 && !state2 && !state3 && !state4 && !state5)
            {
                dat.det_re = 1;
                dat.det_im = 0;
            }

            if (!state2 && !state3 && !state4 && !state5)
                for (i = state1 ? i : 0; i < n; i++)
                {
    	            if (!state1) max = 0;
	
                    for (j = state1 ? j : 0; j < n; j++)
                    {
                        if (!state1)
                        {
                            tmp = Phloat.hypot(a[2 * (i * n + j)], a[2 * (i * n + j) + 1]);

                            if (tmp > max)
                                max = tmp;

                            //STATE(1);
                            if (--count <= 0)
                            {
                                dat.state = 1;
                                goto suspend;
                            }
                        }

	                    //state1:
	                    state1 = false;
	                }
	        
                    scale[i] = max;
                }

            for (j = state2 || state3 || state4 || state5 ? j : 0; j < n; j++)
            {
                if (!state3 && !state4 && !state5)
                {
                    for (i = state2 ? i : 0; i < j; i++)
                    {
                        if (!state2)
                        {
                            sum_re = a[2 * (i * n + j)];
                            sum_im = a[2 * (i * n + j) + 1];
                        }

                        for (k = state2 ? k : 0; k < i; k++)
                        {
                            if (!state2)
                            {
                                xre = a[2 * (i * n + k)];
                                xim = a[2 * (i * n + k) + 1];
                                yre = a[2 * (k * n + j)];
                                yim = a[2 * (k * n + j) + 1];
                                sum_re -= xre * yre - xim * yim;
                                sum_im -= xim * yre + xre * yim;

                                //STATE(2);
                                if (--count <= 0)
                                {
                                    dat.state = 2;
                                    goto suspend;
                                }
                            }

	                        //state2:
	                        state2 = false;

                        }

                        a[2 * (i * n + j)] = sum_re;
                        a[2 * (i * n + j) + 1] = sum_im;
                    }

                    max = 0;
                }   

                if (!state4 && !state5)
                    for (i = state3 ? i : j; i < n; i++)
                    {
                        if (!state3)
                        {
                            sum_re = a[2 * (i * n + j)];
                            sum_im = a[2 * (i * n + j) + 1];
                        }

                        for (k = state3 ? k : 0; k < j; k++)
                        {
                            if (!state3)
                            {
                                xre = a[2 * (i * n + k)];
                                xim = a[2 * (i * n + k) + 1];
                                yre = a[2 * (k * n + j)];
                                yim = a[2 * (k * n + j) + 1];
                                sum_re -= xre * yre - xim * yim;
                                sum_im -= xim * yre + xre * yim;

                                //STATE(3);
                                if (--count <= 0)
                                {
                                    dat.state = 3;
                                    goto suspend;
                                }
                            }

	                        //state3:
	                        state3 = false;
	                    }

                        a[2 * (i * n + j)] = sum_re;
	                    a[2 * (i * n + j) + 1] = sum_im;
	            
                        if (scale[i] == 0)
                        {
		                    imax = i;
		                    break;
    	                }
	
                        tmp = Phloat.hypot(sum_re, sum_im) / scale[i];
	            
                        if (tmp > max)
                        {
		                    imax = i;
		                    max = tmp;
    	                }
	                }

                if (!state5)
                {
                    if (j != imax || state5)
                    {
                        for (k = state4 ? k : 0; k < n; k++)
                        {
                            if (!state4)
                            {
                                tmp = a[2 * (imax * n + k)];

                                a[2 * (imax * n + k)] = a[2 * (j * n + k)];
                                a[2 * (j * n + k)] = tmp;

                                tmp = a[2 * (imax * n + k) + 1];

                                a[2 * (imax * n + k) + 1] = a[2 * (j * n + k) + 1];
                                a[2 * (j * n + k) + 1] = tmp;

                                //STATE(4);
                                if (--count <= 0)
                                {
                                    dat.state = 4;
                                    goto suspend;
                                }
                            }

	                        //state4:
                            state4 = false;

                        }

                        dat.det_re = -dat.det_re;
                        dat.det_im = -dat.det_im;
                        scale[imax] = scale[j];
                    }

                    perm[j] = imax;
                    tmp_re = a[2 * (j * n + j)];
                    tmp_im = a[2 * (j * n + j) + 1];

                    if (tmp_re == 0 && tmp_im == 0)
                    {
                        if (core_settings.matrix_singularmatrix)
                        {
                            scale = null;
                            err = dat.completion(ERR_NONE, ref dat.a, ref perm, 0, 0);
                            dat = null;
                            return err;
                        }
                        else
                        {
                            // For a zero pivot, substitute a small positive number.
                            // I use a number that's about 10^-20 times the size of
                            // the maximum of the original column, with a minimum of
                            // 10^20 / POS_HUGE_PHLOAT.
                            //
                            if (scale[j] == 0)
                                tiny = tiniest;
                            else
                            {
                                tiny = Phloat.pow(10, Phloat.floor(Phloat.log10(scale[j])) - 20);
                                if (tiny < tiniest)
                                    tiny = tiniest;
                            }

                            a[2 * (j * n + j)] = tmp_re = tiny;
                            a[2 * (j * n + j) + 1] = tmp_im = 0;
                        }
                    }

                    tmp = dat.det_re * tmp_re - dat.det_im * tmp_im;
                    dat.det_im = dat.det_im * tmp_re + dat.det_re * tmp_im;
                    dat.det_re = tmp;
                }

                if (j != n - 1 || state5)
                {
                    if (!state5)
                    {
                        tmp = Phloat.hypot(tmp_re, tmp_im);

                        s_re = tmp_re / tmp / tmp;
                        s_im = -tmp_im / tmp / tmp;
                    }
	                
                    for (i = state5 ? i : j + 1; i < n; i++)
                    {
                        if (!state5)
                        {
                            tmp_re = a[2 * (i * n + j)];
                            tmp_im = a[2 * (i * n + j) + 1];
                            a[2 * (i * n + j)] = tmp_re * s_re - tmp_im * s_im;
                            a[2 * (i * n + j) + 1] = tmp_im * s_re + tmp_re * s_im;

                            //STATE(5);
                            if (--count <= 0)
                            {
                                dat.state = 5;
                                goto suspend;
                            }
                        }

	                    //state5:
	                    state5 = false;
	                }
	            }
            }

            scale = null;
            err = dat.completion(ERR_NONE, ref dat.a, ref perm, dat.det_re, dat.det_im);
            dat = null;
            return err;

            suspend:
            dat.i = i;
            dat.imax = imax;
            dat.j = j;
            dat.k = k;
            dat.max = max;
            dat.tmp = tmp;
            dat.tmp_re = tmp_re;
            dat.tmp_im = tmp_im;
            dat.sum_re = sum_re;
            dat.sum_im = sum_im;
            return ERR_INTERRUPTIBLE;
        }
    }
}
