﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

using System.Windows.Navigation;


namespace WP7_42s
{
    public partial class App
    {
        public static int core_program_size(int prgm_index)
        {
            int pc = 0;
            int cmd = 0;
            arg_struct arg = new arg_struct();

            int saved_prgm = current_prgm;
            uint hp42s_code;
            byte code_flags, code_name, code_std_1, code_std_2;
            int size = 0;

            current_prgm = prgm_index;

            do
            {
	            get_next_command(ref pc, ref cmd, arg, 0);

	            hp42s_code = cmd_array[cmd].hp42s_code;

	            code_flags = (byte)(hp42s_code >> 24);
	            code_name = (byte)(hp42s_code >> 16);
	            code_std_1 = (byte)(hp42s_code >> 8);
	            code_std_2 = (byte)(hp42s_code);
	        
                switch (code_flags)
                {
                    case 1:         // A command that requires some special attention
                        if (cmd == CMD_STO)
                        {
		                    if (arg.type == ARGTYPE_NUM && arg.val.num <= 15)
			                    size += 1;
		                    else
			                    goto normal;
		                }
                        else if (cmd == CMD_RCL)
                        {
		                    if (arg.type == ARGTYPE_NUM && arg.val.num <= 15)
			                    size += 1;
		                    else
			                    goto normal;
		                }
                        else if (cmd == CMD_FIX || cmd == CMD_SCI || cmd == CMD_ENG)
                        {
		                    goto normal;
		                }
                        else if (cmd == CMD_SIZE)
                        {
		                    size += 4;
		                }
                        else if (cmd == CMD_LBL)
                        {
		                    if (arg.type == ARGTYPE_NUM)
                            {
			                    if (arg.val.num <= 14)
			                        size += 1;
                                else
            			            goto normal;
		                    }
                            else if (arg.type == ARGTYPE_STR)
                            {
			                    size += arg.length + 4;
		                    }
                            else
			                    goto normal;
		                }
                        else if (cmd == CMD_INPUT)
                        {
		                    goto normal;
		                }
                        else if (cmd == CMD_XEQ)
                        {
		                    if (arg.type == ARGTYPE_NUM || arg.type == ARGTYPE_LCLBL)
                            {
			                    size += 3;
		                    }
                            else if (arg.type == ARGTYPE_STR)
                            {
			                    size += arg.length + 2;
		                    }
                            else
			                    goto normal;
		                }
                        else if (cmd == CMD_GTO)
                        {
        		            if (arg.type == ARGTYPE_NUM && arg.val.num <= 14)
                            {
			                    size += 2;
		                    }
                            else if (arg.type == ARGTYPE_NUM || arg.type == ARGTYPE_LCLBL)
                            {
			                    size += 3;
		                    }
                            else if (arg.type == ARGTYPE_STR)
                            {
			                    size += arg.length + 2;
		                    }
                            else
			                    goto normal;
    		            }
                        else if (cmd == CMD_END)
                        {
                            // Not counted for the line 00 total
		                }
                        else if (cmd == CMD_NUMBER)
                        {
        		            char[] p = phloat2program(arg.val_d);

                            int i = 0;
                            while (p[i++] != 0)
			                    size += 1;

		                    size += 1;
		                }
                        else if (cmd == CMD_STRING)
                        {
		                    size += arg.length + 1;
		                }
                        else if (cmd >= CMD_ASGN01 && cmd <= CMD_ASGN18)
                        {
		                    if (arg.type == ARGTYPE_STR)
			                    size += arg.length + 3;
		                    else
			                    // arg.type == ARGTYPE_COMMAND; we don't use that
			                    // any more, but just to be safe (in case anyone ever
			                    // actually used this in a program), we handle it
			                    // anyway.
			                    //
			                    size += cmd_array[arg.val.cmd].name_length + 3;
		                }
                        else if ((cmd >= CMD_KEY1G && cmd <= CMD_KEY9G) || (cmd >= CMD_KEY1X && cmd <= CMD_KEY9X))
                        {
		                    if (arg.type == ARGTYPE_STR || arg.type == ARGTYPE_IND_STR)
			                    size += arg.length + 3;
		                    else
			                    size += 4;
		                }
                        else if (cmd == CMD_XROM)
                        {
		                    size += 2;
		                }
                        else
                        {
		                    // Shouldn't happen
		                    continue;
		                }
        		        break;

	            case 0:
                normal:
                    if (arg.type == ARGTYPE_STR || arg.type == ARGTYPE_IND_STR)
                    {
		                size += arg.length + 2;
		            }
                    else
                    {
		                size += code_std_1 == 0 ? 1 : 2;
		            
                        if (arg.type != ARGTYPE_NONE)
			                size += 1;
		            }
		            break;

	            case 2:
	            default:
		            // Illegal command
		            continue;
                }
            } while (cmd != CMD_END && pc < prgms[prgm_index].size);

            current_prgm = saved_prgm;
            return size;
        }

        public static void set_running(bool state)
        {
            if (mode_running != state)
            {
                mode_running = state;
                shell_annunciators(-1, -1, -1, state ? 1 : 0, -1, -1);
            }

            if (state)
            {
                // Cancel any pending INPUT command
                input_length = 0;
                mode_goose = -2;
                prgm_highlight_row = 1;
            }
        }

        public static int handle_error(int error)
        {
            if (mode_running)
            {
                if (error == ERR_RUN)
                    error = ERR_NONE;

                if (error == ERR_NONE || error == ERR_NO || error == ERR_YES || error == ERR_STOP)
                    flags.stack_lift_disable = (byte)(mode_disable_stack_lift ? 1 : 0);

                if (error == ERR_NO)
                {
                    if (prgms[current_prgm].text[pc] != CMD_END)
                        pc += get_command_length(current_prgm, pc);
                }
                else if (error == ERR_STOP)
                {
                    if (pc >= prgms[current_prgm].size)
                        pc = -1;

                    set_running(false);
                    
                    return 0;
                }
                else if (error != ERR_NONE && error != ERR_YES)
                {
                    if (flags.error_ignore != 0)
                    {
                        flags.error_ignore = 0;

                        return 1;
                    }
                    if (solve_active() && (error == ERR_OUT_OF_RANGE || error == ERR_DIVIDE_BY_0 || error == ERR_INVALID_DATA || error == ERR_STAT_MATH_ERROR))
                    {
                        unwind_stack_until_solve();
                        error = return_to_solve(1);

                        if (error == ERR_STOP)
                            set_running(false);
                        
                        if (error == ERR_NONE || error == ERR_RUN || error == ERR_STOP)
                            return 0;
                    }

                    pc = oldpc;
                    
                    display_error(error, 1);
                    set_running(false);
                    
                    return 0;
                }
                return 1;
            }
            else if (pending_command == CMD_SST)
            {
                if (error == ERR_RUN)
                    error = ERR_NONE;

                if (error == ERR_NONE || error == ERR_NO || error == ERR_YES || error == ERR_STOP)
                    flags.stack_lift_disable = (byte)(mode_disable_stack_lift ? 1 : 0);

                if (error == ERR_NO)
                {
                    if (prgms[current_prgm].text[pc] != CMD_END)
                        pc += get_command_length(current_prgm, pc);

                    //goto noerr;
                    error = ERR_NONE;

                    if (pc > prgms[current_prgm].size)
                        pc = -1;
                }
                else if (error == ERR_NONE || error == ERR_YES || error == ERR_STOP)
                {
                //noerr:
                    error = ERR_NONE;

                    if (pc > prgms[current_prgm].size)
                        pc = -1;
                }
                else
                {
                    if (flags.error_ignore != 0)
                    {
                        flags.error_ignore = 0;

                        //goto noerr;
                        error = ERR_NONE;

                        if (pc > prgms[current_prgm].size)
                            pc = -1;
                    }
                    if (solve_active() && (error == ERR_OUT_OF_RANGE || error == ERR_DIVIDE_BY_0 || error == ERR_INVALID_DATA || error == ERR_STAT_MATH_ERROR))
                    {
                        unwind_stack_until_solve();
                        error = return_to_solve(1);

                        if (error == ERR_NONE || error == ERR_RUN || error == ERR_STOP)
                        {
                            //goto noerr;
                            error = ERR_NONE;

                            if (pc > prgms[current_prgm].size)
                                pc = -1;
                        }
                    }

                    pc = oldpc;
                    display_error(error, 1);
                }

                return 0;
            }
            else
            {
                if (error == ERR_RUN)
                {
                    set_running(true);

                    error = ERR_NONE;
                }

                if (error == ERR_NONE || error == ERR_NO || error == ERR_YES || error == ERR_STOP)
                    flags.stack_lift_disable = (byte)(mode_disable_stack_lift ? 1 : 0);
                else if (flags.error_ignore != 0)
                {
                    flags.error_ignore = 0;
                    error = ERR_NONE;
                }

                if (error != ERR_NONE && error != ERR_STOP)
                    display_error(error, 1);

                return 0;
            }
        }

        public static bool program_running()
        {
            return mode_running;
        }

        public static void stop_interruptible()
        {
            int error = mode_interruptible(1);

            handle_error(error);
            
            mode_interruptible = null;

            if (mode_running)
                set_running(false);
            else
                shell_annunciators(-1, -1, -1, 0, -1, -1);

            pending_command = CMD_NONE;

            redisplay();
        }

        public static void set_shift(bool state)
        {
            if (mode_shift != state)
            {
                mode_shift = state;
                shell_annunciators(-1, state ? 1 : 0, -1, -1, -1, -1);
            }
        }

        public static int core_powercycle()
        {
            bool need_redisplay = false;

            if (mode_interruptible != null)
                stop_interruptible();

            keybuf_tail = keybuf_head;

            set_shift(false);
            flags.continuous_on = 0;
            pending_command = CMD_NONE;

            if (mode_getkey)
            {
                // A real HP-42S can't be switched off while GETKEY is active: pressing
                // OFF on the keyboard returns code 70 and stops program execution; and
                // when the auto-poweroff timeout expires, code 70 is returned but
                // program execution continues.
                // Since Free42 can be shut down in ways the HP-42S can't (exiting the
                // application, or turning off power on a Palm), I have to fake it a
                // bit; I put 70 in X as if the user had done OFF twice on a real 42S.
                // Note that, as on a real 42S, we don't allow auto-powerdown while
                // GETKEY is waiting; if an auto-powerdown request happens during
                // GETKEY, it returns 70 but program execution is not stopped, and the
                // power stays on (see core_allows_powerdown(), above).
                //
                vartype seventy = App.new_real(70);

                if (seventy != null)
                {
                    recall_result(seventy);
                    flags.stack_lift_disable = 0;
                }
                else
                {
                    display_error(ERR_INSUFFICIENT_MEMORY, 1);
                    flags.auto_exec = 0;
                }

                if (flags.auto_exec == 0)
                    need_redisplay = true;
                
                mode_getkey = false;
            }

            if (flags.auto_exec != 0)
            {
                if (mode_command_entry)
                    finish_command_entry(false);

                if (flags.prgm_mode != 0)
                {
                    if (mode_alpha_entry)
                    {
                        pc = incomplete_saved_pc;
                        prgm_highlight_row = incomplete_saved_highlight_row;
                    }
                    else if (mode_number_entry)
                    {
                        arg_struct arg = new arg_struct();
                        arg.type = ARGTYPE_DOUBLE;
                        arg.val_d = entered_number;

                        store_command(pc, CMD_NUMBER, arg);
                        
                        prgm_highlight_row = 1;
                    }

                    flags.prgm_mode = 0;
                }

                mode_alpha_entry = false;
                mode_number_entry = false;
                
                set_menu(MENULEVEL_ALPHA, (int)MENU.NONE);
                
                flags.alpha_mode = 0;
                set_running(true);
                flags.auto_exec = 0;
                need_redisplay = false;
            }
            else
            {
                if (mode_running)
                {
                    set_running(false);
                    need_redisplay = true;
                }
            }

            if (need_redisplay || bin_dec_mode_switch || state_file_has_old_bcd)
            {
                bin_dec_mode_switch = false;
                state_file_has_old_bcd = false;

                redisplay();
            }

            return mode_running ? 1 : 0;
        }

        public static void sst()
        {
            if (pc >= prgms[current_prgm].size - 2)
            {
                pc = -1;
                prgm_highlight_row = 0;
            }
            else
            {
                if (pc == -1)
                    pc = 0;
                else
                    pc += get_command_length(current_prgm, pc);
         
                prgm_highlight_row = 1;
            }
        }

        public static void bst()
        {
            int line = pc2line(pc);

            if (line == 0)
            {
                pc = prgms[current_prgm].size - 2;
                prgm_highlight_row = 1;
            }
            else
            {
                pc = line2pc(line - 1);
                prgm_highlight_row = 0;
            }
        }

        public static void finish_command_entry(bool refresh)
        {
            if (pending_command == CMD_ASSIGNa)
            {
                pending_command = CMD_NONE;
                start_incomplete_command(CMD_ASSIGNb);
                return;
            }

            mode_command_entry = false;
            
            if (flags.prgm_mode != 0)
            {
                set_menu(MENULEVEL_COMMAND, (int)App.MENU.NONE);

                if (pending_command == CMD_NULL || pending_command == CMD_CANCELLED)
                {
                    pc = incomplete_saved_pc;
                    prgm_highlight_row = incomplete_saved_highlight_row;
                }
                else if (pending_command == CMD_SST || pending_command == CMD_BST)
                {
                    pc = incomplete_saved_pc;
                    prgm_highlight_row = incomplete_saved_highlight_row;

                    if (pending_command == CMD_SST)
                        sst();
                    else
                        bst();

                    repeating = 1;
                    repeating_shift = 1;
                    repeating_key = pending_command == CMD_SST ? KEY_DOWN : KEY_UP;
                    pending_command = CMD_NONE;
                    
                    redisplay();
                }
                else
                {
                    int inserting_an_end = pending_command == CMD_END ? 1 : 0;

                    if ((cmd_array[pending_command].flags & FLAG_IMMED) != 0)
                    {
                        finish_command_entry_2(refresh);
                        return;
                    }

                    store_command(pc, pending_command, pending_command_arg);
                    
                    if (inserting_an_end != 0)      // current_prgm was already incremented by store_command()
                        pc = 0;

                    prgm_highlight_row = 1;
                    pending_command = CMD_NONE;
                    
                    redisplay();
                }
            }
            else
            {
                finish_command_entry_2(refresh);
            }
        }

        public static void finish_command_entry_2(bool refresh)
        {
            if (refresh)
                redisplay();

            if (mode_commandmenu == MENU_CATALOG && (pending_command == CMD_GTO || pending_command == CMD_XEQ))
            {
                // TO DO: also applies to CLP, PRP, and maybe more.
                // Does *not* apply to the program menus displayed by VARMENU,
                // PGMINT, PGMSLV, and... (?)
                //
                int catsect = get_cat_section();

                if (catsect == CATSECT_PGM || catsect == CATSECT_PGM_ONLY)
                {
                    int row = get_cat_row();

                    set_menu(MENULEVEL_TRANSIENT, mode_commandmenu);
                        
                    set_cat_section(catsect);
                    set_cat_row(row);
                        
                    remove_program_catalog = 1;
                }
            }

            if (pending_command == CMD_ASSIGNb || (pending_command >= CMD_ASGN01 && pending_command <= CMD_ASGN18))
            {
                set_menu(MENULEVEL_PLAIN, mode_commandmenu);
                mode_plainmenu_sticky = true;
            }

            set_menu(MENULEVEL_COMMAND, (int)App.MENU.NONE);
                
            if (pending_command == CMD_BST)
            {
                bst();
                
                flags.prgm_mode = 1;
                redisplay();

                flags.prgm_mode = 0;
                pending_command = CMD_CANCELLED;

                repeating = 1;
                repeating_shift = 1;
                repeating_key = KEY_UP;
            }
        }

        public static void start_incomplete_command(int cmd_id)
        {
            int argtype = cmd_array[cmd_id].argtype;

            if ((flags.prgm_mode == 0) && (cmd_array[cmd_id].flags & FLAG_PRGM_ONLY) != 0)
            {
                display_error(ERR_RESTRICTED_OPERATION, 0);
                redisplay();
                return;
            }

            incomplete_command = cmd_id;
            incomplete_ind = 0;
            
            if (argtype == ARG_NAMED || argtype == ARG_PRGM || argtype == ARG_RVAR || argtype == ARG_MAT)
                incomplete_alpha = 1;
            else
                incomplete_alpha = 0;

            incomplete_length = 0;
            incomplete_num = 0;
            
            if (argtype == ARG_NUM9)
                incomplete_maxdigits = 1;
            else if (argtype == ARG_COUNT)
                incomplete_maxdigits = cmd_id == CMD_SIMQ ? 2 : 4;
            else
                incomplete_maxdigits = 2;
            
            incomplete_argtype = argtype;
            mode_command_entry = true;
            
            if (incomplete_command == CMD_ASSIGNa)
            {
                set_catalog_menu(CATSECT_TOP);
                flags.local_label = 0;
            }
            else if (argtype == ARG_CKEY)
                set_menu(MENULEVEL_COMMAND, MENU_CUSTOM1);
            else if (argtype == ARG_MKEY)
                set_menu(MENULEVEL_COMMAND, MENU_BLANK);
            else if (argtype == ARG_VAR)
            {
                if (mode_appmenu == MENU_VARMENU)
                    mode_commandmenu = MENU_VARMENU;
                else if (mode_appmenu == MENU_INTEG_PARAMS)
                    mode_commandmenu = MENU_INTEG_PARAMS;
                else
                    set_catalog_menu(CATSECT_VARS_ONLY);
            }
            else if (argtype == ARG_NAMED)
                set_catalog_menu(CATSECT_VARS_ONLY);
            else if (argtype == ARG_REAL)
            {
                if (mode_appmenu == MENU_VARMENU)
                    mode_commandmenu = MENU_VARMENU;
                else if (mode_appmenu == MENU_INTEG_PARAMS)
                    mode_commandmenu = MENU_INTEG_PARAMS;
                else
                    set_catalog_menu(CATSECT_REAL_ONLY);
            }
            else if (argtype == ARG_RVAR)
            {
                if (App.vars_exist(1, 0, 0) != 0)
                    set_catalog_menu(CATSECT_REAL_ONLY);
                else if (flags.prgm_mode != 0)
                {
                    if (incomplete_command == CMD_MVAR)
                        mode_commandmenu = MENU_ALPHA1;
                }
                else
                {
                    mode_command_entry = false;
                    display_error(ERR_NO_REAL_VARIABLES, 0);
                }
            }
            else if (argtype == ARG_MAT)
            {
                if ((flags.prgm_mode != 0) || (App.vars_exist(0, 0, 1) != 0))
                    set_catalog_menu(CATSECT_MAT_ONLY);
                else if (cmd_id != CMD_DIM)
                {
                    mode_command_entry = false;
                    display_error(ERR_NO_MATRIX_VARIABLES, 0);
                }
            }
            else if (argtype == ARG_LBL || argtype == ARG_PRGM)
                set_catalog_menu(CATSECT_PGM_ONLY);
            else if (cmd_id == CMD_LBL)
                set_menu(MENULEVEL_COMMAND, MENU_ALPHA1);

            redisplay();
        }

        public static int core_menu()
        {
            return (mode_clall || get_front_menu() != (int)MENU.NONE) ? 1 : 0;
        }

        public static int core_repeat()
        {
            keydown(repeating_shift, repeating_key);

            int rpt = repeating;
            repeating = 0;
            return rpt;
        }

        public static void core_keytimeout1()
        {
            if (pending_command == CMD_LINGER1 || pending_command == CMD_LINGER2)
                return;

            if (pending_command == CMD_RUN || pending_command == CMD_SST)
            {
                int saved_pending_command = pending_command;
            
                if (pc == -1)
                    pc = 0;
                
                prgm_highlight_row = 1;
                flags.prgm_mode = 2;    // HACK - magic value to tell redisplay()
                                        // not to suppress option menu highlights
                pending_command = CMD_NONE;
                
                redisplay();
                flags.prgm_mode = 0;

                pending_command = saved_pending_command;
            }
            else if (pending_command != CMD_NONE && pending_command != CMD_CANCELLED && (cmd_array[pending_command].flags & FLAG_NO_SHOW) == 0)
            {
                display_command(0);
                // If the program catalog was left up by GTO or XEQ, don't paint over it
                if (mode_transientmenu == (int)App.MENU.NONE || pending_command == CMD_NULL)
                    display_x(1);

                flush_display();
            }
        }

        public static void core_keytimeout2()
        {
            if (pending_command == CMD_LINGER1 || pending_command == CMD_LINGER2)
                return;

            remove_program_catalog = 0;

            if (pending_command != CMD_NONE && pending_command != CMD_CANCELLED && (cmd_array[pending_command].flags & FLAG_NO_SHOW) == 0)
            {
                clear_row(0);
                draw_string(0, 0, "NULL".ToCharArray(), 4);
                display_x(1);
                flush_display();

                pending_command = CMD_CANCELLED;
            }
        }

        public static bool core_timeout3(int repaint)
        {
            if (mode_pause)
            {
                if (repaint != 0)
                {
                    // The PSE ended normally
                    mode_pause = false;
                    if (mode_goose >= 0)
                        mode_goose = -1 - mode_goose;
                }

                return true;
            }

            // Remove the output of SHOW, MEM, or shift-VARMENU from the display
            if (pending_command == CMD_LINGER1)
                pending_command = CMD_CANCELLED;
            else if (pending_command == CMD_LINGER2)
            {
                flags.message = 0;
                flags.two_line_message = 0;
                pending_command = CMD_NONE;

                if (repaint != 0)
                    redisplay();
            }

            return false;
        }

        public static int core_keydown(int key, ref int enqueued, ref int repeat)
        {
            enqueued = 0;
            repeat = 0;

            if (key == KEY_SHIFT)
            {
                set_shift(!mode_shift);

                return ((mode_running && !mode_getkey && !mode_pause) || keybuf_head != keybuf_tail) ? 1 : 0; ;
            }

            if (mode_pause && key == 0)
                return 0;

            if (mode_pause)
            {
                mode_pause = false;
                set_running(false);

                if (!mode_shift && (key == KEY_RUN || key == KEY_EXIT))
                {
                    redisplay();

                    return 0;
                }
            }

            if (mode_interruptible != null)
            {
                // We're in the middle of an interruptible function
                // (e.g., INVRT, PRP); queue up any keystrokes and invoke
                // the appropriate callback to keep the funtion moving along
                //
                int error;
                int keep_running;

                if (key != 0)
                {
                    // Enqueue...
                    enqueued = 1;

                    if (key == KEY_EXIT || (mode_stoppable && !mode_shift && key == KEY_RUN))
                    {
                        keybuf_tail = keybuf_head;
                        stop_interruptible();

                        return 0;
                    }
                    else
                    {
                        if (((keybuf_head + 1) & 15) != keybuf_tail)
                        {
                            if (mode_shift)
                                key = -key;

                            keybuf[keybuf_head] = key;
                            keybuf_head = (keybuf_head + 1) & 15;
                        }
                    }

                    set_shift(false);
                }

                error = mode_interruptible(0);

                if (error == ERR_INTERRUPTIBLE)
                    // Still not done
                    return 1;

                mode_interruptible = null;

                keep_running = handle_error(error);
                
                if (mode_running)
                {
                    if (keep_running == 0)
                        set_running(false);
                }
                else
                {
                    shell_annunciators(-1, -1, -1, 0, -1, -1);
                    pending_command = CMD_NONE;
                }
                
                if (mode_running || keybuf_tail != keybuf_head)
                    return 1;
                else
                {
                    redisplay();

                    return 0;
                }
            }

            if (mode_running && !mode_getkey)
            {
                // We're running; queue up any keystrokes and invoke
                // continue_running() to keep the program moving along
                //
                if (key != 0)
                {
                    if (key == KEY_EXIT)
                    {
                        keybuf_tail = keybuf_head;
                        set_shift(false);
                        set_running(false);
                        pending_command = CMD_CANCELLED;
                        
                        return 0;
                    }

                    // Enqueue...
                    enqueued = 1;

                    if (!mode_shift && key == KEY_RUN)
                    {
                        keybuf_tail = keybuf_head;
                        set_running(false);
                        redisplay();

                        return 0;
                    }

                    if (((keybuf_head + 1) & 15) != keybuf_tail)
                    {
                        if (mode_shift)
                            key = -key;
                    
                        keybuf[keybuf_head] = key;
                        keybuf_head = (keybuf_head + 1) & 15;
                    }

                    set_shift(false);
                }

                continue_running();
                
                if ((mode_running && !mode_getkey && !mode_pause) || keybuf_tail != keybuf_head)
                    return 1;
                else
                {
                    if (mode_getkey)
                        // Technically, the program is still running, but we turn
                        // off the 'running' annunciator so that the user has some
                        // cue that they may now type. (Actually, I'm doing it
                        // purely because the HP-42S does it, too!)
                        //
                        shell_annunciators(-1, -1, -1, 0, -1, -1);
                    else if (!mode_pause)
                        redisplay();

                    return 0;
                }
            }

            // If we get here, mode_running must be false.
            // or a program is running but hanging in GETKEY;
            //

            if (keybuf_tail != keybuf_head)
            {
                // We're not running, or a program is waiting in GETKEY;
                // feed queued-up keystroke to keydown()
                //
                int oldshift = 0;
                int oldkey = keybuf[keybuf_tail];

                if (oldkey < 0)
                {
                    oldkey = -oldkey;
                    oldshift = 1;
                }
                
                keybuf_tail = (keybuf_tail + 1) & 15;
                // If we're in GETKEY mode, the 'running' annunciator is off;
                // see the code circa 30 lines back.
                // We now turn it back on since program execution resumes.
                //
                if (mode_getkey && mode_running)
                    shell_annunciators(-1, -1, -1, 1, -1, -1);

                // Feed the dequeued key to the usual suspects
                keydown(oldshift, oldkey);
                core_keyup();

                // We've just de-queued a key; may have to enqueue
                // one as well, if the user is actually managing to
                // type while we're unwinding the keyboard buffer
                //
                if (key != 0)
                {
                    if (((keybuf_head + 1) & 15) != keybuf_tail)
                    {
                        if (mode_shift)
                            key = -key;
                 
                        keybuf[keybuf_head] = key;
                        keybuf_head = (keybuf_head + 1) & 15;
                    }

                    set_shift(false);
                }

                return ((mode_running && !mode_getkey) || keybuf_head != keybuf_tail) ? 1: 0;
            }

            // No program is running, or it is running but waiting for a
            // keystroke (GETKEY); handle any new keystroke that has just come in
            //
            if (key != 0)
            {
                int shift = mode_shift ? 1 : 0;

                set_shift(false);
                
                if (mode_getkey && mode_running)
                    shell_annunciators(-1, -1, -1, 1, -1, -1);
                
                keydown(shift, key);
                
                if (repeating != 0)
                {
                    repeat = repeating;
                    repeating = 0;
                }

                return (mode_running && !mode_getkey) ? 1 : 0;
            }

            // Nothing going on at all!
            return 0;
        }

        public static void continue_running()
        {
            int error;

            prgm_start_time = DateTime.Now;

            while (shell_wants_cpu() == 0 && !prgm_block)
            {
                int cmd = 0;
                arg_struct arg = new arg_struct();
                oldpc = pc;
             
                if (pc == -1)
                    pc = 0;
                else if (pc >= prgms[current_prgm].size)
                {
                    pc = -1;
                    set_running(false);
                    return;
                }
                
                get_next_command(ref pc, ref cmd, arg, 1);

                if ((flags.trace_print != 0) && (flags.printer_exists != 0))
                    print_program_line(current_prgm, oldpc);

                mode_disable_stack_lift = false;

                error = cmd_array[cmd].handler(arg);
                
                if (mode_pause)
                {
                    shell_request_timeout3(1000);
                    return;
                }
                
                if (error == ERR_INTERRUPTIBLE)
                    return;
                
                if (handle_error(error) == 0)
                    return;
                
                if (mode_getkey)
                    return;
            }
        }

        public static int core_keyup()
        {
            if (mode_pause)
            {
	            // The only way this can happen is if they key in question was Shift
	            return 0;
            }

            int error = ERR_NONE;

            if (pending_command == CMD_LINGER1 || pending_command == CMD_LINGER2)
            {
	            pending_command = CMD_LINGER2;
	            return (mode_running || keybuf_head != keybuf_tail) ? 1 : 0;
            }

            if (pending_command == CMD_SILENT_OFF)
            {
	            shell_powerdown();
                
                // In prior Navigation implementation, options page was implemented here rather than power off
                // pending_command = CMD_NONE;
                // Main.load_option_page();
                
                return 0;
            }

            if (pending_command == CMD_NONE)
    	        return (mode_running || keybuf_head != keybuf_tail) ? 1 : 0;

            if (remove_program_catalog != 0)
            {
	                if (mode_transientmenu == MENU_CATALOG)
	                    set_menu(MENULEVEL_TRANSIENT, (int)MENU.NONE);
	                else if (mode_plainmenu == MENU_CATALOG)
	                    set_menu(MENULEVEL_PLAIN, (int)MENU.NONE);
	                
                remove_program_catalog = 0;
            }

            if (pending_command == CMD_CANCELLED || pending_command == CMD_NULL)
            {
	            pending_command = CMD_NONE;
	            redisplay();
	            return (mode_running || keybuf_head != keybuf_tail) ? 1 : 0;
            }

            mode_varmenu = pending_command == CMD_VMSTO || pending_command == CMD_VMSTO2
		            || pending_command == CMD_VMSOLVE || pending_command == CMD_VMEXEC;

            if (input_length > 0)
            {
	            // INPUT active
	            if (pending_command == CMD_RUN || pending_command == CMD_SST)
                {
	                int err = generic_sto(input_arg, (char)0);
	            
                    if (((flags.trace_print != 0) || (flags.normal_print != 0)) && (flags.printer_exists != 0))
                    {
		                char[] lbuf = new char[12];
                        char[] rbuf = new char[100];
		                int llen = 0;
                        int rlen;

		                string_copy(lbuf, ref llen, input_name, input_length);

		                lbuf[llen++] = '=';
		                rlen = vartype2string(reg_x, rbuf, 100);
		                
                        print_wide(lbuf, llen, rbuf, rlen);
	                }
	            
                    input_length = 0;
	                
                    if (err != ERR_NONE)
                    {
		                pending_command = CMD_NONE;
		            
                        display_error(err, 1);
		                redisplay();
		            
                        return (mode_running || keybuf_head != keybuf_tail) ? 1 : 0;
	                }
	            }
                else if (pending_command == CMD_GTO || pending_command == CMD_GTODOT
		        || pending_command == CMD_GTODOTDOT || pending_command == CMD_RTN)
	                // NOTE: set_running(true) also ends INPUT mode, so commands that
	                // cause program execution to start do not have to be handled here.
	                //
	                input_length = 0;
            }

            if (pending_command == CMD_VMEXEC)
            {
	            string_copy(reg_alpha, ref reg_alpha_length, pending_command_arg.val.text, pending_command_arg.length);
	            
                //goto do_run;      // Replicated do_run code below
                if (((flags.trace_print != 0) || (flags.normal_print != 0)) && (flags.printer_exists != 0))
                    print_command(pending_command, pending_command_arg);

                pending_command = CMD_NONE;

                if (pc == -1)
                    pc = 0;

                set_running(true);
                return 1;
            }

            if (pending_command == CMD_RUN)
            {
	        //do_run:
	            if (((flags.trace_print != 0) || (flags.normal_print != 0)) && (flags.printer_exists != 0))
	                print_command(pending_command, pending_command_arg);

	            pending_command = CMD_NONE;
	        
                if (pc == -1)
	                pc = 0;

    	        set_running(true);
	            return 1;
            }

            if (pending_command == CMD_SST)
            {
	            int cmd = 0;
	            arg_struct arg = new arg_struct();
	            oldpc = pc;
	        
                if (pc == -1)
	                pc = 0;

	            get_next_command(ref pc, ref cmd, arg, 1);

	            if (((flags.trace_print != 0) || (flags.normal_print != 0)) && (flags.printer_exists != 0))
	                print_program_line(current_prgm, oldpc);

	            mode_disable_stack_lift = false;

	            set_running(true);
	            error = cmd_array[cmd].handler(arg);

	            set_running(false);
	            mode_pause = false;
            }
            else
            {
	            if (((flags.trace_print != 0) || (flags.normal_print != 0)) && (flags.printer_exists != 0))
	                print_command(pending_command, pending_command_arg);

	            mode_disable_stack_lift = false;

	            error = cmd_array[pending_command].handler(pending_command_arg);
	            mode_pause = false;
            }

            if (error == ERR_INTERRUPTIBLE)
            {
	            shell_annunciators(-1, -1, -1, 1, -1, -1);
	            return 1;
            }

            handle_error(error);
            pending_command = CMD_NONE;
            
            if (!mode_getkey && !mode_pause)
	            redisplay();
                
            return ((mode_running && !mode_getkey && !mode_pause) || keybuf_head != keybuf_tail) ? 1 : 0;
        }

        public static void finish_alpha_prgm_line()
        {
            if (entered_string_length == 0)
            {
                pc = incomplete_saved_pc;
                prgm_highlight_row = incomplete_saved_highlight_row;
            }
            else
            {
                arg_struct arg = new arg_struct();
                int i;

                arg.type = ARGTYPE_STR;
                arg.length = (byte)entered_string_length;
                
                for (i = 0; i < entered_string_length; i++)
                    arg.val.text[i] = entered_string[i];
                
                store_command(pc, CMD_STRING, arg);
                prgm_highlight_row = 1;
            }

            mode_alpha_entry = false;
        }

        public static void fix_thousands_separators(char[] buf, ref int bufptr)
        {
            // First, remove the old separators...
            int i = 0;
            int j = 0;
            
            char dot = flags.decimal_point != 0 ? '.' : ',';
            char sep = flags.decimal_point != 0 ? ',' : '.';

            int intdigits = 0;
            int counting_intdigits = 1;
            int nsep;
            
            for (i = 0; i < bufptr; i++)
            {
                char c = buf[i];

                if (c != sep)
                    buf[j++] = c;

                if (c == dot || c == 24)
                    counting_intdigits = 0;
                else if (counting_intdigits != 0 && c >= '0' && c <= '9')
                    intdigits++;
            }

            // Now, put 'em back...
            if (flags.thousands_separators == 0)
            {
                bufptr = j;
                return;
            }

            nsep = (intdigits - 1) / 3;
            
            if (nsep == 0)
            {
                bufptr = j;
                return;
            }
            
            for (i = j - 1; i >= 0; i--)
                buf[i + nsep] = buf[i];
            
            j += nsep;
            
            for (i = 0; i < j; i++)
            {
                char c = buf[i + nsep];
                buf[i] = c;
            
                if (nsep > 0 && c >= '0' && c <= '9')
                {
                    if (intdigits % 3 == 1)
                    {
                        buf[++i] = sep;
                        nsep--;
                    }
                
                    intdigits--;
                }
            }

            bufptr = j;
        }

        public static int find_menu_key(int key)
        {
            switch (key)
            {
                case KEY_SIGMA: return 0;
                case KEY_INV: return 1;
                case KEY_SQRT: return 2;
                case KEY_LOG: return 3;
                case KEY_LN: return 4;
                case KEY_XEQ: return 5;
                default: return -1;
            }
        }

        public static void finish_xeq()
        {
            int prgm = 0;
            int pc = 0;
            int cmd;

            if (find_global_label(pending_command_arg, ref prgm, ref pc) != 0)
                cmd = CMD_NONE;
            else
                cmd = find_builtin(pending_command_arg.val.text, pending_command_arg.length);

            if (cmd == CMD_CLALLa)
            {
                mode_clall = true;
                mode_command_entry = false;
                pending_command = CMD_NONE;

                redisplay();
                return;
            }

            if (cmd == CMD_NONE)
                finish_command_entry(true);
            else
            {
                // Show the entered command in its XEQ "FOO"
                // form, briefly, just to confirm to the user
                // what they've entered; this will be replaced
                // by the matching built-in command after the
                // usual brief delay (timeout1() or the
                // explicit delay, below).
                //
                mode_command_entry = false;
                redisplay();

                if (cmd_array[cmd].argtype == ARG_NONE)
                {
                    pending_command = cmd;
                    pending_command_arg.type = ARGTYPE_NONE;
                    finish_command_entry(false);
                    return;
                }
                else
                {
                    shell_delay(250);

                    pending_command = CMD_NONE;
                    set_menu(MENULEVEL_COMMAND, (int)MENU.NONE);

                    if ((cmd == CMD_CLV || cmd == CMD_PRV) && flags.prgm_mode == 0 && vars_count == 0)
                    {
                        display_error(ERR_NO_VARIABLES, 0);
                        pending_command = CMD_NONE;
                        redisplay();
                    }
                    else
                        start_incomplete_command(cmd);
                    
                    return;
                }
            }
        }

        public static int find_builtin(char[] name, int namelen)
        {
            int i, j;

            for (i = 0; hp41_synonyms[i].cmd_id != CMD_NONE; i++)
            {
	            if (namelen != hp41_synonyms[i].namelen)
	                continue;

	            for (j = 0; j < namelen; j++)
	                if (name[j] != hp41_synonyms[i].name[j])
		                goto nomatch1;

    	        return hp41_synonyms[i].cmd_id;

	        nomatch1:;
            }

            for (i = 0; true; i++)
            {
	            if (i == CMD_OPENF && !core_settings.enable_ext_copan) i += 14;
	            if (i == CMD_DROP && !core_settings.enable_ext_bigstack) i++;
	            if (i == CMD_ACCEL && !core_settings.enable_ext_accel) i++;
	            if (i == CMD_LOCAT && !core_settings.enable_ext_locat) i++;
	            if (i == CMD_HEADING && !core_settings.enable_ext_heading) i++;
	            if (i == CMD_ADATE && !core_settings.enable_ext_time) i += 34;

                if (i == CMD_SENTINEL)
	                break;
	            
                if ((cmd_array[i].flags & FLAG_HIDDEN) != 0)
	                continue;
	        
                if (cmd_array[i].name_length != namelen)
	                continue;
	        
                for (j = 0; j < namelen; j++)
                {
	                byte c1, c2;

	                c1 = (byte)name[j];

	                if (c1 >= 130 && c1 != 138)
		                c1 &= 127;
	                
                    c2 = (byte)cmd_array[i].name[j];

	                if (c2 >= 130 && c2 != 138)
		                c2 &= 127;

	                if (c1 != c2)
		                goto nomatch2;
	            }

                return i;

	        nomatch2:;
            }

            return CMD_NONE;
        }

        public static void do_interactive(int command)
        {
            int err;

            if ((cmd_array[command].flags & (flags.prgm_mode != 0 ? FLAG_NO_PRGM : FLAG_PRGM_ONLY)) != 0)
            {
                display_error(ERR_RESTRICTED_OPERATION, 0);
                redisplay();
                return;
            }

            if (command == CMD_GOTOROW)
            {
                err = docmd_stoel(null);
                
                if (err != ERR_NONE)
                {
                    display_error(err, 1);
                    redisplay();
                    return;
                }
            }
            else if (command == CMD_A_THRU_F)
            {
                set_base(16);

                set_menu(MENULEVEL_APP, MENU_BASE_A_THRU_F);
                redisplay();
                return;
            }
            else if (command == CMD_CLALLa)
            {
                mode_clall = true;
                redisplay();
                return;
            }
            else if (command == CMD_CLV || command == CMD_PRV)
            {
                if (flags.prgm_mode == 0 && vars_count == 0)
                {
                    display_error(ERR_NO_VARIABLES, 0);
                    redisplay();
                    return;
                }
            }
            else if (command == CMD_SST && flags.prgm_mode != 0)
            {
                sst();
                
                redisplay();
                repeating = 1;
                repeating_shift = 1;
                repeating_key = KEY_DOWN;
                return;
            }
            else if (command == CMD_BST)
            {
                bst();
                
                if (flags.prgm_mode == 0)
                {
                    flags.prgm_mode = 1;
                    redisplay();
                    flags.prgm_mode = 0;
                    pending_command = CMD_CANCELLED;
                }
                else
                    redisplay();

                repeating = 1;
                repeating_shift = 1;
                repeating_key = KEY_UP;
                return;
            }

            if (flags.prgm_mode != 0 && (cmd_array[command].flags & FLAG_IMMED) == 0)
            {
                if (command == CMD_RUN)
                    command = CMD_STOP;

                if (cmd_array[command].argtype == ARG_NONE)
                {
                    arg_struct arg = new arg_struct();

                    arg.type = ARGTYPE_NONE;
                    store_command_after(ref pc, command, arg);
                    
                    if (command == CMD_END)
                        pc = 0;
                    
                    prgm_highlight_row = 1;
                    redisplay();
                }
                else
                {
                    incomplete_saved_pc = pc;
                    incomplete_saved_highlight_row = prgm_highlight_row;
                    
                    if (pc == -1)
                        pc = 0;
                    else if (prgms[current_prgm].text[pc] != CMD_END)
                        pc += get_command_length(current_prgm, pc);
                    
                    prgm_highlight_row = 1;
                    start_incomplete_command(command);
                }
            }
            else
            {
                if (cmd_array[command].argtype == ARG_NONE)
                    pending_command = command;
                else
                {
                    if (flags.prgm_mode != 0)
                    {
                        incomplete_saved_pc = pc;
                        incomplete_saved_highlight_row = prgm_highlight_row;
                    }

                    start_incomplete_command(command);
                }
            }
        }

        public static void start_alpha_prgm_line()
        {
            incomplete_saved_pc = pc;
            incomplete_saved_highlight_row = prgm_highlight_row;

            if (pc == -1)
                pc = 0;
            else if (prgms[current_prgm].text[pc] != CMD_END)
                pc += get_command_length(current_prgm, pc);
            
            prgm_highlight_row = 1;
            
            if (cmdline_row == 1)
                display_prgm_line(0, -1);
            
            entered_string_length = 0;
            mode_alpha_entry = true;
        }

        public static bool core_alpha_menu()
        {
            int menu = get_front_menu();
            return menu != (int)MENU.NONE && menu >= (int)MENU.ALPHA1 && menu <= (int)MENU.ALPHA_MISC2;
        }

        public static bool core_hex_menu()
        {
            int menu = get_front_menu();
            return menu != (int)MENU.NONE && menu == (int)MENU.BASE_A_THRU_F;
        }

        public static void set_alpha_entry(bool state)
        {
            mode_alpha_entry = state;
        }
    }
}
