﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        public static int math_asinh_acosh(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim, int do_asinh)
        {
            // (orig author note) TO DO: review; and deal with overflows in intermediate results
            Phloat are, aim, br, bphi;

            // a = z ^ 2 +- 1
            are = xre * xre - xim * xim;

            if (do_asinh != 0)
                are += 1;
            else
                are -= 1;
            
            aim = 2 * xre * xim;

            // b = sqrt(a)
            br = Phloat.sqrt(Phloat.hypot(are, aim));
            bphi = Phloat.atan2(aim, are) / 2;

            // a = b + x
            Phloat.sincos(bphi, ref aim, ref are);
            are = are * br + xre;
            aim = aim * br + xim;

            // y = log(a)
            yre = Phloat.log(Phloat.hypot(are, aim));
            yim = Phloat.atan2(aim, are);
            
            return ERR_NONE;
        }

        public static int math_atanh(Phloat xre, Phloat xim, ref Phloat yre, ref Phloat yim)
        {
            Phloat are;
            Phloat aim;
            Phloat bre;
            Phloat bim;
            Phloat cre;
            Phloat cim;
            Phloat h;

            // (orig author note) TO DO: review, and deal with overflows in intermediate results
            if (xim == 0 && (xre == 1 || xre == -1))
                return ERR_INVALID_DATA;

            // a = 1 + x
            are = 1 + xre;
            aim = xim;

            // b = 1 - x
            bre = 1 - xre;
            bim = -xim;

            // c = a / b
            h = Phloat.hypot(bre, bim);
            cre = (are * bre + aim * bim) / h / h;
            cim = (aim * bre - are * bim) / h / h;

            // y = log(c) / 2
            yre = Phloat.log(Phloat.hypot(cre, cim)) / 2;
            yim = Phloat.atan2(cim, cre) / 2;

            // Theoretically, you could go out of range, but in practice,
            // you can't get close enough to the critical values to cause
            // trouble.
            //
            return ERR_NONE;
        }

        public static int math_gamma(Phloat phx, ref Phloat phgamma)
        {
            if (phx == 0 || (phx < 0 && phx == Phloat.floor(phx)))
                return ERR_INVALID_DATA;

            phgamma = Phloat.gamma(phx);

            int inf = Phloat.p_isinf(phgamma);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    phgamma = inf < 0 ? Phloat.NEG_HUGE_PHLOAT : Phloat.POS_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            return ERR_NONE;
        }

        public static Phloat math_random()
        {
            Phloat n1 = new Phloat(new BCDFloat(9821, 0, 0, 0, 0, 0, 0, 1));
            Phloat n2 = new Phloat(new BCDFloat(2113, 2700, 0, 0, 0, 0, 0, 0));

            random_number = random_number * n1 + n2;
            random_number -= Phloat.floor(random_number);
            
            return random_number;
        }
    }
}
