//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Runtime.Serialization;


namespace WP7_42s
{
    public class Phloat
    {
        public static Phloat PI = new Phloat(new BCDFloat(3, 1415, 9265, 3589, 7932, 3846, 2643, 1));

        public static Phloat POS_HUGE_PHLOAT = new Phloat();
        public static Phloat NEG_HUGE_PHLOAT = new Phloat();
        public static Phloat POS_TINY_PHLOAT = new Phloat();
        public static Phloat NEG_TINY_PHLOAT = new Phloat();

        public BCDFloat bcd = new BCDFloat();

        #region Constructors

        // Static constructor
        static Phloat()
        {
            POS_HUGE_PHLOAT.bcd.d_[0] = 9999;
            POS_HUGE_PHLOAT.bcd.d_[1] = 9999;
            POS_HUGE_PHLOAT.bcd.d_[2] = 9999;
            POS_HUGE_PHLOAT.bcd.d_[3] = 9999;
            POS_HUGE_PHLOAT.bcd.d_[4] = 9999;
            POS_HUGE_PHLOAT.bcd.d_[5] = 9999;
            POS_HUGE_PHLOAT.bcd.d_[6] = 9000;
            POS_HUGE_PHLOAT.bcd.d_[7] = 2500;

            POS_TINY_PHLOAT.bcd.d_[0] = 1;
            POS_TINY_PHLOAT.bcd.d_[1] = 0;
            POS_TINY_PHLOAT.bcd.d_[2] = 0;
            POS_TINY_PHLOAT.bcd.d_[3] = 0;
            POS_TINY_PHLOAT.bcd.d_[4] = 0;
            POS_TINY_PHLOAT.bcd.d_[5] = 0;
            POS_TINY_PHLOAT.bcd.d_[6] = 0;
            int tmp = 2499;
            POS_TINY_PHLOAT.bcd.d_[7] = (ushort)-tmp;

            NEG_HUGE_PHLOAT = 0 - POS_HUGE_PHLOAT;
            NEG_TINY_PHLOAT = 0 - POS_TINY_PHLOAT;
        }

        // Instance constructors
        public Phloat() { }

        public Phloat(int numer, int denom)
        {
            BCDFloat n = new BCDFloat(numer);
            BCDFloat d = new BCDFloat(denom);

            BCDFloat.div(n, d, bcd);
        }

        public Phloat(int i)
        {
            bcd = new BCDFloat(i);
        }

        public Phloat(Int64 i)
        {
                bcd = new BCDFloat(i);
        }

        public Phloat(double d)
        {
            bcd = double2bcd(d);
        }

        public Phloat(Phloat p)
        {
            bcd = new BCDFloat(p.bcd);
        }

        public Phloat(BCDFloat b)
        {
            bcd = new BCDFloat(b.d_);
        }

        #endregion
        
        #region Operators

        public static bool operator ==(Phloat s, Phloat p)
        {
            if ((object)s == null && (object)p == null)
                return true;
            else if ((object)s == null && (object)p != null)
                return false;
            else if ((object)s != null && (object)p == null)
                return false;
            else
                return BCDFloat.equal(s.bcd, p.bcd);
        }

        public static bool operator ==(int x, Phloat y)
        {
            BCDFloat bx = new BCDFloat(x);
            return BCDFloat.equal(bx, y.bcd);
        }

        public static bool operator ==(Phloat y, int x)
        {
            BCDFloat bx = new BCDFloat(x);
            return BCDFloat.equal(bx, y.bcd);
        }

        public static bool operator !=(Phloat s, Phloat p)
        {
            if ((object)s == null && (object)p == null)
                return false;
            else if ((object)s == null && (object)p != null)
                return true;
            else if ((object)s != null && (object)p == null)
                return true;
            else
                return !BCDFloat.equal(s.bcd, p.bcd);
        }

        public static bool operator !=(int x, Phloat y)
        {
            BCDFloat bx = new BCDFloat(x);
            return !BCDFloat.equal(bx, y.bcd);
        }

        public static bool operator !=(Phloat y, int x)
        {
            BCDFloat bx = new BCDFloat(x);
            return !BCDFloat.equal(bx, y.bcd);
        }

        public static bool operator <(Phloat s, Phloat p)
        {
            return BCDFloat.lt(s.bcd, p.bcd);
        }

        public static bool operator <=(Phloat s, Phloat p)
        {
            return BCDFloat.le(s.bcd, p.bcd);
        }

        public static bool operator >(Phloat s, Phloat p)
        {
            return BCDFloat.gt(s.bcd, p.bcd);
        }

        public static bool operator >=(Phloat s, Phloat p)
        {
            return BCDFloat.ge(s.bcd, p.bcd);
        }


        public static Phloat operator *(Phloat s, Phloat p)
        {
            Phloat res = new Phloat();
            BCDFloat.mul(s.bcd, p.bcd, res.bcd);
            return res;
        }

        public static Phloat operator *(int x, Phloat y)
        {
            Phloat res = new Phloat();
            BCDFloat bx = new BCDFloat(x);

            BCDFloat.mul(bx, y.bcd, res.bcd);
            return res;
        }

        public static Phloat operator *(Phloat y, int x)
        {
            Phloat res = new Phloat();
            BCDFloat bx = new BCDFloat(x);

            BCDFloat.mul(bx, y.bcd, res.bcd);
            return res;
        }

        public static Phloat operator /(Phloat s, Phloat p)
        {
            Phloat res = new Phloat();
            BCDFloat.div(s.bcd, p.bcd, res.bcd);
            return res;
        }

        public static Phloat operator /(int x, Phloat y)
        {
            Phloat res = new Phloat();
            BCDFloat bx = new BCDFloat(x);

            BCDFloat.div(bx, y.bcd, res.bcd);
            return res;
        }

        public static Phloat operator /(Phloat y, int x)
        {
            Phloat res = new Phloat();
            BCDFloat bx = new BCDFloat(x);

            BCDFloat.div(y.bcd, bx, res.bcd);
            return res;
        }

        public static Phloat operator /(double x, Phloat y)
        {
            Phloat res =  new Phloat();
            BCDFloat bx = double2bcd(x);

            BCDFloat.div(bx, y.bcd, res.bcd);
            return res;
        }

        public static Phloat operator +(Phloat s, Phloat p)
        {
            Phloat res = new Phloat();
            BCDFloat.add(s.bcd, p.bcd, res.bcd);
            return res;
        }

        public static Phloat operator +(int x, Phloat y)
        {
            Phloat res = new Phloat();
            BCDFloat bx = new BCDFloat(x);

            BCDFloat.add(bx, y.bcd, res.bcd);
            return res;
        }

        public static Phloat operator +(Phloat y, int x)
        {
            Phloat res = new Phloat();
            BCDFloat bx = new BCDFloat(x);

            BCDFloat.add(bx, y.bcd, res.bcd);
            return res;
        }

        public static Phloat operator -(Phloat s, Phloat p)
        {
            Phloat res = new Phloat();
            BCDFloat.sub(s.bcd, p.bcd, res.bcd);
            return res;
        }

        public static Phloat operator -(int x, Phloat y)
        {
            Phloat res = new Phloat();
            BCDFloat bx = new BCDFloat(x);

            BCDFloat.sub(bx, y.bcd, res.bcd);
            return res;
        }

        public static Phloat operator -(Phloat y, int x)
        {
            Phloat res = new Phloat();
            BCDFloat bx = new BCDFloat(x);

            BCDFloat.sub(y.bcd, bx, res.bcd);
            return res;
        }

        public static Phloat operator -(Phloat y)
        {
            Phloat res = new Phloat(y);
            res.bcd.negate();
            return res;
        }

        public static Phloat operator ++(Phloat s)
        {
            Phloat r = new Phloat(s);
            r += new Phloat(1);
            return r;
        }

        public static Phloat operator --(Phloat s)
        {
            Phloat r = new Phloat(s);
            r -= new Phloat(1);
            return r;
        }


        public static implicit operator Phloat(int i)
        {
            return new Phloat(i);
        }

        public static implicit operator Phloat(Int64 i)
        {
            return new Phloat(i);
        }

        public static implicit operator Phloat(double d)
        {
            return new Phloat(double2bcd(d));
        }

        #endregion

        #region Static
        
        public static BCDFloat double2bcd(double dd, bool round = false)
        {
            BCDFloat res = new BCDFloat(dd);

            if (round && !res.isSpecial())
            {
	            // This is used when converting programs from a Free42 Binary state
	            // file. Number literals in programs are rounded to 12 digits, so
	            // what you see really is what you get; without this hack, you'd
	            // get stuff like 0.9 turning into 0.8999999+ but still *looking*
	            // like 0.9!
	            
                int i;

	            for (i = 4; i < App.P; i++)
	                res.d_[i] = 0;

	            ushort s = res.d_[0];
	            ushort d;

	            if (s < 10)
	                d = 10;
	            else if (s < 100)
	                d = 100;
	            else if (s < 1000)
	                d = 1000;
	            else
	                d = 10000;

	            s = res.d_[3];
	            ushort r = (ushort) (s % d);

	            if (r >= (d >> 1))
	                s += d;

	            s -= r;

	            bool carry = s >= 10000;
	            if (carry)
	                s -= 10000;

	            res.d_[3] = s;
	            for (i = 2; carry && i >= 0; i--)
                {
	                s = (ushort) (res.d_[i] + 1);

	                carry = s >= 10000;
	                if (carry)
		                s -= 10000;

	                res.d_[i] = s;
	            }

	            if (carry)
                {
	                for (i = 3; i >= 0; i--)
		                res.d_[i + 1] = res.d_[i];

	                res.d_[0] = 1;
	                res.d_[App.P]++;
	                // No need to check if the exponent is overflowing; the range
	                // of 'double' is too small to cause such problems here.
	            }
            }

            return res;
        }

        public static double bcd2double(BCDFloat b, bool old_bcd)
        {
            if (old_bcd)
	            bcdfloat_old2new(b.d_);

            bool neg = b.neg();

            if (b.isNan())
    	        return double.NaN;
            else if (b.isInf())
    	        return neg ? double.NegativeInfinity : double.PositiveInfinity;

            if (b.d_[0] == 0)
	            return 0;

            // Get rid of neg, nan, and inf flags in exponent, and extend
            // sign of the remaining 13 bits
            short exp = (short) ((((short) b.d_[App.P]) << 3) >> 3);

            // Make sure we get exact results for integers
            if (exp > 0 && exp <= 4)
            {
	            int j;
	            for (j = exp; j < App.P; j++)
	                if (b.d_[j] != 0)
		                goto noninteger;
	            
                Int64 n = 0;
	            for (j = 0; j < exp; j++)
	                n = n * 10000 + b.d_[j];
	            
                return (double) (neg ? -n : n);
            }

            // TO DO: Using a table-based conversion algorithm, like the one I use in
            // the binary version of string2phloat, I could get better accuracy.
            // The current code returns an inexact result for 23.5, for instance
            // (VC++), while the number *is* exactly representable in a 'double'.

            noninteger:
            double res = 0;
            for (int i = 0; i < App.P; i++)
	            res = res * 10000 + b.d_[i];

            if (neg)
	            res = -res;

            return res * Math.Pow(10000.0, (double) (exp - App.P));
        }

        public static void bcdfloat_old2new(ushort[] bcd)
        {
            // Convert old (<= 1.4.51) BCDFloat, where NaN is signalled by
            // (exp & 0x7FFF) == 0x3000, and Infinity is signalled by
            // (exp & 0x7FFF) == 0x3FFF, to the new (>= 1.4.52) BCDFloat, where NaN is
            // signalled by (exp & 0x4000) != 0 and Infinity is signalled by
            // (exp & 0x2000) != 0 (and the exponent field is 2 bits narrower).

            short uexp = (short) (bcd[App.P] & 0x7FFF);

            if (uexp == 0x3000)                 // NaN
    	        bcd[App.P] = 0x4000;
            else if (uexp == 0x3FFF)            // Infinity
	            bcd[App.P] = (ushort) ((bcd[App.P] & 0x8000) | 0x2000);
            else
    	        bcd[App.P] = (ushort) (bcd[App.P] & 0x9FFF);
        }

        public static int string2phloat(char[] buf, int buflen, ref Phloat d)
        {
            char[] tmp = new char[buflen];
            Array.Copy(buf, tmp, buflen);

            return string2phloat(new string(tmp), ref d);
        }

        public static int string2phloat(string buf, ref Phloat d)
        {
            /* Convert string to phloat.
             * Return values:
             * 0: no error
             * 1: positive overflow
             * 2: negative overflow
             * 3: positive underflow
             * 4: negative underflow
             * 5: other error
             */
            // First, convert from HP-42S format to bcdfloat format:
            // strip thousands separators, convert comma to dot if
            // appropriate, and convert char(24) to 'E'.
            // Also, reject numbers with more than 12 digits in the mantissa.
            
            char[] buf2 = new char[100];
            int buflen2 = 0;
            
            char sep = (App.flags.decimal_point != 0) ? ',' : '.';

            int mantdigits = 0;
            bool in_mant = true;
            bool zero = true;
            
            for (int i = 0; i < buf.Length; i++)
            {
    	        char c = buf[i];

	            if (c == sep)
	                continue;

                if (c == ',')
	                c = '.';
    	        else if (c == 24)
                {
	                c = 'E';
	                in_mant = false;
	            } else if (in_mant && c >= '0' && c <= '9')
                {
	                if (++mantdigits > 12)
		                return 5;

	                if (c != '0')
    		            zero = false;
	            }

	            buf2[buflen2++] = c;
            }

            if (!in_mant && mantdigits == 0)
            {
	            // A number like "E4"; not something the HP-41 or HP42S (or Free42, for
	            // that matter) will actually allow you to enter into a program, but
	            // the real calcs do accept this kind of thing (synthetically), and
	            // supply a mantissa of 1 (just like when you start number entry by
	            // pressing EEX (HP-41) or E (HP-42S).
	            for (int i = buflen2 - 1; i >= 0; i--)
	                buf2[i + 1] = buf2[i];

	            buf2[0] = '1';
	            buflen2++;
            }

            buf2[buflen2] = (char) 0;

            BCDFloat b = new BCDFloat(new string(buf2, 0, buflen2));
            
            if (b.isInf())
    	        return b.neg() ? 2 : 1;

            if (!zero && b.isZero())
	            return b.neg() ? 4 : 3;

            d.bcd = b;
            return 0;
        }

        public static int p_isinf(Phloat p)
        {
            if (p.bcd.isInf())
	            return p.bcd.neg() ? -1 : 1;
            else
    	        return 0;
        }

        public static int p_isnan(Phloat p)
        {
            return p.bcd.isNan() ? 1 : 0;
        }

        public static Int64 mant(BCDFloat b)
        {
            Int64 m = 0;
            int e = b.exp();

            if (e > App.P)
	            e = App.P;

            for (int i = 0; i < e; i++)
	            m = m * 10000 + b.d_[i];

            if (b.neg())
	            m = -m;

            return m;
        }

        public static byte[] to_byte_array(Phloat p)
        {
            int index = 0;
            byte[] b = new byte[App.sizeof_Phloat];

            for (int i = 0; i < App.sizeof_Phloat / 2; i++)
            {
                b[index++] = (byte)(p.bcd.d_[i] & 0xff);
                b[index++] = (byte)((p.bcd.d_[i] >> 8) & 0xff);
            }

            return b;
        }

        public static int from_byte_array(Phloat p, byte[] b)
        {
            int index = 0;

            if (b.Length == App.sizeof_Phloat)
            {

                for (int i = 0; i < App.sizeof_Phloat / 2; i++)     // TODO: Verify byte order, index 0, low byte first
                {
                    byte b1 = b[index++];
                    byte b2 = b[index++];

                    p.bcd.d_[i] = (ushort)((b2 << 8) + b1);
                }
            }

            return index;
        }

        public static char to_char(Phloat p)
        {
            return (char) mant(p.bcd);
        }

        public static int to_int(Phloat p)
        {
            return (int) mant(p.bcd);
        }

        public static Int64 to_int8(Phloat p)
        {
            return (Int64) mant(p.bcd);
        }

        public static double to_double(Phloat p)
        {
            return bcd2double(p.bcd, false);
        }

        public static int to_digit(Phloat p)
        {
            return BCD.ifloor(BCD.trunc(bcdmath.fmod(new BCD(p.bcd), new BCD(10))));
        }

        public static Phloat sin(Phloat p)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.sin(new BCD(p.bcd))._v;
            return res;
        }

        public static Phloat cos(Phloat p)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.cos(new BCD(p.bcd))._v;
            return res;
        }

        public static Phloat tan(Phloat p)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.tan(new BCD(p.bcd))._v;
            return res;
        }

        public static Phloat asin(Phloat p)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.asin(new BCD(p.bcd))._v;
            return res;
        }

        public static Phloat acos(Phloat p)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.acos(new BCD(p.bcd))._v;
            return res;
        }

        public static Phloat atan(Phloat p)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.atan(new BCD(p.bcd))._v;
            return res;
        }

        public static void sincos(Phloat phi, ref Phloat s, ref Phloat c)
        {
            BCD p = new BCD(phi.bcd);

            s.bcd = bcdmath.sin(p)._v;
            c.bcd = bcdmath.cos(p)._v;
        }

        public static Phloat hypot(Phloat x, Phloat y)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.hypot(new BCD(x.bcd), new BCD(y.bcd))._v;
            return res;
        }

        public static Phloat atan2(Phloat x, Phloat y)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.atan2(new BCD(x.bcd), new BCD(y.bcd))._v;
            return res;
        }

        public static Phloat sinh(Phloat p)
        {
            // (exp(x)-exp(-x))/2
            BCDFloat temp1 = bcdmath.exp(new BCD(p.bcd))._v;
            BCDFloat temp2 = bcdmath.exp(new BCD((new Phloat(0) - p).bcd))._v;

            BCDFloat temp3 = new BCDFloat();
            BCDFloat.sub(temp1, temp2, temp3);

            BCDFloat two = new BCDFloat(2);

            Phloat res = new Phloat();
            BCDFloat.div(temp3, two, res.bcd);

            return res;
        }

        public static Phloat cosh(Phloat p)
        {
            // (exp(x)+exp(-x))/2
            BCDFloat temp1 = bcdmath.exp(new BCD(p.bcd))._v;
            BCDFloat temp2 = bcdmath.exp(new BCD((new Phloat(0) - p).bcd))._v;

            BCDFloat temp3 = new BCDFloat();
            BCDFloat.add(temp1, temp2, temp3);

            BCDFloat two = new BCDFloat(2);

            Phloat res = new Phloat();
            BCDFloat.div(temp3, two, res.bcd);

            return res;
        }

        public static Phloat tanh(Phloat p)
        {
            // (exp(x)-exp(-x))/(exp(x)+exp(-x))
            BCDFloat temp1 = bcdmath.exp(new BCD(p.bcd))._v;
            BCDFloat temp2 = bcdmath.exp(new BCD((new Phloat(0) - p).bcd))._v;

            BCDFloat temp3 = new BCDFloat();
            BCDFloat.sub(temp1, temp2, temp3);

            if (temp3.isInf())
	            return new Phloat(temp3.neg() ? -1 : 1);

            BCDFloat temp4 = new BCDFloat();
            BCDFloat.add(temp1, temp2, temp4);

            Phloat res = new Phloat();
            BCDFloat.div(temp3, temp4, res.bcd);

            return res;
        }

        public static Phloat asinh(Phloat p)
        {
            // log(sqrt(x^2+1)+x)
            BCDFloat temp1 = new BCDFloat();
            BCDFloat.mul(p.bcd, p.bcd, temp1);

            BCDFloat temp2 = new BCDFloat();
            BCDFloat one = new BCDFloat(1);
            BCDFloat.add(temp1, one, temp2);

            temp1 = BCD.sqrt(new BCD(temp2))._v;
            BCDFloat.add(temp1, p.bcd, temp2);

            Phloat res = new Phloat();
            res.bcd = bcdmath.log(new BCD(temp2))._v;

            return res;
        }

        public static Phloat acosh(Phloat p)
        {
            // log(sqrt(x^2-1)+x)
            BCDFloat temp1 = new BCDFloat();
            BCDFloat.mul(p.bcd, p.bcd, temp1);

            BCDFloat temp2 = new BCDFloat();
            BCDFloat one = new BCDFloat(1);
            BCDFloat.sub(temp1, one, temp2);

            temp1 = BCD.sqrt(new BCD(temp2))._v;
            BCDFloat.add(temp1, p.bcd, temp2);

            Phloat res = new Phloat();
            res.bcd = bcdmath.log(new BCD(temp2))._v;
            return res;
        }

        public static Phloat atanh(Phloat p)
        {
            // log((1+x)/(1-x))/2
            BCDFloat one = new BCDFloat(1);

            BCDFloat temp1 = new BCDFloat();
            BCDFloat temp2 = new BCDFloat();
            BCDFloat temp3 = new BCDFloat();

            BCDFloat.add(one, p.bcd, temp1);
            BCDFloat.sub(one, p.bcd, temp2);
            BCDFloat.div(temp1, temp2, temp3);

            temp1 = bcdmath.log(new BCD(temp3))._v;

            Phloat res = new Phloat();
            BCDFloat two = new BCDFloat(2);

            BCDFloat.div(temp1, two, res.bcd);

            return res;
        }

        public static Phloat log(Phloat p)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.log(new BCD(p.bcd))._v;
            return res;
        }

        public static Phloat log1p(Phloat p)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.ln1p(new BCD(p.bcd))._v;
            return res;
        }

        public static Phloat log10(Phloat p)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.log10(new BCD(p.bcd))._v;
            return res;
        }

        public static Phloat exp(Phloat p)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.exp(new BCD(p.bcd))._v;
            return res;
        }

        public static Phloat expm1(Phloat p)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.expm1(new BCD(p.bcd))._v;
            return res;
        }

        public static Phloat gamma(Phloat p)
        {
            --p;
            Phloat res = new Phloat();
            res.bcd = bcdmath.gammaFactorial(new BCD(p.bcd))._v;
            return res;
        }

        public static Phloat sqrt(Phloat p)
        {
            Phloat res = new Phloat();
            res.bcd = BCD.sqrt(new BCD(p.bcd))._v;
            return res;
        }

        public static Phloat fmod(Phloat x, Phloat y)
        {
            Phloat res = new Phloat();
            res.bcd = bcdmath.fmod(new BCD(x.bcd), new BCD(y.bcd))._v;
            return res;
        }

        public static Phloat fabs(Phloat p)
        {
            Phloat res = new Phloat(p);

            if (!res.bcd.isNan() && res.bcd.neg())
	            res.bcd.negate();

            return res;
        }

        public static Phloat pow(Phloat x, Phloat y)
        {
            Phloat res = new Phloat();

            if (!y.bcd.isSpecial())
            {
	            int iy = BCDFloat.ifloor(y.bcd);

	            BCDFloat by = new BCDFloat(iy);

	            if (BCDFloat.equal(y.bcd, by))
                {
	                res.bcd = BCD.pow(new BCD(x.bcd), iy)._v;
	                return res;
	            }
            }

            res.bcd = bcdmath.pow(new BCD(x.bcd), new BCD(y.bcd))._v;

            return res;
        }

        public static Phloat floor(Phloat p)
        {
            Phloat res = new Phloat();
            BCDFloat.floor(p.bcd, ref res.bcd);
            return res;
        }

        public static int phloat2string(Phloat pd, char[] buf, int buflen, int base_mode, int digits, int dispmode, int thousandssep)
        {
            int chars_so_far = 0;

            if (p_isnan(pd) != 0)
            {
	            App.string2buf(buf, buflen, ref chars_so_far, "<Not a Number>".ToCharArray(), 14);
	            return chars_so_far;
            }

            if (p_isinf(pd) != 0)
            {
	            App.char2buf(buf, buflen, ref chars_so_far, '<');

	            if (pd < new Phloat(0))
	                App.char2buf(buf, buflen, ref chars_so_far, '-');

	            App.string2buf(buf, buflen, ref chars_so_far, "Infinity>".ToCharArray(), 9);
	            return chars_so_far;
            }

            // base_mode: 0=only decimal, 1=all bases, 2=decimal or binary (SHOW)
            int _base = App.get_base();

            if (base_mode == 1 && _base != 10 || base_mode == 2 && _base == 2)
            {
	            Int64 n;
	            int inexact;
                int shift;

	            char[] binbuf = new char[36];
	            int binbufptr = 0;

	            if (pd > 34359738367.0 || pd < -34359738368.0)
                {
	                if (base_mode == 2)
		                goto decimal_after_all;
	                else
                    {
		                App.string2buf(buf, buflen, ref chars_so_far, "<Too Big>".ToCharArray(), 9);
		                return chars_so_far;
	                }
	            }

	            n = to_int8(pd);
	            inexact = (base_mode == 1 && pd != n) ? 1 : 0;

	            n &= 0xfffffffff;
	            shift = _base == 2 ? 1 : _base == 8 ? 3 : 4;

	            while (n != 0)
                {
	                int digit = (int) (n & (_base - 1));
	                char c = (char) (digit < 10 ? '0' + digit : 'A' + digit - 10);

	                binbuf[binbufptr++] = c;
	                n >>= shift;
	            }

	            if (binbufptr == 0)
	                binbuf[binbufptr++] = '0';

	            while (binbufptr > 0)
	                App.char2buf(buf, buflen, ref chars_so_far, binbuf[--binbufptr]);

	            if (inexact != 0)
	                App.char2buf(buf, buflen, ref chars_so_far, (char) (App.flags.decimal_point != 0 ? '.' : ','));

	            return chars_so_far;

	            decimal_after_all:;
            }

            char[] bcd_mantissa = new char[16];
            bcd_mantissa = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0".ToCharArray();

            int bcd_exponent = 0;
            int bcd_mantissa_sign = 0;

            if (pd < 0)
            {
	            pd = -pd;
	            bcd_mantissa_sign = 1;
            }

            if (pd != 0)
            {
    	        int offset;
                int pos = 0;

	            if (pd.bcd.d_[0] >= 1000)
	                offset = 0;
	            else if (pd.bcd.d_[0] >= 100)
	                offset = 1;
	            else if (pd.bcd.d_[0] >= 10)
	                offset = 2;
	            else
	                offset = 3;

	            bcd_exponent = pd.bcd.exp() * 4 - 1 - offset;

	            for (int i = 0; i < 5; i++)
                {
	                short s = (short) pd.bcd.d_[i];

	                for (int j = 0; j < 4; j++)
                    {
		                if (pos == 16)
		                    break;

		                if (offset == 0)
		                    bcd_mantissa[pos++] = (char) (s / 1000);
		                else
		                    offset--;

    		            s = (short) ((s % 1000) * 10);
	                }
	            }
            }

            if (dispmode == 0 || dispmode == 3)
            {
                // FIX and ALL modes

                char[] norm_ip = new char[12];
                char[] norm_fp = new char[27];

                int i;
                int int_digits;
                int frac_digits;

                int digits2;

                if (dispmode == 0)
                    digits2 = digits;
                else
                    digits2 = 11;

                for (i = 0; i < 12; i++)
                    norm_ip[i] = (char)0;

                for (i = 0; i < 27; i++)
                    norm_fp[i] = (char)0;

                if (bcd_exponent > 11 || -bcd_exponent > digits2 + 1)
                    return _do_sci(buf, buflen, chars_so_far, bcd_mantissa, bcd_mantissa_sign, bcd_exponent, dispmode, digits);

                for (i = 0; i < 16; i++)
                {
                    if (i <= bcd_exponent)
                        norm_ip[11 - bcd_exponent + i] = bcd_mantissa[i];
                    else
                        norm_fp[-1 - bcd_exponent + i] = bcd_mantissa[i];
                }

                if (dispmode == 0)
                {
                    // NOTE: I don't simply round norm_fp[digits], because of a
                    // circumstance that cannot happen on a real HP-42S: there
                    // may not be enough positions in the display to show a number
                    // at the requested accuracy (e.g. 123456123456.5 at FIX 01)
                    // so I first calculate how many positions I actually have
                    // available, and then I round to that. As a result of this
                    // rounding, the number may overflow the FIX representation,
                    // in which case I fall back on SCI.

                    int carry;
                    int visdigits = 11 - bcd_exponent;

                    if (visdigits > digits)
                        visdigits = digits;

                    carry = norm_fp[visdigits] >= 5 ? 1 : 0;

                    for (i = visdigits; i < 27; i++)
                        norm_fp[i] = (char)0;

                    if (carry == 0)
                        goto done_rounding;

                    for (i = visdigits - 1; i >= 0; i--)
                    {
                        char c = (char)(norm_fp[i] + 1);

                        if (c < 10)
                        {
                            norm_fp[i] = c;
                            goto done_rounding;
                        }
                        else
                            norm_fp[i] = (char)(c - 10);
                    }

                    for (i = 11; i >= 0; i--)
                    {
                        char c = (char)(norm_ip[i] + 1);

                        if (c < 10)
                        {
                            norm_ip[i] = c;
                            goto done_rounding;
                        }
                        else
                            norm_ip[i] = (char)(c - 10);
                    }

                    // If we get here, the carry went past the 12th integer digit,
                    // and we have to use SCI mode instead.

                    return _do_sci(buf, buflen, chars_so_far, bcd_mantissa, bcd_mantissa_sign, bcd_exponent, dispmode, digits);

                done_rounding: ;
                }
                else
                {
                    // ALL mode: for HP-42S compatibility, round to 12
                    // digits before proceeding.

                    int f = 1000;

                    for (i = 0; i < 39; i++)
                    {
                        char c = i < 12 ? norm_ip[i] : norm_fp[i - 12];

                        if (c != 0 && f == 1000)
                            f = i;

                        if (i == f + 12)
                        {
                            int carry = c >= 5 ? 1 : 0;

                            if (carry != 0)
                            {
                                int j;

                                for (j = i - 1; j >= 0; j--)
                                {
                                    char c2 = j < 12 ? norm_ip[j] : norm_fp[j - 12];
                                    c2++;

                                    if (c2 < 10)
                                        carry = 0;
                                    else
                                    {
                                        c2 -= (char)10;
                                        carry = 1;
                                    }

                                    if (j < 12)
                                        norm_ip[j] = c2;
                                    else
                                        norm_fp[j - 12] = c2;

                                    if (carry == 0)
                                        break;
                                }

                                if (carry != 0)
                                    // Rounding is making the integer part 13 digits
                                    // long; must go to SCI mode.
                                    //
                                    return _do_sci(buf, buflen, chars_so_far, bcd_mantissa, bcd_mantissa_sign, bcd_exponent, dispmode, digits);
                            }
                        }

                        if (i >= f + 12)
                        {
                            if (i < 12)
                                norm_ip[i] = (char)0;
                            else
                                norm_fp[i - 12] = (char)0;
                        }
                    }
                }

                // Check if the number is still within bounds for FIX or ALL
                if (pd != 0)
                {
                    // Make sure that nonzero numbers are not
                    // displayed as zero because of the rounding.

                    for (i = 0; i < 12; i++)
                        if (norm_ip[i] != 0)
                            goto fix_ok;

                    for (i = 0; i < digits2; i++)
                        if (norm_fp[i] != 0)
                            goto fix_ok;

                    // Uh-oh, the number is nonzero, but its rounded representation
                    // is zero. That's not good; use SCI mode instead.
                    return _do_sci(buf, buflen, chars_so_far, bcd_mantissa, bcd_mantissa_sign, bcd_exponent, dispmode, digits);

                fix_ok:

                    if (dispmode == 3)
                    {
                        // Make sure we're not throwing away anything in ALL mode
                        for (i = 11; i < 27; i++)
                            if (norm_fp[i] != 0)
                                return _do_sci(buf, buflen, chars_so_far, bcd_mantissa, bcd_mantissa_sign, bcd_exponent, dispmode, digits);
                    }
                }

                int_digits = 1;

                for (i = 0; i < 12; i++)
                    if (norm_ip[i] != 0)
                    {
                        int_digits = 12 - i;
                        break;
                    }

                if (bcd_mantissa_sign != 0)
                    App.char2buf(buf, buflen, ref chars_so_far, '-');

                for (i = int_digits - 1; i >= 0; i--)
                {
                    if (thousandssep != 0 && i % 3 == 2 && i != int_digits - 1)
                        App.char2buf(buf, buflen, ref chars_so_far, (char)(App.flags.decimal_point != 0 ? ',' : '.'));

                    App.char2buf(buf, buflen, ref chars_so_far, (char)('0' + norm_ip[11 - i]));
                }

                if (dispmode == 0)
                    frac_digits = digits;
                else
                {
                    frac_digits = 0;

                    for (i = 0; i < 27; i++)
                        if (norm_fp[i] != 0)
                            frac_digits = i + 1;
                }

                if (frac_digits + int_digits > 12)
                    frac_digits = 12 - int_digits;

                if (frac_digits > 0 || (dispmode == 0 && thousandssep != 0))
                {
                    App.char2buf(buf, buflen, ref chars_so_far, (char)(App.flags.decimal_point != 0 ? '.' : ','));

                    for (i = 0; i < frac_digits; i++)
                        App.char2buf(buf, buflen, ref chars_so_far, (char)('0' + norm_fp[i]));
                }

                return chars_so_far;
            }
            else
                return _do_sci(buf, buflen, chars_so_far, bcd_mantissa, bcd_mantissa_sign, bcd_exponent, dispmode, digits);
        }

        public static int _do_sci(char[] buf, int buflen, int chars_so_far, char[] bcd_mantissa, int bcd_mantissa_sign, int bcd_exponent, int dispmode, int digits)
        {
            // SCI and ENG modes
            // Also fall-through from FIX and ALL

            int m_digits;
            int carry;
            char[] norm_mantissa = new char[16];
            int norm_exponent;
            int e3;
            int i;

            for (i = 0; i < 16; i++)
                norm_mantissa[i] = bcd_mantissa[i];

            norm_exponent = bcd_exponent;

            if (dispmode == 3)
            {
            // Round to 12 digits before doing anything else;
            // this is needed to handle mantissas like 9.99999999999999,
            // which would otherwise end up getting displayed as
            // 10.0000000000 instead of 10.
            //

                sci_all_round:

                carry = norm_mantissa[12] >= 5 ? 1 : 0;

                for (i = 12; i < 16; i++)
                    norm_mantissa[i] = (char)0;

                if (carry != 0)
                {
                    for (i = 11; i >= 0; i--)
                    {
                        char c = (char)(norm_mantissa[i] + carry);

                        if (c < 10)
                        {
                            norm_mantissa[i] = c;
                            carry = 0;
                            break;
                        }
                        else
                        {
                            norm_mantissa[i] = (char)(c - 10);
                            carry = 1;
                        }
                    }
                }

                if (carry != 0)
                {
                    // Don't round by an additional digit: that would mean
                    // we're rounding the same number twice, which is bad
                    // (think about what happens when you round one digit off
                    // 0.45, twice -- you get first 0.5, then 1... Oops).
                    // So, we start over.

                    for (i = 0; i < 15; i++)
                        norm_mantissa[i + 1] = bcd_mantissa[i];

                    norm_mantissa[0] = (char)0;
                    norm_exponent = bcd_exponent + 1;

                    goto sci_all_round;
                }

                m_digits = 0;

                for (i = 11; i >= 0; i--)
                    if (norm_mantissa[i] != 0)
                    {
                        m_digits = i;
                        break;
                    }
            }
            else
            {
                m_digits = digits;

            sci_round:

                carry = norm_mantissa[m_digits + 1] >= 5 ? 1 : 0;

                for (i = m_digits + 1; i < 16; i++)
                    norm_mantissa[i] = (char)0;

                if (carry != 0)
                {
                    for (i = m_digits; i >= 0; i--)
                    {
                        char c = (char)(norm_mantissa[i] + carry);

                        if (c < 10)
                        {
                            norm_mantissa[i] = c;
                            carry = 0;
                            break;
                        }
                        else
                        {
                            norm_mantissa[i] = (char)(c - 10);
                            carry = 1;
                        }
                    }
                }

                if (carry != 0)
                {
                    // Don't round by an additional digit: that would mean
                    // we're rounding the same number twice, which is bad
                    // (think about what happens when you round one digit off
                    // 0.45, twice -- you get first 0.5, then 1... Oops).
                    // So, we start over.

                    for (i = 0; i < 15; i++)
                        norm_mantissa[i + 1] = bcd_mantissa[i];

                    norm_mantissa[0] = (char)0;
                    norm_exponent = bcd_exponent + 1;

                    goto sci_round;
                }
            }

            if (bcd_mantissa_sign != 0)
                App.char2buf(buf, buflen, ref chars_so_far, '-');

            if (dispmode == 2)
            {
                e3 = norm_exponent % 3;

                if (e3 < 0)
                    e3 += 3;

                if (m_digits < e3)
                    m_digits = e3;

                norm_exponent -= e3;
            }
            else
                e3 = 0;

            for (i = 0; i <= m_digits; i++)
            {
                App.char2buf(buf, buflen, ref chars_so_far, (char)('0' + norm_mantissa[i]));

                if (i == e3)
                    App.char2buf(buf, buflen, ref chars_so_far, (char)(App.flags.decimal_point != 0 ? '.' : ','));
            }

            App.char2buf(buf, buflen, ref chars_so_far, (char)24);


            i = App.int2string(norm_exponent, buf, chars_so_far, buflen);
            chars_so_far += i;

            return chars_so_far;
        }

        #endregion

        public static char[] phloat_text_get(Phloat p)
        {
            char[] text = new char[6];

            for (int i = 0; i < 6; i++)
                text[i] = (char)p.bcd.d_[i];

            return text;
        }

        public static void phloat_text_set(Phloat p, int i, char c)
        {
            p.bcd.d_[i] = c;
        }

        public static ushort phloat_length_get(Phloat p)
        {
            return p.bcd.d_[6];
        }

        public static void phloat_length_set(Phloat p, int len)
        {
            p.bcd.d_[6] = (ushort)len;
        }
    }
}
