﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        public static int generic_add(vartype px, vartype py, ref vartype dst)
        {
            if (px.type == TYPE.REAL && py.type == TYPE.REAL)
            {
	            vartype_real x = (vartype_real) px;
	            vartype_real y = (vartype_real) py;

	            Phloat res = y.x + x.x;

                int inf = Phloat.p_isinf(res);

	            if (inf != 0)
                {
	                if (flags.range_error_ignore != 0)
    		            res = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	                else
		                return ERR_OUT_OF_RANGE;
	            }

                dst = new_real(res);
	        
                if (dst == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            else
	                return ERR_NONE;
            }
            else if (px.type == TYPE.COMPLEX && py.type == TYPE.COMPLEX)
            {
	            vartype_complex x = (vartype_complex) px;
	            vartype_complex y = (vartype_complex) py;
	            
                int inf;
	            Phloat re, im;
	            
                re = x.re + y.re;
	            inf = Phloat.p_isinf(re);

	            if (inf != 0)
                {
	                if (flags.range_error_ignore != 0)
		                re = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	                else
		                return ERR_OUT_OF_RANGE;
	            }

                im = x.im + y.im;
	            inf = Phloat.p_isinf(im);
	        
                if (inf != 0)
                {

                    if (flags.range_error_ignore != 0)
		                im = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	                else
		                return ERR_OUT_OF_RANGE;
	            }
	        
                dst = new_complex(re, im);
	        
                if (dst == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            else
	                return ERR_NONE;
            }
            else if (px.type == TYPE.REAL && py.type == TYPE.COMPLEX)
            {
	            vartype_real x = (vartype_real) px;
	            vartype_complex y = (vartype_complex) py;
	            
                Phloat re = y.re + x.x;
	            int inf = Phloat.p_isinf(re);
	            
                if (inf != 0)
                {
	                if (flags.range_error_ignore != 0)
		                re = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	                else
		                return ERR_OUT_OF_RANGE;
	            }

                dst = new_complex(re, y.im);
	            
                if (dst == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            else
	                return ERR_NONE;
            }
            else if (px.type == TYPE.COMPLEX && py.type == TYPE.REAL)
            {
	            vartype_complex x = (vartype_complex) px;
	            vartype_real y = (vartype_real) py;
	            
                Phloat re = x.re + y.x;
	            int inf = Phloat.p_isinf(re);
	            
                if (inf != 0)
                {
	                if (flags.range_error_ignore != 0)
		                re = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	                else
		                return ERR_OUT_OF_RANGE;
	            }

                dst = new_complex(re, x.im);
	            
                if (dst == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            else
	                return ERR_NONE;
            }
            else if (px.type == TYPE.STRING || py.type == TYPE.STRING)
	            return ERR_ALPHA_DATA_IS_INVALID;
            else
	            return map_binary(px, py, ref dst, add_rr, add_rc, add_cr, add_cc);
        }

        public static int add_rr(Phloat x, Phloat y, ref Phloat z)
        {
            Phloat r;
            int inf;
            
            r = y + x;
            inf = Phloat.p_isinf(r);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    r = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            z = r;
            return ERR_NONE;
        }

        public static int add_rc(Phloat x, Phloat yre, Phloat yim, ref Phloat zre, ref Phloat zim)
        {
            Phloat rre;
            int inf;
            
            rre = yre + x;
            inf = Phloat.p_isinf(rre);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            zre = rre;
            zim = yim;
            
            return ERR_NONE;
        }

        public static int add_cr(Phloat xre, Phloat xim, Phloat y, ref Phloat zre, ref Phloat zim)
        {
            Phloat rre;
            int inf;
            
            rre = y + xre;
            inf = Phloat.p_isinf(rre);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            zre = rre;
            zim = xim;
            
            return ERR_NONE;
        }

        public static int add_cc(Phloat xre, Phloat xim, Phloat yre, Phloat yim, ref Phloat zre, ref Phloat zim)
        {
            Phloat rre, rim;
            int inf;
            
            rre = yre + xre;
            inf = Phloat.p_isinf(rre);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            rim = yim + xim;
            inf = Phloat.p_isinf(rim);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rim = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            zre = rre;
            zim = rim;
            
            return ERR_NONE;
        }

        public static int generic_div(vartype px, vartype py, fp_completion completion)
        {
            if (px.type == TYPE.STRING || py.type == TYPE.STRING)
            {
	            completion(ERR_ALPHA_DATA_IS_INVALID, null);
	            return ERR_ALPHA_DATA_IS_INVALID;
            }
            else if ((px.type == TYPE.REALMATRIX || px.type == TYPE.COMPLEXMATRIX) && (py.type == TYPE.REALMATRIX || py.type == TYPE.COMPLEXMATRIX))
            {
	            return linalg_div(py, px, completion);
            }
            else
            {
	            vartype dst = new vartype();

	            int error = map_binary(px, py, ref dst, div_rr, div_rc, div_cr, div_cc);

	            completion(error, dst);
	            return error;
            }
        }

        public static int div_rr(Phloat x, Phloat y, ref Phloat z)
        {
            Phloat r;
            int inf;
            
            if (x == 0)
                return ERR_DIVIDE_BY_0;
            
            r = y / x;
            inf = Phloat.p_isinf(r);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    r = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            z = r;
            return ERR_NONE;
        }

        public static int div_rc(Phloat x, Phloat yre, Phloat yim, ref Phloat zre, ref Phloat zim)
        {
            Phloat rre;
            Phloat rim;
            int inf;

            if (x == 0)
                return ERR_DIVIDE_BY_0;
            
            rre = yre / x;
            inf = Phloat.p_isinf(rre);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            rim = yim / x;
            inf = Phloat.p_isinf(rim);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rim = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            zre = rre;
            zim = rim;
            
            return ERR_NONE;
        }

        public static int div_cr(Phloat xre, Phloat xim, Phloat y, ref Phloat zre, ref Phloat zim)
        {
            Phloat rre;
            Phloat rim;
            Phloat h;
            int inf;
            
            // TO DO: overflows in intermediate results
            h = xre * xre + xim * xim;
            
            if (h == 0)
                return ERR_DIVIDE_BY_0;
            
            rre = y * xre / h;
            inf = Phloat.p_isinf(rre);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            rim = -y * xim / h;
            inf = Phloat.p_isinf(rim);

            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rim = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            zre = rre;
            zim = rim;
            
            return ERR_NONE;
        }

        public static int div_cc(Phloat xre, Phloat xim, Phloat yre, Phloat yim, ref Phloat zre, ref Phloat zim)
        {
            Phloat rre;
            Phloat rim;
            Phloat h;
            int inf;
            
            // TO DO: overflows in intermediate results
            h = xre * xre + xim * xim;
            
            if (h == 0)
                return ERR_DIVIDE_BY_0;
            
            rre = (xre * yre + xim * yim) / h;
            inf = Phloat.p_isinf(rre);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            rim = (xre * yim - yre * xim) / h;
            inf = Phloat.p_isinf(rim);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rim = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            zre = rre;
            zim = rim;
            
            return ERR_NONE;
        }

        public static int map_binary(vartype src1, vartype src2, ref vartype dst, mappable_rr mrr, mappable_rc mrc, mappable_cr mcr, mappable_cc mcc)
        {
            int error;

            switch (src1.type)
            {
                case TYPE.REAL:
	                switch (src2.type)
                    {
                        case TYPE.REAL:
                            {
                                Phloat r = new Phloat();

                                error = mrr(((vartype_real)src1).x, ((vartype_real)src2).x, ref r);

                                if (error == ERR_NONE)
                                {
                                    dst = new_real(r);

                                    if (dst == null)
                                        return ERR_INSUFFICIENT_MEMORY;
                                }
                                return error;
                            }

                        case TYPE.COMPLEX:
                            {
                                Phloat rre = new Phloat();
                                Phloat rim = new Phloat();

                                error = mrc(((vartype_real)src1).x, ((vartype_complex)src2).re, ((vartype_complex)src2).im, ref rre, ref rim);

                                if (error == ERR_NONE)
                                {
                                    dst = new_complex(rre, rim);

                                    if (dst == null)
                                        return ERR_INSUFFICIENT_MEMORY;
                                }
                                return error;
                            }

                        case TYPE.REALMATRIX:
                            {
		                        vartype_realmatrix sm = (vartype_realmatrix) src2;
		                        vartype_realmatrix dm;
		                        
                                dm = (vartype_realmatrix) new_realmatrix(sm.rows, sm.columns);
		                        
                                if (dm == null)
			                        return ERR_INSUFFICIENT_MEMORY;
		                        
                                int size = sm.rows * sm.columns;
		                        
                                for (int i = 0; i < size; i++)
			                        if (sm.array.is_string[i] != 0)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
			                            free_vartype(ref _vt);
			                            return ERR_ALPHA_DATA_IS_INVALID;
			                        }
		                        
                                for (int i = 0; i < size; i++)
                                {
    			                    error = mrr(((vartype_real) src1).x, sm.array.data[i], ref dm.array.data[i]);
			                    
                                    if (error != ERR_NONE)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
			                            return error;
			                        }
		                        }

                                dst = (vartype) dm;
		                        return ERR_NONE;
                            }

                        case TYPE.COMPLEXMATRIX:
                            {
                                vartype_complexmatrix sm = (vartype_complexmatrix)src2;
                                vartype_complexmatrix dm;

                                dm = (vartype_complexmatrix)new_complexmatrix(sm.rows, sm.columns);

                                if (dm == null)
                                    return ERR_INSUFFICIENT_MEMORY;

                                int size = 2 * sm.rows * sm.columns;

                                for (int i = 0; i < size; i += 2)
                                {
                                    error = mrc(((vartype_real)src1).x, sm.array.data[i], sm.array.data[i + 1], ref dm.array.data[i], ref dm.array.data[i + 1]);

                                    if (error != ERR_NONE)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return error;
                                    }
                                }

                                dst = (vartype)dm;
                                return ERR_NONE;
                            }

                        default:
		                    return ERR_INTERNAL_ERROR;
	                }


                case TYPE.COMPLEX:
	                switch (src2.type)
                    {
                        case TYPE.REAL:
                            {
                                Phloat rre = new Phloat();
                                Phloat rim = new Phloat();

                                error = mcr(((vartype_complex) src1).re, ((vartype_complex) src1).im, ((vartype_real) src2).x, ref rre, ref rim);
		                
                                if (error == ERR_NONE)
                                {
                                    dst = new_complex(rre, rim);

                                    if (dst == null)
			                            return ERR_INSUFFICIENT_MEMORY;
		                        }
		                        return error;
		                    }

		                case TYPE.COMPLEX:
                            {
		                        Phloat rre = new Phloat();
                                Phloat rim = new Phloat();
		                    
                                error = mcc(((vartype_complex) src1).re, ((vartype_complex) src1).im, ((vartype_complex) src2).re,
				                    ((vartype_complex) src2).im, ref rre, ref rim);
		                    
                                if (error == ERR_NONE)
                                {
			                        dst = new_complex(rre, rim);
			                    
                                    if (dst == null)
			                            return ERR_INSUFFICIENT_MEMORY;
		                        }
		                        return error;
		                    }

    		            case TYPE.REALMATRIX:
                            {
		                        vartype_realmatrix sm = (vartype_realmatrix) src2;
		                        vartype_complexmatrix dm;

                                dm = (vartype_complexmatrix) new_complexmatrix(sm.rows, sm.columns);
		                        
                                if (dm == null)
			                        return ERR_INSUFFICIENT_MEMORY;
		                        
                                int size = sm.rows * sm.columns;
		                        
                                for (int i = 0; i < size; i++)
			                        if (sm.array.is_string[i] != 0)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
			                            return ERR_ALPHA_DATA_IS_INVALID;
			                        }
		                        
                                for (int i = 0; i < size; i++)
                                {
			                        error = mcr(((vartype_complex) src1).re, ((vartype_complex) src1).im,
				                        sm.array.data[i], ref dm.array.data[i * 2], ref dm.array.data[i * 2 + 1]);
			                    
                                    if (error != ERR_NONE)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return error;
			                        }
		                        }

		                        dst = (vartype) dm;
		                        return ERR_NONE;
		                    }

		                case TYPE.COMPLEXMATRIX:
                            {
		                        vartype_complexmatrix sm = (vartype_complexmatrix) src2;
		                        vartype_complexmatrix dm;

                                dm = (vartype_complexmatrix) new_complexmatrix(sm.rows, sm.columns);
		                        
                                if (dm == null)
			                        return ERR_INSUFFICIENT_MEMORY;
		                        
                                int size = 2 * sm.rows * sm.columns;
		                        
                                for (int i = 0; i < size; i += 2)
                                {
    			                    error = mcc(((vartype_complex) src1).re, ((vartype_complex) src1).im,
				                        sm.array.data[i], sm.array.data[i + 1], ref dm.array.data[i], ref dm.array.data[i + 1]);
			                    
                                    if (error != ERR_NONE)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return error;
			                        }
		                        }
		                        
                                dst = (vartype) dm;
		                        return ERR_NONE;
		                    }

                        default:
		                    return ERR_INTERNAL_ERROR;
	                }


	            case TYPE.REALMATRIX:
	                switch (src2.type)
                    {
		                case TYPE.REAL:
                            {
		                        vartype_realmatrix sm = (vartype_realmatrix) src1;
		                        vartype_realmatrix dm;

                                dm = (vartype_realmatrix) new_realmatrix(sm.rows, sm.columns);
		                        
                                if (dm == null)
			                        return ERR_INSUFFICIENT_MEMORY;
		                        
                                int size = sm.rows * sm.columns;
		                        
                                for (int i = 0; i < size; i++)
			                        if (sm.array.is_string[i] != 0)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return ERR_ALPHA_DATA_IS_INVALID;
			                        }
		                        
                                for (int i = 0; i < size; i++)
                                {
			                        error = mrr(sm.array.data[i], ((vartype_real) src2).x, ref dm.array.data[i]);
			                    
                                    if (error != ERR_NONE)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return error;
			                        }
		                        }

		                        dst = (vartype) dm;
		                        return ERR_NONE;
		                    }

    		            case TYPE.COMPLEX:
                            {
		                        vartype_realmatrix sm = (vartype_realmatrix) src1;
		                        vartype_complexmatrix dm;

                                dm = (vartype_complexmatrix) new_complexmatrix(sm.rows, sm.columns);
		                        
                                if (dm == null)
			                        return ERR_INSUFFICIENT_MEMORY;
		                            
                                int size = sm.rows * sm.columns;
		                        
                                for (int i = 0; i < size; i++)
			                        if (sm.array.is_string[i] != 0)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return ERR_ALPHA_DATA_IS_INVALID;
			                        }

		                        for (int i = 0; i < size; i++)
                                {
    			                    error = mrc(sm.array.data[i], ((vartype_complex) src2).re, ((vartype_complex) src2).im,
				                        ref dm.array.data[i * 2], ref dm.array.data[i * 2 + 1]);
			                    
                                    if (error != ERR_NONE)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return error;
			                        }
		                        }

		                        dst = (vartype) dm;
		                        return ERR_NONE;
		                    }

		                case TYPE.REALMATRIX:
                            {
		                        vartype_realmatrix sm1 = (vartype_realmatrix) src1;
		                        vartype_realmatrix sm2 = (vartype_realmatrix) src2;
		                        vartype_realmatrix dm;

                                if (sm1.rows != sm2.rows || sm1.columns != sm2.columns)
			                        return ERR_DIMENSION_ERROR;
		                        
                                dm = (vartype_realmatrix) new_realmatrix(sm1.rows, sm1.columns);
		                        
                                if (dm == null)
			                        return ERR_INSUFFICIENT_MEMORY;
		                        
                                int size = sm1.rows * sm1.columns;
		                        
                                for (int i = 0; i < size; i++)
			                        if (sm1.array.is_string[i] != 0 || sm2.array.is_string[i] != 0)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return ERR_ALPHA_DATA_IS_INVALID;
			                        }
		                        
                                for (int i = 0; i < size; i++)
                                {
			                        error = mrr(sm1.array.data[i], sm2.array.data[i], ref dm.array.data[i]);
			                    
                                    if (error != ERR_NONE)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return error;
			                        }
		                        }

		                        dst = (vartype) dm;
		                        return ERR_NONE;
		                    }

		                case TYPE.COMPLEXMATRIX:
                            {
		                        vartype_realmatrix sm1 = (vartype_realmatrix) src1;
		                        vartype_complexmatrix sm2 = (vartype_complexmatrix) src2;
		                        vartype_complexmatrix dm;

                                if (sm1.rows != sm2.rows || sm1.columns != sm2.columns)
			                        return ERR_DIMENSION_ERROR;
		                            
                                dm = (vartype_complexmatrix) new_complexmatrix(sm1.rows, sm1.columns);
		                        
                                if (dm == null)
                                    return ERR_INSUFFICIENT_MEMORY;

                                int size = sm1.rows * sm1.columns;
		                        
                                for (int i = 0; i < size; i++)
			                        if (sm1.array.is_string[i] != 0)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return ERR_ALPHA_DATA_IS_INVALID;
			                        }
		                        
                                for (int i = 0; i < size; i++)
                                {
    			                    error = mrc(sm1.array.data[i], sm2.array.data[i * 2], sm2.array.data[i * 2 + 1],
				                        ref dm.array.data[i * 2], ref dm.array.data[i * 2 + 1]);
			                    
                                    if (error != ERR_NONE)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return error;
			                        }
		                        }

		                        dst = (vartype) dm;
		                        return ERR_NONE;
		                    }

		                default:
		                    return ERR_INTERNAL_ERROR;
	                }


                case TYPE.COMPLEXMATRIX:
	                switch (src2.type)
                    {
		                case TYPE.REAL:
                            {
		                        vartype_complexmatrix sm = (vartype_complexmatrix) src1;
		                        vartype_complexmatrix dm;

                                dm = (vartype_complexmatrix) new_complexmatrix(sm.rows, sm.columns);
		                        
                                if (dm == null)
			                        return ERR_INSUFFICIENT_MEMORY;
		                        
                                int size = 2 * sm.rows * sm.columns;
		                        
                                for (int i = 0; i < size; i += 2)
                                {
    			                    error = mcr(sm.array.data[i], sm.array.data[i + 1], ((vartype_real) src2).x,
				                        ref dm.array.data[i], ref dm.array.data[i + 1]);
			                    
                                    if (error != ERR_NONE)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return error;
			                        }
		                        }

		                        dst = (vartype) dm;
		                        return ERR_NONE;
		                    }

    		            case TYPE.COMPLEX:
                            {
		                        vartype_complexmatrix sm = (vartype_complexmatrix) src1;
		                        vartype_complexmatrix dm;

                                dm = (vartype_complexmatrix) new_complexmatrix(sm.rows, sm.columns);
		                        
                                if (dm == null)
			                        return ERR_INSUFFICIENT_MEMORY;
		                        
                                int size = 2 * sm.rows * sm.columns;
		                        
                                for (int i = 0; i < size; i += 2)
                                {
			                        error = mcc(sm.array.data[i], sm.array.data[i + 1], ((vartype_complex) src2).re,
                                        ((vartype_complex) src2).im, ref dm.array.data[i], ref dm.array.data[i + 1]);
			                    
                                    if (error != ERR_NONE)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return error;
			                        }
		                        }

		                        dst = (vartype) dm;
		                        return ERR_NONE;
		                    }

		                case TYPE.REALMATRIX:
                            {
		                        vartype_complexmatrix sm1 = (vartype_complexmatrix) src1;
		                        vartype_realmatrix sm2 = (vartype_realmatrix) src2;
		                        vartype_complexmatrix dm;

                                if (sm1.rows != sm2.rows || sm1.columns != sm2.columns)
			                        return ERR_DIMENSION_ERROR;
		                        
                                dm = (vartype_complexmatrix) new_complexmatrix(sm1.rows, sm1.columns);
		                        
                                if (dm == null)
			                        return ERR_INSUFFICIENT_MEMORY;
		                        
                                int size = sm1.rows * sm1.columns;
		                        
                                for (int i = 0; i < size; i++)
			                        if (sm2.array.is_string[i] != 0)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return ERR_ALPHA_DATA_IS_INVALID;
			                        }
		                        
                                for (int i = 0; i < size; i++)
                                {
    			                    error = mcr(sm1.array.data[i * 2], sm1.array.data[i * 2 + 1], sm2.array.data[i],
				                        ref dm.array.data[i * 2], ref dm.array.data[i * 2 + 1]);
			                    
                                    if (error != ERR_NONE)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return error;
			                        }
		                        }

		                        dst = (vartype) dm;
		                        return ERR_NONE;
		                    }

		                case TYPE.COMPLEXMATRIX:
                            {
		                        vartype_complexmatrix sm1 = (vartype_complexmatrix) src1;
		                        vartype_complexmatrix sm2 = (vartype_complexmatrix) src2;
		                        vartype_complexmatrix dm;

                                if (sm1.rows != sm2.rows || sm1.columns != sm2.columns)
			                        return ERR_DIMENSION_ERROR;
		                        
                                dm = (vartype_complexmatrix) new_complexmatrix(sm1.rows, sm1.columns);
		                        
                                if (dm == null)
                                    return ERR_INSUFFICIENT_MEMORY;

                                int size = 2 * sm1.rows * sm1.columns;
		                        
                                for (int i = 0; i < size; i += 2)
                                {
    			                    error = mcc(sm1.array.data[i], sm1.array.data[i + 1], sm2.array.data[i],
				                        sm2.array.data[i + 1], ref dm.array.data[i], ref dm.array.data[i + 1]);
			                    
                                    if (error != ERR_NONE)
                                    {
                                        vartype _vt = dm;
                                        dm = null;
                                        free_vartype(ref _vt);
                                        return error;
			                        }
		                        }

		                        dst = (vartype) dm;
		                        return ERR_NONE;
		                    }

		                default:
		                    return ERR_INTERNAL_ERROR;
	                }
	            
                default:
	                return ERR_INTERNAL_ERROR;
            }
        }

        public static int generic_sub(vartype px, vartype py, ref vartype dst)
        {
            if (px.type == TYPE.REAL && py.type == TYPE.REAL)
            {
	            vartype_real x = (vartype_real) px;
	            vartype_real y = (vartype_real) py;

	            Phloat res = y.x - x.x;
	            
                int inf = Phloat.p_isinf(res);
	            
                if (inf != 0)
                {
	                if (flags.range_error_ignore != 0)
                        res = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	                else
		                return ERR_OUT_OF_RANGE;
	            }
	        
                dst = new_real(res);
	        
                if (dst == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            else
    	            return ERR_NONE;
            }
            else if (px.type == TYPE.COMPLEX && py.type == TYPE.COMPLEX)
            {
	            vartype_complex x = (vartype_complex) px;
	            vartype_complex y = (vartype_complex) py;
	            
                int inf;
	            Phloat re, im;
	            
                re = y.re - x.re;
                inf = Phloat.p_isinf(re);
	            
                if (inf != 0)
                {
	                if (flags.range_error_ignore != 0)
                        re = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	                else
		                return ERR_OUT_OF_RANGE;
	            }
	        
                im = y.im - x.im;
                inf = Phloat.p_isinf(im);
	        
                if (inf != 0)
                {
	                if (flags.range_error_ignore != 0)
                        im = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	                else
		                return ERR_OUT_OF_RANGE;
	            }
	        
                dst = new_complex(re, im);
	        
                if (dst == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            else
	                return ERR_NONE;
            }
            else if (px.type == TYPE.REAL && py.type == TYPE.COMPLEX)
            {
	            vartype_real x = (vartype_real) px;
	            vartype_complex y = (vartype_complex) py;
	            
                int inf;
	            Phloat re = y.re - x.x;

                inf = Phloat.p_isinf(re);

	            if (inf != 0)
                {
	                if (flags.range_error_ignore != 0)
                        re = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	                else
		                return ERR_OUT_OF_RANGE;
	            }
	            
                dst = new_complex(re, y.im);
	            
                if (dst == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            else
	                return ERR_NONE;
            }
            else if (px.type == TYPE.COMPLEX && py.type == TYPE.REAL)
            {
	            vartype_complex x = (vartype_complex) px;
	            vartype_real y = (vartype_real) py;
	            
                Phloat re = y.x - x.re;
                int inf = Phloat.p_isinf(re);
	            
                if (inf != 0)
                {
	                if (flags.range_error_ignore != 0)
                        re = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
	                else
		                return ERR_OUT_OF_RANGE;
	            }
	            
                dst = new_complex(re, -x.im);
	            
                if (dst == null)
	                return ERR_INSUFFICIENT_MEMORY;
	            else
	                return ERR_NONE;
            }
            else if (px.type == TYPE.STRING || py.type == TYPE.STRING)
	            return ERR_ALPHA_DATA_IS_INVALID;
            else
	            return map_binary(px, py, ref dst, sub_rr, sub_rc, sub_cr, sub_cc);
        }

        public static int sub_rr(Phloat x, Phloat y, ref Phloat z)
        {
            Phloat r;
            int inf;

            r = y - x;
            inf = Phloat.p_isinf(r);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    r = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            z = r;
            return ERR_NONE;
        }

        public static int sub_rc(Phloat x, Phloat yre, Phloat yim, ref Phloat zre, ref Phloat zim)
        {
            Phloat rre;
            int inf;
            
            rre = yre - x;
            inf = Phloat.p_isinf(rre);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            zre = rre;
            zim = yim;
            return ERR_NONE;
        }

        public static int sub_cr(Phloat xre, Phloat xim, Phloat y, ref Phloat zre, ref Phloat zim)
        {
            Phloat rre;
            int inf;
            
            rre = y - xre;
            inf = Phloat.p_isinf(rre);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            zre = rre;
            zim = -xim;
            return ERR_NONE;
        }

        public static int sub_cc(Phloat xre, Phloat xim, Phloat yre, Phloat yim, ref Phloat zre, ref Phloat zim)
        {
            Phloat rre, rim;
            int inf;
            
            rre = yre - xre;
            inf = Phloat.p_isinf(rre);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            rim = yim - xim;
            inf = Phloat.p_isinf(rim);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rim = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            zre = rre;
            zim = rim;
            return ERR_NONE;
        }

        public static int generic_mul(vartype px, vartype py, fp_completion completion)
        {
            if (px.type == TYPE.STRING || py.type == TYPE.STRING)
            {
	            completion(ERR_ALPHA_DATA_IS_INVALID, null);
	            return ERR_ALPHA_DATA_IS_INVALID;
            }
            else if ((px.type == TYPE.REALMATRIX || px.type == TYPE.COMPLEXMATRIX) && (py.type == TYPE.REALMATRIX || py.type == TYPE.COMPLEXMATRIX))
            {
                return linalg_mul(py, px, completion);
            }
            else
            {
	            vartype dst = new vartype();

	            int error = map_binary(px, py, ref dst, mul_rr, mul_rc, mul_cr, mul_cc);
	            
                completion(error, dst);
	            return error;
            }
        }

        public static int mul_rr(Phloat x, Phloat y, ref Phloat z)
        {
            Phloat r;
            int inf;

            r = y * x;
            inf = Phloat.p_isinf(r);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    r = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }
            
            z = r;
            return ERR_NONE;
        }

        public static int mul_rc(Phloat x, Phloat yre, Phloat yim, ref Phloat zre, ref Phloat zim)
        {
            Phloat rre, rim;
            int inf;

            rre = yre * x;
            inf = Phloat.p_isinf(rre);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            rim = yim * x;
            inf = Phloat.p_isinf(rim);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rim = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            zre = rre;
            zim = rim;
            return ERR_NONE;
        }

        public static int mul_cr(Phloat xre, Phloat xim, Phloat y, ref Phloat zre, ref Phloat zim)
        {
            Phloat rre, rim;
            int inf;
            
            rre = y * xre;
            inf = Phloat.p_isinf(rre);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            rim = y * xim;
            inf = Phloat.p_isinf(rim);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rim = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            zre = rre;
            zim = rim;
            return ERR_NONE;
        }

        public static int mul_cc(Phloat xre, Phloat xim, Phloat yre, Phloat yim, ref Phloat zre, ref Phloat zim)
        {
            Phloat rre, rim;
            int inf;
            
            rre = xre * yre - xim * yim;
            inf = Phloat.p_isinf(rre);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            rim = xre * yim + yre * xim;
            inf = Phloat.p_isinf(rim);
            
            if (inf != 0)
            {
                if (flags.range_error_ignore != 0)
                    rim = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                else
                    return ERR_OUT_OF_RANGE;
            }

            zre = rre;
            zim = rim;
            return ERR_NONE;
        }

        public static int map_unary(vartype src, ref vartype dst, mappable_r mr, mappable_c mc)
        {
            int error;
            
            switch (src.type)
            {
	            case TYPE.REAL:
                    {
	                    Phloat r = new Phloat();

	                    error = mr(((vartype_real) src).x, ref r);
	                    
                        if (error == ERR_NONE)
                        {
		                    dst = new_real(r);
		                    
                            if (dst == null)
		                        return ERR_INSUFFICIENT_MEMORY;
	                    }

	                    return error;
	                }

	            case TYPE.COMPLEX:
                    {
	                    Phloat rre = new Phloat();
                        Phloat rim = new Phloat();

	                    error = mc(((vartype_complex) src).re, ((vartype_complex) src).im, ref rre, ref rim);
	                    
                        if (error == ERR_NONE)
                        {
		                    dst = new_complex(rre, rim);
		                
                            if (dst == null)
		                        return ERR_INSUFFICIENT_MEMORY;
	                    }
	                    
                        return error;
	                }

	            case TYPE.REALMATRIX:
                    {
	                    vartype_realmatrix sm = (vartype_realmatrix) src;
	                    vartype_realmatrix dm;

	                    int size, i;
	                    
                        dm = (vartype_realmatrix) new_realmatrix(sm.rows, sm.columns);
	                    
                        if (dm == null)
		                    return ERR_INSUFFICIENT_MEMORY;
	                    
                        size = sm.rows * sm.columns;
	                    
                        for (i = 0; i < size; i++)
                        {
		                    if (sm.array.is_string[i] != 0)
                            {
                                vartype _vt = (vartype)dm;
                                dm = null;
		                        free_vartype(ref _vt);
		                        return ERR_ALPHA_DATA_IS_INVALID;
		                    }
	                    }

	                    for (i = 0; i < size; i++)
                        {
		                    error = mr(sm.array.data[i], ref dm.array.data[i]);
		                    
                            if (error != ERR_NONE)
                            {
                                vartype _vt = (vartype)dm;
                                dm = null;
                                free_vartype(ref _vt);
                                return error;
		                    }
	                    }

	                    dst = (vartype) dm;
	                    return ERR_NONE;
	                }

	            case TYPE.COMPLEXMATRIX:
                    {
	                    vartype_complexmatrix sm = (vartype_complexmatrix) src;
	                    vartype_complexmatrix dm;

	                    int rows = sm.rows;
	                    int columns = sm.columns;
	                    int size = 2 * rows * columns;
	                    int i;

                        dm = (vartype_complexmatrix) new_complexmatrix(rows, columns);
	                    
                        if (dm == null)
		                    return ERR_INSUFFICIENT_MEMORY;
	                    
                        for (i = 0; i < size; i += 2)
                        {
		                    error = mc(sm.array.data[i], sm.array.data[i + 1], ref dm.array.data[i], ref dm.array.data[i + 1]);
		                
                            if (error != ERR_NONE)
                            {
                                vartype _vt = (vartype)dm;
                                dm = null;
                                free_vartype(ref _vt);
                                return error;
		                    }
	                    }

                        dst = (vartype) dm;
	                    return ERR_NONE;
	                }

	            default:
	                return ERR_INTERNAL_ERROR;
            }
        }

        public static int generic_sto(arg_struct arg, char operation)
        {
            if (arg.type == ARGTYPE_IND_NUM
                || arg.type == ARGTYPE_IND_STK
                || arg.type == ARGTYPE_IND_STR)
            {
                int err = resolve_ind_arg(arg);

                if (err != ERR_NONE)
                    return err;
            }

            if (operation != 0 && reg_x.type == TYPE.STRING)
                return ERR_ALPHA_DATA_IS_INVALID;

            switch (arg.type)
            {
                case ARGTYPE_NUM:
                    {
                        vartype regs = recall_var("REGS".ToCharArray(), 4);
                        
                        if (regs == null)
                            return ERR_SIZE_ERROR;
                        
                        if (regs.type == TYPE.REALMATRIX)
                        {
                            vartype_realmatrix rm = (vartype_realmatrix)regs;

                            int size = rm.rows * rm.columns;
                            int num = arg.val.num;
                            
                            if (num >= size)
                                return ERR_SIZE_ERROR;
                            
                            if (reg_x.type == TYPE.STRING)
                            {
                                vartype_string vs = (vartype_string)reg_x;
                                Phloat ds = rm.array.data[num];
                                int len, i;
                                
                                if (disentangle((vartype)rm) == 0)
                                    return ERR_INSUFFICIENT_MEMORY;
                                
                                len = vs.length;
                                Phloat.phloat_length_set(ds, len);
                                
                                for (i = 0; i < len; i++)
                                    Phloat.phloat_text_set(ds, i, vs.text[i]);
                                
                                rm.array.is_string[num] = (char)1;
                                
                                return ERR_NONE;
                            }
                            else if (reg_x.type == TYPE.REAL)
                            {
                                if (disentangle((vartype)rm) == 0)
                                    return ERR_INSUFFICIENT_MEMORY;
                                
                                if (operation == 0)
                                {
                                    rm.array.data[num] = ((vartype_real)reg_x).x;
                                    rm.array.is_string[num] = (char)0;
                                }
                                else
                                {
                                    Phloat x, n;
                                    int inf;
                                
                                    if (rm.array.is_string[num] != 0)
                                        return ERR_ALPHA_DATA_IS_INVALID;
                                    
                                    x = ((vartype_real)reg_x).x;
                                    n = rm.array.data[num];
                                    
                                    switch (operation)
                                    {
                                        case '/': if (x == 0) return ERR_DIVIDE_BY_0;
                                            n /= x; break;
                                        case '*': n *= x; break;
                                        case '-': n -= x; break;
                                        case '+': n += x; break;
                                    }
                                    
                                    inf = Phloat.p_isinf(n);
                                    
                                    if (inf != 0)
                                    {
                                        if (flags.range_error_ignore != 0)
                                            n = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                                        else
                                            return ERR_OUT_OF_RANGE;
                                    }

                                    rm.array.data[num] = n;
                                }

                                return ERR_NONE;
                            }
                            else
                                return ERR_INVALID_TYPE;
                        }
                        else if (regs.type == TYPE.COMPLEXMATRIX)
                        {
                            vartype_complexmatrix cm = (vartype_complexmatrix)regs;
                            
                            int size = cm.rows * cm.columns;
                            int num = arg.val.num;
                            
                            Phloat re, im;
                            
                            if (num >= size)
                                return ERR_SIZE_ERROR;
                            
                            if (reg_x.type == TYPE.STRING)
                                return ERR_ALPHA_DATA_IS_INVALID;
                            else if (reg_x.type != TYPE.REAL && reg_x.type != TYPE.COMPLEX)
                                return ERR_INVALID_TYPE;

                            if (disentangle((vartype)cm) == 0)
                                return ERR_INSUFFICIENT_MEMORY;
                            
                            if (operation == 0)
                            {
                                if (reg_x.type == TYPE.REAL)
                                {
                                    re = ((vartype_real)reg_x).x;
                                    im = 0;
                                }
                                else
                                {
                                    re = ((vartype_complex)reg_x).re;
                                    im = ((vartype_complex)reg_x).im;
                                }

                                cm.array.data[num * 2] = re;
                                cm.array.data[num * 2 + 1] = im;
                            }
                            else
                            {
                                Phloat nre = cm.array.data[num * 2];
                                Phloat nim = cm.array.data[num * 2 + 1];
                                int inf;
                                
                                if (reg_x.type == TYPE.REAL)
                                {
                                    Phloat x;
                                    x = ((vartype_real)reg_x).x;
                                    
                                    switch (operation)
                                    {
                                        case '/': if (x == 0) return ERR_DIVIDE_BY_0;
                                            nre /= x; nim /= x; break;
                                        case '*': nre *= x; nim *= x; break;
                                        case '-': nre -= x; break;
                                        case '+': nre += x; break;
                                    }
                                }
                                else
                                {
                                    Phloat xre, xim, h, tmp;
                                    
                                    xre = ((vartype_complex)reg_x).re;
                                    xim = ((vartype_complex)reg_x).im;
                                    h = xre * xre + xim * xim;
                                    
                                    // (orig authors note) TO DO: handle overflows in intermediate results
                                    switch (operation)
                                    {
                                        case '/':
                                            if (h == 0)
                                                return ERR_DIVIDE_BY_0;
                                            tmp = (nre * xre + nim * xim) / h;
                                            nim = (xre * nim - xim * nre) / h;
                                            nre = tmp;
                                            break;
                                        case '*':
                                            tmp = nre * xre - nim * xim;
                                            nim = nre * xim + nim * xre;
                                            nre = tmp;
                                            break;
                                        case '-':
                                            nre -= xre;
                                            nim -= xim;
                                            break;
                                        case '+':
                                            nre += xre;
                                            nim += xim;
                                            break;
                                    }
                                }

                                inf = Phloat.p_isinf(nre);
                                
                                if (inf != 0)
                                {
                                    if (flags.range_error_ignore != 0)
                                        nre = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                                    else
                                        return ERR_OUT_OF_RANGE;
                                }

                                inf = Phloat.p_isinf(nim);

                                if (inf != 0)
                                {
                                    if (flags.range_error_ignore != 0)
                                        nim = inf == 1 ? Phloat.POS_HUGE_PHLOAT : Phloat.NEG_HUGE_PHLOAT;
                                    else
                                        return ERR_OUT_OF_RANGE;
                                }

                                cm.array.data[num * 2] = nre;
                                cm.array.data[num * 2 + 1] = nim;
                            }

                            return ERR_NONE;
                        }
                        else
                        {
                            // This should never happen; STO should prevent
                            // "REGS" from being any other type than a real or
                            // complex matrix.
                            //
                            return ERR_INTERNAL_ERROR;
                        }
                    }

                case ARGTYPE_STK:
                    {
                        vartype newval;

                        if (operation == 0)
                        {
                            if (arg.val.stk == 'X')
                            {
                                // STO ST X : no-op!
                                return ERR_NONE;
                            }
                        
                            newval = dup_vartype(reg_x);
                            
                            if (newval == null)
                                return ERR_INSUFFICIENT_MEMORY;
                            
                            switch (arg.val.stk)
                            {
                                case 'Y':
                                    free_vartype(ref reg_y);
                                    reg_y = newval;
                                    break;
                                case 'Z':
                                    free_vartype(ref reg_z);
                                    reg_z = newval;
                                    break;
                                case 'T':
                                    free_vartype(ref reg_t);
                                    reg_t = newval;
                                    break;
                                case 'L':
                                    free_vartype(ref reg_lastx);
                                    reg_lastx = newval;
                                    break;
                            }
                            return ERR_NONE;
                        }
                        else
                        {
                            vartype oldval = new vartype();

                            switch (arg.val.stk)
                            {
                                case 'X': oldval = reg_x; break;
                                case 'Y': oldval = reg_y; break;
                                case 'Z': oldval = reg_z; break;
                                case 'T': oldval = reg_t; break;
                                case 'L': oldval = reg_lastx; break;
                            }
                            
                            temp_arg = arg;
                            
                            return apply_sto_operation(operation, oldval);
                        }
                    }

                case ARGTYPE_STR:
                    {
                        if (operation == 0)
                        {
                            vartype newval;

                            // Only allow matrices to be stored in "REGS"
                            if (arg.length == 4
                                && arg.val.text[0] == 'R'
                                && arg.val.text[1] == 'E'
                                && arg.val.text[2] == 'G'
                                && arg.val.text[3] == 'S'
                                && reg_x.type != TYPE.REALMATRIX
                                && reg_x.type != TYPE.COMPLEXMATRIX)
                                return ERR_RESTRICTED_OPERATION;
                            
                            // When EDITN is active, don't allow the matrix being
                            // edited to be overwritten by a string or scalar.
                            if (matedit_mode == 3 && arg.length == matedit_length
                                && (reg_x.type == TYPE.REAL
                                    || reg_x.type == TYPE.COMPLEX
                                    || reg_x.type == TYPE.STRING))
                            {
                                bool equal = true;
                            
                                for (int i = 0; i < arg.length; i++)
                                    if (arg.val.text[i] != matedit_name[i])
                                    {
                                        equal = false;
                                        break;
                                    }
                                
                                if (equal)
                                    return ERR_RESTRICTED_OPERATION;
                            }

                            newval = dup_vartype(reg_x);
                            
                            if (newval == null)
                                return ERR_INSUFFICIENT_MEMORY;
                            
                            store_var(arg.val.text, arg.length, newval);
                            
                            return ERR_NONE;
                        }
                        else
                        {
                            vartype oldval = recall_var(arg.val.text, arg.length);

                            if (oldval == null)
                                return ERR_NONEXISTENT;
                            
                            temp_arg = arg;
                            
                            return apply_sto_operation(operation, oldval);
                        }
                    }

                default:
                    return ERR_INVALID_TYPE;
            }
        }

        public static int apply_sto_operation(char operation, vartype oldval)
        {
            vartype newval = new vartype();
            int error;
            
            switch (operation)
            {
                case '/':
                    preserve_ij = true;
                    return generic_div(reg_x, oldval, generic_sto_completion);
            
                case '*':
                    preserve_ij = false;
                    return generic_mul(reg_x, oldval, generic_sto_completion);
                
                case '-':
                    preserve_ij = true;
                    error = generic_sub(reg_x, oldval, ref newval);
                    generic_sto_completion(error, newval);
                    return error;
                
                case '+':
                    preserve_ij = true;
                    error = generic_add(reg_x, oldval, ref newval);
                    generic_sto_completion(error, newval);
                    return error;
                
                default:
                    return ERR_INTERNAL_ERROR;
            }
        }

        public static void generic_sto_completion(int error, vartype res)
        {
            if (error != ERR_NONE)
                return;
            
            if (temp_arg.type == ARGTYPE_STK)
            {
                switch (temp_arg.val.stk)
                {
                    case 'X':
                        free_vartype(ref reg_x);
                        reg_x = res;
                        break;
            
                    case 'Y':
                        free_vartype(ref reg_y);
                        reg_y = res;
                        break;
                    
                    case 'Z':
                        free_vartype(ref reg_z);
                        reg_z = res;
                        break;
                    
                    case 'T':
                        free_vartype(ref reg_t);
                        reg_t = res;
                        break;
                    
                    case 'L':
                        free_vartype(ref reg_lastx);
                        reg_lastx = res;
                        break;
                }
            }
            else // temp_arg.type == ARGTYPE_STR
            {
                // If the destination of store_var() is the indexed matrix, it sets I
                // and J to 1. This is *not* the desired behavior for STO+, STO-, and
                // STO/. (It is correct for STO and STO*, since those can cause the
                // destination's dimensions to change.)
                int i = matedit_i;
                int j = matedit_j;
                
                store_var(temp_arg.val.text, temp_arg.length, res);
                
                if (preserve_ij)
                {
                    matedit_i = i;
                    matedit_j = j;
                }
            }
        }

        public static int generic_rcl(arg_struct arg, ref vartype dst)
        {
            int err;
            
            if (arg.type == ARGTYPE_IND_NUM
                || arg.type == ARGTYPE_IND_STK
                || arg.type == ARGTYPE_IND_STR)
            {
                err = resolve_ind_arg(arg);
            
                if (err != ERR_NONE)
                    return err;
            }

            switch (arg.type)
            {
                case ARGTYPE_NUM:
                    {
                        vartype regs = recall_var("REGS".ToCharArray(), 4);
                        
                        if (regs == null)
                            return ERR_SIZE_ERROR;
                        else if (regs.type == TYPE.REALMATRIX)
                        {
                            vartype_realmatrix rm = (vartype_realmatrix)regs;
                            
                            int size = rm.rows * rm.columns;
                            int index = arg.val.num;
                            
                            Phloat ds;
                            
                            if (index >= size)
                                return ERR_SIZE_ERROR;
                            
                            ds = rm.array.data[index];
                            
                            if (rm.array.is_string[index] != 0)
                                dst = new_string(Phloat.phloat_text_get(ds), Phloat.phloat_length_get(ds));
                            else
                                dst = new_real(ds);
                            
                            if (dst == null)
                                return ERR_INSUFFICIENT_MEMORY;
                            
                            return ERR_NONE;
                        }
                        else if (regs.type == TYPE.COMPLEXMATRIX)
                        {
                            vartype_complexmatrix cm = (vartype_complexmatrix)regs;
                            
                            int size = cm.rows * cm.columns;
                            
                            if (arg.val.num >= size)
                                return ERR_SIZE_ERROR;
                            
                            dst = new_complex(cm.array.data[arg.val.num * 2], cm.array.data[arg.val.num * 2 + 1]);
                            
                            if (dst == null)
                                return ERR_INSUFFICIENT_MEMORY;
                            
                            return ERR_NONE;
                        }
                        else
                        {
                            // This should never happen; STO should prevent
                            // "REGS" from being any other type than a real or
                            // complex matrix.
                            //
                            return ERR_INTERNAL_ERROR;
                        }
                    }

                case ARGTYPE_STK:
                    {
                        switch (arg.val.stk)
                        {
                            case 'X': dst = reg_x; break;
                            case 'Y': dst = reg_y; break;
                            case 'Z': dst = reg_z; break;
                            case 'T': dst = reg_t; break;
                            case 'L': dst = reg_lastx; break;
                        }

                        dst = dup_vartype(dst);

                        if (dst == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        
                        return ERR_NONE;
                    }

                case ARGTYPE_STR:
                    {
                        dst = recall_var(arg.val.text, arg.length);
                        
                        if (dst == null)
                            return ERR_NONEXISTENT;
                        
                        dst = dup_vartype(dst);
                        
                        if (dst == null)
                            return ERR_INSUFFICIENT_MEMORY;
                        
                        return ERR_NONE;
                    }

                default:
                    return ERR_INTERNAL_ERROR;
            }
        }
    }
}
