﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        public static command_spec[] cmd_array = new command_spec[]
        {
            new command_spec { name="CLX",                           name_length=3,  handler=docmd_clx,         hp42s_code=0x00000077, argtype=ARG_NONE,  flags=FLAG_NONE }, // CLX
            new command_spec { name="ENT\x00c5R",                    name_length=5,  handler=docmd_enter,       hp42s_code=0x00000083, argtype=ARG_NONE,  flags=FLAG_NONE }, // ENTER
            new command_spec { name="X<>Y",                          name_length=4,  handler=docmd_swap,        hp42s_code=0x00000071, argtype=ARG_NONE,  flags=FLAG_NONE }, // SWAP
            new command_spec { name="R\x000e",                       name_length=2,  handler=docmd_rdn,         hp42s_code=0x00000075, argtype=ARG_NONE,  flags=FLAG_NONE }, // RDN
            new command_spec { name="+/-",                           name_length=3,  handler=docmd_chs,         hp42s_code=0x00000054, argtype=ARG_NONE,  flags=FLAG_NONE }, // CHS
            new command_spec { name="\x0000",                        name_length=1,  handler=docmd_div,         hp42s_code=0x00000043, argtype=ARG_NONE,  flags=FLAG_NONE }, // DIV
            new command_spec { name="\x0001",                        name_length=1,  handler=docmd_mul,         hp42s_code=0x00000042, argtype=ARG_NONE,  flags=FLAG_NONE }, // MUL
            new command_spec { name="-",                             name_length=1,  handler=docmd_sub,         hp42s_code=0x00000041, argtype=ARG_NONE,  flags=FLAG_NONE }, // SUB
            new command_spec { name="+",                             name_length=1,  handler=docmd_add,         hp42s_code=0x00000040, argtype=ARG_NONE,  flags=FLAG_NONE }, // ADD
            new command_spec { name="LASTX",                         name_length=5,  handler=docmd_lastx,       hp42s_code=0x00000076, argtype=ARG_NONE,  flags=FLAG_NONE }, // LASTX
            new command_spec { name="",                              name_length=0,  handler=null,              hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_HIDDEN + FLAG_NO_SHOW }, // SILENT_OFF
            new command_spec { name="",                              name_length=0,  handler=null,              hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_HIDDEN + FLAG_NO_SHOW }, // SILENT_ON
            new command_spec { name="SIN",                           name_length=3,  handler=docmd_sin,         hp42s_code=0x00000059, argtype=ARG_NONE,  flags=FLAG_NONE }, // SIN
            new command_spec { name="COS",                           name_length=3,  handler=docmd_cos,         hp42s_code=0x0000005a, argtype=ARG_NONE,  flags=FLAG_NONE }, // COS
            new command_spec { name="TAN",                           name_length=3,  handler=docmd_tan,         hp42s_code=0x0000005b, argtype=ARG_NONE,  flags=FLAG_NONE }, // TAN
            new command_spec { name="ASIN",                          name_length=4,  handler=docmd_asin,        hp42s_code=0x0000005c, argtype=ARG_NONE,  flags=FLAG_NONE }, // ASIN
            new command_spec { name="ACOS",                          name_length=4,  handler=docmd_acos,        hp42s_code=0x0000005d, argtype=ARG_NONE,  flags=FLAG_NONE }, // ACOS
            new command_spec { name="ATAN",                          name_length=4,  handler=docmd_atan,        hp42s_code=0x0000005e, argtype=ARG_NONE,  flags=FLAG_NONE }, // ATAN
            new command_spec { name="LOG",                           name_length=3,  handler=docmd_log,         hp42s_code=0x00000056, argtype=ARG_NONE,  flags=FLAG_NONE }, // LOG
            new command_spec { name="10^X",                          name_length=4,  handler=docmd_10_pow_x,    hp42s_code=0x00000057, argtype=ARG_NONE,  flags=FLAG_NONE }, // 10_POW_X
            new command_spec { name="LN",                            name_length=2,  handler=docmd_ln,          hp42s_code=0x00000050, argtype=ARG_NONE,  flags=FLAG_NONE }, // LN
            new command_spec { name="E^X",                           name_length=3,  handler=docmd_e_pow_x,     hp42s_code=0x00000055, argtype=ARG_NONE,  flags=FLAG_NONE }, // E_POW_X
            new command_spec { name="SQRT",                          name_length=4,  handler=docmd_sqrt,        hp42s_code=0x00000052, argtype=ARG_NONE,  flags=FLAG_NONE }, // SQRT
            new command_spec { name="X^2",                           name_length=3,  handler=docmd_square,      hp42s_code=0x00000051, argtype=ARG_NONE,  flags=FLAG_NONE }, // SQUARE
            new command_spec { name="1/X",                           name_length=3,  handler=docmd_inv,         hp42s_code=0x00000060, argtype=ARG_NONE,  flags=FLAG_NONE }, // INV
            new command_spec { name="Y^X",                           name_length=3,  handler=docmd_y_pow_x,     hp42s_code=0x00000053, argtype=ARG_NONE,  flags=FLAG_NONE }, // Y_POW_X
            new command_spec { name="%",                             name_length=1,  handler=docmd_percent,     hp42s_code=0x0000004c, argtype=ARG_NONE,  flags=FLAG_NONE }, // PERCENT
            new command_spec { name="PI",                            name_length=2,  handler=docmd_pi,          hp42s_code=0x00000072, argtype=ARG_NONE,  flags=FLAG_NONE }, // PI
            new command_spec { name="C\x00cf\x00cdPL\x00c5X",        name_length=7,  handler=docmd_complex,     hp42s_code=0x0000a072, argtype=ARG_NONE,  flags=FLAG_NONE }, // COMPLEX
            new command_spec { name="STO",                           name_length=3,  handler=docmd_sto,         hp42s_code=0x01810091, argtype=ARG_VAR,   flags=FLAG_NONE }, // STO
            new command_spec { name="STO\x0000",                     name_length=4,  handler=docmd_sto_div,     hp42s_code=0x00850095, argtype=ARG_VAR,   flags=FLAG_NONE }, // STO_DIV
            new command_spec { name="STO\x0001",                     name_length=4,  handler=docmd_sto_mul,     hp42s_code=0x00840094, argtype=ARG_VAR,   flags=FLAG_NONE }, // STO_MUL
            new command_spec { name="STO-",                          name_length=4,  handler=docmd_sto_sub,     hp42s_code=0x00830093, argtype=ARG_VAR,   flags=FLAG_NONE }, // STO_SUB
            new command_spec { name="STO+",                          name_length=4,  handler=docmd_sto_add,     hp42s_code=0x00820092, argtype=ARG_VAR,   flags=FLAG_NONE }, // STO_ADD
            new command_spec { name="RCL",                           name_length=3,  handler=docmd_rcl,         hp42s_code=0x01910090, argtype=ARG_VAR,   flags=FLAG_NONE }, // RCL
            new command_spec { name="RCL\x0000",                     name_length=4,  handler=docmd_rcl_div,     hp42s_code=0x0095f2d4, argtype=ARG_VAR,   flags=FLAG_NONE }, // RCL_DIV
            new command_spec { name="RCL\x0001",                     name_length=4,  handler=docmd_rcl_mul,     hp42s_code=0x0094f2d3, argtype=ARG_VAR,   flags=FLAG_NONE }, // RCL_MUL
            new command_spec { name="RCL-",                          name_length=4,  handler=docmd_rcl_sub,     hp42s_code=0x0093f2d2, argtype=ARG_VAR,   flags=FLAG_NONE }, // RCL_SUB
            new command_spec { name="RCL+",                          name_length=4,  handler=docmd_rcl_add,     hp42s_code=0x0092f2d1, argtype=ARG_VAR,   flags=FLAG_NONE }, // RCL_ADD
            new command_spec { name="FIX",                           name_length=3,  handler=docmd_fix,         hp42s_code=0x01d4009c, argtype=ARG_NUM11, flags=FLAG_NONE }, // FIX
            new command_spec { name="SCI",                           name_length=3,  handler=docmd_sci,         hp42s_code=0x01d5009d, argtype=ARG_NUM11, flags=FLAG_NONE }, // SCI
            new command_spec { name="ENG",                           name_length=3,  handler=docmd_eng,         hp42s_code=0x01d6009e, argtype=ARG_NUM11, flags=FLAG_NONE }, // ENG
            new command_spec { name="ALL",                           name_length=3,  handler=docmd_all,         hp42s_code=0x0000a25d, argtype=ARG_NONE,  flags=FLAG_NONE }, // ALL
            new command_spec { name="\x00ce\x00d5\x00cc\x00cc",      name_length=4,  handler=docmd_null,        hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_HIDDEN }, // NULL
            new command_spec { name="ASTO",                          name_length=4,  handler=docmd_asto,        hp42s_code=0x00b2009a, argtype=ARG_VAR,   flags=FLAG_NONE }, // ASTO
            new command_spec { name="ARCL",                          name_length=4,  handler=docmd_arcl,        hp42s_code=0x00b3009b, argtype=ARG_VAR,   flags=FLAG_NONE }, // ARCL
            new command_spec { name="CLA",                           name_length=3,  handler=docmd_cla,         hp42s_code=0x00000087, argtype=ARG_NONE,  flags=FLAG_NONE }, // CLA
            new command_spec { name="DEG",                           name_length=3,  handler=docmd_deg,         hp42s_code=0x00000080, argtype=ARG_NONE,  flags=FLAG_NONE }, // DEG
            new command_spec { name="RAD",                           name_length=3,  handler=docmd_rad,         hp42s_code=0x00000081, argtype=ARG_NONE,  flags=FLAG_NONE }, // RAD
            new command_spec { name="GRAD",                          name_length=4,  handler=docmd_grad,        hp42s_code=0x00000082, argtype=ARG_NONE,  flags=FLAG_NONE }, // GRAD
            new command_spec { name="RECT",                          name_length=4,  handler=docmd_rect,        hp42s_code=0x0000a25a, argtype=ARG_NONE,  flags=FLAG_NONE }, // RECT
            new command_spec { name="POLAR",                         name_length=5,  handler=docmd_polar,       hp42s_code=0x0000a259, argtype=ARG_NONE,  flags=FLAG_NONE }, // POLAR
            new command_spec { name="SIZE",                          name_length=4,  handler=docmd_size,        hp42s_code=0x01000000, argtype=ARG_COUNT, flags=FLAG_NONE }, // SIZE
            new command_spec { name="QUIET",                         name_length=5,  handler=docmd_quiet,       hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_IMMED }, // QUIET
            new command_spec { name="C\x00d0\x00d8RES",              name_length=6,  handler=docmd_cpxres,      hp42s_code=0x0000a26a, argtype=ARG_NONE,  flags=FLAG_NONE }, // CPXRES
            new command_spec { name="R\x00c5\x00c1\x00ccRES",        name_length=7,  handler=docmd_realres,     hp42s_code=0x0000a26b, argtype=ARG_NONE,  flags=FLAG_NONE }, // REALRES
            new command_spec { name="KEY\x00cc\x00d3\x00ce",         name_length=6,  handler=docmd_keyasn,      hp42s_code=0x0000a263, argtype=ARG_NONE,  flags=FLAG_NONE }, // KEYASN
            new command_spec { name="LCLBL",                         name_length=5,  handler=docmd_lclbl,       hp42s_code=0x0000a264, argtype=ARG_NONE,  flags=FLAG_NONE }, // LCLBL
            new command_spec { name="RDX.",                          name_length=4,  handler=docmd_rdxdot,      hp42s_code=0x0000a25b, argtype=ARG_NONE,  flags=FLAG_NONE }, // RDXDOT
            new command_spec { name="RDX,",                          name_length=4,  handler=docmd_rdxcomma,    hp42s_code=0x0000a25c, argtype=ARG_NONE,  flags=FLAG_NONE }, // RDXCOMMA
            new command_spec { name="CL\x0005",                      name_length=3,  handler=docmd_clsigma,     hp42s_code=0x00000070, argtype=ARG_NONE,  flags=FLAG_NONE }, // CLSIGMA
            new command_spec { name="CLP",                           name_length=3,  handler=docmd_clp,         hp42s_code=0x00f00000, argtype=ARG_PRGM,  flags=FLAG_NONE }, // CLP
            new command_spec { name="CLV",                           name_length=3,  handler=docmd_clv,         hp42s_code=0x00b0f2d8, argtype=ARG_NAMED, flags=FLAG_NONE }, // CLV
            new command_spec { name="CLST",                          name_length=4,  handler=docmd_clst,        hp42s_code=0x00000073, argtype=ARG_NONE,  flags=FLAG_NONE }, // CLST
            new command_spec { name="CLRG",                          name_length=4,  handler=docmd_clrg,        hp42s_code=0x0000008a, argtype=ARG_NONE,  flags=FLAG_NONE }, // CLRG
            new command_spec { name="DEL",                           name_length=3,  handler=docmd_del,         hp42s_code=0x02000000, argtype=ARG_COUNT, flags=FLAG_PRGM_ONLY + FLAG_IMMED }, // DEL
            new command_spec { name="CLK\x00c5Y\x00d3",              name_length=6,  handler=docmd_clkeys,      hp42s_code=0x0000a262, argtype=ARG_NONE,  flags=FLAG_NONE }, // CLKEYS
            new command_spec { name="CLLCD",                         name_length=5,  handler=docmd_cllcd,       hp42s_code=0x0000a763, argtype=ARG_NONE,  flags=FLAG_NONE }, // CLLCD
            new command_spec { name="CLM\x00c5N\x00d5",              name_length=6,  handler=docmd_clmenu,      hp42s_code=0x0000a26d, argtype=ARG_NONE,  flags=FLAG_NONE }, // CLMENU
            new command_spec { name="CLALL",                         name_length=5,  handler=null,              hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_IMMED }, // CLALLa
            new command_spec { name="\x000fDEG",                     name_length=4,  handler=docmd_to_deg,      hp42s_code=0x0000006b, argtype=ARG_NONE,  flags=FLAG_NONE }, // TO_DEG
            new command_spec { name="\x000fRAD",                     name_length=4,  handler=docmd_to_rad,      hp42s_code=0x0000006a, argtype=ARG_NONE,  flags=FLAG_NONE }, // TO_RAD
            new command_spec { name="\x000fHR",                      name_length=3,  handler=docmd_to_hr,       hp42s_code=0x0000006d, argtype=ARG_NONE,  flags=FLAG_NONE }, // TO_HR
            new command_spec { name="\x000fHMS",                     name_length=4,  handler=docmd_to_hms,      hp42s_code=0x0000006c, argtype=ARG_NONE,  flags=FLAG_NONE }, // TO_HMS
            new command_spec { name="\x000fREC",                     name_length=4,  handler=docmd_to_rec,      hp42s_code=0x0000004e, argtype=ARG_NONE,  flags=FLAG_NONE }, // TO_REC
            new command_spec { name="\x000fPOL",                     name_length=4,  handler=docmd_to_pol,      hp42s_code=0x0000004f, argtype=ARG_NONE,  flags=FLAG_NONE }, // TO_POL
            new command_spec { name="IP",                            name_length=2,  handler=docmd_ip,          hp42s_code=0x00000068, argtype=ARG_NONE,  flags=FLAG_NONE }, // IP
            new command_spec { name="FP",                            name_length=2,  handler=docmd_fp,          hp42s_code=0x00000069, argtype=ARG_NONE,  flags=FLAG_NONE }, // FP
            new command_spec { name="RND",                           name_length=3,  handler=docmd_rnd,         hp42s_code=0x0000006e, argtype=ARG_NONE,  flags=FLAG_NONE }, // RND
            new command_spec { name="ABS",                           name_length=3,  handler=docmd_abs,         hp42s_code=0x00000061, argtype=ARG_NONE,  flags=FLAG_NONE }, // ABS
            new command_spec { name="SIGN",                          name_length=4,  handler=docmd_sign,        hp42s_code=0x0000007a, argtype=ARG_NONE,  flags=FLAG_NONE }, // SIGN
            new command_spec { name="MOD",                           name_length=3,  handler=docmd_mod,         hp42s_code=0x0000004b, argtype=ARG_NONE,  flags=FLAG_NONE }, // MOD
            new command_spec { name="SF",                            name_length=2,  handler=docmd_sf,          hp42s_code=0x00a000a8, argtype=ARG_NUM99, flags=FLAG_NONE }, // SF
            new command_spec { name="CF",                            name_length=2,  handler=docmd_cf,          hp42s_code=0x00a100a9, argtype=ARG_NUM99, flags=FLAG_NONE }, // CF
            new command_spec { name="FS?",                           name_length=3,  handler=docmd_fs_t,        hp42s_code=0x00a400ac, argtype=ARG_NUM99, flags=FLAG_NONE }, // FS_T
            new command_spec { name="FC?",                           name_length=3,  handler=docmd_fc_t,        hp42s_code=0x00a500ad, argtype=ARG_NUM99, flags=FLAG_NONE }, // FC_T
            new command_spec { name="FS?C",                          name_length=4,  handler=docmd_fsc_t,       hp42s_code=0x00a200aa, argtype=ARG_NUM99, flags=FLAG_NONE }, // FSC_T
            new command_spec { name="FC?C",                          name_length=4,  handler=docmd_fcc_t,       hp42s_code=0x00a300ab, argtype=ARG_NUM99, flags=FLAG_NONE }, // FCC_T
            new command_spec { name="COMB",                          name_length=4,  handler=docmd_comb,        hp42s_code=0x0000a06f, argtype=ARG_NONE,  flags=FLAG_NONE }, // COMB
            new command_spec { name="PERM",                          name_length=4,  handler=docmd_perm,        hp42s_code=0x0000a070, argtype=ARG_NONE,  flags=FLAG_NONE }, // PERM
            new command_spec { name="N!",                            name_length=2,  handler=docmd_fact,        hp42s_code=0x00000062, argtype=ARG_NONE,  flags=FLAG_NONE }, // FACT
            new command_spec { name="GAM\x00cd\x00c1",               name_length=5,  handler=docmd_gamma,       hp42s_code=0x0000a074, argtype=ARG_NONE,  flags=FLAG_NONE }, // GAMMA
            new command_spec { name="RAN",                           name_length=3,  handler=docmd_ran,         hp42s_code=0x0000a071, argtype=ARG_NONE,  flags=FLAG_NONE }, // RAN
            new command_spec { name="SEED",                          name_length=4,  handler=docmd_seed,        hp42s_code=0x0000a073, argtype=ARG_NONE,  flags=FLAG_NONE }, // SEED
            new command_spec { name="LBL",                           name_length=3,  handler=docmd_lbl,         hp42s_code=0x010000cf, argtype=ARG_OTHER, flags=FLAG_PRGM_ONLY }, // LBL
            new command_spec { name="RTN",                           name_length=3,  handler=docmd_rtn,         hp42s_code=0x00000085, argtype=ARG_NONE,  flags=FLAG_NONE }, // RTN
            new command_spec { name="INPUT",                         name_length=5,  handler=docmd_input,       hp42s_code=0x01c5f2d0, argtype=ARG_VAR,   flags=FLAG_PRGM_ONLY }, // INPUT
            new command_spec { name="VIEW",                          name_length=4,  handler=docmd_view,        hp42s_code=0x00800098, argtype=ARG_VAR,   flags=FLAG_NONE }, // VIEW
            new command_spec { name="AVIEW",                         name_length=5,  handler=docmd_aview,       hp42s_code=0x0000007e, argtype=ARG_NONE,  flags=FLAG_NONE }, // AVIEW
            new command_spec { name="XEQ",                           name_length=3,  handler=docmd_xeq,         hp42s_code=0x01a700ae, argtype=ARG_LBL,   flags=FLAG_NONE }, // XEQ
            new command_spec { name="PROM\x00d0\x00d4",              name_length=6,  handler=docmd_prompt,      hp42s_code=0x0000008e, argtype=ARG_NONE,  flags=FLAG_NONE }, // PROMPT
            new command_spec { name="PSE",                           name_length=3,  handler=docmd_pse,         hp42s_code=0x00000089, argtype=ARG_NONE,  flags=FLAG_NONE }, // PSE
            new command_spec { name="ISG",                           name_length=3,  handler=docmd_isg,         hp42s_code=0x00960096, argtype=ARG_REAL,  flags=FLAG_NONE }, // ISG
            new command_spec { name="DSE",                           name_length=3,  handler=docmd_dse,         hp42s_code=0x00970097, argtype=ARG_REAL,  flags=FLAG_NONE }, // DSE
            new command_spec { name="AIP",                           name_length=3,  handler=docmd_aip,         hp42s_code=0x0000a631, argtype=ARG_NONE,  flags=FLAG_NONE }, // AIP
            new command_spec { name="XTOA",                          name_length=4,  handler=docmd_xtoa,        hp42s_code=0x0000a66f, argtype=ARG_NONE,  flags=FLAG_NONE }, // XTOA
            new command_spec { name="AGRA\x00d0\x00c8",              name_length=6,  handler=docmd_agraph,      hp42s_code=0x0000a764, argtype=ARG_NONE,  flags=FLAG_NONE }, // AGRAPH
            new command_spec { name="PIXEL",                         name_length=5,  handler=docmd_pixel,       hp42s_code=0x0000a765, argtype=ARG_NONE,  flags=FLAG_NONE }, // PIXEL
            new command_spec { name="BEEP",                          name_length=4,  handler=docmd_beep,        hp42s_code=0x00000086, argtype=ARG_NONE,  flags=FLAG_NONE }, // BEEP
            new command_spec { name="TONE",                          name_length=4,  handler=docmd_tone,        hp42s_code=0x00d7009f, argtype=ARG_NUM9,  flags=FLAG_NONE }, // TONE
            new command_spec { name="MVAR",                          name_length=4,  handler=docmd_mvar,        hp42s_code=0x00900000, argtype=ARG_RVAR,  flags=FLAG_NONE }, // MVAR
            new command_spec { name="VARM\x00c5\x00ce\x00d5",        name_length=7,  handler=docmd_varmenu,     hp42s_code=0x00c1f2f8, argtype=ARG_PRGM,  flags=FLAG_NONE }, // VARMENU
            new command_spec { name="GETK\x00c5\x00d9",              name_length=6,  handler=docmd_getkey,      hp42s_code=0x0000a26e, argtype=ARG_NONE,  flags=FLAG_NONE }, // GETKEY
            new command_spec { name="MENU",                          name_length=4,  handler=docmd_menu,        hp42s_code=0x0000a25e, argtype=ARG_NONE,  flags=FLAG_PRGM_ONLY }, // MENU
            new command_spec { name="KEYG",                          name_length=4,  handler=null,              hp42s_code=0x02000000, argtype=ARG_MKEY,  flags=FLAG_NONE }, // KEYG
            new command_spec { name="KEYX",                          name_length=4,  handler=null,              hp42s_code=0x02000000, argtype=ARG_MKEY,  flags=FLAG_NONE }, // KEYX
            new command_spec { name="X=0?",                          name_length=4,  handler=docmd_x_eq_0,      hp42s_code=0x00000067, argtype=ARG_NONE,  flags=FLAG_NONE }, // X_EQ_0
            new command_spec { name="X\x000c0?",                     name_length=4,  handler=docmd_x_ne_0,      hp42s_code=0x00000063, argtype=ARG_NONE,  flags=FLAG_NONE }, // X_NE_0
            new command_spec { name="X<0?",                          name_length=4,  handler=docmd_x_lt_0,      hp42s_code=0x00000066, argtype=ARG_NONE,  flags=FLAG_NONE }, // X_LT_0
            new command_spec { name="X>0?",                          name_length=4,  handler=docmd_x_gt_0,      hp42s_code=0x00000064, argtype=ARG_NONE,  flags=FLAG_NONE }, // X_GT_0
            new command_spec { name="X\x00090?",                     name_length=4,  handler=docmd_x_le_0,      hp42s_code=0x0000007b, argtype=ARG_NONE,  flags=FLAG_NONE }, // X_LE_0
            new command_spec { name="X\x000b0?",                     name_length=4,  handler=docmd_x_ge_0,      hp42s_code=0x0000a25f, argtype=ARG_NONE,  flags=FLAG_NONE }, // X_GE_0
            new command_spec { name="X=Y?",                          name_length=4,  handler=docmd_x_eq_y,      hp42s_code=0x00000078, argtype=ARG_NONE,  flags=FLAG_NONE }, // X_EQ_Y
            new command_spec { name="X\x000cY?",                     name_length=4,  handler=docmd_x_ne_y,      hp42s_code=0x00000079, argtype=ARG_NONE,  flags=FLAG_NONE }, // X_NE_Y
            new command_spec { name="X<Y?",                          name_length=4,  handler=docmd_x_lt_y,      hp42s_code=0x00000044, argtype=ARG_NONE,  flags=FLAG_NONE }, // X_LT_Y
            new command_spec { name="X>Y?",                          name_length=4,  handler=docmd_x_gt_y,      hp42s_code=0x00000045, argtype=ARG_NONE,  flags=FLAG_NONE }, // X_GT_Y
            new command_spec { name="X\x0009Y?",                     name_length=4,  handler=docmd_x_le_y,      hp42s_code=0x00000046, argtype=ARG_NONE,  flags=FLAG_NONE }, // X_LE_Y
            new command_spec { name="X\x000bY?",                     name_length=4,  handler=docmd_x_ge_y,      hp42s_code=0x0000a260, argtype=ARG_NONE,  flags=FLAG_NONE }, // X_GE_Y
            new command_spec { name="PR\x0005",                      name_length=3,  handler=docmd_prsigma,     hp42s_code=0x0000a752, argtype=ARG_NONE,  flags=FLAG_NONE }, // PRSIGMA
            new command_spec { name="PRP",                           name_length=3,  handler=docmd_prp,         hp42s_code=0x02000000, argtype=ARG_PRGM,  flags=FLAG_IMMED }, // PRP
            new command_spec { name="PRV",                           name_length=3,  handler=docmd_prv,         hp42s_code=0x00b1f2d9, argtype=ARG_NAMED, flags=FLAG_NONE }, // PRV
            new command_spec { name="PRST\x00cb",                    name_length=5,  handler=docmd_prstk,       hp42s_code=0x0000a753, argtype=ARG_NONE,  flags=FLAG_NONE }, // PRSTK
            new command_spec { name="PRA",                           name_length=3,  handler=docmd_pra,         hp42s_code=0x0000a748, argtype=ARG_NONE,  flags=FLAG_NONE }, // PRA
            new command_spec { name="PRX",                           name_length=3,  handler=docmd_prx,         hp42s_code=0x0000a754, argtype=ARG_NONE,  flags=FLAG_NONE }, // PRX
            new command_spec { name="PRUSR",                         name_length=5,  handler=docmd_prusr,       hp42s_code=0x0000a761, argtype=ARG_NONE,  flags=FLAG_NONE }, // PRUSR
            new command_spec { name="LIST",                          name_length=4,  handler=docmd_list,        hp42s_code=0x02000000, argtype=ARG_COUNT, flags=FLAG_IMMED }, // LIST
            new command_spec { name="ADV",                           name_length=3,  handler=docmd_adv,         hp42s_code=0x0000008f, argtype=ARG_NONE,  flags=FLAG_NONE }, // ADV
            new command_spec { name="PRLCD",                         name_length=5,  handler=docmd_prlcd,       hp42s_code=0x0000a762, argtype=ARG_NONE,  flags=FLAG_NONE }, // PRLCD
            new command_spec { name="DELAY",                         name_length=5,  handler=docmd_delay,       hp42s_code=0x0000a760, argtype=ARG_NONE,  flags=FLAG_NONE }, // DELAY
            new command_spec { name="P\x00d2ON",                     name_length=4,  handler=docmd_pon,         hp42s_code=0x0000a75e, argtype=ARG_NONE,  flags=FLAG_NONE }, // PON
            new command_spec { name="P\x00d2OFF",                    name_length=5,  handler=docmd_poff,        hp42s_code=0x0000a75f, argtype=ARG_NONE,  flags=FLAG_NONE }, // POFF
            new command_spec { name="MAN",                           name_length=3,  handler=docmd_man,         hp42s_code=0x0000a75b, argtype=ARG_NONE,  flags=FLAG_NONE }, // MAN
            new command_spec { name="NORM",                          name_length=4,  handler=docmd_norm,        hp42s_code=0x0000a75c, argtype=ARG_NONE,  flags=FLAG_NONE }, // NORM
            new command_spec { name="TRACE",                         name_length=5,  handler=docmd_trace,       hp42s_code=0x0000a75d, argtype=ARG_NONE,  flags=FLAG_NONE }, // TRACE
            new command_spec { name="\x0005+",                       name_length=2,  handler=docmd_sigmaadd,    hp42s_code=0x00000047, argtype=ARG_NONE,  flags=FLAG_NONE }, // SIGMAADD
            new command_spec { name="\x0005-",                       name_length=2,  handler=docmd_sigmasub,    hp42s_code=0x00000048, argtype=ARG_NONE,  flags=FLAG_NONE }, // SIGMASUB
            new command_spec { name="GTO",                           name_length=3,  handler=docmd_gto,         hp42s_code=0x01a60000, argtype=ARG_LBL,   flags=FLAG_NONE }, // GTO
            new command_spec { name="END",                           name_length=3,  handler=docmd_end,         hp42s_code=0x01000000, argtype=ARG_NONE,  flags=FLAG_NONE }, // END
            new command_spec { name="",                              name_length=0,  handler=docmd_number,      hp42s_code=0x01000000, argtype=ARG_NONE,  flags=FLAG_HIDDEN }, // NUMBER
            new command_spec { name="",                              name_length=0,  handler=docmd_string,      hp42s_code=0x01000000, argtype=ARG_NONE,  flags=FLAG_HIDDEN }, // STRING
            new command_spec { name="RUN",                           name_length=3,  handler=null,              hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_HIDDEN }, // RUN
            new command_spec { name="SST",                           name_length=3,  handler=null,              hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_NONE }, // SST
            new command_spec { name="GTO .",                         name_length=5,  handler=docmd_gtodot,      hp42s_code=0x02000000, argtype=ARG_OTHER, flags=FLAG_IMMED }, // GTODOT
            new command_spec { name="GTO ..",                        name_length=6,  handler=docmd_gtodotdot,   hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_IMMED }, // GTODOTDOT
            new command_spec { name="STOP",                          name_length=4,  handler=docmd_stop,        hp42s_code=0x00000084, argtype=ARG_NONE,  flags=FLAG_NONE }, // STOP
            new command_spec { name="NEW\x00cd\x00c1\x00d4",         name_length=6,  handler=docmd_newmat,      hp42s_code=0x0000a6da, argtype=ARG_NONE,  flags=FLAG_NONE }, // NEWMAT
            new command_spec { name="R^",                            name_length=2,  handler=docmd_rup,         hp42s_code=0x00000074, argtype=ARG_NONE,  flags=FLAG_NONE }, // RUP
            new command_spec { name="RE\x00c1L?",                    name_length=5,  handler=docmd_real_t,      hp42s_code=0x0000a265, argtype=ARG_NONE,  flags=FLAG_NONE }, // REAL_T
            new command_spec { name="CPX?",                          name_length=4,  handler=docmd_cpx_t,       hp42s_code=0x0000a267, argtype=ARG_NONE,  flags=FLAG_NONE }, // CPX_T
            new command_spec { name="STR?",                          name_length=4,  handler=docmd_str_t,       hp42s_code=0x0000a268, argtype=ARG_NONE,  flags=FLAG_NONE }, // STR_T
            new command_spec { name="MAT?",                          name_length=4,  handler=docmd_mat_t,       hp42s_code=0x0000a266, argtype=ARG_NONE,  flags=FLAG_NONE }, // MAT_T
            new command_spec { name="DIM?",                          name_length=4,  handler=docmd_dim_t,       hp42s_code=0x0000a6e7, argtype=ARG_NONE,  flags=FLAG_NONE }, // DIM_T
            new command_spec { name="AS\x00d3\x00c9GN",              name_length=6,  handler=null,              hp42s_code=0x02000000, argtype=ARG_NAMED, flags=FLAG_NONE }, // ASSIGNa
            new command_spec { name="",                              name_length=0,  handler=null,              hp42s_code=0x02000000, argtype=ARG_CKEY,  flags=FLAG_HIDDEN }, // ASSIGNb
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn01,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN01
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn02,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN02
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn03,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN03
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn04,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN04
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn05,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN05
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn06,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN06
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn07,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN07
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn08,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN08
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn09,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN09
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn10,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN10
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn11,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN11
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn12,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN12
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn13,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN13
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn14,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN14
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn15,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN15
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn16,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN16
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn17,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN17
            new command_spec { name="",                              name_length=0,  handler=docmd_asgn18,      hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // ASGN18
            new command_spec { name="ON",                            name_length=2,  handler=docmd_on,          hp42s_code=0x0000a270, argtype=ARG_NONE,  flags=FLAG_NONE }, // ON
            new command_spec { name="OFF",                           name_length=3,  handler=docmd_off,         hp42s_code=0x0000008d, argtype=ARG_NONE,  flags=FLAG_NONE }, // OFF
            new command_spec { name="KEY 1 GTO",                     name_length=9,  handler=docmd_key1g,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY1G
            new command_spec { name="KEY 2 GTO",                     name_length=9,  handler=docmd_key2g,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY2G
            new command_spec { name="KEY 3 GTO",                     name_length=9,  handler=docmd_key3g,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY3G
            new command_spec { name="KEY 4 GTO",                     name_length=9,  handler=docmd_key4g,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY4G
            new command_spec { name="KEY 5 GTO",                     name_length=9,  handler=docmd_key5g,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY5G
            new command_spec { name="KEY 6 GTO",                     name_length=9,  handler=docmd_key6g,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY6G
            new command_spec { name="KEY 7 GTO",                     name_length=9,  handler=docmd_key7g,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY7G
            new command_spec { name="KEY 8 GTO",                     name_length=9,  handler=docmd_key8g,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY8G
            new command_spec { name="KEY 9 GTO",                     name_length=9,  handler=docmd_key9g,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY9G
            new command_spec { name="KEY 1 XEQ",                     name_length=9,  handler=docmd_key1x,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY1X
            new command_spec { name="KEY 2 XEQ",                     name_length=9,  handler=docmd_key2x,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY2X
            new command_spec { name="KEY 3 XEQ",                     name_length=9,  handler=docmd_key3x,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY3X
            new command_spec { name="KEY 4 XEQ",                     name_length=9,  handler=docmd_key4x,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY4X
            new command_spec { name="KEY 5 XEQ",                     name_length=9,  handler=docmd_key5x,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY5X
            new command_spec { name="KEY 6 XEQ",                     name_length=9,  handler=docmd_key6x,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY6X
            new command_spec { name="KEY 7 XEQ",                     name_length=9,  handler=docmd_key7x,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY7X
            new command_spec { name="KEY 8 XEQ",                     name_length=9,  handler=docmd_key8x,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY8X
            new command_spec { name="KEY 9 XEQ",                     name_length=9,  handler=docmd_key9x,       hp42s_code=0x01000000, argtype=ARG_LBL,   flags=FLAG_HIDDEN }, // KEY9X
            new command_spec { name="",                              name_length=0,  handler=null,              hp42s_code=0x02000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // VMEXEC
            new command_spec { name="STO",                           name_length=3,  handler=docmd_vmsto,       hp42s_code=0x02000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // VMSTO
            new command_spec { name="\x0005REG",                     name_length=4,  handler=docmd_sigma_reg,   hp42s_code=0x00d30099, argtype=ARG_NUM99, flags=FLAG_NONE }, // SIGMAREG
            new command_spec { name="\x0005R\x00c5G?",               name_length=5,  handler=docmd_sigma_reg_t, hp42s_code=0x0000a678, argtype=ARG_NONE,  flags=FLAG_NONE }, // SIGMAREG_T
            new command_spec { name="CLD",                           name_length=3,  handler=docmd_cld,         hp42s_code=0x0000007f, argtype=ARG_NONE,  flags=FLAG_NONE }, // CLD
            new command_spec { name="ACOSH",                         name_length=5,  handler=docmd_acosh,       hp42s_code=0x0000a066, argtype=ARG_NONE,  flags=FLAG_NONE }, // ACOSH
            new command_spec { name="ALEN\x00c7",                    name_length=5,  handler=docmd_aleng,       hp42s_code=0x0000a641, argtype=ARG_NONE,  flags=FLAG_NONE }, // ALENG
            new command_spec { name="ALL\x0005",                     name_length=4,  handler=docmd_allsigma,    hp42s_code=0x0000a0ae, argtype=ARG_NONE,  flags=FLAG_NONE }, // ALLSIGMA
            new command_spec { name="AND",                           name_length=3,  handler=docmd_and,         hp42s_code=0x0000a588, argtype=ARG_NONE,  flags=FLAG_NONE }, // AND
            new command_spec { name="AOFF",                          name_length=4,  handler=docmd_aoff,        hp42s_code=0x0000008b, argtype=ARG_NONE,  flags=FLAG_NONE }, // AOFF
            new command_spec { name="AON",                           name_length=3,  handler=docmd_aon,         hp42s_code=0x0000008c, argtype=ARG_NONE,  flags=FLAG_NONE }, // AON
            new command_spec { name="AROT",                          name_length=4,  handler=docmd_arot,        hp42s_code=0x0000a646, argtype=ARG_NONE,  flags=FLAG_NONE }, // AROT
            new command_spec { name="ASHF",                          name_length=4,  handler=docmd_ashf,        hp42s_code=0x00000088, argtype=ARG_NONE,  flags=FLAG_NONE }, // ASHF
            new command_spec { name="ASINH",                         name_length=5,  handler=docmd_asinh,       hp42s_code=0x0000a064, argtype=ARG_NONE,  flags=FLAG_NONE }, // ASINH
            new command_spec { name="AT\x00c1NH",                    name_length=5,  handler=docmd_atanh,       hp42s_code=0x0000a065, argtype=ARG_NONE,  flags=FLAG_NONE }, // ATANH
            new command_spec { name="ATOX",                          name_length=4,  handler=docmd_atox,        hp42s_code=0x0000a647, argtype=ARG_NONE,  flags=FLAG_NONE }, // ATOX
            new command_spec { name="BASE+",                         name_length=5,  handler=docmd_baseadd,     hp42s_code=0x0000a0e6, argtype=ARG_NONE,  flags=FLAG_NONE }, // BASEADD
            new command_spec { name="BASE-",                         name_length=5,  handler=docmd_basesub,     hp42s_code=0x0000a0e7, argtype=ARG_NONE,  flags=FLAG_NONE }, // BASESUB
            new command_spec { name="BASE\x0001",                    name_length=5,  handler=docmd_basemul,     hp42s_code=0x0000a0e8, argtype=ARG_NONE,  flags=FLAG_NONE }, // BASEMUL
            new command_spec { name="BASE\x0000",                    name_length=5,  handler=docmd_basediv,     hp42s_code=0x0000a0e9, argtype=ARG_NONE,  flags=FLAG_NONE }, // BASEDIV
            new command_spec { name="B\x00c1\x00d3\x00c5+/-",        name_length=7,  handler=docmd_basechs,     hp42s_code=0x0000a0ea, argtype=ARG_NONE,  flags=FLAG_NONE }, // BASECHS
            new command_spec { name="BEST",                          name_length=4,  handler=docmd_best,        hp42s_code=0x0000a09f, argtype=ARG_NONE,  flags=FLAG_NONE }, // BEST
            new command_spec { name="BINM",                          name_length=4,  handler=docmd_binm,        hp42s_code=0x0000a0e5, argtype=ARG_NONE,  flags=FLAG_NONE }, // BINM
            new command_spec { name="BIT?",                          name_length=4,  handler=docmd_bit_t,       hp42s_code=0x0000a58c, argtype=ARG_NONE,  flags=FLAG_NONE }, // BIT_T
            new command_spec { name="BST",                           name_length=3,  handler=null,              hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_NONE }, // BST
            new command_spec { name="CORR",                          name_length=4,  handler=docmd_corr,        hp42s_code=0x0000a0a7, argtype=ARG_NONE,  flags=FLAG_NONE }, // CORR
            new command_spec { name="COSH",                          name_length=4,  handler=docmd_cosh,        hp42s_code=0x0000a062, argtype=ARG_NONE,  flags=FLAG_NONE }, // COSH
            new command_spec { name="CROSS",                         name_length=5,  handler=docmd_cross,       hp42s_code=0x0000a6ca, argtype=ARG_NONE,  flags=FLAG_NONE }, // CROSS
            new command_spec { name="CUST\x00cf\x00cd",              name_length=6,  handler=docmd_custom,      hp42s_code=0x0000a26f, argtype=ARG_NONE,  flags=FLAG_NONE }, // CUSTOM
            new command_spec { name="DECM",                          name_length=4,  handler=docmd_decm,        hp42s_code=0x0000a0e3, argtype=ARG_NONE,  flags=FLAG_NONE }, // DECM
            new command_spec { name="DELR",                          name_length=4,  handler=docmd_delr,        hp42s_code=0x0000a0ab, argtype=ARG_NONE,  flags=FLAG_NONE }, // DELR
            new command_spec { name="DET",                           name_length=3,  handler=docmd_det,         hp42s_code=0x0000a6cc, argtype=ARG_NONE,  flags=FLAG_NONE }, // DET
            new command_spec { name="DIM",                           name_length=3,  handler=docmd_dim,         hp42s_code=0x00c4f2ec, argtype=ARG_MAT,   flags=FLAG_NONE }, // DIM
            new command_spec { name="DOT",                           name_length=3,  handler=docmd_dot,         hp42s_code=0x0000a6cb, argtype=ARG_NONE,  flags=FLAG_NONE }, // DOT
            new command_spec { name="EDIT",                          name_length=4,  handler=docmd_edit,        hp42s_code=0x0000a6e1, argtype=ARG_NONE,  flags=FLAG_NONE }, // EDIT
            new command_spec { name="EDITN",                         name_length=5,  handler=docmd_editn,       hp42s_code=0x00c6f2ef, argtype=ARG_MAT,   flags=FLAG_NONE }, // EDITN
            new command_spec { name="EXITA\x00cc\x00cc",             name_length=7,  handler=docmd_exitall,     hp42s_code=0x0000a26c, argtype=ARG_NONE,  flags=FLAG_NONE }, // EXITALL
            new command_spec { name="EXPF",                          name_length=4,  handler=docmd_expf,        hp42s_code=0x0000a0a0, argtype=ARG_NONE,  flags=FLAG_NONE }, // EXPF
            new command_spec { name="E^X-\x00b1",                    name_length=5,  handler=docmd_e_pow_x_1,   hp42s_code=0x00000058, argtype=ARG_NONE,  flags=FLAG_NONE }, // E_POW_X_1
            new command_spec { name="FCSTX",                         name_length=5,  handler=docmd_fcstx,       hp42s_code=0x0000a0a8, argtype=ARG_NONE,  flags=FLAG_NONE }, // FCSTX
            new command_spec { name="FCSTY",                         name_length=5,  handler=docmd_fcsty,       hp42s_code=0x0000a0a9, argtype=ARG_NONE,  flags=FLAG_NONE }, // FCSTY
            new command_spec { name="FNRM",                          name_length=4,  handler=docmd_fnrm,        hp42s_code=0x0000a6cf, argtype=ARG_NONE,  flags=FLAG_NONE }, // FNRM
            new command_spec { name="GETM",                          name_length=4,  handler=docmd_getm,        hp42s_code=0x0000a6e8, argtype=ARG_NONE,  flags=FLAG_NONE }, // GETM
            new command_spec { name="GROW",                          name_length=4,  handler=docmd_grow,        hp42s_code=0x0000a6e3, argtype=ARG_NONE,  flags=FLAG_NONE }, // GROW
            new command_spec { name="HEXM",                          name_length=4,  handler=docmd_hexm,        hp42s_code=0x0000a0e2, argtype=ARG_NONE,  flags=FLAG_NONE }, // HEXM
            new command_spec { name="HMS+",                          name_length=4,  handler=docmd_hmsadd,      hp42s_code=0x00000049, argtype=ARG_NONE,  flags=FLAG_NONE }, // HMSADD
            new command_spec { name="HMS-",                          name_length=4,  handler=docmd_hmssub,      hp42s_code=0x0000004a, argtype=ARG_NONE,  flags=FLAG_NONE }, // HMSSUB
            new command_spec { name="I+",                            name_length=2,  handler=docmd_i_add,       hp42s_code=0x0000a6d2, argtype=ARG_NONE,  flags=FLAG_NONE }, // I_ADD
            new command_spec { name="I-",                            name_length=2,  handler=docmd_i_sub,       hp42s_code=0x0000a6d3, argtype=ARG_NONE,  flags=FLAG_NONE }, // I_SUB
            new command_spec { name="INDEX",                         name_length=5,  handler=docmd_index,       hp42s_code=0x0087f2da, argtype=ARG_MAT,   flags=FLAG_NONE }, // INDEX
            new command_spec { name="INSR",                          name_length=4,  handler=docmd_insr,        hp42s_code=0x0000a0aa, argtype=ARG_NONE,  flags=FLAG_NONE }, // INSR
            new command_spec { name="INTEG",                         name_length=5,  handler=docmd_integ,       hp42s_code=0x00b6f2ea, argtype=ARG_RVAR,  flags=FLAG_NONE }, // INTEG
            new command_spec { name="INV\x00d2\x00d4",               name_length=5,  handler=docmd_invrt,       hp42s_code=0x0000a6ce, argtype=ARG_NONE,  flags=FLAG_NONE }, // INVRT
            new command_spec { name="J+",                            name_length=2,  handler=docmd_j_add,       hp42s_code=0x0000a6d4, argtype=ARG_NONE,  flags=FLAG_NONE }, // J_ADD
            new command_spec { name="J-",                            name_length=2,  handler=docmd_j_sub,       hp42s_code=0x0000a6d5, argtype=ARG_NONE,  flags=FLAG_NONE }, // J_SUB
            new command_spec { name="LINF",                          name_length=4,  handler=docmd_linf,        hp42s_code=0x0000a0a1, argtype=ARG_NONE,  flags=FLAG_NONE }, // LINF
            new command_spec { name="LIN\x0005",                     name_length=4,  handler=docmd_linsigma,    hp42s_code=0x0000a0ad, argtype=ARG_NONE,  flags=FLAG_NONE }, // LINSIGMA
            new command_spec { name="LN1+\x00d8",                    name_length=5,  handler=docmd_ln_1_x,      hp42s_code=0x00000065, argtype=ARG_NONE,  flags=FLAG_NONE }, // LN_1_X
            new command_spec { name="LOGF",                          name_length=4,  handler=docmd_logf,        hp42s_code=0x0000a0a2, argtype=ARG_NONE,  flags=FLAG_NONE }, // LOGF
            new command_spec { name="MEAN",                          name_length=4,  handler=docmd_mean,        hp42s_code=0x0000007c, argtype=ARG_NONE,  flags=FLAG_NONE }, // MEAN
            new command_spec { name="NOT",                           name_length=3,  handler=docmd_not,         hp42s_code=0x0000a587, argtype=ARG_NONE,  flags=FLAG_NONE }, // NOT
            new command_spec { name="OCTM",                          name_length=4,  handler=docmd_octm,        hp42s_code=0x0000a0e4, argtype=ARG_NONE,  flags=FLAG_NONE }, // OCTM
            new command_spec { name="OLD",                           name_length=3,  handler=docmd_old,         hp42s_code=0x0000a6db, argtype=ARG_NONE,  flags=FLAG_NONE }, // OLD
            new command_spec { name="OR",                            name_length=2,  handler=docmd_or,          hp42s_code=0x0000a589, argtype=ARG_NONE,  flags=FLAG_NONE }, // OR
            new command_spec { name="P\x00c7\x00cdSLV",              name_length=6,  handler=docmd_pgmslv,      hp42s_code=0x00b5f2e9, argtype=ARG_PRGM,  flags=FLAG_NONE }, // PGMSLV
            new command_spec { name="P\x00c7\x00cdINT",              name_length=6,  handler=docmd_pgmint,      hp42s_code=0x00b4f2e8, argtype=ARG_PRGM,  flags=FLAG_NONE }, // PGMINT
            new command_spec { name="POSA",                          name_length=4,  handler=docmd_posa,        hp42s_code=0x0000a65c, argtype=ARG_NONE,  flags=FLAG_NONE }, // POSA
            new command_spec { name="PUTM",                          name_length=4,  handler=docmd_putm,        hp42s_code=0x0000a6e9, argtype=ARG_NONE,  flags=FLAG_NONE }, // PUTM
            new command_spec { name="PWRF",                          name_length=4,  handler=docmd_pwrf,        hp42s_code=0x0000a0a3, argtype=ARG_NONE,  flags=FLAG_NONE }, // PWRF
            new command_spec { name="RCLEL",                         name_length=5,  handler=docmd_rclel,       hp42s_code=0x0000a6d7, argtype=ARG_NONE,  flags=FLAG_NONE }, // RCLEL
            new command_spec { name="RCLIJ",                         name_length=5,  handler=docmd_rclij,       hp42s_code=0x0000a6d9, argtype=ARG_NONE,  flags=FLAG_NONE }, // RCLIJ
            new command_spec { name="RNRM",                          name_length=4,  handler=docmd_rnrm,        hp42s_code=0x0000a6ed, argtype=ARG_NONE,  flags=FLAG_NONE }, // RNRM
            new command_spec { name="ROTXY",                         name_length=5,  handler=docmd_rotxy,       hp42s_code=0x0000a58b, argtype=ARG_NONE,  flags=FLAG_NONE }, // ROTXY
            new command_spec { name="RSUM",                          name_length=4,  handler=docmd_rsum,        hp42s_code=0x0000a6d0, argtype=ARG_NONE,  flags=FLAG_NONE }, // RSUM
            new command_spec { name="R<>R",                          name_length=4,  handler=docmd_swap_r,      hp42s_code=0x0000a6d1, argtype=ARG_NONE,  flags=FLAG_NONE }, // SWAP_R
            new command_spec { name="SDEV",                          name_length=4,  handler=docmd_sdev,        hp42s_code=0x0000007d, argtype=ARG_NONE,  flags=FLAG_NONE }, // SDEV
            new command_spec { name="SINH",                          name_length=4,  handler=docmd_sinh,        hp42s_code=0x0000a061, argtype=ARG_NONE,  flags=FLAG_NONE }, // SINH
            new command_spec { name="SLOPE",                         name_length=5,  handler=docmd_slope,       hp42s_code=0x0000a0a4, argtype=ARG_NONE,  flags=FLAG_NONE }, // SLOPE
            new command_spec { name="SOLVE",                         name_length=5,  handler=docmd_solve,       hp42s_code=0x00b7f2eb, argtype=ARG_RVAR,  flags=FLAG_NONE }, // SOLVE
            new command_spec { name="STOEL",                         name_length=5,  handler=docmd_stoel,       hp42s_code=0x0000a6d6, argtype=ARG_NONE,  flags=FLAG_NONE }, // STOEL
            new command_spec { name="STOIJ",                         name_length=5,  handler=docmd_stoij,       hp42s_code=0x0000a6d8, argtype=ARG_NONE,  flags=FLAG_NONE }, // STOIJ
            new command_spec { name="SUM",                           name_length=3,  handler=docmd_sum,         hp42s_code=0x0000a0a5, argtype=ARG_NONE,  flags=FLAG_NONE }, // SUM
            new command_spec { name="TANH",                          name_length=4,  handler=docmd_tanh,        hp42s_code=0x0000a063, argtype=ARG_NONE,  flags=FLAG_NONE }, // TANH
            new command_spec { name="TRANS",                         name_length=5,  handler=docmd_trans,       hp42s_code=0x0000a6c9, argtype=ARG_NONE,  flags=FLAG_NONE }, // TRANS
            new command_spec { name="UVEC",                          name_length=4,  handler=docmd_uvec,        hp42s_code=0x0000a6cd, argtype=ARG_NONE,  flags=FLAG_NONE }, // UVEC
            new command_spec { name="WM\x00c5\x00c1N",               name_length=5,  handler=docmd_wmean,       hp42s_code=0x0000a0ac, argtype=ARG_NONE,  flags=FLAG_NONE }, // WMEAN
            new command_spec { name="WRAP",                          name_length=4,  handler=docmd_wrap,        hp42s_code=0x0000a6e2, argtype=ARG_NONE,  flags=FLAG_NONE }, // WRAP
            new command_spec { name="X<>",                           name_length=3,  handler=docmd_x_swap,      hp42s_code=0x008600ce, argtype=ARG_VAR,   flags=FLAG_NONE }, // X_SWAP
            new command_spec { name="XOR",                           name_length=3,  handler=docmd_xor,         hp42s_code=0x0000a58a, argtype=ARG_NONE,  flags=FLAG_NONE }, // XOR
            new command_spec { name="YINT",                          name_length=4,  handler=docmd_yint,        hp42s_code=0x0000a0a6, argtype=ARG_NONE,  flags=FLAG_NONE }, // YINT
            new command_spec { name="\x000fDEC",                     name_length=4,  handler=docmd_to_dec,      hp42s_code=0x0000005f, argtype=ARG_NONE,  flags=FLAG_NONE }, // TO_DEC
            new command_spec { name="\x000fOCT",                     name_length=4,  handler=docmd_to_oct,      hp42s_code=0x0000006f, argtype=ARG_NONE,  flags=FLAG_NONE }, // TO_OCT
            new command_spec { name="\x0010",                        name_length=1,  handler=docmd_left,        hp42s_code=0x0000a6dc, argtype=ARG_NONE,  flags=FLAG_NONE }, // LEFT
            new command_spec { name="^",                             name_length=1,  handler=docmd_up,          hp42s_code=0x0000a6de, argtype=ARG_NONE,  flags=FLAG_NONE }, // UP
            new command_spec { name="\x000e",                        name_length=1,  handler=docmd_down,        hp42s_code=0x0000a6df, argtype=ARG_NONE,  flags=FLAG_NONE }, // DOWN
            new command_spec { name="\x000f",                        name_length=1,  handler=docmd_right,       hp42s_code=0x0000a6dd, argtype=ARG_NONE,  flags=FLAG_NONE }, // RIGHT
            new command_spec { name="%CH",                           name_length=3,  handler=docmd_percent_ch,  hp42s_code=0x0000004d, argtype=ARG_NONE,  flags=FLAG_NONE }, // PERCENT_CH
            new command_spec { name="SIMQ",                          name_length=4,  handler=docmd_simq,        hp42s_code=0x02000000, argtype=ARG_COUNT, flags=FLAG_HIDDEN + FLAG_NO_PRGM }, // SIMQ
            new command_spec { name="MATA",                          name_length=4,  handler=docmd_mata,        hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_HIDDEN + FLAG_NO_PRGM }, // MATA
            new command_spec { name="MATB",                          name_length=4,  handler=docmd_matb,        hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_HIDDEN + FLAG_NO_PRGM }, // MATB
            new command_spec { name="MATX",                          name_length=4,  handler=docmd_matx,        hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_HIDDEN + FLAG_NO_PRGM }, // MATX
            new command_spec { name="GOTO\x00a0\x00d2\x00ef\x00f7",  name_length=8,  handler=null,              hp42s_code=0x02000000, argtype=ARG_COUNT, flags=FLAG_HIDDEN }, // GOTOROW
            new command_spec { name="GOTO Column",                   name_length=11, handler=null,              hp42s_code=0x02000000, argtype=ARG_COUNT, flags=FLAG_HIDDEN }, // GOTOCOLUMN
            new command_spec { name="A...F",                         name_length=5,  handler=null,              hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_HIDDEN + FLAG_NO_PRGM }, // A_THRU_F
            new command_spec { name="CLALL",                         name_length=5,  handler=docmd_clall,       hp42s_code=0x02000000, argtype=ARG_NONE,  flags=FLAG_HIDDEN }, // CLALLb
            new command_spec { name="P\x00c7\x00cdSLV",              name_length=6,  handler=docmd_pgmslvi,     hp42s_code=0x02000000, argtype=ARG_PRGM,  flags=FLAG_HIDDEN }, // PGMSLVi
            new command_spec { name="P\x00c7\x00cdINT",              name_length=6,  handler=docmd_pgminti,     hp42s_code=0x02000000, argtype=ARG_PRGM,  flags=FLAG_HIDDEN }, // PGMINTi
            new command_spec { name="STO",                           name_length=3,  handler=docmd_vmsto2,      hp42s_code=0x02000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // VMSTO2
            new command_spec { name="SOLVE",                         name_length=5,  handler=docmd_vmsolve,     hp42s_code=0x02000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // VMSOLVE
            new command_spec { name="[MAX]",                         name_length=5,  handler=docmd_max,         hp42s_code=0x0000a6eb, argtype=ARG_NONE,  flags=FLAG_NONE }, // MAX
            new command_spec { name="[MIN]",                         name_length=5,  handler=docmd_min,         hp42s_code=0x0000a6ea, argtype=ARG_NONE,  flags=FLAG_NONE }, // MIN
            new command_spec { name="[FIND]",                        name_length=6,  handler=docmd_find,        hp42s_code=0x0000a6ec, argtype=ARG_NONE,  flags=FLAG_NONE }, // FIND
            new command_spec { name="XROM",                          name_length=4,  handler=docmd_xrom,        hp42s_code=0x01000000, argtype=ARG_OTHER, flags=FLAG_HIDDEN }, // XROM

            // Here endeth the original Free42 function table.
            // There are two extensions to this table out there in the wild today:
            // Underhill's COPAN extensions for surveyors
            // (http://www.underhill.ca/Software/Free42/Free42Copan.php),
            // and Byron Foster's Big Stack in 42s for iPhone
            // (http://free42iphone.googlecode.com/).
            // I'm merging both of these extensions into the main source repository --
            // not the actual implementations, but the function table entries, so that
            // programs created by these extended versions can be shared more easily.
            // Note that I had to move Byron's DROP extension from position 315 to
            // 329 to resolve the clash with the older Underhill extensions.
            // The function handlers (i.e. the docmd_ functions) are #defined as
            // docmd_xrom if the appropriate extensions are not present, so executing
            // them will raise a Nonexistent error, but you will see the function
            // *names* from this table in the actual program listings, regardless of
            // whether the extensions are present or not.
            // UPDATE: To support "pure" HP-42S behavior, all extensions can be
            // disabled at runtime, using the core_settings.enable_ext_* flags.
            // When an extension is disabled, its commands disappear from the FCN
            // catalog, are not recognized by XEQ, are displayed as their XROM
            // equivalents in programs, and raise a Nonexistent error when trying to
            // execute them from programs.
            // When a shell disables or enables an extension in response to the user
            // changing a setting in the Preferences dialog, it should call redisplay()
            // to make sure the display reflects the new setting.
            //

            // Underhill's COPAN
            new command_spec { name="OPENF", name_length=5, handler=docmd_openf,  hp42s_code=0x0000a7c1, argtype=ARG_NONE, flags=FLAG_NONE }, // OPENF
            new command_spec { name="CLOSF", name_length=5, handler=docmd_closef, hp42s_code=0x0000a7c2, argtype=ARG_NONE, flags=FLAG_NONE }, // CLOSF
            new command_spec { name="READP", name_length=5, handler=docmd_readp,  hp42s_code=0x0000a7c3, argtype=ARG_NONE, flags=FLAG_NONE }, // READP
            new command_spec { name="WRITP", name_length=5, handler=docmd_writp,  hp42s_code=0x0000a7c4, argtype=ARG_NONE, flags=FLAG_NONE }, // WRITP
            new command_spec { name="GETXY", name_length=5, handler=docmd_getxy,  hp42s_code=0x0000a7c5, argtype=ARG_NONE, flags=FLAG_NONE }, // GETXY
            new command_spec { name="PUTXY", name_length=5, handler=docmd_putxy,  hp42s_code=0x0000a7c6, argtype=ARG_NONE, flags=FLAG_NONE }, // PUTXY
            new command_spec { name="CLRP",  name_length=4, handler=docmd_clrp,   hp42s_code=0x0000a7c7, argtype=ARG_NONE, flags=FLAG_NONE }, // CLRP
            new command_spec { name="CLRD",  name_length=4, handler=docmd_clrd,   hp42s_code=0x0000a7c8, argtype=ARG_NONE, flags=FLAG_NONE }, // CLRD
            new command_spec { name="APPD",  name_length=4, handler=docmd_appd,   hp42s_code=0x0000a7c9, argtype=ARG_NONE, flags=FLAG_NONE }, // APPD
            new command_spec { name="GETN",  name_length=4, handler=docmd_getn,   hp42s_code=0x0000a7ca, argtype=ARG_NONE, flags=FLAG_NONE }, // GETN
            new command_spec { name="PUTN",  name_length=4, handler=docmd_putn,   hp42s_code=0x0000a7cb, argtype=ARG_NONE, flags=FLAG_NONE }, // PUTN
            new command_spec { name="GETZ",  name_length=4, handler=docmd_getz,   hp42s_code=0x0000a7cc, argtype=ARG_NONE, flags=FLAG_NONE }, // GETZ
            new command_spec { name="PUTZ",  name_length=4, handler=docmd_putz,   hp42s_code=0x0000a7cd, argtype=ARG_NONE, flags=FLAG_NONE }, // PUTZ
            new command_spec { name="DELP",  name_length=4, handler=docmd_delp,   hp42s_code=0x0000a7ce, argtype=ARG_NONE, flags=FLAG_NONE }, // DELP

            // Byron Foster's DROP for Bigstack
            new command_spec { name="DROP",  name_length=4, handler=docmd_drop,   hp42s_code=0x0000a271, argtype=ARG_NONE, flags=FLAG_NONE }, // DROP

            // Accelerometer, GPS, and compass support
            new command_spec { name="ACCEL",                  name_length=5, handler=docmd_accel,       hp42s_code=0x0000a7cf, argtype=ARG_NONE,  flags=FLAG_NONE }, // ACCEL
            new command_spec { name="LOCAT",                  name_length=5, handler=docmd_locat,       hp42s_code=0x0000a7d0, argtype=ARG_NONE,  flags=FLAG_NONE }, // LOCAT
            new command_spec { name="H\x00c5\x00c1D\x00c9NG", name_length=7, handler=docmd_heading,     hp42s_code=0x0000a7d1, argtype=ARG_NONE,  flags=FLAG_NONE }, // HEADING

            // Time Module & CX Time support
            new command_spec { name="ADATE",                  name_length=5, handler=docmd_adate,      hp42s_code=0x0000a681, argtype=ARG_NONE, flags=FLAG_NONE }, // ADATE
            new command_spec { name="AL\x00cdCAT",            name_length=6, handler=docmd_almcat,     hp42s_code=0x0000a682, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // ALMCAT
            new command_spec { name="AL\x00cdN\x00cfW",       name_length=6, handler=docmd_almnow,     hp42s_code=0x0000a683, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // ALMNOW
            new command_spec { name="ATIME",                  name_length=5, handler=docmd_atime,      hp42s_code=0x0000a684, argtype=ARG_NONE, flags=FLAG_NONE }, // ATIME
            new command_spec { name="AT\x00c9\x00cd\x00c524", name_length=7, handler=docmd_atime24,    hp42s_code=0x0000a685, argtype=ARG_NONE, flags=FLAG_NONE }, // ATIME24
            new command_spec { name="CL\x00cb12",             name_length=5, handler=docmd_clk12,      hp42s_code=0x0000a686, argtype=ARG_NONE, flags=FLAG_NONE }, // CLK12
            new command_spec { name="CL\x00cb24",             name_length=5, handler=docmd_clk24,      hp42s_code=0x0000a687, argtype=ARG_NONE, flags=FLAG_NONE }, // CLK24
            new command_spec { name="CLKT",                   name_length=4, handler=docmd_clkt,       hp42s_code=0x0000a688, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // CLKT
            new command_spec { name="CL\x00cbTD",             name_length=5, handler=docmd_clktd,      hp42s_code=0x0000a689, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // CLKTD
            new command_spec { name="CL\x00cf\x00c3K",        name_length=5, handler=docmd_clock,      hp42s_code=0x0000a68a, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // CLOCK
            new command_spec { name="CORR\x00c5\x00c3\x00d4", name_length=7, handler=docmd_correct,    hp42s_code=0x0000a68b, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // CORRECT
            new command_spec { name="DATE",                   name_length=4, handler=docmd_date,       hp42s_code=0x0000a68c, argtype=ARG_NONE, flags=FLAG_NONE }, // DATE
            new command_spec { name="DATE+",                  name_length=5, handler=docmd_date_plus,  hp42s_code=0x0000a68d, argtype=ARG_NONE, flags=FLAG_NONE }, // DATE_PLUS
            new command_spec { name="DDAYS",                  name_length=5, handler=docmd_ddays,      hp42s_code=0x0000a68e, argtype=ARG_NONE, flags=FLAG_NONE }, // DDAYS
            new command_spec { name="DMY",                    name_length=3, handler=docmd_dmy,        hp42s_code=0x0000a68f, argtype=ARG_NONE, flags=FLAG_NONE }, // DMY
            new command_spec { name="DOW",                    name_length=3, handler=docmd_dow,        hp42s_code=0x0000a690, argtype=ARG_NONE, flags=FLAG_NONE }, // DOW
            new command_spec { name="MDY",                    name_length=3, handler=docmd_mdy,        hp42s_code=0x0000a691, argtype=ARG_NONE, flags=FLAG_NONE }, // MDY
            new command_spec { name="RCLAF",                  name_length=5, handler=docmd_rclaf,      hp42s_code=0x0000a692, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // RCLAF
            new command_spec { name="RC\x00ccSW",             name_length=5, handler=docmd_rclsw,      hp42s_code=0x0000a693, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // RCLSW
            new command_spec { name="R\x00d5NSW",             name_length=5, handler=docmd_runsw,      hp42s_code=0x0000a694, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // RUNSW
            new command_spec { name="SETAF",                  name_length=5, handler=docmd_setaf,      hp42s_code=0x0000a695, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // SETAF
            new command_spec { name="S\x00c5\x00d4DATE",      name_length=7, handler=docmd_setdate,    hp42s_code=0x0000a696, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // SETDATE
            new command_spec { name="S\x00c5TIME",            name_length=6, handler=docmd_setime,     hp42s_code=0x0000a697, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // SETIME
            new command_spec { name="SE\x00d4SW",             name_length=5, handler=docmd_setsw,      hp42s_code=0x0000a698, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // SETSW
            new command_spec { name="ST\x00cf\x00d0SW",       name_length=6, handler=docmd_stopsw,     hp42s_code=0x0000a699, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // STOPSW
            new command_spec { name="SW",                     name_length=2, handler=docmd_sw,         hp42s_code=0x0000a69a, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // SW
            new command_spec { name="T+X",                    name_length=3, handler=docmd_t_plus_x,   hp42s_code=0x0000a69b, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // T_PLUS_X
            new command_spec { name="TIME",                   name_length=4, handler=docmd_time,       hp42s_code=0x0000a69c, argtype=ARG_NONE, flags=FLAG_NONE }, // TIME
            new command_spec { name="XYZALM",                 name_length=6, handler=docmd_xyzalm,     hp42s_code=0x0000a69d, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // XYZALM
            new command_spec { name="CLAL\x00cdA",            name_length=6, handler=docmd_clalma,     hp42s_code=0x0000a69f, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // CLALMA
            new command_spec { name="CLAL\x00cdX",            name_length=6, handler=docmd_clalmx,     hp42s_code=0x0000a6a0, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // CLALMX
            new command_spec { name="CLRALMS",                name_length=7, handler=docmd_clralms,    hp42s_code=0x0000a6a1, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // CLRALMS
            new command_spec { name="RCLALM",                 name_length=6, handler=docmd_rclalm,     hp42s_code=0x0000a6a2, argtype=ARG_NONE, flags=FLAG_HIDDEN }, // RCLALM
            new command_spec { name="SWPT",                   name_length=4, handler=docmd_swpt,       hp42s_code=0x0000a6a3, argtype=ARG_NONE, flags=FLAG_HIDDEN }  // SWPT
        };


        public static int[] fcn_cat = new int[]
        {
            App.CMD_ABS,      App.CMD_ACOS,      App.CMD_ACOSH,    App.CMD_ADV,        App.CMD_AGRAPH,  App.CMD_AIP,
            App.CMD_ALENG,    App.CMD_ALL,       App.CMD_ALLSIGMA, App.CMD_AND,        App.CMD_AOFF,    App.CMD_AON,
            App.CMD_ARCL,     App.CMD_AROT,      App.CMD_ASHF,     App.CMD_ASIN,       App.CMD_ASINH,   App.CMD_ASSIGNa,
            App.CMD_ASTO,     App.CMD_ATAN,      App.CMD_ATANH,    App.CMD_ATOX,       App.CMD_AVIEW,   App.CMD_BASEADD,
            App.CMD_BASESUB,  App.CMD_BASEMUL,   App.CMD_BASEDIV,  App.CMD_BASECHS,    App.CMD_BEEP,    App.CMD_BEST,
            App.CMD_BINM,     App.CMD_BIT_T,     App.CMD_BST,      App.CMD_CF,         App.CMD_CLA,     App.CMD_CLALLa,
            App.CMD_CLD,      App.CMD_CLKEYS,    App.CMD_CLLCD,    App.CMD_CLMENU,     App.CMD_CLP,     App.CMD_CLRG,
            App.CMD_CLST,     App.CMD_CLV,       App.CMD_CLX,      App.CMD_CLSIGMA,    App.CMD_COMB,    App.CMD_COMPLEX,
            App.CMD_CORR,     App.CMD_COS,       App.CMD_COSH,     App.CMD_CPXRES,     App.CMD_CPX_T,   App.CMD_CROSS,
            App.CMD_CUSTOM,   App.CMD_DECM,      App.CMD_DEG,      App.CMD_DEL,        App.CMD_DELAY,   App.CMD_DELR,
            App.CMD_DET,      App.CMD_DIM,       App.CMD_DIM_T,    App.CMD_DOT,        App.CMD_DSE,     App.CMD_EDIT,
            App.CMD_EDITN,    App.CMD_END,       App.CMD_ENG,      App.CMD_ENTER,      App.CMD_EXITALL, App.CMD_EXPF,
            App.CMD_E_POW_X,  App.CMD_E_POW_X_1, App.CMD_FC_T,     App.CMD_FCC_T,      App.CMD_FCSTX,   App.CMD_FCSTY,
            App.CMD_FIX,      App.CMD_FNRM,      App.CMD_FP,       App.CMD_FS_T,       App.CMD_FSC_T,   App.CMD_GAMMA,
            App.CMD_GETKEY,   App.CMD_GETM,      App.CMD_GRAD,     App.CMD_GROW,       App.CMD_GTO,     App.CMD_HEXM,
            App.CMD_HMSADD,   App.CMD_HMSSUB,    App.CMD_I_ADD,    App.CMD_I_SUB,      App.CMD_INDEX,   App.CMD_INPUT,
            App.CMD_INSR,     App.CMD_INTEG,     App.CMD_INVRT,    App.CMD_IP,         App.CMD_ISG,     App.CMD_J_ADD,
            App.CMD_J_SUB,    App.CMD_KEYASN,    App.CMD_KEYG,     App.CMD_KEYX,       App.CMD_LASTX,   App.CMD_LBL,
            App.CMD_LCLBL,    App.CMD_LINF,      App.CMD_LINSIGMA, App.CMD_LIST,       App.CMD_LN,      App.CMD_LN_1_X,
            App.CMD_LOG,      App.CMD_LOGF,      App.CMD_MAN,      App.CMD_MAT_T,      App.CMD_MEAN,    App.CMD_MENU,
            App.CMD_MOD,      App.CMD_MVAR,      App.CMD_FACT,     App.CMD_NEWMAT,     App.CMD_NORM,    App.CMD_NOT,
            App.CMD_OCTM,     App.CMD_OFF,       App.CMD_OLD,      App.CMD_ON,         App.CMD_OR,      App.CMD_PERM,
            App.CMD_PGMINT,   App.CMD_PGMSLV,    App.CMD_PI,       App.CMD_PIXEL,      App.CMD_POLAR,   App.CMD_POSA,
            App.CMD_PRA,      App.CMD_PRLCD,     App.CMD_POFF,     App.CMD_PROMPT,     App.CMD_PON,     App.CMD_PRP,
            App.CMD_PRSTK,    App.CMD_PRUSR,     App.CMD_PRV,      App.CMD_PRX,        App.CMD_PRSIGMA, App.CMD_PSE,
            App.CMD_PUTM,     App.CMD_PWRF,      App.CMD_QUIET,    App.CMD_RAD,        App.CMD_RAN,     App.CMD_RCL,
            App.CMD_RCL_ADD,  App.CMD_RCL_SUB,   App.CMD_RCL_MUL,  App.CMD_RCL_DIV,    App.CMD_RCLEL,   App.CMD_RCLIJ,
            App.CMD_RDXCOMMA, App.CMD_RDXDOT,    App.CMD_REALRES,  App.CMD_REAL_T,     App.CMD_RECT,    App.CMD_RND,
            App.CMD_RNRM,     App.CMD_ROTXY,     App.CMD_RSUM,     App.CMD_RTN,        App.CMD_SWAP_R,  App.CMD_RUP,
            App.CMD_RDN,      App.CMD_SCI,       App.CMD_SDEV,     App.CMD_SEED,       App.CMD_SF,      App.CMD_SIGN,
            App.CMD_SIN,      App.CMD_SINH,      App.CMD_SIZE,     App.CMD_SLOPE,      App.CMD_SOLVE,   App.CMD_SQRT,
            App.CMD_SST,      App.CMD_STO,       App.CMD_STO_ADD,  App.CMD_STO_SUB,    App.CMD_STO_MUL, App.CMD_STO_DIV,
            App.CMD_STOEL,    App.CMD_STOIJ,     App.CMD_STOP,     App.CMD_STR_T,      App.CMD_SUM,     App.CMD_TAN,
            App.CMD_TANH,     App.CMD_TONE,      App.CMD_TRACE,    App.CMD_TRANS,      App.CMD_UVEC,    App.CMD_VARMENU,
            App.CMD_VIEW,     App.CMD_WMEAN,     App.CMD_WRAP,     App.CMD_X_SWAP,     App.CMD_SWAP,    App.CMD_X_LT_0,
            App.CMD_X_LT_Y,   App.CMD_X_LE_0,    App.CMD_X_LE_Y,   App.CMD_X_EQ_0,     App.CMD_X_EQ_Y,  App.CMD_X_NE_0,
            App.CMD_X_NE_Y,   App.CMD_X_GT_0,    App.CMD_X_GT_Y,   App.CMD_X_GE_0,     App.CMD_X_GE_Y,  App.CMD_XEQ,
            App.CMD_XOR,      App.CMD_XTOA,      App.CMD_SQUARE,   App.CMD_YINT,       App.CMD_Y_POW_X, App.CMD_INV,
            App.CMD_10_POW_X, App.CMD_ADD,       App.CMD_SUB,      App.CMD_MUL,        App.CMD_DIV,     App.CMD_CHS,
            App.CMD_SIGMAADD, App.CMD_SIGMASUB,  App.CMD_SIGMAREG, App.CMD_SIGMAREG_T, App.CMD_TO_DEC,  App.CMD_TO_DEG,
            App.CMD_TO_HMS,   App.CMD_TO_HR,     App.CMD_TO_OCT,   App.CMD_TO_POL,     App.CMD_TO_RAD,  App.CMD_TO_REC,
            App.CMD_LEFT,     App.CMD_UP,        App.CMD_DOWN,     App.CMD_RIGHT,      App.CMD_PERCENT, App.CMD_PERCENT_CH
        };

        // SLR - Flags are manually set here in static initialization to match program implementations and values set in init_state()
        public static extension_struct[] extensions = new extension_struct[]
        {
            new extension_struct { first_cmd=CMD_OPENF,   last_cmd=CMD_DELP,    enable_flag=false },     // core_settings.enable_ext_copan
            new extension_struct { first_cmd=CMD_DROP,    last_cmd=CMD_DROP,    enable_flag=false },     // core_settings.enable_ext_bigstack
            new extension_struct { first_cmd=CMD_ACCEL,   last_cmd=CMD_ACCEL,   enable_flag=false },     // core_settings.enable_ext_accel
            new extension_struct { first_cmd=CMD_LOCAT,   last_cmd=CMD_LOCAT,   enable_flag=false },     // core_settings.enable_ext_locat
            new extension_struct { first_cmd=CMD_HEADING, last_cmd=CMD_HEADING, enable_flag=false },     // core_settings.enable_ext_heading
            new extension_struct { first_cmd=CMD_ADATE,   last_cmd=CMD_SWPT,    enable_flag=true },     // core_settings.enable_ext_time
            new extension_struct { first_cmd=CMD_NULL,    last_cmd=CMD_NULL,    enable_flag=false }
        };



        public static byte[,] bigchars = new byte[130, 5]
        {
	        { 0x08, 0x08, 0x2a, 0x08, 0x08 },
	        { 0x22, 0x14, 0x08, 0x14, 0x22 },
	        { 0x10, 0x20, 0x7f, 0x01, 0x01 },
	        { 0x20, 0x40, 0x3e, 0x01, 0x02 },
	        { 0x55, 0x2a, 0x55, 0x2a, 0x55 },
	        { 0x41, 0x63, 0x55, 0x49, 0x63 },
	        { 0x7f, 0x7f, 0x3e, 0x1c, 0x08 },
	        { 0x04, 0x7c, 0x04, 0x7c, 0x04 },
	        { 0x30, 0x48, 0x45, 0x40, 0x20 },
	        { 0x50, 0x58, 0x54, 0x52, 0x51 },
	        { 0x0f, 0x08, 0x00, 0x78, 0x28 },
	        { 0x51, 0x52, 0x54, 0x58, 0x50 },
	        { 0x14, 0x34, 0x1c, 0x16, 0x14 },
	        { 0x20, 0x70, 0xa8, 0x20, 0x3f },
	        { 0x10, 0x20, 0x7f, 0x20, 0x10 },
	        { 0x08, 0x08, 0x2a, 0x1c, 0x08 },
	        { 0x08, 0x1c, 0x2a, 0x08, 0x08 },
	        { 0x7e, 0x20, 0x20, 0x1e, 0x20 },
	        { 0x48, 0x7e, 0x49, 0x41, 0x02 },
	        { 0x00, 0x0e, 0x0a, 0x0e, 0x00 },
	        { 0x78, 0x16, 0x15, 0x16, 0x78 },
	        { 0x7c, 0x0a, 0x11, 0x22, 0x7d },
	        { 0x7c, 0x13, 0x12, 0x13, 0x7c },
	        { 0x60, 0x50, 0x58, 0x64, 0x42 },
	        { 0x3e, 0x2a, 0x2a, 0x22, 0x00 },
	        { 0x7e, 0x09, 0x7f, 0x49, 0x41 },
	        { 0x60, 0x00, 0x60, 0x00, 0x60 },
	        { 0x1f, 0x15, 0x71, 0x50, 0x50 },
	        { 0x3c, 0x43, 0x42, 0x43, 0x3c },
	        { 0x3c, 0x41, 0x40, 0x41, 0x3c },
	        { 0x55, 0x2a, 0x55, 0x2a, 0x55 },
	        { 0x3c, 0x3c, 0x3c, 0x3c, 0x3c },
	        { 0x00, 0x00, 0x00, 0x00, 0x00 },
	        { 0x00, 0x00, 0x5f, 0x00, 0x00 },
	        { 0x00, 0x07, 0x00, 0x07, 0x00 },
	        { 0x14, 0x7f, 0x14, 0x7f, 0x14 },
	        { 0x24, 0x2a, 0x7f, 0x2a, 0x12 },
	        { 0x23, 0x13, 0x08, 0x64, 0x62 },
	        { 0x36, 0x49, 0x56, 0x20, 0x50 },
	        { 0x00, 0x00, 0x07, 0x00, 0x00 },
	        { 0x00, 0x1c, 0x22, 0x41, 0x00 },
	        { 0x00, 0x41, 0x22, 0x1c, 0x00 },
	        { 0x08, 0x2a, 0x1c, 0x2a, 0x08 },
	        { 0x08, 0x08, 0x3e, 0x08, 0x08 },
	        { 0x00, 0xb0, 0x70, 0x00, 0x00 },
	        { 0x08, 0x08, 0x08, 0x08, 0x00 },
	        { 0x00, 0x60, 0x60, 0x00, 0x00 },
	        { 0x20, 0x10, 0x08, 0x04, 0x02 },
	        { 0x3e, 0x51, 0x49, 0x45, 0x3e },
	        { 0x00, 0x42, 0x7f, 0x40, 0x00 },
	        { 0x62, 0x51, 0x49, 0x49, 0x46 },
	        { 0x22, 0x49, 0x49, 0x49, 0x36 },
	        { 0x18, 0x14, 0x12, 0x7f, 0x10 },
	        { 0x27, 0x45, 0x45, 0x45, 0x39 },
	        { 0x3c, 0x4a, 0x49, 0x49, 0x30 },
	        { 0x01, 0x71, 0x09, 0x05, 0x03 },
	        { 0x36, 0x49, 0x49, 0x49, 0x36 },
	        { 0x06, 0x49, 0x49, 0x29, 0x1e },
	        { 0x00, 0x36, 0x36, 0x00, 0x00 },
	        { 0x00, 0xb6, 0x76, 0x00, 0x00 },
	        { 0x08, 0x14, 0x22, 0x41, 0x00 },
	        { 0x14, 0x14, 0x14, 0x14, 0x14 },
	        { 0x41, 0x22, 0x14, 0x08, 0x00 },
	        { 0x02, 0x01, 0x51, 0x09, 0x06 },
	        { 0x3e, 0x41, 0x5d, 0x55, 0x5e },
	        { 0x7e, 0x09, 0x09, 0x09, 0x7e },
	        { 0x7f, 0x49, 0x49, 0x49, 0x36 },
	        { 0x3e, 0x41, 0x41, 0x41, 0x22 },
	        { 0x7f, 0x41, 0x41, 0x22, 0x1c },
	        { 0x7f, 0x49, 0x49, 0x49, 0x41 },
	        { 0x7f, 0x09, 0x09, 0x09, 0x01 },
	        { 0x3e, 0x41, 0x41, 0x51, 0x72 },
	        { 0x7f, 0x08, 0x08, 0x08, 0x7f },
	        { 0x00, 0x41, 0x7f, 0x41, 0x00 },
	        { 0x30, 0x40, 0x40, 0x40, 0x3f },
	        { 0x7f, 0x08, 0x14, 0x22, 0x41 },
	        { 0x7f, 0x40, 0x40, 0x40, 0x40 },
	        { 0x7f, 0x02, 0x0c, 0x02, 0x7f },
	        { 0x7f, 0x04, 0x08, 0x10, 0x7f },
	        { 0x3e, 0x41, 0x41, 0x41, 0x3e },
	        { 0x7f, 0x09, 0x09, 0x09, 0x06 },
	        { 0x3e, 0x41, 0x51, 0x21, 0x5e },
	        { 0x7f, 0x09, 0x19, 0x29, 0x46 },
	        { 0x26, 0x49, 0x49, 0x49, 0x32 },
	        { 0x01, 0x01, 0x7f, 0x01, 0x01 },
	        { 0x3f, 0x40, 0x40, 0x40, 0x3f },
	        { 0x07, 0x18, 0x60, 0x18, 0x07 },
	        { 0x7f, 0x20, 0x18, 0x20, 0x7f },
	        { 0x63, 0x14, 0x08, 0x14, 0x63 },
	        { 0x03, 0x04, 0x78, 0x04, 0x03 },
	        { 0x61, 0x51, 0x49, 0x45, 0x43 },
	        { 0x00, 0x7f, 0x41, 0x41, 0x00 },
	        { 0x02, 0x04, 0x08, 0x10, 0x20 },
	        { 0x00, 0x41, 0x41, 0x7f, 0x00 },
	        { 0x04, 0x02, 0x7f, 0x02, 0x04 },
	        { 0x80, 0x80, 0x80, 0x80, 0x80 },
	        { 0x00, 0x03, 0x04, 0x00, 0x00 },
	        { 0x20, 0x54, 0x54, 0x54, 0x78 },
	        { 0x7f, 0x44, 0x44, 0x44, 0x38 },
	        { 0x38, 0x44, 0x44, 0x44, 0x44 },
	        { 0x38, 0x44, 0x44, 0x44, 0x7f },
	        { 0x38, 0x54, 0x54, 0x54, 0x58 },
	        { 0x00, 0x08, 0x7e, 0x09, 0x02 },
	        { 0x18, 0xa4, 0xa4, 0xa4, 0x78 },
	        { 0x7f, 0x04, 0x04, 0x04, 0x78 },
	        { 0x00, 0x44, 0x7d, 0x40, 0x00 },
	        { 0x00, 0x40, 0x80, 0x84, 0x7d },
	        { 0x7f, 0x10, 0x28, 0x44, 0x00 },
	        { 0x00, 0x41, 0x7f, 0x40, 0x00 },
	        { 0x7c, 0x04, 0x38, 0x04, 0x7c },
	        { 0x7c, 0x04, 0x04, 0x04, 0x78 },
	        { 0x38, 0x44, 0x44, 0x44, 0x38 },
	        { 0xfc, 0x24, 0x24, 0x24, 0x18 },
	        { 0x18, 0x24, 0x24, 0x24, 0xfc },
	        { 0x7c, 0x08, 0x04, 0x04, 0x04 },
	        { 0x48, 0x54, 0x54, 0x54, 0x24 },
	        { 0x00, 0x04, 0x3f, 0x44, 0x20 },
	        { 0x3c, 0x40, 0x40, 0x40, 0x7c },
	        { 0x1c, 0x20, 0x40, 0x20, 0x1c },
	        { 0x3c, 0x40, 0x30, 0x40, 0x3c },
	        { 0x44, 0x28, 0x10, 0x28, 0x44 },
	        { 0x1c, 0xa0, 0xa0, 0xa0, 0x7c },
	        { 0x44, 0x64, 0x54, 0x4c, 0x44 },
	        { 0x08, 0x36, 0x41, 0x41, 0x00 },
	        { 0x00, 0x00, 0x7f, 0x00, 0x00 },
	        { 0x00, 0x41, 0x41, 0x36, 0x08 },
	        { 0x08, 0x04, 0x08, 0x10, 0x08 },
	        { 0x7f, 0x08, 0x08, 0x08, 0x08 },
	        { 0x28, 0x00, 0x00, 0x00, 0x00 },
	        { 0x04, 0x08, 0x70, 0x08, 0x04 }
        };

        public static byte[] smallchars = new byte[329]
        {
	        0x00, 0x00, 0x00,
	        0x5c,
	        0x06, 0x00, 0x06,
	        0x28, 0x7c, 0x28, 0x7c, 0x28,
	        0x08, 0x54, 0x7c, 0x54, 0x20,
	        0x24, 0x10, 0x48,
	        0x30, 0x4c, 0x50, 0x20, 0x50,
	        0x08, 0x04,
	        0x38, 0x44,
	        0x44, 0x38,
	        0x54, 0x38, 0x54,
	        0x10, 0x38, 0x10,
	        0x40, 0x20,
	        0x10, 0x10, 0x10,
	        0x40,
	        0x60, 0x10, 0x0c,
	        0x38, 0x44, 0x38,
	        0x48, 0x7c, 0x40,
	        0x74, 0x54, 0x5c,
	        0x44, 0x54, 0x7c,
	        0x1c, 0x10, 0x7c,
	        0x5c, 0x54, 0x74,
	        0x7c, 0x54, 0x74,
	        0x64, 0x14, 0x0c,
	        0x7c, 0x54, 0x7c,
	        0x5c, 0x54, 0x7c,
	        0x28,
	        0x40, 0x28,
	        0x10, 0x28, 0x44,
	        0x28, 0x28, 0x28,
	        0x44, 0x28, 0x10,
	        0x08, 0x04, 0x54, 0x08,
	        0x38, 0x44, 0x54, 0x58,
	        0x78, 0x14, 0x78,
	        0x7c, 0x54, 0x28,
	        0x38, 0x44, 0x44,
	        0x7c, 0x44, 0x38,
	        0x7c, 0x54, 0x44,
	        0x7c, 0x14, 0x04,
	        0x7c, 0x44, 0x54, 0x74,
	        0x7c, 0x10, 0x7c,
	        0x7c,
	        0x60, 0x40, 0x7c,
	        0x7c, 0x10, 0x28, 0x44,
	        0x7c, 0x40, 0x40,
	        0x7c, 0x08, 0x10, 0x08, 0x7c,
	        0x7c, 0x18, 0x30, 0x7c,
	        0x7c, 0x44, 0x7c,
	        0x7c, 0x14, 0x1c,
	        0x38, 0x44, 0x24, 0x58,
	        0x7c, 0x14, 0x6c,
	        0x48, 0x54, 0x24,
	        0x04, 0x7c, 0x04,
	        0x7c, 0x40, 0x7c,
	        0x1c, 0x60, 0x1c,
	        0x7c, 0x20, 0x10, 0x20, 0x7c,
	        0x6c, 0x10, 0x6c,
	        0x0c, 0x70, 0x0c,
	        0x64, 0x54, 0x4c,
	        0x7c, 0x44,
	        0x0c, 0x10, 0x60,
	        0x44, 0x7c,
	        0x10, 0x08, 0x7c, 0x08, 0x10,
	        0x40, 0x40, 0x40,
	        0x04, 0x08,
	        0x10, 0x6c, 0x44,
	        0x6c,
	        0x44, 0x6c, 0x10,
	        0x10, 0x08, 0x10, 0x20, 0x10,
	        0x54, 0x28, 0x54, 0x28, 0x54,
	        0x10, 0x54, 0x10,
	        0x28, 0x10, 0x28,
	        0x10, 0x20, 0x7c, 0x04, 0x04, 0x04,
	        0x20, 0x40, 0x38, 0x04, 0x08,
	        0x44, 0x6c, 0x54, 0x44,
	        0x08, 0x78, 0x08, 0x78, 0x08,
	        0x50, 0x58, 0x54,
	        0x3c, 0x20, 0x00, 0x78, 0x28,
	        0x54, 0x58, 0x50,
	        0x28, 0x68, 0x38, 0x2c, 0x28,
	        0x10, 0x20, 0x7c, 0x20, 0x10,
	        0x10, 0x10, 0x54, 0x38, 0x10,
	        0x10, 0x38, 0x54, 0x10, 0x10,
	        0x78, 0x20, 0x38, 0x20,
	        0x1c, 0x14, 0x1c,
	        0x1c, 0x08, 0x08,
	        0x60, 0x00, 0x60, 0x00, 0x60,
	        0x60, 0x50, 0x58, 0x64, 0x40,
	        0x74, 0x28, 0x28, 0x74,
	        0x34, 0x48, 0x48, 0x34,
	        0x34, 0x40, 0x40, 0x34,
	        0x7c, 0x12, 0x24, 0x7a,
	        0x50, 0x78, 0x54, 0x04,
	        0x20, 0x54, 0x40, 0x20,
	        0x78, 0x14, 0x7c, 0x54,
	        0x38, 0x38, 0x38,
	        0x70, 0x2c, 0x70,
	        0x58, 0x24, 0x54, 0x48
        };

        public static short[] smallchars_offset = new short[99]
        {
	          0,
	          3,
	          4,
	          7,
	         12,
	         17,
	         20,
	         25,
	         27,
	         29,
	         31,
	         34,
	         37,
	         39,
	         42,
	         43,
	         46,
	         49,
	         52,
	         55,
	         58,
	         61,
	         64,
	         67,
	         70,
	         73,
	         76,
	         77,
	         79,
	         82,
	         85,
	         88,
	         92,
	         96,
	         99,
	        102,
	        105,
	        108,
	        111,
	        114,
	        118,
	        121,
	        122,
	        125,
	        129,
	        132,
	        137,
	        141,
	        144,
	        147,
	        151,
	        154,
	        157,
	        160,
	        163,
	        166,
	        171,
	        174,
	        177,
	        180,
	        182,
	        185,
	        187,
	        192,
	        195,
	        197,
	        200,
	        201,
	        204,
	        209,
	        214,
	        217,
	        220,
	        226,
	        231,
	        235,
	        240,
	        243,
	        248,
	        251,
	        256,
	        261,
	        266,
	        271,
	        275,
	        278,
	        281,
	        286,
	        291,
	        295,
	        299,
	        303,
	        307,
	        311,
	        315,
	        319,
	        322,
	        325,
	        329
        };

        public static byte[] smallchars_map = new byte[128]
        {
	        /*   0 */  70,
	        /*   1 */  71,
	        /*   2 */  72,
	        /*   3 */  73,
	        /*   4 */  69,
	        /*   5 */  74,
	        /*   6 */  97,
	        /*   7 */  75,
	        /*   8 */  93,
	        /*   9 */  76,
	        /*  10 */  77,
	        /*  11 */  78,
	        /*  12 */  79,
	        /*  13 */   0,
	        /*  14 */  80,
	        /*  15 */  81,
	        /*  16 */  82,
	        /*  17 */  83,
	        /*  18 */  92,
	        /*  19 */  84,
	        /*  20 */  96,
	        /*  21 */  91,
	        /*  22 */  88,
	        /*  23 */  87,
	        /*  24 */  37,
	        /*  25 */  94,
	        /*  26 */  86,
	        /*  27 */  69,
	        /*  28 */  89,
	        /*  29 */  90,
	        /*  30 */  69,
	        /*  31 */  95,
	        /*  32 */   0,
	        /*  33 */   1,
	        /*  34 */   2,
	        /*  35 */   3,
	        /*  36 */   4,
	        /*  37 */   5,
	        /*  38 */   6,
	        /*  39 */   7,
	        /*  40 */   8,
	        /*  41 */   9,
	        /*  42 */  10,
	        /*  43 */  11,
	        /*  44 */  12,
	        /*  45 */  13,
	        /*  46 */  14,
	        /*  47 */  15,
	        /*  48 */  16,
	        /*  49 */  17,
	        /*  50 */  18,
	        /*  51 */  19,
	        /*  52 */  20,
	        /*  53 */  21,
	        /*  54 */  22,
	        /*  55 */  23,
	        /*  56 */  24,
	        /*  57 */  25,
	        /*  58 */  26,
	        /*  59 */  27,
	        /*  60 */  28,
	        /*  61 */  29,
	        /*  62 */  30,
	        /*  63 */  31,
	        /*  64 */  32,
	        /*  65 */  33,
	        /*  66 */  34,
	        /*  67 */  35,
	        /*  68 */  36,
	        /*  69 */  37,
	        /*  70 */  38,
	        /*  71 */  39,
	        /*  72 */  40,
	        /*  73 */  41,
	        /*  74 */  42,
	        /*  75 */  43,
	        /*  76 */  44,
	        /*  77 */  45,
	        /*  78 */  46,
	        /*  79 */  47,
	        /*  80 */  48,
	        /*  81 */  49,
	        /*  82 */  50,
	        /*  83 */  51,
	        /*  84 */  52,
	        /*  85 */  53,
	        /*  86 */  54,
	        /*  87 */  55,
	        /*  88 */  56,
	        /*  89 */  57,
	        /*  90 */  58,
	        /*  91 */  59,
	        /*  92 */  60,
	        /*  93 */  61,
	        /*  94 */  62,
	        /*  95 */  63,
	        /*  96 */  64,
	        /*  97 */  33,
	        /*  98 */  34,
	        /*  99 */  35,
	        /* 100 */  36,
	        /* 101 */  37,
	        /* 102 */  38,
	        /* 103 */  39,
	        /* 104 */  40,
	        /* 105 */  41,
	        /* 106 */  42,
	        /* 107 */  43,
	        /* 108 */  44,
	        /* 109 */  45,
	        /* 110 */  46,
	        /* 111 */  47,
	        /* 112 */  48,
	        /* 113 */  49,
	        /* 114 */  50,
	        /* 115 */  51,
	        /* 116 */  52,
	        /* 117 */  53,
	        /* 118 */  54,
	        /* 119 */  55,
	        /* 120 */  56,
	        /* 121 */  57,
	        /* 122 */  58,
	        /* 123 */  65,
	        /* 124 */  66,
	        /* 125 */  67,
	        /* 126 */  68,
	        /* 127 */  85
        };
    }
}
