﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;

namespace WP7_42s
{
    public partial class App
    {
        public static App.vartype new_real(Phloat value)
        {
            App.vartype_real r = new App.vartype_real();
            r.type = App.TYPE.REAL;
            r.x = value;

            return r;
        }

        public static App.vartype new_complex(Phloat re, Phloat im)
        {
            App.vartype_complex c = new App.vartype_complex();
            c.type = App.TYPE.COMPLEX;
            
            c.re = re;
            c.im = im;

            return c;
        }

        public static App.vartype new_string(char[] text, int length)
        {
            App.vartype_string s = new App.vartype_string();

            s.type = App.TYPE.STRING;
            s.length = length > 6 ? 6 : length;

            for (int i = 0; i < s.length; i++)
		        s.text[i] = text[i];

            return s;
        }

        public static App.vartype new_realmatrix(int rows, int columns)
        {
            App.vartype_realmatrix rm = new App.vartype_realmatrix();

            rm.type = App.TYPE.REALMATRIX;
            rm.rows = rows;
            rm.columns = columns;
            int sz = rows * columns;

            rm.array = new App.realmatrix_data();
            rm.array.data = new Phloat[sz];
            rm.array.is_string = new char[sz];

            for (int i = 0; i < sz; i++)
                rm.array.data[i] = 0;

            for (int i = 0; i < sz; i++)
                rm.array.is_string[i] = '\0';

            rm.array.refcount = 1;

            return rm;
        }

        public static App.vartype new_complexmatrix(int rows, int columns)
        {
            App.vartype_complexmatrix cm = new App.vartype_complexmatrix();

            cm.type = App.TYPE.COMPLEXMATRIX;
            cm.rows = rows;
            cm.columns = columns;
            int sz = rows * columns * 2;

            cm.array = new App.complexmatrix_data();
            cm.array.data = new Phloat[sz];

            for (int i = 0; i < sz; i++)
                cm.array.data[i] = 0;

            cm.array.refcount = 1;

            return cm;
        }

        public static App.vartype new_matrix_alias(App.vartype m)
        {
            if (m.type == App.TYPE.REALMATRIX)
            {
                App.vartype_realmatrix rm1 = (App.vartype_realmatrix) m;
                App.vartype_realmatrix rm2 = new App.vartype_realmatrix();

                rm2.type = rm1.type;
                rm2.rows = rm1.rows;
                rm2.columns = rm1.columns;
                rm2.array = rm1.array;

                rm2.array.refcount++;

                return rm2;
            }
            else if (m.type == App.TYPE.COMPLEXMATRIX)
            {
                App.vartype_complexmatrix cm1 = (App.vartype_complexmatrix) m;
                App.vartype_complexmatrix cm2 = new App.vartype_complexmatrix();

                cm2.type = cm1.type;
                cm2.rows = cm1.rows;
                cm2.columns = cm1.columns;
                cm2.array = cm1.array;

                cm2.array.refcount++;

                return cm2;
            }
            else
                return null;
        }

        public static void free_vartype(ref App.vartype v)
        {
            if (v == null)
                return;

            switch (v.type)
            {
                case App.TYPE.REALMATRIX:
                    ((App.vartype_realmatrix) v).array.refcount--;
                    break;

                case App.TYPE.COMPLEXMATRIX:
                    ((App.vartype_complexmatrix) v).array.refcount--;
                    break;
            }

            v = null;
        }

        public static App.vartype dup_vartype(App.vartype v)
        {
            if (v == null)
		        return null;

            switch (v.type)
	        {
	            case App.TYPE.REAL:
			        return new_real(((App.vartype_real) v).x);

	            case App.TYPE.COMPLEX:
                    return new_complex(((App.vartype_complex)v).re, ((App.vartype_complex)v).im);

	            case App.TYPE.REALMATRIX:
			        App.vartype_realmatrix rm = (App.vartype_realmatrix) v;
			        App.vartype_realmatrix rm2 = new App.vartype_realmatrix();

			        rm2.type = App.TYPE.REALMATRIX;
			        rm2.rows = rm.rows;
			        rm2.columns = rm.columns;
			        rm2.array = rm.array;
			        rm.array.refcount++;
			
			        return rm2;

	            case App.TYPE.COMPLEXMATRIX:
			        App.vartype_complexmatrix cm = (App.vartype_complexmatrix) v;
			        App.vartype_complexmatrix cm2 = new App.vartype_complexmatrix();

			        cm2.type = App.TYPE.COMPLEXMATRIX;
			        cm2.rows = cm.rows;
			        cm2.columns = cm.columns;
			        cm2.array = cm.array;
			        cm.array.refcount++;
			
			        return cm2;

	            case App.TYPE.STRING:
                    return new_string(((App.vartype_string)v).text, ((App.vartype_string)v).length);

	            default:
	                return null;
            }
        }

        public static int disentangle(App.vartype v)
        {
            switch (v.type)
            {
                case App.TYPE.REALMATRIX:
                    App.vartype_realmatrix rm = (App.vartype_realmatrix) v;

                    if (rm.array.refcount == 1)
                        return 1;
                    else
                    {
                        App.realmatrix_data md = new App.realmatrix_data();

                        int sz = rm.rows * rm.columns;

                        md.data = new Phloat[sz];
                        md.is_string = new char[sz];

                        for (int i = 0; i < sz; i++)
                            md.data[i] = rm.array.data[i];

                        for (int i = 0; i < sz; i++)
                            md.is_string[i] = rm.array.is_string[i];

                        md.refcount = 1;
                        rm.array.refcount--;
                        rm.array = md;

                        return 1;
                    }

                case App.TYPE.COMPLEXMATRIX:
                    {
                        App.vartype_complexmatrix cm = (App.vartype_complexmatrix) v;

                        if (cm.array.refcount == 1)
                            return 1;
                        else
                        {
                            App.complexmatrix_data md = new App.complexmatrix_data();

                            int sz = cm.rows * cm.columns * 2;

                            md.data = new Phloat[sz];

                            for (int i = 0; i < sz; i++)
                                md.data[i] = cm.array.data[i];

                            md.refcount = 1;
                            cm.array.refcount--;
                            cm.array = md;

                            return 1;
                        }
                    }

                case App.TYPE.REAL:
                case App.TYPE.COMPLEX:
                case App.TYPE.STRING:
                default:
                    return 1;
            }
        }

        public static int lookup_var(char[] name, int namelength)
        {
            for (int i = 0; i < App.vars_count; i++)
	        {
		        if (App.vars[i].length == namelength)
		        {
			        for (int j = 0; j < namelength; j++)
				        if (App.vars[i].name[j] != name[j])
					        goto nomatch;
		
			        return i;
		        }
	        nomatch:;
            }

	        return -1;
        }

        public static App.vartype recall_var(char[] name, int namelength)
        {
            int varindex = lookup_var(name, namelength);

            if (varindex == -1)
		        return null;
            else
		        return App.vars[varindex].value;
        }

        public static void store_var(char[] name, int namelength, App.vartype value)
        {
            int varindex = lookup_var(name, namelength);

            if (varindex == -1)
	        {
		        if (App.vars_count == App.vars_capacity)
		        {
		            App.vars_capacity += 25;
                    Array.Resize(ref App.vars, App.vars_capacity);

                    for (int i = vars.Length - 25; i < vars.Length; i++)
                        vars[i] = new var_struct();
		        }

		        varindex = App.vars_count++;
		        App.vars[varindex].length = (byte) namelength;

		        for (int i = 0; i < namelength; i++)
                    App.vars[varindex].name[i] = name[i];
                //App.vars[varindex].name = name;                 // Don't need to copy chars, just assign to vars struct
            }
	        else
	        {
		        if ((App.matedit_mode != 0) && (App.string_equals(name, namelength, App.matedit_name, App.matedit_length) != 0))
		        {
			        // We're replacing the indexed matrix; need to reset I and J

			        // TO DO: proper handling of modes 2 and 3 (edit & editn); this code
			        // really only takes care of the mode 1 case, because it does not
			        // deal with the contents of X, nor with leaving the editor if the
			        // replacement value is not a matrix (or is that forbidden? Maybe
			        // that should cause a "Restricted Operation" error... And that
			        // means I'll have to fix store_var() so that it returns an error
			        // code -- which it should anyway; right now it does not check for
			        // memory allocation failure when growing the variables array, and
			        // that is evil).
			        //
			        
                    if (value.type == App.TYPE.REALMATRIX || value.type == App.TYPE.COMPLEXMATRIX)
				        App.matedit_i = App.matedit_j = 0;
			        else
				        App.matedit_mode = 0;
		        }

		        free_vartype(ref App.vars[varindex].value);
            }

            App.vars[varindex].value = value;

            App.update_catalog();
        }

        public static int purge_var(char[] name, int namelength)
        {
            int varindex = lookup_var(name, namelength);

            if (varindex == -1)
		        return 0;

            free_vartype(ref App.vars[varindex].value);

            for (int i = varindex; i < App.vars_count - 1; i++)
            {
                App.vars[i] = App.vars[i + 1];
                App.vars[i + 1] = new var_struct();
            }

            App.vars_count--;
            App.update_catalog();
    
	        return 1;
        }

        public static void purge_all_vars()
        {
            for (int i = 0; i < App.vars_count; i++)
                free_vartype(ref App.vars[i].value);

            App.vars_count = 0;
        }

        public static int vars_exist(int real, int cpx, int matrix)
        {
            for (int i = 0; i < App.vars_count; i++)
            {
                switch (App.vars[i].value.type)
                {
                    case App.TYPE.REAL:
                    case App.TYPE.STRING:
                        if (real != 0)
                            return 1;
                        else
                            break;

                    case App.TYPE.COMPLEX:
                        if (cpx != 0)
                            return 1;
                        else
                            break;

                    case App.TYPE.REALMATRIX:
                    case App.TYPE.COMPLEXMATRIX:
                        if (matrix != 0)
                            return 1;
                        else
                            break;
                }
            }

            return 0;
        }

        public static int contains_no_strings(App.vartype_realmatrix rm)
        {
            int size = rm.rows * rm.columns;

	        for (int i = 0; i < size; i++)
		        if (rm.array.is_string[i] != '\0')
		            return 0;

            return 1;
        }

        public static App.ERR matrix_copy(ref App.vartype dst, App.vartype src)
        {
            int size, i;

	        if (src.type == App.TYPE.REALMATRIX)
	        {
		        App.vartype_realmatrix s = (App.vartype_realmatrix) src;

		        if (dst.type == App.TYPE.REALMATRIX)
		        {
			        App.vartype_realmatrix d = (App.vartype_realmatrix) dst;

			        if ((s.rows != d.rows) || (s.columns != d.columns))
				        return App.ERR.DIMENSION_ERROR;

			        size = s.rows * s.columns;
			
			        for (i = 0; i < size; i++)
			        {
				        d.array.is_string[i] = s.array.is_string[i];
				        d.array.data[i] = s.array.data[i];
			        }

			        return App.ERR.NONE;
		        }
		        else if (dst.type == App.TYPE.COMPLEXMATRIX)
		        {
			        App.vartype_complexmatrix d = (App.vartype_complexmatrix) dst;

			        if ((s.rows != d.rows) || (s.columns != d.columns))
				        return App.ERR.DIMENSION_ERROR;
			
			        if (contains_no_strings(s) == 0)
				        return App.ERR.ALPHA_DATA_IS_INVALID;

			        size = s.rows * s.columns;

			        for (i = 0; i < size; i++)
			        {
				        d.array.data[2 * i] = s.array.data[i];
				        d.array.data[2 * i + 1] = 0;
			        }

			        return App.ERR.NONE;
		        }
		        else
		            return App.ERR.INVALID_TYPE;
            }
	        else if ((src.type == App.TYPE.COMPLEXMATRIX) && (dst.type == App.TYPE.COMPLEXMATRIX))
	        {
		        App.vartype_complexmatrix s = (App.vartype_complexmatrix) src;
		        App.vartype_complexmatrix d = (App.vartype_complexmatrix) dst;

		        if ((s.rows != d.rows) || (s.columns != d.columns))
			        return App.ERR.DIMENSION_ERROR;

		        size = s.rows * s.columns * 2;

		        for (i = 0; i < size; i++)
			        d.array.data[i] = s.array.data[i];
		
		        return App.ERR.NONE;
            }
	        else
		        return App.ERR.INVALID_TYPE;
        }
    }
}
