﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

namespace WP7_42s
{
    public partial class App
    {
        public const string STATE_FILE = "state.bin";
        public const string PRINT_FILE = "42s_print.txt";
        public const string IMG_FILE = "lcd";

        public const off_methods OFF_METHOD = off_methods.display_off;

        public const int FREE42_MAGIC = 0x466b3432;
        public const int FREE42_VERSION = 16;

        public const int printPixelMul = 2;
        public const int print_entry_limit = 500;               // TODO: 1000 stutters on scrolling in emulator. Test in real device if larger values desired

        public static string[] weekdaynames = new string[7] { "SUN", "MON", "TUE", "WED", "THU", "FRI", "SAT" };

        public const int sizeof_Phloat = (App.P + 1) * sizeof(ushort);

        public const int NEG = 0x8000;
        public const int P = 7;
        public const int BASE = 10000;
        public const int EXPMASK = 0x1fff;
        public const int EXPMASKNEG = (NEG | EXPMASK);
        public const int POS_INF_EXP = 0x2000;
        public const int NEG_INF_EXP = 0xA000;
        public const int NAN_EXP = 0x4000;
        public const int BCD_CONST_PI = 0;
        public const int BCD_CONST_PI2 = 1;
        public const int BCD_CONST_PIBY2 = 2;
        public const int BCD_CONST_PIBY32 = 3;
        public const int BCD_CONST_SINTAB = 4;

        public const int BCD_CONST_COSTAB = (BCD_CONST_SINTAB+8);
        public const int BCD_CONST_LN2 = (BCD_CONST_COSTAB+8);
        public const int BCD_CONST_ONEOTWO = (BCD_CONST_LN2+1);
        public const int BCD_CONST_LN10 = (BCD_CONST_ONEOTWO+1);
        public const int BCD_CONST_LN2OLN10 = (BCD_CONST_ONEOTWO+2);
        public const int BCD_CONST_ATANLIM = (BCD_CONST_LN2OLN10+1);
        public const int BCD_CONST_PIBY32A = (BCD_CONST_ATANLIM+1);
        public const int BCD_CONST_PIBY32B = (BCD_CONST_PIBY32A+1);
        public const int BCD_CONST_HUNDREDTH = (BCD_CONST_PIBY32B+1);
        public const int BCD_CONST_HALF = (BCD_CONST_HUNDREDTH+1);
        public const int BCD_CONST_TWOHALF = (BCD_CONST_HALF+1);
        public const int BCD_CONST_TEN = (BCD_CONST_TWOHALF + 1);
        public const int BCD_CONST_SINPOLY = (BCD_CONST_TEN+1);
        public const int BCD_CONST_COSPOLY = (BCD_CONST_SINPOLY+7);
        public const int BCD_CONST_LOGPOLY = (BCD_CONST_COSPOLY+7);
        public const int BCD_CONST_LOGCK = (BCD_CONST_LOGPOLY+4);
        public const int BCD_CONST_EXP1 = (BCD_CONST_LOGCK+64);

        public const int BCD2_CONST_LANCZOS = 0;
        public const int K = 12;
        public const int GG = 12;
        public const double PI = 3.1415926535897932384626433;

        public const int CMD_NONE = -1;
        public const int CMD_CANCELLED = -2;
        public const int CMD_LINGER1 = -3;
        public const int CMD_LINGER2 = -4;

        public const int CMD_CLX = 0;
        public const int CMD_ENTER = 1;
        public const int CMD_SWAP = 2;
        public const int CMD_RDN = 3;
        public const int CMD_CHS = 4;
        public const int CMD_DIV = 5;
        public const int CMD_MUL = 6;
        public const int CMD_SUB = 7;
        public const int CMD_ADD = 8;
        public const int CMD_LASTX = 9;
        public const int CMD_SILENT_OFF = 10;
        public const int CMD_SILENT_ON = 11;
        public const int CMD_SIN = 12;
        public const int CMD_COS = 13;
        public const int CMD_TAN = 14;
        public const int CMD_ASIN = 15;
        public const int CMD_ACOS = 16;
        public const int CMD_ATAN = 17;
        public const int CMD_LOG = 18;
        public const int CMD_10_POW_X = 19;
        public const int CMD_LN = 20;
        public const int CMD_E_POW_X = 21;
        public const int CMD_SQRT = 22;
        public const int CMD_SQUARE = 23;
        public const int CMD_INV = 24;
        public const int CMD_Y_POW_X = 25;
        public const int CMD_PERCENT = 26;
        public const int CMD_PI = 27;
        public const int CMD_COMPLEX = 28;
        public const int CMD_STO = 29;
        public const int CMD_STO_DIV = 30;
        public const int CMD_STO_MUL = 31;
        public const int CMD_STO_SUB = 32;
        public const int CMD_STO_ADD = 33;
        public const int CMD_RCL = 34;
        public const int CMD_RCL_DIV = 35;
        public const int CMD_RCL_MUL = 36;
        public const int CMD_RCL_SUB = 37;
        public const int CMD_RCL_ADD = 38;
        public const int CMD_FIX = 39;
        public const int CMD_SCI = 40;
        public const int CMD_ENG = 41;
        public const int CMD_ALL = 42;
        public const int CMD_NULL = 43;
        public const int CMD_ASTO = 44;
        public const int CMD_ARCL = 45;
        public const int CMD_CLA = 46;
        public const int CMD_DEG = 47;
        public const int CMD_RAD = 48;
        public const int CMD_GRAD = 49;
        public const int CMD_RECT = 50;
        public const int CMD_POLAR = 51;
        public const int CMD_SIZE = 52;
        public const int CMD_QUIET = 53;
        public const int CMD_CPXRES = 54;
        public const int CMD_REALRES = 55;
        public const int CMD_KEYASN = 56;
        public const int CMD_LCLBL = 57;
        public const int CMD_RDXDOT = 58;
        public const int CMD_RDXCOMMA = 59;
        public const int CMD_CLSIGMA = 60;
        public const int CMD_CLP = 61;
        public const int CMD_CLV = 62;
        public const int CMD_CLST = 63;
        public const int CMD_CLRG = 64;
        public const int CMD_DEL = 65;
        public const int CMD_CLKEYS = 66;
        public const int CMD_CLLCD = 67;
        public const int CMD_CLMENU = 68;
        public const int CMD_CLALLa = 69;
        public const int CMD_TO_DEG = 70;
        public const int CMD_TO_RAD = 71;
        public const int CMD_TO_HR = 72;
        public const int CMD_TO_HMS = 73;
        public const int CMD_TO_REC = 74;
        public const int CMD_TO_POL = 75;
        public const int CMD_IP = 76;
        public const int CMD_FP = 77;
        public const int CMD_RND = 78;
        public const int CMD_ABS = 79;
        public const int CMD_SIGN = 80;
        public const int CMD_MOD = 81;
        public const int CMD_SF = 82;
        public const int CMD_CF = 83;
        public const int CMD_FS_T = 84;
        public const int CMD_FC_T = 85;
        public const int CMD_FSC_T = 86;
        public const int CMD_FCC_T = 87;
        public const int CMD_COMB = 88;
        public const int CMD_PERM = 89;
        public const int CMD_FACT = 90;
        public const int CMD_GAMMA = 91;
        public const int CMD_RAN = 92;
        public const int CMD_SEED = 93;
        public const int CMD_LBL = 94;
        public const int CMD_RTN = 95;
        public const int CMD_INPUT = 96;
        public const int CMD_VIEW = 97;
        public const int CMD_AVIEW = 98;
        public const int CMD_XEQ = 99;
        public const int CMD_PROMPT = 100;
        public const int CMD_PSE = 101;
        public const int CMD_ISG = 102;
        public const int CMD_DSE = 103;
        public const int CMD_AIP = 104;
        public const int CMD_XTOA = 105;
        public const int CMD_AGRAPH = 106;
        public const int CMD_PIXEL = 107;
        public const int CMD_BEEP = 108;
        public const int CMD_TONE = 109;
        public const int CMD_MVAR = 110;
        public const int CMD_VARMENU = 111;
        public const int CMD_GETKEY = 112;
        public const int CMD_MENU = 113;
        public const int CMD_KEYG = 114;
        public const int CMD_KEYX = 115;
        public const int CMD_X_EQ_0 = 116;
        public const int CMD_X_NE_0 = 117;
        public const int CMD_X_LT_0 = 118;
        public const int CMD_X_GT_0 = 119;
        public const int CMD_X_LE_0 = 120;
        public const int CMD_X_GE_0 = 121;
        public const int CMD_X_EQ_Y = 122;
        public const int CMD_X_NE_Y = 123;
        public const int CMD_X_LT_Y = 124;
        public const int CMD_X_GT_Y = 125;
        public const int CMD_X_LE_Y = 126;
        public const int CMD_X_GE_Y = 127;
        public const int CMD_PRSIGMA = 128;
        public const int CMD_PRP = 129;
        public const int CMD_PRV = 130;
        public const int CMD_PRSTK = 131;
        public const int CMD_PRA = 132;
        public const int CMD_PRX = 133;
        public const int CMD_PRUSR = 134;
        public const int CMD_LIST = 135;
        public const int CMD_ADV = 136;
        public const int CMD_PRLCD = 137;
        public const int CMD_DELAY = 138;
        public const int CMD_PON = 139;
        public const int CMD_POFF = 140;
        public const int CMD_MAN = 141;
        public const int CMD_NORM = 142;
        public const int CMD_TRACE = 143;
        public const int CMD_SIGMAADD = 144;
        public const int CMD_SIGMASUB = 145;
        public const int CMD_GTO = 146;
        public const int CMD_END = 147;
        public const int CMD_NUMBER = 148;
        public const int CMD_STRING = 149;
        public const int CMD_RUN = 150;
        public const int CMD_SST = 151;
        public const int CMD_GTODOT = 152;
        public const int CMD_GTODOTDOT = 153;
        public const int CMD_STOP = 154;
        public const int CMD_NEWMAT = 155;
        public const int CMD_RUP = 156;
        public const int CMD_REAL_T = 157;
        public const int CMD_CPX_T = 158;
        public const int CMD_STR_T = 159;
        public const int CMD_MAT_T = 160;
        public const int CMD_DIM_T = 161;
        public const int CMD_ASSIGNa = 162;
        public const int CMD_ASSIGNb = 163;
        public const int CMD_ASGN01 = 164;
        public const int CMD_ASGN02 = 165;
        public const int CMD_ASGN03 = 166;
        public const int CMD_ASGN04 = 167;
        public const int CMD_ASGN05 = 168;
        public const int CMD_ASGN06 = 169;
        public const int CMD_ASGN07 = 170;
        public const int CMD_ASGN08 = 171;
        public const int CMD_ASGN09 = 172;
        public const int CMD_ASGN10 = 173;
        public const int CMD_ASGN11 = 174;
        public const int CMD_ASGN12 = 175;
        public const int CMD_ASGN13 = 176;
        public const int CMD_ASGN14 = 177;
        public const int CMD_ASGN15 = 178;
        public const int CMD_ASGN16 = 179;
        public const int CMD_ASGN17 = 180;
        public const int CMD_ASGN18 = 181;
        public const int CMD_ON = 182;
        public const int CMD_OFF = 183;
        public const int CMD_KEY1G = 184;
        public const int CMD_KEY2G = 185;
        public const int CMD_KEY3G = 186;
        public const int CMD_KEY4G = 187;
        public const int CMD_KEY5G = 188;
        public const int CMD_KEY6G = 189;
        public const int CMD_KEY7G = 190;
        public const int CMD_KEY8G = 191;
        public const int CMD_KEY9G = 192;
        public const int CMD_KEY1X = 193;
        public const int CMD_KEY2X = 194;
        public const int CMD_KEY3X = 195;
        public const int CMD_KEY4X = 196;
        public const int CMD_KEY5X = 197;
        public const int CMD_KEY6X = 198;
        public const int CMD_KEY7X = 199;
        public const int CMD_KEY8X = 200;
        public const int CMD_KEY9X = 201;
        public const int CMD_VMEXEC = 202;
        public const int CMD_VMSTO = 203;
        public const int CMD_SIGMAREG = 204;
        public const int CMD_SIGMAREG_T = 205;
        public const int CMD_CLD = 206;
        public const int CMD_ACOSH = 207;
        public const int CMD_ALENG = 208;
        public const int CMD_ALLSIGMA = 209;
        public const int CMD_AND = 210;
        public const int CMD_AOFF = 211;
        public const int CMD_AON = 212;
        public const int CMD_AROT = 213;
        public const int CMD_ASHF = 214;
        public const int CMD_ASINH = 215;
        public const int CMD_ATANH = 216;
        public const int CMD_ATOX = 217;
        public const int CMD_BASEADD = 218;
        public const int CMD_BASESUB = 219;
        public const int CMD_BASEMUL = 220;
        public const int CMD_BASEDIV = 221;
        public const int CMD_BASECHS = 222;
        public const int CMD_BEST = 223;
        public const int CMD_BINM = 224;
        public const int CMD_BIT_T = 225;
        public const int CMD_BST = 226;
        public const int CMD_CORR = 227;
        public const int CMD_COSH = 228;
        public const int CMD_CROSS = 229;
        public const int CMD_CUSTOM = 230;
        public const int CMD_DECM = 231;
        public const int CMD_DELR = 232;
        public const int CMD_DET = 233;
        public const int CMD_DIM = 234;
        public const int CMD_DOT = 235;
        public const int CMD_EDIT = 236;
        public const int CMD_EDITN = 237;
        public const int CMD_EXITALL = 238;
        public const int CMD_EXPF = 239;
        public const int CMD_E_POW_X_1 = 240;
        public const int CMD_FCSTX = 241;
        public const int CMD_FCSTY = 242;
        public const int CMD_FNRM = 243;
        public const int CMD_GETM = 244;
        public const int CMD_GROW = 245;
        public const int CMD_HEXM = 246;
        public const int CMD_HMSADD = 247;
        public const int CMD_HMSSUB = 248;
        public const int CMD_I_ADD = 249;
        public const int CMD_I_SUB = 250;
        public const int CMD_INDEX = 251;
        public const int CMD_INSR = 252;
        public const int CMD_INTEG = 253;
        public const int CMD_INVRT = 254;
        public const int CMD_J_ADD = 255;
        public const int CMD_J_SUB = 256;
        public const int CMD_LINF = 257;
        public const int CMD_LINSIGMA = 258;
        public const int CMD_LN_1_X = 259;
        public const int CMD_LOGF = 260;
        public const int CMD_MEAN = 261;
        public const int CMD_NOT = 262;
        public const int CMD_OCTM = 263;
        public const int CMD_OLD = 264;
        public const int CMD_OR = 265;
        public const int CMD_PGMSLV = 266;
        public const int CMD_PGMINT = 267;
        public const int CMD_POSA = 268;
        public const int CMD_PUTM = 269;
        public const int CMD_PWRF = 270;
        public const int CMD_RCLEL = 271;
        public const int CMD_RCLIJ = 272;
        public const int CMD_RNRM = 273;
        public const int CMD_ROTXY = 274;
        public const int CMD_RSUM = 275;
        public const int CMD_SWAP_R = 276;
        public const int CMD_SDEV = 277;
        public const int CMD_SINH = 278;
        public const int CMD_SLOPE = 279;
        public const int CMD_SOLVE = 280;
        public const int CMD_STOEL = 281;
        public const int CMD_STOIJ = 282;
        public const int CMD_SUM = 283;
        public const int CMD_TANH = 284;
        public const int CMD_TRANS = 285;
        public const int CMD_UVEC = 286;
        public const int CMD_WMEAN = 287;
        public const int CMD_WRAP = 288;
        public const int CMD_X_SWAP = 289;
        public const int CMD_XOR = 290;
        public const int CMD_YINT = 291;
        public const int CMD_TO_DEC = 292;
        public const int CMD_TO_OCT = 293;
        public const int CMD_LEFT = 294;
        public const int CMD_UP = 295;
        public const int CMD_DOWN = 296;
        public const int CMD_RIGHT = 297;
        public const int CMD_PERCENT_CH = 298;
        public const int CMD_SIMQ = 299;
        public const int CMD_MATA = 300;
        public const int CMD_MATB = 301;
        public const int CMD_MATX = 302;
        public const int CMD_GOTOROW = 303;
        public const int CMD_GOTOCOLUMN = 304;
        public const int CMD_A_THRU_F = 305;
        public const int CMD_CLALLb = 306;
        public const int CMD_PGMSLVi = 307;
        public const int CMD_PGMINTi = 308;
        public const int CMD_VMSTO2 = 309;
        public const int CMD_VMSOLVE = 310;
        public const int CMD_MAX = 311;
        public const int CMD_MIN = 312;
        public const int CMD_FIND = 313;
        public const int CMD_XROM = 314;
        public const int CMD_OPENF = 315;
        public const int CMD_CLOSEF = 316;
        public const int CMD_READP = 317;
        public const int CMD_WRITEP = 318;
        public const int CMD_GETXY = 319;
        public const int CMD_PUTXY = 320;
        public const int CMD_CLRP = 321;
        public const int CMD_CLRD = 322;
        public const int CMD_APPD = 323;
        public const int CMD_GETN = 324;
        public const int CMD_PUTN = 325;
        public const int CMD_GETZ = 326;
        public const int CMD_PUTZ = 327;
        public const int CMD_DELP = 328;
        public const int CMD_DROP = 329;
        public const int CMD_ACCEL = 330;
        public const int CMD_LOCAT = 331;
        public const int CMD_HEADING = 332;
        public const int CMD_ADATE = 333;
        public const int CMD_ALMCAT = 334;
        public const int CMD_ALMNOW = 335;
        public const int CMD_ATIME = 336;
        public const int CMD_ATIME24 = 337;
        public const int CMD_CLK12 = 338;
        public const int CMD_CLK24 = 339;
        public const int CMD_CLKT = 340;
        public const int CMD_CLKTD = 341;
        public const int CMD_CLOCK = 342;
        public const int CMD_CORRECT = 343;
        public const int CMD_DATE = 344;
        public const int CMD_DATE_PLUS = 345;
        public const int CMD_DDAYS = 346;
        public const int CMD_DMY = 347;
        public const int CMD_DOW = 348;
        public const int CMD_MDY = 349;
        public const int CMD_RCLAF = 350;
        public const int CMD_RCLSW = 351;
        public const int CMD_RUNSW = 352;
        public const int CMD_SETAF = 353;
        public const int CMD_SETDATE = 354;
        public const int CMD_SETIME = 355;
        public const int CMD_SETSW = 356;
        public const int CMD_STOPSW = 357;
        public const int CMD_SW = 358;
        public const int CMD_T_PLUS_X = 359;
        public const int CMD_TIME = 360;
        public const int CMD_XYZALM = 361;
        public const int CMD_CLALMA = 362;
        public const int CMD_CLALMX = 363;
        public const int CMD_CLRALMS = 364;
        public const int CMD_RCLALM = 365;
        public const int CMD_SWPT = 366;
        public const int CMD_SENTINEL = 367;
        public const int ARG_NONE = 0;
        public const int ARG_VAR = 1;
        public const int ARG_REAL = 2;
        public const int ARG_NAMED = 3;
        public const int ARG_NUM9 = 4;
        public const int ARG_NUM11 = 5;
        public const int ARG_NUM99 = 6;
        public const int ARG_COUNT = 7;
        public const int ARG_LBL = 8;
        public const int ARG_CKEY = 9;
        public const int ARG_MKEY = 10;
        public const int ARG_PRGM = 11;
        public const int ARG_RVAR = 12;
        public const int ARG_MAT = 13;
        public const int ARG_OTHER = 14;
        public const int FLAG_NONE = 0;
        public const int FLAG_PRGM_ONLY = 1;
        public const int FLAG_IMMED = 2;
        public const int FLAG_HIDDEN = 4;
        public const int FLAG_NO_PRGM = 8;
        public const int FLAG_NO_SHOW = 16;
        public const int ARGTYPE_NONE = 0;
        public const int ARGTYPE_NUM = 1;
        public const int ARGTYPE_NEG_NUM = 2;
        public const int ARGTYPE_STK = 3;
        public const int ARGTYPE_STR = 4;
        public const int ARGTYPE_IND_NUM = 5;
        public const int ARGTYPE_IND_STK = 6;
        public const int ARGTYPE_IND_STR = 7;
        public const int ARGTYPE_COMMAND = 8;
        public const int ARGTYPE_LCLBL = 9;
        public const int ARGTYPE_DOUBLE = 10;
        public const int ARGTYPE_LBLINDEX = 11;
        public const int ERR_NONE = 0;
        public const int ERR_ALPHA_DATA_IS_INVALID = 1;
        public const int ERR_INSUFFICIENT_MEMORY = 2;
        public const int ERR_NOT_YET_IMPLEMENTED = 3;
        public const int ERR_OUT_OF_RANGE = 4;
        public const int ERR_DIVIDE_BY_0 = 5;
        public const int ERR_INVALID_TYPE = 6;
        public const int ERR_INVALID_DATA = 7;
        public const int ERR_DIMENSION_ERROR = 8;
        public const int ERR_SIZE_ERROR = 9;
        public const int ERR_INTERNAL_ERROR = 10;
        public const int ERR_NONEXISTENT = 11;
        public const int ERR_RESTRICTED_OPERATION = 12;
        public const int ERR_YES = 13;
        public const int ERR_NO = 14;
        public const int ERR_STOP = 15;
        public const int ERR_LABEL_NOT_FOUND = 16;
        public const int ERR_NO_REAL_VARIABLES = 17;
        public const int ERR_NO_COMPLEX_VARIABLES = 18;
        public const int ERR_NO_MATRIX_VARIABLES = 19;
        public const int ERR_NO_MENU_VARIABLES = 20;
        public const int ERR_STAT_MATH_ERROR = 21;
        public const int ERR_INVALID_FORECAST_MODEL = 22;
        public const int ERR_SOLVE_INTEG_RTN_LOST = 23;
        public const int ERR_SINGULAR_MATRIX = 24;
        public const int ERR_SOLVE_SOLVE = 25;
        public const int ERR_INTEG_INTEG = 26;
        public const int ERR_RUN = 27;
        public const int ERR_INTERRUPTED = 28;
        public const int ERR_PRINTING_IS_DISABLED = 29;
        public const int ERR_INTERRUPTIBLE = 30;
        public const int ERR_NO_VARIABLES = 31;
        public const int KEY_SIGMA = 1;
        public const int KEY_INV = 2;
        public const int KEY_SQRT = 3;
        public const int KEY_LOG = 4;
        public const int KEY_LN = 5;
        public const int KEY_XEQ = 6;
        public const int KEY_STO = 7;
        public const int KEY_RCL = 8;
        public const int KEY_RDN = 9;
        public const int KEY_SIN = 10;
        public const int KEY_COS = 11;
        public const int KEY_TAN = 12;
        public const int KEY_ENTER = 13;
        public const int KEY_SWAP = 14;
        public const int KEY_CHS = 15;
        public const int KEY_E = 16;
        public const int KEY_BSP = 17;
        public const int KEY_UP = 18;
        public const int KEY_7 = 19;
        public const int KEY_8 = 20;
        public const int KEY_9 = 21;
        public const int KEY_DIV = 22;
        public const int KEY_DOWN = 23;
        public const int KEY_4 = 24;
        public const int KEY_5 = 25;
        public const int KEY_6 = 26;
        public const int KEY_MUL = 27;
        public const int KEY_SHIFT = 28;
        public const int KEY_1 = 29;
        public const int KEY_2 = 30;
        public const int KEY_3 = 31;
        public const int KEY_SUB = 32;
        public const int KEY_EXIT = 33;
        public const int KEY_0 = 34;
        public const int KEY_DOT = 35;
        public const int KEY_RUN = 36;
        public const int KEY_ADD = 37;
        public const int MENU_ALPHA1 = 0;
        public const int MENU_ALPHA2 = 1;
        public const int MENU_ALPHA_ABCDE1 = 2;
        public const int MENU_ALPHA_ABCDE2 = 3;
        public const int MENU_ALPHA_FGHI = 4;
        public const int MENU_ALPHA_JKLM = 5;
        public const int MENU_ALPHA_NOPQ1 = 6;
        public const int MENU_ALPHA_NOPQ2 = 7;
        public const int MENU_ALPHA_RSTUV1 = 8;
        public const int MENU_ALPHA_RSTUV2 = 9;
        public const int MENU_ALPHA_WXYZ = 10;
        public const int MENU_ALPHA_PAREN = 11;
        public const int MENU_ALPHA_ARROW = 12;
        public const int MENU_ALPHA_COMP = 13;
        public const int MENU_ALPHA_MATH = 14;
        public const int MENU_ALPHA_PUNC1 = 15;
        public const int MENU_ALPHA_PUNC2 = 16;
        public const int MENU_ALPHA_MISC1 = 17;
        public const int MENU_ALPHA_MISC2 = 18;
        public const int MENU_ST = 19;
        public const int MENU_IND_ST = 20;
        public const int MENU_IND = 21;
        public const int MENU_MODES1 = 22;
        public const int MENU_MODES2 = 23;
        public const int MENU_DISP = 24;
        public const int MENU_CLEAR1 = 25;
        public const int MENU_CLEAR2 = 26;
        public const int MENU_CONVERT1 = 27;
        public const int MENU_CONVERT2 = 28;
        public const int MENU_FLAGS = 29;
        public const int MENU_PROB = 30;
        public const int MENU_CUSTOM1 = 31;
        public const int MENU_CUSTOM2 = 32;
        public const int MENU_CUSTOM3 = 33;
        public const int MENU_PGM_FCN1 = 34;
        public const int MENU_PGM_FCN2 = 35;
        public const int MENU_PGM_FCN3 = 36;
        public const int MENU_PGM_FCN4 = 37;
        public const int MENU_PGM_XCOMP0 = 38;
        public const int MENU_PGM_XCOMPY = 39;
        public const int MENU_PRINT1 = 40;
        public const int MENU_PRINT2 = 41;
        public const int MENU_PRINT3 = 42;
        public const int MENU_TOP_FCN = 43;
        public const int MENU_CATALOG = 44;
        public const int MENU_BLANK = 45;
        public const int MENU_PROGRAMMABLE = 46;
        public const int MENU_VARMENU = 47;
        public const int MENU_STAT1 = 48;
        public const int MENU_STAT2 = 49;
        public const int MENU_STAT_CFIT = 50;
        public const int MENU_STAT_MODL = 51;
        public const int MENU_MATRIX1 = 52;
        public const int MENU_MATRIX2 = 53;
        public const int MENU_MATRIX3 = 54;
        public const int MENU_MATRIX_SIMQ = 55;
        public const int MENU_MATRIX_EDIT1 = 56;
        public const int MENU_MATRIX_EDIT2 = 57;
        public const int MENU_BASE = 58;
        public const int MENU_BASE_A_THRU_F = 59;
        public const int MENU_BASE_LOGIC = 60;
        public const int MENU_SOLVE = 61;
        public const int MENU_INTEG = 62;
        public const int MENU_INTEG_PARAMS = 63;
        public const int TYPE_NULL = 0;
        public const int TYPE_REAL = 1;
        public const int TYPE_COMPLEX = 2;
        public const int TYPE_REALMATRIX = 3;
        public const int TYPE_COMPLEXMATRIX = 4;
        public const int TYPE_STRING = 5;
        public const int CATSECT_TOP = 0;
        public const int CATSECT_FCN = 1;
        public const int CATSECT_PGM = 2;
        public const int CATSECT_REAL = 3;
        public const int CATSECT_CPX = 4;
        public const int CATSECT_MAT = 5;
        public const int CATSECT_PGM_ONLY = 6;
        public const int CATSECT_REAL_ONLY = 7;
        public const int CATSECT_MAT_ONLY = 8;
        public const int CATSECT_VARS_ONLY = 9;
        public const int CATSECT_PGM_SOLVE = 10;
        public const int CATSECT_PGM_INTEG = 11;
        public const int MENULEVEL_COMMAND = 0;
        public const int MENULEVEL_ALPHA = 1;
        public const int MENULEVEL_TRANSIENT = 2;
        public const int MENULEVEL_PLAIN = 3;
        public const int MENULEVEL_APP = 4;
        public const int FLAGOP_SF = 0;
        public const int FLAGOP_CF = 1;
        public const int FLAGOP_FS_T = 2;
        public const int FLAGOP_FC_T = 3;
        public const int FLAGOP_FSC_T = 4;
        public const int FLAGOP_FCC_T = 5;
        public const int DISP_ROWS = 2;
        public const int DISP_COLUMNS = 22;
        public const int DIR_LEFT = 0;
        public const int DIR_RIGHT = 1;
        public const int DIR_UP = 2;
        public const int DIR_DOWN = 3;
        public const int MODEL_NONE = -1;
        public const int MODEL_LIN = 0;
        public const int MODEL_LOG = 1;
        public const int MODEL_EXP = 2;
        public const int MODEL_PWR = 3;
        public const int VARS_INCREMENT = 25;
        public const int PRGMS_INCREMENT = 10;
        public const int LABELS_INCREMENT = 10;
        public const int MAX_RTNS = 8;
        public const int SOLVE_VERSION = 4;
        public const int INTEG_VERSION = 2;
        public const int NUM_SHADOWS = 10;
        public const int ROMB_K = 5;
        public const int ROMB_MAX = 20;
        public const int SOLVE_ROOT = 0;
        public const int SOLVE_SIGN_REVERSAL = 1;
        public const int SOLVE_EXTREMUM = 2;
        public const int SOLVE_BAD_GUESSES = 3;
        public const int SOLVE_CONSTANT = 4;
        public const int IMGTYPE_MONO = 1;
        public const int IMGTYPE_GRAY = 2;
        public const int IMGTYPE_COLORMAPPED = 3;
        public const int IMGTYPE_TRUECOLOR = 4;
        public const int KEYMAP_MAX_MACRO_LENGTH = 31;
        public const int MAX_LOADSTRING = 100;
        public const int FILENAMELEN = 256;
        public const int PRINT_LINES = 16384;
        public const int PRINT_BYTESPERLINE = 36;
        public const int PRINT_SIZE = 589824;
        public const int SHELL_VERSION = 7;
        public const int SKIN_MAX_MACRO_LENGTH = 31;
    }
}
