﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

namespace WP7_42s
{
    public partial class App
    {
        public delegate int fp_CMD(arg_struct x);
        public delegate int fp_generic(int x);

        // Signatures for functions mapped by map_binary
        public delegate int mappable_rr(Phloat x, Phloat y, ref Phloat z);
        public delegate int mappable_rc(Phloat x, Phloat yre, Phloat yim, ref Phloat zre, ref Phloat zim);
        public delegate int mappable_cr(Phloat xre, Phloat xim, Phloat y, ref Phloat zre, ref Phloat zim);
        public delegate int mappable_cc(Phloat xre, Phloat xim, Phloat yre, Phloat yim, ref Phloat zre, ref Phloat zim);

        // Signatures for functions mapped by map_unary
        public delegate int mappable_r(Phloat x, ref Phloat z);
        public delegate int mappable_c(Phloat xre, Phloat xim, ref Phloat zre, ref Phloat zim);

        public delegate void fp_completion(int error, vartype res);
        public delegate int fp_completion_rm(int error, ref vartype_realmatrix a, ref int[] perm, Phloat det);
        public delegate void fp_completion_rm2(int error, ref vartype_realmatrix a, ref int[] perm, vartype_realmatrix b);
        public delegate int fp_completion_cm(int error, ref vartype_complexmatrix a, ref int[] perm, Phloat det_re, Phloat det_im);
        public delegate void fp_completion_cm2(int error, ref vartype_complexmatrix a, ref int[] perm, vartype_complexmatrix b);

        public delegate int fp_completion_rc();
        public delegate void fp_completion_rc2(int error, ref vartype_realmatrix a, ref int[] perm, vartype_complexmatrix b);


        // Print support

        public class prp_data_struct
        {
            public char[] buf;
            public int len;
            public int saved_prgm;
            public int cmd;
            public arg_struct arg;
            public int line;
            public int pc;
            public int lines;
            public int width;
            public int first;
            public int trace;
            public int normal;

            public prp_data_struct()
            {
                buf = new char[100];
                arg = new arg_struct();
            }
        }




        // ***** Back-substitution *****

        public class backsub_rr_data_struct
        {
            public vartype_realmatrix a;
            public int[] perm;
            public vartype_realmatrix b;
            public int i;
            public int ii;
            public int j;
            public int ll;
            public int k;
            public Phloat sum;
            public int state;
            public fp_completion_rm2 completion;
        }

        public class backsub_cc_data_struct
        {
            public vartype_complexmatrix a;
            public int[] perm;
            public vartype_complexmatrix b;
            public int i;
            public int ii;
            public int j;
            public int ll;
            public int k;
            public Phloat sum_re;
            public Phloat sum_im;
            public int state;
            public fp_completion_cm2 completion;
        }

        public class backsub_rc_data_struct
        {
            public vartype_realmatrix a;
            public int[] perm;
            public vartype_complexmatrix b;
            public int i;
            public int ii;
            public int j;
            public int ll;
            public int k;
            public Phloat sum_re;
            public Phloat sum_im;
            public int state;
            public fp_completion_rc2 completion;
        } ;



        // ***** LU decomposition *****

        public class lu_r_data_struct
        {
            public vartype_realmatrix a;
            public int[] perm;
            public Phloat det;
            public int i;
            public int imax;
            public int j;
            public int k;
            public Phloat max;
            public Phloat tmp;
            public Phloat sum;
            public Phloat[] scale;
            public int state;
            public fp_completion_rm completion;
        }

        public class lu_c_data_struct
        {
            public vartype_complexmatrix a;
            public int[] perm;
            public Phloat det_re;
            public Phloat det_im;
            public int i;
            public int imax;
            public int j;
            public int k;
            public Phloat max;
            public Phloat tmp;
            public Phloat tmp_re;
            public Phloat tmp_im;
            public Phloat sum_re;
            public Phloat sum_im;
            public Phloat[] scale;
            public int state;
            public fp_completion_cm completion;
        }


        // ***** Matrix-matrix multiplication *****

        public class mul_rr_data_struct
        {
            public vartype_realmatrix left;
            public vartype_realmatrix right;
            public vartype result;
            public int i;
            public int j;
            public int k;
            public Phloat sum;
            public fp_completion completion;
        }

        public class mul_cc_data_struct
        {
            public vartype_complexmatrix left;
            public vartype_complexmatrix right;
            public vartype result;
            public int i;
            public int j;
            public int k;
            public Phloat sum_re;
            public Phloat sum_im;
            public fp_completion completion;
        }

        public class mul_cr_data_struct
        {
            public vartype_complexmatrix left;
            public vartype_realmatrix right;
            public vartype result;
            public int i;
            public int j;
            public int k;
            public Phloat sum_re;
            public Phloat sum_im;
            public fp_completion completion;
        }

        public class mul_rc_data_struct
        {
            public vartype_realmatrix left;
            public vartype_complexmatrix right;
            public vartype result;
            public int i;
            public int j;
            public int k;
            public Phloat sum_re;
            public Phloat sum_im;
            public fp_completion completion;
        }










        public struct synonym_spec
        {
            public string name;            // Length 7
            public int namelen;
            public int cmd_id;
        }

        public struct command_spec
        {
            public string name;            // len 12
            public int name_length;
            public fp_CMD handler;
            public uint hp42s_code;
            public int argtype;
            public int flags;
        }

        public struct extension_struct
        {
            public int first_cmd;
            public int last_cmd;
            public bool enable_flag;
        }



        // Programs
        public class prgm_struct        // Must be class to allow references
        {
            public int capacity;        // TODO: evaluate all structs - should they be classes? should they initialize arrays or others?
            public int size;
            public int lclbl_invalid;
            public byte[] text;
        }

        public class label_struct       // Must be class to allow references
        {
            public byte length;
            public char[] name;        // Set to length 7
            public int prgm;
            public int pc;

            public label_struct()
            {
                name = new char[7];
            }
        }

        // Variables
        public class var_struct
        {
            public byte length;
            public char[] name;        // Set to length 7
            public vartype value;

            public var_struct()
            {
                name = new char[7];
            }
        }

        public class arg_struct
        {
            public byte type;
            public byte length;
            public int target;

            public class val_struct
            {
                public int num;
                public char[] text = new char[15];
                public char stk;
                public int cmd;        // For backward compatibility only!
                public char lclbl;
            }
            
            public val_struct val = new val_struct();

            // This used to be a member of the 'val' union, but once I changed it
            // from 'double' to 'phloat', that was no longer possible.
            public Phloat val_d = new Phloat();
        }

        public struct core_settings_struct
        {
            public bool matrix_singularmatrix;
            public bool matrix_outofrange;
            public bool raw_text;
            public bool auto_repeat;
            public bool enable_ext_copan;
            public bool enable_ext_bigstack;
            public bool enable_ext_accel;
            public bool enable_ext_locat;
            public bool enable_ext_heading;
            public bool enable_ext_time;
        }


        public class menu_item_spec
        {
            public MENU menuid;
            public string title;
        }


        public class menu_spec
        {
            public MENU parent;
            public MENU next;
            public MENU prev;

            public menu_item_spec[] child;
        }



        // FLAGS
        // Note: flags whose names start with VIRTUAL_ are named here for reference
        // only; they are actually handled by virtual_flag_handler(). Setting or
        // clearing them in 'flags' has no effect.
        // Flags whose names are the letter 'f' followed by two digits have no
        // specified meaning according to the HP-42S manual; they are either user flags
        // or reserved.
        //
        public struct flags_struct
        {
            // farray union temporarily removed until needed
            // removed f sub-struct for now
            public byte f00;
            public byte f01;
            public byte f02;
            public byte f03;
            public byte f04;
            public byte f05;
            public byte f06;
            public byte f07;
            public byte f08;
            public byte f09;
            public byte f10;
            public byte auto_exec;
            public byte double_wide_print;
            public byte lowercase_print;
            public byte f14;
            public byte trace_print; // 'normal_print' ignored if this is set
            public byte normal_print;
            public byte f17;
            public byte f18;
            public byte f19;
            public byte f20;
            public byte printer_enable;
            public byte numeric_data_input;
            public byte alpha_data_input;
            public byte range_error_ignore;
            public byte error_ignore;
            public byte audio_enable;
            public byte VIRTUAL_custom_menu;
            public byte decimal_point;
            public byte thousands_separators;
            public byte stack_lift_disable;
            public byte f31;
            public byte f32;
            public byte f33;
            public byte agraph_control1; // 0 (default): dst = dst | src, 1: dst = src,
            public byte agraph_control0; // 2: dst = dst & ~src, 3: dst = dst ^ src
            public byte digits_bit3;
            public byte digits_bit2;
            public byte digits_bit1;
            public byte digits_bit0;
            public byte fix_or_all;
            public byte eng_or_all;
            public byte grad;
            public byte rad;
            public byte continuous_on;
            public byte VIRTUAL_solving;
            public byte VIRTUAL_integrating;
            public byte VIRTUAL_variable_menu;
            public byte alpha_mode;
            public byte VIRTUAL_low_battery;
            public byte message;
            public byte two_line_message;
            public byte prgm_mode;
            public byte VIRTUAL_input;
            public byte f54;
            public byte printer_exists;
            public byte lin_fit;
            public byte log_fit;
            public byte exp_fit;
            public byte pwr_fit;
            public byte all_sigma;
            public byte log_fit_invalid;
            public byte exp_fit_invalid;
            public byte pwr_fit_invalid;
            public byte f64;
            public byte VIRTUAL_matrix_editor;
            public byte grow;
            public byte f67;
            public byte base_bit0; // Note: dec=0, bin=1, oct=7, hex=15
            public byte base_bit1;
            public byte base_bit2;
            public byte base_bit3;
            public byte local_label;
            public byte polar;
            public byte real_result_only;
            public byte VIRTUAL_programmable_menu;
            public byte matrix_edge_wrap;
            public byte matrix_end_wrap;
            public byte f78;
            public byte f79;
            public byte f80;
            public byte f81;
            public byte f82;
            public byte f83;
            public byte f84;
            public byte f85;
            public byte f86;
            public byte f87;
            public byte f88;
            public byte f89;
            public byte f90;
            public byte f91;
            public byte f92;
            public byte f93;
            public byte f94;
            public byte f95;
            public byte f96;
            public byte f97;
            public byte f98;
            public byte f99;
        }


        public class vartype
        {
            public TYPE type;
        }

        public class vartype_real : vartype
        {
            public Phloat x;
        }

        public class vartype_complex : vartype
        {
            public Phloat re;
            public Phloat im;
        }


        public class matrix_data
        {
            public int refcount;
        }


        public class realmatrix_data : matrix_data
        {
            public Phloat[] data;
            public char[] is_string;
        }

        public class vartype_realmatrix : vartype
        {
            public int rows;
            public int columns;
            public realmatrix_data array;
        }


        public class complexmatrix_data : matrix_data
        {
            public Phloat[] data;
        }

        public class vartype_complexmatrix : vartype
        {
            public int rows;
            public int columns;
            public complexmatrix_data array;
        }


        public class vartype_string : vartype
        {
            public int length;
            public char[] text = new char[6];
        }

        public struct matrix_persister
        {
            public TYPE type;
            public int rows;
            public int columns;
        }

        // Integrator
        public class integ_state
        {
            public int version;
            public char[] prgm_name;               // len 7
            public int prgm_length;
            public char[] active_prgm_name;        // len 7
            public int active_prgm_length;
            public char[] var_name;                // len 7
            public int var_length;
            public int keep_running;
            public int prev_prgm;
            public int prev_pc;
            public int state;
            public Phloat llim, ulim, acc;
            public Phloat a, b, eps;
            public int n, m, i, k;
            public Phloat h, sum;
            public Phloat[] c;                     // len ROMB_K
            public Phloat[] s;                     // len ROMB_K+1
            public int nsteps;
            public Phloat p;
            public Phloat t, u;
            public Phloat prev_int;
            public int evalCount;

            public integ_state()
            {
                prgm_name = new char[7];
                active_prgm_name = new char[7];
                var_name = new char[7];

                c = new Phloat[ROMB_K];
                for (int i = 0; i < c.Length; i++)
                    c[i] = new Phloat();

                s = new Phloat[ROMB_K+1];
                for (int i = 0; i < s.Length; i++)
                    s[i] = new Phloat();

                llim = new Phloat();
                ulim = new Phloat();
                acc = new Phloat();
                a = new Phloat();
                b = new Phloat();
                eps = new Phloat();
                h = new Phloat();
                sum = new Phloat();
                p = new Phloat();
                t = new Phloat();
                u = new Phloat();
                prev_int = new Phloat();
            }
        }


        // Solver
        public class solve_state
        {
            public int version;
            public char[] prgm_name;                // len 7
            public int prgm_length;
            public char[] active_prgm_name;         // len 7
            public int active_prgm_length;
            public char[] var_name;                 // len 7
            public int var_length;
            public int keep_running;
            public int prev_prgm;
            public int prev_pc;
            public int state;
            public int which;
            public int toggle;
            public int retry_counter;
            public Phloat retry_value;
            public Phloat x1, x2, x3;
            public Phloat fx1, fx2;
            public Phloat prev_x, curr_x, curr_f;
            public Phloat xm, fxm;
            public char[][] shadow_name;            // len1: NUM_SHADOWS, len2: 7
            public int[] shadow_length;             // len NUM_SHADOWS
            public Phloat[] shadow_value;           // len NUM_SHADOWS
            public uint last_disp_time;

            public solve_state()
            {
                prgm_name = new char[7];
                active_prgm_name = new char[7];
                var_name = new char[7];

                shadow_name = new char[NUM_SHADOWS][];
                for (int i = 0; i < shadow_name.Length; i++)
                    shadow_name[i] = new char[7];

                shadow_length = new int[NUM_SHADOWS];
                shadow_value = new Phloat[NUM_SHADOWS];
                for (int i = 0; i < shadow_value.Length; i++)
                    shadow_value[i] = new Phloat();

                retry_value = new Phloat();
                x1 = new Phloat();
                x2 = new Phloat();
                x3 = new Phloat();
                fx1 = new Phloat();
                fx2 = new Phloat();
                prev_x = new Phloat();
                curr_x = new Phloat();
                curr_f = new Phloat();
                xm = new Phloat();
                fxm = new Phloat();
            }
        }


        public class keymap_entry
        {
            public bool ctrl;
            public bool alt;
            public bool shift;
            public bool cshift;

            public int keycode;

            public byte[] macro;
        }



        public struct sigma_label_spec
        {
            public char[] text;
            public int length;
        }


        public class sum_struct
        {
            public Phloat x;
            public Phloat x2;
            public Phloat y;
            public Phloat y2;
            public Phloat xy;
            public Phloat n;
            public Phloat lnx;
            public Phloat lnx2;
            public Phloat lny;
            public Phloat lny2;
            public Phloat lnxlny;
            public Phloat xlny;
            public Phloat ylnx;

            public sum_struct()
            {
                x = new Phloat(0);
                x2 = new Phloat(0);
                y = new Phloat(0);
                y2 = new Phloat(0);
                xy = new Phloat(0);
                n = new Phloat(0);
                lnx = new Phloat(0);
                lnx2 = new Phloat(0);
                lny = new Phloat(0);
                lny2 = new Phloat(0);
                lnxlny = new Phloat(0);
                xlny = new Phloat(0);
                ylnx = new Phloat(0);
            }
        }



        public class model_struct
        {
            public Phloat x;
            public Phloat x2;
            public Phloat y;
            public Phloat y2;
            public Phloat xy;
            public Phloat n;

            public int ln_before;
            public int exp_after;
            public int valid;

            public Phloat slope;
            public Phloat yint;

            public model_struct()
            {
                x = new Phloat(0);
                x2 = new Phloat(0);
                y = new Phloat(0);
                y2 = new Phloat(0);
                xy = new Phloat(0);
                n = new Phloat(0);

                ln_before = 0;
                exp_after = 0;
                valid = 0;

                slope = new Phloat(0);
                yint = new Phloat(0);
            }
        }
    }
}