//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Runtime.InteropServices;
using System.Windows.Threading;
using System.Windows.Controls;
using System.Collections;
using System.Collections.Generic;
using System.Windows.Media.Imaging;


namespace WP7_42s
{
    public partial class App
    {
        public static bool trialMode = false;

        public static Display Disp;     // Store a public referenece to the display here for easy access
        public static MainPage Main;    // Store a public referenece to the main screen here for easy access

        public static DateTime prgm_start_time;
        public static TimeSpan prgm_time_slice = new TimeSpan(0, 0, 0, 0, 100);
        public static bool prgm_block = false;
        public static bool calc_off = false;
        public static List<int> tones;

        public static bool showSplash = true;


        // printing support
        public static prp_data_struct prp_data;
        public static int imgNum = 0;
        public static int imgQuality = 90;
        public static bool scrollPrint = false;




        // Pref settings
        public static double prefBeepVol;
        public static double prefKeyClickVol;
        public static bool prefKeyHighlightEn;


        //------------------------------------------------------------------------------------------
        //
        // Hard reset init
        //

        // Registers
        public static vartype reg_x;
        public static vartype reg_y;
        public static vartype reg_z;
        public static vartype reg_t;
        public static vartype reg_lastx;

        // Alpha
        public static int reg_alpha_length;
        public static char[] reg_alpha;             // len 44
        public static int reg_alpha_len_max;

        // Variables
        public static int vars_capacity;
        public static int vars_count;
        public static var_struct[] vars;

        // Programs
        public static int prgms_capacity;
        public static int prgms_count;
        public static prgm_struct[] prgms;
        public static int labels_capacity;
        public static int labels_count;
        public static label_struct[] labels;

        public static int pending_command;

        // Matrix editor / matrix indexing
        public static int matedit_mode;         // 0=off, 1=index, 2=edit, 3=editn
        public static char[] matedit_name;
        public static int matedit_length;
        public static vartype matedit_x;
        public static int matedit_i;
        public static int matedit_j;
        public static int matedit_prev_appmenu;

        // INPUT
        public static char[] input_name;
        public static int input_length;
        public static arg_struct input_arg;

        // BASE application
        public static int baseapp;

        // Flags
        public static flags_struct flags;

        public static bool mode_clall;
        public static bool mode_command_entry;
        public static bool mode_number_entry;
        public static bool mode_alpha_entry;
        public static bool mode_shift;
        public static int mode_appmenu;
        public static int mode_plainmenu;
        public static int mode_transientmenu;
        public static int mode_alphamenu;
        public static int mode_commandmenu;
        public static bool mode_running;
        public static bool mode_varmenu;
        public static bool mode_getkey;
        public static bool mode_pause;

        public static int current_prgm;         // = -1;
        public static int pc;
        public static int prgm_highlight_row;

        public static int varmenu_length;
        public static char[] varmenu = new char[7];
        public static int varmenu_rows;
        public static int varmenu_row;
        public static int[] varmenu_labellength;    // size int[6];
        public static char[][] varmenu_labeltext;   // size char[6][7];
        public static int varmenu_role;

        public static bool mode_updown;
        public static int mode_sigma_reg;
        public static int mode_goose;
        public static bool mode_time_clktd;
        public static bool mode_time_clk24;
        public static bool mode_time_dmy;

        public static core_settings_struct core_settings;


        
        //------------------------------------------------------------------------------------------
        //
        // Static init only
        //
        
        public static int array_count;
        public static int array_list_capacity;
        public static matrix_data[] array_list;

        public static int varray_count;
        public static int varray_list_capacity;
        public static vartype[] varray_list;



        public static bool mode_plainmenu_sticky;


        public static Phloat entered_number = new Phloat();
        public static int entered_string_length;
        public static char[] entered_string = new char[15];

        public static arg_struct pending_command_arg = new arg_struct();
        public static int xeq_invisible;

        public static int incomplete_command;
        public static int incomplete_ind;
        public static int incomplete_alpha;
        public static int incomplete_length;
        public static int incomplete_maxdigits;
        public static int incomplete_argtype;
        public static int incomplete_num;
        public static char[] incomplete_str = new char[7];
        public static int incomplete_saved_pc;
        public static int incomplete_saved_highlight_row;

        public static char[] cmdline = new char[100];
        public static int cmdline_length;
        public static int cmdline_row;

        
        public static Random rand = null;


        // Random number generator
        public static Phloat random_number = new Phloat(0);
        public static Phloat rnd_multiplier;


        // NORM & TRACE mode: number waiting to be printed
        public static int deferred_print = 0;

        // Keystroke buffer - holds keystrokes received while
        // there is a program running.
        //
        public static int keybuf_head = 0;
        public static int keybuf_tail = 0;
        public static int[] keybuf = new int[16];



        public static int rtn_sp;
        public static int[] rtn_prgm = new int[App.MAX_RTNS];
        public static int[] rtn_pc = new int[App.MAX_RTNS];



        public static bool bin_dec_mode_switch;
        public static bool state_file_has_old_bcd = false;

        public static fp_generic mode_interruptible = null;
        public static bool mode_disable_stack_lift;   // transient
        public static bool mode_stoppable;
        public static int oldpc;

        public static vartype prv_var;
        public static int prv_index;

        public static int repeating = 0;
        public static int repeating_shift;
        public static int repeating_key;

        public static int remove_program_catalog = 0;


        // Shell

        public static DispatcherTimer timer;
        public static DispatcherTimer timer3;

        public static int running = 0;
        public static int enqueued = 0;
        public static int ckey = 0;
        public static bool mouse_key;
        public static int active_keycode = 0;
        public static bool ctrl_down = false;
        public static bool alt_down = false;
        public static bool shift_down = false;
        public static byte[] macro = null;

        public static keymap_entry[] keymap = null;
        public static bool just_pressed_shift = false;


        // core_math1

        public static integ_state integ = new integ_state();
        public static solve_state solve = new solve_state();

        // linalg

        public static fp_completion linalg_div_completion;
        public static vartype linalg_div_left;
        public static vartype linalg_div_result;

        public static backsub_rr_data_struct backsub_rr_data;
        public static lu_r_data_struct lu_r_data;

        public static backsub_cc_data_struct backsub_cc_data;
        public static lu_c_data_struct lu_c_data;

        public static backsub_rc_data_struct backsub_rc_data;

        public static mul_rr_data_struct mul_rr_data;
        public static mul_cc_data_struct mul_cc_data;
        public static mul_cr_data_struct mul_cr_data;
        public static mul_rc_data_struct mul_rc_data;

        
        // sto_rcl

        public static arg_struct temp_arg;
        public static bool preserve_ij;
        public static vartype temp_v;



        // misc

        public static int prusr_state;
        public static int prusr_index;

        public static sum_struct sum = new sum_struct();
        public static model_struct model = new model_struct();

        public static bool linalg_det_prev_sm_err;
        public static fp_completion linalg_det_completion;
        public static fp_completion linalg_inv_completion;
        public static vartype linalg_inv_result;

        public static vartype matx_v;








        // display

        public static byte[] display = new byte[272];

        public static bool is_dirty = false;
        public static int dirty_top;
        public static int dirty_left;
        public static int dirty_bottom;
        public static int dirty_right;

        public static int[] catalogmenu_section = new int[5];
        public static int[] catalogmenu_rows = new int[5];
        public static int[] catalogmenu_row = new int[5];
        public static int[,] catalogmenu_item = new int[5, 6];

        public static int[,] custommenu_length = new int[3, 6];
        public static string[,] custommenu_label = new string[3, 6];

        public static arg_struct[] progmenu_arg = new arg_struct[9];
        public static int[] progmenu_is_gto = new int[9];
        public static int[] progmenu_length = new int[6];
        public static string[] progmenu_label = new string[6];

        public static int appmenu_exitcallback = 0;

        public static uint lastgoosetime = 0;










        public static sigma_label_spec[] sigma_labels = new sigma_label_spec[]
        {
            new sigma_label_spec { text="\x0005X=".ToCharArray(),           length=3 },
            new sigma_label_spec { text="\x0005X^2=".ToCharArray(),         length=5 },
            new sigma_label_spec { text="\x0005Y=".ToCharArray(),           length=3 },
            new sigma_label_spec { text="\x0005Y^2=".ToCharArray(),         length=5 },
            new sigma_label_spec { text="\x0005XY=".ToCharArray(),          length=4 },
            new sigma_label_spec { text="N=".ToCharArray(),                 length=2 },
            new sigma_label_spec { text="\x0005LN(X)=".ToCharArray(),       length=7 },
            new sigma_label_spec { text="\x0005LN(X)^2=".ToCharArray(),     length=9 },
            new sigma_label_spec { text="\x0005LN(Y)=".ToCharArray(),       length=7 },
            new sigma_label_spec { text="\x0005LN(Y)^2=".ToCharArray(),     length=9 },
            new sigma_label_spec { text="\x0005LN(X)LN(Y)=".ToCharArray(), length=12 },
            new sigma_label_spec { text="\x0005XLN(Y)=".ToCharArray(),      length=8 },
            new sigma_label_spec { text="\x0005YLN(X)=".ToCharArray(),      length=8 }
        };


        public static string[] solve_message = new string[]
        {
            null,
            "Sign Reversal",
            "Extremum",
            "Bad Guess(es)",
            "Constant?"
        };

        

        public static synonym_spec[] hp41_synonyms = new synonym_spec[]
        {
            new synonym_spec { name="/",      namelen=1, cmd_id=CMD_DIV     },
            new synonym_spec { name="*",      namelen=1, cmd_id=CMD_MUL     },
            new synonym_spec { name="CHS",    namelen=3, cmd_id=CMD_CHS     },
            new synonym_spec { name="DEC",    namelen=3, cmd_id=CMD_TO_DEC  },
            new synonym_spec { name="D-R",    namelen=3, cmd_id=CMD_TO_RAD  },
            new synonym_spec { name="ENTER^", namelen=6, cmd_id=CMD_ENTER   },
            new synonym_spec { name="FACT",   namelen=4, cmd_id=CMD_FACT    },
            new synonym_spec { name="FRC",    namelen=3, cmd_id=CMD_FP      },
            new synonym_spec { name="HMS",    namelen=3, cmd_id=CMD_TO_HMS  },
            new synonym_spec { name="HR",     namelen=2, cmd_id=CMD_TO_HR   },
            new synonym_spec { name="INT",    namelen=3, cmd_id=CMD_IP      },
            new synonym_spec { name="OCT",    namelen=3, cmd_id=CMD_TO_OCT  },
            new synonym_spec { name="P-R",    namelen=3, cmd_id=CMD_TO_REC  },
            new synonym_spec { name="R-D",    namelen=3, cmd_id=CMD_TO_DEG  },
            new synonym_spec { name="RDN",    namelen=3, cmd_id=CMD_RDN     },
            new synonym_spec { name="R-P",    namelen=3, cmd_id=CMD_TO_POL  },
            new synonym_spec { name="ST+",    namelen=3, cmd_id=CMD_STO_ADD },
            new synonym_spec { name="ST/",    namelen=3, cmd_id=CMD_STO_DIV },
            new synonym_spec { name="ST*",    namelen=3, cmd_id=CMD_STO_MUL },
            new synonym_spec { name="ST-",    namelen=3, cmd_id=CMD_STO_SUB },
            new synonym_spec { name="X<=0?",  namelen=5, cmd_id=CMD_X_LE_0  },
            new synonym_spec { name="X<=Y?",  namelen=5, cmd_id=CMD_X_LE_Y  },
            new synonym_spec { name="",       namelen=0, cmd_id=CMD_NONE    }
        };
        
        public static string[] errors = new string[]
        {
            /* NONE */                      null,
            /* ALPHA_DATA_IS_INVALID */     "Alpha Data Is Invalid",
            /* INSUFFICIENT_MEMORY */       "Insufficient Memory",
            /* NOT_YET_IMPLEMENTED */       "Not Yet Implemented",
            /* OUT_OF_RANGE */              "Out of Range",
            /* DIVIDE_BY_0 */               "Divide by 0",
            /* INVALID_TYPE */              "Invalid Type",
            /* INVALID_DATA */              "Invalid Data",
            /* DIMENSION_ERROR */           "Dimension Error",
            /* SIZE_ERROR */                "Size Error",
            /* INTERNAL_ERROR */            "Internal Error",
            /* NONEXISTENT */               "Nonexistent",
            /* RESTRICTED_OPERATION */      "Restricted Operation",
            /* YES */                       "Yes",
            /* NO */                        "No",
            /* STOP */                      null,
            /* LABEL_NOT_FOUND */           "Label Not Found",
            /* NO_REAL_VARIABLES */         "No Real Variables",
            /* NO_COMPLEX_VARIABLES */      "No Complex Variables",
            /* NO_MATRIX_VARIABLES */       "No Matrix Variables",
            /* NO_MENU_VARIABLES */         "No Menu Variables",
            /* STAT_MATH_ERROR */           "Stat Math Error",
            /* INVALID_FORECAST_MODEL */    "Invalid Forecast Model",
            /* SOLVE_INTEG_RTN_LOST */      "Solve/Integ RTN Lost",
            /* SINGULAR_MATRIX */           "Singular Matrix",
            /* SOLVE_SOLVE */               "Solve(Solve)",
            /* INTEG_INTEG */               "Integ(Integ)",
            /* RUN */                       null,
            /* INTERRUPTED */               "Interrupted",
            /* PRINTING_IS_DISABLED */      "Printing Is Disabled",
            /* INTERRUPTIBLE */             null,
            /* NO_VARIABLES */              "No Variables"
        };



        public static menu_spec[] menus = new menu_spec[]
        {
            new menu_spec
            { /* MENU_ALPHA1 */
                parent=MENU.NONE, next=MENU.ALPHA2, prev=MENU.ALPHA2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.ALPHA_ABCDE1, title="ABCDE" },
			        new menu_item_spec { menuid=MENU.ALPHA_FGHI,   title="FGHI"  },
			        new menu_item_spec { menuid=MENU.ALPHA_JKLM,   title="JKLM"  },
			        new menu_item_spec { menuid=MENU.ALPHA_NOPQ1,  title="NOPQ"  },
			        new menu_item_spec { menuid=MENU.ALPHA_RSTUV1, title="RSTUV" },
			        new menu_item_spec { menuid=MENU.ALPHA_WXYZ,   title="WXYZ"  }
                }
            },
            new menu_spec
            { /* MENU_ALPHA2 */
                parent=MENU.NONE, next=MENU.ALPHA1, prev=MENU.ALPHA1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.ALPHA_PAREN, title="( [ {"     },
			        new menu_item_spec { menuid=MENU.ALPHA_ARROW, title="\x10^\x0e" },      // 020=0x10, 016=0x0e
			        new menu_item_spec { menuid=MENU.ALPHA_COMP,  title="< = >"     },
			        new menu_item_spec { menuid=MENU.ALPHA_MATH,  title="MATH"      },
			        new menu_item_spec { menuid=MENU.ALPHA_PUNC1, title="PUNC"      },
			        new menu_item_spec { menuid=MENU.ALPHA_MISC1, title="MISC"      }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_ABCDE1 */
                parent=MENU.ALPHA1, next=MENU.ALPHA_ABCDE2, prev=MENU.ALPHA_ABCDE2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="A" },
			        new menu_item_spec { menuid=MENU.NONE, title="B" },
			        new menu_item_spec { menuid=MENU.NONE, title="C" },
			        new menu_item_spec { menuid=MENU.NONE, title="D" },
			        new menu_item_spec { menuid=MENU.NONE, title="E" },
			        new menu_item_spec { menuid=MENU.NONE, title=" " }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_ABCDE2 */
                parent=MENU.ALPHA1, next=MENU.ALPHA_ABCDE1, prev=MENU.ALPHA_ABCDE1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="\x16" },                  // 026=0x16
			        new menu_item_spec { menuid=MENU.NONE, title="\x14" },                  // 024=0x14
			        new menu_item_spec { menuid=MENU.NONE, title="\x19" },                  // 031=0x19
			        new menu_item_spec { menuid=MENU.NONE, title=" "    },
			        new menu_item_spec { menuid=MENU.NONE, title=" "    },
			        new menu_item_spec { menuid=MENU.NONE, title=" "    }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_FGHI */
                parent=MENU.ALPHA1, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="F" },
			        new menu_item_spec { menuid=MENU.NONE, title="G" },
			        new menu_item_spec { menuid=MENU.NONE, title="H" },
			        new menu_item_spec { menuid=MENU.NONE, title="I" },
			        new menu_item_spec { menuid=MENU.NONE, title=" " },
			        new menu_item_spec { menuid=MENU.NONE, title=" " }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_JKLM */
                parent=MENU.ALPHA1, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="J" },
			        new menu_item_spec { menuid=MENU.NONE, title="K" },
			        new menu_item_spec { menuid=MENU.NONE, title="L" },
			        new menu_item_spec { menuid=MENU.NONE, title="M" },
			        new menu_item_spec { menuid=MENU.NONE, title=" " },
			        new menu_item_spec { menuid=MENU.NONE, title=" " }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_NOPQ1 */
                parent=MENU.ALPHA1, next=MENU.ALPHA_NOPQ2, prev=MENU.ALPHA_NOPQ2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="N" },
			        new menu_item_spec { menuid=MENU.NONE, title="O" },
			        new menu_item_spec { menuid=MENU.NONE, title="P" },
			        new menu_item_spec { menuid=MENU.NONE, title="Q" },
			        new menu_item_spec { menuid=MENU.NONE, title=" " },
			        new menu_item_spec { menuid=MENU.NONE, title=" " }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_NOPQ2 */
                parent=MENU.ALPHA1, next=MENU.ALPHA_NOPQ1, prev=MENU.ALPHA_NOPQ1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="\x15" },                  // 025=0x15
			        new menu_item_spec { menuid=MENU.NONE, title="\x1c" },                  // 034=0x1c
			        new menu_item_spec { menuid=MENU.NONE, title=" "    },
			        new menu_item_spec { menuid=MENU.NONE, title=" "    },
			        new menu_item_spec { menuid=MENU.NONE, title=" "    },
			        new menu_item_spec { menuid=MENU.NONE, title=" "    }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_RSTUV1 */
                parent=MENU.ALPHA1, next=MENU.ALPHA_RSTUV2, prev=MENU.ALPHA_RSTUV2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="R" },
			        new menu_item_spec { menuid=MENU.NONE, title="S" },
			        new menu_item_spec { menuid=MENU.NONE, title="T" },
			        new menu_item_spec { menuid=MENU.NONE, title="U" },
			        new menu_item_spec { menuid=MENU.NONE, title="V" },
			        new menu_item_spec { menuid=MENU.NONE, title=" " }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_RSTUV2 */
                parent=MENU.ALPHA1, next=MENU.ALPHA_RSTUV1, prev=MENU.ALPHA_RSTUV1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title=" "    },
			        new menu_item_spec { menuid=MENU.NONE, title=" "    },
			        new menu_item_spec { menuid=MENU.NONE, title=" "    },
			        new menu_item_spec { menuid=MENU.NONE, title="\x1d" },                  // 035=0x1d
			        new menu_item_spec { menuid=MENU.NONE, title=" "    },
			        new menu_item_spec { menuid=MENU.NONE, title=" "    }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_WXYZ */
                parent=MENU.ALPHA1, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="W" },
			        new menu_item_spec { menuid=MENU.NONE, title="X" },
			        new menu_item_spec { menuid=MENU.NONE, title="Y" },
			        new menu_item_spec { menuid=MENU.NONE, title="Z" },
			        new menu_item_spec { menuid=MENU.NONE, title=" " },
			        new menu_item_spec { menuid=MENU.NONE, title=" " }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_PAREN */
                parent=MENU.ALPHA2, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="(" },
			        new menu_item_spec { menuid=MENU.NONE, title=")" },
			        new menu_item_spec { menuid=MENU.NONE, title="[" },
			        new menu_item_spec { menuid=MENU.NONE, title="]" },
			        new menu_item_spec { menuid=MENU.NONE, title="{" },
			        new menu_item_spec { menuid=MENU.NONE, title="}" }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_ARROW */
                parent=MENU.ALPHA2, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="\x10" },                  // 020=0x10
			        new menu_item_spec { menuid=MENU.NONE, title="^"    },
			        new menu_item_spec { menuid=MENU.NONE, title="\x0e" },                  // 016=0x0e
			        new menu_item_spec { menuid=MENU.NONE, title="\x0f" },                  // 017=0x0f
			        new menu_item_spec { menuid=MENU.NONE, title=" "    },
			        new menu_item_spec { menuid=MENU.NONE, title=" "    }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_COMP */
                parent=MENU.ALPHA2, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="="    },
			        new menu_item_spec { menuid=MENU.NONE, title="\x0c" },                  // 014=0x0c
			        new menu_item_spec { menuid=MENU.NONE, title="<"    },
			        new menu_item_spec { menuid=MENU.NONE, title=">"    },
			        new menu_item_spec { menuid=MENU.NONE, title="\x09" },                  // 011=0x09
			        new menu_item_spec { menuid=MENU.NONE, title="\x0b" }                   // 013=0x0b
                }
            },
            new menu_spec
            { /* MENU_ALPHA_MATH */
                parent=MENU.ALPHA2, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="\x05" },                  // 005=0x05
			        new menu_item_spec { menuid=MENU.NONE, title="\x03" },                  // 003=0x03
			        new menu_item_spec { menuid=MENU.NONE, title="\x02" },                  // 002=0x02
			        new menu_item_spec { menuid=MENU.NONE, title="\x17" },                  // 027=0x17
			        new menu_item_spec { menuid=MENU.NONE, title="\x13" },                  // 023=0x13
			        new menu_item_spec { menuid=MENU.NONE, title="\x11" }                   // 021=0x11
                }
            },
            new menu_spec
            { /* MENU_ALPHA_PUNC1 */
                parent=MENU.ALPHA2, next=MENU.ALPHA_PUNC2, prev=MENU.ALPHA_PUNC2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title=","  },
			        new menu_item_spec { menuid=MENU.NONE, title=";"  },
			        new menu_item_spec { menuid=MENU.NONE, title=":"  },
			        new menu_item_spec { menuid=MENU.NONE, title="!"  },
			        new menu_item_spec { menuid=MENU.NONE, title="?"  },
			        new menu_item_spec { menuid=MENU.NONE, title="\"" }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_PUNC2 */
                parent=MENU.ALPHA2, next=MENU.ALPHA_PUNC1, prev=MENU.ALPHA_PUNC1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="\x1a" },                  // 032=0x1a
			        new menu_item_spec { menuid=MENU.NONE, title="_"    },
			        new menu_item_spec { menuid=MENU.NONE, title="`"    },
			        new menu_item_spec { menuid=MENU.NONE, title="'"    },
			        new menu_item_spec { menuid=MENU.NONE, title="\x08" },                  // 010=0x08
			        new menu_item_spec { menuid=MENU.NONE, title="\x0a" }                   // 012=0x0a
                }
            },
            new menu_spec
            { /* MENU_ALPHA_MISC1 */
                parent=MENU.ALPHA2, next=MENU.ALPHA_MISC2, prev=MENU.ALPHA_MISC2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="$"    },
			        new menu_item_spec { menuid=MENU.NONE, title="*"    },
			        new menu_item_spec { menuid=MENU.NONE, title="#"    },
			        new menu_item_spec { menuid=MENU.NONE, title="/"    },
			        new menu_item_spec { menuid=MENU.NONE, title="\x1f" },                  // 037=0x1f
			        new menu_item_spec { menuid=MENU.NONE, title=" "    }
                }
            },
            new menu_spec
            { /* MENU_ALPHA_MISC2 */
                parent=MENU.ALPHA2, next=MENU.ALPHA_MISC1, prev=MENU.ALPHA_MISC1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="\x12" },                  // 022=0x12
			        new menu_item_spec { menuid=MENU.NONE, title="&"    },
			        new menu_item_spec { menuid=MENU.NONE, title="@"    },
			        new menu_item_spec { menuid=MENU.NONE, title="\\"   },
			        new menu_item_spec { menuid=MENU.NONE, title="~"    },
			        new menu_item_spec { menuid=MENU.NONE, title="|"    }
                }
            },
            new menu_spec
            { /* MENU_ST */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="ST L" },
			        new menu_item_spec { menuid=MENU.NONE, title="ST X" },
			        new menu_item_spec { menuid=MENU.NONE, title="ST Y" },
			        new menu_item_spec { menuid=MENU.NONE, title="ST Z" },
			        new menu_item_spec { menuid=MENU.NONE, title="ST T" },
			        new menu_item_spec { menuid=MENU.NONE, title=""     }
                }
            },
            new menu_spec
            { /* MENU_IND_ST */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="IND"  },
			        new menu_item_spec { menuid=MENU.NONE, title="ST L" },
			        new menu_item_spec { menuid=MENU.NONE, title="ST X" },
			        new menu_item_spec { menuid=MENU.NONE, title="ST Y" },
			        new menu_item_spec { menuid=MENU.NONE, title="ST Z" },
			        new menu_item_spec { menuid=MENU.NONE, title="ST T" }
                }
            },
            new menu_spec
            { /* MENU_IND */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.NONE, title="IND"  },
			        new menu_item_spec { menuid=MENU.NONE, title="" },
			        new menu_item_spec { menuid=MENU.NONE, title="" },
			        new menu_item_spec { menuid=MENU.NONE, title="" },
			        new menu_item_spec { menuid=MENU.NONE, title="" },
			        new menu_item_spec { menuid=MENU.NONE, title="" }
                }
            },
            new menu_spec
            { /* MENU_MODES1 */
                parent=MENU.NONE, next=MENU.MODES2, prev=MENU.MODES2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x2000 + CMD.DEG),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.RAD),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.GRAD),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.RECT),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.POLAR), title="" }
                }
            },
            new menu_spec
            { /* MENU_MODES2 */
                parent=MENU.NONE, next=MENU.MODES1, prev=MENU.MODES1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SIZE),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.QUIET),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.CPXRES),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.REALRES), title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.KEYASN),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.LCLBL),   title="" }
                }
            },
            new menu_spec
            { /* MENU_DISP */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x2000 + CMD.FIX),      title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.SCI),      title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.ENG),      title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.ALL),      title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.RDXDOT),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.RDXCOMMA), title="" }
                }
            },
            new menu_spec
            { /* MENU_CLEAR1 */
                parent=MENU.NONE, next=MENU.CLEAR2, prev=MENU.CLEAR2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CLSIGMA), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CLP),     title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CLV),     title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CLST),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CLA),     title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CLX),     title="" }
                }
            },
            new menu_spec
            { /* MENU_CLEAR2 */
                parent=MENU.NONE, next=MENU.CLEAR1, prev=MENU.CLEAR1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CLRG),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.DEL),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CLKEYS), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CLLCD),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CLMENU), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CLALLa), title="" }
                }
            },
            new menu_spec
            { /* MENU_CONVERT1 */
                parent=MENU.NONE, next=MENU.CONVERT2, prev=MENU.CONVERT2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.TO_DEG), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.TO_RAD), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.TO_HR),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.TO_HMS), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.TO_REC), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.TO_POL), title="" }
                }
            },
            new menu_spec
            { /* MENU_CONVERT2 */
                parent=MENU.NONE, next=MENU.CONVERT1, prev=MENU.CONVERT1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.IP),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.FP),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.RND),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.ABS),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SIGN), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.MOD),  title="" }
                }
            },
            new menu_spec
            { /* MENU_FLAGS */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SF),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CF),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.FS_T),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.FC_T),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.FSC_T), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.FCC_T), title="" }
                }
            },
            new menu_spec
            { /* MENU_PROB */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.COMB),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PERM),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.FACT),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.GAMMA), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.RAN),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SEED),  title="" }
                }
            },
            new menu_spec
            { /* MENU_CUSTOM1 */
                parent=MENU.NONE, next=MENU.CUSTOM2, prev=MENU.CUSTOM3, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" }
                }
            },
            new menu_spec
            { /* MENU_CUSTOM2 */
                parent=MENU.NONE, next=MENU.CUSTOM3, prev=MENU.CUSTOM1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" }
                }
            },
            new menu_spec
            { /* MENU_CUSTOM3 */
                parent=MENU.NONE, next=MENU.CUSTOM1, prev=MENU.CUSTOM2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" }
                }
            },
            new menu_spec
            { /* MENU_PGM_FCN1 */
                parent=MENU.NONE, next=MENU.PGM_FCN2, prev=MENU.PGM_FCN4, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.LBL),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.RTN),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.INPUT), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.VIEW),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.AVIEW), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.XEQ),   title="" }
                }
            },
            new menu_spec
            { /* MENU_PGM_FCN2 */
                parent=MENU.NONE, next=MENU.PGM_FCN3, prev=MENU.PGM_FCN1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=MENU.PGM_XCOMP0,     title="X?0" },
			        new menu_item_spec { menuid=MENU.PGM_XCOMPY,     title="X?Y" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PROMPT), title=""    },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PSE),    title=""    },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.ISG),    title=""    },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.DSE),    title=""    }
                }
            },
            new menu_spec
            { /* MENU_PGM_FCN3 */
                parent=MENU.NONE, next=MENU.PGM_FCN4, prev=MENU.PGM_FCN2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.AIP),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.XTOA),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.AGRAPH), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PIXEL),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.BEEP),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.TONE),   title="" }
                }
            },
            new menu_spec
            { /* MENU_PGM_FCN4 */
                parent=MENU.NONE, next=MENU.PGM_FCN1, prev=MENU.PGM_FCN3, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.MVAR),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.VARMENU), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.GETKEY),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.MENU),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.KEYG),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.KEYX),    title="" }
                }
            },
            new menu_spec
            { /* MENU_PGM_XCOMP0 */
                parent=MENU.PGM_FCN2, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.X_EQ_0), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.X_NE_0), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.X_LT_0), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.X_GT_0), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.X_LE_0), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.X_GE_0), title="" }
                }
            },
            new menu_spec
            { /* MENU_PGM_XCOMPY */
                parent=MENU.PGM_FCN2, prev=MENU.NONE, next=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.X_EQ_Y), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.X_NE_Y), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.X_LT_Y), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.X_GT_Y), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.X_LE_Y), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.X_GE_Y), title="" }
                }
            },
            new menu_spec
            { /* MENU_PRINT1 */
                parent=MENU.NONE, next=MENU.PRINT2, prev=MENU.PRINT3, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PRSIGMA), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PRP),     title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PRV),     title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PRSTK),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PRA),     title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PRX),     title="" }
                }
            },
            new menu_spec
            { /* MENU_PRINT2 */
                parent=MENU.NONE, next=MENU.PRINT3, prev=MENU.PRINT1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PRUSR), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.LIST),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.ADV),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PRLCD), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.DELAY), title="" }
                }
            },
            new menu_spec
            { /* MENU_PRINT3 */
                parent=MENU.NONE, next=MENU.PRINT1, prev=MENU.PRINT2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x2000 + CMD.PON),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.POFF),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.MAN),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.NORM),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.TRACE), title="" }
                }
            },
            new menu_spec
            { /* MENU_TOP_FCN */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SIGMAADD), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.INV),      title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SQRT),     title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.LOG),      title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.LN),       title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.XEQ),      title="" }
                }
            },
            new menu_spec
            { /* MENU_CATALOG */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" }
                }
            },
            new menu_spec
            { /* MENU_BLANK */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" }
                }
            },
            new menu_spec
            { /* MENU_PROGRAMMABLE */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" }
                }
            },
            new menu_spec
            { /* MENU_VARMENU */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" },
			        new menu_item_spec { menuid=0, title="" }
                }
            },
            new menu_spec
            { /* MENU_STAT1 */
                parent=MENU.NONE, next=MENU.STAT2, prev=MENU.STAT2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SIGMAADD), title=""     },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SUM),      title=""     },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.MEAN),     title=""     },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.WMEAN),    title=""     },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SDEV),     title=""     },
			        new menu_item_spec { menuid=MENU.STAT_CFIT,        title="CFIT" }
                    }
            },
            new menu_spec
            { /* MENU_STAT2 */
                parent=MENU.NONE, next=MENU.STAT1, prev=MENU.STAT1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x2000 + CMD.ALLSIGMA),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.LINSIGMA),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL),       title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL),       title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SIGMAREG),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SIGMAREG_T), title="" }
                }
            },
            new menu_spec
            { /* MENU_STAT_CFIT */
                parent=MENU.STAT1, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.FCSTX), title=""     },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.FCSTY), title=""     },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SLOPE), title=""     },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.YINT),  title=""     },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CORR),  title=""     },
			        new menu_item_spec { menuid=MENU.STAT_MODL,     title="MODL" }
                }
            },
            new menu_spec
            { /* MENU_STAT_MODL */
                parent=MENU.STAT_CFIT, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x2000 + CMD.LINF), title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.LOGF), title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.EXPF), title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.PWRF), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.BEST), title="" }
                }
            },
            new menu_spec
            { /* MENU_MATRIX1 */
                parent=MENU.NONE, next=MENU.MATRIX2, prev=MENU.MATRIX3, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NEWMAT), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.INVRT),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.DET),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.TRANS),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SIMQ),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.EDIT),   title="" }
                }
            },
            new menu_spec
            { /* MENU_MATRIX2 */
                parent=MENU.NONE, next=MENU.MATRIX3, prev=MENU.MATRIX1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.DOT),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.CROSS), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.UVEC),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.DIM),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.INDEX), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.EDITN), title="" }
                }
            },
            new menu_spec
            { /* MENU_MATRIX3 */
                parent=MENU.NONE, next=MENU.MATRIX1, prev=MENU.MATRIX2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.STOIJ), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.RCLIJ), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.STOEL), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.RCLEL), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PUTM),  title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.GETM),  title="" }
                }
            },
            new menu_spec
            { /* MENU_MATRIX_SIMQ */
                parent=MENU.MATRIX1, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.MATA), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.MATB), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.MATX), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL), title="" }
                }
            },
            new menu_spec
            { /* MENU_MATRIX_EDIT1 */
                parent=MENU.NONE, next=MENU.MATRIX_EDIT2, prev=MENU.MATRIX_EDIT2, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.LEFT),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.OLD),     title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.UP),      title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.DOWN),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.GOTOROW), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.RIGHT),   title="" }
                }
            },
            new menu_spec
            { /* MENU_MATRIX_EDIT2 */
                parent=MENU.NONE, next=MENU.MATRIX_EDIT1, prev=MENU.MATRIX_EDIT1, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.INSR), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.DELR), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL), title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.WRAP), title="" },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.GROW), title="" }
                }
            },
            new menu_spec
            { /* MENU_BASE */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.A_THRU_F), title=""      },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.HEXM),     title=""      },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.DECM),     title=""      },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.OCTM),     title=""      },
			        new menu_item_spec { menuid=(MENU)(0x2000 + CMD.BINM),     title=""      },
			        new menu_item_spec { menuid=MENU.BASE_LOGIC,       title="LOGIC" }
                }
            },
            new menu_spec
            { /* MENU_BASE_A_THRU_F */
                parent=MENU.BASE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=0, title="A" },
			        new menu_item_spec { menuid=0, title="B" },
			        new menu_item_spec { menuid=0, title="C" },
			        new menu_item_spec { menuid=0, title="D" },
			        new menu_item_spec { menuid=0, title="E" },
			        new menu_item_spec { menuid=0, title="F" }
                }
            },
            new menu_spec
            { /* MENU_BASE_LOGIC */
                parent=MENU.BASE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.AND),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.OR),    title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.XOR),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NOT),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.BIT_T), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.ROTXY), title="" }
                }
            },
            new menu_spec
            { /* MENU_SOLVE */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.MVAR),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PGMSLV), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.SOLVE),  title="" }
                }
            },
            new menu_spec
            { /* MENU_INTEG */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=(MENU)(0x1000 + CMD.MVAR),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL),   title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.PGMINT), title="" },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.INTEG),  title="" }
                }
            },
            new menu_spec
            { /* MENU.INTEG_PARAMS */
                parent=MENU.NONE, next=MENU.NONE, prev=MENU.NONE, child=new menu_item_spec[6]
		        {
                    new menu_item_spec { menuid=0,                 title="LLIM" },
			        new menu_item_spec { menuid=0,                 title="ULIM" },
			        new menu_item_spec { menuid=0,                 title="ACC"  },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL), title=""     },
			        new menu_item_spec { menuid=(MENU)(0x1000 + CMD.NULL), title=""     },
			        new menu_item_spec { menuid=0,                 title="\x03" }           // 003 = 0x03
                }
            }
        };






        public static void clear_all_prgms()
        {
            if (prgms != null)
            {
                for (int i = 0; i < prgms_count; i++)
                    if (prgms[i].text != null)
                        prgms[i].text = null;

                prgms = null;
            }

            prgms = null;
            prgms_capacity = 0;
            prgms_count = 0;

            if (labels != null)
                labels = null;

            labels = null;
            labels_capacity = 0;
            labels_count = 0;
        }

        public static int find_global_label(arg_struct arg, ref int prgm, ref int pc)
        {
            char[] name = arg.val.text;
            int namelen = arg.length;
            
            for (int i = labels_count - 1; i >= 0; i--)
            {
	            char[] labelname;

                if (labels[i].length != namelen)
	                continue;

	            labelname = labels[i].name;

	            for (int j = 0; j < namelen; j++)
	                if (labelname[j] != name[j])
    		            goto nomatch;

	            prgm = labels[i].prgm;
	            pc = labels[i].pc;
	            return 1;

    	        nomatch:;
            }

            return 0;
        }

        public static int get_command_length(int prgm_index, int pc)
        {
            prgm_struct prgm = prgms[prgm_index];

            int pc2 = pc;
            int command = prgm.text[pc2++];
            int argtype = prgm.text[pc2++];

            command |= (argtype & 240) << 4;
            argtype &= 15;

            if ((command == CMD_GTO || command == CMD_XEQ) && (argtype == ARGTYPE_NUM || argtype == ARGTYPE_LCLBL))
                pc2 += 4;

            switch (argtype)
            {
                case ARGTYPE_NUM:
                case ARGTYPE_NEG_NUM:
                case ARGTYPE_IND_NUM:
                    {
                        while ((prgm.text[pc2++] & 128) == 0) ;
                        break;
                    }
                case ARGTYPE_STK:
                case ARGTYPE_IND_STK:
                case ARGTYPE_COMMAND:
                case ARGTYPE_LCLBL:
                    pc2++;
                    break;
                case ARGTYPE_STR:
                case ARGTYPE_IND_STR:
                    {
                        pc2 += prgm.text[pc2] + 1;
                        break;
                    }
                case ARGTYPE_DOUBLE:
                    pc2 += sizeof_Phloat;
                    break;
            }

            return pc2 - pc;
        }

        public static void get_next_command(ref int pc, ref int command, arg_struct arg, int find_target)
        {
            prgm_struct prgm = prgms[current_prgm];

            int i;
            int target_pc;
            
            int orig_pc = pc;

            command = prgm.text[pc++];
            arg.type = prgm.text[pc++];

            command |= (arg.type & 240) << 4;
            arg.type &= 15;

            if ((command == CMD_GTO || command == CMD_XEQ) && (arg.type == ARGTYPE_NUM || arg.type == ARGTYPE_LCLBL))
            {
    	        if (find_target != 0)
                {
	                target_pc = 0;
	                
                    for (i = 0; i < 4; i++)
		                target_pc = (target_pc << 8) | prgm.text[pc++];

	                if (target_pc != -1)
                    {
		                arg.target = target_pc;
		                find_target = 0;
	                }
    	        }
                else
	                pc += 4;
            }
            else
            {
	            find_target = 0;
	            arg.target = -1;
            }

            switch (arg.type)
            {
	            case ARGTYPE_NUM:
	            case ARGTYPE_NEG_NUM:
	            case ARGTYPE_IND_NUM:
                    int num = 0;
	                byte c;

	                do
                    {
		                c = prgm.text[pc++];
		                num = (num << 7) | (c & 127);
	                }
                    while ((c & 128) == 0);

	                if (arg.type == ARGTYPE_NEG_NUM)
                    {
		                arg.type = ARGTYPE_NUM;
		                num = -num;
	                }

	                arg.val.num = num;
	                break;

                case ARGTYPE_STK:
	            case ARGTYPE_IND_STK:
	                arg.val.stk = (char)prgm.text[pc++];
	                break;

	            case ARGTYPE_COMMAND:
	                arg.val.cmd = prgm.text[pc++];
	                break;

	            case ARGTYPE_LCLBL:
	                arg.val.lclbl = (char)prgm.text[pc++];
	                break;

	            case ARGTYPE_STR:
	            case ARGTYPE_IND_STR:
	                arg.length = prgm.text[pc++];
	                for (i = 0; i < arg.length; i++)
		                arg.val.text[i] = (char)prgm.text[pc++];
	                break;

                case ARGTYPE_DOUBLE:
                    byte[] b = new byte[sizeof_Phloat];
                    Array.Copy(prgm.text, pc, b, 0, sizeof_Phloat);
                    pc += Phloat.from_byte_array(arg.val_d, b);
                    break;
            }

            if (command == CMD_NUMBER && arg.type != ARGTYPE_DOUBLE)
            {
	            // argtype is ARGTYPE_NUM; convert to phloat
	            arg.val_d = arg.val.num;
	            arg.type = ARGTYPE_DOUBLE;
            }
    
            if (find_target != 0)
            {
	            target_pc = find_local_label(arg);

	            arg.target = target_pc;

	            for (i = 5; i >= 2; i--)
                {
	                prgm.text[orig_pc + i] = (byte)target_pc;
	                target_pc >>= 8;
	            }

	            prgm.lclbl_invalid = 0;
            }
        }

        public static int find_local_label(arg_struct arg)
        {
            int orig_pc = pc;
            int search_pc;
            int wrapped = 0;

            prgm_struct prgm = prgms[current_prgm];

            if (orig_pc == -1)
	            orig_pc = 0;

            search_pc = orig_pc;

            while (wrapped == 0 || search_pc < orig_pc)
            {
	            int command, argtype;

	            if (search_pc >= prgm.size - 2)
                {
	                if (orig_pc == 0)
		               break;

	                search_pc = 0;
	                wrapped = 1;
	            }

	            command = prgm.text[search_pc];
	            argtype = prgm.text[search_pc + 1];

	            command |= (argtype & 240) << 4;
	            argtype &= 15;

	            if (command == CMD_LBL && argtype == arg.type)
                {
	                if (argtype == ARGTYPE_NUM)
                    {
		                int num = 0;
		                byte c;

		                int pos = search_pc + 2;

		                do
                        {
		                    c = prgm.text[pos++];
		                    num = (num << 7) | (c & 127);
		                }
                        while ((c & 128) == 0);

        		        if (num == arg.val.num)
		                    return search_pc;
	                }
                    else
                    {
		                char lclbl = (char)prgm.text[search_pc + 2];

		                if (lclbl == arg.val.lclbl)
		                    return search_pc;
	                }
	        }

	        search_pc += get_command_length(current_prgm, search_pc);
            }

            return -2;
        }

        public static int mvar_prgms_exist()
        {
            for (int i = 0; i < labels_count; i++)
                if (label_has_mvar(i) != 0)
                    return 1;

            return 0;
        }

        public static int label_has_mvar(int lblindex)
        {
            int saved_prgm;
            int pc;
            int command = 0;
            arg_struct arg = new arg_struct();

            if (labels[lblindex].length == 0)
                return 0;

            saved_prgm = current_prgm;
            
            current_prgm = labels[lblindex].prgm;
            pc = labels[lblindex].pc;

            pc += get_command_length(current_prgm, pc);
            get_next_command(ref pc, ref command, arg, 0);
            
            current_prgm = saved_prgm;
            
            return (command == CMD_MVAR) ? 1 : 0;
        }

        public static int pc2line(int pc)
        {
            if (pc == -1)
                return 0;
            else
                return pc_line_convert(pc, 1);
        }

        public static int line2pc(int line)
        {
            if (line == 0)
                return -1;
            else
                return pc_line_convert(line, 0);
        }

        public static int pc_line_convert(int loc, int loc_is_pc)
        {
            int pc = 0;
            int line = 1;

            prgm_struct prgm = prgms[current_prgm];

            while (true)
            {
                if (loc_is_pc != 0)
                {
                    if (pc >= loc)
                        return line;
                }
                else
                {
                    if (line >= loc)
                        return pc;
                }

                if (prgm.text[pc] == CMD_END)
                    return (loc_is_pc != 0) ? line : pc;

                pc += get_command_length(current_prgm, pc);
                line++;
            }
        }

        public static void store_command(int pc, int command, arg_struct arg)
        {
            byte[] buf = new byte[100];
            int bufptr = 0;
            int i;
            int pos;
            prgm_struct prgm = prgms[current_prgm];

            // We should never be called with pc = -1, but just to be safe...
            if (pc == -1)
	            pc = 0;

            if (arg.type == ARGTYPE_NUM && arg.val.num < 0)
            {
	            arg.type = ARGTYPE_NEG_NUM;
	            arg.val.num = -arg.val.num;
            }
            else if (command == CMD_NUMBER)
            {
	            // arg.type is always ARGTYPE_DOUBLE for CMD_NUMBER, but for storage
	            // efficiency, we handle integers specially and store them as
	            // ARGTYPE_NUM or ARGTYPE_NEG_NUM instead.
	            //
	            int n = Phloat.to_int(arg.val_d);

	            if (n == arg.val_d && n != 0x80000000)      // TODO: Why compare against -0 ???
                {
	                if (n >= 0)
                    {
		                arg.val.num = n;
		                arg.type = ARGTYPE_NUM;
	                }
                    else
                    {
		                arg.val.num = -n;
		                arg.type = ARGTYPE_NEG_NUM;
	                }
	            }
            }
            else if (arg.type == ARGTYPE_LBLINDEX)
            {
	            int li = arg.val.num;
	            arg.length = labels[li].length;

	            for (i = 0; i < arg.length; i++)
    	            arg.val.text[i] = labels[li].name[i];

	            arg.type = ARGTYPE_STR;
            }

            buf[bufptr++] = (byte)(command & 255);
            buf[bufptr++] = (byte)(arg.type | ((command & ~255) >> 4));

            // If the program is nonempty, it must already contain an END,
            // since that's the very first thing that gets stored in any new
            // program. In this case, we need to split the program.
            //
            if (command == CMD_END && prgm.size > 0)
            {
	            if (prgms_count == prgms_capacity)
                {
                    prgms_capacity += 10;
	                prgm_struct[] new_prgms = new prgm_struct[prgms_capacity];

                    for (i = 0; i <= current_prgm; i++)
		                new_prgms[i] = prgms[i];

	                for (i = current_prgm + 1; i < prgms_count; i++)
		                new_prgms[i + 1] = prgms[i];

	                prgms = new_prgms;
	                prgm = prgms[current_prgm];
	            }
                else
                {
                    for (i = prgms_count - 1; i > current_prgm; i--)
                    {
                        prgms[i + 1] = prgms[i];
                        prgms[i] = new prgm_struct();
                    }
	            }
    
                prgms_count++;
	            
                prgm_struct new_prgm = prgms[current_prgm + 1];

                new_prgm.size = prgm.size - pc;
	            new_prgm.capacity = (new_prgm.size + 511) & ~511;
                new_prgm.text = new byte[new_prgm.capacity];

	            for (i = pc; i < prgm.size; i++)
	                new_prgm.text[i - pc] = prgm.text[i];

	            current_prgm++;

	            // Truncate the previously 'current' program and append an END.
	            // No need to check the size against the capacity and grow the
	            // program; since it contained an END before, it still has the
	            // capacity for one now;
	            //
	            prgm.size = pc;
	            prgm.text[prgm.size++] = CMD_END;
	            prgm.text[prgm.size++] = ARGTYPE_NONE;

	            if ((flags.trace_print != 0) || (flags.normal_print != 0))
	                print_program_line(current_prgm - 1, pc);

	            rebuild_label_table();
	            invalidate_lclbls(current_prgm);
	            invalidate_lclbls(current_prgm - 1);
	            clear_all_rtns();
	            
                draw_varmenu();
	            
                return;
            }

            if ((command == CMD_GTO || command == CMD_XEQ) && (arg.type == ARGTYPE_NUM || arg.type == ARGTYPE_LCLBL))
	            for (i = 0; i < 4; i++)
	                buf[bufptr++] = 255;

            switch (arg.type)
            {
                case ARGTYPE_NUM:
	            case ARGTYPE_NEG_NUM:
	            case ARGTYPE_IND_NUM:
	                int num = arg.val.num;

                    char[] tmpbuf = new char[5];
	                int tmplen = 0;

	                while (num > 127)
                    {
		                tmpbuf[tmplen++] = (char)(num & 127);
		                num >>= 7;
	                }

	                tmpbuf[tmplen++] = (char)num;
	                tmpbuf[0] |= (char)128;

	                while (--tmplen >= 0)
		                buf[bufptr++] = (byte)tmpbuf[tmplen];
	                break;

                case ARGTYPE_STK:
	            case ARGTYPE_IND_STK:
	                buf[bufptr++] = (byte)arg.val.stk;
	                break;

	            case ARGTYPE_STR:
	            case ARGTYPE_IND_STR:
	                buf[bufptr++] = arg.length;
	                
                    for (i = 0; i < arg.length; i++)
		                buf[bufptr++] = (byte)arg.val.text[i];
                    break;

                case ARGTYPE_LCLBL:
	                buf[bufptr++] = (byte)arg.val.lclbl;
    	            break;
	
                case ARGTYPE_DOUBLE:
                    byte[] b = Phloat.to_byte_array(arg.val_d);
	                for (i = 0; i < sizeof_Phloat; i++)
		                buf[bufptr++] = b[i];
                    break;
            }

            if (bufptr + prgm.size > prgm.capacity)
            {
	            prgm.capacity += 512;
	            byte[] newtext = new byte[prgm.capacity];
	            
                for (pos = 0; pos < pc; pos++)
	                newtext[pos] = prgm.text[pos];
	            
                for (pos = pc; pos < prgm.size; pos++)
	                newtext[pos + bufptr] = prgm.text[pos];
	            
                prgm.text = newtext;
            }
            else
            {
	            for (pos = prgm.size - 1; pos >= pc; pos--)
	                prgm.text[pos + bufptr] = prgm.text[pos];
            }
            
            for (pos = 0; pos < bufptr; pos++)
	            prgm.text[pc + pos] = buf[pos];

            prgm.size += bufptr;

            if (command != CMD_END && ((flags.trace_print != 0) || (flags.normal_print != 0)))
	            print_program_line(current_prgm, pc);
    
            if (command == CMD_END || (command == CMD_LBL && arg.type == ARGTYPE_STR))
	            rebuild_label_table();
            else
    	        update_label_table(current_prgm, pc, bufptr);

            invalidate_lclbls(current_prgm);
            clear_all_rtns();
            
            draw_varmenu();
        }

        public static void rebuild_label_table()
        {
            // TO DO -- this is *not* efficient; inserting and deleting ENDs and
            // global LBLs should not cause every single program to get rescanned!
            // But, I don't feel like dealing with that at the moment, so just
            // this ugly brute force approach for now.
            //
            int prgm_index;
            int pc;
            labels_count = 0;
            
            for (prgm_index = 0; prgm_index < prgms_count; prgm_index++)
            {
                prgm_struct prgm = prgms[prgm_index];
                pc = 0;
                
                while (pc < prgm.size)
                {
                    int command = prgm.text[pc];
                    int argtype = prgm.text[pc + 1];

                    command |= (argtype & 240) << 4;
                    argtype &= 15;

                    if (command == CMD_END || (command == CMD_LBL && argtype == ARGTYPE_STR))
                    {
                        label_struct newlabel = new label_struct();

                        if (labels_count == labels_capacity)
                        {
                            labels_capacity += 50;

                            label_struct[] newlabels = new label_struct[labels_capacity];
                            for (int x = 0; x < newlabels.Length; x++)
                                newlabels[x] = new label_struct();

                            for (int i = 0; i < labels_count; i++)
                                newlabels[i] = labels[i];

                            labels = newlabels;
                        }

                        newlabel = labels[labels_count++];

                        if (command == CMD_END)
                            newlabel.length = 0;
                        else
                        {
                            newlabel.length = prgm.text[pc + 2];

                            for (int i = 0; i < newlabel.length; i++)
                                newlabel.name[i] = (char)(prgm.text[pc + 3 + i]);
                        }

                        newlabel.prgm = prgm_index;
                        newlabel.pc = pc;
                    }

                    pc += get_command_length(prgm_index, pc);
                }
            }
        }

        public static void update_label_table(int prgm, int pc, int inserted)
        {
            for (int i = 0; i < labels_count; i++)
            {
                if (labels[i].prgm > prgm)
                    return;

                if (labels[i].prgm == prgm && labels[i].pc >= pc)
                    labels[i].pc += inserted;
            }
        }

        public static void invalidate_lclbls(int prgm_index)
        {
            prgm_struct prgm = prgms[prgm_index];

            if (prgm.lclbl_invalid == 0)
            {
                int pc2 = 0;

                while (pc2 < prgm.size)
                {
                    int command = prgm.text[pc2];
                    int argtype = prgm.text[pc2 + 1];

                    command |= (argtype & 240) << 4;
                    argtype &= 15;
                    
                    if ((command == CMD_GTO || command == CMD_XEQ) && (argtype == ARGTYPE_NUM || argtype == ARGTYPE_LCLBL))
                    {
                        // A dest_pc value of -1 signals 'unknown',
                        // -2 means 'nonexistent', and anything else is
                        // the pc where the destination label is found.
                        //
                        for (int pos = pc2 + 2; pos < pc2 + 6; pos++)
                            prgm.text[pos] = 255;
                    }

                    pc2 += get_command_length(prgm_index, pc2);
                }

                prgm.lclbl_invalid = 1;
            }
        }

        public static void clear_all_rtns()
        {
            rtn_sp = 0;
        }

        public static bool solve_active()
        {
            for (int i = 0; i < rtn_sp; i++)
                if (rtn_prgm[i] == -2)
                    return true;
            
            return false;
        }

        public static void unwind_stack_until_solve()
        {
            while (rtn_prgm[--rtn_sp] != -2) ;
        }

        public static void store_command_after(ref int pc, int command, arg_struct arg)
        {
            if (pc == -1)
                pc = 0;
            else if (prgms[current_prgm].text[pc] != CMD_END)
                pc += get_command_length(current_prgm, pc);

            store_command(pc, command, arg);
        }

        public static void delete_command(int pc)
        {
            prgm_struct prgm = prgms[current_prgm];

            int command = prgm.text[pc];
            int argtype = prgm.text[pc + 1];
            int length = get_command_length(current_prgm, pc);
            int pos;

            command |= (argtype & 240) << 4;
            argtype &= 15;

            if (command == CMD_END)
            {
    	        int newsize;

	            if (current_prgm == prgms_count - 1)
	                // Don't allow deletion of last program's END.
	                return;

	            prgm_struct nextprgm = prgms[current_prgm + 1];

                prgm.size -= 2;
	            newsize = prgm.size + nextprgm.size;

	            if (newsize > prgm.capacity)
                {
	                int newcapacity = (newsize + 511) & ~511;

                    byte[] newtext = new byte[newcapacity];
	                
                    for (pos = 0; pos < prgm.size; pos++)
		                newtext[pos] = prgm.text[pos];

                    prgm.text = newtext;
	                prgm.capacity = newcapacity;
	            }

    	        for (pos = 0; pos < nextprgm.size; pos++)
	                prgm.text[prgm.size++] = nextprgm.text[pos];

	            nextprgm.text = null;

                for (pos = current_prgm + 1; pos < prgms_count - 1; pos++)
                {
                    prgms[pos] = prgms[pos + 1];
                    prgms[pos + 1] = new prgm_struct();
                }

    	        prgms_count--;
	
                rebuild_label_table();
	            invalidate_lclbls(current_prgm);
	            clear_all_rtns();
	        
                draw_varmenu();
	            return;
            }

            for (pos = pc; pos < prgm.size - length; pos++)
	            prgm.text[pos] = prgm.text[pos + length];

            prgm.size -= length;
            
            if (command == CMD_LBL && argtype == ARGTYPE_STR)
	            rebuild_label_table();
            else
	            update_label_table(current_prgm, pc, -length);

            invalidate_lclbls(current_prgm);
            clear_all_rtns();
            
            draw_varmenu();
        }

        public static int push_rtn_addr(int prgm, int pc)
        {
            int err = ERR_NONE;
            
            if (rtn_sp == MAX_RTNS)
            {
                if (rtn_prgm[0] == -2 || rtn_prgm[0] == -3)
                    err = ERR_SOLVE_INTEG_RTN_LOST;
                
                for (int i = 0; i < MAX_RTNS - 1; i++)
                {
                    rtn_prgm[i] = rtn_prgm[i + 1];
                    rtn_pc[i] = rtn_pc[i + 1];
                }
                
                rtn_sp--;
            }

            rtn_prgm[rtn_sp] = prgm;
            rtn_pc[rtn_sp] = pc;
            rtn_sp++;
            
            return err;
        }

        public static int clear_prgm(arg_struct arg)
        {
            int prgm_index = 0;
            int i, j;
            
            if (arg.type == ARGTYPE_LBLINDEX)
	            prgm_index = labels[arg.val.num].prgm;
            else if (arg.type == ARGTYPE_STR)
            {
	            if (arg.length == 0)
                {
	                if (current_prgm < 0 || current_prgm >= prgms_count)
		                return ERR_INTERNAL_ERROR;
	                
                    prgm_index = current_prgm;
	            }
                else
                {
                    for (i = labels_count - 1; i >= 0; i--)
                        if (string_equals(arg.val.text, arg.length, labels[i].name, labels[i].length) != 0)
		                    goto found;
	                
                    return ERR_LABEL_NOT_FOUND;
	                
                found:
	                prgm_index = labels[i].prgm;
	            }
            }

            clear_all_rtns();
            
            if (prgm_index == current_prgm)
	            pc = -1;
            else if (current_prgm > prgm_index)
	            current_prgm--;

            prgms[prgm_index].text = null;

            for (i = prgm_index; i < prgms_count - 1; i++)
            {
                prgms[i] = prgms[i + 1];
                prgms[i + 1] = new prgm_struct();
            }
            
            prgms_count--;
            i = j = 0;
            
            while (j < labels_count)
            {
                if (j > i)
                {
                    labels[i] = labels[j];
                    labels[j] = new label_struct();
                }
	        
                j++;
	            
                if (labels[i].prgm > prgm_index)
                {
	                labels[i].prgm--;
	                i++;
	            }
                else if (labels[i].prgm < prgm_index)
	                i++;
            }

            labels_count = i;
            
            if (prgms_count == 0 || prgm_index == prgms_count)
            {
	            int saved_prgm = current_prgm;
	            int saved_pc = pc;
	            
                goto_dot_dot();
	            
                current_prgm = saved_prgm;
	            pc = saved_pc;
            }

            update_catalog();
            return ERR_NONE;
        }

        public static void goto_dot_dot()
        {
            int command = 0;
            arg_struct arg = new arg_struct();
            
            if (prgms_count != 0)
            {
                // Check if last program is empty
                pc = 0;
                current_prgm = prgms_count - 1;
                
                get_next_command(ref pc, ref command, arg, 0);
                
                if (command == CMD_END)
                {
                    pc = -1;
                    return;
                }
            }

            if (prgms_count == prgms_capacity)
            {
                int i;
                prgms_capacity += 10;

                prgm_struct[] newprgms = new prgm_struct[prgms_capacity];
                for (int x = 0; x < newprgms.Length; x++)
                    newprgms[x] = new prgm_struct();
                
                // TO DO - handle memory allocation failure
                for (i = 0; i < prgms_count; i++)
                    newprgms[i] = prgms[i];

                prgms = newprgms;
            }

            current_prgm = prgms_count++;
            prgms[current_prgm].capacity = 0;
            prgms[current_prgm].size = 0;
            prgms[current_prgm].lclbl_invalid = 1;
            prgms[current_prgm].text = null;
            command = CMD_END;
            arg.type = ARGTYPE_NONE;

            store_command(0, command, arg);
            pc = -1;
        }

        public static void clear_prgm_lines(int count)
        {
            int frompc, deleted, i, j;
            
            if (pc == -1)
                pc = 0;
            
            frompc = pc;
            
            while (count > 0)
            {
                int command = 0;
                arg_struct arg = new arg_struct();
            
                get_next_command(ref pc, ref command, arg, 0);
                
                if (command == CMD_END)
                {
                    pc -= 2;
                    break;
                }

                count--;
            }

            deleted = pc - frompc;

            for (i = pc; i < prgms[current_prgm].size; i++)
                prgms[current_prgm].text[i - deleted] = prgms[current_prgm].text[i];

            prgms[current_prgm].size -= deleted;
            pc = frompc;

            i = j = 0;
            
            while (j < labels_count)
            {
                if (j > i)
                {
                    labels[i] = labels[j];
                    labels[j] = new label_struct();
                }
            
                j++;
                
                if (labels[i].prgm == current_prgm)
                {
                    if (labels[i].pc < frompc)
                        i++;
                    else if (labels[i].pc >= frompc + deleted)
                    {
                        labels[i].pc -= deleted;
                        i++;
                    }
                }
                else
                    i++;
            }
            labels_count = i;

            invalidate_lclbls(current_prgm);
            clear_all_rtns();
        }

        public static void pop_rtn_addr(ref int prgm, ref int pc)
        {
            if (rtn_sp == 0)
            {
                prgm = -1;
                pc = -1;
            }
            else
            {
                rtn_sp--;
                prgm = rtn_prgm[rtn_sp];
                pc = rtn_pc[rtn_sp];
            }
        }

        public static bool integ_active()
        {
            int i;

            for (i = 0; i < rtn_sp; i++)
                if (rtn_prgm[i] == -3)
                    return true;
            
            return false;
        }
    }
}
