﻿//
// Porting and modifications made by Rayzzor Studios, 2011
//

/***************************************************************************** 
* Free42 -- an HP-42S calculator simulator 
* Copyright (C) 2004-2010 Thomas Okken 
* 
* This program is free software; you can redistribute it and/or modify 
* it under the terms of the GNU General Public License, version 2, 
* as published by the Free Software Foundation. 
* 
* This program is distributed in the hope that it will be useful, 
* but WITHOUT ANY WARRANTY; without even the implied warranty of 
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
* GNU General Public License for more details. 
* 
* You should have received a copy of the GNU General Public License 
* along with this program; if not, see http://www.gnu.org/licenses/. 
*****************************************************************************/

using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;

using System.Windows.Threading;
using System.Windows.Navigation;
using System.Threading;
using System.Windows.Media.Imaging;
using System.IO;
using System.IO.IsolatedStorage;
using System.Collections;
using System.Collections.Generic;
using Microsoft.Phone.Marketplace;
using Microsoft.Xna.Framework.GamerServices;

namespace WP7_42s
{
    public partial class App
    {
        //-------------------------------------------------------------------------------------
        //
        // Shell items
        //
        //-------------------------------------------------------------------------------------

        public static int ann_updown = 0;
        public static int ann_shift = 0;
        public static int ann_print = 0;
        public static int ann_run = 0;
        public static int ann_battery = 0;
        public static int ann_g = 0;
        public static int ann_rad = 0;


        // shell_annunciators()
        // Invoked by the emulator core to change the state of the display
        // annunciators (up/down, shift, print, run, battery, (g)rad).
        // Every parameter can have values 0 (turn off), 1 (turn on), or -1 (leave
        // unchanged).
        // The battery annunciator is missing from the list; this is the only one of
        // the lot that the emulator core does not actually have any control over, and
        // so the shell is expected to handle that one by itself.
        //
        public static void shell_annunciators(int updn, int shf, int prt, int run, int g, int rad)
        {
            if (updn != -1 && ann_updown != updn)
                ann_updown = updn;

            if (shf != -1 && ann_shift != shf)
                ann_shift = shf;

            if (prt != -1 && ann_print != prt)
                ann_print = prt;

            if (run != -1 && ann_run != run)
                ann_run = run;

            if (g != -1 && ann_g != g)
                ann_g = g;

            if (rad != -1 && ann_rad != rad)
                ann_rad = rad;

            update_annunciators();
        }


        public static void update_annunciators()
        {
            Disp._UpDown.Opacity = ann_updown;
            Main.UpDownHighlight.Opacity = prefKeyHighlightEn ? ann_updown : 0;

            Disp._Shift.Opacity = ann_shift;
            Main.ShiftHighlight.Opacity = prefKeyHighlightEn ? ann_shift : 0;
            
            Disp._Print.Opacity = ann_print;
            Disp._Run.Opacity = ann_run;

            if (ann_g != 0)
            {
                Disp._GRAD.Text = "GRAD";
                Disp._GRAD.Opacity = ann_g;
            }
            else
            {
                Disp._GRAD.Text = "RAD";
                Disp._GRAD.Opacity = ann_rad;
            }
        }

        public static void shell_keydown()
        {
            if (calc_off && ckey != KEY_EXIT)
                return;

            KillTimer(ref timer);

            if (timer3 != null && ckey != 28)   // 28 = SHIFT
            {
                KillTimer(ref timer3);

                core_timeout3(0);
            }

            shell_beeper(11);      // Do keyclick if enabled

            int repeat = 0;
            
            running = core_keydown(ckey, ref enqueued, ref repeat);

            if (running == 0)
            {
                if (repeat != 0)
                    SetTimer(ref timer, repeat == 1 ? 1000 : 500, repeater);
                else if (enqueued == 0)
                    SetTimer(ref timer, 250, timeout1);
            }
        }

        public static void shell_keyup()
        {
            if (calc_off && ckey != KEY_EXIT)
                return;

            if (calc_off)
            {
                turn_on();
                return;
            }

            ckey = 0;

            KillTimer(ref timer);
            
            if (enqueued == 0)
                running = core_keyup();
        }

        public static void SetTimer(ref DispatcherTimer t, int timeout, EventHandler callback)
        {
            t = new DispatcherTimer();
            t.Interval = new TimeSpan(0, 0, 0, 0, timeout);
            t.Tick += callback;

            t.Start();
        }

        public static void KillTimer(ref DispatcherTimer t)
        {
            if (t != null)
                t.Stop();

            t = null;
        }

        public static void repeater(Object sender, EventArgs e)
        {
	        KillTimer(ref timer);

	        int repeat = core_repeat();

	        if (repeat != 0)
		        SetTimer(ref timer, repeat == 1 ? 200 : 100, repeater);
	        else
		        SetTimer(ref timer, 250, timeout1);
        }

        public static void timeout1(Object sender, EventArgs e)
        {
	        KillTimer(ref timer);

	        if (ckey != 0)
            {
		        core_keytimeout1();
		    
                SetTimer(ref timer, 1750, timeout2);
	        }
        }

        public static void timeout2(Object sender, EventArgs e)
        {
	        KillTimer(ref timer);

	        if (ckey != 0)
		        core_keytimeout2();
        }

        public static void timeout3(Object sender, EventArgs e)
        {
	        KillTimer(ref timer3);

	        bool keep_running = core_timeout3(1);

	        if (keep_running)
            {
		        running = 1;
		        // Post dummy message to get the message loop moving again
	            // PostMessage(hMainWnd, WM_USER, 0, 0);                    // Win32'ism.  Nothing to do here for Silverlight. Timer set up to emulate.
	        }
        }

        // Callback to ask the shell to call core_timeout3() after the given number of
        // milliseconds. If there are keystroke events during that time, the timeout is
        // cancelled. (Pressing 'shift' does not cancel the timeout.)
        // This function supports the delay after SHOW, MEM, and shift-VARMENU.
        //
        public static void shell_request_timeout3(int delay)
        {
            if (timer3 != null)
                KillTimer(ref timer3);

            SetTimer(ref timer3, delay, timeout3);
        }

        public static int shell_wants_cpu()
        {
            TimeSpan exe_time = DateTime.Now - prgm_start_time;
            if (exe_time > prgm_time_slice)
                return 1;
            else
                return 0;
        }

        public static void shell_powerdown()
        {
            if (OFF_METHOD == off_methods.exit)
                throw new QuitException();              // Exit application?  This may disqualify Marketplace acceptance.
            else                                        // Alternative is to emulate a powerdown without exiting...
                turn_off();                             // ...which is what we do here.
        }

        public static void turn_off()
        {
            calc_off = true;
            Disp.Disable();
            Main.show_on();
        }

        public static void turn_on()
        {
            Main.hide_on();
            Disp.Enable();
            calc_off = false;
        }

        public static uint shell_get_mem()
        {
            return (uint)System.GC.GetTotalMemory(false);         // TODO: Any way to figure out available memory?
        }

        public static void shell_delay(int duration)
        {
            Thread.Sleep(duration);
        }

        public static int shell_low_battery()
        {
            return 0;                                                   // TODO: Implement
        }

        public static void shell_print(char[] text, int length, byte[] bits, int bytesperline, int x, int y, int width, int height)
        {
            char[] tmp = new char[bits.Length];

            for (int i = 0; i < bits.Length; i++)
                tmp[i] = (char)bits[i];

            shell_print(text, length, tmp, bytesperline, x, y, width, height);
        }

        public static string hp2ascii(char[] src, int srclen)
        {
            string esc = "";
            string result = "";
            char c;
            int s, d = 0;
            
            for (s = 0; s < srclen; s++)
            {
	            c = src[s];

                if (c >= 130 && c != 138)
	                c &= (char)127;
	        
                switch ((int)c)
                {
	                // NOTE: this code performs the following 12 translations
	                // that are not ASCII, but seem to be widely accepted --
	                // that is, they looked OK when I tried them in several
	                // fonts in Windows and Linux, and in Memo Pad on the Palm:
	                //
	                //   0: 247 (0367) divide
	                //   1: 215 (0327) multiply
	                //   8: 191 (0277) upside-down question mark
	                //  17: 181 (0265) lowercase mu
	                //  18: 163 (0243) sterling
	                //  19: 176 (0260) degree
	                //  20: 197 (0305) Aring
	                //  21: 209 (0321) Ntilde
	                //  22: 196 (0304) Aumlaut
	                //  25: 198 (0306) AE
	                //  28: 214 (0326) Oumlaut
	                //  29: 220 (0334) Uumlaut
	                //
	                // Two additional candidates are these:
	                //
	                //  26: 133 (0205) ellipsis
	                //  31: 149 (0225) bullet
	                //
	                // I'm not using those last two because support for them is not
	                // as good: they exist in Times New Roman and Adobe Courier
	                // (tested on Windows and Linux, respectively) and on the Palm,
	                // but are missing from Windows Fixedsys (the default Notepad
	                // font, so that is a concern!) and X11 lucidatypewriter and
	                // fixed.
	                // Note that 133 and 149 are both in the 128-159 range, that
	                // is, the Ctrl+Meta range, which is unused in many fonts.
	                // Eventually, I should probably support several translation
	                // modes: raw, pure ASCII (only emit codes 32-126 and 10),
	                // non-pure as below, and more aggressive non-pure (using the
	                // ellipsis and fatdot codes, and maybe others). Then again,
	                // maybe not. :-)
	                //
	                case  0:   esc = "\x00f7"; break;
	                case  1:   esc = "\x00d7"; break;
	                case  2:   esc = "\\sqrt"; break;
	                case  3:   esc = "\\int"; break;
	                case  4:   esc = "\\gray1"; break;
	                case  5:   esc = "\\Sigma"; break;
	                case  6:   esc = ">"; break;
	                case  7:   esc = "\\pi"; break;
	                case  8:   esc = "\x00bf"; break;
	                case  9:   esc = "<="; break;
	                case 11:   esc = ">="; break;
	                case 12:   esc = "!="; break;
	                case 13:   esc = "\\r"; break;
	                case 14:   esc = "v"; break;
	                case 15:   esc = "->"; break;
	                case 16:   esc = "<-"; break;
	                case 17:   esc = "\x00b5"; break;
	                case 18:   esc = "\x00a3"; break;
	                case 19:   esc = "\x00b0"; break;
	                case 20:   esc = "\x00c5"; break;
	                case 21:   esc = "\x00d1"; break;
	                case 22:   esc = "\x00c4"; break;
	                case 23:   esc = "\\angle"; break;
	                case 24:   esc = "E"; break;
	                case 25:   esc = "\x00c6"; break;
	                case 26:   esc = "..."; break;
	                case 27:   esc = "\\esc"; break;
	                case 28:   esc = "\x00d6"; break;
	                case 29:   esc = "\x00dc"; break;
	                case 30:   esc = "\\gray2"; break;
	                case 31:   esc = "\\bullet"; break;
	                case '\\': esc = "\\\\"; break;
	                case 127:  esc = "|-"; break;
	                case 128:  esc = ":"; break;
	                case 129:  esc = "y"; break;
	                case 138:  esc = "\\LF"; break;
	                default:   result += c; continue;
	            }

                if (esc.Length > 0)
                    result += esc;
            }

            return result;
        }

        public static void shell_print(char[] text, int length, char[] bits, int bytesperline, int x, int y, int width, int height)
        {
            Color bg = (Color)Main.Resources["PhoneBackgroundColor"];
            Color fg = (Color)Main.Resources["PhoneForegroundColor"];

            WriteableBitmap newBitmap = new WriteableBitmap(143 * printPixelMul, height * printPixelMul);

            for (int row = 0; row < height; row++)
            {
		        for (int col = 0; col < 143; col++)
                {
			        bool bit;
			    
                    if (col < width)
                    {
				        char c = bits[(y + row) * bytesperline + ((x + col) >> 3)];
				        bit = (c & (1 << ((x + col) & 7))) != 0;
			        }
                    else
				        bit = false;

                    Color pixel = bit ? fg : bg;

                    long pixelVal = pixel.A;
                    pixelVal = pixelVal << 8 | pixel.R;
                    pixelVal = pixelVal << 8 | pixel.G;
                    pixelVal = pixelVal << 8 | pixel.B;

                    for (int rowMul = 0; rowMul < printPixelMul; rowMul++)
                        for (int colMul = 0; colMul < printPixelMul; colMul++)
                            newBitmap.Pixels[(row * printPixelMul + rowMul) * (143 * printPixelMul) + (col * printPixelMul + colMul)] = (int)pixelVal;
                }
            }

            // Add current print addition
            Image i = new Image();
            i.Source = newBitmap;
            Main.printStack.Children.Add(i);

            // Limit printout length on screen
            while (Main.printStack.Children.Count > print_entry_limit)
                Main.printStack.Children.RemoveAt(0);

            // Ensure child layout is complete and scroll to bottom using a very large offset
            Main.printStack.UpdateLayout();
            Main.printScroll.ScrollToVerticalOffset(1000000);

            
            

            // Print textfile & graphics

            IsolatedStorageFile isf = IsolatedStorageFile.GetUserStoreForApplication();
            IsolatedStorageFileStream isfs = null;

            try
            {
                isfs = new IsolatedStorageFileStream(App.PRINT_FILE, FileMode.Append, isf);
                StreamWriter ws = new StreamWriter(isfs);

                string line = hp2ascii(text, length);

                if (line == "<lcd>")
                {
                    string imgFileName = string.Format("{0}{1:0000}.jpg", IMG_FILE, imgNum++);
                    line = string.Format("<lcd: {0}>", imgFileName);
                    IsolatedStorageFileStream img = new IsolatedStorageFileStream(imgFileName, FileMode.Create, isf);
                    
                    newBitmap.SaveJpeg(img, newBitmap.PixelWidth, newBitmap.PixelHeight, 0, imgQuality);
                    img.Flush();
                    img.Close();
                }

                ws.WriteLine(line);
            }
            catch (Exception e)
            {
//                MessageBox.Show("Write error: " + e.ToString());
            }

            if (isfs != null)
            {
                isfs.Flush();
                isfs.Close();
            }
        }

        public static double shell_random_seed()
        {
            return ((double)rand.Next()) / (Int16.MaxValue + 1.0);      // TODO: Int16 or Int32 MaxValue???
        }

        public static uint shell_milliseconds()
        {
            return (uint)Environment.TickCount;
        }

        public static void shell_beeper(int t)
        {
            prgm_block = true;

            if (t == 11)
                tones.Insert(0, t);
            else
                tones.Add(t);
        }

        public static void core_init(int version)
        {
            // If trial mode, don't load state
            if (trialMode)
                hard_reset();
            else
                if (!load_state(version))
                    hard_reset();

            repaint_display();

            shell_annunciators(
                mode_updown ? 1 : 0,
                mode_shift ? 1 : 0,
                0,          // Print
                mode_running ? 1 : 0,
                flags.grad,
                (flags.rad != 0) || (flags.grad != 0) ? 1 : 0
                );
        }

        public static void shell_get_time_date(ref uint time, ref uint date, ref int weekday)
        {
            DateTime st = DateTime.Now;

            time = (uint)(st.Hour * 1000000 + st.Minute * 10000 + st.Second * 100 + st.Millisecond / 10);
            date = (uint)(st.Year * 10000 + st.Month * 100 + st.Day);
            weekday = (int)st.DayOfWeek;
        }
    }
}
