﻿// Copyright Rayzzor Studios 2011 
// 
// This file is part of 42S.
// 
// 42S is free software: you can redistribute it and/or modify 
// it under the terms of the GNU General Public License as published by 
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version. 
// 
// 42S is distributed in the hope that it will be useful, 
// but WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the 
// GNU General Public License for more details. 
// 
// You should have received a copy of the GNU General Public License 
// along with 42S. If not, see <http://www.gnu.org/licenses/>. 

using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.ComponentModel;

namespace WP7_42s
{
	public partial class VolSliderControl : UserControl
	{
		public VolSliderControl()
		{
			// Required to initialize variables
			InitializeComponent();

            // Initialize dependency properties
            Slider.Value = Value;
            tbHeader.Text = Header;
		}


        #region Control Events

        //
        // Value changed event
        //

        public delegate void ValueHandler(object sender, EventArgs e);
        public event ValueHandler ValueChanged;

        void RaiseValueChanged(RoutedEventArgs e)
        {
            if (ValueChanged != null)
                ValueChanged(this, e);
        }

        #endregion


        #region Dependency Properties

        //
        // Heading text dependency property
        //

        [Category("Common Properties")]
        [Description("Control title")]
        public string Header
        {
            get { return (string)GetValue(HeaderProperty); }
            set { SetValue(HeaderProperty, value); }
        }

        public static readonly DependencyProperty HeaderProperty = DependencyProperty.Register(
            "Header",
            typeof(string),
            typeof(VolSliderControl),
            new PropertyMetadata("Header", HeaderPropertyChanged));

        private static void HeaderPropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
                ((VolSliderControl)d).tbHeader.Text = (string)e.NewValue;
        }


        //
        // Control value dependency property
        //

        [Category("Common Properties")]
        [Description("Set initial slider value in the range of 0 to 1")]
        public double Value
        {
            get { return (double)GetValue(ValueProperty); }
            set { SetValue(ValueProperty, value); }
        }

        public static readonly DependencyProperty ValueProperty = DependencyProperty.Register(
            "Value",
            typeof(double),
            typeof(VolSliderControl),
            new PropertyMetadata(1.0, ValuePropertyChanged));

        private static void ValuePropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e)
        {
            if ((double)e.NewValue > 1 || (double)e.NewValue < 0)
                d.SetValue(e.Property, e.OldValue);
            else
                ((VolSliderControl)d).Slider.Value = (double)e.NewValue;
        }

        #endregion


        private void Slider_ValueChanged(object sender, System.Windows.RoutedPropertyChangedEventArgs<double> e)
		{
            // Restrict slider to small change increments
            Slider.Value = Math.Round(Slider.Value / Slider.SmallChange) * Slider.SmallChange;

            // Update dependency property
            Value = Slider.Value;

            // Raise control's event
            RaiseValueChanged(e);
		}
	}
}